Package: TreeDist
Title: Calculate and Map Distances Between Phylogenetic Trees
Version: 2.14.1
Authors@R: c(person("Martin R.", "Smith",
                    email = "martin.smith@durham.ac.uk", 
                    role = c("aut", "cre", "cph", "prg"), 
                    comment = c(ORCID = "0000-0001-5660-1727")),
             person("Roy", "Jonker", email = "roy_jonker@magiclogic.com",
                    role = c("prg", "cph"), comment = "LAP algorithm"),
             person("Yong", "Yang", email = "yongyanglink@gmail.com",
                    role = c("ctb", "cph"), comment = "LAP algorithm"),
             person("Yi", "Cao", role = c("ctb", "cph"), comment = "LAP algorithm"),
             person("Neil", "Kaye", role = c("cph"), comment = "Mercator image")
            )
License: GPL (>= 3)
Description: Implements measures of tree similarity, including 
  information-based generalized Robinson-Foulds distances
  (Phylogenetic Information Distance, Clustering Information Distance,
  Matching Split Information Distance; Smith 2020)
  <doi:10.1093/bioinformatics/btaa614>; 
  Jaccard-Robinson-Foulds distances (Bocker et al. 2013)
  <doi:10.1007/978-3-642-40453-5_13>, 
  including the Nye et al. (2006) metric <doi:10.1093/bioinformatics/bti720>;
  the Matching Split Distance (Bogdanowicz & Giaro 2012)
  <doi:10.1109/TCBB.2011.48>;
  the Hierarchical Mutual Information (Perotti et al. 2015)
  <doi:10.1103/PhysRevE.92.062825>;
  Maximum Agreement Subtree distances;
  the Kendall-Colijn (2016) distance <doi:10.1093/molbev/msw124>, and the
  Nearest Neighbour Interchange (NNI) distance, approximated per Li et al. 
  (1996) <doi:10.1007/3-540-61332-3_168>.
  Includes tools for visualizing mappings of tree space (Smith 2022)
  <doi:10.1093/sysbio/syab100>,
  for identifying islands of trees (Silva and Wilkinson 2021)
  <doi:10.1093/sysbio/syab015>,
  for calculating the median of sets of trees,
  and for computing the information content of trees and splits.
Copyright: Jonker-Volgenant Linear Assignment Problem implementation by
        Roy Jonker modified by Yong Yang and Yi Cao.
URL: https://ms609.github.io/TreeDist/,
        https://github.com/ms609/TreeDist/
BugReports: https://github.com/ms609/TreeDist/issues/
Additional_repositories: https://ms609.github.io/packages/
Depends: R (>= 4.0), stats,
Imports: ape (>= 5.0), cli (>= 3.0), colorspace, Rdpack (>= 0.7),
        shiny, shinyjs, TreeTools (>= 2.1.0),
Suggests: bookdown, cluster, ggplot2, hypervolume, kdensity, knitr,
        MASS, parallel, phangorn (>= 2.2.1), plotly, PlotTools,
        protoclust, Quartet, readxl, rmarkdown, Rcpp (>= 1.0.8), rgl,
        Rogue, spelling, TBRDist, testthat (>= 3.0), Ternary (>=
        1.1.2), TreeDistData (> 0.1.0), TreeSearch (>= 1.4.0), Umatrix,
        vdiffr (>= 1.0.0), withr,
LinkingTo: Rcpp, TreeTools (>= 2.1.0),
RdMacros: Rdpack
VignetteBuilder: knitr
Config/Needs/app/optional: uwot
Config/Needs/check: rcmdcheck
Config/Needs/coverage: covr
Config/Needs/memcheck: pkgdown, testthat
Config/Needs/metadata: codemetar
Config/Needs/revdeps: revdepcheck
Config/Needs/website: openssl, pkgdown, remotes, shinylive
Config/roxygen2/version: 8.0.0
Config/testthat/parallel: false
Config/testthat/edition: 3
SystemRequirements: C++17, pandoc-citeproc
ByteCompile: true
Encoding: UTF-8
Language: en-GB
X-schema.org-keywords: phylogenetics, tree-distance
NeedsCompilation: yes
Packaged: 2026-06-10 10:46:37 UTC; pjjg18
Author: Martin R. Smith [aut, cre, cph, prg] (ORCID:
    <https://orcid.org/0000-0001-5660-1727>),
  Roy Jonker [prg, cph] (LAP algorithm),
  Yong Yang [ctb, cph] (LAP algorithm),
  Yi Cao [ctb, cph] (LAP algorithm),
  Neil Kaye [cph] (Mercator image)
Maintainer: Martin R. Smith <martin.smith@durham.ac.uk>
Repository: CRAN
Date/Publication: 2026-06-10 12:10:08 UTC

# Generated by roxygen2: do not edit by hand

S3method(ClusteringEntropy,"NULL")
S3method(ClusteringEntropy,Splits)
S3method(ClusteringEntropy,list)
S3method(ClusteringEntropy,multiPhylo)
S3method(ClusteringEntropy,phylo)
S3method(ClusteringInfo,"NULL")
S3method(ClusteringInfo,Splits)
S3method(ClusteringInfo,list)
S3method(ClusteringInfo,multiPhylo)
S3method(ClusteringInfo,phylo)
S3method(DistanceFromMedian,dist)
S3method(DistanceFromMedian,numeric)
S3method(KCDiameter,list)
S3method(KCDiameter,multiPhylo)
S3method(KCDiameter,numeric)
S3method(KCDiameter,phylo)
S3method(KMeansPP,dist)
S3method(KMeansPP,matrix)
S3method(KMeansPP,numeric)
S3method(MeanMSTEdge,dist)
S3method(MeanMSTEdge,numeric)
S3method(MeanNN,dist)
S3method(MeanNN,numeric)
S3method(NNIDiameter,list)
S3method(NNIDiameter,multiPhylo)
S3method(NNIDiameter,numeric)
S3method(NNIDiameter,phylo)
S3method(NTip,HPart)
S3method(RenumberTips,HPart)
S3method(SPRDist,list)
S3method(SPRDist,multiPhylo)
S3method(SPRDist,phylo)
S3method(SplitwiseInfo,"NULL")
S3method(SplitwiseInfo,Splits)
S3method(SplitwiseInfo,list)
S3method(SplitwiseInfo,multiPhylo)
S3method(SplitwiseInfo,phylo)
S3method(as.HPart,HPart)
S3method(as.HPart,default)
S3method(as.HPart,list)
S3method(as.HPart,phylo)
S3method(as.phylo,HPart)
S3method(clone,HPart)
S3method(median,multiPhylo)
S3method(plot,HPart)
S3method(print,HPart)
export(.ThreeDPlotServer)
export(.TreeDistance)
export(AHMI)
export(AllSplitPairings)
export(CalculateTreeDistance)
export(ClusteringEntropy)
export(ClusteringInfo)
export(ClusteringInfoDist)
export(ClusteringInfoDistance)
export(CompareAll)
export(ConsensusInfo)
export(DifferentPhylogeneticInfo)
export(DisplayMatching)
export(DistFromMed)
export(DistanceFromMedian)
export(EHMI)
export(Entropy)
export(ExpectedVariation)
export(GeneralizedRF)
export(GetParallel)
export(HH)
export(HMI)
export(HierarchicalMutualInfo)
export(HierarchicalMutualInformation)
export(InfoRobinsonFoulds)
export(InfoRobinsonFouldsSplits)
export(Islands)
export(JaccardRobinsonFoulds)
export(JaccardSplitSimilarity)
export(KCDiameter)
export(KCVector)
export(KMeansPP)
export(KendallColijn)
export(LAPJV)
export(LnTreesConsistentWithTwoSplits)
export(Log2TreesConsistentWithTwoSplits)
export(MASTInfo)
export(MASTSize)
export(MCITree)
export(MSTSegments)
export(MapTrees)
export(MappingQuality)
export(MatchingSplitDistance)
export(MatchingSplitDistanceSplits)
export(MatchingSplitInfo)
export(MatchingSplitInfoDistance)
export(MatchingSplitInfoSplits)
export(MeanCentDist)
export(MeanCentroidDist)
export(MeanCentroidDistance)
export(MeanMSTEdge)
export(MeanNN)
export(MeilaMutualInformation)
export(MeilaVariationOfInformation)
export(MutualClusteringInfo)
export(MutualClusteringInfoSplits)
export(MutualClusteringInformation)
export(NNIDiameter)
export(NNIDist)
export(NormalizeInfo)
export(Ntropy)
export(NyeSimilarity)
export(NyeSplitSimilarity)
export(PathDist)
export(PathVector)
export(PhylogeneticInfoDistance)
export(Plot3)
export(PlotMatching)
export(Project)
export(ProjectionQuality)
export(ReduceTrees)
export(ReportMatching)
export(RobinsonFoulds)
export(RobinsonFouldsInfo)
export(RobinsonFouldsMatching)
export(RobinsonFouldsSplits)
export(SPRDist)
export(SelfHMI)
export(SetParallel)
export(SharedPhylogeneticInfo)
export(SharedPhylogeneticInfoSplits)
export(SpectralClustering)
export(SpectralEigens)
export(SplitDifferentInformation)
export(SplitEntropy)
export(SplitSharedInformation)
export(SplitVector)
export(SplitsCompatible)
export(SplitwiseInfo)
export(StartParallel)
export(StopParallel)
export(StrainCol)
export(SumOfRanges)
export(SumOfVariances)
export(SumOfVars)
export(TransferConsensus)
export(TransferDist)
export(TransferDistSplits)
export(TransferDistance)
export(TreeDistPlot)
export(TreeDistance)
export(TreesConsistentWithTwoSplits)
export(VisualiseMatching)
export(VisualizeMatching)
export(as.HPart)
export(clone)
export(entropy_int)
export(is.HPart)
importFrom(Rdpack,reprompt)
importFrom(TreeTools,AllAncestors)
importFrom(TreeTools,Consensus)
importFrom(TreeTools,DropTip)
importFrom(TreeTools,FirstMatchingSplit)
importFrom(TreeTools,KeepTip)
importFrom(TreeTools,KeepTipPostorder)
importFrom(TreeTools,LnRooted.int)
importFrom(TreeTools,LnTreesMatchingSplit)
importFrom(TreeTools,Log2Rooted.int)
importFrom(TreeTools,Log2TreesMatchingSplit)
importFrom(TreeTools,Log2Unrooted)
importFrom(TreeTools,Log2Unrooted.int)
importFrom(TreeTools,MSTEdges)
importFrom(TreeTools,MSTLength)
importFrom(TreeTools,MatchStrings)
importFrom(TreeTools,NRooted)
importFrom(TreeTools,NSplits)
importFrom(TreeTools,NTip)
importFrom(TreeTools,PairwiseDistances)
importFrom(TreeTools,PectinateTree)
importFrom(TreeTools,Postorder)
importFrom(TreeTools,PostorderOrder)
importFrom(TreeTools,Preorder)
importFrom(TreeTools,RenumberTips)
importFrom(TreeTools,RootOnNode)
importFrom(TreeTools,RootTree)
importFrom(TreeTools,SplitFrequency)
importFrom(TreeTools,SplitInformation)
importFrom(TreeTools,SplitsInBinaryTree)
importFrom(TreeTools,StarTree)
importFrom(TreeTools,TipLabels)
importFrom(TreeTools,TipsInSplits)
importFrom(TreeTools,TreeIsRooted)
importFrom(TreeTools,TreesMatchingSplit)
importFrom(TreeTools,as.ClusterTable)
importFrom(TreeTools,as.Splits)
importFrom(TreeTools,edge_to_splits)
importFrom(ape,Nnode.phylo)
importFrom(ape,as.phylo)
importFrom(ape,drop.tip)
importFrom(ape,edgelabels)
importFrom(ape,nodelabels)
importFrom(ape,plot.phylo)
importFrom(cli,cli_alert_danger)
importFrom(cli,cli_alert_info)
importFrom(cli,cli_alert_success)
importFrom(cli,cli_progress_along)
importFrom(cli,cli_progress_bar)
importFrom(cli,cli_progress_update)
importFrom(colorspace,qualitative_hcl)
importFrom(colorspace,sequential_hcl)
importFrom(grDevices,colorRamp)
importFrom(grDevices,hcl.colors)
importFrom(grDevices,rgb)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(graphics,rect)
importFrom(graphics,segments)
importFrom(parallel,makeCluster)
importFrom(parallel,parCapply)
importFrom(parallel,parLapply)
importFrom(parallel,stopCluster)
importFrom(shiny,runApp)
importFrom(shinyjs,useShinyjs)
importFrom(stats,as.dist)
importFrom(stats,dist)
importFrom(stats,kmeans)
importFrom(stats,median)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,var)
importFrom(utils,combn)
importFrom(utils,install.packages)
useDynLib(TreeDist, .registration = TRUE)

## ----generate-trees-----------------------------------------------------------
# Load trees
library("TreeTools", quietly = TRUE)
batch1 <- as.phylo(1:60, 8) # Generate 60 similar trees 
batch2 <- as.phylo(seq(200, 800, length.out = 30), 8) # A separate batch of 30 trees
styles <- c(1, 2) # Select plotting colours / symbols
treeStyle <- rep(styles, c(length(batch1), length(batch2)))

# Calculate distances
library("TreeDist")
distances <- ClusteringInfoDistance(c(batch1, batch2))

# Construct over-simple 2D PCoA mapping
mapping <- cmdscale(distances, k = 2)

## ----plot-mapping, fig.align = "center"---------------------------------------
# Plot mapping
par(mar = rep(0, 4))
plot(mapping,
     asp = 1, # Preserve aspect ratio - do not distort distances
     ann = FALSE, axes = FALSE, # Don't label axes: dimensions are meaningless
     col = treeStyle, # Colour
     pch = treeStyle # Plotting symbol
     )
legend("left", c("Batch 1", "Batch 2"), col = styles, pch = styles)

## ----compare-dist-from-median, fig.align = "center"---------------------------
# Calculate median trees
median1 <- median(batch1)
median2 <- median(batch2)

# Compute distance from each tree to the median of its batch
dist1 <- ClusteringInfoDist(batch1, median1)
dist2 <- ClusteringInfoDist(batch2, median2)

# Set resolution of histogram
nBreaks <- 10
breaks <- seq(0, max(dist1, dist2), length.out = nBreaks)

# Plot first distance set
hist(dist1, col = "#00000022", breaks = breaks,
     main = "Distance from median of batch",
     xlab = "Clustering information distance",
     ylim = c(0, 25) # Omit this line to infer Y axis limit from first batch.
     )

# Add second distance set
hist(dist2, col = "#ff000022", breaks = breaks, add = TRUE)

# Add legend
legend("topleft", c("Batch 1", "Batch 2"),
       fill = c("#00000022",  "#ff000022"))

## ----rogue-detection----------------------------------------------------------
# Create tree set with a rogue taxon
batch3 <- AddTipEverywhere(as.phylo(7, 7), "t8")

# Set up plotting area
par(mfrow = c(2, 2), mar = rep(0.4, 4))

# Plot naive strict consensus
plot(consensus(batch1, p = 1))
plot(consensus(batch3, p = 1))

if (requireNamespace("Rogue", quietly = TRUE)) {
  cons1 <- ConsensusWithout(batch1, p = 1,
                            Rogue::QuickRogue(batch1, p = 1)[-1, "taxon"])
  cons3 <- ConsensusWithout(batch3, p = 1,
                            Rogue::QuickRogue(batch3, p = 1)[-1, "taxon"])
  
  # The information content of each tree gives a measure of its resolution,
  # accounting for omitted rogue leaves
  SplitwiseInfo(cons1) # 8.5 bits
  SplitwiseInfo(cons3) # 15.1 bits: higher resolution indicates that these
  # trees are more similar, notwithstanding rogue taxa.
  
  # Plot the trees
  plot(cons1)
  plot(cons3)
} else {
  message("The package 'Rogue' is required to run this example.")
}

## ----init, message=FALSE, warning=FALSE, echo = FALSE-------------------------
library('ape')
library('TreeDist')
tree1 <- read.tree(text='((A, B), ((C, (D, E)), (F, (G, (H, I)))));')
tree2 <- read.tree(text='((A, B), ((C, D, (E, I)), (F, (G, H))));')
AtoJ <-  read.tree(text='(((((A, B), C), D), E), (F, (G, (H, (I, J)))));')
swapAJ <-  read.tree(text='(((((J, B), C), D), E), (F, (G, (H, (I, A)))));')

## ----fig.width=6, fig.align='center'------------------------------------------
VisualizeMatching(SharedPhylogeneticInfo, tree1, tree2, 
                  Plot = TreeDistPlot, matchZeros = FALSE)
SharedPhylogeneticInfo(tree1, tree2)

## ----total-phylo-info---------------------------------------------------------
SplitwiseInfo(tree1)

## ----fig.width=6, fig.align='center'------------------------------------------
VisualizeMatching(SharedPhylogeneticInfo, AtoJ, swapAJ,
                  Plot = TreeDistPlot, matchZeros = FALSE, prune = c(5, 18))

## ----fig.width=6, fig.align='center'------------------------------------------
VisualizeMatching(MutualClusteringInfo, AtoJ, swapAJ,
                  Plot = TreeDistPlot, matchZeros = FALSE, prune = c(5, 18))
MutualClusteringInfo(AtoJ, swapAJ)

## ----fig.width=6, fig.align='center'------------------------------------------
VisualizeMatching(MutualClusteringInfo, tree1, tree2, 
                  Plot = TreeDistPlot, matchZeros = FALSE)

## ----total-mci----------------------------------------------------------------
ClusteringEntropy(tree1)

## ----fig.width=6, fig.align='center'------------------------------------------
VisualizeMatching(NyeSimilarity, tree1, tree2, 
                  Plot = TreeDistPlot, matchZeros = FALSE)
NyeSimilarity(tree1, tree2, normalize = FALSE)

## ----fig.width=6, fig.align='center'------------------------------------------
JaccardRobinsonFoulds(tree1, tree2, k = 1)
VisualizeMatching(JaccardRobinsonFoulds, tree1, tree2,
                  Plot = TreeDistPlot, matchZeros = FALSE)

JRF2 <- function(...) JaccardRobinsonFoulds(k = 2, ...)
JRF2(tree1, tree2)
VisualizeMatching(JRF2, tree1, tree2,
                  Plot = TreeDistPlot, matchZeros = FALSE)

## ----fig.width=6, fig.height = 4,fig.align = 'center', echo = FALSE-----------
oldPar <- par(mar = c(4, 4, 0, 0))
x <- seq(1, 20, by = 0.1)
cbPalette8 <- c("#000000", "#E69F00", "#56B4E9", "#009E73", "#F0E442",
                "#0072B2", "#D55E00", "#CC79A7") # dput(Ternary::cbPalette8)
plot(0, xlim = range(x), ylim = c(3.5, 9.5),
     xlab = expression(italic('k')),
     ylab = 'Jaccard\u2013Robinson\u2013Foulds distance',
     pch = 3, type = 'n')

abline(h = RobinsonFoulds(tree1, tree2),
       lty = 'dashed', col = cbPalette8[6])
abline(h = NyeSimilarity(tree1, tree2, similarity = FALSE),
       lty = 'dotted', col = cbPalette8[8])

lines(x = x, y = vapply(x, JaccardRobinsonFoulds, tree1 = tree1, tree2 = tree2,
                        allowConflict = TRUE, 0), col = cbPalette8[5])
lines(x = x, y = vapply(x, JaccardRobinsonFoulds, tree1 = tree1, tree2 = tree2,
                        0), col = cbPalette8[4])

legend('right', c('Robinson\u2013Foulds', 'JRF, no conflict', 'JRF, conflict ok', 'Nye et al.'),
       lty = c('dashed', 'solid', 'solid', 'dotted'), 
       col = cbPalette8[c(6, 4, 5, 8)], bty = 'n')
par(oldPar)

## ----fig.width=6, fig.align='center'------------------------------------------
MatchingSplitDistance(tree1, tree2)
VisualizeMatching(MatchingSplitDistance, tree1, tree2,
                  Plot = TreeDistPlot, matchZeros = FALSE)

## ----fig.width=6, fig.align='center'------------------------------------------
MatchingSplitDistance(read.tree(text='((a, b, c, d, e, f), (g, h, i, j));'),
                      read.tree(text='((a, b, c, d, e, i, j), (g, h, f));'))

## -----------------------------------------------------------------------------
TreeTools::SplitInformation(5, 2)

## ----fig.width=6, fig.align='center'------------------------------------------
MatchingSplitInfoDistance(tree1, tree2)
VisualizeMatching(MatchingSplitInfoDistance, tree1, tree2,
                  Plot = TreeDistPlot, matchZeros = FALSE)

## -----------------------------------------------------------------------------
library("TreeTools", quietly = TRUE)
library("TreeDist")
treesMatchingSplit <- c(
  AB.CDEF = TreesMatchingSplit(2, 4),
  ABC.DEF = TreesMatchingSplit(3, 3)
)
treesMatchingSplit

proportionMatchingSplit <- treesMatchingSplit / NUnrooted(6)
proportionMatchingSplit

splitInformation <- -log2(proportionMatchingSplit)
splitInformation

treesMatchingBoth <- TreesConsistentWithTwoSplits(6, 2, 3)
combinedInformation <- -log2(treesMatchingBoth / NUnrooted(6))

sharedInformation <- sum(splitInformation) - combinedInformation
sharedInformation

# Or more concisely:
SplitSharedInformation(n = 6, 2, 3)

## ----mackay-8-1, echo=FALSE, fig.width=4, out.width='50%', fig.height=3, fig.align='center'----
library('TreeDist')
H <- function(inBracket) {
  expression(paste(italic('H'), plain('('), italic(inBracket), plain(')')))
}
oldPar <- par(mar = c(3.1, 0.1, 0, 0.1))
joint <- Entropy(c(1, 2, 1, 2) / 6)
plot(0, type = 'n', xlim = c(0, joint), ylim = c(5, 0), axes = FALSE)
axis(1, at = c(0, 0.5, 1, 1.5, round(joint, 2)))
mtext('Entropy / bits', 1, 2)
rect(joint - 1, 3.1, 1, 3.9, col = "#56B4E9")
rect(0, 0.1, joint - 1, 0.9, col = "#F0E442", border = NA)
rect(1, 1.1, joint, 1.9, col = "#F0E442", border = NA)
rect(joint - 1, 1.1, 1, 1.9, col = "#56B4E9", border = NA)
rect(joint - 1, 0.1, 1, 0.9, col = "#56B4E9", border = NA)
text(1, 3.5, pos=4,
     expression(paste(italic(I), plain('('), italic('A;B'), plain(')'))))



rect(0, 2.1, joint, 2.9)
text(joint / 2, 2.5, 
     expression(paste(italic('H'), plain('('), italic('A, B'), plain(')'))))

rect(0, 0.1, 1, 0.9)
text(0.5, 0.5, 
     expression(paste(italic('H'), plain('('), italic(A), plain(')'))))


rect(joint - 1, 1.1, joint - 0, 1.9)
text(joint - 0.5, 1.5, 
     expression(paste(italic('H'), plain('('), italic(B), plain(')'))))


rect(0, 4.1, joint - 1, 4.9, col = "#F0E442")
rect(1, 4.1, joint, 4.9,     col = "#F0E442")
text(joint / 2, 4.5, 
     expression(paste(italic('H'['D']), plain('('), italic('A, B'), plain(')'))))
par(oldPar)

## ----init, echo=FALSE, warning=FALSE, message=FALSE---------------------------
library('ape')
library('TreeDist')
standardMargin <- c(0.4, 0.4, 0.8, 0.4)
cbPalette8 <- c("#000000", "#E69F00", "#56B4E9", "#009E73", "#F0E442",
                "#0072B2", "#D55E00", "#CC79A7")
TwoTreePlot <- function() par(mfrow = c(1, 2), mar = standardMargin, cex = 0.9)

Plot <- function(tree, tree2 = NULL, highlight = character(0),
                  prune = if (is.null(tree2)) integer(0) 
                          else list(integer(0), integer(0))
                  ) {
  ABCDEFGHIJK <- LETTERS[1:11]
  if (is.null(tree2)) {
    if(!all(tree$tip.label %in% 1:11)) {
      tree$tip.label <- match(tree$tip.label, ABCDEFGHIJK)
    }
    
    highlightTip <- match(highlight, ABCDEFGHIJK)
    TreeDistPlot(tree, bold = highlightTip, prune = prune,
                 graft = which(tree$edge[, 2] == 
                                 match(highlightTip, tree$tip.label)))
  } else {
    par(mfrow = c(1, 2), mar = rep(0.4, 4))
    Plot(tree, highlight = highlight, prune = prune[[1]])
    Plot(tree2, highlight = highlight, prune = prune[[2]])
  }
}

## ----echo=FALSE, fig.width = 6, fig.align = 'center'--------------------------
startPar <- TwoTreePlot()
palette <- colorspace::qualitative_hcl(5, c=42, l=88)
TreeDistPlot(balancedTree <- 
               read.tree(text="(((1, 2), (3, 4)), ((5, 6), (7, 8)));"))
edgelabels(1:5, c(1, 2, 5, 9, 12), adj = c(1.5, 0.5), bg = palette[1]) # Splits
edgelabels(1:6, edge = c(1, 2, 5, 8, 9, 12), bg = palette[2],
           adj = c(-0.5, 0.5)) # Internal edges
nodelabels(1:7, bg = palette[3], adj=c(1, 0.5)) # Nodes
nodelabels(1:6, node = 10:15, bg = palette[4], adj=c(-1, 0.5)) # Clades
legend('bottomleft', col=palette[1:2], pch=15, bty='n',
       c('Splits (5)', 'Internal edges (6)'))
legend('bottomright', col=palette[3:4], pch=15, bty='n',
       c('Nodes (7)', 'Clades (6)'))

TreeDistPlot(read.tree(text="((1, 2, 3, 4), (5, 6, 7, 8));"))
edgelabels(1, 1, adj = c(1.5, 0.5), bg = palette[1])
edgelabels(1:2, edge = c(1, 6), adj = c(-0.5, 0.5), bg = palette[2])
nodelabels(1:3, bg = palette[3], adj = c(1, 0.5))
nodelabels(1:2, node=10:11, bg = palette[4], adj = c(-1, 0.5))


## ----two-trees, echo=FALSE, fig.height = 3, fig.width=6, fig.align = 'center'----
TwoTreePlot()
TreeDistPlot(balancedTree, edge.color = cbPalette8[3])
ape::edgelabels(1:5, c(1, 2, 5, 9, 12), bg = palette[1], adj=c(0.5, -0.25))
ape::edgelabels(c('4:4', rep('2:6', 4)), c(1, 2, 5, 9, 12),
                adj=c(0.5, 1.25), frame = 'n')
ape::edgelabels(c(5.53, rep(3.46, 4)), c(1, 2, 5, 9, 12),
                adj=c(0.5, 2.75), frame = 'n')
legend('topleft', 'Balanced tree', bty='n', inset=c(-0.05, -0.03))

TreeDistPlot(caterpillarTree <- 
               ape::read.tree(text="(1, (2, (3, (4, (5, (6, (7, 8)))))));"),
             edge.color = cbPalette8[2])
ape::edgelabels(1:5, bg = palette[1], c(4, 6, 8, 10, 12), adj = c(0.5, -0.25))
ape::edgelabels(c('2|6', '3|5', '4|4', '3|5', '2|6'), edge = c(4, 6, 8, 10, 12),
                frame = 'n', adj = c(0.5, 1.25))
ape::edgelabels(c(3.46, 5.04, 5.53, 5.04, 3.46), edge = c(4, 6, 8, 10, 12),
                frame = 'n', adj = c(0.5, 2.75))
legend('topleft', 'Asymmetric tree', bty = 'n', inset = c(-0.05, -0.03))

## ----ic-of-splits, display='asis', echo=FALSE, warnings=FALSE-----------------
splitSmall <- 2:4
splitLarge <- 8L - splitSmall

rootedTrees <- c(
  '1' = 1,
  '2' = 1 * 1,
  '3' = 3 * 1,
  '4' = 5 * 3 * 1,
  '5' = 7 * 5 * 3 * 1,
  '6' = 9 * 7 * 5 * 3 * 1
)

matchingTrees <- rootedTrees[splitSmall] * rootedTrees[splitLarge]
names(matchingTrees) <- paste0("Split size: ", splitSmall, ':', splitLarge)

matchingP <- matchingTrees / (11 * 9 * 7 * 5 * 3 * 1 * 1)

ic <- -log2(matchingP)

knitr::kable(cbind(
      'Matching trees' = paste(matchingTrees, "/ 10 395"),
      '_P_(Match in random tree)' = signif(matchingP, 3),
      'Phylogenetic information content' = paste(signif(ic, 3), 'bits')))

## ----all-8-tip-trees, echo=FALSE, cache=TRUE, fig.width=4, fig.height=4, fig.align='center'----
calculate <- FALSE
if (calculate) {
  # Generate all eight-leaf trees
  all8 <- as.phylo(seq_len(NUnrooted(8)), tipLabels = 1:8)
  inBalanced <- RobinsonFoulds(all8, balancedTree, similarity = TRUE) / 2
  inCaterpillar <- RobinsonFoulds(all8, caterpillarTree, similarity = TRUE) / 2
} else {
  inBalanced <- rep(0:5, c(7088, 2708, 512, 76, 10, 1))
  inCaterpillar <- rep(0:5, c(8162, 1808, 350, 64, 10, 1))
}

par(cex = 0.8)
sch <- hist(inCaterpillar + 0.7, breaks = 0:18 / 3 - (1/6),
            main = "8-leaf trees with N common splits", cex.main = 1,
            xlim = c(0, 6), axes = FALSE,
            xlab = "Splits in common", ylab = "Number of trees")
sbh <- hist(inBalanced + 0.4, breaks = 0:18 / 3 - (1/6), plot = FALSE)
plot(sch, col = paste0(cbPalette8[2], "44"), add = TRUE)
plot(sbh, col = paste0(cbPalette8[3], "44"), add = TRUE)
text(1/6, 100, paste0("Balanced: ", sum(inBalanced == 0)), 
     pos = 4, srt = 90, cex = 0.7)
text(0.5, 100, paste0("Asymmetric: ", sum(inCaterpillar == 0)), 
     pos = 4, srt = 90, cex = 0.7)


legend("topright", pch = 22,
       pt.cex = 2, col = "black",
       pt.bg = paste0(cbPalette8[2:3], "44"), bty = "n",
       c("Asymmetric", "Balanced"))

axis(1, at = 0:5 + 0.5, labels = 0:5)
axis(2)

## ----fig.height = 3, fig.width=6, fig.align = 'center'------------------------
tree1 <- ape::read.tree(text='(1, (2, (3, (4, (5, (6, (7, 8)))))));')
tree2 <- ape::read.tree(text='(1, (2, (3, (4, (5, (7, (6, 8)))))));')
tree3 <- ape::read.tree(text='(1, (2, (3, (5, (4, (6, (7, 8)))))));')

VisualizeMatching(InfoRobinsonFoulds, tree1, tree2, 
                  Plot = TreeDistPlot, prune = 12)

## ----fig.height = 3, fig.width=6, fig.align = 'center'------------------------
VisualizeMatching(InfoRobinsonFoulds, tree1, tree3, 
                  Plot = TreeDistPlot, prune = 8)

## ----fig.height = 3, fig.width=6, fig.align = 'center'------------------------
tree1 <- ape::read.tree(text='(1, (2, (3, (4, (5, (6, (7, 8)))))));')
tree2 <- ape::read.tree(text='(8, (1, (2, (3, (4, (5, (6, 7)))))));')

VisualizeMatching(RobinsonFouldsMatching, tree1, tree2, Plot = TreeDistPlot)

## ----fig.height = 3, fig.width=6, fig.align = 'center'------------------------
tree1 <- ape::read.tree(text='((A, B), ((C, (D, E)), (F, (G, (H, I)))));')
tree2 <- ape::read.tree(text='((A, B), ((C, D, (E, I)), (F, (G, H))));')

Plot(tree1, tree2, highlight = 'I', prune = list(8, integer(0)))

## ----fig.height = 3, fig.width=6, fig.align = 'center'------------------------
TwoTreePlot()
VisualizeMatching(RobinsonFouldsMatching, tree1, tree2)

## ----fig.height = 3, fig.width=6, fig.align = 'center'------------------------
TwoTreePlot()
VisualizeMatching(RobinsonFouldsMatching,
                  drop.tip(tree1, 'I'),
                  drop.tip(tree2, 'I'))

## ----fig.height = 3, fig.width=6, fig.align = 'center'------------------------
TwoTreePlot()
VisualizeMatching(SharedPhylogeneticInfo, tree1, tree2)

## -----------------------------------------------------------------------------
summary(TreeTools::as.Splits(tree1, LETTERS[1:9]))
summary(TreeTools::as.Splits(tree2, LETTERS[1:9]))

## -----------------------------------------------------------------------------
attributes(SharedPhylogeneticInfo(tree1, tree2, reportMatching = TRUE))

## ----echo=FALSE---------------------------------------------------------------
par(startPar)

# options("TreeDist.logging" = TRUE) # Log function entry and exit
logging <- isTRUE(getOption("TreeDist.logging"))

if (logging) {
  .DateTime <- function(time = Sys.time()) { # Copy, because not exported
    format(time, "%Y-%m-%d %T")
  }
  LogMsg <- function(...) {
    message(.DateTime(), ": ", ...)
  }
} else {
  LogMsg <- function(...) {}
}

suppressPackageStartupMessages({
  library("shiny",     exclude = "runExample")
  library("shinyjs",   exclude = "runExample")
  library("TreeTools")
  library("TreeDist")
  # ape is a hard dep of TreeTools/TreeDist; library() is needed because
  # TreeTools imports as.phylo via importFrom(ape, as.phylo) for internal use
  # only (no re-export), so the generic is not on the search path after
  # library(TreeTools) alone.
  library("ape")
})

.Install <- if (isTRUE(getOption("webr.initialized"))) webr::install else install.packages

# Reliable webR/WebAssembly detector. getOption("webr.initialized") is not
# dependable at server runtime, but the wasm32 architecture always is.
.runningWebR <- identical(R.version[["arch"]], "wasm32")

# Shinylive/webR resolves neither the "app" nor "session" bindCache scope (nor
# an explicit cachem object), so every bindCache(cache = "app") reactive in the
# server would error with
#   resolve_cache_object: `cache` must be "app", "session", or a cache object
# Caching is only a speed optimisation, so under webR run the reactive uncached
# (webR is compute-bound anyway). Server deployments keep Shiny's app cache.
bindCache <- function(x, ...) {
  if (.runningWebR) x else shiny::bindCache(x, ...)
}

for (.pkg in c("cluster", "protoclust", "MASS", "Quartet", "readxl")) {
  if (!requireNamespace(.pkg, quietly = TRUE)) .Install(.pkg)
}

# Indirect reference to hide from shinylive's static package scanner
.uwot_pkg <- paste0("uw", "ot")

# uwot is optional (Suggests) and absent from the webR binary repo. Under webR,
# requireNamespace() for a missing package emits an install warning and can
# return a non-logical, so `!requireNamespace(...)` errored ("invalid argument
# type"). Resolve availability once, defensively, and reuse it.
.uwotAvailable <- isTRUE(suppressWarnings(tryCatch(
  requireNamespace(.uwot_pkg, quietly = TRUE),
  error = function(e) FALSE
)))

# Allow large files to be submitted
options(shiny.maxRequestSize = 100 * 1024^2)

palettes <- list(
  "#91aaa7",
  c("#2a7df6", "#b9e17b"),
  c("#be83ae", "#2ea7af", "#fbcdcf"),
  c("#72c5a9", "#b7c5ff", "#dbdabb", "#a28bac"),
  c("#59c4c0", "#ea9a9a", "#7998a6", "#e9d7a9", "#9c9379"),
  c("#e8b889", "#67c6eb", "#e5d5c2", "#938fba", "#64b69a", "#779c7b"),
  c("#c4808f", "#5ba08f", "#f0a693", "#ccd7fe", "#cdb87e", "#c6aae2", "#d2dad8"),
  c("#d0847f", "#63a5d7", "#d7b981", "#5a9bbb", "#9bb67e", "#dea6d5", "#91967e", "#ca7f96"),
  c("#8b93a8", "#ccb97e", "#8e9dd7", "#57a384", "#dbb1e7", "#2da7af", "#d68986", "#75d2f9", "#e4d1f0"),
  c("#dfcf92", "#40b3cb", "#b88a61", "#ecb2e0", "#d6dbbc", "#a28bae", "#edcfeb", "#7498ab", "#b187a0", "#8f939c"),
  c("#a98f70", "#7be5e2", "#d295c0", "#9ae2bd", "#d3b7f1", "#eca88d", "#8993cd", "#ffc7bb", "#8695a8", "#b3e1df", "#b6878a"),
  c("#eaa9d3", "#7ac09b", "#fdaca8", "#8ce7e4", "#eed69b", "#70a4d9", "#e8d6ba", "#589bbb", "#959672", "#d0dbd1", "#9b9282", "#d9d9c6"),
  c("#7498ab", "#e5bd8a", "#7ed8ff", "#de8f8e", "#46bac6", "#ffc0d3", "#57ae96", "#f7cddd", "#50a098", "#b58a6d", "#add49d", "#a18da1", "#cedad9"),
  c("#8097a4", "#d0dea9", "#a78cc3", "#aee4bf", "#bb82a8", "#5dc9c6", "#b88690", "#26a3b9", "#ad8e6f", "#a4e2ef", "#869a65", "#efcfdd", "#60a089", "#9e927b"),
  c("#b9aae5", "#bbd69c", "#e2adde", "#77a777", "#f8abc8", "#8ee7ce", "#f2a1a5", "#81bdf1", "#f2bb91", "#b8dcfe", "#aeb276", "#f2cdef", "#e8d6b2", "#8d92b0", "#b7878d"),
  c("#c3d79b", "#b28cc0", "#64a985", "#e3a7d4", "#2ea2aa", "#e69892", "#85c6f9", "#fbd1a0", "#7696be", "#89996c", "#ddcdff", "#719d89", "#f5cde6", "#b6e0da", "#e8d4cd", "#b5ddfa"),
  c("#a98d83", "#84e1ff", "#bb8964", "#46b1d1", "#ffbfa5", "#6199c0", "#bbcb8f", "#bf82ab", "#85ddc4", "#eea0ba", "#c1d8ff", "#c3818b", "#c5c6ff", "#999388", "#e8cbff", "#ffb5b6", "#d2dad7"),
  c("#faccde", "#60a987", "#c6abe4", "#6f9e77", "#c48093", "#a5e5d3", "#cc8f6f", "#499fae", "#d9dca6", "#7796b8", "#bee1ba", "#b4daff", "#919583", "#e2d3e9", "#47a19b", "#ebd4bc", "#7c9993", "#a9e3e0"),
  c("#739e6e", "#ffbfd9", "#43b6bb", "#e8ad88", "#5e9bce", "#c2af75", "#a8e0fe", "#fad0a8", "#679e8d", "#ffc7b1", "#abe5c0", "#ac8d78", "#c5dddc", "#a48f84", "#cadfb0", "#899694", "#fdcdc1", "#d1dad5", "#dfd8c4"),
  c("#6e9c93", "#ffb4b3", "#7ec6a2", "#eeccfe", "#cddb9d", "#8a90c5", "#dcb983", "#77bff0", "#f0ab92", "#90ddff", "#f1d3a9", "#b5c2fe", "#c1e1b7", "#7596ba", "#bce1c4", "#a88c96", "#5a9daf", "#b18b80", "#d4d6f3", "#949577"),
  c("#e7d6bb", "#749ed5", "#f9d29d", "#67b3e2", "#d09773", "#65ccec", "#d38585", "#7fe8ef", "#cf8190", "#94e8cd", "#ae8cc1", "#b3cf95", "#cbc0fc", "#94a66c", "#eeccff", "#ada368", "#e9a6ce", "#48a297", "#ffc1df", "#799c7a", "#facbe0", "#5d9e9a", "#ffc6c1", "#619bb0", "#fccdcb", "#7197bb", "#b1e4c3", "#9390b1", "#c3e0c0", "#a98c90", "#ade3ce", "#9c927d", "#c2dafe", "#869881", "#e6d3dc", "#6e9ba4", "#bde0d0", "#8196a4", "#b2e1df", "#b9deea")
)

badToGood <- rev(c("#1AB958", "#23B956", "#2BB954", "#31B952", "#37B850", "#3CB84E", "#41B84C", "#45B74A", "#49B749", "#4DB747", "#51B645", "#54B643", "#58B641", "#5BB53F", "#5FB53D", "#62B53C", "#65B43A", "#68B438", "#6BB336", "#6DB335", "#70B333", "#73B231", "#76B230", "#78B12E", "#7BB12C", "#7DB02B", "#80B029", "#82AF28", "#85AF26", "#87AE25", "#8AAE23", "#8CAD22", "#8EAD21", "#91AC1F", "#93AC1E", "#95AB1D", "#97AB1C", "#9AAA1B", "#9CAA1A", "#9EA919", "#A0A918", "#A2A818", "#A4A717", "#A6A716", "#A8A616", "#AAA616", "#ACA515", "#AEA415", "#B0A415", "#B2A315", "#B4A315", "#B6A216", "#B8A116", "#B9A117", "#BBA017", "#BD9F18", "#BF9F18", "#C19E19", "#C29D1A", "#C49D1B", "#C69C1C", "#C79B1D", "#C99A1E", "#CB9A1F", "#CC9920", "#CE9822", "#CF9823", "#D19724", "#D29625", "#D49626", "#D59528", "#D79429", "#D8932A", "#D9932C", "#DB922D", "#DC912E", "#DD9130", "#DF9031", "#E08F33", "#E18F34", "#E28E35", "#E38D37", "#E58C38", "#E68C3A", "#E78B3B", "#E88A3D", "#E98A3E", "#EA8940", "#EB8841", "#EC8843", "#ED8744", "#EE8746", "#EE8647", "#EF8549", "#F0854A", "#F1844C", "#F2844D", "#F2834F", "#F38350", "#F48252", "#F48253", "#F58155", "#F58157", "#F68058", "#F6805A", "#F77F5B", "#F77F5D", "#F87E5E"))

Reference <- function(authors, year, title, journal = "",
                       volume = NULL, pages = NULL, doi = NULL,
                       publisher = NULL, editors = NULL) {
  nAuth <- length(authors)
  if (nAuth > 1L) {
    authors <- paste(paste0(authors[-nAuth], collapse = ", "), "&amp;", authors[nAuth])
  }
  nEd <- length(editors)
  if (nEd > 1L) {
    editors <- paste(paste0(editors[-nEd], collapse = ", "), "&amp;", editors[nEd])
  } else if (nEd < 1) {
    editors <- ""
  }
  paste0("<p class=\"reference\">", authors, " (", year, "). &ldquo;", title,
         "&rdquo;. ",
         if (editors != "") paste0("In: ", editors, " (eds). ") else "",
         if (journal != "") paste0("<i>", journal, "</i>. ") else "",
         if (is.null(volume)) "" else paste0("<b>", volume, "</b>:"),
         if (is.null(publisher)) "" else paste0(publisher, ". "),
         if (is.null(pages)) "" else paste0(paste0(pages, collapse = "&ndash;"), ". "),
         if (is.null(doi)) "" else paste0(
           "doi:<a href=\"https://doi.org/", doi, "\" title=\"CrossRef\">",
                  doi, "</a>. "), 
         "</p>")
}


Arthur2007 <- Reference(
  c("Arthur, D.", "Vassilvitskii, S"),
  title = "k-means++: the advantages of careful seeding",
  year = 2007,
  journal = "Proceedings of the Eighteenth Annual ACM-SIAM Symposium on Discrete Algorithms",
  pages = c(1027, 1035)
)
Bien2011 <- Reference(
  c("Bien, J.", "Tibshirani, R."),
  title = "Hierarchical clustering with prototypes via minimax linkage",
  year = 2011,
  volume = 106,
  doi = "10.1198/jasa.2011.tm10183",
  pages = c(1075, 1084),
  journal = "Journal of the American Statistical Association"
)
Day1985 <- Reference(
  title = "Optimal algorithms for comparing trees with labeled leaves",
  author = "Day, W.H.E.", year = 1985,
  volume = 2,
  pages = c(7, 28),
  doi = "10.1007/BF01908061",
  journal = "Journal of Classification"
)
Estabrook1985 <- Reference(
  c("Estabrook, G.F.", "McMorris, F.R.", "Meacham, C.A."), 1985,
  title = "Comparison of undirected phylogenetic trees based on subtrees of four evolutionary units",
  volume = 34,
  pages = c(193, 200),
  doi = "10.2307/sysbio/34.2.193",
  journal = "Systematic Zoology"
)
Farris1973 <- Reference(title = "On comparing the shapes of taxonomic trees",
                        author = "Farris, J.S.",
                        year = 1973,
                        volume = 22,
                        pages = c(50, 54),
                        doi = "10.2307/2412378",
                        journal = "Systematic Zoology")
Gower1966 <- Reference(  title = "Some distance properties of latent root and vector methods used in multivariate analysis",
                         author = "Gower, J.C.",
                         year = 1966,
                         volume = 53,
                         pages = c(325, 338),
                         doi = "10.2307/2333639",
                         journal = "Biometrika")
Gower1969 <- Reference(
  title = "Minimum spanning trees and single linkage cluster analysis",
  author = c("Gower, J.C.", "Ross, G.J.S."),
  year = 1969, volume = 18, pages = c(54, 64), doi = "10.2307/2346439",
  journal = "Journal of the Royal Statistical Society. Series C (Applied Statistics)")
Kendall2016 <- Reference(
  c("Kendall, M.", "Colijn, C"), 2016,
  "Mapping phylogenetic trees to reveal distinct patterns of evolution",
  "Molecular Biology and Evolution", 33, c(2735, 2743),
  doi = "10.1093/molbev/msw124")
Kruskal1964 <- Reference(
  title = "Multidimensional scaling by optimizing goodness of fit to a nonmetric hypothesis",
  author = "Kruskal, J.B.", year = 1964, volume = 29, pages = c(1, 27),
  doi = "10.1007/BF02289565", journal = "Psychometrika")
Kaski2003 <- Reference(
  title = "Trustworthiness and metrics in visualizing similarity of gene expression",
  author = c("Kaski, S.", "Nikkil&auml;, J.", "Oja, M.", "Venna, J.",
             "T&ouml;r&ouml;nen, P.", "Castr&eacute;n, E."),
  year = 2003, volume = 4, pages = 48, doi = "10.1186/1471-2105-4-48",
  journal = "BMC Bioinformatics")
Maechler2019 <- Reference(
  title = "cluster: cluster analysis basics and extensions", year = 2019,
  author = c("Maechler, M.", "Rousseeuw, P.", "Struyf, A.", "Hubert, M.", "Hornik, K."),
  journal = "Comprehensive R Archive Network")
McInnes2020 <- Reference(
  title = "UMAP: uniform manifold approximation and projection for dimension reduction",
  author = c("McInnes, L.", "Healy, J.", "Melville, J."),
  year = 2020,
  journal = "arXiv", pages = "1802.03426")
Murtagh1983 <- Reference(
  title = "A survey of recent advances in hierarchical clustering algorithms",
  author = "Murtagh, F.", year = 1983, volume = 26, pages = c(354, 359),
  doi = "10.1093/comjnl/26.4.354", journal = "The Computer Journal")
RCoreTeam <- Reference(
  author = "R Core Team", year = 2020,
  title = "R: A language and environment for statistical computing",
  publisher = "R Foundation for Statistical Computing, Vienna, Austria")
Sammon1969 <- Reference(
  title = "A nonlinear mapping for data structure analysis",
  author = "Sammon, J.W.", year = 1969,
  volume = "C-18", pages = c(401, 409),
  doi = "10.1109/T-C.1969.222678", journal = "IEEE Transactions on Computers")
Venables2002 <- Reference(
  title = "Modern Applied Statistics with S. Fourth edition",
  author = c("Venables, W.N.", "Ripley, B.D."),
  publisher = "Springer, New York",
  year = 2002)
Venna2001 <- Reference(
  title = "Neighborhood preservation in nonlinear projection methods: an experimental study",
  author = c("Venna, J.", "Kaski, S."), year = 2001, pages = c(485, 491),
  journal = "Lecture Notes in Computer Science: Artificial Neural Networks&mdash;ICANN 2001",
  editors = c("Dorffner, G.", "Bischof, H.", "Hornik, K."),
  publisher = "Springer, Berlin",
  doi = "10.1007/3-540-44668-0_68")

 
Sand2014 <- Reference(
    author = c("Sand, A.", "Holt, M.K.", "Johansen, J.", "Brodal, G.S.",
               "Mailund, T.", "Pedersen, C.N.S."),
    doi = "10.1093/bioinformatics/btu157",
    journal = "Bioinformatics",
    pages = c(2079, 2080),
    title = "tqDist: a library for computing the quartet and triplet distances between binary or general trees",
    volume = 30,
    year = 2014
)
Smith2020 <- Reference("Smith, M.R.", 2020,
  "Information theoretic Generalized Robinson-Foulds metrics for comparing phylogenetic trees",
  "Bioinformatics", 36, c(5007, 5013),
  doi = "10.1093/bioinformatics/btaa614")
Smith2021 <- Reference("*Smith, M.R.", 2022,
  "Robust analysis of phylogenetic tree space",
  "Systematic Biology", 71, c(1255, 1270),
  doi = "10.1093/sysbio/syab100")
SmithDist <- Reference("Smith, M.R.", 2020,
  "TreeDist: distances between phylogenetic trees",
  doi = "10.5281/zenodo.3528123", "Comprehensive R Archive Network")
SmithQuartet <- Reference("Smith, M.R.", 2019,
  "Quartet: comparison of phylogenetic trees using quartet and split measures",
  "Comprehensive R Archive Network", doi = "10.5281/zenodo.2536318")
Steel1993 <- Reference(
  c("Steel, M.", "Penny, D"), 1993,
  "Distributions of tree comparison metrics\u2212some new results",
  "Systematic Biology", 42, c(126, 141),
  doi = "10.1093/sysbio/42.2.126")
Stockham2002 <- Reference(
  author = c("Stockham, C.", "Wang, L.-S.", "Warnow, T."), 2002,
  "Statistically based postprocessing of phylogenetic analysis by clustering",
  "Bioinformatics", 18, c("S285", "S293"),
  doi = "10.1093/bioinformatics/18.suppl_1.S285")
Ward1963 <- Reference("Ward, J.H.", 1963,
                      "Hierarchical grouping to optimize an objective function",
                      "Journal of the American Statistical Association",
                      58, c(236, 244),
                      doi = "10.1080/01621459.1963.10500845")
Robinson1981 <- Reference(
title = "Comparison of phylogenetic trees",
volume = 53,
doi = "10.1016/0025-5564(81)90043-2",
journal = "Mathematical Biosciences",
author = c("Robinson, D.F.", "Foulds, L.R."),
year = 1981,
pages = c(131, 147)
)


ui <- fluidPage(
  useShinyjs(),
    column(3, 
      tabsetPanel(
        tabPanel("Input",
          textOutput(outputId = "treesInMemory"),
          fileInput("treefile", "Load trees", placeholder = "No tree file selected"),
          textOutput(outputId = "treesStatus"),
          textOutput(outputId = "keptTrees"),
          sliderInput(inputId = "keepTrees",
                      label = "Trees to analyse:",
                      min = 1,
                      max = 90,
                      value = c(1, 90)),
          sliderInput(inputId = "thinTrees",
                      label = "Sample every:",
                      min = 0,
                      max = 12,
                      round = FALSE,
                      step = 0.01,
                      pre = "2^",
                      value = 0),
          textOutput(outputId = "thinnedTrees"),
          actionButton("replaceTrees", "Replace existing"),
          actionButton("addTrees", "Add batch to existing"),
          
          tags$p("For custom point colours, upload a file with a value for ",
                 "each tree in a separate row: "),
          fileInput("pt.data", "Point property file",
                    accept = c(".csv", ".txt", ".xls", ".xlsx")),
          checkboxInput("pt.data.subsample", "Subsample point properties", TRUE),
          tags$p("Subsampling not supported for multiple batches."),
          tagAppendAttributes(textOutput("pt.data.status"), class = "message"),
          HTML("<p>Do <a href=\"https://github.com/ms609/TreeTools/issues/new?title=Treespace",
               "app:\" title=\"Suggest improvements\">report</a> any bugs or",
               "feature requests to the maintainer (<a",
               "href=\"https://smithlabdurham.github.io/\"",
               "title=\"Martin Smith\">Martin R. Smith</a>).</p>"),
        ),
        tabPanel("Analysis",
           selectInput("distance", "Distance method",
                       choices = list("Clustering information" = "cid",
                                      "Phylogenetic information" = "pid",
                                      "Quartet (slow)" = "qd",
                                      "Kendall\u2212Colijn (rooted)" = "kc",
                                      "Path" = "path",
                                      "Robinson\u2212Foulds" = "rf"),
                       selected = "cid"),
          
          textOutput(outputId = "mappingStatus"),
          fluidRow(plotOutput(outputId = "pcQuality", height = "72px")),
          selectInput("mapping", "Mapping method",
                       choices = list("Princ. comps (classical MDS)" = "pca",
                                      "Sammon mapping mMDS" = "nls",
                                      "Kruskal-1 nmMDS (slow)" = "k",
                                      "t-UMAP" = "tumap",
                                      "UMAP" = "umap"
                                      ),
                       selected = "pca"),
          hidden(sliderInput(inputId = "nNeighb", "Nearest neighbours",
                             min = 2, max = 100, value = 15)),
          
          checkboxGroupInput("clustering", "Clustering methods",
                             choices = list(
                               "Partitioning around medoids" = "pam",
                               "Hierarchical, minimax linkage" = "hmm",
                               "Hierarchical, single linkage" = "hsi",
                               "Hierarchical, complete linkage" = "hco",
                               "Hierarchical, average linkage" = "hav",
                               "Hierarchical, median linkage" = "hmd",
                               "Hierarchical, centroid linkage" = "hct",
                               "Hierarchical, Ward d\ub2 linkage" = "hwd",
                               "K-means++" = "kmn",
                               "Spectral" = "spec"),
                             selected = c("pam", "hmm", "kmn")),
          sliderInput(inputId = "maxClusters",
                      label = "Maximum cluster number:",
                      min = 2,
                      max = 20,
                      value = 8),
          fluidRow(plotOutput(outputId = "clustQuality", height = "80px")),
        ),
        tabPanel("Display",
          sliderInput(inputId = "dims",
                      label = "Dimensions to plot:",
                      min = 2, max = 15, value = 5),
          plotOutput("howManyDims", width = "100%", height = "180px"),
          
          sliderInput(inputId = "mst",
                      label = "Minimum spanning tree size",
                      min = 0,
                      max = 90,
                      value = 90),
          checkboxGroupInput("display", "Display settings",
                             list(
                               "Depict MST stress" = "mstStrain",
                               "Clusters' convex hulls" = "hulls",
                               "Cluster consensus trees" = "cons",
                               "Tree numbers" = "labelTrees",
                               "Interactive 3D plot" = "show3d"
                             ),
                             character(0)),
          sliderInput("pt.cex", "Point size",
                      min = 0, max = 4, value = 1, 
                      step = 0.01),
          sliderInput("pt.opacity", "Point opacity",
                      min = 0, max = 255, value = 196,
                      step = 1),
          selectInput("pt.pch", "Point symbols:",
                      list(
                        "Solid circle" = "16",
                        "Open ring" = "1",
                        "One per cluster" = "clust",
                        "One per batch" = "entry"
                      ), "16"),
          selectInput("pt.col", "Colour points by:",
                      list("Cluster membership" = "clust", 
                           "Batch" = "entry",
                           "Sequence within batches" = "seq",
                           "Distance from focal tree" = "dist",
                           "Custom categorical" = "discrete",
                           "Custom continuous" = "continuous"), "clust"),
          tagAppendAttributes(textOutput("pt.col.status"), class = "message"),
          plotOutput("pt.col.scale", height = "36px"),
          hidden(sliderInput("which.tree", "Focal tree index",
                             min = 1, max = 90, value = 1, step = 1)),
        )
      ),
    ),
    
    column(9,
      fluidRow(id = "plotConfig",
        tags$span("Plot size:", id = "plotSizeSpan"),
        sliderInput(inputId = "plotSize", label = NULL,
                    width = "200px",
                    min = 100, max = 2000,
                    post = "px", value = 600),
        tags$span("Save as: "),
        downloadButton("savePdf", "PDF"),
        downloadButton("savePng", "PNG"),
        downloadButton("saveCons", "Cluster consensus trees"),
      ),
      fluidRow(
        plotOutput(outputId = "distPlot", height = "0px"),
        uiOutput(outputId = "threeDPlot", style = "height: 600px; width: 600px;"),
        hidden(plotOutput("clustCons", height = "200px")),
        htmlOutput("references"),
      ),
  )
)

server <- function(input, output, session) {
  #treeNumbers <- c(1:220)
  treeNumbers <- c(1:50, 401:440)
  
  r <- reactiveValues(allTrees = as.phylo(treeNumbers, 8),
                      entryOrder = rep(1, length(treeNumbers)),
                      treesUpdated = Sys.time())
  
  ##############################################################################
  # Load trees
  ##############################################################################
  treeFile <- reactive({
    fileInput <- input$treefile
    if (is.null(input)) return("No tree file selected.")
    tmpFile <- fileInput$datapath
    if (is.null(tmpFile)) return ("No trees found.")
    if (length(grep("#NEXUS", toupper(readLines(tmpFile, 1L)),
                    fixed = TRUE)) > 0) {
      ret <- c(read.nexus(tmpFile))
    } else {
      ret <- ReadTntTree(tmpFile)
      if (length(ret) == 0) ret <- read.tree(tmpFile)
    }
    
    if (!inherits(ret, c("phylo", "multiPhylo"))) {
      return("Could not read trees from file")
    }
    nTrees <- length(ret)
    if (nTrees) {
      updateSliderInput(session, "keepTrees",
                        value = c(0, nTrees),
                        max = nTrees)
      updateSliderInput(session, "thinTrees",
                        value = log2(nTrees / 100),
                        max = max(0, signif(log2(nTrees) - 1L, 3)))
    }
    ret
  })
  
  output$treesStatus <- renderText({
    if (is.character(treeFile())) return(treeFile())
    paste(length(treeFile()), "trees in file.")
  })
  
  treesLoaded <- reactive({
    !is.character(treeFile())
  })
  
  keptRange <- reactive({
    if (treesLoaded()) {
      nTrees <- length(treeFile())
      input$keepTrees
    } else {
      c(1, length(r$allTrees))
    }
  })
  
  output$keptTrees <- renderText({
    nTrees <- length(treeFile())
    if (treesLoaded()) {
      paste0("Keeping trees ", keptRange()[[1]], " to ", keptRange()[[2]], ".")
    } else {
      "Using example trees."
    }
  })
  
  output$thinnedTrees <- renderText({
    if (input$thinTrees == 1) "" else {
      nTrees <- length(thinnedTrees())
      if (nTrees > 1500) {
        paste("Selecting", nTrees, "trees.  I hope you're feeling patient.")
      } else if (nTrees > 500) {
        paste("Selecting", nTrees, "trees.  May be slow.")
      } else {
        paste("Selecting", nTrees, "trees.")
      }
    }
  })
  
  thinnedTrees <- reactive({
    as.integer(seq(keptRange()[[1]], keptRange()[[2]], by = 2 ^ input$thinTrees))
  })
  
  TreeNumberUpdate <- function() {
    nTrees <- length(r$allTrees)
    updateSliderInput(session, "mst", value = min(500, nTrees), max = nTrees)
    updateSliderInput(session, "which.tree", max = nTrees)
    updateSliderInput(session, "dims",
                      max = nProjDim(),
                      value = min(input$dims, nProjDim()))
    updateSliderInput(session, "nNeighb",
                      max = nTrees,
                      value = min(nNeighb(), nTrees, 100))
  }
  
  observeEvent(input$addTrees, {
    newTrees <- treeFile()[thinnedTrees()]
    if (length(r$allTrees) == 0) {
      r$entryOrder <- rep(1L, length(newTrees))
      r$allTrees <- c(newTrees)
    } else {
      r$entryOrder <- c(r$entryOrder,
                        rep(1L + max(r$entryOrder), length(newTrees)))
      r$allTrees <- c(r$allTrees, newTrees)
    }
    TreeNumberUpdate()
  })
  
  observeEvent(input$replaceTrees, {
    r$allTrees <- c(treeFile()[thinnedTrees()])
    r$entryOrder <- rep(1L, length(r$allTrees))
    TreeNumberUpdate()
  })
  
  observeEvent(r$allTrees, {
    r$treesUpdated <- Sys.time()
  })
  
  output$treesInMemory <- renderText({
    paste(length(r$allTrees), "trees in memory.")
  })
  
  nLeaves <- reactive(max(NTip(r$allTrees)))
  
  ##############################################################################
  # Calculate distances
  ##############################################################################

  chosenDistance <- function() {
    switch(input$distance, 
           cid = "Clustering information distance",
           pid = "Phylogenetic information distance",
           qd = "Quartet divergence",
           kc = "Kendall\u2212Colijn distance",
           path = "Path distance",
           rf = "Robinson\u2212Foulds distance")
  }
  
  distances <- bindCache(
    reactive({
      if (length(r$allTrees) > 1L) {
        LogMsg("Calculating distances: ", input$distance, " _ ",
               .DateTime(r$treesUpdated))
        withProgress(
          message = "Calculating distances", value = 0.99,
          switch(
            input$distance,
            "cid" = ClusteringInfoDistance(r$allTrees),
            "pid" = PhylogeneticInfoDistance(r$allTrees),
            "qd" = as.dist(Quartet::QuartetDivergence(
              Quartet::ManyToManyQuartetAgreement(r$allTrees),
              similarity = FALSE)
            ),
            "kc" = KendallColijn(r$allTrees),
            "path" = PathDist(r$allTrees),
            "rf" = RobinsonFoulds(r$allTrees)
          )
        )
      } else {
        matrix(0, 0, 0)
      }
    }),
    r$treesUpdated,
    input$distance
  )
  
  ##############################################################################
  # Mapping
  ##############################################################################
  maxProjDim <- reactive({
    min(15L, length(r$allTrees) - 1L)
  })
  
  nProjDim <- reactive({
    dim(mapping())[[2]]
  })
  
  dims <- debounce(reactive({
    if (mode3D()) 3L else {
      min(input$dims, maxProjDim())
    }
  }), 400)
  
  mappingUsesNeighb <- function() {
    input$mapping %in% c("umap", "tumap")
  }
  
  observeEvent(input$mapping, {
    if (mappingUsesNeighb()) {
      showElement("nNeighb")
    } else {
      hideElement("nNeighb")
    }
  })
  
  nNeighb <- debounce(reactive(input$nNeighb), 300)
  
  observe({
    if (!.uwotAvailable) {
      updateSelectInput(session, "mapping",
                        choices = c("Principal Components (PCA)" = "pca", 
                                    "Kruskal's non-metric MDS" = "k", 
                                    "Sammon's non-linear mapping" = "nls")
      )
    }
  })
  
  mapping <- bindCache(
    reactive({
      if (maxProjDim() > 1L) {
        LogMsg("Mapping distances")
        withProgress(
          message = "Mapping distances",
          value = 0.99,
          {
            uwot <- .uwotAvailable
            switch(
              input$mapping,
              "pca" = cmdscale(distances(), k = maxProjDim()),
              "k" = MASS::isoMDS(distances(), k = maxProjDim())$points,
              "nls" = MASS::sammon(distances(), k = maxProjDim())$points,
              "tumap" = if (uwot) {
                getFromNamespace("tumap", .uwot_pkg)(distances(), verbose = FALSE,
                            n_neighbors = nNeighb(),
                            n_components = maxProjDim())
              } else {
                showNotification(paste(.uwot_pkg, "package unavailable. Defaulting to PCA."), type = "error")
                cmdscale(distances(), k = maxProjDim())
              },
              "umap" = if (uwot) {
                getFromNamespace("umap", .uwot_pkg)(distances(), verbose = FALSE,
                           a = 1.8956, b = 0.8006,
                           approx_pow = TRUE,
                           n_neighbors = nNeighb(),
                           n_components = maxProjDim())
              } else {
                showNotification(paste(.uwot_pkg, "package unavailable. Defaulting to PCA."), type = "error")
                cmdscale(distances(), k = maxProjDim())
              }
            )
          }
        )
      } else {
        matrix(0, 0, 0)
      }
    }),
    r$treesUpdated,
    input$mapping,
    input$distance,
    nNeighb()
  )
  
  
  projQual <- bindCache(
    reactive({
      LogMsg("Estimate quality of mapping with dimensions: ",
             paste0(dim(mapping()), collapse = ", "))
      withProgress(message = "Estimating mapping quality", {
        vapply(seq_len(nProjDim()), function(k) {
          incProgress(1 / nProjDim())
          MappingQuality(distances(), dist(mapping()[, seq_len(k)]), 10)
        }, numeric(4))
      })
    }),
    nProjDim(),
    input$mapping,
    input$distance,
    nNeighb()
  )
    
  
  LogScore <- function(x) {
    (-(log10(1 - x + 1e-2))) / 2
  }
  
  output$pcQuality <- renderPlot({
    par(mar = c(2, 0, 0, 0))
    nStop <- length(badToGood)
    
    plot(NULL, xlim = c(0, 1), ylim = c(-1.5, 2.5),
         ann = FALSE, axes = FALSE)
    x <- seq.int(from = 0, to = 1, length.out = nStop + 1)
    segments(x[-(nStop + 1)], numeric(nStop), x[-1], lwd = 5, col = badToGood)
    
    logScore <- LogScore(projQual()["TxC", nProjDim()])
    lines(rep(logScore, 2), c(-1, 1), lty = 3)
    
    tickPos <- c(0, 0.5, 0.7, 0.8, 0.9, 0.95, 1.0)
    ticks <- LogScore(tickPos)
    
    axis(1, at = ticks, labels = NA, line = 0)
    axis(1, tick = FALSE, at = ticks, labels = tickPos, line = 0)
    axis(1, line = -1, tick = FALSE,
         at = ticks[-1] - ((ticks[-1] - ticks[-length(ticks)]) / 2),
         labels = c("", "dire", "", "ok", "gd", "excellent"))
    axis(3, at = 0.5, tick = FALSE, line = -2, 
         paste0(dims(), "D mapping quality (trustw. \ud7 contin.):"))
  })
  
  
  output$howManyDims <- renderPlot({
    par(mar = c(2.5, 2.5, 0, 0), xpd = NA, mgp = c(1.5, 0.5, 0))
    txc <- projQual()["TxC", ]
    nStop <- length(badToGood)
    
    plot(txc, type = "n", ylim = c(min(txc, 0.5), 1),
         frame.plot = FALSE, axes = FALSE,
         xlab = "Dimension", ylab = "Trustw. \uD7 Contin.")
    par(xpd = FALSE)
    axis(1, 1:14)
    axis(2)
    tickPos <- c(0, 0.5, 0.7, 0.8, 0.9, 0.95, 1.0)
    mids <- c(0.6, 0.75, 0.85, 0.925)
    text(rep(15, 4), mids, pos = 2, cex = 0.8,
         col = badToGood[nStop * LogScore(mids)],
         c("Essentially random", "Dangerous", "Usable", "Good"))
    text(1, 0.975, pos = 4, "Excellent", cex = 0.8, 
         col = badToGood[LogScore(0.975) * nStop])
    for (i in tickPos[-1]) {
      abline(h = i, lty = 3, col = badToGood[LogScore(i) * nStop])
    }
    points(txc, type = "b")
    txcNow <- txc[dims()]
    
    points(dims(), txcNow, pch = 16, col = badToGood[LogScore(txcNow) * nStop],
           cex = 1.6)
  })
  
  ##############################################################################
  # Clusterings
  ##############################################################################
  maxClust <- reactive(min(input$maxClusters, length(r$allTrees) - 1L))
  
  possibleClusters <- reactive(2:maxClust())
  
  pamClusters <- bindCache(
    reactive({
      lapply(possibleClusters(), function(k) {
        cluster::pam(distances(), k = k)
      })
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  pamSils <- bindCache(
    reactive({
      vapply(pamClusters(), function(pamCluster) {
        mean(cluster::silhouette(pamCluster)[, 3])
      }, double(1))
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  hmmClusters <- bindCache(
    reactive({
      hTree <- protoclust::protoclust(distances())
      lapply(possibleClusters(), function(k) cutree(hTree, k = k))
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  hmmSils <- bindCache(
    reactive({
      vapply(hmmClusters(), function(hCluster) {
        mean(cluster::silhouette(hCluster, distances())[, 3])
      }, double(1))
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  hwdClusters <- bindCache(
    reactive({
      hTree <- stats::hclust(distances(), method = "ward.D2")
      lapply(possibleClusters(), function(k) cutree(hTree, k = k))
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  hwdSils <- bindCache(
    reactive({
      vapply(hwdClusters(), function(hCluster) {
        mean(cluster::silhouette(hCluster, distances())[, 3])
      }, double(1))
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  hsiClusters <- bindCache(
    reactive({
      hTree <- stats::hclust(distances(), method = "single")
      lapply(possibleClusters(), function(k) cutree(hTree, k = k))
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  hsiSils <- bindCache(
    reactive({
      vapply(hsiClusters(), function(hCluster) {
        mean(cluster::silhouette(hCluster, distances())[, 3])
      }, double(1))
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  hcoClusters <- bindCache(
    reactive({
      hTree <- stats::hclust(distances(), method = "complete")
      lapply(possibleClusters(), function(k) cutree(hTree, k = k))
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  hcoSils <- bindCache(
    reactive({
      vapply(hcoClusters(), function(hCluster) {
        mean(cluster::silhouette(hCluster, distances())[, 3])
      }, double(1))
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  havClusters <- bindCache(
    reactive({
      hTree <- stats::hclust(distances(), method = "average")
      lapply(possibleClusters(), function(k) cutree(hTree, k = k))
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  havSils <- bindCache(
    reactive({
      vapply(havClusters(), function(hCluster) {
        mean(cluster::silhouette(hCluster, distances())[, 3])
      }, double(1))
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  hmdClusters <- bindCache(
    reactive({
      hTree <- stats::hclust(distances(), method = "median")
      lapply(possibleClusters(), function(k) cutree(hTree, k = k))
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  hmdSils <- bindCache(
    reactive({
      vapply(hmdClusters(), function(hCluster) {
        mean(cluster::silhouette(hCluster, distances())[, 3])
      }, double(1))
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  hctClusters <- bindCache(
    reactive({
      hTree <- stats::hclust(distances() ^ 2, method = "centroid")
      lapply(possibleClusters(), function(k) cutree(hTree, k = k))
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  hctSils <- bindCache(
    reactive({
      vapply(hctClusters(), function(hCluster) {
        mean(cluster::silhouette(hCluster, distances())[, 3])
      }, double(1))
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  kppClusters <- bindCache(
    reactive({
      lapply(possibleClusters(), function(k) KMeansPP(distances(), k))
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  kppSils <- bindCache(
    reactive({
      vapply(kppClusters(), function(kCluster) {
        mean(cluster::silhouette(kCluster$cluster, distances())[, 3])
      }, double(1))
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  specClusters <- bindCache(
    reactive({
      spectralEigens <- SpectralEigens(
        distances(),
        nn = min(ncol(as.matrix(distances())) - 1L, 10),
        nEig = 3L
      )
      lapply(possibleClusters(), function(k) {
        cluster::pam(spectralEigens, k = k)
      })
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  specSils <- bindCache(
    reactive({
      vapply(specClusters(), function(cluster) {
        mean(cluster::silhouette(cluster$cluster, distances())[, 3])
      }, double(1))
    }),
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  .Ascending <- function(x) {
    order(unique(x))[x]
  }
  
  clusterings <- bindCache(
    reactive({
      if (maxClust() > 1) {
        
        hSil <- pamSil <- havSil <- kSil <- specSil <-
          hsiSil <- hcoSil <- hmdSil <- hctSil <- hwdSil <- -99
        
        nMethodsChecked <- 10
        methInc <- 1 / nMethodsChecked
        
        withProgress(message = "Clustering", {
          
          incProgress(methInc, detail = "PAM clustering")
          if ("pam" %in% input$clustering) {
            bestPam <- which.max(pamSils())
            pamSil <- pamSils()[bestPam]
            pamCluster <- pamClusters()[[bestPam]]$cluster
          }
          
          incProgress(methInc, detail = "minimax clustering")
          if ("hmm" %in% input$clustering) {
            bestH <- which.max(hmmSils())
            hSil <- hmmSils()[bestH]
            hCluster <- hmmClusters()[[bestH]]
          }
          
          incProgress(methInc, detail = "Ward D\ub2 clustering")
          if ("hwd" %in% input$clustering) {
            bestHwd <- which.max(hwdSils())
            hwdSil <- hwdSils()[bestHwd]
            hwdCluster <- hwdClusters()[[bestHwd]]
          }
          
          incProgress(methInc, detail = "single clustering")
          if ("hsi" %in% input$clustering) {
            bestHsi <- which.max(hsiSils())
            hsiSil <- hsiSils()[bestHsi]
            hsiCluster <- hsiClusters()[[bestHsi]]
          }
          
          incProgress(methInc, detail = "complete clustering")
          if ("hco" %in% input$clustering) {
            bestHco <- which.max(hcoSils())
            hcoSil <- hcoSils()[bestHco]
            hcoCluster <- hcoClusters()[[bestHco]]
          }
          
          incProgress(methInc, detail = "average clustering")
          if ("hav" %in% input$clustering) {
            bestHav <- which.max(havSils())
            havSil <- havSils()[bestHav]
            havCluster <- havClusters()[[bestHav]]
          }
          
          incProgress(methInc, detail = "median clustering")
          if ("hmd" %in% input$clustering) {
            bestHmd <- which.max(hmdSils())
            hmdSil <- hmdSils()[bestHmd]
            hmdCluster <- hmdClusters()[[bestHmd]]
          }
          
          incProgress(methInc, detail = "centroid clustering")
          if ("hct" %in% input$clustering) {
            bestHct <- which.max(hctSils())
            hctSil <- hctSils()[bestHct]
            hctCluster <- hctClusters()[[bestHct]]
          }
          
          incProgress(methInc, detail = "K-means++ clustering")
          if ("kmn" %in% input$clustering) {
            bestK <- which.max(kppSils())
            kSil <- kppSils()[bestK]
            kCluster <- kppClusters()[[bestK]]$cluster
          }
          
          incProgress(methInc, detail = "spectral clustering")
          if ("spec" %in% input$clustering) {
            bestSpec <- which.max(specSils())
            specSil <- specSils()[bestSpec]
            specCluster <- specClusters()[[bestSpec]]$cluster
          }
          
          bestCluster <- c("none", "pam", "hmm", "hsi", "hco", "hav",
                           "hmd", "hct", "hwd", "kmn", "spec")[
                             which.max(c(0, pamSil, hSil, hsiSil, hcoSil,
                                         havSil, hmdSil, hctSil,
                                         hwdSil, kSil, specSil))]
          
        })
      } else {
        bestCluster <- "none"
      }
      
      list(
        method = switch(bestCluster,
                        pam = "part. around medoids",
                        hmm = "minimax linkage",
                        hsi = "single linkage",
                        hco = "complete linkage",
                        hav = "average linkage",
                        hmd = "median linkage",
                        hct = "centroid linkage",
                        hwd = "Ward d\ub2 linkage",
                        kmn = "K-means++",
                        spec = "spectral",
                        "no attempt to find any"),
        n = 1 + switch(bestCluster,
                       pam = bestPam,
                       hmm = bestH,
                       hsi = bestHsi,
                       hco = bestHco,
                       hav = bestHav,
                       hmd = bestHmd,
                       hct = bestHct,
                       hwd = bestHwd,
                       har = bestHav, kmn = bestK,
                       spec = bestSpec,
                       0),
        sil = switch(bestCluster,
                     pam = pamSil,
                     hmm = hSil,
                     hsi = hsiSil,
                     hco = hcoSil,
                     hav = havSil,
                     hmd = hmdSil,
                     hct = hctSil,
                     hwd = hwdSil,
                     kmn = kSil, spec = specSil, 
                     0), 
        cluster = .Ascending(switch(
          bestCluster,
          pam = pamCluster,
          hmm = hCluster,
          hsi = hsiCluster,
          hco = hcoCluster,
          hav = havCluster,
          hmd = hmdCluster,
          hct = hctCluster,
          hwd = hwdCluster,
          kmn = kCluster,
          spec = specCluster,
          1))
      )
    }),
    input$clustering,
    maxClust(),
    r$treesUpdated,
    input$distance
  )
  
  mstSize <- debounce(reactive(input$mst), 100)
  
  mstEnds <- bindCache(
    reactive({
      dist <- as.matrix(distances())
      nRows <- dim(dist)[[1]]
      withProgress(message = "Calculating MST", {
        selection <- unique(round(seq.int(1, nRows,
                                          length.out = max(2L, mstSize()))))
        edges <- MSTEdges(dist[selection, selection])
        edges[, 1] <- selection[edges[, 1]]
        edges[, 2] <- selection[edges[, 2]]
        edges
      })
    }),
    r$treesUpdated,
    input$distance,
    mstSize()
  )
  
  output$clustQuality <- renderPlot({
    par(mar = c(2, 0.5, 0, 0.5), xpd = NA, mgp = c(2, 1, 0))
    cl <- clusterings()
    sil <- cl$sil
    if (length(sil) == 0) sil <- -0.5
    nStop <- 400
    range <- c(0.5, 1)
    negScale <- hcl.colors(nStop, "plasma")[seq(range[[1]] * nStop, 1,
                                            length.out = nStop * range[[1]])]
    posScale <- hcl.colors(nStop, "viridis")
    
    plot(seq(-range[[1]], range[[2]], length.out = nStop * sum(range)),
         rep(0, nStop * sum(range)),
         pch = 15, col = c(negScale, posScale),
         ylim = c(-2, 2),
         ann = FALSE, axes = FALSE)
    lines(c(sil, sil), c(-1, 1), lty = 3)
    
    ticks <- c(-0.5, 0, 0.25, 0.5, 0.7, 1)
    axis(1, at = ticks, line = -1)
    axis(1, tick = FALSE, at = ticks[-1] - ((ticks[-1] - ticks[-6]) / 2),
         labels = c("Structure:", "none", "weak", "ok", "strong"),
         line = -2)
    
    axis(1, tick = FALSE, line = 0, at = 0.25,
         labels = paste0("Silhouette coefficient (", round(sil, 3), ")"))
    
    
    axis(3, tick = FALSE, line = -2, at = 0.25,
         labels = if (sil > 0.25) 
           paste0(cl$n, " clusters found with ", cl$method) else 
             paste0("No meaningful clusters found"))
    
  })
  
  observeEvent(input$display, {
    if ("cons" %in% input$display) {
      show("clustCons")
    } else {
      hide("clustCons")
    }
  }, ignoreInit = TRUE, ignoreNULL = FALSE)
  
  consRows <- reactive({
    cl <- clusterings()
    if (cl$sil > 0.25) ceiling(cl$n / 3) else 1L
  })
  
  consSize <- reactive({
    nLeaves() * 12 * consRows()
  })
  
  PlotClusterCons <- function() {
    cl <- clusterings()
    par(mar = c(0.2, 0, 0.2, 0), xpd = NA)
    par(cex = 0.9)
    if (cl$sil > 0.25) {
      par(mfrow = c(consRows(), ceiling(cl$n / consRows())))
      for (i in seq_len(cl$n)) {
        col <- palettes[[min(length(palettes), cl$n)]][i]
        tr <- ape::consensus(r$allTrees[cl$cluster == i])
        tr$edge.length <- rep(1, dim(tr$edge)[[1]])
        plot(tr, edge.width = 2, font = 1, cex = 1,
             edge.color = col, tip.color = col)
      }
    } else {
      tr <- ape::consensus(r$allTrees)
      tr$edge.length <- rep(1, dim(tr$edge)[[1]])
      plot(tr,edge.width = 2, font = 1, cex = 1,
           edge.color = palettes[[1]],
           tip.color = palettes[[1]])
    }
  }
  
  output$clustCons <- renderPlot({
    if ("cons" %in% input$display) {
      PlotClusterCons()
    }
  }, height = consSize)
  
  ##############################################################################
  # Plot settings: point style
  ##############################################################################
  PointDataStatus <- function(...) {
    msg <- paste0(...)
    output$pt.data.status <- renderText(msg)
    output$pt.col.status <- renderText(msg)
  }
  
  pointData <- reactive({
    fp <- input$pt.data$datapath
    PointDataStatus("")
    extension <- if(is.null(fp)) "" else substr(fp, nchar(fp) - 3, nchar(fp))
    ret <- switch(extension,
                  ".csv" = read.csv(fp),
                  ".txt" = read.table(fp),
                  ".xls" = readxl::read_excel(fp),
                  "xlsx" = readxl::read_excel(fp),
                  {
                    PointDataStatus("Unrecognized file extension.")
                    matrix(0)
                  }
    )
    
    if (input$pt.data.subsample) ret[thinnedTrees(), 1] else ret[, 1]
  })
  
  ContinuousPtCol <- function(dat, bigDark = FALSE) {
    show("pt.col.scale")
    scale <- substr(hcl.colors(256, "plasma"), 1, 7)
    if (bigDark) scale <- rev(scale)
    output$pt.col.scale <- renderPlot({
      par(mar = c(1, 1, 0, 1))
      plot(1:256, rep(0, 256), pch = 15, col = scale,
           ylim = c(-1, 0), ann = FALSE, axes = FALSE)
      axis(1, at = c(1, 256), labels = signif(range(dat), 4), line = -1)
    })
    scale[cut(dat, 256)]
  }
  
  observeEvent(input$pt.col, {
    if (input$pt.col == "dist") show("which.tree") else hide("which.tree")
  })
  
  pointCols <- reactive({
    switch(input$pt.col,
           "clust" = {
             hide("pt.col.scale")
             cl <- clusterings()
             if (cl$sil > 0.25) {
               palettes[[min(length(palettes), cl$n)]][cl$cluster]
             } else palettes[[1]]
           },
           "entry" = {
             hide("pt.col.scale")
             n <- r$entryOrder
             palettes[[max(n)]][n]
           },
           "seq" = {
             ContinuousPtCol(seq_along(r$allTrees))
           },
           "dist" = {
             dat <- as.matrix(distances())[input$which.tree, ]
             message(paste0(round(dat, 2)[1:10], collapse = ", "))
             message(cut(dat, 256)[1:10])
             ContinuousPtCol(as.matrix(distances())[input$which.tree, ],
                             bigDark = TRUE)
           },
           "discrete" = {
             dat <- pointData()
             categories <- unique(dat)
             nCol <- length(categories)
             if (nCol <= length(palettes)) {
               PointDataStatus("")
               show("pt.col.scale")
               output$pt.col.scale <- renderPlot({
                 par(mar = c(1, 0, 0, 0))
                 plot(seq_len(nCol), rep(0, nCol), pch = 15, ylim = c(-1, 0),
                      col = palettes[[nCol]], ann = FALSE, axes = FALSE)
                 axis(1, at = seq_len(nCol), labels = categories,
                      tick = FALSE, line = -1)
               })
               palettes[[nCol]][match(dat, categories)]
             } else {
               hide("pt.col.scale")
               PointDataStatus(nCol, " categories is too many to display;",
                               "did you mean continuous?")
               c(palettes[[length(palettes)]], "#000000")[
                 pmin(match(dat, categories), length(palettes[[length(palettes)]]) + 1L)]
             }
           },
           "continuous" = {
             dat <- pointData()
             if (is.numeric(dat)) {
               ContinuousPtCol(dat)
             } else {
               hide("pt.col.scale")
               PointDataStatus("Point data are not numeric.",
                               "Try selecting \"Custom categorical\".")
               "#000000"
             }
           }
    )
  })
  
  pchs <- reactive({
    switch(input$pt.pch,
           "1" = 1,
           "16" = 16,
           "clust" = {
             cl <- clusterings()
             if (cl$sil > 0.25) {
               cl$cluster - 1
             } else 16
           },
           "entry" = {
             r$entryOrder - 1
           }
    )
  })
  
  ##############################################################################
  # Render plot
  ##############################################################################
  treespacePlot <- function() {
    cl <- clusterings()
    proj <- mapping()
    
    nDim <- min(dims(), nProjDim())
    plotSeq <- matrix(0, nDim, nDim)
    nPanels <- nDim * (nDim - 1L) / 2L
    plotSeq[upper.tri(plotSeq)] <- seq_len(nPanels)
    layout(t(plotSeq[-nDim, -1]))
    par(mar = rep(0.2, 4))
    withProgress(message = "Drawing plot", {
      for (i in 2:nDim) for (j in seq_len(i - 1)) {
        incProgress(1 / nPanels)
        # Set up blank plot
        plot(proj[, j], proj[, i], ann = FALSE, axes = FALSE, frame.plot = TRUE,
             type = "n", asp = 1, xlim = range(proj), ylim = range(proj))
        
        # Plot MST
        if (mstSize() > 0) {
          MSTSegments(proj[, c(j, i)], mstEnds(),
                      col = if ("mstStrain" %in% input$display) {
                        StrainCol(distances(), proj[, c(j, i)])
                      } else "#bbbbbb")
        }
        
        # Add points
        points(proj[, j], proj[, i], pch = pchs(),
               col = paste0(pointCols(), as.hexmode(input$pt.opacity)),
               cex = input$pt.cex)
        
        if ("hulls" %in% input$display && cl$sil > 0.25) {
          # Mark clusters
          for (clI in seq_len(cl$n)) {
            inCluster <- cl$cluster == clI
            clusterX <- proj[inCluster, j]
            clusterY <- proj[inCluster, i]
            hull <- chull(clusterX, clusterY)
            polygon(clusterX[hull], clusterY[hull], lty = 1, lwd = 2,
                    border = palettes[[min(length(palettes), cl$n)]][clI])
            #border = "#54de25bb")
          }
        }
        if ("labelTrees" %in% input$display) {
          text(proj[, j], proj[, i], thinnedTrees())
        }
      }
    })
  }
  
  mode3D <- reactive("show3d" %in% input$display)
  PlotSize <- debounce(reactive(input$plotSize), 100)
  output$distPlot <- renderPlot({
    if (!mode3D()) {
      if (inherits(distances(), "dist")) {
        treespacePlot()
        output$mappingStatus <- renderText("")
      } else {
        output$mappingStatus <- renderText("No distances available.")
      }
    }
  }, width = PlotSize, height = PlotSize)
  
  .ThreeDPlotServer(
    input, output,
    distances   = distances,
    clusterings = clusterings,
    mapping     = mapping,
    mstEnds     = mstEnds,
    mstSize     = mstSize,
    pointCols   = pointCols,
    thinnedTrees = thinnedTrees,
    StrainCol   = StrainCol,
    PlotSize    = PlotSize,
    mode3D      = mode3D
  )
  
  output$savePng <- downloadHandler(
    filename = "TreeSpace.png",
    content = function(file) {
      png(file, width = input$plotSize, height = input$plotSize)
      treespacePlot()
      dev.off()
  })
  
  output$savePdf <- downloadHandler(
    filename = "TreeSpace.pdf",
    content = function(file) {
      pdf(file, title = paste0("Tree space mapping"))
      treespacePlot()
      dev.off()
  })
  
  output$saveCons <- downloadHandler(
    filename = "ClusterConsensusTrees.pdf",
    content = function(file) {
      cl <- clusterings()
      pdf(file, title = if (cl$sil > 0.25) {
        paste0("Consensus trees for ", cl$n, " clusters found with ", cl$method,
               " using ", chosenDistance())
      } else {
        paste0("Consensus of all trees (no meaningful clusters found using ",
               chosenDistance(), ")")
      },
      width = 8, height = consSize() / 100, pointsize = 10)
      PlotClusterCons()
      dev.off()
  })
  
  
  
  ##############################################################################
  # References
  ##############################################################################
  output$references <- renderUI({
    tags$div(style = paste0(
      "position: relative; top: ",
      (input$plotSize - 600) +
        if ("cons" %in% input$display) consSize() - 200 else 0,
      "px"),
    tagList(
      tags$h2("References for methods used"),
      #HTML(paste0("<h2>References", input$plotSize, "</h2>")),
      tags$h3("Tree space construction"),
      HTML(paste0(Smith2021,
                  Kaski2003, Venna2001, RCoreTeam)),
      HTML(if (mstSize() > 0) Gower1969),
      HTML(if(input$distance == "qd") SmithQuartet),
      tags$h3("Tree distance"),
      HTML(switch(input$distance,
             "cid" = paste0(Smith2020, SmithDist),
             "pid" = paste0(Smith2020, SmithDist),
             "qd" = paste0(Estabrook1985, Sand2014, SmithQuartet),
             "kc" = paste0(Kendall2016),
             "path" = paste0(Farris1973, Steel1993, SmithDist),
             "rf" = paste0(Robinson1981, Day1985, SmithDist))
      ),
      tags$h3("Mapping"),
      HTML(switch(input$mapping,
                  "pca" = Gower1966,
                  "k" = paste0(Kruskal1964, Venables2002),
                  "nls" = paste0(Sammon1969, Venables2002),
                  "umap" = McInnes2020
                  )
           ),
      tags$h3("Clustering"),
      HTML(paste(c(pam = paste0("Partitioning around medoids: ", Maechler2019),
        hmm = paste0("Hierarchical, minimax linkage: ", Bien2011, Murtagh1983),
        hsi = "",#paste0("Hierarchical, single linkage:"),
        hco = "",#paste0("Hierarchical, complete linkage:"),
        hav = "",#paste0("Hierarchical, average linkage:"),
        hmd = "",#paste0("Hierarchical, median linkage:"),
        hct = "",#paste0("Hierarchical, centroid linkage:"),
        hwd = paste0("Hierarchical, Ward d\ub2 linkage: ", Ward1963),
        kmn = paste0("K-means++:", Arthur2007),
        spec = ""#paste0("Spectral:")
        )[input$clustering])),
      HTML(paste("Cluster consensus trees:", Stockham2002))
    ))
  })
}

shinyApp(ui = ui, server = server)

## ----secret-header, echo=FALSE------------------------------------------------
set.seed(0)
protoclust <- if (requireNamespace("protoclust", quietly = TRUE)) {
  protoclust::protoclust
} else {
  hclust
}

## ----generate-trees-----------------------------------------------------------
library("TreeTools", quietly = TRUE)
treeNumbers <- c(1:220)
trees <- as.phylo(treeNumbers, 8)
spectrum <- hcl.colors(220, "plasma")
treeCols <- spectrum[treeNumbers]

## ----calculate-distances, message=FALSE---------------------------------------
library("TreeDist")
distances <- ClusteringInfoDistance(trees)

## ----map----------------------------------------------------------------------
mapping <- cmdscale(distances, k = 12)

## ----plot-mapping-2d, fig.asp = 1, fig.width = 3, fig.align="center"----------
par(mar = rep(0, 4))
plot(mapping,
     asp = 1, # Preserve aspect ratio - do not distort distances
     ann = FALSE, axes = FALSE, # Don't label axes: dimensions are meaningless
     col = treeCols, pch = 16
     )

## ----clustering, fig.align="center"-------------------------------------------
possibleClusters <- 2:10

# Partitioning around medoids
pamClusters <- lapply(possibleClusters, function(k) cluster::pam(distances, k = k))
pamSils <- vapply(pamClusters, function(pamCluster) {
  mean(cluster::silhouette(pamCluster)[, 3])
}, double(1))

bestPam <- which.max(pamSils)
pamSil <- pamSils[bestPam]
pamCluster <- pamClusters[[bestPam]]$cluster

# Hierarchical clustering
hTree <- protoclust(distances)
hClusters <- lapply(possibleClusters, function(k) cutree(hTree, k = k))
hSils <- vapply(hClusters, function(hCluster) {
  mean(cluster::silhouette(hCluster, distances)[, 3])
}, double(1))

bestH <- which.max(hSils)
hSil <- hSils[bestH]
hCluster <- hClusters[[bestH]]

# k-means++ clustering
kClusters <- lapply(possibleClusters, function(k) KMeansPP(distances, k = k))
kSils <- vapply(kClusters, function(kCluster) {
  mean(cluster::silhouette(kCluster$cluster, distances)[, 3])
}, double(1))

bestK <- which.max(kSils)
kSil <- kSils[bestK]
kCluster <- kClusters[[bestK]]$cluster

plot(pamSils ~ possibleClusters,
     xlab = "Number of clusters", ylab = "Silhouette coefficient",
     ylim = range(c(pamSils, hSils)))
points(hSils ~ possibleClusters, pch = 2, col = 2)
points(kSils ~ possibleClusters, pch = 3, col = 3)
legend("topright", c("PAM", "Hierarchical", "k-means++"),
       pch = 1:3, col = 1:3)

## ----chosen-cluster-----------------------------------------------------------
nClusters <- 2
whichResult <- match(nClusters, possibleClusters)
cluster <- hClusters[[whichResult]]

## ----h-tree, fig.align="center"-----------------------------------------------
class(hTree) <- "hclust"
par(mar = c(0, 0, 0, 0))
plot(hTree, labels = FALSE, main = "")
points(seq_along(trees), rep(1, length(trees)), pch = 16,
       col = spectrum[hTree$order])

## ----cluster-consensus, fig.align="center"------------------------------------
par(mfrow = c(1, 2), mar = rep(0.2, 4))
col1 <- spectrum[mean(treeNumbers[cluster == 1])]
col2 <- spectrum[mean(treeNumbers[cluster == 2])]
plot(consensus(trees[cluster == 1], p = 0.5),
     edge.color = col1, edge.width = 2, tip.color = col1)
plot(consensus(trees[cluster == 2], p = 0.5),
     edge.color = col2, edge.width = 2, tip.color = col2)

## ----island-id, fig.asp = 1, fig.width = 3, fig.align="center"----------------
par(mar = rep(0, 4))
# set a threshold corresponding to the width of the "moat" between islands
threshold <- 1.8
island <- Islands(distances, threshold)

# See how many trees are on each island
table(island)

# Let's ignore the small islands for now
largeIsle <- Islands(distances, threshold, smallest = 5)

# Colour trees according to their island
plot(mapping,
     asp = 1, # Preserve aspect ratio - do not distort distances
     ann = FALSE, axes = FALSE, # Don't label axes: dimensions are meaningless
     col = ifelse(is.na(largeIsle), "grey", largeIsle + 1),
     pch = 16
     )

## ----island-consensus, fig.align="center"-------------------------------------
par(mfrow = c(1, 2), mar = rep(0.2, 4))
plot(consensus(trees[!is.na(largeIsle) & largeIsle == 1], p = 0.5),
     edge.color = 2, edge.width = 2, tip.color = 2)
plot(consensus(trees[!is.na(largeIsle) & largeIsle == 2], p = 0.5),
     edge.color = 3, edge.width = 2, tip.color = 3)

## ----how-many-dims, fig.align="center"----------------------------------------
txc <- vapply(seq_len(ncol(mapping)), function(k) {
  newDist <- dist(mapping[, seq_len(k)])
  MappingQuality(distances, newDist, 10)["TxC"]
}, 0)
plot(txc, xlab = "Dimension")
abline(h = 0.9, lty = 2)

## ----calculate-MST------------------------------------------------------------
mstEnds <- MSTEdges(distances)

## ----plot-mapping-5d, fig.asp = 1, fig.align="center"-------------------------
nDim <- which.max(txc > 0.9)
plotSeq <- matrix(0, nDim, nDim)
plotSeq[upper.tri(plotSeq)] <- seq_len(nDim * (nDim - 1) / 2)
plotSeq <- t(plotSeq[-nDim, -1])
plotSeq[nDim * 1:3] <- (nDim * (nDim - 1) / 2) + 1:3
layout(plotSeq)
par(mar = rep(0.1, 4))

for (i in 2:nDim) for (j in seq_len(i - 1)) {
  # Set up blank plot
  plot(mapping[, j], mapping[, i], ann = FALSE, axes = FALSE, frame.plot = TRUE,
       type = "n", asp = 1, xlim = range(mapping), ylim = range(mapping))
  
  # Plot MST
  MSTSegments(mapping[, c(j, i)], mstEnds,
              col = StrainCol(distances, mapping[, c(j, i)]))
  
  # Add points
  points(mapping[, j], mapping[, i], pch = 16, col = treeCols)

  # Mark clusters
  for (clI in unique(cluster)) {
    inCluster <- cluster == clI
    clusterX <- mapping[inCluster, j]
    clusterY <- mapping[inCluster, i]
    hull <- chull(clusterX, clusterY)
    polygon(clusterX[hull], clusterY[hull], lty = 1, lwd = 2,
            border = "#54de25bb")
    text(mean(clusterX), mean(clusterY), clI, col = "#54de25bb", font = 2)
  }
}
# Annotate dimensions
plot(0, 0, type = "n", ann = FALSE, axes = FALSE)
text(0, 0, "Dimension 2")
plot(0, 0, type = "n", ann = FALSE, axes = FALSE)
text(0, 0, "Dimension 3")
plot(0, 0, type = "n", ann = FALSE, axes = FALSE)
text(0, 0, "Dimension 4")

## ----pid, fig.asp = 1, fig.width = 4, fig.align = "center", echo = FALSE, message = FALSE----
library("TreeDist")
pid_distances <- PhylogeneticInfoDistance(trees)
pid_mapping <- cmdscale(pid_distances, k = 6)
pid_cluster <- cutree(protoclust(pid_distances), k = 2)

par(mar = rep(0, 4))
plot(pid_mapping, ann = FALSE, axes = FALSE, asp = 1,
     col = treeCols, pch = 16)
MSTEdges(pid_distances, TRUE, pid_mapping[, 1], pid_mapping[, 2],
         col = "#bbbbbb", lty = 1)


pid_clusters <- seq_along(unique(pid_cluster))
for (clI in pid_clusters) {
  inCluster <- pid_cluster == clI
  clusterX <- pid_mapping[inCluster, 1]
  clusterY <- pid_mapping[inCluster, 2]
  hull <- chull(clusterX, clusterY)
  polygon(clusterX[hull], clusterY[hull], lty = 1, lwd = 2,
          border = "#54de25bb")
}

## ----hypervolume, message = FALSE---------------------------------------------
hypervolumeInstalled <- requireNamespace("hypervolume", quietly = TRUE)
if (hypervolumeInstalled) {
  library("hypervolume")
  hv1 <- hypervolume_gaussian(pid_mapping[pid_cluster == 1, 1:3],
                              verbose = FALSE)
  hv2 <- hypervolume_gaussian(pid_mapping[pid_cluster == 2, 1:3],
                              verbose = FALSE)
  hv_dist <- hypervolume_distance(hv1, hv2)
  capture.output(
    hyperset <- hypervolume_set(hv1, hv2, verbose = FALSE,
                              check.memory = FALSE)
  ) -> XX_VerboseNotRespected
  hv_overlap <- hypervolume_overlap_statistics(hyperset)
  hv_dist
  hv_overlap
} else {
  print("Install the 'hypervolume' package to run this example")
}


## ----sum-of-ranges------------------------------------------------------------
SumOfRanges(pid_mapping, pid_cluster)

## ----cluster-size-------------------------------------------------------------
SumOfVariances(pid_mapping, pid_cluster)
MeanCentroidDistance(pid_mapping, pid_cluster)
DistanceFromMedian(pid_mapping, pid_cluster)

## ----cluster-density----------------------------------------------------------
MeanNN(pid_mapping, pid_cluster)
MeanMSTEdge(pid_mapping, pid_cluster)

## ----umatrix, fig.asp = 1, fig.align = "center"-------------------------------
umatrixInstalled <- requireNamespace("Umatrix", quietly = TRUE)
if (umatrixInstalled) {
  map <- Umatrix::esomTrain(as.matrix(distances), Key = seq_along(trees),
                            Epochs = 5, # Increase for better results
                            Lines = 42,
                            Columns = 42,
                            Toroid = FALSE)
  
  Umatrix::plotMatrix(Matrix = map$Umatrix,
                      Toroid = FALSE, FixedRatio = TRUE,
                      TransparentContours = FALSE, Clean = TRUE) +
  ggplot2::geom_point(data = data.frame(x = map$BestMatches[, 3],
                                        y = map$BestMatches[, 2]),
                      shape = 19, color = treeCols, size = 2)

} else {
  message("Install the 'Umatrix' package to run this example")
}

## ----fig.width=6, out.width="90%", fig.align="center"-------------------------
library("TreeDist")
tree1 <- ape::read.tree(text = '(A, ((B, ((C, D), (E, F))), (G, (H, (I, J, K)))));')
tree2 <- ape::read.tree(text = '(A, (B, (C, D, E, (J, K)), (F, (G, H, I))));')
VisualizeMatching(NyeSimilarity, tree1, tree2,
                  Plot = TreeDistPlot, matchZeros = FALSE)

## -----------------------------------------------------------------------------
NyeSimilarity(tree1, tree2, normalize = FALSE) / 8
NyeSimilarity(tree1, tree2, normalize = 8)

## -----------------------------------------------------------------------------
NyeSimilarity(tree1, tree2,
                  normalize = min(TreeTools::NSplits(list(tree1, tree2))))

## -----------------------------------------------------------------------------
NyeSimilarity(tree1, tree2, normalize = min)

## -----------------------------------------------------------------------------
NyeSimilarity(list(tree1, tree2), list(tree1, tree2), normalize = pmin)

## -----------------------------------------------------------------------------
NyeSimilarity(tree1, tree2, normalize = TRUE)

## -----------------------------------------------------------------------------
TreeTools::NSplits(tree1)

## -----------------------------------------------------------------------------
NyeSimilarity(tree1, tree2, normalize = TreeTools::NSplits(tree1))

## -----------------------------------------------------------------------------
if (requireNamespace("Quartet", quietly = TRUE)) {
  library("Quartet", exclude = "RobinsonFoulds")
  expectedQD <- 2 / 3
  normalizedQD <- QuartetDivergence(QuartetStatus(tree1, tree2),
                                    similarity = FALSE) / expectedQD
}

## ----fig.width=7, fig.height=4, message=FALSE---------------------------------
if (requireNamespace("TreeDistData", quietly = TRUE)) {
  library("TreeDistData", exclude = "PairwiseDistances")
  data("randomTreeDistances", package = "TreeDistData")
  
  methods <- c("pid", "cid", "nye", "qd")
  methodCol <- c(pid = "#e15659", cid = "#58a14e", nye = "#edc949",
                 qd = "#af7aa1")
  
  oldPar <- par(cex = 0.7, mar = c(5, 5, 0.01, 0.01))
  nLeaves <- as.integer(dimnames(randomTreeDistances)[[3]])
  plot(nLeaves, type = "n", randomTreeDistances["pid", "mean", ],
       ylim = c(0.54, 1),
       xlab = "Number of leaves",
       ylab = "Normalized distance between random tree pairs")
  
  for (method in methods) {
    dat <- randomTreeDistances[method, , ]
    lines(nLeaves, dat["50%", ], pch = 1, col = methodCol[method])
    polygon(c(nLeaves, rev(nLeaves)), c(dat["25%", ], rev(dat["75%", ])),
            border = NA, col = paste0(methodCol[method], "55"))
    
  }
  
  text(202, randomTreeDistances[methods, "50%", "200"] + 0.02, 
       c("Different phylogenetic information", 
         "Clustering information distance",
         expression(paste(plain("Nye "), italic("et al."))),
         "Quartet divergence"
         ), col = methodCol[methods], pos = 2)
  par(oldPar)
}

## ----eval = FALSE-------------------------------------------------------------
# expectedCID <- randomTreeDistances["cid", "mean", "9"]
# ClusteringInfoDistance(tree1, tree2, normalize = TRUE) / expectedCID

## ----fig.align="center", fig.height=1.8, fig.width=6, out.width="80%"---------
testTrees <- list(
  trueTree = ape::read.tree(text = '(a, (b, (c, (d, (e, (f, (g, h)))))));'),
  lackRes = ape::read.tree(text = '(a, (b, c, (d, e, (f, g, h))));'),
  smallErr = ape::read.tree(text = '(a, (c, (b, (d, (f, (e, (g, h)))))));'),
  bigErr = ape::read.tree(text = '(a, (c, (((b, d), (f, h)), (e, g))));')
)
VisualizeMatching(MutualClusteringInfo, testTrees$trueTree, testTrees$lackRes)
points(4, 7.5, pch = 2, cex = 3, col = "#E69F00", xpd = NA)
VisualizeMatching(MutualClusteringInfo, testTrees$trueTree, testTrees$smallErr)
points(4, 7.5, pch = 3, cex = 3, col = "#56B4E9", xpd = NA)
VisualizeMatching(MutualClusteringInfo, testTrees$trueTree, testTrees$bigErr)
points(4, 7.5, pch = 4, cex = 3, col = "#009E73", xpd = NA)

## ----fig.width=4, fig.align="center", fig.asp=1, out.width="50%"--------------
if (requireNamespace("Ternary", quietly = TRUE)) {
  library("Ternary")
  oldPar <- par(mar = rep(0.1, 4))
  TernaryPlot(alab = "Absent information", blab = "Shared information",
              clab = "Misinformation",
              lab.cex = 0.8, lab.offset = 0.18,
              point = "left", clockwise = FALSE,
              grid.col = "#dedede", grid.minor.lines = 0,
              axis.labels = 0:10 / 10, axis.col = "#aaaaaa")
  
  HorizontalGrid()
  correct <- MutualClusteringInfo(testTrees$trueTree, testTrees)
  resolved <- ClusteringEntropy(testTrees)
  unresolved <- resolved["trueTree"] - resolved
  incorrect <- resolved - correct
  TernaryPoints(cbind(unresolved, correct, incorrect), 
                pch = 1:4, cex = 2, col = Ternary::cbPalette8[1:4])
  par(oldPar)
}

## -----------------------------------------------------------------------------
set.seed(0)
trueTree <- TreeTools::RandomTree(20, root = TRUE)

## -----------------------------------------------------------------------------
treeSearchInstalled <- requireNamespace("TreeSearch", quietly = TRUE)
if (treeSearchInstalled) {
  library("TreeSearch", quietly = TRUE) # for TBR, NNI
  oneAway <- structure(lapply(seq_len(200), function(x) {
    tbrTree <- TBR(trueTree)
    ape::consensus(list(tbrTree,
                        NNI(tbrTree),
                        NNI(tbrTree),
                        NNI(tbrTree)))
  }), class = "multiPhylo")
} else {
  message("Install \"TreeSearch\" to run this example")
}

## -----------------------------------------------------------------------------
if (treeSearchInstalled) {
  threeAway <- structure(lapply(seq_len(200), function(x) {
    tbrTree <- TBR(TBR(TBR(trueTree)))
    ape::consensus(list(tbrTree, 
                        NNI(NNI(tbrTree)),
                        NNI(NNI(tbrTree)),
                        NNI(NNI(tbrTree))))
  }), class = "multiPhylo")
}

## -----------------------------------------------------------------------------
if (treeSearchInstalled) {
  correct1 <- MutualClusteringInfo(trueTree, oneAway)
  correct3 <- MutualClusteringInfo(trueTree, threeAway)
}

## -----------------------------------------------------------------------------
if (treeSearchInstalled) {
  infoInTree1 <- ClusteringEntropy(oneAway)
  infoInTree3 <- ClusteringEntropy(threeAway)
}

## -----------------------------------------------------------------------------
if (treeSearchInstalled) {
  unresolved1 <- ClusteringEntropy(trueTree) - infoInTree1
  unresolved3 <- ClusteringEntropy(trueTree) - infoInTree3
}

## -----------------------------------------------------------------------------
if (treeSearchInstalled) {
  incorrect1 <- infoInTree1 - correct1
  incorrect3 <- infoInTree3 - correct3
}

## ----collapse=TRUE------------------------------------------------------------
col1 <- hcl(200, alpha = 0.9)
col3 <- hcl(40, alpha = 0.9)
spec1 <- matrix(col2rgb(col1, alpha = TRUE), nrow = 4, ncol = 181)
spec3 <- matrix(col2rgb(col3, alpha = TRUE), nrow = 4, ncol = 181)
spec1[4, ] <- spec3[4, ] <- 0:180
ColToHex <- function(x) rgb(x[1], x[2], x[3], x[4], maxColorValue = 255)
spec1 <- apply(spec1, 2, ColToHex)
spec3 <- apply(spec3, 2, ColToHex)

## ----fig.width=7, fig.align="center", fig.asp=5/7, out.width="70%"------------
if (treeSearchInstalled && requireNamespace("Ternary", quietly = TRUE)) {
  layout(matrix(c(1, 2), ncol = 2), widths = c(5, 2))
  oldPar <- par(mar = rep(0, 4))
  TernaryPlot(alab = "Information absent in degraded tree", 
              blab = "\n\nCorrect information in degraded tree", 
              clab = "Misinformation in degraded tree",
              point = "left", clockwise = FALSE, grid.minor.lines = 0,
              axis.labels = 0:10 / 10)
  
  HorizontalGrid()
  
  coords1 <- cbind(unresolved1, correct1, incorrect1)
  coords3 <- cbind(unresolved3, correct3, incorrect3)
  
  ColourTernary(TernaryDensity(coords1, resolution = 20), spectrum = spec1)
  ColourTernary(TernaryDensity(coords3, resolution = 20), spectrum = spec3)
  
  TernaryDensityContour(coords3, col = col3, nlevels = 4)
  TernaryDensityContour(coords1, col = col1, nlevels = 4)
  
  if (requireNamespace("kdensity", quietly = TRUE)) {
    library("kdensity")
    HorizontalKDE <- function(dat, col, add = FALSE) {
      lty <- 1
      lwd <- 2
      kde <- kdensity(dat)
      kdeRange <- kdensity:::get_range(kde)
      if (add) {
        lines(kde(kdeRange), kdeRange, col = col, lty = lty, lwd = lwd)
      } else {
        plot(kde(kdeRange), kdeRange, col = col, lty = lty, lwd = lwd, 
             ylim = c(0, 1), main = "", axes = FALSE, type = "l")
      }
      # abline(h = 0:10 / 10) # Useful for confirming alignment
    }
  
    par(mar = c(1.8, 0, 1.8, 0)) # align plot limits with ternary plot
    HorizontalKDE(correct1 / infoInTree1, col1, add = FALSE)
    HorizontalKDE(correct3 / infoInTree3, col3, add = TRUE)
    mtext("\u2192 Normalized tree quality \u2192", 2)
  }
  par(oldPar)
} else {
  message("Install \"TreeSearch\" and \"Ternary\" to generate this plot")
}

## ----initialize---------------------------------------------------------------
library("TreeTools", quietly = TRUE)
balAL <- BalancedTree(letters[1:12])
balAJ <- DropTip(balAL, letters[11:12])
balCL <- DropTip(balAL, letters[1:2])
balGL <- DropTip(balAL, letters[1:6])

pecAL <- PectinateTree(letters[1:12])
pecAF <- DropTip(pecAL, letters[7:12])
pecAJ <- DropTip(pecAL, letters[11:12])
pecCL <- DropTip(pecAL, letters[1:2])

treeList <- list(balAL = balAL, balAJ = balAJ, balCL = balCL, balGL = balGL,
                 pecAL = pecAL, pecAJ = pecAJ, pecCL = pecCL, pecAF = pecAF)

# Define a function to plot two trees
Plot2 <- function(t1, t2, ..., main2 = "") {
  oPar <- par(mfrow = c(1, 2), mar = rep(1, 4), cex = 0.9)
  on.exit(par(oPar)) # Restore original parameters
  plot(t1, ...)
  plot(t2, main = main2)
}

## ----drop-some-plot, fig.width = 8--------------------------------------------
Plot2(balAL, balCL,
      main = "balAL", main2 = "balCL",
      font = c(rep(4, 2), rep(3, 10)) # Emphasize missing tips
      )

## ----drop-some----------------------------------------------------------------
library("TreeDist")
commonTips <- intersect(TipLabels(balAL), TipLabels(balCL))

# How much information is in tree balCL?
ClusteringEntropy(balCL)
# All the information in balCL is also present in balAL
MutualClusteringInfo(balCL, balAL)
# balAL also contains information about leaves A and B,
# so contains more information in total
ClusteringEntropy(balAL)

## ----full-list----------------------------------------------------------------
# Information in common (bits)
raw <- MutualClusteringInfo(treeList)
heatmap(raw, symm = TRUE)
# Normalized
normalized <- MutualClusteringInfo(treeList, normalize = TRUE)
heatmap(normalized, symm = TRUE)

#' Shared information content of two splits
#' 
#' Calculate the phylogenetic information shared, or not shared, between two
#' splits.
#' See the [accompanying vignette](
#' https://ms609.github.io/TreeDist/articles/information.html)
#' for definitions.
#' 
#' 
#' Split _S1_ divides _n_ leaves into two splits, _A1_ and _B1_.
#' Split _S2_ divides the same leaves into the splits _A2_ and _B2_.
#' 
#' Splits must be named such that _A1_ fully overlaps with _A2_: 
#' that is to say, all taxa in _A1_ are also in _A2_, or _vice versa_.
#' Thus, all taxa in the smaller of _A1_ and _A2_ also occur in the larger.
#' 
#' @param n Integer specifying the number of leaves
#' @param A1,A2 Integers specifying the number of taxa in _A1_ and _A2_, 
#' once the splits have been arranged such that _A1_ fully overlaps with _A2_.
#' @return 
#' `TreesConsistentWithTwoSplits()` returns the number of unrooted bifurcating
#' trees consistent with two splits.
#' 
#' `SplitSharedInformation()` returns the phylogenetic information that two splits
#' have in common \insertCite{Meila2007}{TreeDist}, in bits.
#' 
#' `SplitDifferentInformation()` returns the amount of phylogenetic information
#' distinct to one of the two splits, in bits.
#' 
#' @examples 
#'   # Eight leaves, labelled A to H.
#'   # Split 1: ABCD|EFGH
#'   # Split 2: ABC|DEFGH
#'   # Let A1 = ABCD (four taxa), and A2 = ABC (three taxa).
#'   # A1 and A2 overlap (both contain ABC).
#'   
#'   TreesConsistentWithTwoSplits(n = 8, A1 = 4, A2 = 3)
#'   SplitSharedInformation(n = 8, A1 = 4, A2 = 3)
#'   SplitDifferentInformation(n = 8, A1 = 4, A2 = 3)
#'
#'   # If splits are identical, then their shared information is the same
#'   # as the information of either split:
#'   SplitSharedInformation(n = 8, A1 = 3, A2 = 3)
#'   TreeTools::SplitInformation(3, 5)
#' @template MRS
#'   
#' @references \insertAllCited{}
#' 
#' @family information functions
#' @importFrom TreeTools Log2TreesMatchingSplit Log2Unrooted
#' @export
SplitSharedInformation <- function(n, A1, A2 = A1) {
  Log2Unrooted(n) +
    Log2TreesConsistentWithTwoSplits(n, A1, A2) -
    Log2TreesMatchingSplit(A1, n - A1) -
    Log2TreesMatchingSplit(A2, n - A2)
}

#' @describeIn SplitSharedInformation Different information between two splits.
#' @importFrom TreeTools SplitInformation
#' @export
SplitDifferentInformation <- function(n, A1, A2 = A1) {
  Log2TreesMatchingSplit(A1, n - A1) +
    Log2TreesMatchingSplit(A2, n - A2) -
    (2 * Log2TreesConsistentWithTwoSplits(n, A1, A2))
  
}

#' Use variation of clustering information to compare pairs of splits
#' 
#' Compare a pair of splits viewed as clusterings of taxa, using the variation
#' of clustering information proposed by \insertCite{Meila2007}{TreeDist}.
#' 
#' This is equivalent to the mutual clustering information
#' \insertCite{Vinh2010}{TreeDist}.
#' For the total information content, multiply the VoI by the number of leaves.
#' 
#' @template split12Params
#' 
#' @return `MeilaVariationOfInformation()` returns the variation of (clustering)
#' information, measured in bits.
#' 
#' @references \insertAllCited{}
#' 
#' @examples 
#' # Maximum variation = information content of each split separately
#' A <- TRUE
#' B <- FALSE
#' MeilaVariationOfInformation(c(A, A, A, B, B, B), c(A, A, A, A, A, A))
#' Entropy(c(3, 3) / 6) + Entropy(c(0, 6) / 6)
#' 
#' # Minimum variation = 0
#' MeilaVariationOfInformation(c(A, A, A, B, B, B), c(A, A, A, B, B, B))
#' 
#' # Not always possible for two evenly-sized splits to reach maximum
#' # variation of information
#' Entropy(c(3, 3) / 6) * 2  # = 2
#' MeilaVariationOfInformation(c(A, A, A,B ,B, B), c(A, B, A, B, A, B)) # < 2
#' 
#' # Phylogenetically uninformative groupings contain spliting information
#' Entropy(c(1, 5) / 6)
#' MeilaVariationOfInformation(c(B, A, A, A, A, A), c(A, A, A, A, A, B))
#' @template MRS
#' 
#' @encoding UTF-8
#' @export
MeilaVariationOfInformation <- function(split1, split2) {
  if (length(split1) != length(split2)) stop("Split lengths differ")
  n <- length(split1)
  
  associationMatrix <- c(sum(split1 & split2), sum(split1 & !split2),
                         sum(!split1 & split2), sum(!split1 & !split2))
  probabilities <- associationMatrix / n
  p1 <- probabilities[[1]] + probabilities[[3]]
  p2 <- probabilities[[1]] + probabilities[[2]]
  
  jointEntropies <- Entropy(probabilities)
  
  # Return:
  jointEntropies + jointEntropies - Entropy(c(p1, 1 - p1)) - 
    Entropy(c(p2, 1 - p2))
}

#' @rdname MeilaVariationOfInformation
#' @return `MeilaMutualInformation()` returns the mutual information, 
#' measured in bits.
#' @export
MeilaMutualInformation <- function(split1, split2) {
  if (length(split1) != length(split2)) stop("Split lengths differ")
  n <- length(split1)
  
  associationMatrix <- c(sum(split1 & split2), sum(split1 & !split2),
                         sum(!split1 & split2), sum(!split1 & !split2))
  probabilities <- associationMatrix / n
  p1 <- probabilities[[1]] + probabilities[[3]]
  p2 <- probabilities[[1]] + probabilities[[2]]
  
  jointEntropies <- Entropy(probabilities)
  mutualInformation <- Entropy(c(p1, 1 - p1)) + Entropy(c(p2, 1 - p2)) -
    jointEntropies
  
  # Return:
  if (abs(mutualInformation) < .Machine[["double.eps"]]^0.5) 0 else mutualInformation
}

#' Variation of information for all split pairings
#' 
#' Calculate the variation of clustering information
#' \insertCite{Meila2007}{TreeDist} for each possible pairing of
#' non-trivial splits on _n_ leaves \insertCite{SmithDist}{TreeDist},
#' tabulating the number of pairings with each similarity.
#' 
#' @param n Integer specifying the number of leaves in a tree.
#' 
#' @return `AllSplitPairings()` returns a named vector. The name of each 
#' element corresponds to a certain variation of information, in bits; the
#' value of each element specifies the number of pairings of non-trivial
#' splits that give rise to that variation of information.
#' Split `AB|CD`  is treated as distinct from `CD|AB`.  If pairing
#' `AB|CD`=`CD|AB` is considered equivalent to `CD|AB`=`CD|AB` (etc), then
#' values should be divided by four.
#' 
#' @examples
#' AllSplitPairings(6)
#' # Treat equivalent splits as identical by dividing by four:
#' AllSplitPairings(6) / 4L
#' @template MRS
#' 
#' @references \insertAllCited{}
#' 
#' @encoding UTF-8
#' @export
AllSplitPairings <- local({
  # Simple memoization cache
  cache <- new.env(parent = emptyenv())
  
  function(n) {
    # Check cache first
    key <- as.character(n)
    if (exists(key, envir = cache)) {
      return(get(key, envir = cache))
    }
    
    # Compute result
    if (n < 4L) stop("No informative splits with < 4 taxa")
    
    # smallHalves <- 1L + seq_len(ceiling(n / 2) - 2L)
    dataRows <- 2L
  
    unevenPairs <- matrix(
      # For i in 2:largestSmallSplit
      #TODO: Make faster by not calculating bottom triangle
      unlist(lapply(1L + seq_len(n - 3L), function(inA) {
        # For j in 2:(n - 2)
        nCa <- choose(n, inA)
        outA <- n - inA
        hA <- Entropy(c(inA, outA) / n)
        unlist(lapply(1L + seq_len(n - 3L), function(inB) {
          outB <- n - inB
          hB <- Entropy(c(inB, outB) / n)
          vapply(max(0, inA + inB - n):min(inA, inB), function(inAB) {
            association <- c(inAB, inA - inAB, inB - inAB, n + inAB - inA - inB)
            jointEntropies <- Entropy(association / n)
            
            c(#inA, inB, inAB, 
              #npairs = NPartitionPairs(association), nis = choose(n, i),
              nTotal = nCa * choose(inA, inAB) * choose(outA, inB - inAB),
              VoI = jointEntropies + jointEntropies - hA - hB)
          }, double(dataRows))
        }))
      })), dataRows, dimnames=list(c("nTotal", "VoI"), NULL))
    
    result <- tapply(unevenPairs["nTotal", ], unevenPairs["VoI", ], sum)
    
    # Cache the result
    assign(key, result, envir = cache)
    
    return(result)
  }
})

#' Entropy of two splits
#' 
#' Calculate the entropy, joint entropy, entropy distance and information 
#' content of two splits, treating each split as a division of _n_ leaves into
#' two groups.
#' Further details are available in a 
#' [vignette](https://ms609.github.io/TreeDist/articles/information.html),
#' \insertCite{Mackay2003;textual}{TreeDist} and
#' \insertCite{Meila2007;textual}{TreeDist}.
#' 
#' @template split12Params
#' 
#' @return A numeric vector listing, in bits:
#'  * `H1` The entropy of split 1;
#'  * `H2` The entropy of split 2;
#'  * `H12` The joint entropy of both splits;
#'  * `I` The mutual information of the splits;
#'  * `Hd` The entropy distance (variation of information) of the splits.
#' 
#' @references \insertAllCited{}
#' 
#' @examples
#' A <- TRUE
#' B <- FALSE
#' SplitEntropy(c(A, A, A, B, B, B), c(A, A, B, B, B, B))
#' @template MRS
#' @encoding UTF-8
#' @family information functions
#' @export
SplitEntropy <- function(split1, split2 = split1) {
  A1A2 <- sum(split1 & split2)
  A1B2 <- sum(split1 & !split2)
  B1A2 <- sum(!split1 & split2)
  B1B2 <- sum(!split1 & !split2)
  overlaps <- c(A1A2, A1B2, B1A2, B1B2)
  
  A1 <- A1A2 + A1B2
  A2 <- A1A2 + B1A2
  B1 <- B1A2 + B1B2
  B2 <- A1B2 + B1B2
  n <- A1 + B1
  
  h1 <- Entropy(c(A1, B1) / n)
  h2 <- Entropy(c(A2, B2) / n)
  jointH <- Entropy(overlaps[overlaps > 0L] / n)
  sharedInformation <- h1 + h2 - jointH
  entropyDistance <- jointH - sharedInformation
  
  # Return:
  c(H1 = h1, H2 = h2, H12 = jointH, I = sharedInformation,
    Hd = entropyDistance)
}

#' @describeIn SplitSharedInformation Number of trees consistent with two 
#' splits.
#' @importFrom TreeTools TreesMatchingSplit NRooted
#' @export
TreesConsistentWithTwoSplits <- function(n, A1, A2 = A1) {
  
  smallSplit <- min(A1, A2)
  bigSplit <- max(A1, A2)
  
  if (smallSplit == 0) return (TreesMatchingSplit(bigSplit, n - bigSplit))
  if (bigSplit == n) return (TreesMatchingSplit(smallSplit, n - smallSplit))
  
  overlap <- bigSplit - smallSplit
  
  #  Here are two spits:
  #  AA OOO BBBBBB
  #  11 111 000000
  #  11 000 000000
  #  
  #  There are (2O - 5)!! unrooted trees of the overlapping taxa (O)
  #  There are (2A - 5)!! unrooted trees of A
  #  There are (2B - 5)!! unrooted trees of B
  #  
  #  There are 2A - 3 places on A that A can be attached to O.
  #  There are 2O - 3 places on O to which A can be attached.
  #  
  #  There are 2B - 3 places in B that B can be attached to (O+A).
  #  There are 2O - 3 + 2 places on O + A to which B can be attached:
  #   2O - 3 places on O, plus the two new edges created when A was joined to O.
  #  
  #  (2A - 3)(2A - 5)!! == (2A - 3)!!
  #  (2B - 3)(2B - 5)!! == (2B - 3)!!
  #  (2O - 1)(2O - 3)(2O - 5)!! == (2O - 1)!!
  #  
  #  We therefore want NRooted(A) * NRooted(O + 1) * NRooted(B)
  #  
  #  O = overlap = bigSplit - smallSplit
  #  bigSplit - overlap = smallSplit = either A or B
  #  n - bigSplit = either B or A
  
  
  # Return:
  NRooted(overlap + 1L) * 
    NRooted(smallSplit) *
    NRooted(n - bigSplit)
}

#' @describeIn SplitSharedInformation Natural logarithm of 
#' `TreesConsistentWithTwoSplits()`.
#' @importFrom TreeTools LnTreesMatchingSplit LnRooted.int
#' @export
LnTreesConsistentWithTwoSplits <- function(n, A1, A2 = A1) {
  smallSplit <- min(A1, A2)
  bigSplit <- max(A1, A2)
  
  # Return:
  if (smallSplit == 0) {
    LnTreesMatchingSplit(bigSplit, n - bigSplit)
  } else if (bigSplit == n) {
    LnTreesMatchingSplit(smallSplit, n - smallSplit)
  } else {
    LnRooted.int(bigSplit - smallSplit + 1L) + 
      LnRooted.int(smallSplit) + 
      LnRooted.int(n - bigSplit)
  }
}

#' @describeIn SplitSharedInformation Base two logarithm of 
#' `TreesConsistentWithTwoSplits()`.
#' @importFrom TreeTools Log2TreesMatchingSplit Log2Rooted.int
#' @export
Log2TreesConsistentWithTwoSplits <- function(n, A1, A2 = A1) {
  smallSplit <- min(A1, A2)
  bigSplit <- max(A1, A2)
  
  # Return:
  if (smallSplit == 0) {
    Log2TreesMatchingSplit(bigSplit, n - bigSplit)
  } else if (bigSplit == n) {
    Log2TreesMatchingSplit(smallSplit, n - smallSplit)
  } else {
    Log2Rooted.int(bigSplit - smallSplit + 1L) + 
      Log2Rooted.int(smallSplit) + 
      Log2Rooted.int(n - bigSplit)
  }
}

#' @describeIn SplitSharedInformation Base 2 logarithm of 
#' `TreesConsistentWithTwoSplits()`.
#' @importFrom TreeTools Log2TreesMatchingSplit Log2Rooted.int
#' @export
Log2TreesConsistentWithTwoSplits <- function(n, A1, A2 = A1) {
  smallSplit <- min(A1, A2)
  bigSplit <- max(A1, A2)
  
  # Return:
  if (smallSplit == 0) {
    Log2TreesMatchingSplit(bigSplit, n - bigSplit)
  } else if (bigSplit == n) {
    Log2TreesMatchingSplit(smallSplit, n - smallSplit)
  } else {
    Log2Rooted.int(bigSplit - smallSplit + 1L) + 
      Log2Rooted.int(smallSplit) + 
      Log2Rooted.int(n - bigSplit)
  }
}

library("TreeTools")

test_that("Entropy is calculated correctly", {
  expect_equal(Entropy(rep(0.5, 2)), 1)
  expect_equal(Entropy(c(1/4, 1/4, 0, 1/4, 0, 1/4)), 2)
  expect_equal(entropy_int(as.integer(c(1, 1, 0, 1, 1))), 2)
  expect_equal(Ntropy(c(0, 1, 1)), 1)
  expect_equal(Ntropy(c(0, 65536 + 1, 65536 + 1)), 1)
})

test_that("AllSplitPairings() counted correctly", {
  expect_error(AllSplitPairings(3))
  for (n in 4:10) {
    totalSplits <- sum(choose(n, 2:(n - 2)))
    expect_equal(totalSplits * totalSplits, sum(AllSplitPairings(n)))
  }
})

test_that("AllSplitPairings() cache lookup succeeds", {
  expect_equal(AllSplitPairings(12), AllSplitPairings(12))
})

test_that("Removing contradictions improves scores", {
  # Imagine 200 taxa, split 
  #   ...00000111111.....
  #   ...00011001111.....
  # We can delete taxa 99 and 100 to produce matching splits of
  # 198=100:98, or we can delete taxa 99-102 to produce 196=96:96.
  
  Test <- function(nTaxa, nContra) {
    nInSplit <- nTaxa / 2L
    split1 <- split2 <- c(rep(TRUE, nInSplit), rep(FALSE, nInSplit))
    flips <- nInSplit + ((1-nContra):nContra)
    nonFlips <- c(seq_len(nContra), nTaxa + 1 - seq_len(nContra))
    split2[flips] <- !split2[flips]
 
    expect_true(
      SharedPhylogeneticInfoSplits(as.Splits(split1[-flips]), as.Splits(split2[-flips]))
      >
      SharedPhylogeneticInfoSplits(as.Splits(split1[-nonFlips]), as.Splits(split2[-nonFlips]))
    )
  }
  
  Test(200, 2)
  Test(200, 1)
  Test(200, 10)
})


test_that("TreesConsistentWithTwoSplits works", {
  
  Test <- function(n, a, b, score) {
    logScore <- log(score)
    l2Score <- log2(score)
    
    expect_equal(TreesConsistentWithTwoSplits(n, a, b), score)
    expect_equal(TreesConsistentWithTwoSplits(n, b, a), score)
    expect_equal(TreesConsistentWithTwoSplits(n, n - a, n - b), score)
    expect_equal(TreesConsistentWithTwoSplits(n, n - b, n - a), score)
    expect_equal(Log2TreesConsistentWithTwoSplits(n, a, b), l2Score)
    expect_equal(Log2TreesConsistentWithTwoSplits(n, b, a), l2Score)
    expect_equal(Log2TreesConsistentWithTwoSplits(n, n - a, n - b), l2Score)
    expect_equal(Log2TreesConsistentWithTwoSplits(n, n - b, n - a), l2Score)
    expect_equal(LnTreesConsistentWithTwoSplits(n, a, b), logScore)
    expect_equal(LnTreesConsistentWithTwoSplits(n, b, a), logScore)
    expect_equal(LnTreesConsistentWithTwoSplits(n, n - a, n - b), logScore)
    expect_equal(LnTreesConsistentWithTwoSplits(n, n - b, n - a), logScore)
  }
  
  Test(8, 3, 0, 315)
  Test(8, 8, 0, NUnrooted(8))
  Test(8, 3, 3, TreesMatchingSplit(3, 5))
  Test(10, 5, 2, 1575)
  Test(9, 5, 3, 135)
  Test(8, 7, 3, 315)
})

test_that("MeilaMutualInformation", {
  expect_error(MeilaMutualInformation(c(T,T,T), c(T,T,T,T)))
  expect_equal(0, MeilaMutualInformation(c(T,T,T,F,F), c(F,F,F,F,F)))
  expect_equal(0.4199732, tolerance = 1e-6,
               MeilaMutualInformation(c(T,T,T,F,F), c(F,T,T,F,F)))
})

test_that("MeilaVariationOfInformation", {
  expect_equal(1L, MeilaVariationOfInformation(c(T,T,T,F,F,F), c(T,T,T,T,T,T)))
  expect_equal(0L, MeilaVariationOfInformation(c(T,T,T,F,F,F), c(T,T,T,F,F,F)))
  expect_equal(11.01955 / 6, MeilaVariationOfInformation(c(T,T,T,F,F,F), c(T,F,T,F,T,F)))
  expect_equal(7.219281 / 6, MeilaVariationOfInformation(c(F,T,T,T,T,T), c(T,T,T,T,T,F)))
})

test_that("SplitEntropy", {
  expect_equal(SplitEntropy(c(rep(TRUE, 5), rep(FALSE, 6)), c(rep(TRUE, 5), rep(FALSE, 6))),
               SplitEntropy(c(rep(TRUE, 5), rep(FALSE, 6))))
  expect_equal(c(H1 = 0.994, H2 = 0.994, H12 = 1.348, I = 0.6394, Hd = 0.709),
               SplitEntropy(c(rep(TRUE, 5), rep(FALSE, 6)), c(rep(TRUE, 6), rep(FALSE, 5))),
               tolerance = 0.001)
})

.onUnload <- function(libpath) {
  StopParallel(quietly = TRUE)
  library.dynam.unload("TreeDist", libpath)
}

## Reminders when releasing for CRAN
release_questions <- function() {
  c(
    "Is the code free of #TODOs?",
    "Have you updated inst/REFERENCES.bib with a full citation to Smith & Donoghue 2022?"
    )
}

# Additional steps:
#
# codemeta::write_codemeta()
# 
# Propagate changes in README.md to R/TreeDist-package.R

# Additional tests:
#
# run_examples()
# build_vignettes()
# build_manual() # PDF support for special characters

# # Unnecessary:
# # tools::resaveRdaFiles("R", compress="auto") - is default bzip2 the optimal?
# # tools::checkRdaFiles("R") - set optimal compression in `data-raw`

.Bercow <- function (x) order(order(x, method = "radix"), method = "radix")

.Triangle <- function (n) n * (n + 1) / 2

.TrustSum <- function(r, rHat, N, k) {
  sum(vapply(seq_len(N), function(i) {
    Uk <- rHat[, i] <= k & r[, i] > k
    sum(r[Uk, i] - k, na.rm = TRUE)
  }, double(1)))
}

.MMax <- function (N, k) {
  UkMax <- min(k, N - k)
  if (UkMax < 1) {
    warning("All points are nearest neighbours. Decrease `neighbours`.")
  }
  # Return:
  N * ((UkMax * (N - k - 1)) - .Triangle(UkMax - 1))
}

#' Faithfulness of mapped distances
#' 
#' `MappingQuality()` calculates the trustworthiness and continuity
#' of mapped distances \insertCite{Venna2001,Kaski2003}{TreeDist}.
#' Trustworthiness measures, on a scale from 0--1,
#' the degree to which points that are nearby in a mapping are truly close
#' neighbours; continuity, the extent to which points that are truly nearby 
#' retain their close spatial proximity in a mapping.
#'
#' 
#' @param original,mapped Square matrix or `dist` object containing 
#' original / mapped pairwise distances.
#' @param neighbours Integer specifying number of nearest neighbours to use in
#' calculation.  This should typically be small relative to the number of
#' points.
#' 
#' @return `MappingQuality()` returns a named vector of length four, 
#' containing the entries: `Trustworthiness`, `Continuity`, `TxC` 
#' (the product of these values), and `sqrtTxC` (its square root).
#' 
#' @examples
#' library("TreeTools", quietly = TRUE)
#' trees <- as.phylo(0:10, nTip = 8)
#' distances <- ClusteringInfoDistance(trees)
#' mapping <- cmdscale(distances)
#' MappingQuality(distances, dist(mapping), 4)
#' @template MRS
#' 
#' @references
#' \insertAllCited{}
#' @family tree space functions
#' @export
MappingQuality <- function(original, mapped, neighbours = 10L) {
  originalRank <- apply(as.matrix(original), 2, .Bercow) - 1
  mappedRank <- apply(as.matrix(mapped), 2, .Bercow) - 1
  diag(originalRank) <- diag(mappedRank) <- NA
  
  if (!identical(dim(originalRank), dim(mappedRank))) {
    stop("Original and mapped distances must have the same dimensions")
  }
  
  N <- dim(originalRank)[[2]]
  k <- neighbours
  MMax <- .MMax(N, k)
  
  trust <- 1 - (.TrustSum(originalRank, mappedRank, N, neighbours) / MMax)
  cont <- 1 - (.TrustSum(mappedRank, originalRank, N, neighbours) / MMax)
  txc <- trust * cont
  c("Trustworthiness" = trust,
    "Continuity" = cont,
    "TxC" = txc,
    "sqrtTxC" = sqrt(txc))
}

#' @rdname MappingQuality
#' @export
ProjectionQuality <- MappingQuality

library("TreeTools", quietly = TRUE)

test_that("MappingQuality() fails gracefully", {
  expect_error(ProjectionQuality(dist(1:10), dist(1:5)),
               "have the same dimensions")
})

test_that("Trustworthiness range", {
  
  for (N in c(5, 16, 50)) {
    k <- floor(seq(1, (N - 1) / 2, length.out = 5))
    UkMax <- pmin(k, N - k)
    maxVal <- N * ((UkMax * (N - k - 1)) - .Triangle(UkMax - 1))
    expect_equal(maxVal, (N * k * (2 * N - 3 * k - 1)) / 2)
  }
  
  trees <- as.phylo(1:12, 10)
  dists <- ClusteringInfoDist(trees)
  mapped <- dist(cmdscale(dists, k = 6))
  
  r <- apply(as.matrix(dists), 2, .Bercow) - 1
  diag(r) <- NA
  rHat <- apply(as.matrix(mapped), 2, .Bercow) - 1
  diag(r) <- NA
  
  rPrime <- 12 - r
  
  .Trustworthiness <- function (x, y, k) {
    MappingQuality(x, y, k)[["Trustworthiness"]]
  }
  
  expect_warning(expect_equal(.Trustworthiness(r, r, k = 12), NaN),
                 "All points are nearest neighbours")
  
  expect_equal(.Trustworthiness(r, r, k = 1), 1)
  expect_equal(.Trustworthiness(r, r, k = 5), 1)
  expect_equal(.Trustworthiness(r, r, k = 10), 1)
  
  expect_lte(.Trustworthiness(r, rHat, k = 1), 1)
  expect_lte(.Trustworthiness(r, rHat, k = 5), 1)
  expect_lte(.Trustworthiness(r, rHat, k = 10), 1)
  
  expect_gt(.Trustworthiness(r, rHat, k = 1), 0)
  expect_gt(.Trustworthiness(r, rHat, k = 5), 0)
  expect_gt(.Trustworthiness(r, rHat, k = 10), 0)
  
  expect_equal(.TrustSum(r, rPrime, 12, k = 1), .MMax(12, 1))
  expect_equal(.TrustSum(r, rPrime, 12, k = 5), .MMax(12, 5))
  expect_equal(.TrustSum(r, rPrime, 12, k = 10), .MMax(12, 10))
  
})

#' Generalized Robinson&ndash;Foulds distance
#' 
#' An internal function to calculate Generalized Robinson&ndash;Foulds
#' distances from splits.
#'
#' Note that no checks will be made to confirm that `splits1` and `splits2`
#' contain the same leaves in the same order.
#' This is the responsibility of the calling function.
#' 
#' @inheritParams SharedPhylogeneticInfoSplits
#' @param nTip Integer specifying the number of leaves in each split.
#' @param PairScorer function taking four arguments, `splits1`, `splits2`,
#' `nSplits1`, `nSplits2`, which should return the score of each pair of splits
#' in a two-dimensional matrix.  Additional parameters may be specified via 
#' \code{...}.
#' @param maximize Logical specifying whether the optimal matching maximizes
#' or minimizes the scores obtained by `PairScorer()`.
#' @param \dots Additional parameters to `PairScorer()`.
#' 
#' @return A numeric value specifying the score of the tree pairs under the 
#' specified pair scorer. If `reportMatching = TRUE`, attribute also list:
#' 
#' - `matching`: which split in `splits2` is optimally matched to each split in 
#'   `split1` (`NA` if not matched);
#' 
#' - `matchedSplits`: Textual representation of each match
#'
#' - `matchedScores`: Scores for matched split.
#' 
#' - `pairScores`: Calculated scores for each possible matching of each split.
#'
#' @keywords internal
#' @template MRS
#' @encoding UTF-8
#' @export
GeneralizedRF <- function(splits1, splits2, nTip, PairScorer, 
                           maximize, reportMatching, ...) {
  .ValidateSplitArgs(splits1, splits2, nTip)
  nSplits1 <- dim(splits1)[[1]]
  nSplits2 <- dim(splits2)[[1]]
  
  solution <- PairScorer(splits1, splits2, nTip, ...)
  ret <- solution[["score"]]
  
  if (reportMatching) {
    matching <- solution[["matching"]]
    matching[matching > nSplits2 | matching == 0L] <- NA
    if (nSplits1 < nSplits2) {
      matching <- matching[seq_len(nSplits1)]
    }
    attr(ret, "matching") <- matching
    
    # If reporting matching, we're not worried about performance
    pairScores <- matrix(0, nSplits1, nSplits2)
    for (i in seq_len(nSplits1)) {
      for (j in seq_len(nSplits2)) {
        pairScores[i, j] <- PairScorer(splits1[i, , drop = FALSE], 
                                       splits2[j, , drop = FALSE],
                                       nTip = nTip, ...)[["score"]]
      }
    }
    
    if (!is.null(attr(splits1, "tip.label"))) {
      matched1 <- !is.na(matching)
      matched2 <- matching[matched1]
      matched1 <- which(matched1)
      
      # Textual representation of matchings
      attr(ret, "matchedSplits") <- 
        ReportMatching(splits1[[matched1]],
                       splits2[[matched2]],
                       realMatch = if (maximize) {
                         pairScores[matrix(c(matched1, matched2), ncol = 2L)] > 0
                       } else rep(TRUE, length(matched1)))
    }
    
    attr(ret, "matchedScores") <- pairScores[
      matrix(c(seq_along(matching), matching), ncol = 2L)]
    
    attr(ret, "pairScores") <- pairScores
  }
  
  # Return:
  ret
}


#' @importFrom cli cli_progress_along
.MaxValue <- function(tree1, tree2, Value) {
  lab1 <- TipLabels(tree1)
  sameTips <- .AllTipsSame(lab1, TipLabels(tree2))
  if (!is.null(tree2)) {
    if (!sameTips) {
      trees <- .SharedOnly(tree1, tree2)
      tree1 <- trees[[1]]
      tree2 <- trees[[2]]
    }
  }
  
  if (is.null(tree2)) {
    if (sameTips) {
      value1 <- Value(tree1)
      # Much faster to discard unneeded than to only calculate required
      maxValue <- outer(value1, value1, "+")[, , drop = TRUE]
      maxValue[lower.tri(maxValue)]
    } else {
      # !sameTips implies that tree1 contains multiple trees
      pairs <- combn(seq_along(tree1), 2)
      nPairs <- dim(pairs)[[2]]
      
      vapply(cli_progress_along(seq_len(nPairs), "Calc max value"),
             function(pair) {
               i <- pairs[1, pair]
               j <- pairs[2, pair]
               common <- intersect(lab1[[i]], lab1[[j]])
               Value(KeepTip(tree1[[i]], common)) +
                 Value(KeepTip(tree1[[j]], common))
             }, double(1))
      
    }
  } else {
    value1 <- Value(tree1)
    if (sameTips) {
      outer(value1, Value(tree2), "+")[, , drop = TRUE]
    } else {
      value1 + Value(tree2)
    }
  }
}

# Fast path for *Distance() functions: computes pairwise distances and
# per-tree info in a single as.Splits() conversion.  Returns NULL when the
# fast path is not applicable.
#' @importFrom TreeTools as.Splits TipLabels
.FastDistPath <- function(tree1, tree2, reportMatching,
                          cpp_batch_fn, cpp_entropy_fn) {
  if (!is.null(tree2) || reportMatching) return(NULL)
  if (inherits(tree1, c("phylo", "Splits"))) return(NULL) # nocov
  if (!is.null(getOption("TreeDist-cluster"))) return(NULL)
  
  labs <- TipLabels(tree1)
  if (is.list(labs)) {
    # nocov start
    if (!all(vapply(labs[-1], setequal, logical(1), labs[[1]]))) return(NULL)
    tipLabels <- labs[[1]]
    # nocov end
  } else {
    tipLabels <- labs
  }
  nTip <- length(tipLabels)
  if (nTip < 4) return(NULL) # nocov
  .CheckMaxTips(nTip)
  
  splits_list <- as.Splits(tree1, tipLabels = tipLabels)
  n_threads <- as.integer(getOption("mc.cores", 1L))
  
  info_vec <- cpp_batch_fn(splits_list, as.integer(nTip), n_threads)
  entropies <- cpp_entropy_fn(splits_list, as.integer(nTip))
  
  list(
    info = structure(info_vec, class = "dist",
                     Size = length(tree1), Labels = names(tree1),
                     Diag = FALSE, Upper = FALSE),
    entropies = entropies
  )
}

# Fast path for cross-pairs (ManyMany): avoids duplicate as.Splits() calls by
# computing both pairwise distances and per-tree entropies in a single pass.
# Returns NULL when not applicable. When applicable, returns list with:
#   $dists: nA × nB matrix of pairwise distances
#   $info1: per-tree entropies for tree1 (length nA)
#   $info2: per-tree entropies for tree2 (length nB)
#' @importFrom TreeTools as.Splits TipLabels
.FastManyManyPath <- function(tree1, tree2, reportMatching,
                              cpp_cross_pairs_fn, cpp_entropy_fn) {
  if (is.null(tree2) || reportMatching) return(NULL)
  if (inherits(tree1, c("phylo", "Splits")) || inherits(tree2, c("phylo", "Splits"))) {
    return(NULL) # nocov
  }
  if (!is.null(getOption("TreeDist-cluster"))) return(NULL)
  
  lab1 <- TipLabels(tree1)
  lab2 <- TipLabels(tree2)
  
  # Check tip label agreement
  if (is.list(lab1)) {
    if (!all(vapply(lab1[-1], setequal, logical(1), lab1[[1]]))) return(NULL) # nocov
    tipLabels1 <- lab1[[1]] # nocov
  } else {
    tipLabels1 <- lab1
  }
  
  if (is.list(lab2)) {
    if (!all(vapply(lab2[-1], setequal, logical(1), lab2[[1]]))) return(NULL) # nocov
    tipLabels2 <- lab2[[1]] # nocov
  } else {
    tipLabels2 <- lab2
  }
  
  # Only use fast path if both collections have the same tip set
  if (!setequal(tipLabels1, tipLabels2)) return(NULL)
  
  nTip <- length(tipLabels1)
  if (nTip < 4) return(NULL)
  .CheckMaxTips(nTip)
  
  splits1 <- as.Splits(tree1, tipLabels = tipLabels1)
  splits2 <- as.Splits(tree2, tipLabels = tipLabels1)  # Use tipLabels1 to ensure order consistency
  n_threads <- as.integer(getOption("mc.cores", 1L))
  
  dists <- cpp_cross_pairs_fn(splits1, splits2, as.integer(nTip), n_threads)
  info1 <- cpp_entropy_fn(splits1, as.integer(nTip))
  info2 <- cpp_entropy_fn(splits2, as.integer(nTip))
  
  # Add row/column names to the distance matrix
  rownames(dists) <- names(tree1)
  colnames(dists) <- names(tree2)
  
  list(
    dists = dists,
    info1 = info1,
    info2 = info2
  )
}

# Lower-tri pairwise sums: outer(x, x, "+")[lower.tri(.)]
.PairwiseSums <- function(x) {
  g <- outer(x, x, "+")
  g[lower.tri(g)]
}

# Floor sub-noise distances to zero before normalization.
# Two sources of numerical noise scale with treesIndependentInfo:
#  (1) LAP int64 cost-matrix quantization in *Splits scoring; per-cell
#      truncation of up to (max_possible / BIG) bits, summed over n_splits.
#  (2) Float-accumulation drift between independently-built tables (e.g.
#      InfoRobinsonFoulds vs cpp_splitwise_info_batch sum the same per-split
#      info contributions, but using different lookup-table constructions).
# Both grow with the magnitude of the answer, so an absolute sqrt(eps)
# tolerance becomes too tight beyond a few thousand tips. Scaling by
# treesIndependentInfo self-adjusts; pmax(1, ·) preserves the original
# tolerance for tiny trees where these errors are negligible anyway.
.FloorNumericalNoise <- function(ret, treesIndependentInfo) {
  tol <- pmax(1, treesIndependentInfo) * .Machine[["double.eps"]] ^ 0.5
  ret[ret < tol] <- 0
  ret
}

.AllTipsSame <- function(x, y) {
  if (is.list(x)) {
    xPrime <- x[[1]]
    if (length(x) > 1 && !all(vapply(x[-1], setequal, logical(1), xPrime))) {
      return(FALSE)
    }
  } else {
    xPrime <- x
  }
  if (is.null(y)) {
    TRUE
  } else {
    if (is.list(y)) {
      all(vapply(y, setequal, logical(1), xPrime))
    } else {
      setequal(xPrime, y)
    }
  }
}

#' Are splits compatible?
#' 
#' Determine whether splits are compatible (concave); i.e. they can both occur
#' on a single tree.
#' 
#' @template split12Params
#' @return `SplitsCompatible()` returns a logical specifying whether the splits
#' provided are compatible with one another.
#' 
#' @examples 
#' A <- TRUE
#' B <- FALSE
#' SplitsCompatible(c(A, A, A, B, B, B),
#'                  c(A, A, B, B, B, B))
#' SplitsCompatible(c(A, A, A, B, B, B),
#'                  c(A, A, B, B, B, A))
#' @template MRS
#' @export
SplitsCompatible <- function(split1, split2) {
  # Return:
  (
    all (split1[split2]) ||
    all(split1[!split2]) ||
    all(!split1[split2]) ||
    all(!split1[!split2])
  )
}

#' Calculate distances in parallel
#' 
#' Accelerate distance calculation by employing multiple \acronym{CPU} workers.
#' 
#' ## OpenMP (recommended for all split-based metrics)
#' 
#' When the package is built with \acronym{OpenMP} support (the default on
#' Linux and Windows; optional on macOS), all pairwise split-based distance
#' calculations use an efficient multi-threaded batch path automatically —
#' no cluster setup is required.  The affected functions are:
#' 
#' - [`ClusteringInfoDistance()`] / [`MutualClusteringInfo()`]
#' - [`SharedPhylogeneticInfo()`] / [`DifferentPhylogeneticInfo()`]
#' - [`MatchingSplitInfo()`] / [`MatchingSplitInfoDistance()`]
#' - [`MatchingSplitDistance()`]
#' - [`InfoRobinsonFoulds()`]
#' - [`NyeSimilarity()`]
#' - [`JaccardRobinsonFoulds()`]
#' 
#' The number of \acronym{OpenMP} threads is controlled by the standard
#' `"mc.cores"` option:
#' 
#' ```r
#' options(mc.cores = parallel::detectCores())  # use all available cores
#' options(mc.cores = 4L)                        # or a fixed number
#' ```
#' 
#' The default is `1` (single-threaded).
#' 
#' ## R parallel cluster
#' 
#' `StartParallel()` creates an R socket cluster (via [`makeCluster()`]) and
#' registers it for use by TreeDist.  `SetParallel()` registers a pre-existing
#' cluster.  `StopParallel()` stops the cluster and releases resources.
#' 
#' **When to use `StartParallel()`:** for metrics that do not have an
#' \acronym{OpenMP} batch path, namely tree-object-based distances such as
#' [`NNIDist()`] and [`MASTSize()`] / [`MASTInfo()`], or any function called
#' via [`CompareAll()`].  R-cluster parallelism carries a serialisation overhead
#' of ~2–3 s, so it is only beneficial for large problems.
#' 
#' **When _not_ to use `StartParallel()`:** for the split-based metrics listed
#' above.  Registering a cluster disables the \acronym{OpenMP} batch path for
#' those functions, replacing a thread-local C++ loop with inter-process
#' communication — which is slower at every problem size measured.  Call
#' `StopParallel()` before computing split-based distances if a cluster is
#' active.
#' 
#' @param \dots Parameters to pass to [`makeCluster()`].
#' @param cl An existing cluster.
#' 
#' @examples
#' # OpenMP parallelism: set mc.cores before calling any split-based metric.
#' options(mc.cores = 2L)
#' # MutualClusteringInfo(trees)  # uses 2 OpenMP threads automatically
#' options(mc.cores = NULL)  # restore default (single-threaded)
#' 
#' if (interactive()) {
#'   # R cluster: beneficial for NNIDist, MASTSize/MASTInfo, CompareAll(), etc.
#'   # Do NOT activate while computing split-based distances (MCI, SPI, MSI, …)
#'   # as it bypasses the faster OpenMP path.
#'   library("TreeTools", quietly = TRUE)
#'   nCores <- ceiling(parallel::detectCores() / 2)
#'   StartParallel(nCores) # Takes a few seconds to set up processes
#'   GetParallel()
#'   CompareAll(as.phylo(0:6, 100), NNIDist)
#'   StopParallel() # Returns system resources
#' }
#' @template MRS
#' @importFrom parallel makeCluster
#' @importFrom cli cli_alert_success cli_alert_danger
#' @export
StartParallel <- function(...) {
  cl <- makeCluster(...)
  cli_alert_success("Started cluster")
  options("TreeDist-cluster" = cl)
}

#' @rdname StartParallel
#' @return `StartParallel()` and `SetParallel()` return the previous value of
#' `options("TreeDist-cluster")`.
#' @export
SetParallel <- function(cl) {
  options("TreeDist-cluster" = cl)
}

#' @rdname StartParallel
#' @importFrom cli cli_alert_info
#' @return `GetParallel()` returns the currently specified cluster.
#' @export
GetParallel <- function(cl) {
  ret <- getOption("TreeDist-cluster")
  if (is.null(ret)) {
    cli_alert_info("No cluster currently specified")
  }
  ret
}

#' @rdname StartParallel
#' @param quietly Logical; if `TRUE`, do not warn when no cluster was running.
#' @export
#' @importFrom parallel stopCluster
#' @return `StopParallel()` returns `TRUE` if a cluster was destroyed,
#' `FALSE` otherwise.
StopParallel <- function(quietly = FALSE) {
  cluster <- getOption("TreeDist-cluster")
  if (!is.null(cluster)) {
    stopCluster(cluster)
    cli_alert_success("Cluster destroyed")
    options("TreeDist-cluster" = NULL)
    TRUE
  } else {
    if (!quietly) {
      cli_alert_danger("No cluster running")
    }
    FALSE
  }
}

test_that("Parallelization works", {
  library("TreeTools")
  trees <- as.phylo(0:20, 20)
  
  suppressMessages({
  expect_equal(options("TreeDist-cluster"), StartParallel(2))
  cl <- getOption("TreeDist-cluster")
  
  parallel <- ClusteringInfoDistance(trees)
  SetParallel(NULL)
  expect_null(GetParallel())
  expect_equal(ClusteringInfoDistance(trees), parallel)

  expect_equal(options("TreeDist-cluster"), SetParallel(cl))
  expect_equal(cl, getOption("TreeDist-cluster"))
  expect_equal(cl, GetParallel())
  parallel <- CompareAll(trees, NNIDist)
  SetParallel(NULL)
  expect_equal(CompareAll(trees, NNIDist), parallel)
  expect_false(StopParallel())
  SetParallel(cl)
  expect_true(StopParallel())
    })
})

#' Nye _et al_. (2006) tree comparison
#' 
#' `NyeSimilarity()` and `NyeSplitSimilarity()` implement the 
#' [Generalized Robinson&ndash;Foulds](https://ms609.github.io/TreeDist/articles/Robinson-Foulds.html#generalized-robinson-foulds-distances)
#' tree comparison metric of \insertCite{Nye2006;textual}{TreeDist}.
#' In short, this finds the optimal matching that pairs each branch from
#' one tree with a branch in the second, where matchings are scored according to
#' the size of the largest split that is consistent with both of them,
#' normalized against the Jaccard index.
#' A more detailed account is available in the 
#' [vignettes](https://ms609.github.io/TreeDist/articles/Generalized-RF.html#nye-et-al--tree-similarity-metric).
#' 
#' The measure is defined as a similarity score.  If `similarity = FALSE`, the
#' similarity score will be converted into a distance by doubling it and
#' subtracting it from the number of splits present in both trees.
#' This ensures consistency with `JaccardRobinsonFoulds`.
#' 
#' Note that `NyeSimilarity(tree1, tree2)` is equivalent to, but 
#' slightly faster than, \code{\link{JaccardRobinsonFoulds}
#' (tree1, tree2, k = 1, allowConflict = TRUE)}.
#'  
#' @inheritParams RobinsonFoulds
#' @param normalizeMax When calculating similarity, normalize against the 
#' maximum number of splits that could have been present (`TRUE`),
#'  or the number of splits that were actually observed (`FALSE`)?  
#' Defaults to the number of splits in the better-resolved tree; set
#'  `normalize = pmin.int` to use the number of splits in the less resolved
#'  tree.
#' 
#' @section Normalization:
#' 
#' If `normalize = TRUE` and `similarity = TRUE`, then results will be rescaled
#' from zero to one by dividing by the mean number of splits in the two trees
#' being compared.
#'  
#' You may wish to normalize instead against the number of splits present
#' in the smaller tree, which represents the maximum value possible for a pair
#' of trees with the specified topologies (`normalize = pmin.int`); the
#' number of splits in the most resolved tree (`normalize = pmax.int`);
#' or the maximum value possible for any pair of trees with  _n_ leaves, 
#' _n_ - 3 (`normalize = TreeTools::NTip(tree1) - 3L`).
#' 
#' If `normalize = TRUE` and `similarity = FALSE`, then results will be rescaled
#' from zero to one by dividing by the total number of splits in the pair
#' of trees being considered.
#' 
#' Trees need not contain identical leaves; scores are based on the leaves that
#' trees hold in common.  Check for unexpected differences in tip labelling
#' with `setdiff(TipLabels(tree1), TipLabels(tree2))`.
#' 
#' @templateVar returns `NyeSimilarity()` returns
#' @template distReturn
#' 
#' @examples 
#' library("TreeTools")
#' NyeSimilarity(BalancedTree(8), PectinateTree(8))
#' VisualizeMatching(NyeSimilarity, BalancedTree(8), PectinateTree(8))

#' NyeSimilarity(as.phylo(0:5, nTip = 8), PectinateTree(8))
#' NyeSimilarity(as.phylo(0:5, nTip = 8), similarity = FALSE)
#' @template MRS
#' 
#' @references \insertAllCited{}
#' @family tree distances
#' 
#' @encoding UTF-8
#' @importFrom TreeTools NSplits SplitsInBinaryTree
#' @export
NyeSimilarity <- function(tree1, tree2 = NULL, similarity = TRUE,
                           normalize = FALSE,
                           normalizeMax = !is.logical(normalize),
                           reportMatching = FALSE,
                           diag = TRUE) {
  unnormalized <- CalculateTreeDistance(NyeSplitSimilarity, tree1, tree2, 
                                        reportMatching)
  if (similarity) {
    InfoInTree <- if (normalizeMax) SplitsInBinaryTree else NSplits
    if (diag && is.null(tree2)) {
      unnormalized <- as.matrix(unnormalized)
      diag(unnormalized) <- InfoInTree(tree1)
      tree2 <- tree1
    }
    
    # Return:
    NormalizeInfo(unnormalized, tree1, tree2, how = normalize,
                  InfoInTree = InfoInTree, Combine = .MeanOfTwo)
  } else {
    unnormalized <- .MaxValue(tree1, tree2, NSplits) - 
      (unnormalized + unnormalized)
    
    # Return:
    NormalizeInfo(unnormalized, tree1, tree2, how = normalize,
                  InfoInTree = NSplits, Combine = "+")
  }
}

.MeanOfTwo <- function(x, y) (x + y) / 2L

#' @rdname NyeSimilarity
#' @inheritParams SharedPhylogeneticInfoSplits
#' @export
NyeSplitSimilarity <- function(splits1, splits2, 
                                nTip = attr(splits1, "nTip"),
                                reportMatching = FALSE) {
  GeneralizedRF(splits1, splits2, nTip, cpp_jaccard_similarity, k = 1L,
                allowConflict = TRUE, maximize = TRUE,
                reportMatching = reportMatching)
}

#' Jaccard&ndash;Robinson&ndash;Foulds metric
#' 
#' Calculate the 
#' [Jaccard&ndash;Robinson&ndash;Foulds metric](https://ms609.github.io/TreeDist/articles/Generalized-RF.html#jaccard-robinson-foulds-metric)
#' \insertCite{Bocker2013}{TreeDist}, a 
#' [Generalized Robinson&ndash;Foulds metric](https://ms609.github.io/TreeDist/articles/Robinson-Foulds.html#generalized-robinson-foulds-distances).
#' 
#' In short, the Jaccard&ndash;Robinson&ndash;Foulds
#' metric is a generalized Robinson-Foulds
#' metric: it finds the optimal matching that pairs each split in one tree with
#' a similar split in the second.
#' Matchings are scored according to the size of the largest split that is 
#' consistent with both of them, normalized against the Jaccard index, and 
#' raised to an arbitrary exponent. 
#' A more detailed explanation is provided in the 
#' [vignettes](https://ms609.github.io/TreeDist/articles/Generalized-RF.html#jaccard-robinson-foulds-metric).
#' 
#' By default, conflicting splits may be paired. 
#' 
#' Note that the settings `k = 1, allowConflict = TRUE, similarity = TRUE`
#' give the similarity metric of \insertCite{Nye2006;textual}{TreeDist};
#' a slightly faster implementation of this metric is available as
#' [`NyeSimilarity()`].
#' 
#' The examples section below details how to visualize matchings with 
#' non-default parameter values.
#' 
#' Trees need not contain identical leaves; scores are based on the leaves that
#' trees hold in common.  Check for unexpected differences in tip labelling
#' with `setdiff(TipLabels(tree1), TipLabels(tree2))`.
#' 
#' @inheritParams RobinsonFoulds
#' @param k An arbitrary exponent to which to raise the Jaccard index.
#' Integer values greater than one are anticipated by B&ouml;cker _et al_.
#' The Nye _et al_. metric uses `k = 1`.
#' As k increases towards infinity, the metric converges to the Robinson&ndash;Foulds
#' metric.
#' @param allowConflict Logical specifying whether to allow conflicting splits
#' to be paired. If `FALSE`, such pairings will be allocated a similarity
#' score of zero.
#' 
#' @section Normalization:
#' 
#' If `normalize = TRUE`, then results will be rescaled from zero to one by
#' dividing by the maximum possible value for trees of the given topologies,
#' which is equal to the sum of the number of splits in each tree. 
#' You may wish to normalize instead against the maximum number of splits
#' present in a pair of trees with _n_ leaves, by specifying 
#' `normalize = n - 3`.
#' 
#' @templateVar returns `JaccardRobinsonFoulds()` returns
#' @template distReturn
#' 
#' @references \insertAllCited{}
#' 
#' @examples 
#' set.seed(2)
#' tree1 <- ape::rtree(10)
#' tree2 <- ape::rtree(10)
#' JaccardRobinsonFoulds(tree1, tree2, k = 2, allowConflict = FALSE)
#' JaccardRobinsonFoulds(tree1, tree2, k = 2, allowConflict = TRUE)
#' 
#' JRF2 <- function(tree1, tree2, ...) 
#'   JaccardRobinsonFoulds(tree1, tree2, k = 2, allowConflict = FALSE, ...)
#'   
#' VisualizeMatching(JRF2, tree1, tree2, matchZeros = FALSE)
#' @template MRS
#' 
#' @family tree distances
#' 
#' @encoding UTF-8
#' @importFrom TreeTools NSplits
#' @export
JaccardRobinsonFoulds <- function(tree1, tree2 = NULL, k = 1L, 
                                   allowConflict = TRUE, similarity = FALSE,
                                   normalize = FALSE, reportMatching = FALSE) {
  unnormalized <- CalculateTreeDistance(JaccardSplitSimilarity, tree1, tree2, 
                                        k = k, allowConflict = allowConflict, 
                                        reportMatching = reportMatching) * 2L
  
  if (!similarity) {
    unnormalized <- .MaxValue(tree1, tree2, NSplits) - unnormalized
  }

  NormalizeInfo(unnormalized, tree1, tree2, how = normalize,
                InfoInTree = NSplits, Combine = "+")
}

#' @rdname JaccardRobinsonFoulds
#' @inheritParams SharedPhylogeneticInfoSplits
#' @export
JaccardSplitSimilarity <- function(splits1, splits2,
                                    nTip = attr(splits1, "nTip"),
                                    k = 1L, allowConflict = TRUE,
                                    reportMatching = FALSE) {
  GeneralizedRF(splits1, splits2, nTip, cpp_jaccard_similarity, k = k,
                allowConflict = allowConflict, maximize = TRUE,
                reportMatching = reportMatching)
}

library("TreeTools")

test_that("VisualizeMatching() works", {
  tree1 <- PectinateTree(as.character(1:11))
  tree2 <- tree1
  tree2$tip.label[c(11, 1)] <- tree1$tip.label[c(1, 11)]
  tree2r <- CollapseNode(tree2, 20:21)
  
  Minus <- function(...) {
    x <- MutualClusteringInfo(...)
    attr(x, "pairScores") <- -attr(x, "pairScores")
    x
  }
  expect_error(VisualizeMatching(Minus, PectinateTree(8), BalancedTree(8), 
                                 setPar = FALSE))
  
  skip_if_not_installed("vdiffr")
  skip_if(packageVersion("graphics") < "4.3")
  skip_if(packageVersion("vdiffr") < "1.0")
  
  vdiffr::expect_doppelganger("Test VM", TestVM <- function() {
    VisualizeMatching(MutualClusteringInfo, tree1, tree2, 
                      setPar = TRUE, precision = 3, matchZeros = FALSE,
                      Plot = plot.phylo) -> x
    expect_equal(sum(attr(x, "matchedScores")), x[[1]])
  })
  
  vdiffr::expect_doppelganger("Test VMr", function() {
    VisualizeMatching(MutualClusteringInfo, tree1, tree2r,
                      setPar = TRUE, precision = 3, matchZeros = TRUE, 
                      Plot = plot.phylo, cex = 1.5)
  })
  
  vdiffr::expect_doppelganger("Visualize MCI matching", function() {
    par(mfrow = c(2, 2), mar = rep(0.1, 4), cex = 1.5)
    tree1 <- ape::read.tree(text="((1, 2), ((3, (4, 5)), (6, (7, (8, 9)))));")
    tree2 <- ape::read.tree(text="((1, 2), ((3, 4, (5, 9)), (6, (7, 8))));")
    VisualizeMatching(MutualClusteringInfo, tree1, tree2,
                      setPar = FALSE, precision = 3L,
                      Plot = TreeDistPlot,
                      matchZeros = FALSE,
                      leaveRoom = FALSE)
    VisualizeMatching(MutualClusteringInfo, tree2, tree1,
                      setPar = FALSE, precision = 3,
                      Plot = TreeDistPlot,
                      matchZeros = FALSE,
                      leaveRoom = FALSE)
  })
  
  vdiffr::expect_doppelganger("RF: Collapse a node", function() {
    par(mfrow = c(2, 2), mar = rep(0.1, 4), cex = 1.5)
    tree1 <- ape::read.tree(text="((1, 2), ((6, (7, 8)), (3, 4, (5, 9))));")
    tree2 <- ape::read.tree(text="((1, 2), ((3, (4, (5, 9))), (6, (7, 8))));")
    VisualizeMatching(RobinsonFouldsMatching, tree1, tree2,
                      setPar = FALSE, precision = 3,
                      edge.cex = 1,
                      Plot = TreeDistPlot,
                      matchZeros = TRUE,
                      leaveRoom = FALSE)
    VisualizeMatching(RobinsonFouldsMatching, tree2, tree1,
                      setPar = FALSE, precision = 3,
                      Plot = TreeDistPlot,
                      matchZeros = FALSE,
                      leaveRoom = FALSE)
  })
  
  vdiffr::expect_doppelganger("Hidden edge labels", function() {
    par(mfrow = c(2, 2), mar = rep(0.1, 4), cex = 1.2)
    tree1 <- ape::read.tree(text="((1, 2), ((6, (7, 8)), (3, 4, (5, 9))));")
    tree2 <- ape::read.tree(text="((1, 2), ((3, (4, (5, 9))), (6, (7, 8))));")
    VisualizeMatching(RobinsonFouldsMatching, tree1, tree2,
                      setPar = FALSE, precision = 3,
                      edge.cex = 0,
                      value.cex = 2,
                      Plot = TreeDistPlot,
                      matchZeros = TRUE,
                      leaveRoom = FALSE)
    VisualizeMatching(RobinsonFouldsMatching, tree1, tree2,
                      setPar = FALSE, precision = 3,
                      edge.cex = 2,
                      value.cex = FALSE,
                      Plot = TreeDistPlot,
                      matchZeros = TRUE,
                      leaveRoom = FALSE)
  })
  
  
  vdiffr::expect_doppelganger("RF: Collapse and change", function() {
    par(mfrow = c(2, 2), mar = rep(0.1, 4), cex = 1.5)
    tree1 <- ape::read.tree(text="((1, 2), ((6, (7, 8)), (3, 4, (5, 9))));")
    tree2 <- ape::read.tree(text="((1, 2), ((3, (4, (5, 9))), ((6, 7), 8)));")
    VisualizeMatching(RobinsonFouldsMatching, tree1, tree2,
                      setPar = FALSE, precision = 3L,
                      Plot = TreeDistPlot,
                      matchZeros = TRUE,
                      leaveRoom = FALSE)
    VisualizeMatching(RobinsonFouldsMatching, tree2, tree1,
                      setPar = FALSE, precision = 3L,
                      Plot = TreeDistPlot,
                      matchZeros = FALSE,
                      leaveRoom = FALSE)
  })
  
  vdiffr::expect_doppelganger("RF VM Single splits; plainEdges", function() {
    par(mfrow = c(2, 2), mar = rep(0.1, 4), cex = 1.5)
    tree1 <- ape::read.tree(text = "((1, 2), (3, 4, 5, 6, 7, 8));")
    tree2 <- ape::read.tree(text = "((1, 2, 3), (4, 5, 6, 7, 8));")
    VisualizeMatching(RobinsonFouldsMatching, tree1, tree2,
                      setPar = FALSE,
                      Plot = TreeDistPlot,
                      matchZeros = TRUE,
                      plainEdges = TRUE,
                      edge.width = NULL,
                      leaveRoom = FALSE)
    VisualizeMatching(RobinsonFouldsMatching, tree2, tree1,
                      setPar = FALSE,
                      Plot = TreeDistPlot,
                      matchZeros = FALSE,
                      plainEdges = FALSE,
                      leaveRoom = FALSE)
  })
  
  vdiffr::expect_doppelganger("JRF VM matchZeros FALSE", function() {
    JRF2 <- function(tree1, tree2, ...) 
      JaccardRobinsonFoulds(tree1, tree2, k = 2, allowConflict = FALSE, ...)
    
    tree1 <- RootTree(as.phylo(704564, 10), paste0("t", c(1, 4, 5, 8, 9)))
    tree2 <- RootTree(as.phylo(20165 , 10), paste0("t", c(1, 4)))
    VisualizeMatching(JRF2, tree1, tree2, matchZeros = FALSE)
  })
})

test_that("VisualizeMatching() handles unrooted trees", {
  skip_if_not_installed("graphics", "4.3")
  skip_if_not_installed("vdiffr", "1.0")
  
  vdiffr::expect_doppelganger("VM unrooted", function() {
    par(mfrow = c(2, 2), mar = rep(0.1, 4), cex = 1.5)
    tree1 <- UnrootTree(BalancedTree(1:5))
    tree2 <- UnrootTree(PectinateTree(1:5))
    VisualizeMatching(RobinsonFouldsMatching, tree1, tree2,
                      edge.frame = "n",
                      setPar = FALSE,
                      Plot = TreeDistPlot)
  })
  
  vdiffr::expect_doppelganger("VM one rooted", function() {
    par(mfrow = c(2, 2), mar = rep(0.1, 4), cex = 1.5)
    tree1 <- UnrootTree(BalancedTree(1:5))
    tree2 <- PectinateTree(1:5)
    VisualizeMatching(RobinsonFouldsMatching, tree1, tree2,
                      setPar = FALSE,
                      Plot = TreeDistPlot)
  })
})

#' Add minimum spanning tree to plot, colouring by stress
#' 
#' To identify strain in a multidimensional scaling of distances, it can be
#' useful to plot a minimum spanning tree
#' \insertCite{Gower1966,SmithSpace}{TreeDist}.  Colouring each edge of the
#' tree according to its strain can identify areas where the mapping is
#' stretched or compressed.
#' 
#' @param mapping Two-column matrix giving _x_ and _y_ coordinates of plotted
#' points.
#' @param mstEnds Two-column matrix identifying rows of `mapping` at end of
#' each edge of the MST, as output by [`TreeTools::MSTEdges()`].
#' @param distances Matrix or `dist` object giving original distances between
#' each pair of points.
#' @param palette Vector of colours with which to colour edges.
#' @param \dots Additional arguments to [`segments()`].
#'
#' @examples
#' set.seed(0)
#' library("TreeTools", quietly = TRUE)
#' distances <- ClusteringInfoDist(as.phylo(5:16, 8))
#' mapping <- cmdscale(distances, k = 2)
#' mstEnds <- MSTEdges(distances)
#' 
#' # Set up blank plot
#' plot(mapping, asp = 1, frame.plot = FALSE, ann = FALSE, axes = FALSE,
#'      type = "n")
#' # Add MST
#' MSTSegments(mapping, mstEnds, 
#'             col = StrainCol(distances, mapping, mstEnds))
#' # Add points at end so they overprint the MST
#' points(mapping)
#' PlotTools::SpectrumLegend(
#'  "bottomleft",
#'  legend = c("Extended", "Median", "Contracted"),
#'  bty = "n",     # No box
#'  y.intersp = 2, # Expand in Y direction
#'  palette = hcl.colors(256L, "RdYlBu", rev = TRUE)
#' )
#' @template MRS
#' @references \insertAllCited{}
#' @family tree space functions
#' @importFrom graphics segments
#' @export
MSTSegments <- function(mapping, mstEnds, ...) {
  segments(mapping[mstEnds[, 1], 1], mapping[mstEnds[, 1], 2],
           mapping[mstEnds[, 2], 1], mapping[mstEnds[, 2], 2], ...)
}

#' @rdname MSTSegments
#' @return `StrainCol()` returns a vector in which each entry is selected from
#' `palette`, with an attribute `logStrain` denoting the logarithm of the
#' mapped over original distance, shifted such that the median value is zero.
#' Palette colours are assigned centred on the median value, with entries
#' early in `palette` assigned to edges in which the ratio of mapped
#' distance to original distance is small.
#' @importFrom grDevices hcl.colors
#' @importFrom TreeTools MSTEdges
#' @export
StrainCol <- function(distances, mapping, mstEnds = MSTEdges(distances),
                      palette = rev(hcl.colors(256L, "RdYlBu"))) {
  distMat <- as.matrix(distances)
  logStrain <- apply(mstEnds, 1, function(ends) {
    orig <- distMat[ends[[1]], ends[[2]]]
    mapped <- sum((mapping[ends[[1]], ] - mapping[ends[[2]], ]) ^ 2)
    (
      log(mapped) / 2 # sqrt
    ) - log(orig) # High when mapping extends original distances
  })
  strain <- logStrain - median(logStrain[is.finite(logStrain)])
  # Infinite values arise when orig == 0
  maxVal <- max(abs(strain[is.finite(strain)])) + sqrt(.Machine[["double.eps"]])
  nCols <- length(palette)
  bins <- cut(strain, seq(-maxVal, maxVal, length.out = nCols))
  
  # Return:
  structure(palette[bins],
            logStrain = strain)
}

test_that("MST example plots as expected", {
  skip_if_not_installed("graphics", "4.3")
  skip_if_not_installed("vdiffr", "1.0")
  vdiffr::expect_doppelganger("MST example plot", function() {
    distances <- structure(
      c(3, 2.3, 2.3, 2.3, 3, 1.7, 2.3, 2.3, 2.3, 4.5, 4.5, 1.7, 1.7, 2.3, 3,
        2.3, 1.7, 2.3, 2.3, 3.8, 4.4, 1.6, 3.1, 2.3, 2.6, 3, 1.5, 2.6, 4.7,
        4.6, 2.6, 2.3, 1.5, 3, 2.6, 1.5, 4.3, 4.7, 1.7, 2.6, 1.5, 3, 1.6, 4.6,
        4.7, 2.3, 2.3, 1.7, 1.7, 4.4, 3.8, 3.1, 3.1, 1.5, 4.4, 4.4, 3.1, 2.6,
        4.2, 4.8, 3, 4.8, 4.2, 4.7, 4.3, 1.6),
      class = "dist", Size = 12L, Diag = FALSE, Upper = FALSE
    ) # dput(round(ClusteringInfoDist(as.phylo(5:16, 8)), 1))
    
    mapping <- structure(
      c(0.75, 0.36, 0.89, 0.85, 0.89, 0.36, 0.64, 0.6, 0.6, 0.85, -3.4, -3.4,
        0, -0.25, 1.33, 0.39, -1.33, 0.25, 0, -1.52, 1.52, -0.39, -0.58, 0.58),
      dim = c(12L, 2L)
    ) # dput(round(cmdscale(distances, k = 2), 2))
    
    mstEnds <- MSTEdges(distances)
    
    # Set up blank plot
    plot(mapping, asp = 1, frame.plot = FALSE, ann = FALSE, axes = FALSE,
         type = "n")
    
    # Add MST
    MSTSegments(mapping, mstEnds, 
                col = StrainCol(distances, mapping, mstEnds))
    
    # Add points at end so they overprint the MST
    points(mapping)
    PlotTools::SpectrumLegend(
      "bottomleft",
      legend = c("Extended", "Median", "Contracted"),
      bty = "n",
      y.intersp = 2,
      lend = "square",
      palette = hcl.colors(256L, "RdYlBu", rev = TRUE)
    )
  })
})

test_that("StrainCol() handles zeroes", {
  distances <- dist(c(1, 1, 10, 100))
  mapping <- cmdscale(distances)
  expect_equal(attr(StrainCol(distances, mapping), "logStrain")[[1]], Inf)
})

#' Median of a set of trees
#' 
#' Calculate the single binary tree that represents the geometric median -- 
#' an "average" -- of a forest of tree topologies.
#' 
#' The geometric median is the tree that exhibits the shortest average distance
#' from each other tree topology in the set.
#' It represents an "average" of a set of trees, though note that an unsampled 
#' tree may be closer to the geometric "centre of gravity" of the input set --
#' such a tree would not be considered.
#' 
#' The result will depend on the metric chosen to calculate distances between
#' tree topologies. In the absence of a natural metric of tree topologies, 
#' the default choice is [`ClusteringInfoDistance()`] -- which discards
#' branch length information.
#' If specifying a different function, be sure that it returns a difference,
#' rather than a similarity.
#' 
#' 
#' @param x Object of class `multiPhylo` containing phylogenetic trees.
#' @param na.rm,\dots Unused; included for consistency with default function..
#' @param Distance Function to calculate distances between each pair
#' of trees in `x`.
#' @param index Logical: if `TRUE`, return the index of the median tree(s);
#' if `FALSE`, return the tree itself.
#' @param breakTies Logical: if `TRUE`, return a single tree with the minimum
#'  score; if `FALSE`, return all tied trees.
#' 
#' @return `median()` returns an object of class `phylo` 
#' corresponding to the geometric median of a set of trees:
#' that is, the tree whose average distance from all other trees in the set
#' is lowest.  
#' If multiple trees tie in their average distance, the first will be returned,
#' unless `breakTies = FALSE`, in which case an object of class `multiPhylo`
#' containing all such trees will be returned.
#' 
#' @examples
#' library("TreeTools", quietly = TRUE)
#' tenTrees <- as.phylo(1:10, nTip = 8)
#' 
#' # Default settings:
#' median(tenTrees)
#' 
#' # Robinson-Foulds distances include ties:
#' median(tenTrees, Distance = RobinsonFoulds, breakTies = FALSE)
#' 
#' # Be sure to use a distance function, rather than a similarity:
#' NyeDistance <- function(...) NyeSimilarity(..., similarity = FALSE)
#' median(tenTrees, Distance = NyeDistance)
#' 
#' # To analyse a list of trees that is not of class multiPhylo:
#' treeList <- lapply(1:10, as.phylo, nTip = 8)
#' class(treeList)
#' median(structure(treeList, class = "multiPhylo"))
#' @template MRS
#' 
#' @seealso Consensus methods:
#'   [`ape::consensus()`],
#'   [`TreeTools::ConsensusWithout()`]
#'   
#' @importFrom stats median
#' @family tree space functions
#' @encoding UTF-8
#' @export
median.multiPhylo <- function(x, na.rm = FALSE,
                              Distance = ClusteringInfoDistance,
                              index = FALSE,
                              breakTies = TRUE, ...) {
  distances <- unname(colSums(as.matrix(Distance(x))))
  
  # Return:
  if (breakTies) {
    if (index) which.min(distances) else x[[which.min(distances)]]
  } else {
    chosen <- distances == min(distances)
    if (index) {
      which(chosen)
    } else {
      if (sum(chosen) == 1L) {
        x[[which(chosen)]]
      } else {
        x[chosen]
      }
    }
  }
}

#' TreeDist: Distances between Phylogenetic Trees
#' 
#' 'TreeDist' is an R package that implements a suite of metrics that quantify the
#' topological distance between pairs of unweighted phylogenetic trees.
#' It also includes a simple "Shiny" application to allow the visualization of
#' distance-based tree spaces, and functions to calculate the information content
#' of trees and splits.
#' 
#' "TreeDist" primarily employs metrics in the category of
#' "generalized Robinson&ndash;Foulds distances": they are based on comparing splits
#' (bipartitions) between trees, and thus reflect the relationship data within 
#' trees, with no reference to branch lengths.
#' Detailed documentation and usage instructions are 
#' [available online](https://ms609.github.io/TreeDist/) or in the vignettes.
#' 
#' 
#' ## Generalized RF distances
#' 
#' The [Robinson&ndash;Foulds distance](https://ms609.github.io/TreeDist/articles/Robinson-Foulds.html)
#' simply tallies the number of non-trivial splits (sometimes inaccurately
#' termed clades, nodes or edges) that occur in both trees -- any splits that are
#' not perfectly identical contributes one point to the distance score of zero, 
#' however similar or different they are.
#' By overlooking potential similarities between almost-identical splits, 
#' this conservative approach has undesirable properties.
#' 
#' ["Generalized" RF metrics](https://ms609.github.io/TreeDist/articles/Generalized-RF.html)
#' generate _matchings_ that pair each split in one tree with a similar split 
#' in the other.
#' Each pair of splits is assigned a similarity score; the sum of these scores in
#' the optimal matching then quantifies the similarity between two trees.
#' 
#' Different ways of calculating the the similarity between a pair of splits
#' lead to different tree distance metrics, implemented in the functions below:
#'   
#' * [`MutualClusteringInfo()`](https://ms609.github.io/TreeDist/reference/TreeDistance.html), [`SharedPhylogeneticInfo()`](https://ms609.github.io/TreeDist/reference/TreeDistance.html)
#' 
#'   + Smith (2020) scores matchings based on the amount of information
#'     that one partition contains about the other.  The Mutual Phylogenetic
#'     Information assigns zero similarity to split pairs that cannot
#'     both exist on a single tree; The Mutual 
#'     Clustering Information metric is more forgiving, and exhibits more 
#'     desirable behaviour; it is the recommended metric for tree comparison.
#'     (Its complement, [`ClusteringInfoDistance()`](https://ms609.github.io/TreeDist/reference/TreeDistance.html), returns a tree 
#'       distance.)
#' 
#' * [`NyeSimilarity()`](https://ms609.github.io/TreeDist/reference/NyeSimilarity.html)
#' 
#'   + Nye _et al._ (2006) score matchings according to the size of the largest 
#'     split that is consistent with both of them, normalized against 
#'     the Jaccard index.  This approach is extended by B&ouml;cker _et al_. (2013)
#'     with the Jaccard&ndash;Robinson&ndash;Foulds metric (function 
#'     [`JaccardRobinsonFoulds()`](https://ms609.github.io/TreeDist/reference/JaccardRobinsonFoulds.html)).
#' 
#' * [`MatchingSplitDistance()`](https://ms609.github.io/TreeDist/reference/MatchingSplitDistance.html)
#' 
#'   + Bogdanowicz and Giaro (2012) and  Lin _et al._ (2012) independently proposed
#'     counting the number of "mismatched" leaves in a pair of splits.
#'     [`MatchingSplitInfoDistance()`](https://ms609.github.io/TreeDist/reference/TreeDistance.html)
#'     provides an information-based equivalent (Smith 2020).
#' 
#' 
#' The package also implements the variation of the path distance 
#' proposed by Kendal and Colijn (2016) (function
#' [`KendallColijn()`](https://ms609.github.io/TreeDist/reference/KendallColijn.html)),
#' approximations of the Nearest-Neighbour Interchange (NNI) distance (function
#' [`NNIDist()`](https://ms609.github.io/TreeDist/reference/NNIDist.html); 
#' following Li _et al._ (1996)), and calculates the size (function
#' [`MASTSize()`](https://ms609.github.io/TreeDist/reference/MASTSize.html)) and 
#' information content (function
#' [`MASTInfo()`](https://ms609.github.io/TreeDist/reference/MASTSize.html)) of the 
#' Maximum Agreement Subtree.
#' 
#' For an implementation of the Tree Bisection and Reconnection (TBR) distance, see 
#' the package '[TBRDist](https://ms609.github.io/TBRDist/index.html)'.
#' 
#' 
#' # Tree space analysis
#' 
#' Map tree spaces and readily visualize mapped landscapes, avoiding
#' common analytical pitfalls (Smith, forthcoming),
#' using the inbuilt graphical user interface:
#'   
#'   ```r
#' TreeDist::MapTrees()
#' ```
#' 
#' Serious analysts should consult the
#' [vignette](https://ms609.github.io/TreeDist/articles/treespace.html)
#' for a command-line interface.
#' 
#' 
#' @seealso
#' 
#' Further documentation is available in the 
#' [package vignettes](https://ms609.github.io/TreeDist/articles/), visible from
#' R using `vignette(package = "TreeDist")`.
#' 
#' Other R packages implementing tree distance functions include:
#'   
#'   * [ape](https://emmanuelparadis.github.io/):
#'     - `cophenetic.phylo()`: Cophenetic distance
#'     - `dist.topo()`: Path (topological) distance, Robinson&ndash;Foulds distance.
#'   * [phangorn](https://cran.r-project.org/package=phangorn)
#'     - `treedist()`: Path, Robinson&ndash;Foulds and approximate SPR distances.
#'   * [Quartet](https://ms609.github.io/Quartet/): Triplet and Quartet distances, 
#'   using the tqDist algorithm.
#'   * [TBRDist](https://ms609.github.io/TBRDist/): TBR and SPR distances on 
#'   unrooted trees, using the 'uspr' C library.
#'   * [distory](https://cran.r-project.org/package=distory) (unmaintained): 
#'     Geodesic distance
#' 
#' @references
#' 
#' - \insertRef{Bocker2013}{TreeDist}
#' 
#' - \insertRef{Bogdanowicz2012}{TreeDist}
#' 
#' - \insertRef{Kendall2016}{TreeDist}
#' 
#' - \insertRef{Li1996}{TreeDist}
#' 
#' - \insertRef{Lin2012}{TreeDist}
#' 
#' - \insertRef{Nye2006}{TreeDist}
#' 
#' - \insertRef{SmithDist}{TreeDist}
#' 
#' - \insertRef{SmithSpace}{TreeDist}
#' 
#' @encoding UTF-8
#' @keywords internal
"_PACKAGE"

# Suppress "NOTE: Nothing imported from Rdpack":
#' @importFrom Rdpack reprompt
NULL

# The following block is used by usethis to automatically manage
# roxygen namespace tags. Modify with care!
## usethis namespace: start
## usethis namespace: end
NULL

#' Eigenvalues for spectral clustering
#'
#' Spectral clustering emphasizes nearest neighbours when forming clusters;
#' it avoids some of the issues that arise from clustering around means / 
#' medoids.
#'
#' @param D Square matrix or `dist` object containing Euclidean distances
#'   between data points.
#' @param nn Integer specifying number of nearest neighbours to consider
#' @param nEig Integer specifying number of eigenvectors to retain.
#' @author Adapted by MRS from script by [Nura
#'   Kawa](https://rpubs.com/nurakawa/spectral-clustering)
#' @return `SpectralEigens()` returns spectral eigenvalues that can then be
#'   clustered using a method of choice.
#' @examples
#' library("TreeTools", quietly = TRUE)
#' trees <- as.phylo(0:18, nTip = 8)
#' distances <- ClusteringInfoDistance(trees)
#' eigens <- SpectralEigens(distances)
#' # Perform clustering:
#' clusts <- KMeansPP(dist(eigens), k = 3)
#' plot(eigens, pch = 15, col = clusts$cluster)
#' plot(cmdscale(distances), pch = 15, col = clusts$cluster)
#' @family tree space functions
#' @export
SpectralEigens <- function(D, nn = 10L, nEig = 2L) {
  if (nEig < 1) {
    stop("nEig must be positive")
  }
  
  MutualKnnGraph <- function(D, nn) {
    D <- as.matrix(D)
    dims <- dim(D)
    
    # intialize the knn matrix
    knn_mat <- matrix(FALSE, nrow = dims[[1]], ncol = dims[[2]])
    
    # find the 10 nearest neighbours for each point
    for (i in seq_len(nrow(D))) {
      neighbor_index <- order(D[i, ])[2:(nn + 1)]
      knn_mat[i, ][neighbor_index] <- TRUE
    }
    
    # Now we note that i,j are neighbours iff K[i,j] = 1 or K[j, i] = 1
    knn_mat <- knn_mat | t(knn_mat) # find mutual knn
    
    # Return:
    knn_mat
  }
  
  GraphLaplacian <- function(W) {
    stopifnot(nrow(W) == ncol(W))
    
    g <- colSums(W) # degrees of vertices
    n <- dim(W)[[1]]
    D_half <- diag(1 / sqrt(g))
    
    # Return:
    diag(n) - D_half %*% W %*% D_half
  }
  
  W <- MutualKnnGraph(D, nn) # 1. matrix of similarities
  L <- GraphLaplacian(W) # 2. compute graph laplacian
  ei <- eigen(L, symmetric = TRUE) # 3. Compute the eigenvectors and values of L
  
  nL <- dim(L)[[1]]
  
  if (nL > nEig) {
    # Return the eigenvectors of the n_eig smallest eigenvalues:
    ei[["vectors"]][, nL - (0:(nEig - 1))]
  } else {
    ei[["vectors"]]
  }
}

#' @export
#' @rdname SpectralEigens
SpectralClustering <- function(D, nn = 10L, nEig = 2L) {
  .Deprecated("SpectralEigens") # 2021-07
  SpectralEigens(D = D, nn = nn, nEig = nEig)
}

d <- dist(c(1:20, 60:41))

test_that("Spectral clustering fails gracefully", {
  expect_error(SpectralEigens(d, nEig = 0), "nEig must be.*positive")
})

test_that("Spectral clustering works", {
  allEig <- SpectralEigens(d, nEig = Inf)
  expect_equal(dim(allEig), c(40, 40))
  expect_equal(abs(SpectralEigens(d, nEig = 2)), abs(allEig[, 40:39]),
               tolerance = sqrt(.Machine[["double.eps"]]))

  expect_warning(deprecated <- SpectralClustering(d, nEig = Inf),
                 "'SpectralClustering' is deprecated.")
  # Can't check raw equality because signs may flip
  expect_true(all.equal(crossprod(deprecated), crossprod(allEig),
                        tolerance = sqrt(.Machine[["double.eps"]])))
})

#' Information content of splits within a tree
#' 
#' Sum the entropy (`ClusteringEntropy()`), clustering information content
#' (`ClusteringInfo()`), or phylogenetic information content (`SplitwiseInfo()`)
#' across each split within a phylogenetic tree,
#' or the consensus of a set of phylogenetic trees (`ConsensusInfo()`).
#' This value will be greater than the total information 
#' content of the tree where a tree contains multiple splits, as 
#' these splits are not independent and thus contain mutual information that is
#' counted more than once
#' 
#' @section Clustering information:
#' 
#' Clustering entropy addresses the question "how much information is contained
#' in the splits within a tree". Its approach is complementary to the 
#' phylogenetic information content, used in [`SplitwiseInfo()`].
#' In essence, it asks, given a split that subdivides the leaves of a tree into
#' two partitions, how easy it is to predict which partition a randomly drawn 
#' leaf belongs to \insertCite{@Meila2007; @Vinh2010}{TreeDist}.
#' 
#' 
#' Formally, the entropy of a split _S_ that divides _n_ leaves into two
#' partitions of sizes _a_ and _b_ is given by 
#' _H(S)_ = - _a/n_ log _a/n_ - _b/n_ log _b/n_.
#' 
#' Base 2 logarithms are conventionally used, such that entropy is measured in
#' bits.  
#' Entropy denotes the number of bits that are necessary to encode the outcome
#' of a random variable: here, the random variable is "what partition does a
#' randomly selected leaf belong to". 
#' 
#' An even split has an entropy of 1 bit: there is no better way of encoding 
#' an outcome than using one bit to specify which of the two partitions the 
#' randomly selected leaf belongs to.
#' 
#' An uneven split has a lower entropy: membership of the larger partition is
#' common, and thus less surprising; it can be signified using fewer bits in an
#' optimal compression system.
#' 
#' If this sounds confusing, let's consider creating a code to transmit the 
#' cluster label of two randomly selected leaves.  One straightforward
#' option would be to use 
#' 
#' - `00` = "Both leaves belong to partition A"
#' 
#' - `11` = "Both leaves belong to partition B"
#' 
#' - `01` = 'First leaf in A, second in B`
#' 
#' - `10` = 'First leaf in B, second in A`
#' 
#' This code uses two bits to transmit the partition labels of two leaves.
#' If partitions A and B are equiprobable, this is the optimal code; our 
#' entropy -- the average information content required per leaf -- is 1 bit.
#' 
#' 
#' Alternatively, we could use the (suboptimal) code
#' 
#' - `0` = "Both leaves belong to partition A"
#' 
#' - `111` = "Both leaves belong to partition B"
#' 
#' - `101` = 'First leaf in A, second in B`
#' 
#' - `110` = 'First leaf in B, second in A`
#' 
#' If A is much larger than B, then most pairs of leaves will require just
#' a single bit (code `0`). The additional bits when 1+ leaf belongs to B
#' may be required sufficiently rarely that the average message 
#' requires fewer than two bits for two leaves, so the entropy is less than 
#' 1 bit.  (The optimal coding strategy will depend on the exact sizes
#' of A and B.)
#' 
#' 
#' As entropy measures the bits required to transmit the cluster label of each 
#' leaf \insertCite{@Vinh2010: p. 2840}{TreeDist}, the information content of 
#' a split is its entropy multiplied by the number of leaves. 
#' 
#' @section Phylogenetic information:
#' 
#' Phylogenetic information expresses the information content of a split
#' in terms of the probability that a uniformly selected tree will contain it
#' \insertCite{Thorley1998}{TreeDist}.
#' 
#' @section Consensus information:
#' 
#' The information content of splits in a consensus tree is calculated by
#' interpreting support values (i.e. the proportion of trees containing
#' each split in the consensus) as probabilities that the true tree contains
#' that split, following \insertCite{SmithCons;textual}{TreeDist}.
#' 
#' @return `SplitwiseInfo()`, `ClusteringInfo()` and `ClusteringEntropy()`
#' return the splitwise information content of the tree -- or of each split
#' in turn, if `sum = FALSE` -- in bits.
#' 
#' @return `ConsensusInfo()` returns the splitwise information content of the
#' majority rule consensus of `trees`.
#' 
#' @seealso 
#' 
#' An introduction to the phylogenetic information content of a split is given
#' in [`SplitInformation()`](https://ms609.github.io/TreeTools/reference/SplitInformation.html)
#' and in a [package vignette](https://ms609.github.io/TreeDist/articles/information.html).
#' 
#' @examples
#' library("TreeTools", quietly = TRUE)
#' 
#' SplitwiseInfo(PectinateTree(8))
#' tree <- read.tree(text = "(a, b, (c, (d, e, (f, g)0.8))0.9);")
#' SplitwiseInfo(tree)
#' SplitwiseInfo(tree, TRUE)
#' @template MRS
#' 
#' @references
#' \insertAllCited{}
#' 
#' @family information functions
#' @encoding UTF-8
#' @name TreeInfo
NULL


#' @param x A tree of class `phylo`, a list of trees, or a `multiPhylo` object.
#' @param p A vector of probabilities corresponding to each split in `x`. 
#' Specify `TRUE` to calculate this vector using the node labels of each tree.
#' @param sum Logical: if `TRUE`, sum the information content of each split to
#' provide the total splitwise information content of the tree.
#' @rdname TreeInfo
#' @export
SplitwiseInfo <- function(x, p = NULL, sum = TRUE) {
  UseMethod("SplitwiseInfo")
}

.GetPFromLabels <- function(tree, p, splits = as.Splits(tree)) {
  if (length(p) == 1L) { # length(NULL) == 0
    if (p == FALSE) {
      p <- NULL
    } else {
      np <- tree[["node.label"]][as.integer(names(splits)) - NTip(tree)]
      if (is.null(np)) {
        np <- rep_len(p, length(splits))
      }
      p <- as.double(np) / p
      p[is.na(p)] <- 1
      if (any(p > 1)) {
        stop("Nodes must be labelled with probabilities <= 1")
      }
    }
  }
  # Return:
  p
}

#' @export
SplitwiseInfo.phylo <- function(x, p = NULL, sum = TRUE) {
  splits <- as.Splits(x)
  SplitwiseInfo.Splits(splits, .GetPFromLabels(x, p, splits), sum)
}
  
#' @export
SplitwiseInfo.multiPhylo <- function(x, p = NULL, sum = TRUE) {
  vapply(as.Splits(x), SplitwiseInfo, double(1L), p, sum)
}

#' @export
SplitwiseInfo.list <- SplitwiseInfo.multiPhylo

#' @importFrom TreeTools Log2Rooted.int Log2Unrooted.int TipsInSplits
#' @export
SplitwiseInfo.Splits <- function(x, p = NULL, sum = TRUE) {
  nTip <- attr(x, "nTip")
  inSplit <- TipsInSplits(x)
  
  if (is.null(p)) {
    ret <- Log2Unrooted.int(nTip) -
      vapply(inSplit, Log2Rooted.int, 0) -
      vapply(nTip - inSplit, Log2Rooted.int, 0)
    if (sum) sum(ret) else ret
  } else {
    q <- 1L - p
    qNonZero <- as.logical(1L - p)
    q <- q[qNonZero]
    
    l2n <- Log2Unrooted.int(nTip)
    
    l2nConsistent <- vapply(inSplit, Log2Rooted.int, 0) +
      vapply(nTip - inSplit, Log2Rooted.int, 0)
    
    l2pConsistent <- l2nConsistent - l2n
    l2pInconsistent <- log2(-expm1(l2pConsistent[qNonZero] * log(2)))
    
    l2nInconsistent <- l2pInconsistent + l2n
    
    # Return:
    if (sum) {
      sum(l2n * length(inSplit),
          p * (log2(p) - l2nConsistent),
          q * (log2(q) - l2nInconsistent))
    } else {
      ret <- l2n + (p * (log2(p) - l2nConsistent))
      ret[names(l2nInconsistent)] <- ret[names(l2nInconsistent)] +
        (q * (log2(q) - l2nInconsistent))
      ret
    }
  }
}

#' @export
SplitwiseInfo.NULL <- function(x, p = NULL, sum = TRUE) NULL

#' @examples
#' 
#' # Clustering entropy of an even split = 1 bit
#' ClusteringEntropy(TreeTools::as.Splits(c(rep(TRUE, 4), rep(FALSE, 4))))
#' 
#' # Clustering entropy of an uneven split < 1 bit
#' ClusteringEntropy(TreeTools::as.Splits(c(rep(TRUE, 2), rep(FALSE, 6))))
#' 
#' tree1 <- TreeTools::BalancedTree(8)
#' tree2 <- TreeTools::PectinateTree(8)
#' 
#' ClusteringInfo(tree1)
#' ClusteringEntropy(tree1)
#' ClusteringInfo(list(one = tree1, two = tree2))
#' 
#' ClusteringInfo(tree1) + ClusteringInfo(tree2)
#' ClusteringEntropy(tree1) + ClusteringEntropy(tree2)
#' ClusteringInfoDistance(tree1, tree2)
#' MutualClusteringInfo(tree1, tree2)
#' 
#' # Clustering entropy with uncertain splits
#' tree <- ape::read.tree(text = "(a, b, (c, (d, e, (f, g)0.8))0.9);")
#' ClusteringInfo(tree)
#' ClusteringInfo(tree, TRUE)
#' @importFrom TreeTools as.Splits
#' @rdname TreeInfo
#' @export
ClusteringEntropy <- function(x, p = NULL, sum = TRUE) {
  UseMethod("ClusteringEntropy")
}

#' @rdname TreeInfo
#' @export
ClusteringInfo <- function(x, p = NULL, sum = TRUE) {
  UseMethod("ClusteringInfo")
}

#' @rdname TreeInfo
#' @export
ClusteringEntropy.phylo <- function(x, p = NULL, sum = TRUE) {
  splits <- as.Splits(x)
  ClusteringEntropy.Splits(splits, p = .GetPFromLabels(x, p, splits), sum)
}

#' @rdname TreeInfo
#' @export
ClusteringEntropy.list <- function(x, p = NULL, sum = TRUE)
    vapply(as.Splits(x), ClusteringEntropy.Splits, double(1L), p = p, sum)

#' @rdname TreeInfo
#' @export
ClusteringEntropy.multiPhylo <- ClusteringEntropy.list

#' @rdname TreeInfo
#' @export
ClusteringEntropy.Splits <- function(x, p = NULL, sum = TRUE) {
  nLeaves <- attr(x, "nTip")
  if (is.null(nLeaves) || !is.finite(nLeaves)) {
    stop("`x` must have a valid 'nTip' attribute.")
  }
  if (nLeaves <= 0) {
    return(if (sum) 0 else double(0))
  }
  
  inSplit <- TipsInSplits(x)  # integer vector, names correspond to splits
  nSplits <- length(inSplit)
  
  H <- tryCatch(
    binary_entropy_counts(inSplit, as.integer(nLeaves)),
    error = function(e) {
      pi <- inSplit / nLeaves
      H <- numeric(nSplits)
      nz <- (pi > 0) & (pi < 1)
      pz <- pi[nz]
      H[nz] <- -(pz * log(pz) + (1 - pz) * log1p(-pz))
      H[!nz] <- 0.0
      H
    }
  )
  
  # Weighting
  if (is.null(p)) p <- 1
  if (!(length(p) %in% c(1L, nSplits))) {
    stop("`p` must be length 1 or length equal to number of splits (", nSplits, ").")
  }
  ret <- H * p
  
  # Return:
  if (isTRUE(sum)) {
    sum(ret)
  } else {
    setNames(ret, names(inSplit))
  }
}

#' @export
ClusteringEntropy.NULL <- function(x, p = NULL, sum = TRUE) NULL

#' @rdname TreeInfo
#' @export
ClusteringInfo.phylo <- function(x, p = NULL, sum = TRUE) {
  splits <- as.Splits(x)
  ClusteringInfo.Splits(splits, p = .GetPFromLabels(x, p, splits), sum)
}

#' @rdname TreeInfo
#' @export
ClusteringInfo.list <- function(x, p = NULL, sum = TRUE)
    vapply(as.Splits(x), ClusteringInfo.Splits, double(1L), p = p, sum)

#' @rdname TreeInfo
#' @export
ClusteringInfo.multiPhylo <- ClusteringInfo.list

#' @rdname TreeInfo
#' @export
ClusteringInfo.Splits <- function(x, p = NULL, sum = TRUE) {
  nLeaves <- attr(x, "nTip")
  ClusteringEntropy.Splits(x, p, sum) * nLeaves
}

#' @export
ClusteringInfo.NULL <- function(x, p = NULL, sum = TRUE) NULL

#' @rdname TreeInfo
#' 
#' @param trees List of `phylo` objects, optionally with class `multiPhylo`.
#' @param info Abbreviation of "phylogenetic" or "clustering", specifying
#' the concept of information to employ.
#' @param p Scalar from 0.5 to 1 specifying minimum proportion of trees that
#' must contain a split for it to appear within the consensus.
#' @param check.tips Logical specifying whether to renumber leaves such that
#' leaf numbering is consistent in all trees.
#' 
#' @examples
#' 
#' # Support-weighted information content of a consensus tree
#' set.seed(0)
#' trees <- list(RandomTree(8), RootTree(BalancedTree(8), 1), PectinateTree(8))
#' cons <- consensus(trees, p = 0.5)
#' p <- SplitFrequency(cons, trees) / length(trees)
#' plot(cons)
#' LabelSplits(cons, signif(SplitwiseInfo(cons, p, sum = FALSE), 4))
#' ConsensusInfo(trees)
#' LabelSplits(cons, signif(ClusteringInfo(cons, p, sum = FALSE), 4))
#' ConsensusInfo(trees, "clustering")
#' @export
ConsensusInfo <- function(trees, info = "phylogenetic", p = 0.5,
                           check.tips = TRUE) {
  safeP <- min(1, max(0.5, p))
  if (safeP != p) {
    warning("`p` coerced to range (0.5 <= p <= 1) in ConsensusInfo()")
  }
  
  mode <- pmatch(tolower(info),
                 c("phylogenetic", "clustering", "spic", "scic")) %% 2
  if (is.na(mode)) {
    stop("`info` must be 'phylogenetic' or 'clustering'")
  }
  if (inherits(trees, "phylo")) {
    # Convert to length-1 multiphylo object
    trees <- c(trees)
  }
  if (length(trees) == 1L) {
    return((if (mode) SplitwiseInfo else ClusteringInfo)(trees))
  }
  if (check.tips) {
    trees <- RenumberTips(trees, trees[[1]])
  }
  consensus_info(trees, mode == 1L, p = safeP)
}

# R-level validation wrapper; shadows RcppExports version.
# C++ ASSERT() is compiled out in release builds.
consensus_info <- function(trees, phylo, p) {
  if (p > 1 + 1e-15) {
    stop("p must be <= 1.0 in consensus_info()")
  }
  if (p < 0.5) {
    stop("p must be >= 0.5 in consensus_info()")
  }
  nTip <- NTip(trees[[1]])
  # CT_MAX_LEAVES = 16383 in information.h (lookup table size limit)
  if (nTip > 16383L) {
    stop("This many leaves are not yet supported")
  }
  .Call(`_TreeDist_consensus_info`, trees, phylo, p)
}

#' Maximum Clade Information Tree
#'
#' Analogous to the Maximum Clade Credibility tree:
#' select the tree from a posterior distribution whose clades have the
#' highest information content.
#' Generate the MCC tree by specifying `info = "credibility"`.
#'
#' @inheritParams ConsensusInfo
#' @return `MCITree()` returns the tree with the highest information
#' content, selected from `trees`.
#' @examples
#' library("TreeTools", quietly = TRUE)
#' trees <- as.phylo(24:40, 16)
#'
#' # Maximum Clade Information tree
#' mci <- MCITree(trees)
#' SplitwiseInfo(mci)
#' plot(mci)
#' p <- SplitFrequency(mci, trees) / length(trees)
#' LabelSplits(mci, round(p * 100), "%", bg = SupportColor(p))
#'
#' \donttest{
#' # Compare with Maximum Clade Credibility tree
#' mcc <- MCITree(trees, "credibility")
#' plot(mcc)
#' p <- SplitFrequency(mcc, trees) / length(trees)
#' LabelSplits(mcc, round(p * 100), "%", bg = SupportColor(p))
#' SplitwiseInfo(mcc)
#' }
#'
#' @template MRS
#' @family summary trees
#' @importFrom TreeTools as.Splits SplitFrequency
#' @export
MCITree <- function(trees, info = "phylogenetic", check.tips = TRUE) {
  mode <- pmatch(tolower(info),
                 c("phylogenetic", "clustering", "credibility",
                   "spic", "scic"))
  if (is.na(mode)) {
    stop("`info` must be \"phylogenetic\", \"clustering\" or \"credibility\"")
  }
  if (inherits(trees, "phylo")) {
    return(trees)
  }
  if (length(trees) == 1L) {
    return(trees[[1]])
  }
  if (check.tips) {
    trees <- RenumberTips(trees, trees[[1]])
  }

  nTree <- length(trees)
  treeSplits <- as.Splits(trees)
  InfoFn <- switch(mode,
    function(sp, p) SplitwiseInfo(sp, p, sum = TRUE),
    function(sp, p) ClusteringInfo(sp, p, sum = TRUE),
    function(sp, p) sum(p),
    function(sp, p) SplitwiseInfo(sp, p, sum = TRUE),
    function(sp, p) ClusteringInfo(sp, p, sum = TRUE)
  )

  treeInfo <- vapply(treeSplits, function(sp) {
    p <- SplitFrequency(sp, trees) / nTree
    InfoFn(sp, p)
  }, double(1))

  chosen <- which.max(treeInfo)
  nBest <- sum(treeInfo == max(treeInfo))
  if (nBest > 1L) {
    message(nBest, " trees tied for best score; returning tree ", chosen)
  }

  trees[[chosen]]
}

test_that("MCITree returns one of the input trees", {
  trees <- as.phylo(24:40, 16)
  mci <- MCITree(trees)
  expect_true(any(vapply(trees, all.equal, TRUE, mci) == "TRUE"))
})

test_that("MCITree modes give valid results", {
  trees <- as.phylo(24:40, 16)
  for (mode in c("phylogenetic", "clustering", "credibility")) {
    result <- MCITree(trees, mode)
    expect_s3_class(result, "phylo")
  }
})

test_that("MCITree aliases match base modes", {
  trees <- as.phylo(24:40, 16)
  expect_equal(MCITree(trees, "spic"), MCITree(trees, "phylogenetic"))
  expect_equal(MCITree(trees, "scic"), MCITree(trees, "clustering"))
})

test_that("MCITree handles edge cases", {
  tree <- PectinateTree(8)
  expect_identical(MCITree(tree), tree)

  trees <- c(PectinateTree(8), BalancedTree(8))
  expect_identical(MCITree(trees[1]), trees[[1]])
})

test_that("MCITree rejects invalid info", {
  trees <- as.phylo(24:40, 16)
  expect_error(MCITree(trees, "invalid"), "must be")
})

test_that("MCITree credibility selects highest-support tree", {
  # 10 copies of tree A + 1 copy of tree B:
  # tree A's splits have frequency 10/11; tree B's have ~1/11.
  # The credibility tree should be one of the A copies.
  treeA <- PectinateTree(8)
  treeB <- BalancedTree(8)
  trees <- c(rep(list(treeA), 10), list(treeB))
  class(trees) <- "multiPhylo"
  mcc <- MCITree(trees, "credibility")
  expect_equal(mcc, treeA)
})

test_that("MCITree messages on ties", {
  # Identical trees should tie
  tree <- PectinateTree(8)
  trees <- rep(list(tree), 5)
  class(trees) <- "multiPhylo"
  expect_message(MCITree(trees), "tied")
})

test_that("Split info calculated", {
  library("ape")
  trees <- rev(list( # rev() because trees are emplaced backwards
    read.tree(text = "(a, (b, (c, (z, ((d, e), (f, (g, x)))))));"),
    read.tree(text = "(a, (b, (c, (z, ((d, e), (f, (g, x)))))));"),
    read.tree(text = "(a, (b, ((c, z), ((d, e), ((f, x), g)))));"),
    read.tree(text = "(a, (b, ((c, z), ((d, (e, x)), (f, g)))));"),
    read.tree(text = "(a, ((b, x), ((c, z), ((d, e), (f, g)))));")
  ))
  split_p <- c(ab = 0.8, d..x = 0.8, fgx = 0.6, de = 0.8, cz = 0.6)
  
  trees[] <- lapply(trees, RenumberTips, c(letters[1:7], "x", "z"))
  expect_equal(consensus_info(trees, TRUE, 1), 0)
  expect_equal(SplitwiseInfo(consensus(trees, p = 0.5, rooted = TRUE),
                             p = split_p),
               ConsensusInfo(trees, p = 0.5))
  expect_warning(expect_equal(ConsensusInfo(trees, p = 0.4),
                              ConsensusInfo(trees, p = 0.5)))
  expect_equal(SplitwiseInfo(consensus(trees, p = 0.7),
                             p = split_p[split_p > 0.7]),
               consensus_info(trees, TRUE, 0.7))
  expect_error(consensus_info(trees, TRUE, 7))
  expect_error(consensus_info(trees, TRUE, -7))
  expect_error(
    consensus_info(list(rtree(20000), rtree(20000)), TRUE, p = 1),
    "This many leaves are not yet supported"
  )
  expect_error(
    consensus_info(list(rtree(20000), rtree(20000)), FALSE, p = 1),
    "This many leaves are not yet supported"
  )

  
  # Expected:
  # - tablei = 0: 
  #              1 2 3 4
  #   g..x : 2   x . . .
  #   f..x : 3   x x . .  ! 3|6, 60%
  #   d..e : 4   x x . x  ! 2|7, 80%
  #   d..x : 4   x x x .  ! 4|5, 80%
  #   z..x : 2   x . . .
  #   c..x : 4   x x x .  ! 7|2, 80%
  # 
  # - tablei = 1:
  #   Pass
  #   
  # - tablei = 2:
  #              3 4
  #   c..z : 3   x x  ! 2|7, 60%
  #   f..x : 1   . .  
  # 
  # - tablei = 3: Terminate.
  #
  
  p_in <- c(7, 5, 3, 2, 2) / 9
  p_out <- 1 - p_in
  expect_equal(sum(apply(rbind(p_in, p_out), 2, Entropy) * split_p)
               * NTip(trees[[1]]),
               consensus_info(trees, FALSE, 0.5))
  
  # Even number of trees: cz, with p == 0.5, not in consensus.
  split_p <- c(ab = 0.8, d..x = 0.8, fgx = 0.6, de = 0.8, cz = 0.6)
  expect_equal(SplitwiseInfo(consensus(trees[-1], p = 0.5),
                             p = c(ab = 1, d..x = 1, fgx = 3/4, de = 3/4)),
               consensus_info(trees[-1], TRUE, p = 0.5))
  expect_equal(SplitwiseInfo(consensus(trees[-1], p = 0.8),
                             p = c(ab = 1, d..x = 1)),
               consensus_info(trees[-1], TRUE, p = 0.8))
  expect_warning(expect_equal(ConsensusInfo(trees[-1], p = 80),
                              ConsensusInfo(trees[-1], p = 1)))
})

library("TreeTools")

test_that("ClusteringInfo() fails gracefully", {
  bal6 <- as.Splits(BalancedTree(6))
  bad6 <- bal6
  attr(bad6, "nTip") <- NULL
  expect_error(ClusteringEntropy(bad6), "nTip")
  expect_error(ClusteringEntropy(bal6, p = c(3, 5)), "number of splits .3\\)")
})

test_that("SplitwiseInfo() / ClusteringInfo() handle probabilities", {
  Tree <- function(txt) ape::read.tree(text = txt)
  tree <- Tree("((a, b)60, (c, d)60);")
  treeP <- Tree("((a, b)0.60, (c, d)0.60);")
  treeProfile <- list(Tree("((a, b), (c, d));"),
                      Tree("(a, b, c, d);"),
                      Tree("((a, d), (c, b));"))[c(1, 1, 1, 2, 3)]
  
  Test <- function(Expect, tree, p = NULL, ...) {
    Expect(..., SplitwiseInfo(tree, p))
    Expect(..., ClusteringInfo(tree, p))
    Expect(..., ClusteringEntropy(tree, p))
  }
  Clust <- function(tree, ...) {
    expect_equal(ClusteringInfo(tree, ..., sum = TRUE),
                 sum(ClusteringInfo(tree, ..., sum = FALSE)))
    expect_equal(ClusteringEntropy(tree, ..., sum = TRUE),
                 sum(ClusteringEntropy(tree, ..., sum = FALSE)))
    expect_equal(ClusteringInfo(tree, ...) / NTip(tree),
                 ClusteringEntropy(tree, ...))
  }
  Test(expect_error, tree, TRUE)
  Test(expect_null, tree = NULL)
  expect_gt(SplitwiseInfo(tree), SplitwiseInfo(tree, 100))
  expect_gt(ClusteringInfo(tree), ClusteringInfo(tree, 100))
  expect_gt(ClusteringEntropy(tree), ClusteringEntropy(tree, 100))
  expect_equal(SplitwiseInfo(tree), SplitwiseInfo(tree, p = FALSE))
  expect_equal(ClusteringInfo(tree), ClusteringInfo(tree, p = FALSE))
  expect_equal(ClusteringEntropy(tree), ClusteringEntropy(tree, p = FALSE))
  
  expect_equal(0, SplitwiseInfo(tree, 60 * 3),
               tolerance = sqrt(.Machine$double.eps))
  expect_equal(1 / 3, Clust(tree, 60 * 3))
  
  p <- 1.0 * c(1, 0, 0) + 0.0 * c(0, 1/2, 1/2)
  expect_equal(log2(3), SplitwiseInfo(tree))
  expect_equal(1, Clust(tree))
  
  p <- 0.6 * c(1, 0, 0) + 0.4 * c(0, 1/2, 1/2)
  expect_equal(sum(p), 1)
  expectation <- log2(3) + sum(p * log2(p))
  expect_equal(expectation, SplitwiseInfo(tree, 100))
  expect_equal(0.6, Clust(tree, 100))
  
  
  expect_equal(SplitwiseInfo(tree, 100), SplitwiseInfo(treeP, TRUE))
  expect_equal(Clust(tree, 100), Clust(treeP, TRUE))
  
  expect_equal(SplitwiseInfo(tree, 100), ConsensusInfo(treeProfile, "p"))
  expect_equal(ClusteringInfo(tree, 100), ConsensusInfo(treeProfile, "c"))
  
  expect_equal(SplitwiseInfo(Tree("(a, b, (c, (d, e)0.8)0.75);"), TRUE),
               SplitwiseInfo(Tree("(a, b, (c, d, e)0.75);"), TRUE) +
                 SplitwiseInfo(Tree("(a, b, c, (d, e)0.8);"), TRUE))
  expect_equal(Clust(Tree("(a, b, (c, (d, e)0.8)0.75);"), TRUE),
               Clust(Tree("(a, b, (c, d, e)0.75);"), TRUE) +
                 Clust(Tree("(a, b, c, (d, e)0.8);"), TRUE))
  expect_equal(SplitwiseInfo(Tree("(a, b, (c, (d, e)0.8)0.75);"), TRUE),
               SplitwiseInfo(Tree("(a, b, (c, (d, e)));"), c(0.75, 0.8)))
  expect_equal(Clust(Tree("(a, b, (c, (d, e)0.8)0.75);"), TRUE),
               Clust(Tree("(a, b, (c, (d, e)));"), c(0.75, 0.8)))
  expect_equal(SplitwiseInfo(Tree("(a, b, (c, (d, e)0.8));"), TRUE),
               SplitwiseInfo(Tree("(a, b, (c, (d, e)));"), c(1, 0.8)))
  expect_equal(Clust(Tree("(a, b, (c, (d, e)0.8));"), TRUE),
               Clust(Tree("(a, b, (c, (d, e)));"), c(1, 0.8)))
  expect_equal(SplitwiseInfo(Tree("(a, b, (c, (d, e)));")),
               SplitwiseInfo(Tree("(a, b, (c, (d, e)));"), TRUE))
  expect_equal(Clust(Tree("(a, b, (c, (d, e)));")),
               Clust(Tree("(a, b, (c, (d, e)));"), TRUE))
  expect_equal(SplitwiseInfo(Tree("(a, b, (c, (d, e)));"), TRUE),
               SplitwiseInfo(Tree("(a, b, (c, (d, e)));"), c(1, 1)))
  expect_equal(Clust(Tree("(a, b, (c, (d, e)));"), TRUE),
               Clust(Tree("(a, b, (c, (d, e)));"), c(1, 1)))
})


test_that("SplitwiseInfo() / ClusteringInfo(sum = FALSE)", {
  splits <- as.Splits(BalancedTree(8))
  Test <- function(x) {
    expect_equal(length(x), length(splits))
    expect_equal(names(x), names(splits))
  }
  p <- c(1, 1, 0.5, 0.6, 0.7)
  Test(SplitwiseInfo(BalancedTree(8), sum = FALSE))
  Test(SplitwiseInfo(BalancedTree(8), p = p, sum = FALSE))
  Test(ClusteringInfo(BalancedTree(8), sum = FALSE))
  Test(ClusteringInfo(BalancedTree(8), p = p, sum = FALSE))
})

test_that("SplitwiseInfo() can't be improved by dropping resolved tip", {
  # Test arguably redundant, but a useful reminder of a closed optimization
  # possibility
  b8With <- as.Splits(c(T, T, T, T, F, F, F, F))
  b8Without <- as.Splits(c(T, T, T, F, F, F, F))
  i8With <- as.Splits(c(T, T, T, T, T, T, F, F))
  i8Without <- as.Splits(c(T, T, T, T, T, F, F))
  expect_lt(SplitwiseInfo(b8With, p = 0.5), SplitwiseInfo(b8Without))
  expect_lt(SplitwiseInfo(i8With, p = 0.5), SplitwiseInfo(i8Without))
  
  balancedWithout <- BalancedTree(32)
  balancedWith <- AddTip(balancedWithout, 32)
  p <- double(30) + 1
  p[c(58, 62, 64, 65) - 35] <- 0.5
  expect_lt(SplitwiseInfo(balancedWith, p = p), SplitwiseInfo(balancedWithout))
})

test_that("ClusteringInfo() method works", {
  trees <- list(BalancedTree(8), PectinateTree(8))
  expect_equal(vapply(trees, ClusteringInfo, 0),
               ClusteringInfo(structure(trees, class = "multiPhylo")))
  expect_equal(vapply(trees, ClusteringInfo, 0),
               ClusteringInfo(trees))
  
  
  # Recall: Trees are read in backwards.
  trees <- list(RandomTree(8), BalancedTree(8), PectinateTree(8))
  # Consensus assumes trees are rooted (!)
  cons <- consensus(lapply(trees, RootTree, 1), p = 0.5)
  p <- SplitFrequency(cons, trees) / length(trees)
  expect_equal(SplitwiseInfo(cons, p), ConsensusInfo(trees, "spic"))
  expect_equal(ClusteringInfo(cons, p), ConsensusInfo(trees, "scic"))
  
})


test_that("ConsensusInfo() is robust", {
  trees <- list(ape::read.tree(text = "(a, (b, (c, (d, (e, X)))));"),
                ape::read.tree(text = "((a, X), (b, (c, (d, e))));"))
  expect_equal(0, ConsensusInfo(trees, "cl"))
  expect_error(ConsensusInfo(trees, "ERROR"))
  # multiPhylo vs phylo
  expect_equal(ConsensusInfo(trees[1]), ConsensusInfo(trees[[1]]))
})

test_that("ConsensusInfo() generates correct value", {
  trees <- list(ape::read.tree(text = "((a, b), (c, d));"),
                ape::read.tree(text = "((a, c), (b, d));"),
                ape::read.tree(text = "((a, d), (c, b));"))
  expect_equal(0, ConsensusInfo(trees))
  expect_equal(0, ConsensusInfo(trees, "cl"))
  expect_equal(log2(3), ConsensusInfo(trees[1]))
  expect_equal(4, ConsensusInfo(trees[1], "cl"))
  expect_equal(log2(3), ConsensusInfo(trees[c(1, 1)]))
  expect_equal(4, ConsensusInfo(trees[c(1, 1)], "cl"))
  
  expect_equal(Entropy(c(1, 1, 1) / 3) - Entropy(c(1/2, 1/2, 9)/10),
               ConsensusInfo(trees[c(rep(1, 9), 2)]))
})

#' Graphical user interface for mapping distances and analysing
#' tree space
#' 
#' `MapTrees()` launches a "Shiny" application for the visualization and
#' evaluation of tree spaces.
#' 
#' # Input tab
#' 
#' The input tab allows for the upload of sets of phylogenetic trees from file.
#' Trees at the start or end of a file can be excluded, and the number of trees
#' can be brought down to a manageable number by uniformly subsampling every
#' _n_th tree.  Samples of c. 100 trees can be analysed in seconds;
#' analysis of larger samples will take longer, particularly with slower
#' methods (e.g. quartet distances; Kruskal-1 MDS; large minimum spanning trees).
#' 
#' Different batches can be plotted with different colours / symbols.
#' 
#' If each tree is associated with a property -- for example, the data or method 
#' used to generate it, or its stratigraphic congruence -- a list of properties 
#' for each tree, with one entry per line/row, can be uploaded along with
#' the trees.  Points in tree space can then be styled according to the 
#' corresponding property.
#' 
#' If trees are subsampled (using the "Sample every" slider), then the values
#' in the tree properties file can also be subsampled accordingly.
#' Unfortunately there is not yet support for multiple point property files;
#' one file will be applied to all trees, in the sequence that they were added
#' to memory.
#' 
#' 
#' # Analysis tab
#' 
#' Select from a suite of distance methods: clustering information and 
#' phylogenetic information are quick and satisfactory; quartet is slow but
#' gives slightly better mappings; path is very fast but may not reflect
#' evolutionary signal very well; and Robinson--Foulds should probably never
#' be used for analysis; it is included for comparison purposes.
#' 
#' Principle components mappings should suffice for most purposes; 
#' Sammon and Kruskal mappings are slower and seldom differ by much,
#' in character or quality, but may emphasize outliers more.
#' 
#' Partitioning around medoids or minimax-linkage hierarchical clustering
#' will typically find a close-to-optimal clustering where one exists;
#' select additional methods for a more exhaustive search.  
#' To avoid redundant calculation, clusterings are only updated when
#' "recalculate clustering" is clicked, or the "maximum cluster number" slider
#' is modified; clustering solutions using more than this many clusters are
#' not considered
#' Clusterings with silhouette coefficients < 0.25 are unlikely to represent 
#' genuine structure and are not reported or depicted.
#' 
#' 
#' # Display tab
#' 
#' Up to 15 dimensions can be depicted; the quality of a mapping -- that is, 
#' the faithfulness of mapped distances to true tree-to-tree distances --
#' is quantified by the product of the Trustworthiness and Continuity metrics,
#' which should exceed 0.9 (at least).
#' 
#' An interactive 3D plot can be explored by dragging the mouse and scrolling,
#' but do be careful to check that three dimensions are enough to depict your
#' data accurately.
#' 
#' The minimum spanning tree is the shortest possible line selecting the chosen
#' subsample of trees; if it takes a convoluted zig-zagging route, then the 
#' mapping is doing a poor job of reflecting true tree to tree distances.
#' 
#' Convex hulls are the smallest polygons enclosing all points in each cluster;
#' they are handy for spotting clusters, but their area does not correspond
#' to a genuine quantity, so should not be interpreted.
#' 
#' Tree numbers correspond to the sequence of trees in their original input 
#' file, before subsampling.
#' 
#' Each tree is denoted by a point, whose symbol can be styled according to
#' cluster membership or according to the file that contains the tree,
#' with each click of "Add to existing" on the input tab constituting a 
#' new batch with a new symbol.
#' 
#' Points can be coloured according to a category -- the cluster or batch to
#' which they belong, or custom data provided in the Point Property File
#' on the input tab -- or continuously, either by the sequence in which they
#' were added to memory, or according to custom data.
#' 
#' 
#' # Exporting tree spaces
#' 
#' A mapping can be saved to PDF or as a PNG bitmap at the size selected.
#' 
#' 
#' # References
#' 
#' A list of references employed when constructing the tree space is populated
#' according to the methods used; it would be appropriate to cite and briefly 
#' discuss these studies in any publication using figures generated using
#' this application.  The application itself can be cited using 
#' \insertCite{SmithDist,SmithSpace;textual}{TreeDist}
#'  
#' 
#' @seealso
#' Full detail of tree space analysis in R is provided in the accompanying
#' [vignette](https://ms609.github.io/TreeDist/articles/treespace.html).
#' 
#' @references 
#' \insertAllCited{}
#' 
#' @template MRS
#' @family tree space functions
#' 
#' @importFrom shiny runApp
#' @importFrom shinyjs useShinyjs
#' @export
MapTrees <- function() {
  appDir <- system.file("treespace", package = "TreeDist")
  if (appDir == "") {
    stop("Could not find example directory. Try re-installing 'TreeDist'.", 
         call. = FALSE)
  }
  
  shiny::runApp(appDir, display.mode = "normal")
}

#' @rdname MapTrees
#' @export
Project <- function() {
  .Deprecated("MapTrees")
  MapTrees()
}

# Benchmark: streaming pmin vs apply-rebuild in KMeansPP initialisation
#
# Compares old (apply-rebuild) vs new (streaming pmin) seeding loop.
# Also verifies bit-identical output under fixed set.seed().

library(TreeDist)

# ---- helpers ----------------------------------------------------------------

old_seed_loop <- function(d, n, k, nstart) {
  best_centres <- NULL
  best_score <- Inf
  for (s in seq_len(nstart)) {
    centres <- c(sample.int(n, 1L), numeric(k - 1L))
    for (i in 2:k) {
      p <- apply(d[centres, , drop = FALSE], 2, min) ^ 2
      centres[i] <- sample.int(n, 1L, prob = p)
    }
    score <- sum(apply(d[centres, , drop = FALSE], 2, min) ^ 2)
    if (score < best_score) { best_score <- score; best_centres <- centres }
  }
  best_centres
}

new_seed_loop <- function(d, n, k, nstart) {
  best_centres <- NULL
  best_score <- Inf
  for (s in seq_len(nstart)) {
    centres <- integer(k)
    centres[1L] <- sample.int(n, 1L)
    min_d <- d[centres[1L], ]
    for (i in 2:k) {
      p <- min_d ^ 2
      centres[i] <- sample.int(n, 1L, prob = p)
      min_d <- pmin.int(min_d, d[centres[i], ])
    }
    score <- sum(min_d ^ 2)
    if (score < best_score) { best_score <- score; best_centres <- centres }
  }
  best_centres
}

# ---- identity check ---------------------------------------------------------

check_identity <- function(n, k, nstart, seed = 42L) {
  set.seed(seed); x <- matrix(rnorm(n * 5), n, 5)
  d <- as.matrix(dist(x))
  # old loop stores centres as double (c(sample.int, numeric(...))); new as
  # integer. Coerce both to integer before comparing.
  set.seed(seed + 1L); old <- as.integer(old_seed_loop(d, n, k, nstart))
  set.seed(seed + 1L); new <- as.integer(new_seed_loop(d, n, k, nstart))
  identical(old, new)
}

cat("Identity checks:\n")
for (cfg in list(c(200, 10), c(500, 20))) {
  n <- cfg[1]; k <- cfg[2]
  ok <- check_identity(n, k, nstart = 10L)
  cat(sprintf("  N=%d k=%d: %s\n", n, k, if (ok) "PASS" else "FAIL"))
}

# ---- benchmark --------------------------------------------------------------

run_bench <- function(n, k, nstart = 10L, times = 5L) {
  set.seed(99L); x <- matrix(rnorm(n * 5), n, 5)
  d <- as.matrix(dist(x))

  if (requireNamespace("microbenchmark", quietly = TRUE)) {
    mb <- microbenchmark::microbenchmark(
      old = { set.seed(1L); old_seed_loop(d, n, k, nstart) },
      new = { set.seed(1L); new_seed_loop(d, n, k, nstart) },
      times = times
    )
    med <- tapply(mb$time, mb$expr, median)
    ratio <- med[["old"]] / med[["new"]]
    cat(sprintf("  N=%d k=%d: old %.2f ms | new %.2f ms | ratio %.2fx\n",
                n, k,
                med[["old"]] / 1e6,
                med[["new"]] / 1e6,
                ratio))
  } else {
    t_old <- median(replicate(times, {
      set.seed(1L); system.time(old_seed_loop(d, n, k, nstart))["elapsed"]
    }))
    t_new <- median(replicate(times, {
      set.seed(1L); system.time(new_seed_loop(d, n, k, nstart))["elapsed"]
    }))
    ratio <- t_old / t_new
    cat(sprintf("  N=%d k=%d: old %.4f s | new %.4f s | ratio %.2fx\n",
                n, k, t_old, t_new, ratio))
  }
}

cat("\nBenchmark (seeding loop only, nstart=10):\n")
for (cfg in list(c(200, 10), c(500, 20))) {
  run_bench(n = cfg[1], k = cfg[2])
}

#' Cluster size statistics
#' @name cluster-statistics
#' 
#' @param x Matrix in which each row lists the coordinates of a point
#' in a Euclidian space; or, where supported, `dist` object specifying
#' distances between each pair of points.
#' @param cluster Optional integer vector specifying the cluster or group to
#' which each row in `x` belongs.
#' 
#' @examples
#' points <- rbind(matrix(1:16, 4), rep(1, 4), matrix(1:32, 8, 4) / 10)
#' cluster <- rep(1:3, c(4, 1, 8))
#' 
#' plot(
#'   points[, 1:2], # Plot first two dimensions of four-dimensional space
#'   col = cluster, pch = cluster, # Style by cluster membership
#'   asp = 1, # Fix aspect ratio to avoid distortion
#'   ann = FALSE, frame = FALSE # Simple axes
#' )
#' 
#' @family tree space functions
#' @family cluster functions
#' @template MRS
NULL

#' @return `SumOfRanges()` returns a numeric specifying the sum of ranges
#' within each cluster across all dimensions.
#' @rdname cluster-statistics
#' @examples SumOfRanges(points, cluster)
#' @export
SumOfRanges <- function(x, cluster = 1) {
  if (is.null(dim(x))) {
    warning(paste0("`x` lacks dimensions. ",
                   "Did you subset without specifying `drop = FALSE`?"))
    x <- matrix(x, 1)
  }
  
  # Return:
  vapply(seq_along(unique(cluster)), 
         function(i) .SumOfRanges(x[cluster == i, , drop = FALSE]),
         numeric(1))
}

.SumOfRanges <- function(x) {
  ranges <- apply(x, 2, range)
  
  # Return:
  sum(ranges[2, ] - ranges[1, ])
}

#' @rdname cluster-statistics
#' @return `SumOfVariances()` returns a numeric specifying the sum of variances
#' within each cluster across all dimensions.
#' @examples SumOfVariances(points, cluster)
#' @importFrom stats var
#' @export
SumOfVariances <- function(x, cluster = 1) {
  if (is.null(dim(x))) {
    warning(paste0("`x` lacks dimensions. ",
                   "Did you subset without specifying `drop = FALSE`?"))
    x <- matrix(x, 1)
  }
  
  # Return:
  vapply(seq_along(unique(cluster)), 
         function(i) .SumOfVariances(x[cluster == i, , drop = FALSE]),
         numeric(1))
}

#' @rdname cluster-statistics
#' @export
SumOfVars <- SumOfVariances

.SumOfVariances <- function(x) {
  sum(apply(x, 2, var))
}


#' @rdname cluster-statistics
#' @return `MeanCentroidDistance()` returns a numeric specifying the mean
#' distance from the centroid to points in each cluster.
#' @examples MeanCentroidDistance(points, cluster)
#' @export
MeanCentroidDistance <- function(x, cluster = 1, Average = mean) {
  if (is.null(dim(x))) {
    warning(paste0("`x` lacks dimensions. ",
                   "Did you subset without specifying `drop = FALSE`?"))
    x <- matrix(x, 1)
  }
  
  # Return:
  vapply(seq_along(unique(cluster)),
         function(i) .MeanCentroidDist(x[cluster == i, , drop = FALSE],
                                       Average = Average),
         numeric(1))
}

#' @rdname cluster-statistics
#' @export
MeanCentDist <- MeanCentroidDistance

#' @rdname cluster-statistics
#' @export
MeanCentroidDist <- MeanCentroidDistance

.MeanCentroidDist <- function(x, Average = mean) {
  recentred <- t(t(x) - apply(x, 2, mean))
  
  # Return:
  Average(sqrt(rowSums(recentred ^ 2)))
}

#' @rdname cluster-statistics
#' @param Average Function to use to summarize distances. Defaults to `mean`;
#' specifying `median` returns a value akin to the median absolute divergence
#' (see [`mad`]).
#' @return `DistanceFromMedian()` returns a numeric specifying the mean distance
#' of each point (except the median) from the median point of its cluster.
#' @examples DistanceFromMedian(points, cluster)
#' @export
DistanceFromMedian <- function(x, cluster = 1, Average = mean)
  UseMethod("DistanceFromMedian")

#' @rdname cluster-statistics
#' @export
DistFromMed <- DistanceFromMedian

#' @export
DistanceFromMedian.dist <- function(x, cluster = 1, Average = mean) {
  d <- as.matrix(x)
  vapply(seq_along(unique(cluster)),
         function(i) {
           .DistanceFromMedian.dist(
             d[cluster == i, cluster == i, drop = FALSE], Average = Average)
         },
         numeric(1))
}

#' @export
DistanceFromMedian.numeric <- function(x, cluster = 1, Average = mean) {
  if (is.null(dim(x))) {
    warning(paste0("`x` lacks dimensions. ",
                   "Did you subset without specifying `drop = FALSE`?"))
    x <- matrix(x, 1)
  }
  
  # Return:
  vapply(seq_along(unique(cluster)),
         function(i) .DistanceFromMedian(x[cluster == i, , drop = FALSE],
                                         Average = Average),
         numeric(1))
}

.DistanceFromMedian <- function(x, Average = mean) {
  if (dim(x)[[1]] > 1) {
    .DistanceFromMedian.dist(as.matrix(dist(x)), Average = Average)
  } else {
    NA_real_
  }
}

.DistanceFromMedian.dist <- function(d, Average = mean) {
  if (dim(d)[[1]] > 1) {
    medPoint <- which.min(unname(colSums(d)))
    Average(d[medPoint, -medPoint])
  } else {
    NA_real_
  }
}

#' @rdname cluster-statistics
#' @return `MeanNN()` returns a numeric specifying the mean distance from each
#' point within a cluster to its nearest neighbour.
#' @examples MeanNN(points, cluster)
#' @export
MeanNN <- function(x, cluster = 1, Average = mean) UseMethod("MeanNN")

#' @export
MeanNN.dist <- function(x, cluster = 1, Average = mean) {
  d <- as.matrix(x)
  diag(d) <- NA_real_
  vapply(seq_along(unique(cluster)),
         function(i) .MeanNN.dist(d[cluster == i, cluster == i, drop = FALSE],
                                  Average = Average),
         numeric(1))
}

.MeanNN.dist <- function(x, Average = mean) {
  if (dim(x)[[1]] > 1) {
    Average(apply(x, 1, min, na.rm = TRUE))
  } else {
    NA_real_
  }
}

#' @export
MeanNN.numeric <- function(x, cluster = 1, Average = mean) {
  if (is.null(dim(x))) {
    warning(paste0("`x` lacks dimensions. ",
                   "Did you subset without specifying `drop = FALSE`?"))
    x <- matrix(x, 1)
  }
  
  # Return:
  vapply(seq_along(unique(cluster)),
         function(i) .MeanNN(x[cluster == i, , drop = FALSE], Average = Average),
         numeric(1))
}

.MeanNN <- function(x, Average = mean) {
  if (dim(x)[[1]] > 1) {
    distances <- as.matrix(dist(x))
    diag(distances) <- NA_real_
    
    # Return:
    Average(apply(distances, 1, min, na.rm = TRUE))
    
  } else {
    # Return:
    NA_real_
  }
}

#' @rdname cluster-statistics
#' @return `MeanMSTEdge()` returns a numeric specifying the mean length of an
#' edge in the minimum spanning tree of points within each cluster.
#' @examples MeanMSTEdge(points, cluster)
#' @export
MeanMSTEdge <- function(x, cluster = 1) UseMethod("MeanMSTEdge")

#' @export
MeanMSTEdge.dist <- function(x, cluster = 1) {
  d <- as.matrix(x)
  diag(d) <- NA_real_
  vapply(seq_along(unique(cluster)),
         function(i) .MeanMSTEdge.dist(d[cluster == i, cluster == i, drop = FALSE]),
         numeric(1))
}

#' @importFrom stats as.dist
.MeanMSTEdge.dist <- function(x) {
  n <- dim(x)[[1]]
  # Return:
  if (n > 1) {
    MSTLength(as.dist(x)) / (n - 1)
  } else {
    NA_real_
  }
}

#' @export
MeanMSTEdge.numeric <- function(x, cluster = 1) {
  if (is.null(dim(x))) {
    warning(paste0("`x` lacks dimensions. ",
                   "Did you subset without specifying `drop = FALSE`?"))
    x <- matrix(x, 1)
  }
  
  # Return:
  vapply(seq_along(unique(cluster)),
         function(i) .MeanMSTEdge(x[cluster == i, , drop = FALSE]),
         numeric(1))
}

#' @importFrom TreeTools MSTLength
.MeanMSTEdge <- function(x) {
  n <- dim(x)[[1]]
  # Return:
  if (n > 1) {
    MSTLength(dist(x)) / (n - 1)
  } else {
    NA_real_
  }
}

points <- rbind(matrix(1:16, 4),
                rep(1, 4),
                matrix(1:32, 8, 4) / 10)
cluster <- rep(1:3, c(4, 1, 8))

test_that("SumOfRanges()", {
  expect_warning(expect_equal(SumOfRanges(points[cluster == 2, ]), 0),
                 "lacks dimensions")
  expect_equal(SumOfRanges(points, cluster), c(3 * 4, 0, 0.7 * 4))
  expect_equal(
    sapply(1:3, function(i) SumOfRanges(points[cluster == i, , drop = FALSE])),
    SumOfRanges(points, cluster)
  )
})

test_that("SumOfVariances()", {
  expect_warning(expect_equal(SumOfVariances(points[cluster == 2, ]), NA_real_),
                 "lacks dimensions")
  expect_equal(SumOfVars(points, cluster),
               c(var(1:4) * 4, NA_real_, var(1:8 / 10) * 4))
  expect_equal(
    sapply(1:3, function(i) SumOfVars(points[cluster == i, , drop = FALSE])),
    SumOfVariances(points, cluster)
  )
})

test_that("MeanCentroidDistance()", {
  expect_warning(expect_equal(MeanCentroidDistance(points[cluster == 2, ]), 0),
                 "lacks dimensions")
  
  expect_equal(MeanCentroidDistance(points, cluster), c(2, 0, 0.4))
  expect_equal(
    sapply(1:3, function(i) MeanCentDist(points[cluster == i, , drop = FALSE])),
    MeanCentroidDist(points, cluster)
  )
})

test_that("DistanceFromMedian()", {
  expect_warning(
    expect_equal(DistanceFromMedian(points[cluster == 2, ]), NA_real_),
    "lacks dimensions"
  )
  
  expect_equal(DistanceFromMedian(points, cluster),
               c((2 + 2 + 4) / 3,
                 NA_real_,
                 sum(0.6, 0.4, 0.2, 0.2, 0.4, 0.6, 0.8) / 7)
  )
  
  expect_equal(DistanceFromMedian(points, cluster, median),
               c(median(c(2, 2, 4)),
                 NA_real_,
                 median(c(0.6, 0.4, 0.2, 0.2, 0.4, 0.6, 0.8)))
  )
  
  expect_equal(
    sapply(1:3, function(i) DistFromMed(points[cluster == i, , drop = FALSE])),
    DistanceFromMedian(points, cluster)
  )
  expect_equal(DistFromMed(dist(points), cluster),
               DistanceFromMedian(points, cluster))
})

test_that("MeanNN()", {
  expect_warning(expect_equal(MeanNN(points[cluster == 2, ]), NA_real_),
                 "lacks dimensions")
  expect_equal(MeanNN(points, cluster), c(2, NA, 0.2))
  expect_equal(
    sapply(1:3, function(i) MeanNN(points[cluster == i, , drop = FALSE])),
    MeanNN(points, cluster)
  )
  expect_equal(MeanNN(dist(points), cluster), MeanNN(points, cluster))
})

test_that("MeanMSTEdge()", {
  expect_warning(expect_equal(MeanMSTEdge(points[cluster == 2, ]), NA_real_),
                 "lacks dimensions")
  expect_equal(MeanMSTEdge(points, cluster),
               c(2, NA, 0.2))
  expect_equal(
    sapply(1:3, function(i) MeanMSTEdge(points[cluster == i, , drop = FALSE])),
    MeanMSTEdge(points, cluster)
  )
  expect_equal(MeanMSTEdge(dist(points), cluster), MeanMSTEdge(points, cluster))
})

#' Visualize a matching
#' 
#' Depict the splits that are matched between two trees using a specified 
#' [Generalized Robinson&ndash;Foulds](
#' https://ms609.github.io/TreeDist/articles/Generalized-RF.html)
#' similarity measure.
#' 
#' Note that when visualizing a Robinson&ndash;Foulds distance (using 
#' `Func = RobinsonFouldsMatching`), matched splits are assigned a _similarity_
#' score of 1, which is deducted from the total number of splits to calculate 
#' the Robinson&ndash;Foulds _distance_.  Unmatched splits thus contribute one to 
#' total tree distance.
#' 
#' @param Func Function used to construct tree similarity.
#' @param tree1,tree2 Trees of class `phylo`, with identical leaf labels.
#' @param setPar Logical specifying whether graphical parameters should be 
#' set to display trees side by side.
#' @param precision Integer specifying number of significant figures to display
#' when reporting matching scores.
#' @param Plot Function to use to plot trees.
#' @param matchZeros Logical specifying whether to pair splits with zero
#' similarity (`TRUE`), or leave them unpaired (`FALSE`).
#' @param plainEdges Logical specifying whether to plot edges with a uniform
#' width and colour (`TRUE`), or whether to draw edge widths according to the
#' similarity of the associated splits (`FALSE`).
#' @param edge.cex Character expansion for edge labels.
#' If `FALSE`, suppress edge labels.
#' @param value.cex Character expansion for values on edge labels.
#' If `FALSE`, values are not displayed.
#' @param edge.frame Character specifying the kind of frame to be printed around
#' the text of the edge labels.  Choose an abbreviation of `"rect"`, `"circle"`,
#' or `"none"`.
#' @param edge.width,edge.color,\dots Additional parameters to send to `Plot()`.
#' 
#' @returns `VisualizeMatching()` invisibly returns the matching of splits
#' between `tree1` and `tree2` (i.e. 
#' `Func(tree1, tree2, reportMatching = TRUE)`)
#' 
#' @examples 
#' tree1 <- TreeTools::BalancedTree(6)
#' tree2 <- TreeTools::PectinateTree(6)
#' 
#' VisualizeMatching(RobinsonFouldsMatching, tree1, tree2)
#' matching <- VisualizeMatching(SharedPhylogeneticInfo, tree1, tree2,
#'                               matchZeros = FALSE)
#' attributes(matching)
#' @template MRS
#' @encoding UTF-8
#' @importFrom ape nodelabels edgelabels plot.phylo
#' @importFrom colorspace qualitative_hcl sequential_hcl
#' @importFrom graphics par
#' @importFrom TreeTools as.Splits
#' @aliases VisualiseMatching PlotMatching DisplayMatching
#' @export
VisualizeMatching <- function (Func, tree1, tree2, setPar = TRUE,
                               precision = 3L, Plot = plot.phylo,
                               matchZeros = TRUE, plainEdges = FALSE,
                               edge.cex = par("cex"),
                               value.cex = edge.cex * 0.8,
                               edge.frame = "rect",
                               edge.width = 1, edge.color = "black",
                               ...) 
{
  splits1 <- as.Splits(tree1)
  edge1 <- tree1[["edge"]]
  child1 <- edge1[, 2]
  nTip <- attr(splits1, "nTip")
  splitEdges1 <- vapply(as.integer(rownames(splits1)),
                        function(node) which(child1 == node), integer(1))
  
  splits2 <- as.Splits(tree2, tipLabels = tree1)
  edge2 <- tree2[["edge"]]
  child2 <- edge2[, 2]
  splitEdges2 <- vapply(as.integer(rownames(splits2)),
                        function(node) which(child2 == node), integer(1))
  
  matching <- Func(tree1, tree2, reportMatching = TRUE)
  pairings <- attr(matching, "matching")
  scores <- attr(matching, "pairScores")
  pairScores <- signif(mapply(function(i, j) scores[i, j],
                              seq_along(pairings), pairings), precision)
  
  faint <- "#aaaaaa"
  
  if (setPar) {
    origPar <- par(mfrow = c(1, 2), mar = rep(0.5, 4))
    on.exit(par(origPar))
  }
  
  .LabelEdge <- function(label, edges, frame = "n", ...) {
    if (edge.cex) {
      edgelabels(text = label, edge = edges, frame = frame,
                 cex = edge.cex, adj = c(0.5, -0.2), ...)
    }
  }
  .LabelValue <- function(label, edges, frame = "n", ...) {
    if (value.cex) {
      edgelabels(text = label, edge = edges, frame = frame,
                 cex = value.cex, adj = c(0.5, 1.1), ...)
    }
  }
  
  .LabelUnpaired <- function(splitEdges, unpaired) {
    if (any(unpaired)) {
      .LabelEdge(label = expression("-"), edges = splitEdges[unpaired],
                 frame = "n", col = faint)
      .LabelValue(label = "0", edges = splitEdges[unpaired],
                  frame = "n", col = faint)
    }
  }
  
  if (plainEdges) {
    Plot(tree1, edge.width = edge.width, edge.color = edge.color, ...)
  } else {
    Normalize <- function(scores, na.rm = FALSE) {
      if (length(scores) == 0) return(scores)
      if (na.rm) {
        scores <- scores[!is.na(scores)]
      } else {
        scores[is.na(scores)] <- 0
      }
      if (any(scores < 0)) {
        stop("Negative scores not supported")                                   # nocov
      }
      if (max(scores) == 0) {
        return(scores)
      }
      if (min(scores) == max(scores)) {
        return(rep(1L, length(scores)))
      }
      scores / max(scores)
    }
    
    OtherRootEdge <- function(splitNodes, edge) {
      parent <- edge[, 1]
      child <- edge[, 2]
      rootEdges <- which(parent == min(parent))
      rootChildren <- child[rootEdges]
      treeIsRooted <- length(rootChildren) < 3
      if (treeIsRooted) {
        splitEdges <- vapply(splitNodes, match, table = child, 0)
        got <- rootChildren %in% splitNodes
        if (any(got)) {
          if (sum(got) != 1) {
            warning("Unexpected polytomy")                                      # nocov
          }
          c(score = as.integer(which(splitNodes %in% rootChildren[got])),
            edge = rootEdges[!got])
        } else {
          # `edge` is not a root edge
          c(score = NA_integer_, edge = NA_integer_)
        }
      } else {
        # Tree is unrooted => there is no root edge at all
        c(score = NA_integer_, edge = NA_integer_)
      }
    }
    edgeColPalette <- sequential_hcl(n = 256L, palette = "Viridis")
    
    EdgyPlot <- function(tree, splits, edge, splitEdges,
                         normalizedScores, ...) {
      splitNodes <- as.integer(names(splits))
      ore <- OtherRootEdge(splitNodes, edge)
      if (length(normalizedScores) && !is.na(ore[[1]])) {
        ns <- c(normalizedScores, normalizedScores[ore["score"]])
        se <- c(splitEdges, ore[[2]])
      } else {
        ns <- normalizedScores
        se <- splitEdges
      }
      
      edge.width <- rep(1, nrow(edge))
      edge.width[se] <- 1 + (10 * ns)
      edge.color <- rep("black", nrow(edge))
      edge.color[se] <- edgeColPalette[1 + ceiling(255 * ns)]
      
      Plot(tree, edge.width = edge.width, edge.color = edge.color, ...)
    }
    
    EdgyPlot(tree1, splits1, edge1, splitEdges1, Normalize(pairScores), ...)
  }
  paired1 <- if (matchZeros) {
    !is.na(pairings)
  } else {
    !is.na(pairings) & pairScores > 0
  }
  palette <- qualitative_hcl(sum(paired1), c = 42, l = 88)
  pairedPairScores <- pairScores[paired1]
  pairLabels <- seq_len(sum(paired1))
  if (any(pairLabels)) {
    .LabelEdge(pairLabels, splitEdges1[paired1], frame = edge.frame,
               bg = palette)
    .LabelValue(pairedPairScores, splitEdges1[paired1],
                col = ifelse(pairedPairScores, "black", faint))
  }
  .LabelUnpaired(splitEdges1, !paired1)
  
  
  paired2 <- seq_along(splitEdges2) %in% pairings[paired1]
  pairNames2 <- pairings[paired1]
  
  if (plainEdges) {
    Plot(tree2, edge.width = edge.width, edge.color = edge.color, ...)
  } else {
    EdgyPlot(tree2, splits2[[pairNames2]], edge2, splitEdges2[pairNames2],
             Normalize(pairedPairScores, na.rm = TRUE), ...)
  }
  if (any(pairLabels)) {
    .LabelEdge(pairLabels, splitEdges2[pairNames2], frame = edge.frame,
               bg = palette)
    .LabelValue(pairedPairScores, splitEdges2[pairNames2],
               col = ifelse(pairedPairScores, "black", faint))
  }
  .LabelUnpaired(splitEdges2, !paired2)
  
  # Return:
  invisible(matching)
}

#' @export
PlotMatching <- VisualizeMatching
#' @export
DisplayMatching <- VisualizeMatching
#' @export
VisualiseMatching <- VisualizeMatching

#' Maximum Agreement Subtree size
#' 
#' Calculate the size or phylogenetic information content 
#' \insertCite{Steel2006}{TreeDist}
#' of the maximum agreement subtree between two phylogenetic trees, i.e.
#' the largest tree that can be obtained from both `tree1` and `tree2` by
#' deleting, but not rearranging, leaves, using the algorithm of
#' \insertCite{Valiente2009;textual}{TreeDist}.
#' 
#' Implemented for trees with up to 4096 tips.  Contact the maintainer if you
#' need to process larger trees.
#' 
#' @param tree1,tree2 Trees of class `phylo`, or lists of such trees to undergo
#' pairwise comparison.
#' @param rooted Logical specifying whether to treat the trees as rooted.
#' 
#' @return `MASTSize()` returns an integer specifying the number of leaves in
#' the maximum agreement subtree.
#' 
#' @examples
#'  # for as.phylo, BalancedTree, PectinateTree:
#' library("TreeTools", quietly = TRUE)
#'
#' MASTSize(PectinateTree(8), BalancedTree(8))
#' MASTInfo(PectinateTree(8), BalancedTree(8))
#' 
#' MASTSize(BalancedTree(7), as.phylo(0:3, 7))
#' MASTSize(as.phylo(0:3, 7), PectinateTree(7))
#' 
#' MASTInfo(BalancedTree(7), as.phylo(0:3, 7))
#' MASTInfo(as.phylo(0:3, 7), PectinateTree(7))
#'
#' MASTSize(list(Bal = BalancedTree(7), Pec = PectinateTree(7)),
#'          as.phylo(0:3, 7))
#' MASTInfo(list(Bal = BalancedTree(7), Pec = PectinateTree(7)),
#'          as.phylo(0:3, 7))
#' 
#' CompareAll(as.phylo(0:4, 8), MASTSize)
#' CompareAll(as.phylo(0:4, 8), MASTInfo)
#' @template MRS
#' 
#' @seealso [`phangorn::mast()`], a slower implementation that also lists the
#' leaves contained within the subtree.
#' 
#' @references \insertAllCited{}
#' 
#' @family tree distances
#' @export
MASTSize <- function(tree1, tree2 = tree1, rooted = TRUE) {
  .TreeDistance(.MASTSizeSingle, tree1, tree2, rooted = rooted,
                # Checks not necessary, as tip labels need not match.
                checks = FALSE)
}

#' @importFrom ape drop.tip
#' @importFrom TreeTools Postorder RenumberTips TreeIsRooted RootOnNode
.MASTSizeSingle <- function(tree1, tree2, rooted = TRUE,
                             tipLabels = tree1[["tip.label"]],
                             ...) {
  label1 <- tipLabels
  label2 <- tree2[["tip.label"]]
  
  tree1 <- drop.tip(tree1, setdiff(label1, label2))
  label1 <- tree1[["tip.label"]]
  tree2 <- RenumberTips(drop.tip(tree2, setdiff(label2, label1)), label1)
  
  nTip <- length(label1)
  
  if (!rooted) {
    if (!TreeIsRooted(tree1)) {
      tree1 <- RootOnNode(tree1, node = tree1[["edge"]][nTip + nTip - 2L], TRUE)
    }
    postorderEdge1 <- Postorder(tree1[["edge"]])
    tree2 <- Preorder(tree2)
    max(vapply(tree2[["edge"]][, 2], function(node)
      .MASTSizeEdges(postorderEdge1,
                     RootOnNode(tree2, node = node, TRUE)[["edge"]],
                     nTip = nTip), 0L))
  } else {
    if (!TreeIsRooted(tree1) || !TreeIsRooted(tree2)) {
      stop("Both trees must be rooted if rooted = TRUE")
    }
    .MASTSizeEdges(Postorder(tree1[["edge"]]), tree2[["edge"]], nTip)
  }
}

#' Calculate MAST size from edge matrices.
#' 
#' Internal function.
#' 
#' @param edge1 Edge matrix of tree 1. **Must be in postorder!**
#' @param edge2 Edge matrix of tree 2.
#' @param nTip Integer specifying the number of leaves in each split.
#' @keywords internal
.MASTSizeEdges <- function(edge1, edge2, nTip) {
  if (nrow(edge1) != nrow(edge2)) {
    stop("Both trees must contain the same number of edges.")
  }
  if (nTip > 4096L) {
    stop("Tree too large; please contact maintainer for advice.")
  }
  cpp_mast(edge1 - 1L, Postorder(edge2) - 1L, nTip)
}

#' @rdname MASTSize
#' @return `MASTInfo()` returns a vector or matrix listing the phylogenetic
#' information content, in bits, of the maximum agreement subtree.
#' @importFrom TreeTools Log2Rooted.int Log2Unrooted.int
#' @export
MASTInfo <- function(tree1, tree2 = tree1, rooted = TRUE) {
  size <- MASTSize(tree1, tree2, rooted = rooted)
  ret <- if (rooted) Log2Rooted.int(size) else Log2Unrooted.int(size)
  if (!is.null(attributes(size))) attributes(ret) <- attributes(size)
  # Return:
  ret
}

library("TreeTools")

test_that("MAST fails gracefully", {
  # R-level guard catches edge count mismatch
  expect_error(.MASTSizeEdges(BalancedTree(7)$edge, BalancedTree(8)$edge, 7),
               "same number of edges")
  expect_error(MASTSize(BalancedTree(8), UnrootTree(BalancedTree(8))))
  expect_error(MASTSize(BalancedTree(10000), PectinateTree(10000))) # Too large
  
})

test_that("MAST works", {
  tree1 <- BalancedTree(8L)
  tree2 <- PectinateTree(8L)
  expect_equal(8L, MASTSize(tree1, tree1, rooted = TRUE))
  expect_equal(8L, MASTSize(tree1, tree1, rooted = FALSE))
  expect_equal(4L, MASTSize(tree1, tree2, rooted = TRUE))
  expect_equal(6L, MASTSize(tree1, tree2, rooted = FALSE))
  
  expect_equal(MASTSize(BalancedTree(7), as.phylo(0:3, 7)),
               MASTSize(as.phylo(0:3, 7), BalancedTree(7)))
  
  expect_equal(MASTSize(list(BalancedTree(7), PectinateTree(7)), as.phylo(0:3, 7))[1, ],
               MASTSize(as.phylo(0:3, 7), BalancedTree(7)))
  
  expect_equal(MASTInfo(BalancedTree(7), as.phylo(0:3, 7)),
               MASTInfo(as.phylo(0:3, 7), BalancedTree(7)))
  
  expect_equal(MASTInfo(list(BalancedTree(7), PectinateTree(7)), as.phylo(0:3, 7))[1, ],
               MASTInfo(as.phylo(0:3, 7), BalancedTree(7)))
  
})

test_that("MAST supports funnily-ordered edges", {
  tree1 <- BalancedTree(8L)
  tree2 <- PectinateTree(8L)
  tree1$edge <- tree1$edge[c(1:7 * 2, (7:1 * 2) - 1), ]
  tree2$edge <- tree2$edge[c(1:7 * 2, (7:1 * 2) - 1), ]
  expect_equal(6L, MASTSize(Postorder(tree1), Postorder(tree2), rooted = FALSE))
  expect_equal(6L, MASTSize(tree1, tree2, rooted = FALSE))
})

test_that("MAST size calculated correctly on small trees", {
  library("TreeTools")
  #expect_equal(4L, MASTSize(as.phylo(0, 5), as.phylo(1, 5)))
  t1 <- structure(list(edge = structure(c(6L, 6L, 7L, 7L, 8L, 8L, 9L, 
                                          9L, 1L, 7L, 2L, 8L, 3L, 9L, 4L, 5L),
                                        .Dim = c(8L, 2L)), Nnode = 4L, 
                       tip.label = c("t3", "t5", "t4", "t1", "t2")),
                  class = "phylo", order = "cladewise")
  t2 <- structure(list(edge = structure(c(6L, 9L, 9L, 7L, 7L, 8L, 8L, 
                                          6L, 9L, 2L, 7L, 3L, 8L, 4L, 5L, 1L), 
                                        .Dim = c(8L, 2L)), 
                       tip.label = c("t3", "t4", "t1", "t2", "t5"), Nnode = 4L), 
                  class = "phylo", order = "cladewise")
  t1 <- RenumberTips(t1, paste0("t", 1:5))
  
  t2 <- RenumberTips(t2, t1)
  t2 <- Preorder(t2)

  ME <- function(e, node) {
    expect_equal(e, 
                 .MASTSizeEdges(Postorder(t1$edge),
                                RootOnNode(t2, node = node, TRUE)$edge,
                                nTip = 5L))
  }
  
  ME(2L, 1)
  ME(3L, 2)
  ME(4L, 3)
  ME(2L, 4)
  ME(2L, 5)
  ME(4L, 6)
  ME(4L, 7)
  ME(3L, 8)
  ME(3L, 9)
  
  expect_equal(4L, MASTSize(t1, t2, rooted = FALSE))
})

#' Transfer dissimilarity between phylogenetic trees
#'
#' Compute the transfer dissimilarity between phylogenetic trees, as defined
#' by \insertCite{Takazawa2026;textual}{TreeDist}.
#' The transfer dissimilarity uses the transfer distance
#' \insertCite{Lemoine2018}{TreeDist} to compare bipartitions, providing a
#' finer-grained measure than the Robinson--Foulds distance.  Each split in
#' each tree is scored by how many taxa must be moved to match its closest
#' counterpart in the other tree, and these scores are summed.
#'
#' The `scaled` variant divides each split's contribution by its depth minus
#' one, giving equal weight to all splits regardless of their depth (analogous
#' to the Robinson--Foulds distance).  The `unscaled` variant uses raw transfer
#' distances, giving more weight to deep splits.  Neither variant satisfies
#' the triangle inequality for trees with six or more tips.
#'
#' @inheritParams TreeDistance
#' @param scale Logical; if `TRUE` (default), use the scaled transfer
#'   dissimilarity.  If `FALSE`, use the unscaled transfer dissimilarity.
#'
#' @return `TransferDist()` returns an object of class `dist` (if `tree2` is
#'   `NULL`), a numeric matrix (if both `tree1` and `tree2` are lists), or a
#'   numeric value (for a single pair).  If `reportMatching = TRUE`, the
#'   return value carries `matching` and `pairScores` attributes.
#'
#' @section Normalization:
#'
#' When `normalize = TRUE`, the scaled transfer dissimilarity is divided by
#' `2 * (n - 3)`, placing it in the range \[0, 1\].  The unscaled version is
#' divided by the maximum possible unscaled dissimilarity
#' (following \insertCite{Takazawa2026;textual}{TreeDist}).
#'
#' @examples
#' library(TreeTools)
#' TransferDist(BalancedTree(8), PectinateTree(8))
#' TransferDist(BalancedTree(8), PectinateTree(8), scale = FALSE)
#'
#' # All-pairs
#' TransferDist(as.phylo(0:5, 8))
#'
#' @references
#' \insertAllCited{}
#'
#' @family tree distances
#'
#' @importFrom TreeTools as.Splits TipLabels NSplits
#' @export
TransferDist <- function(tree1, tree2 = NULL, scale = TRUE,
                         normalize = FALSE, reportMatching = FALSE) {
  
  # --- Fast path: all-pairs (tree2 = NULL) ---
  if (is.null(tree2) && !reportMatching) {
    fast <- .TransferDistAllPairs(tree1, scale)
    if (!is.null(fast)) {
      if (!isFALSE(normalize)) {
        nTip <- length(TipLabels(tree1[[1]]))
        denom <- .TransferNormDenom(nTip, scale)
        fast <- fast / denom
      }
      return(fast)
    }
  }
  
  # --- Fast path: cross-pairs ---
  if (!is.null(tree2) && !reportMatching) {
    fast <- .TransferDistCrossPairs(tree1, tree2, scale)
    if (!is.null(fast)) {
      if (!isFALSE(normalize)) {
        nTip <- length(TipLabels(
          if (inherits(tree1, c("phylo", "Splits"))) tree1 else tree1[[1]]))
        denom <- .TransferNormDenom(nTip, scale)
        fast <- fast / denom
      }
      return(fast)
    }
  }
  
  # --- Generic fallback via CalculateTreeDistance ---
  # Capture `scale` in the closure for the Splits-level function
  Func <- function(splits1, splits2, nTip, reportMatching = FALSE) {
    TransferDistSplits(splits1, splits2, nTip = nTip,
                       scale = scale, reportMatching = reportMatching)
  }
  unnormalized <- CalculateTreeDistance(Func, tree1, tree2, reportMatching)
  
  if (!isFALSE(normalize)) {
    nTip <- length(TipLabels(
      if (inherits(tree1, c("phylo", "Splits"))) tree1 else tree1[[1]]))
    denom <- .TransferNormDenom(nTip, scale)
    unnormalized <- unnormalized / denom
  }
  
  unnormalized
}

#' @rdname TransferDist
#' @export
TransferDistance <- TransferDist

#' @rdname TransferDist
#' @inheritParams SharedPhylogeneticInfoSplits
#' @export
TransferDistSplits <- function(splits1, splits2,
                                nTip = attr(splits1, "nTip"),
                                scale = TRUE,
                                reportMatching = FALSE) {
  .ValidateSplitArgs(splits1, splits2, nTip)
  solution <- cpp_transfer_dist_scored(splits1, splits2,
                                        nTip = as.integer(nTip),
                                        scale = scale)
  ret <- solution[["score"]]
  
  if (reportMatching) {
    nSplits1 <- nrow(splits1)
    nSplits2 <- nrow(splits2)
    matching <- solution[["matching"]]
    matching[matching > nSplits2 | matching == 0L] <- NA
    if (nSplits1 < nSplits2) {
      matching <- matching[seq_len(nSplits1)]
    }
    attr(ret, "matching") <- matching
    
    # Compute full pairwise score matrix for reportMatching
    pairScores <- matrix(0, nSplits1, nSplits2)
    for (i in seq_len(nSplits1)) {
      for (j in seq_len(nSplits2)) {
        # Per-pair: the transfer distance contribution
        # This is the individual δ(b_i, b*_j) / (depth(b_i) - 1) for scaled
        # or min(δ(b_i, b*_j), depth(b_i) - 1) for unscaled
        pair_res <- cpp_transfer_dist_scored(
          splits1[i, , drop = FALSE],
          splits2[j, , drop = FALSE],
          nTip = as.integer(nTip),
          scale = scale
        )
        pairScores[i, j] <- pair_res[["score"]]
      }
    }
    
    if (!is.null(attr(splits1, "tip.label"))) {
      matched1 <- !is.na(matching)
      matched2 <- matching[matched1]
      matched1 <- which(matched1)
      attr(ret, "matchedSplits") <-
        ReportMatching(splits1[[matched1]], splits2[[matched2]],
                       realMatch = rep(TRUE, length(matched1)))
    }
    
    attr(ret, "matchedScores") <- pairScores[
      matrix(c(seq_along(matching), matching), ncol = 2L)]
    attr(ret, "pairScores") <- pairScores
  }
  
  ret
}


# ============================================================================
# Internal helpers
# ============================================================================

# All-pairs fast path
.TransferDistAllPairs <- function(tree1, scale) {
  if (inherits(tree1, c("phylo", "Splits"))) return(NULL)
  if (length(tree1) < 2L) return(NULL)
  
  tipLabels <- TipLabels(tree1[[1]])
  if (is.null(tipLabels)) return(NULL)
  nTip <- length(tipLabels)
  if (nTip < 4L) return(NULL)
  .CheckMaxTips(nTip)
  
  # Check all trees share same tip set
  allLabels <- TipLabels(tree1)
  if (is.list(allLabels)) {
    if (!all(vapply(allLabels[-1], setequal, logical(1), tipLabels))) {
      return(NULL)
    }
  }
  
  splitsList <- lapply(tree1, function(tr) {
    unclass(as.Splits(tr, tipLabels))
  })
  
  nThreads <- max(1L, getOption("TreeDist.threads",
                                 getOption("mc.cores", 1L)))
  
  dists <- cpp_transfer_dist_all_pairs(splitsList, nTip, scale, nThreads)
  
  N <- length(tree1)
  attr(dists, "Size") <- N
  attr(dists, "Diag") <- FALSE
  attr(dists, "Upper") <- FALSE
  nms <- names(tree1)
  if (!is.null(nms)) attr(dists, "Labels") <- nms
  class(dists) <- "dist"
  dists
}


# Cross-pairs fast path
.TransferDistCrossPairs <- function(tree1, tree2, scale) {
  single1 <- inherits(tree1, c("phylo", "Splits"))
  single2 <- inherits(tree2, c("phylo", "Splits"))
  if (single1 && single2) return(NULL) # use generic path for single pair
  
  trees1 <- if (single1) list(tree1) else tree1
  trees2 <- if (single2) list(tree2) else tree2
  
  tipLabels <- TipLabels(trees1[[1]])
  if (is.null(tipLabels)) return(NULL)
  nTip <- length(tipLabels)
  if (nTip < 4L) return(NULL)
  .CheckMaxTips(nTip)
  
  # Check all trees share same tip set
  allLabels1 <- TipLabels(trees1)
  allLabels2 <- TipLabels(trees2)
  if (is.list(allLabels1)) {
    if (!all(vapply(allLabels1, setequal, logical(1), tipLabels))) return(NULL)
  }
  if (is.list(allLabels2)) {
    if (!all(vapply(allLabels2, setequal, logical(1), tipLabels))) return(NULL)
  } else {
    if (!setequal(allLabels2, tipLabels)) return(NULL)
  }
  
  splits1 <- lapply(trees1, function(tr) unclass(as.Splits(tr, tipLabels)))
  splits2 <- lapply(trees2, function(tr) unclass(as.Splits(tr, tipLabels)))
  
  nThreads <- max(1L, getOption("TreeDist.threads",
                                 getOption("mc.cores", 1L)))
  
  mat <- cpp_transfer_dist_cross_pairs(splits1, splits2, nTip, scale, nThreads)
  
  rownames(mat) <- names(trees1)
  colnames(mat) <- names(trees2)
  
  # If one input was a single tree, simplify to vector
  if (single1) return(mat[1, ])
  if (single2) return(mat[, 1])
  mat
}


# Normalization denominator
.TransferNormDenom <- function(nTip, scale) {
  if (scale) {
    # Scaled: each tree contributes at most (n-3) branches × 1.0
    2.0 * (nTip - 3L)
  } else {
    # Unscaled: maximum dissimilarity between two caterpillar trees
    # Takazawa 2026: (n^2 - 2n + 4)/4 for even n, (n^2 - 2n + 5)/4 for odd n
    if (nTip %% 2L == 0L) {
      (nTip^2 - 2 * nTip + 4) / 4
    } else {
      (nTip^2 - 2 * nTip + 5) / 4
    }
  }
}

test_that("TransferDist() returns 0 for identical trees", {
  tr <- TreeTools::BalancedTree(8)
  expect_equal(0, TransferDist(tr, tr))
  expect_equal(0, TransferDist(tr, tr, scale = FALSE))
})


test_that("TransferDist() is symmetric", {
  bal8 <- TreeTools::BalancedTree(8)
  pec8 <- TreeTools::PectinateTree(8)
  expect_equal(TransferDist(bal8, pec8), TransferDist(pec8, bal8))
  expect_equal(TransferDist(bal8, pec8, scale = FALSE),
               TransferDist(pec8, bal8, scale = FALSE))
})


test_that("TransferDist() scaled is bounded by RF", {
  # The scaled transfer dissimilarity is bounded above by the RF distance
  bal8 <- TreeTools::BalancedTree(8)
  pec8 <- TreeTools::PectinateTree(8)
  td_scaled <- TransferDist(bal8, pec8)
  rf <- RobinsonFoulds(bal8, pec8)
  expect_lte(td_scaled, rf)
})


test_that("TransferDist() normalization", {
  bal8 <- TreeTools::BalancedTree(8)
  pec8 <- TreeTools::PectinateTree(8)

  # Scaled normalization denominator = 2 * (n - 3) = 10
  td_raw <- TransferDist(bal8, pec8)
  td_norm <- TransferDist(bal8, pec8, normalize = TRUE)
  expect_equal(td_norm, td_raw / (2 * (8 - 3)))
  expect_true(td_norm >= 0)
  expect_true(td_norm <= 1)

  # Unscaled normalization
  td_raw_u <- TransferDist(bal8, pec8, scale = FALSE)
  td_norm_u <- TransferDist(bal8, pec8, scale = FALSE, normalize = TRUE)
  # Even n: denom = (n^2 - 2n + 4) / 4 = (64 - 16 + 4) / 4 = 13
  expect_equal(td_norm_u, td_raw_u / 13)
})


test_that("TransferDist() all-pairs", {
  trees <- as.phylo(0:5, 8)

  # All-pairs
  d <- TransferDist(trees)
  expect_s3_class(d, "dist")
  expect_equal(attr(d, "Size"), 6L)

  # Check diagonal = 0 implicitly (dist objects don't store diagonal)
  # Check that batch matches pairwise
  dm <- as.matrix(d)
  for (i in 1:5) {
    for (j in (i + 1):6) {
      expected <- TransferDist(trees[[i]], trees[[j]])
      expect_equal(dm[i, j], expected, tolerance = 1e-12,
                   info = paste("pair", i, j))
    }
  }
})


test_that("TransferDist() cross-pairs", {
  trees1 <- as.phylo(0:2, 8)
  trees2 <- as.phylo(3:5, 8)

  mat <- TransferDist(trees1, trees2)
  expect_equal(dim(mat), c(3, 3))

  # Check matches pairwise
  for (i in 1:3) {
    for (j in 1:3) {
      expected <- TransferDist(trees1[[i]], trees2[[j]])
      expect_equal(mat[i, j], expected, tolerance = 1e-12,
                   info = paste("cross", i, j))
    }
  }
})


test_that("TransferDist() single tree vs list", {
  tr <- TreeTools::BalancedTree(8)
  trees <- as.phylo(0:3, 8)

  d1 <- TransferDist(tr, trees)
  d2 <- TransferDist(trees, tr)

  expect_length(d1, 4)
  expect_length(d2, 4)

  for (i in seq_along(trees)) {
    expected <- TransferDist(tr, trees[[i]])
    expect_equal(d1[i], expected, tolerance = 1e-12)
    expect_equal(d2[i], expected, tolerance = 1e-12)
  }
})


test_that("TransferDist() scaled vs unscaled relationship", {
  bal8 <- TreeTools::BalancedTree(8)
  pec8 <- TreeTools::PectinateTree(8)

  # Both should be non-negative
  td_s <- TransferDist(bal8, pec8, scale = TRUE)
  td_u <- TransferDist(bal8, pec8, scale = FALSE)
  expect_true(td_s >= 0)
  expect_true(td_u >= 0)

  # For identical trees, both = 0
  expect_equal(0, TransferDist(bal8, bal8, scale = TRUE))
  expect_equal(0, TransferDist(bal8, bal8, scale = FALSE))
})


test_that("TransferDist() reportMatching", {
  bal8 <- TreeTools::BalancedTree(8)
  pec8 <- TreeTools::PectinateTree(8)

  res <- TransferDist(bal8, pec8, reportMatching = TRUE)
  expect_equal(res[[1]], TransferDist(bal8, pec8))

  matching <- attr(res, "matching")
  expect_true(!is.null(matching))

  pairScores <- attr(res, "pairScores")
  expect_true(!is.null(pairScores))
  expect_equal(nrow(pairScores), TreeTools::NSplits(bal8))
  expect_equal(ncol(pairScores), TreeTools::NSplits(pec8))
})


test_that("TransferDist() handles star trees", {
  star <- TreeTools::StarTree(8)
  bal8 <- TreeTools::BalancedTree(8)

  # Star tree has no internal splits; dissimilarity = sum over binary tree only
  td <- TransferDist(bal8, star)
  expect_true(td >= 0)

  # Star vs star = 0
  expect_equal(0, TransferDist(star, star))
})


test_that("TransferDist() handles small trees", {
  # 4 tips: minimal non-trivial case
  tr1 <- TreeTools::BalancedTree(4)
  tr2 <- TreeTools::PectinateTree(4)

  td <- TransferDist(tr1, tr2)
  expect_true(is.finite(td))
  expect_true(td >= 0)
})


test_that("TransferDist() consistent with TransferConsensus internals", {
  skip_if_not_installed("TreeTools")

  # The transfer dissimilarity for a single pair should be consistent with
  # how TransferConsensus computes distances internally
  set.seed(6419)
  trees <- structure(lapply(1:10, function(i) ape::rtree(12)),
                     class = "multiPhylo")

  # All-pairs: check symmetry and non-negativity
  d <- TransferDist(trees)
  dm <- as.matrix(d)
  expect_true(all(dm >= -1e-12))
  expect_true(isSymmetric(dm, tol = 1e-12))
})


# =========================================================================
# Additional coverage: TransferDistance alias
# =========================================================================

test_that("TransferDistance() is an alias for TransferDist()", {
  bal8 <- TreeTools::BalancedTree(8)
  pec8 <- TreeTools::PectinateTree(8)
  expect_equal(TransferDistance(bal8, pec8), TransferDist(bal8, pec8))
  expect_equal(TransferDistance(bal8, pec8, scale = FALSE),
               TransferDist(bal8, pec8, scale = FALSE))
})


# =========================================================================
# cpp_transfer_dist() direct (both scaled + unscaled + matching)
# =========================================================================

test_that("cpp_transfer_dist() returns both scores and matchings", {
  bal8 <- TreeTools::BalancedTree(8)
  pec8 <- TreeTools::PectinateTree(8)
  tipLabels <- TreeTools::TipLabels(bal8)

  sp1 <- unclass(as.Splits(bal8, tipLabels))
  sp2 <- unclass(as.Splits(pec8, tipLabels))

  res <- TreeDist:::cpp_transfer_dist(sp1, sp2, 8L)

  expect_true(res$score_scaled >= 0)
  expect_true(res$score_unscaled >= 0)
  expect_length(res$matching_xy, nrow(sp1))
  expect_length(res$matching_yx, nrow(sp2))

  # Consistent with TransferDist
  expect_equal(res$score_scaled,
               TransferDist(bal8, pec8, scale = TRUE), tolerance = 1e-12)
  expect_equal(res$score_unscaled,
               TransferDist(bal8, pec8, scale = FALSE), tolerance = 1e-12)

  # Identical trees → distance 0, all matchings defined
  res_id <- TreeDist:::cpp_transfer_dist(sp1, sp1, 8L)
  expect_equal(res_id$score_scaled, 0)
  expect_equal(res_id$score_unscaled, 0)
})

test_that("cpp_transfer_dist() handles star trees", {
  star <- TreeTools::StarTree(8)
  bal8 <- TreeTools::BalancedTree(8)
  tipLabels <- TreeTools::TipLabels(bal8)

  sp_star <- unclass(as.Splits(star, tipLabels))
  sp_bal <- unclass(as.Splits(bal8, tipLabels))

  # Star vs binary: only one direction contributes
  res <- TreeDist:::cpp_transfer_dist(sp_bal, sp_star, 8L)
  expect_true(res$score_scaled >= 0)
  expect_true(res$score_unscaled >= 0)
  expect_length(res$matching_xy, nrow(sp_bal))
  # matching_yx has 0 elements for star tree
  expect_length(res$matching_yx, 0)

  # Star vs star: both scores = 0
  res_ss <- TreeDist:::cpp_transfer_dist(sp_star, sp_star, 8L)
  expect_equal(res_ss$score_scaled, 0)
  expect_equal(res_ss$score_unscaled, 0)
})

test_that("column mismatch caught by R-level guard", {
  sp1 <- matrix(as.raw(0), nrow = 1, ncol = 1)
  sp2 <- matrix(as.raw(0), nrow = 1, ncol = 2)
  expect_error(
    TreeDist:::.ValidateSplitArgs(sp1, sp2, 8L),
    "same number of tips"
  )
})


# =========================================================================
# cpp_transfer_dist_scored() edge cases
# =========================================================================

test_that("column mismatch caught by R-level guard (scored)", {
  sp1 <- matrix(as.raw(0), nrow = 1, ncol = 1)
  sp2 <- matrix(as.raw(0), nrow = 1, ncol = 2)
  expect_error(
    TreeDist:::.ValidateSplitArgs(sp1, sp2, 8L),
    "same number of tips"
  )
})

test_that("cpp_transfer_dist_scored() with asymmetric split counts", {
  bal8 <- TreeTools::BalancedTree(8)
  pec8 <- TreeTools::PectinateTree(8)
  tipLabels <- TreeTools::TipLabels(bal8)

  sp_bal <- unclass(as.Splits(bal8, tipLabels))
  sp_pec <- unclass(as.Splits(pec8, tipLabels))

  # BalancedTree(8) has fewer splits than PectinateTree(8)
  res <- TreeDist:::cpp_transfer_dist_scored(sp_bal, sp_pec, 8L, TRUE)
  expect_true(res$score >= 0)
  # matching length = max(nSplits1, nSplits2)
  expect_equal(length(res$matching),
               max(nrow(sp_bal), nrow(sp_pec)))
})


# =========================================================================
# Normalization edge cases
# =========================================================================

test_that(".TransferNormDenom() for odd and even nTip", {
  # Even nTip = 8: (64 - 16 + 4) / 4 = 13
  expect_equal(TreeDist:::.TransferNormDenom(8, FALSE), 13)
  # Odd nTip = 9: (81 - 18 + 5) / 4 = 17
  expect_equal(TreeDist:::.TransferNormDenom(9, FALSE), 17)
  # Scaled: 2 * (n - 3)
  expect_equal(TreeDist:::.TransferNormDenom(8, TRUE), 10)
  expect_equal(TreeDist:::.TransferNormDenom(9, TRUE), 12)
})

test_that("TransferDist() normalize with odd nTip", {
  trees <- as.phylo(0:3, nTip = 9)
  # Scaled normalization
  d <- TransferDist(trees, normalize = TRUE)
  d_raw <- TransferDist(trees, normalize = FALSE)
  denom <- 2 * (9 - 3)
  expect_equal(as.numeric(d), as.numeric(d_raw) / denom, tolerance = 1e-12)

  # Unscaled normalization with odd nTip
  d_u <- TransferDist(trees, scale = FALSE, normalize = TRUE)
  d_u_raw <- TransferDist(trees, scale = FALSE, normalize = FALSE)
  denom_u <- (81 - 18 + 5) / 4
  expect_equal(as.numeric(d_u), as.numeric(d_u_raw) / denom_u, tolerance = 1e-12)
})

test_that("TransferDist() normalize with cross-pairs", {
  trees1 <- as.phylo(0:2, 8)
  trees2 <- as.phylo(3:5, 8)

  mat_raw <- TransferDist(trees1, trees2)
  mat_norm <- TransferDist(trees1, trees2, normalize = TRUE)
  denom <- 2 * (8 - 3)
  expect_equal(mat_norm, mat_raw / denom, tolerance = 1e-12)

  # Unscaled cross-pairs normalize
  mat_raw_u <- TransferDist(trees1, trees2, scale = FALSE)
  mat_norm_u <- TransferDist(trees1, trees2, scale = FALSE, normalize = TRUE)
  denom_u <- 13
  expect_equal(mat_norm_u, mat_raw_u / denom_u, tolerance = 1e-12)
})

test_that("TransferDist() normalize via generic fallback (reportMatching)", {
  bal8 <- TreeTools::BalancedTree(8)
  pec8 <- TreeTools::PectinateTree(8)

  # reportMatching forces generic path; normalize divides after
  res <- TransferDist(bal8, pec8, normalize = TRUE, reportMatching = TRUE)
  raw <- TransferDist(bal8, pec8, reportMatching = TRUE)
  denom <- 2 * (8 - 3)
  expect_equal(res[[1]], raw[[1]] / denom, tolerance = 1e-12)

  # Unscaled + normalize + reportMatching
  res_u <- TransferDist(bal8, pec8, scale = FALSE,
                         normalize = TRUE, reportMatching = TRUE)
  raw_u <- TransferDist(bal8, pec8, scale = FALSE, reportMatching = TRUE)
  expect_equal(res_u[[1]], raw_u[[1]] / 13, tolerance = 1e-12)
})


# =========================================================================
# Fast path edge cases
# =========================================================================

test_that(".TransferDistAllPairs() returns NULL for edge cases", {
  # Single phylo (not multiPhylo)
  tr <- TreeTools::BalancedTree(8)
  expect_null(TreeDist:::.TransferDistAllPairs(tr, TRUE))

  # < 2 trees
  expect_null(TreeDist:::.TransferDistAllPairs(
    structure(list(tr), class = "multiPhylo"), TRUE))

  # < 4 tips
  small <- as.phylo(0:2, nTip = 3)
  expect_null(TreeDist:::.TransferDistAllPairs(small, TRUE))

  # Mismatched tip labels
  t1 <- TreeTools::BalancedTree(paste0("a", 1:6))
  t2 <- TreeTools::BalancedTree(paste0("b", 1:6))
  mixed <- structure(list(t1, t2), class = "multiPhylo")
  expect_null(TreeDist:::.TransferDistAllPairs(mixed, TRUE))
})

test_that(".TransferDistCrossPairs() returns NULL for edge cases", {
  tr1 <- TreeTools::BalancedTree(8)
  tr2 <- TreeTools::PectinateTree(8)

  # Both single → delegate to generic
  expect_null(TreeDist:::.TransferDistCrossPairs(tr1, tr2, TRUE))

  # < 4 tips
  small1 <- as.phylo(0:1, nTip = 3)
  small2 <- as.phylo(2:3, nTip = 3)
  expect_null(TreeDist:::.TransferDistCrossPairs(small1, small2, TRUE))

  # Mismatched tip labels
  t1 <- structure(list(TreeTools::BalancedTree(paste0("a", 1:6))),
                  class = "multiPhylo")
  t2 <- structure(list(TreeTools::BalancedTree(paste0("b", 1:6))),
                  class = "multiPhylo")
  expect_null(TreeDist:::.TransferDistCrossPairs(t1, t2, TRUE))
})

test_that("TransferDist() all-pairs with star trees in mix", {
  star <- TreeTools::StarTree(8)
  bal <- TreeTools::BalancedTree(8)
  pec <- TreeTools::PectinateTree(8)
  trees <- structure(list(star, bal, pec), class = "multiPhylo")

  d <- TransferDist(trees)
  dm <- as.matrix(d)
  expect_equal(dm[1, 1], 0)  # star self
  expect_true(dm[1, 2] >= 0) # star vs binary
  expect_equal(dm[2, 3], TransferDist(bal, pec), tolerance = 1e-12)
})

test_that("TransferDist() all-pairs with exactly 2 trees", {
  trees <- as.phylo(0:1, nTip = 8)
  d <- TransferDist(trees)
  expect_s3_class(d, "dist")
  expect_equal(attr(d, "Size"), 2L)
  expect_equal(as.numeric(d),
               TransferDist(trees[[1]], trees[[2]]), tolerance = 1e-12)
})

test_that("TransferDist() all-pairs with exactly 4 tips", {
  trees <- as.phylo(0:3, nTip = 4)
  d <- TransferDist(trees)
  expect_s3_class(d, "dist")
  dm <- as.matrix(d)
  expect_true(all(dm >= -1e-12))
})


# =========================================================================
# TransferDistSplits edge cases
# =========================================================================

test_that("TransferDistSplits() reportMatching with various configurations", {
  bal8 <- TreeTools::BalancedTree(8)
  pec8 <- TreeTools::PectinateTree(8)
  tipLabels <- TreeTools::TipLabels(bal8)

  sp_bal <- as.Splits(bal8, tipLabels)
  sp_pec <- as.Splits(pec8, tipLabels)

  # Unscaled reportMatching
  res <- TransferDistSplits(sp_bal, sp_pec, scale = FALSE,
                             reportMatching = TRUE)
  expect_true(is.numeric(res))
  expect_true(!is.null(attr(res, "matching")))
  expect_true(!is.null(attr(res, "pairScores")))
  expect_true(!is.null(attr(res, "matchedScores")))

  # Scaled reportMatching (already tested above, but confirm matchedSplits)
  res_s <- TransferDistSplits(sp_bal, sp_pec, scale = TRUE,
                               reportMatching = TRUE)
  expect_true(!is.null(attr(res_s, "matchedSplits")))
})

test_that("TransferDistSplits() with asymmetric nSplits (nSplits1 < nSplits2)", {
  # Use majority-rule consensus (fewer splits) vs fully resolved tree
  trees <- as.phylo(0:9, 8)
  mr <- Consensus(trees, p = 0.5)
  bal8 <- TreeTools::BalancedTree(8)
  tipLabels <- TreeTools::TipLabels(bal8)

  sp_mr <- as.Splits(mr, tipLabels)
  sp_bal <- as.Splits(bal8, tipLabels)

  # Confirm asymmetry
  expect_lt(nrow(sp_mr), nrow(sp_bal))

  res <- TransferDistSplits(sp_mr, sp_bal, reportMatching = TRUE)
  # Matching should be truncated to nSplits1
  matching <- attr(res, "matching")
  expect_equal(length(matching), nrow(sp_mr))
  expect_true(!is.null(attr(res, "pairScores")))
})

test_that("TransferDist() for 3-tip trees uses fallback path", {
  # 3 tips: fast path returns NULL, falls through to generic
  tr1 <- TreeTools::BalancedTree(3)
  tr2 <- TreeTools::PectinateTree(3)
  td <- TransferDist(tr1, tr2)
  expect_true(is.finite(td))
  expect_true(td >= 0)
})

test_that("TransferDist() with named tree list preserves labels", {
  trees <- as.phylo(0:3, nTip = 8)
  names(trees) <- paste0("tree_", 1:4)

  d <- TransferDist(trees)
  expect_equal(attr(d, "Labels"), paste0("tree_", 1:4))
})

test_that("TransferDist() cross-pairs with single Splits input", {
  bal8 <- TreeTools::BalancedTree(8)
  tipLabels <- TreeTools::TipLabels(bal8)
  sp <- as.Splits(bal8, tipLabels)
  trees <- as.phylo(0:3, nTip = 8)

  # Single Splits object vs list of trees
  result <- TransferDist(sp, trees)
  expect_length(result, 4)
  expect_true(all(result >= 0))
})

test_that("TransferDist fast-path guards handle edge cases", {
  # Tiny trees (< 4 tips) fall through to generic path
  tiny <- structure(rep(list(ape::read.tree(text = "(a,b,c);")), 3),
                    class = "multiPhylo")
  d <- TransferDist(tiny)
  expect_s3_class(d, "dist")

  # Cross-pairs with mixed-label trees fall through to generic path
  t1 <- ape::read.tree(text = "((a,b),(c,d));")
  t2 <- ape::read.tree(text = "((a,b),(e,f));")
  expect_true(is.numeric(TransferDist(t1, t2)))
})

#' Find islands from distance matrix
#' 
#' `Islands()` assigns a set of objects to islands, such that all elements
#' within an island can form a connected graph in which each edge is no longer
#' than `threshold` distance units \insertRef{Silva2021}{TreeDist}.
#' 
#' @inheritParams SpectralEigens
#' @param threshold Elements greater than `threshold` distance units will not be
#' assigned to the same island.
#' @param dense Logical; if `FALSE`, each island will be named according to the
#' index of its lowest-indexed member; if `TRUE`, each island will be numbered
#' sequentially from `1`, ordered by the index of the lowest-indexed member.
#' @param smallest Integer; Islands comprising no more than `smallest` elements
#' will be assigned to island `NA`.
#' @return `Islands()` returns a vector listing the island to which
#' each element is assigned.
#' @references \insertAllCited{}
#' @examples
#' library("TreeTools", quietly = TRUE)
#' # Generate a set of trees
#' trees <- as.phylo(as.TreeNumber(BalancedTree(16)) + c(-(40:20), 70:105), 16)
#' 
#' # Calculate distances between trees
#' distances <- ClusteringInfoDist(trees)
#' summary(distances)
#' 
#' # Assign trees to islands
#' isle <- Islands(distances, quantile(distances, 0.1))
#' table(isle)
#' 
#' # Indicate island membership on 2D mapping of tree distances
#' mapping <- cmdscale(distances, 2)
#' plot(mapping, col = isle + 1,
#'      asp = 1, # Preserve aspect ratio - do not distort distances
#'      ann = FALSE, axes = FALSE, # Don't label axes: dimensions are meaningless)
#'      pch = 16 # Plotting character: Filled circle
#' )
#' 
#' # Compare strict consensus with island consensus trees
#' oPar <- par(mfrow = c(2, 2), mai = rep(0.1, 4))
#' plot(Consensus(trees), main = "Strict")
#' plot(Consensus(trees[isle == 1]), edge.col = 2, main = "Island 1")
#' plot(Consensus(trees[isle == 2]), edge.col = 3, main = "Island 2")
#' plot(Consensus(trees[isle == 3]), edge.col = 4, main = "Island 3")
#' 
#' # Restore graphical parameters
#' par(oPar)
#' @template MRS
#' @family tree space functions
#' @export
Islands <- function(D, threshold, dense = TRUE, smallest = 0) {
  linked <- as.matrix(D) <= threshold
  n <- dim(linked)[[1]]
  ret <- integer(n)
  i <- 1
  repeat {
    links <- seq_len(n) == i
    repeat {
      nowLinked <- colSums(linked[links, , drop = FALSE]) > 0
      if (any(nowLinked[!links])) {
        # Added to island
        links <- nowLinked
      } else {
        break
      }
    }
    ret[links] <- i
    i <- which.min(ret)
    if (ret[[i]]) break
  }
  tab <- table(ret, dnn = NULL)
  ret[ret %in% as.integer(names(tab)[tab < smallest])] <- NA
  
  if (dense) {
    as.integer(factor(rank(ret, ties.method = "min", na.last = "keep")))
  } else {
    ret
  }
}

test_that("Islands() works", {
  library("TreeTools", quietly = TRUE)
  trees <- as.phylo(as.TreeNumber(BalancedTree(16)) + c(70:78, -(39:30), 80:99), 16)
  distances <- ClusteringInfoDist(trees)
  expect_equal(Islands(distances, 2.5), rep(c(1, 2, 1), c(9, 10, 20)))
  expect_equal(Islands(distances, 2.5, FALSE), rep(c(1, 10, 1), c(9, 10, 20)))
  
  
  trees <- as.phylo(c(0:10, 1000:1005, 2000:2019), 16)
  distances <- ClusteringInfoDist(trees)
  expect_equal(Islands(distances, 2.5, TRUE, smallest = 6),
               rep(c(1, 2, 3), c(11, 6, 20)))
  expect_equal(Islands(distances, 2.5, TRUE, smallest = 7),
               rep(c(1, NA, 2), c(11, 6, 20)))
  expect_equal(Islands(distances, 2.5, FALSE, smallest = 6),
               rep(c(1, 12, 18), c(11, 6, 20)))
  expect_equal(Islands(distances, 2.5, FALSE, smallest = 7),
               rep(c(1, NA, 18), c(11, 6, 20)))
})

#' Kendall&ndash;Colijn distance
#' 
#' Calculate the Kendall&ndash;Colijn tree distance, a measure related to the 
#' path difference. 
#' 
#' The Kendall&ndash;Colijn distance works by measuring, for each pair of
#' leaves, the distance from the most recent common ancestor of those leaves
#' and the root node.
#' For a given tree, this produces a vector of values recording the
#' distance-from-the-root of each most recent common ancestor of each pair of
#' leaves.
#' 
#' Two trees are compared by taking the Euclidean distance between the
#' respective vectors.  This is calculated by taking the square root of the sum 
#' of the squares of the differences between the vectors.
#' 
#' An analogous distance can be created from any vector representation of a
#' tree.
#' The split size vector metric \insertCite{SmithSpace}{TreeDist} is an attempt
#' to mimic the Kendall&nbsp;Colijn metric in situations where the position of
#' the root should not be afforded special significance; and the path distance
#' \insertCite{Steel1993}{TreeDist} is a familiar alternative whose underlying
#' vector measures the distance of the last common ancestor of each pair
#' of leaves from the leaves themselves, i.e. the length of the path from one 
#' leaf to another.
#' 
#' 
#' None of these vector-based methods performs as well as other tree distances
#' in measuring similarities in the relationships implied by a pair of trees
#' \insertCite{SmithDist}{TreeDist}; in particular, the Kendall&nbsp;Colijn
#' metric is strongly influenced by tree balance, and may not be appropriate
#' for a suite of common applications \insertCite{SmithSpace}{TreeDist}.
#' 
#' @template tree12ListParams
#' @param Vector Function converting a tree to a numeric vector.
#' 
#' `KCVector`, the default, returns the number of edges between the common
#' ancestor of each pair of leaves and the root of the tree
#' \insertCite{@per @Kendall2016}{TreeDist}.
#' 
#' `PathVector` returns the number of edges between each pair of leaves
#' \insertCite{@per @Steel1993}{TreeDist}.
#' 
#' `SplitVector` returns the size of the smallest split that contains each
#' pair of leaves (per \insertCite{SmithSpace;nobrackets}{TreeDist}).
#' 
#' @templateVar returns `KendallColijn()` returns
#' @template distReturn
#' 
#' @examples 
#' KendallColijn(TreeTools::BalancedTree(8), TreeTools::PectinateTree(8))
#'
#' set.seed(0)
#' KendallColijn(TreeTools::BalancedTree(8), lapply(rep(8, 3), ape::rtree))
#' KendallColijn(lapply(rep(8, 4), ape::rtree))
#' 
#' KendallColijn(lapply(rep(8, 4), ape::rtree), Vector = SplitVector)
#' 
#' # Notice that changing tree shape close to the root results in much
#' # larger differences
#' tree1 <- ape::read.tree(text = "(a, (b, (c, (d, (e, (f, (g, h)))))));")
#' tree2 <- ape::read.tree(text = "(a, ((b, c), (d, (e, (f, (g, h))))));")
#' tree3 <- ape::read.tree(text = "(a, (b, (c, (d, (e, ((f, g), h))))));")
#' trees <- c(tree1, tree2, tree3)
#' KendallColijn(trees)
#' KendallColijn(trees, Vector = SplitVector)
#' @template MRS
#' 
#' @seealso [`treespace::treeDist`](https://CRAN.R-project.org/package=treespace/vignettes/introduction.html)
#' is a more sophisticated, if more cumbersome, implementation that supports 
#' lambda > 0, i.e. use of edge lengths in tree comparison.
#' 
#' @references \insertAllCited{}
#' 
#' @family tree distances
#' @importFrom utils combn
#' @encoding UTF-8
#' @export
KendallColijn <- function(tree1, tree2 = NULL, Vector = KCVector) {
  
  FunValue <- function(nTip) double(nTip * (nTip - 1L) / 2L)
  
  if (inherits(tree1, "phylo")) {
    if (inherits(tree2, "phylo")) {
      if (length(tree1[["tip.label"]]) != length(tree2[["tip.label"]]) || 
          length(setdiff(tree1[["tip.label"]], tree2[["tip.label"]])) > 0) {
        stop("Leaves must bear identical labels.")
      }
      .EuclideanDistance(Vector(tree1) - Vector(tree2))
    } else {
      if (is.null(tree2)) {
        0
      } else {
        apply(Vector(tree1) - vapply(tree2, Vector,
                                     FunValue(length(tree1[["tip.label"]]))),
              2L, .EuclideanDistance)
      }
    }
  } else {
    if (inherits(tree2, "phylo")) {
      apply(Vector(tree2) - vapply(tree1, Vector,
                                   FunValue(length(tree2[["tip.label"]]))),
            2L, .EuclideanDistance)
    } else if (is.null(tree2)) {
      
      treeVec <- vapply(tree1, Vector, FunValue(length(tree1[[1]][["tip.label"]])))
      nTree <- length(tree1)
      
      # C++ pairwise Euclidean distance (returns dist-order vector)
      structure(class = "dist", Size = nTree,
                Diag = FALSE, Upper = FALSE,
                pair_diff_euclidean(treeVec))
    }
      else {
      vector1 <- vapply(tree1, Vector, FunValue(length(tree1[[1]][["tip.label"]])))
      vector2 <- vapply(tree2, Vector, FunValue(length(tree2[[1]][["tip.label"]])))
      ret <- vec_diff_euclidean(vector1, vector2)
      n1 <- names(tree1)
      n2 <- names(tree2)
      if (!is.null(n1) || !is.null(n2)) {
        dimnames(ret) <- list(n1, n2)
      }
      ret
    }
  }
}

.EuclideanDistance <- function(x) sqrt(sum(x * x))

#' @describeIn KendallColijn Creates a vector that characterises a rooted tree,
#' as described in \insertCite{Kendall2016;textual}{TreeDist}.
#' @param tree A tree of class \code{\link[ape:read.tree]{phylo}}.
#' @importFrom TreeTools Preorder
#' @export
KCVector <- function(tree) {
  tree <- Preorder(tree)
  edge <- tree[["edge"]]
  nTip <- edge[1L, 1L] - 1L
  tipOrder <- order(tree[["tip.label"]])
  structure(cpp_kc_vector(edge, tipOrder), Size = nTip, class = "dist")
}

#' @describeIn KendallColijn Creates a vector reporting the number of edges
#' between each pair of leaves, per the path metric of
#' \insertCite{Steel1993;textual}{TreeDist}.
#' @importFrom TreeTools AllAncestors Preorder
#' @importFrom utils combn
#' @export
PathVector <- function(tree) {
  if (!inherits(tree, "phylo")) {
    stop("`tree` must be of class `phylo`")
  }
  edge <- tree[["edge"]]
  structure(path_vector(edge), Size = NTip(tree), class = "dist")
}

#' @describeIn KendallColijn Creates a vector reporting the smallest split
#' containing each pair of leaves, per the metric proposed in
#' \insertCite{SmithSpace;textual}{TreeDist}.
#' @importFrom TreeTools as.Splits
#' @export
SplitVector <- function(tree) {
  tipLabel <- tree[["tip.label"]]
  nTip <- length(tipLabel)
  splits <- as.logical(as.Splits(tree, tipLabel[order(tipLabel)]))
  splits <- rbind(splits, !splits)
  inSplit <- rowSums(splits)
  is <- combn(seq_len(nTip), 2)
  
  smallestSplit <- apply(is, 2, function(i)
    min(inSplit[splits[, i[[1]]] & splits[, i[[2]]]], nTip - 1L))
  
  structure(smallestSplit, Size = nTip, class = "dist")
}

#' Approximate diameter of the Kendall&ndash;Collijn metric
#'
#' @return `KCDiameter()` returns the value of the Kendall & Colijn's (2016)
#' metric distance between two pectinate trees with _n_ leaves ordered in
#' the opposite direction, which I suggest (without any attempt at a proof) may
#' be a useful proxy for the diameter (i.e. maximum value) of the K&ndash;C
#' metric.
#'
#' @examples
#' KCDiameter(4)
#' KCDiameter(trees)
#' @importFrom TreeTools PectinateTree
#' @rdname KendallColijn
#' @export
KCDiameter <- function(tree) UseMethod("KCDiameter")

#' @importFrom TreeTools NTip
#' @export
KCDiameter.phylo <- function(tree) {
  KCDiameter.numeric(NTip(tree))
}

#' @export
KCDiameter.numeric <- function(tree) {
  nTip <- as.integer(tree)
  mat <- matrix(seq_len(nTip), nTip, nTip)
  Euclid <- function(x, y) sqrt(sum((x - y) ^ 2))
  
  # Return:
  Euclid(nTip - mat[lower.tri(mat)] + 1L, t(mat)[lower.tri(mat)])
}

#' @export
KCDiameter.list <- function(tree) {
  vapply(tree, KCDiameter, double(1))
}

#' @export
KCDiameter.multiPhylo <- function(tree) {
  vapply(tree, KCDiameter, double(1))
}

library("TreeTools", quietly = TRUE)

test_that("KC fails gracefully", {
  expect_error(KendallColijn(BalancedTree(8), BalancedTree(1:8)),
               "Leaves must bear identical labels.")
})

test_that("KC vector calculations", {
  
  bal7 <- ape::read.tree(text = "(((t1,t2),(t3,t4)),((t5,t6),t7));")
  expect_error(PathVector(bal7$edge), "must be of class\\b")
  bal7b <- ape::read.tree(text = "(((t5,t6),t7), ((t1,t2),(t3,t4)));")
  expect_equal(PathVector(UnrootTree(bal7)),
               PathVector(UnrootTree(RootTree(bal7, 1))))
  expect_equal(PathVector(UnrootTree(bal7)),
               PathVector(UnrootTree(RenumberTips(bal7b, bal7))))
  expect_equal(as.numeric(PathVector(RenumberTips(bal7b, bal7))),
               c(2, 4, 4, 6, 6, 5,
                    4, 4, 6, 6, 5,
                       2, 6, 6, 5,
                          6, 6, 5,
                             2, 3,
                                3))
  
  expect_equal(SplitVector(bal7), SplitVector(RootTree(bal7, 1)))
  expect_equal(SplitVector(bal7), SplitVector(bal7b))
  expect_equal(as.numeric(SplitVector(BalancedTree((9)))),
               c(2, 3, 5, 5, 7, 7, 7, 7,
                    3, 5, 5, 7, 7, 7, 7,
                       5, 5, 7, 7, 7, 7,
                          2, 6, 6, 6, 6,
                             6, 6, 6, 6,
                                2, 4, 4,
                                   4, 4,
                                      2))
})

test_that("KC distances with special vectors", {
  trees <- as.phylo(1:20, 12)
  expect_equal(PathDist(trees), KendallColijn(trees, Vector = PathVector),
               ignore_attr = TRUE)
})

test_that("KCDiameter() calculated", {
  Test <- function(nTip) {
    tips <- seq_len(nTip)
    expect_equal(KendallColijn(PectinateTree(tips), PectinateTree(rev(tips))),
                 KCDiameter(nTip))
  }
  Test(4)
  Test(40)
  tree1 <- ape::read.tree(text = "(a, (b, (c, (d, (e, (f, (g, h)))))));")
  expect_equal(KendallColijn(tree1), 0)
  tree2 <- ape::read.tree(text = "(a, ((b, c), (d, (e, (f, (g, h))))));")
  tree3 <- ape::read.tree(text = "(a, (b, (c, (d, (e, (f, g))))));")
  trees <- c(tree1, tree2, tree3)
  expect_equal(KCDiameter(trees),
               c(KCDiameter(tree1), KCDiameter(tree2), KCDiameter(tree3)))
  expect_equal(KCDiameter(list(tree1, tree2, tree3)), KCDiameter(trees))
})

test_that("PathDist()", {
  library("TreeTools", quietly = TRUE)
  t05 <- as.phylo(0:5, 6)
  bal6 <- BalancedTree(paste0("t", 6:1))
  vec6 <- PathVector(RenumberTips(bal6, t05))
  expect_equal(PathDist(t05, bal6),
               sqrt(colSums((vapply(t05, PathVector, vec6) - vec6) ^ 2)))
  expect_equal(PathDist(bal6, t05), PathDist(t05, bal6))
  
  expect_equal(PathDist(bal6, PectinateTree(6)),
               PathDist(list(bal6, PectinateTree(6)))[1],
               ignore_attr = TRUE)
  
  trees <- as.phylo(1:8, 29)
  expect_equal(unname(as.matrix(PathDist(trees))), PathDist(trees, trees))
  expect_equal(PathDist(trees), CompareAll(trees, PathDist),
               ignore_attr = TRUE)
})

test_that("PathDist() equivalent to path.dist()", {
  skip_if_not_installed("phangorn")
  
  t1 <- Postorder(as.phylo(0:5, 6))
  t2 <- Postorder(BalancedTree(6))
  t3 <- Postorder(PectinateTree(6))
  expect_equal(PathDist(UnrootTree(t1), UnrootTree(t2)),
               phangorn::path.dist(t1, t2))
  expect_equal(PathDist(UnrootTree(t2), UnrootTree(t3)),
               phangorn::path.dist(t3, t2))
  
  postTrees <- Postorder(as.phylo(0:5, 182))
  expect_equal(PathDist(postTrees), phangorn::path.dist(postTrees))
})

#' @rdname TreeDistance
#' @export
MatchingSplitInfo <- function(tree1, tree2 = NULL, normalize = FALSE,
                               reportMatching = FALSE, diag = TRUE) {
  unnormalized <- CalculateTreeDistance(MatchingSplitInfoSplits, tree1,
                                        tree2, reportMatching)
  
  if (diag && is.null(tree2)) {
    unnormalized <- as.matrix(unnormalized)
    diag(unnormalized) <- SplitwiseInfo(tree1)
    tree2 <- tree1
  }
  # Return:
  NormalizeInfo(unnormalized, tree1, tree2, how = normalize,
                InfoInTree = SplitwiseInfo, Combine = .PairMean)
}

#' @rdname TreeDistance
#' @export
MatchingSplitInfoDistance <- function(tree1, tree2 = NULL, 
                                       normalize = FALSE,
                                       reportMatching = FALSE) {
  
  # Fast path (all-pairs): same tips, no matching — avoids duplicate as.Splits()
  fast <- .FastDistPath(tree1, tree2, reportMatching,
                        cpp_msi_all_pairs,
                        cpp_splitwise_info_batch)
  if (!is.null(fast)) {
    msi <- fast[["info"]]
    treesIndependentInfo <- .PairwiseSums(fast[["entropies"]])
    
    ret <- .FloorNumericalNoise(treesIndependentInfo - msi - msi, treesIndependentInfo)
    ret <- NormalizeInfo(ret, tree1, tree2, how = normalize,
                         infoInBoth = treesIndependentInfo,
                         InfoInTree = SplitwiseInfo, Combine = "+")
    attributes(ret) <- attributes(msi)
    return(ret)
  }

  # Fast path (cross-pairs): same tips, no matching — avoids duplicate as.Splits()
  fast_many <- .FastManyManyPath(tree1, tree2, reportMatching,
                                 cpp_msi_cross_pairs,
                                 cpp_splitwise_info_batch)
  if (!is.null(fast_many)) {
    msi <- fast_many[["dists"]]
    info1 <- fast_many[["info1"]]
    info2 <- fast_many[["info2"]]
    treesIndependentInfo <- outer(info1, info2, "+")

    ret <- .FloorNumericalNoise(treesIndependentInfo - msi - msi, treesIndependentInfo)
    ret <- NormalizeInfo(ret, tree1, tree2, how = normalize,
                         infoInBoth = treesIndependentInfo,
                         InfoInTree = SplitwiseInfo, Combine = "+")
    return(ret)
  }

  msi <- MatchingSplitInfo(tree1, tree2, normalize = FALSE, diag = FALSE,
                           reportMatching = reportMatching)

  treesIndependentInfo <- .MaxValue(tree1, tree2, SplitwiseInfo)
  ret <- .FloorNumericalNoise(treesIndependentInfo - msi - msi, treesIndependentInfo)
  ret <- NormalizeInfo(ret, tree1, tree2, how = normalize,
                       infoInBoth = treesIndependentInfo,
                       InfoInTree = SplitwiseInfo, Combine = "+")
  attributes(ret) <- attributes(msi)
  # Return:
  ret
}


#' @rdname TreeDistance
#' @export
MatchingSplitInfoSplits <- function(splits1, splits2,
                                     nTip = attr(splits1, "nTip"),
                                     reportMatching = FALSE) {
  
  GeneralizedRF(splits1, splits2, nTip, cpp_msi_distance, maximize = TRUE,
                reportMatching = reportMatching)
}

# Validate that nTip does not exceed the supported tip-count ceiling.
# cpp_sl_max_tips() > 2048L iff TreeTools >= 2.3.0 raised the stack threshold
# and provides heap-backed split storage; accept up to 32767 tips in that case.
# Otherwise cap at the compiled SL_MAX_TIPS.
# Called from every distance entry point before any C++ work.
.CheckMaxTips <- function(nTip) {
  if (is.na(nTip)) return(invisible(NULL))
  sl_max <- cpp_sl_max_tips()
  if (sl_max > 2048L) {
    if (nTip > 32767L) {
      stop("Trees with ", nTip,
           " tips are not yet supported (maximum 32767).")
    }
  } else if (nTip > sl_max) {
    # else-if fires only when sl_max <= 2048 (TreeTools < 2.3.0)
    stop(
      "Trees with ", nTip, " tips exceed the compiled limit of ",
      sl_max, " tips.",
      "\nUpdate TreeTools and reinstall TreeDist to support more tips."
    )
  }
}

#' Wrapper for tree distance calculations
#' 
#' Calls tree distance functions from trees or lists of trees
#' 
#' @inheritParams TreeDistance
#' @param Func Tree distance function.
#' @param \dots Additional arguments to `Func`.
#' 
#' @template MRS
#' @keywords internal
#' @importFrom TreeTools as.Splits TipLabels
#' @importFrom utils combn
#' @export
# Keep in sync with C++ guard: min(SL_MAX_TIPS, int16_t::max()).
CalculateTreeDistance <- function(Func, tree1, tree2 = NULL,
                                  reportMatching = FALSE, ...) {
  supportedClasses <- c("phylo", "Splits")
  
  single1 <- inherits(tree1, supportedClasses)
  if (!single1 && !inherits(tree1[[1]], supportedClasses)) {
    stop("`tree1` must be a tree, or list of trees, with class ", 
         paste(supportedClasses, collapse = " / "))
  }
  labels1 <- TipLabels(tree1)
  if (is.list(labels1)) {
    if (!all(vapply(labels1[-1], setequal, logical(1), labels1[[1]]))) {
      nTip <- NA
    } else {
      labels1 <- labels1[[1]]
      nTip <- length(labels1)
    }
  } else {
    nTip <- length(labels1)
  }
  
  null2 <- is.null(tree2)
  if (!null2) {
    single2 <- inherits(tree2, supportedClasses)
    if (!single2 && !inherits(tree2[[1]], supportedClasses)) {
      stop("`tree2` must be a tree, or list of trees, with class ", 
           paste(supportedClasses, collapse = " / "))
    }
    labels2 <- TipLabels(tree2)
    if (is.list(labels2)) {
      if (!all(vapply(labels2[-1], setequal, logical(1), labels2[[1]]))) {
        nTip <- NA
      } else {
        labels2 <- labels2[[1]]
      }
    }
    
    if (!setequal(labels1, labels2)) {
      nTip <- NA
    }
  }
  
  if (single1) {
    if (null2) {
      dist(0)
    } else if (single2) {
      .SplitDistanceOneOne(Func, tree1, tree2, tipLabels = labels1, nTip,
                           reportMatching = reportMatching, ...)
    } else {
      .SplitDistanceOneMany(Func, oneSplit = tree1, manySplits = tree2,
                            tipLabels = labels1, nTip = nTip, ...)
    }
  } else {
    if (is.null(tree2)) {
      .SplitDistanceAllPairs(Func, tree1, tipLabels = labels1, nTip, ...)
    } else if (single2) {
      .SplitDistanceOneMany(Func, oneSplit = tree2, manySplits = tree1,
                            tipLabels = labels2, nTip = nTip, ...)
    } else {
      .SplitDistanceManyMany(Func, tree1, tree2, tipLabels = labels1,
                             nTip = nTip, ...)
    }
  }
}

.SplitDistanceOneOne <- function(Func, split1, split2, tipLabels, 
                                 nTip = length(tipLabels), reportMatching,
                                 ...) {
  if (is.na(nTip)) {
    common <- intersect(TipLabels(split1), TipLabels(split2))
    split1 <- KeepTip(split1, common)
    split2 <- KeepTip(split2, common)
    nTip <- length(common)
  }
  if (nTip < 4) {
    0
  } else {
    Func(as.Splits(split1, asSplits = reportMatching),
         as.Splits(split2, tipLabels = tipLabels, asSplits = reportMatching),
         nTip = nTip, reportMatching = reportMatching, ...)
  }
}

#' @importFrom stats setNames
.SplitDistanceOneMany <- function(Func, oneSplit, manySplits, 
                                  tipLabels, nTip = length(tipLabels), ...) {
  if (is.na(nTip)) {
    oneLabels <- TipLabels(oneSplit)
    manyLabels <- TipLabels(manySplits)
    common <- if (is.list(manyLabels)) {
      lapply(manyLabels, intersect, oneLabels)
    } else {
      list(intersect(manyLabels, oneLabels))
    }
    setNames(unlist(.mapply(function(s2, keep) {
      Func(
        as.Splits(KeepTip(oneSplit, keep), tipLabels = keep, asSplits = FALSE),
        as.Splits(KeepTip(s2, keep), tipLabels = keep, asSplits = FALSE),
        nTip = length(keep),
        ...
      )
    }, dots = list(manySplits, common), MoreArgs = NULL)), names(manySplits))
  } else {
    s1 <- as.Splits(oneSplit, tipLabels = tipLabels, asSplits = FALSE)
    vapply(manySplits,
           function(s2) Func(s1, as.Splits(s2, tipLabels = tipLabels,
                                           asSplits = FALSE),
                              nTip = nTip, ...),
           double(1))
  }
}

#' @importFrom parallel parCapply
#' @importFrom cli cli_progress_bar cli_progress_update
.SplitDistanceAllPairs <- function(Func, splits1, tipLabels,
                                   nTip = length(tipLabels), ...) {
  cluster <- getOption("TreeDist-cluster")

  # Fast paths: use OpenMP batch functions when all trees share the same tip
  # set and no R-level cluster has been configured.  Each branch mirrors the
  # generic path exactly but avoids per-pair R overhead.
  .CheckMaxTips(nTip)
  if (!is.na(nTip) && is.null(cluster)) {
    .n_threads <- as.integer(getOption("mc.cores", 1L))
    .batch_result <- if (identical(Func, MutualClusteringInfoSplits)) {
      splits <- as.Splits(splits1, tipLabels = tipLabels, asSplits = FALSE)
      cpp_mutual_clustering_all_pairs(splits, as.integer(nTip), .n_threads)

    } else if (identical(Func, InfoRobinsonFouldsSplits)) {
      splits <- as.Splits(splits1, tipLabels = tipLabels, asSplits = FALSE)
      cpp_rf_info_all_pairs(splits, as.integer(nTip), .n_threads)

    } else if (identical(Func, MatchingSplitDistanceSplits)) {
      splits <- as.Splits(splits1, tipLabels = tipLabels, asSplits = FALSE)
      cpp_msd_all_pairs(splits, as.integer(nTip), .n_threads)

    } else if (identical(Func, MatchingSplitInfoSplits)) {
      splits <- as.Splits(splits1, tipLabels = tipLabels, asSplits = FALSE)
      cpp_msi_all_pairs(splits, as.integer(nTip), .n_threads)

    } else if (identical(Func, SharedPhylogeneticInfoSplits)) {
      splits <- as.Splits(splits1, tipLabels = tipLabels, asSplits = FALSE)
      cpp_shared_phylo_all_pairs(splits, as.integer(nTip), .n_threads)

    } else if (identical(Func, NyeSplitSimilarity)) {
      splits <- as.Splits(splits1, tipLabels = tipLabels, asSplits = FALSE)
      cpp_jaccard_all_pairs(splits, as.integer(nTip),
                            k = 1.0, allow_conflict = TRUE, .n_threads)

    } else if (identical(Func, JaccardSplitSimilarity)) {
      dots <- list(...)
      splits <- as.Splits(splits1, tipLabels = tipLabels, asSplits = FALSE)
      cpp_jaccard_all_pairs(
        splits, as.integer(nTip),
        k            = as.double(if ("k" %in% names(dots)) dots[["k"]] else 1L),
        allow_conflict = as.logical(
          if ("allowConflict" %in% names(dots)) dots[["allowConflict"]] else TRUE),
        .n_threads
      )

    } else {
      NULL
    }

    if (!is.null(.batch_result)) {
      return(structure(.batch_result,
                       class  = "dist",
                       Size   = length(splits1),
                       Labels = names(splits1),
                       Diag   = FALSE,
                       Upper  = FALSE))
    }
  }

  if (is.na(nTip)) {
    splits <- lapply(splits1, as.Splits)
    
    .PairDist <- function(i) {
      s1 <- splits[[i[[1]]]]
      s2 <- splits[[i[[2]]]]
      common <- intersect(TipLabels(s1), TipLabels(s2))
      Func(KeepTip(s1, common), KeepTip(s2, common),
           nTip = length(common), reportMatching = FALSE, ...)
    }
    
  } else {
    splits <- as.Splits(splits1, tipLabels = tipLabels, asSplits = FALSE)
    
    .PairDist <- function(i) {
      Func(splits[[i[[1]]]], splits[[i[[2]]]],
           nTip = nTip, reportMatching = FALSE, ...)
    }
    
  }
  
  nSplits <- length(splits)
  is <- combn(seq_len(nSplits), 2)
  
  if (is.null(cluster)) {
    cli_progress_bar("Calculating distances", total = ncol(is))
  }
  
  .CliPairDist <- function(i) { # TODO if cli possible from parallel, merge.
    cli_progress_update(1, .envir = parent.frame(2))
    .PairDist(i)
  }
  
  ret <- structure(class = "dist", Size = nSplits,
                   Labels = names(splits1),
                   Diag = FALSE, Upper = FALSE,
                   if (is.null(cluster)) {
                     apply(is, 2, .CliPairDist)
                   } else {
                     parCapply(cluster, is, .PairDist)
                   })
  # Return:
  ret
}

#' @importFrom stats setNames
.SplitDistanceManyMany <- function(Func, splits1, splits2, 
                                   tipLabels, nTip = length(tipLabels), ...) {
  .CheckMaxTips(nTip)
  if (is.na(nTip)) {
    tipLabels <- union(unlist(tipLabels, use.names = FALSE),
                       unlist(TipLabels(splits2), use.names = FALSE))
    splits1 <- as.Splits(splits1, tipLabels = tipLabels, asSplits = TRUE)
    splits2 <- as.Splits(splits2, tipLabels = tipLabels, asSplits = TRUE)
    vapply(splits1, function(s1) {
        l1 <- TipLabels(s1)
        vapply(splits2, function(s2) {
          common <- intersect(l1, TipLabels(s2))
          Func(KeepTip(s1, common), KeepTip(s2, common),
               nTip = length(common))#, ...)
        }, double(1))
      },
      setNames(double(length(splits2)), names(splits2))
    )
  } else {
    # Fast path: use OpenMP batch functions when available
    .n_threads <- as.integer(getOption("mc.cores", 1L))
    .batch_result <- if (identical(Func, MutualClusteringInfoSplits)) {
      s1 <- as.Splits(splits1, tipLabels = tipLabels, asSplits = FALSE)
      s2 <- as.Splits(splits2, tipLabels = tipLabels, asSplits = FALSE)
      cpp_mutual_clustering_cross_pairs(s1, s2, as.integer(nTip), .n_threads)

    } else if (identical(Func, InfoRobinsonFouldsSplits)) {
      s1 <- as.Splits(splits1, tipLabels = tipLabels, asSplits = FALSE)
      s2 <- as.Splits(splits2, tipLabels = tipLabels, asSplits = FALSE)
      cpp_rf_info_cross_pairs(s1, s2, as.integer(nTip), .n_threads)

    } else if (identical(Func, MatchingSplitDistanceSplits)) {
      s1 <- as.Splits(splits1, tipLabels = tipLabels, asSplits = FALSE)
      s2 <- as.Splits(splits2, tipLabels = tipLabels, asSplits = FALSE)
      cpp_msd_cross_pairs(s1, s2, as.integer(nTip), .n_threads)

    } else if (identical(Func, MatchingSplitInfoSplits)) {
      s1 <- as.Splits(splits1, tipLabels = tipLabels, asSplits = FALSE)
      s2 <- as.Splits(splits2, tipLabels = tipLabels, asSplits = FALSE)
      cpp_msi_cross_pairs(s1, s2, as.integer(nTip), .n_threads)

    } else if (identical(Func, SharedPhylogeneticInfoSplits)) {
      s1 <- as.Splits(splits1, tipLabels = tipLabels, asSplits = FALSE)
      s2 <- as.Splits(splits2, tipLabels = tipLabels, asSplits = FALSE)
      cpp_shared_phylo_cross_pairs(s1, s2, as.integer(nTip), .n_threads)

    } else if (identical(Func, NyeSplitSimilarity)) {
      s1 <- as.Splits(splits1, tipLabels = tipLabels, asSplits = FALSE)
      s2 <- as.Splits(splits2, tipLabels = tipLabels, asSplits = FALSE)
      cpp_jaccard_cross_pairs(s1, s2, as.integer(nTip),
                              k = 1.0, allow_conflict = TRUE, .n_threads)

    } else if (identical(Func, JaccardSplitSimilarity)) {
      dots <- list(...)
      s1 <- as.Splits(splits1, tipLabels = tipLabels, asSplits = FALSE)
      s2 <- as.Splits(splits2, tipLabels = tipLabels, asSplits = FALSE)
      cpp_jaccard_cross_pairs(
        s1, s2, as.integer(nTip),
        k = as.double(if ("k" %in% names(dots)) dots[["k"]] else 1L),
        allow_conflict = as.logical(
          if ("allowConflict" %in% names(dots)) dots[["allowConflict"]] else TRUE),
        .n_threads
      )

    } else {
      NULL
    }

    if (!is.null(.batch_result)) {
      dimnames(.batch_result) <- list(names(splits1), names(splits2))
      return(.batch_result)
    }

    splits1 <- as.Splits(splits1, tipLabels = tipLabels, asSplits = FALSE)
    splits2 <- as.Splits(splits2, tipLabels = tipLabels, asSplits = FALSE)
    matrix(
      unlist(.mapply(Func, list(rep(splits2, each = length(splits1)),
                         splits1, nTip = nTip, ...), NULL)),
      length(splits1),
      length(splits2),
      dimnames = list(names(splits1), names(splits2))
    )
  }
}

#' Calculate distance between trees, or lists of trees
#' @template MRS
#' @importFrom TreeTools TipLabels
#' @param checks Logical specifying whether to perform basic sanity checks to
#' avoid crashes in C++.
#' @keywords internal
#' @seealso [`CalculateTreeDistance()`]
#' @export
.TreeDistance <- function(Func, tree1, tree2, checks = TRUE, ...) {
  single1 <- inherits(tree1, "phylo")
  labels1 <- TipLabels(tree1, single = TRUE)
  nTip <- length(labels1)
  
  single2 <- inherits(tree2, "phylo")
  labels2 <- TipLabels(tree2, single = TRUE)
  
  if (checks) {
    if (length(setdiff(labels1, labels2)) > 0) {
      stop("Leaves must bear identical labels.")
    }
    
    if (!single1) {
      .CheckLabelsSame(lapply(tree1, TipLabels))
    }
    
    if (!single2) {
      .CheckLabelsSame(lapply(tree2, TipLabels))
    }
  }
  
  if (single1) {
    if (single2) {
      Func(tree1, tree2, tipLabels = labels1, nTip = nTip, ...)
    } else {
      .TreeDistanceOneMany(Func, oneTree = tree1, manyTrees = tree2,
                           tipLabels = labels1, nTip = nTip, ...)
    }
  } else {
    if (single2) {
      .TreeDistanceOneMany(Func, oneTree = tree2, tipLabels = labels2, 
                           nTip = nTip, manyTrees = tree1, ...)
    } else {
      .TreeDistanceManyMany(Func, tree1, tree2, tipLabels = labels1,
                            nTip = nTip, ...)
    }
  }
}

.TreeDistanceOneMany <- function(Func, oneTree, manyTrees, tipLabels, 
                                  nTip = length(tipLabels), 
                                  FUN.VALUE = Func(manyTrees[[1]], oneTree,
                                                   nTip = nTip, 
                                                   tipLabels = tipLabels, ...),
                                  ...) {
  vapply(manyTrees, Func, oneTree, tipLabels = tipLabels,
                            nTip = nTip, ..., FUN.VALUE = FUN.VALUE)
}

.TreeDistanceManyMany <- function(Func, trees1, trees2, tipLabels, 
                                   nTip = length(tipLabels),
                                   FUN.VALUE = Func(trees1[[1]], trees2[[1]],
                                                    nTip = nTip, 
                                                    tipLabels = tipLabels, ...),
                                   ...) {
  if (identical(trees1, trees2)) {
    CompareAll(trees1, Func, nTip = nTip, tipLabels = tipLabels,
               FUN.VALUE = FUN.VALUE, ...)
  } else {
    seqAlong1 <- seq_along(trees1)
    names(seqAlong1) <- names(trees1)
    value <- vapply(seqAlong1, function(x) FUN.VALUE, FUN.VALUE)
    ret <- vapply(trees2, .TreeDistanceOneMany, Func = Func, manyTrees = trees1,
           tipLabels = tipLabels, nTip = nTip, FUN.VALUE = value, ...)
    if (length(dim(ret)) == 3) {
      aperm(ret, c(2, 3, 1))
    } else {
      ret
    }
  }
}

# Validate split-matrix arguments before passing to C++.
# On ARM, Rcpp::stop() can call std::terminate() instead of returning
# a proper R error, so all expected-input validation must happen in R.
.ValidateSplitArgs <- function(x, y, nTip) {
  if (ncol(x) != ncol(y)) {
    stop("Input splits must address same number of tips.")
  }
  .CheckMaxTips(nTip)
}

.CheckLabelsSame <- function(labelList) {
  nTip <- unique(lengths(labelList))
  if (length(nTip) != 1) {
    stop("All trees must contain the same number of leaves.")
  }
  tipLabel <- unique(lapply(labelList, sort))
  if (length(tipLabel) != 1L) {
    stop("All trees must bear identical labels. Found:\n >  ", 
         paste0(lapply(tipLabel, paste, collapse = " "), 
               " (", lapply(tipLabel, class), ")",
                collapse = "\n >  "))
  }
}

#' Entropy in bits
#' 
#' Calculate the entropy of a vector of probabilities, in bits.
#' Probabilities should sum to one.
#' Probabilities equalling zero will be ignored.
#' 
#' @param \dots Series of numerics, or single numeric vector, specifying
#' probabilities of outcomes (for `Entropy()`) or counts (for `Ntropy()`).
#' 
#' @return `Entropy()` and `Ntropy()` return the entropy of the specified
#' probabilities or counts, in bits.
#' 
#' @examples
#' Entropy(1/2, 0, 1/2) # = 1
#' Entropy(rep(1/4, 4)) # = 2
#' Ntropy(c(2, 2, 0, 2, 2)) # = 2
#' @template MRS
#' @export
Entropy <- function(...) {
  p <- c(...)
  p <- p[p > 0]
  -sum(p * log2(p))
}

#' @rdname Entropy
#' @export
Ntropy <- function(...) {
  entropy_int(as.integer(c(...)))
}

#' Distances between each pair of trees
#' 
#' Calculate the distance between each tree in a list, and each other tree
#' in the same list.
#' 
#' `CompareAll()` is not limited to tree comparisons:
#' `Func` can be any symmetric function.
#'
#' @param x List of trees, in the format expected by `Func()`.
#' @param Func distance function returning distance between two trees,
#' e.g. [`path.dist()`][phangorn::treedist].
#' @param FUN.VALUE Format of output of `Func()`, to be passed to [`vapply()`]. 
#' If unspecified, calculated by running `Func(x[[1]], x[[1]])`.
#' @param \dots Additional parameters to pass to `Func()`.
#' @return `CompareAll()` returns a distance matrix of class `dist` detailing
#' the distance between each pair of trees.
#' Identical trees are assumed to have zero distance.
#' 
#' @examples
#' # Generate a list of trees to compare
#' library("TreeTools", quietly = TRUE)
#' trees <- list(bal1 = BalancedTree(1:8), 
#'               pec1 = PectinateTree(1:8),
#'               pec2 = PectinateTree(c(4:1, 5:8)))
#' 
#' # Compare each tree with each other tree
#' CompareAll(trees, NNIDist)
#'   
#' # Providing FUN.VALUE yields a modest speed gain:
#' dist <- CompareAll(trees, NNIDist, FUN.VALUE = integer(7))
#'   
#' # View distances as a matrix
#' as.matrix(dist$lower)
#' @template MRS
#' @family pairwise tree distances
#' @importFrom cli cli_progress_bar cli_progress_update
#' @importFrom parallel parLapply
#' @importFrom stats dist
#' @export
CompareAll <- function(x, Func, FUN.VALUE = Func(x[[1]], x[[1]], ...),
                        ...) {
  nTree <- length(x)
  countUp <- seq_len(nTree - 1)
  i <- x[rep(countUp, rev(countUp))]
  j <- x[unlist(sapply(countUp, function(n) n + seq_len(nTree - n)))]
  
  cluster <- getOption("TreeDist-cluster")
  ret <- if (is.null(cluster)) {
    cli_progress_bar("Comparing", total = length(i))
    vapply(seq_along(i), function(k) {
      cli_progress_update(1, .envir = parent.frame(2))
      Func(i[[k]], j[[k]], ...)
      }, FUN.VALUE)
  } else {
    do.call("cbind", 
            parLapply(cluster, seq_along(i), 
                      function(k, Func) Func(i[[k]], j[[k]]),
                      Func = Func))
  }
  
  .WrapReturn <- function(dists) {
    structure(dists,
              Size = nTree,
              Labels = names(x),
              Diag = FALSE,
              Upper = TRUE,
              class = "dist")
  }
  
  # Return:
  if (length(FUN.VALUE) == 1) {
    .WrapReturn(ret)
  } else {
    structure(lapply(seq_len(dim(ret)[[1]]), 
                     function(i) .WrapReturn(unlist(ret[i, ]))),
              names = rownames(ret))
  }
}

#' Normalize tree distances
#' 
#' `NormalizeInfo()` is an internal function used to normalize information
#' against a reference, such as the total information present in a pair of
#' trees.
#' 
#' The unnormalized value(s) are normalized by dividing by a denominator
#' calculated based on the `how` parameter.  Valid options include:
#' 
#' \describe{
#'  \item{`FALSE`}{No normalization is performed; the unnormalized values
#'  are returned.}
#'  \item{`TRUE`}{Unless `infoInBoth` is specified, the information in
#'  each tree is computed using `InfoInTree()`, and the two values combined
#'  using `Combine()`.}
#'  \item{A numeric value, vector or matrix}{`how` is used as the denominator;
#'  the returned value is `unnormalized / how`.}
#'  \item{A function}{Unless `infoInBoth` is specified, the information in
#'  each tree is computed using `InfoInTree()`, and the two values combined
#'  using `how`.  `NormalizeInfo(how = Func)` is thus equivalent to
#'  `NormalizeInfo(how = TRUE, Combine = Func)`.}
#' }
#' 
#' @param unnormalized Numeric value, vector or matrix to be normalized.
#' @param tree1,tree2 Trees from which `unnormalized` was calculated.
#' @param InfoInTree Function to calculate the information content of each tree.
#' @param infoInBoth Optional numeric specifying information content of both
#' trees independently. If unspecified (`NULL`), this will be calculated using
#' the method specified by `how`.
#' @param how Method for normalization, perhaps specified using the `normalize`
#' argument to a tree distance function.  See details for options.
#' @param \dots Additional parameters to `InfoInTree()` or `how()`.
#' @returns `NormalizeInfo()` returns an object corresponding to the normalized
#' values of `unnormalized`.
#' @examples
#' library("TreeTools", quietly = TRUE)
#' pair1 <- c(BalancedTree(9), StarTree(9))
#' pair2 <- c(BalancedTree(9), PectinateTree(9))
#' 
#' # We'll let the number of nodes define the total information in a tree
#' Nnode(pair1)
#' Nnode(pair2)
#' 
#' # Let's normalize a unit distance
#' rawDist <- cbind(c(1, 1), c(1, 1))
#' 
#' # With `Combine = "+"`, the maximum distance is the sum of
#' # the information in each tree
#' denominator <- outer(Nnode(pair1), Nnode(pair2), "+")
#' 
#' NormalizeInfo(rawDist, pair1, pair2, InfoInTree = ape::Nnode, Combine = "+")
#' rawDist / denominator
#' 
#' 
#' # A denominator can be specified manually using `how`:
#' NormalizeInfo(rawDist, pair1, pair2, InfoInTree = ape::Nnode, how = 16)
#' rawDist / 16
#' 
#' 
#' # `how` also allows the denominator to be computed from trees:
#' outer(Nnode(pair1), Nnode(pair2), pmin)
#' NormalizeInfo(rawDist, pair1, pair2, InfoInTree = ape::Nnode, how = pmin)
#' rawDist / outer(Nnode(pair1), Nnode(pair2), pmin)
#' 
#' @keywords internal
#' @template MRS
#' @importFrom TreeTools KeepTip TipLabels
#' @export
NormalizeInfo <- function(unnormalized, tree1, tree2, InfoInTree,
                          infoInBoth = NULL, how = TRUE, Combine = "+", ...) {
  
  if (is.logical(how) && how == FALSE) {
    # Return:
    unnormalized
  } else {
    
    CombineInfo <- function(tree1Info, tree2Info, Combiner = Combine,
                            pairwise = FALSE) {

      if (length(tree1Info) == 1 || length(tree2Info) == 1 || pairwise) {
        unlist(.mapply(Combiner, dots = list(tree1Info, tree2Info), NULL))
      } else {
        ret <- outer(tree1Info, tree2Info, Combiner)
        if (inherits(unnormalized, "dist")) ret[lower.tri(ret)] else ret
      }
    }
    
    lab1 <- TipLabels(tree1)
    lab2 <- TipLabels(tree2)
    sameLabels <- .AllTipsSame(lab1, lab2)
    
    if (!sameLabels) {
      trees <- .SharedOnly(tree1, tree2, lab1, lab2)
      tree1 <- trees[[1]]
      tree2 <- trees[[2]]
    }
    
    if (is.logical(how)) {
      # how == FALSE case handled early above
      if (is.null(infoInBoth)) {
        info1 <- InfoInTree(tree1, ...)
        info2 <- if (is.null(tree2)) {
          info1
        } else {
          InfoInTree(tree2, ...)
        }
        infoInBoth <- CombineInfo(info1, info2, pairwise = !sameLabels)
      }
    } else if (is.function(how)) {
      if (is.null(infoInBoth)) {
        info1 <- InfoInTree(tree1, ...)
        info2 <- if (is.null(tree2)) info1 else InfoInTree(tree2, ...)
        infoInBoth <- CombineInfo(info1, info2, Combiner = how,
                                  pairwise = !sameLabels)
      }
    } else {
      infoInBoth <- how
    }
    # Return:
    unnormalized / infoInBoth
  }
}

# We only call this function when not all trees contain identical leaf sets
#' @importFrom TreeTools KeepTip TipLabels
.SharedOnly <- function(tree1, tree2,
                        lab1 = TipLabels(tree1),
                        lab2 = TipLabels(tree2)) {
  if (is.null(tree2)) {
    # Case: N trees vs themselves
    # We require a triangular matrix suitable for a dist object
    if (.MultipleTrees(tree1)) {
      pairs <- combn(seq_along(tree1), 2)
      nPairs <- dim(pairs)[[2]]
      ret <- list(vector("list", nPairs), vector("list", nPairs))
      
      for (n in seq_len(nPairs)) {
        i <- pairs[1, n]
        j <- pairs[2, n]
        common <- intersect(lab1[[i]], lab1[[j]])
        ret[[1]][[n]] <- KeepTip(tree1[[i]], common)
        ret[[2]][[n]] <- KeepTip(tree1[[j]], common)
      }
      
      # Return:
      ret
    } else {
      return(list(NULL, NULL))
    }
  } else if (.MultipleTrees(tree1)) {
    if (.MultipleTrees(tree2)) {
      # Case: N vs M
      n1 <- length(tree1)
      n2 <- length(tree2)
      nPairs <- n1 * n2
      ret <- list(vector("list", nPairs), vector("list", nPairs))
      pair1 <- rep(tree1, each = n2)
      pair2 <- rep.int(tree2, times = n1)
      lab1 <- if (is.list(lab1)) {
        rep(lab1, each = n2)
      } else {
        rep.int(list(lab1), times = n1 * n2)
      }
      lab2 <- if (is.list(lab2)) {
        rep.int(lab2, times = n1)
      } else {
        rep.int(list(lab2), times = n1 * n2)
      }
      for (i in seq_len(nPairs)) {
        common <- intersect(lab1[[i]], lab2[[i]])
        ret[[1]][[i]] <- KeepTip(pair1[[i]], common)
        ret[[2]][[i]] <- KeepTip(pair2[[i]], common)
      }
      # Return:
      ret
    } else {
      # Case: N vs 1
      if (inherits(tree2, "multiPhylo")) {
        tree2 <- tree2[[1]]
      }
      commonLeaves <- lapply(if (is.list(lab1)) lab1 else list(lab1),
                             intersect, lab2)
      list(.mapply(KeepTip, list(tree1, commonLeaves), NULL),
           lapply(commonLeaves, function (common) KeepTip(tree2, common)))
    }
  } else {
    if (.MultipleTrees(tree2)) {
      # Case: 1 vs N
      if (inherits(tree1, "multiPhylo")) {
        tree1 <- tree1[[1]]
      }
      commonLeaves <- lapply(if (is.list(lab2)) lab2 else list(lab2),
                             intersect, lab1)
      list(lapply(commonLeaves, function (common) KeepTip(tree1, common)),
           .mapply(KeepTip, list(tree2, commonLeaves), NULL))
    } else {
      # Case: 1 vs 1
      commonLeaves <- intersect(lab1, lab2)
      list(KeepTip(tree1, commonLeaves), KeepTip(tree2, commonLeaves))
    }
  }
}

.MultipleTrees <- function(tree) {
  is.list(tree) && inherits(tree[[1]], "phylo") && length(tree) > 1
}

#' List clades as text
#' @param splits1,splits2 Logical matrices with columns specifying membership
#' of each corresponding matched clade.
#' @return `ReportMatching` returns a character vector describing each pairing 
#' in a matching.
#'   
#' @seealso [`VisualizeMatching`]
#' @template MRS
#' @keywords internal
#' @export
ReportMatching <- function(splits1, splits2, realMatch = TRUE) {
  paste(as.character(splits1), ifelse(realMatch, "=>", ".."),
        as.character(splits2))
}

library("TreeTools", quietly = TRUE)


test_that("Tree normalization works", {
  expect_equal(NormalizeInfo(unnormalized = 5,
                             tree1 = BalancedTree(3),
                             tree2 = BalancedTree(5),
                             how = TRUE,
                             InfoInTree = function(x, y) NTip(x) + NTip(y),
                             y = SingleTaxonTree()),
               5 / ((3 + 1) + (3 + 1)))
  expect_equal(NormalizeInfo(unnormalized = 0:4,
                             tree1 = lapply(1:5, BalancedTree),
                             tree2 = BalancedTree(3),
                             InfoInTree = NTip,
                             Combine = min),
               0:4 / c(1, 2, 3, 3, 3))
  expect_equal(NormalizeInfo(unnormalized = 3, 1, 1, how = 4L), 3 / 4)
  expect_equal(as.dist(matrix(0:24 / 10, 5, 5)),
               NormalizeInfo(as.dist(matrix(0:24, 5, 5)),
                             rep(5, 5), rep(5, 5), InfoInTree = I),
               ignore_attr = TRUE)
  expect_equal(matrix(1:20 / 10, 5, 4) * c(10,10,1,1,1, 10,10,1,1,1, rep(1,10)),
               NormalizeInfo(1:20, c(1, 1, 10, 10, 10), c(1, 1, 10, 10), 
                             InfoInTree = I, Combine = pmax))
})

test_that("CalculateTreeDistance() errs appropriately", {
  expect_error(CalculateTreeDistance(RobinsonFouldsSplits, "Not a tree"))
  expect_error(CalculateTreeDistance(RobinsonFouldsSplits, "Not a tree", BalancedTree(8)))
  expect_error(CalculateTreeDistance(RobinsonFouldsSplits, BalancedTree(8), "Not a tree"))
})

test_that("cpp_sl_max_tips() returns a valid value", {
  expect_true(is.integer(TreeDist:::cpp_sl_max_tips()))
  expect_true(TreeDist:::cpp_sl_max_tips() >= 2048L)
})

test_that("Tip-count guard is applied consistently", {
  expect_no_error(.CheckMaxTips(1000L))

  # Direct C++ calls bypass the R guard; check_ntip() always says "not yet supported".
  cppErrMsg <- "Trees with 327.. tips are not yet supported \\(maximum 32767\\)"
  splits8 <- unclass(as.Splits(BalancedTree(8)))
  expect_error(cpp_robinson_foulds_distance(splits8, splits8, 32768L), cppErrMsg)
  expect_error(cpp_robinson_foulds_info(splits8, splits8, 32768L), cppErrMsg)
})

test_that("Tip-count guard accepts up to 32767 tips (TreeTools >= 2.3.0)", {
  skip_if(TreeDist:::cpp_sl_max_tips() <= 2048L,
          "Compiled against TreeTools < 2.3.0 (SL_MAX_TIPS = 2048)")
  expect_no_error(.CheckMaxTips(32705L))
  expect_no_error(.CheckMaxTips(32767L))
  rErrMsg <- "Trees with 327.. tips are not yet supported \\(maximum 32767\\)"
  expect_error(.CheckMaxTips(32768L), rErrMsg)

  trees <- list(BalancedTree(8), PectinateTree(8))
  class(trees) <- "multiPhylo"
  expect_error(
    .SplitDistanceAllPairs(RobinsonFouldsSplits, trees, letters[1:8], 32768L),
    rErrMsg
  )
})

test_that("Tip-count guard caps at SL_MAX_TIPS (TreeTools < 2.3.0)", {
  skip_if(TreeDist:::cpp_sl_max_tips() > 2048L,
          "Compiled against TreeTools >= 2.3.0 (SL_MAX_TIPS > 2048)")
  rErrMsg <- "Trees with 327.. tips exceed the compiled limit of 2048"
  expect_error(.CheckMaxTips(32705L), rErrMsg)

  trees <- list(BalancedTree(8), PectinateTree(8))
  class(trees) <- "multiPhylo"
  expect_error(
    .SplitDistanceAllPairs(RobinsonFouldsSplits, trees, letters[1:8], 32768L),
    rErrMsg
  )
})

test_that("Tip-count guard: TT < 2.3.0 code path (mocked)", {
  local_mocked_bindings(
    cpp_sl_max_tips = function() 2048L,
    .package = "TreeDist"
  )
  expect_no_error(.CheckMaxTips(100L))
  expect_error(.CheckMaxTips(3000L),
               "Trees with 3000 tips exceed the compiled limit of 2048")
  trees <- list(BalancedTree(8), PectinateTree(8))
  class(trees) <- "multiPhylo"
  expect_error(
    .SplitDistanceAllPairs(RobinsonFouldsSplits, trees, letters[1:8], 3000L),
    "Trees with 3000 tips exceed the compiled limit of 2048"
  )
})


test_that("CalculateTreeDistance() handles splits appropriately", {
  set.seed(101)
  tree10 <- ape::rtree(10)
  tree10.1 <- ape::rtree(10)
  splits10 <- as.Splits(tree10)
  splits10.1 <- as.Splits(tree10.1)
  trees10.3 <- list(BalancedTree(tree10),
                    PectinateTree(tree10),
                    ape::rtree(10, br = NULL))
  trees10.4 <- list(PectinateTree(tree10),
                    BalancedTree(tree10),
                    ape::rtree(10, br = NULL),
                    ape::rtree(10, br = NULL))
  splits10.3 <- as.Splits(trees10.3)
  splits10.4 <- as.Splits(trees10.4)
  
  expect_equal(0, CalculateTreeDistance(RobinsonFouldsSplits, splits10, splits10))
  expect_equal(dist(0), CalculateTreeDistance(RobinsonFouldsSplits, splits10, NULL))
  expect_equal(seq_len(7),
               attr(CalculateTreeDistance(RobinsonFouldsSplits, 
                                          splits10, splits10, TRUE), "matching")
               )
  expect_equal(CalculateTreeDistance(RobinsonFouldsSplits, splits10, splits10.1),
               CalculateTreeDistance(RobinsonFouldsSplits, splits10.1, splits10))
  expect_equal(CalculateTreeDistance(RobinsonFouldsSplits, splits10, splits10.3),
               CalculateTreeDistance(RobinsonFouldsSplits, splits10.3, splits10))
  
  expect_equal(ignore_attr = TRUE,
    CalculateTreeDistance(RobinsonFouldsSplits, splits10.3, splits10.3),
    CalculateTreeDistance(RobinsonFouldsSplits, splits10.3, trees10.3))
  
  expect_equal(ignore_attr = TRUE,
    as.matrix(CalculateTreeDistance(RobinsonFouldsSplits, trees10.3)),
    as.matrix(CalculateTreeDistance(RobinsonFouldsSplits, trees10.3, trees10.3)))
  
  expect_equal(ignore_attr = TRUE,
    as.matrix(CalculateTreeDistance(RobinsonFouldsSplits, trees10.3, trees10.3))[, c(1, 3, 2)],
    CalculateTreeDistance(RobinsonFouldsSplits, trees10.3, trees10.3[c(1, 3, 2)]))
  
  mat <- matrix(NA, 3, 4)
  for (i in 1:3) for (j in 1:4) {
    mat[i, j] <- CalculateTreeDistance(RobinsonFouldsSplits, splits10.3[[i]], splits10.4[[j]])
  }
  
  expect_equal(mat,
               CalculateTreeDistance(RobinsonFouldsSplits, splits10.3, splits10.4),
               ignore_attr = TRUE)
  
  expect_equal(ignore_attr = TRUE,
    CalculateTreeDistance(RobinsonFouldsSplits, splits10.4, splits10.3),
    t(CalculateTreeDistance(RobinsonFouldsSplits, splits10.3, splits10.4)))
})

test_that("Matches are reported", {
  # Trees copied from test-tree_distance.R
  treeSym8 <- ape::read.tree(text="((e, (f, (g, h))), (((a, b), c), d));")
  treeBal8 <- ape::read.tree(text="(((e, f), (g, h)), ((a, b), (c, d)));")
  treeTwoSplits <- ape::read.tree(text="(((a, b), c, d), (e, f, g, h));")
  
  A <- TRUE
  B <- FALSE
  splitsA <- as.Splits(c(A, A, B, B, B, B, B, B), tipLabels = letters[1:8])
  splitsB <- as.Splits(matrix(c(A, A, A, A, A, A, B, B, 
                          A, A, A, B, B, B, B, B,
                          B, B, A, A, A, A, A, A,
                          A, A, A, A, A, B, A, B), nrow = 4, byrow = TRUE),
                       tipLabels = letters[1:8])
  match <- TreeTools::match # Avoid being re-masked by "base"
  
  expect_equal("a b | c d e f g h => a b c | d e f g h",
               ReportMatching(splitsA, splitsB[[2]]))
  
  splits1 <- as.Splits(treeSym8)
  splits2 <- as.Splits(treeBal8)
  
  matchedSplits <- match(splits1, splits2)
  cs <- CompatibleSplits(splits1, splits2)
  cs[, matchedSplits] <- FALSE
  unmatched <- is.na(matchedSplits)
  matchedSplits[unmatched] <- apply(cs[unmatched, ], 1, which)
  expect_equal(matchedSplits,
               attr(GeneralizedRF(as.Splits(treeSym8), as.Splits(treeBal8), 8L, 
                                  cpp_shared_phylo, maximize = TRUE, 
                                  reportMatching = TRUE),
                    "matching"))
    
  expect_equal(matchedSplits,
               attr(SharedPhylogeneticInfoSplits(as.Splits(treeSym8),
                                                 as.Splits(treeBal8),
                                                 reportMatching = TRUE),
                    "matching"))
  
  tree1 <- PectinateTree(letters[1:8])
  tree2 <- BalancedTree(letters[8:1])
  splits1 <- as.Splits(tree1)
  splits2 <- as.Splits(tree2, tree1)
  
  .ExpectAtOK <- function(at) {
    expect_equal(4L, length(at))
    expect_equal(names(at),
                 c("matching", "matchedSplits", "matchedScores", "pairScores"))
  }
  
  Test <- function(Func, relaxed = FALSE, ...) {
    
    at <- attributes(Func(tree1, tree2, reportMatching = TRUE, ...))
    .ExpectAtOK(at)
    
    matchedSplits <- match(splits1, splits2)
    if (relaxed) {
      expect_equal(matchedSplits[!is.na(matchedSplits)],
                   as.integer(at[["matching"]][c(1, 3, 5)]))
    } else {
      cs <- CompatibleSplits(splits1, splits2)
      cs[, matchedSplits] <- FALSE
      unmatched <- is.na(matchedSplits)
      matchedSplits[unmatched] <- apply(cs[unmatched, ], 1, which)
      
      expect_equal(matchedSplits, as.integer(at$matching))
    }
    ghSplit <- at[["matchedSplits"]][
      match(as.Splits(c(rep(FALSE, 6), TRUE, TRUE), letters[1:8]),
            splits1[[which(!is.na(matchedSplits))]])]
    expect_equal(ghSplit, "g h | a b c d e f => g h | a b c d e f")
    
    at <- attributes(Func(treeSym8, treeTwoSplits, reportMatching = TRUE, ...))
    .ExpectAtOK(at)
    expect_equal(as.integer(at[["matching"]]),
                 match(as.Splits(treeSym8), as.Splits(treeTwoSplits, treeSym8)))
    expect_equal(at[["matchedSplits"]][[2]],
                 "a b | e f g h c d => a b | e f g h c d")
  }
  
  Test(SharedPhylogeneticInfo)
  Test(DifferentPhylogeneticInfo)
  Test(MatchingSplitInfo)
  Test(MatchingSplitInfoDistance)
  Test(MutualClusteringInfo)
  Test(ClusteringInfoDistance)
  
  Test(NyeSimilarity)
  Test(MatchingSplitDistance)
  Test(JaccardRobinsonFoulds, k = 2, allowConflict = FALSE)
  Test(JaccardRobinsonFoulds, k = 2, allowConflict = TRUE)
  Test(RobinsonFoulds, relaxed = TRUE)
  Test(InfoRobinsonFoulds, relaxed = TRUE)
  
  # Matching Split Distance matches differently:  
  at <- attributes(MatchingSplitDistance(treeSym8, treeBal8, 
                                         reportMatching = TRUE))
  .ExpectAtOK(at)
  expect_equal(as.integer(at$matching), c(1:3, 5:4))
  expect_equal(at[["matchedSplits"]][[5]],
               "a b | e f g h c d => a b | e f g h c d")
  
  # Zero match:
  expect_true(attr(SharedPhylogeneticInfo(
    ape::read.tree(text = "((a, b), (c, d));"),
    ape::read.tree(text = "((a, c), (b, d));"), 
    reportMatching = TRUE),
    "matchedSplits") %in% c(
      "a b | c d .. a c | b d",
      "a b | c d .. b d | a c",
      "c d | a b .. a c | b d",
      "c d | a b .. b d | a c"))
})

test_that("Matchings are calculated in both directions", {
  tree1 <- ape::read.tree(text="((1, 2), ((3, 4, (5, 9)), (6, (7, 8))));")
  tree2 <- ape::read.tree(text="((1, 2), ((3, (4, 5)), (6, (7, (8, 9)))));")
  splits1 <- as.Splits(tree1)
  splits2 <- as.Splits(tree2)
  nMatches <- min(length(splits1), length(splits2))
  
  Test <- function(Func, ...) {
    matching12 <- Func(tree1, tree2, reportMatching = TRUE, ...)
    expect_equal(nMatches, length(attr(matching12, "matchedSplits")))
    
    matching21 <- Func(tree2, tree1, reportMatching = TRUE, ...)
    expect_equal(nMatches, length(attr(matching21, "matchedSplits")))
  }
  
  Test(SharedPhylogeneticInfo)
  Test(DifferentPhylogeneticInfo)
  Test(MatchingSplitInfo)
  Test(MatchingSplitInfoDistance)
  Test(MutualClusteringInfo)
  Test(ClusteringInfoDistance)
  
  Test(NyeSimilarity)
  Test(JaccardRobinsonFoulds, k = 2, allowConflict = FALSE)
  Test(JaccardRobinsonFoulds, k = 2, allowConflict = TRUE)
  Test(MatchingSplitDistance)
  
  nMatches <- 1L
  Test(RobinsonFoulds)
  Test(InfoRobinsonFoulds)
  
})

test_that(".TreeDistance() supports all sizes", {
  expect_equal(rbind(
    bal = MASTSize(BalancedTree(7), as.phylo(0:3, 7)),
    pec = MASTSize(as.phylo(0:3, 7), PectinateTree(7))),
    MASTSize(list(bal = BalancedTree(7), pec = PectinateTree(7)),
             as.phylo(0:3, 7)))
  expect_equal(t(NNIDist(BalancedTree(7), as.phylo(0:3, 7))),
               NNIDist(list(bal = BalancedTree(7), pec = PectinateTree(7)),
                       as.phylo(0:3, 7))[1, , ])
  expect_error(.TreeDistance(RobinsonFoulds, PectinateTree(1:6),
                             PectinateTree(6)))
})

test_that("Unrooteds are handled by MAST", {
  trees <- list(UnrootTree(BalancedTree(8)), UnrootTree(PectinateTree(8)))
  expect_equal(6L, as.integer(MASTSize(trees, rooted = FALSE)),
               ignore_attr = TRUE)
})

test_that("Entropy() supports dots input", {
  expect_identical(2, Entropy(rep(1/4, 4)))
  expect_identical(1, Entropy(0, .5, 1/2))
})

test_that(".SharedOnly() works", {
  ah <- letters[1:8]
  bi <- letters[2:9]
  bh <- letters[2:8]
  ch <- letters[3:8]
  ci <- letters[3:9]
  balAH <- BalancedTree(ah)
  pecAH <- PectinateTree(ah)
  pecBI <- PectinateTree(bi)
  balCH <- BalancedTree(ch)
  pecCI <- PectinateTree(ci)
  
  expect_equal( # 1 v 1
    .SharedOnly(balAH, balCH),
    list(KeepTip(balAH, intersect(ah, ch)),
         KeepTip(balCH, intersect(ah, ch)))
  )
  expect_equal( # 1 v N
    .SharedOnly(balAH, c(balAH, balCH)),
    list(list(balAH, KeepTip(balAH, ch)),
         list(balAH, balCH))
  )
  expect_equal( # N v 1
    .SharedOnly(c(balAH, balCH), balAH),
    list(list(balAH, balCH),
         list(balAH, KeepTip(balAH, ch)))
  )
  expect_equal( # N v N, all tips differ
    .SharedOnly(c(balAH, balCH), c(pecBI, pecCI)),
    list(
      list(KeepTip(balAH, bh), KeepTip(balAH, ch),
           balCH, balCH),
      list(KeepTip(pecBI, bh), KeepTip(pecCI, ch),
           KeepTip(pecBI, ch), KeepTip(pecCI, ch))
    )
  )
  expect_equal( # N v N, each list same tips
    .SharedOnly(c(balAH, pecAH), c(pecBI, pecBI)),
    list(
      list(KeepTip(balAH, bh), KeepTip(balAH, bh),
           KeepTip(pecAH, bh), KeepTip(pecAH, bh)),
      list(KeepTip(pecBI, bh), KeepTip(pecBI, bh),
           KeepTip(pecBI, bh), KeepTip(pecBI, bh))
    )
  )
  expect_equal( # N v M, first tips same
    .SharedOnly(c(balAH, balAH), c(pecBI, pecCI, balAH)),
    list(
      list(KeepTip(balAH, bh), KeepTip(balAH, ch), balAH,
           KeepTip(balAH, bh), KeepTip(balAH, ch), balAH),
      list(KeepTip(pecBI, bh), KeepTip(pecCI, ch), balAH,
           KeepTip(pecBI, bh), KeepTip(pecCI, ch), balAH)
    )
  )
  expect_equal( # N v M, second tips same
    .SharedOnly(c(pecBI, pecCI, balAH), c(balAH, balAH)),
    list(
      list(KeepTip(pecBI, bh), KeepTip(pecBI, bh),
           KeepTip(pecCI, ch), KeepTip(pecCI, ch),
           balAH, balAH),
      list(KeepTip(balAH, bh), KeepTip(balAH, bh),
           KeepTip(balAH, ch), KeepTip(balAH, ch),
           balAH, balAH)
    )
  )
  expect_equal(.SharedOnly(balAH, NULL), list(NULL, NULL))
  expect_equal(
    .SharedOnly(c(balAH, pecBI, balCH), NULL),
    list(
      list(KeepTip(balAH, bh), KeepTip(balAH, ch),
           KeepTip(pecBI, ch)),
      list(KeepTip(pecBI, bh), balCH,
           balCH)
    )
  )
  
  t1 <- structure(list(
    t1_1 = structure(list(edge = structure(
      c(7L, 8L, 9L, 9L, 8L, 7L, 10L, 10L, 11L, 11L, 8L, 9L, 1L, 6L, 2L, 10L, 
        3L, 11L, 4L, 5L), dim = c(10L, 2L)),
      Nnode = 5L, tip.label = c("1", "2", "3", "4", "5", "6")),
      order = "preorder", class = "phylo"),
    t1_2 = structure(list(edge = structure(
      c(7L, 8L, 8L, 7L, 9L, 10L, 10L, 9L, 11L, 11L, 8L, 1L, 6L, 9L, 10L, 2L, 3L,
        11L, 4L, 5L), dim = c(10L, 2L)),
      Nnode = 5L, tip.label = c("1", "2", "3", "4", "5", "6")),
      order = "preorder", class = "phylo")),
    firstHit = c(seed = 0, final = 2), class = "multiPhylo")
  t2 <- structure(list(
    t2_1 = structure(list(
      edge = structure(c(6L, 6L, 7L, 7L, 8L, 8L, 9L, 9L, 1L, 7L, 2L, 8L, 3L, 9L,
                         4L, 5L), dim = c(8L, 2L)),
      Nnode = 4L, tip.label = c("2", "3", "4", "5", "6")), order = "preorder",
      class = "phylo")),
    firstHit = c(seed = 0, final = 1), class = "multiPhylo")
  expect_equal(class(.SharedOnly(t1, t2)[[2]][[1]]), class(t2[[1]]))
  expect_equal(class(.SharedOnly(t2, t1)[[1]][[1]]), class(t2[[1]]))
  expect_equal(class(.SharedOnly(t2, t2)[[1]][[1]]), class(t2[[1]]))
  
})

#' Helper function for shiny 3D plotting
#' @importFrom utils install.packages
#' @keywords internal
#' @export
.ThreeDPlotServer <- function(input, output, distances, clusterings,
                             mapping, mstEnds, mstSize,
                             pointCols, thinnedTrees,
                             StrainCol, PlotSize, mode3D) {
  output$threeDPlot <- shiny::renderUI({
    plotSize <- sprintf( "height: %dpx; width: %dpx;", PlotSize(), PlotSize())
    
    if (!mode3D()) {
      return(shiny::div(style = plotSize))
    }
    
    if (!requireNamespace("rgl", quietly = TRUE)) {
      is_wasm <- identical(Sys.getenv("R_WASM"), "1")
      if (!is_wasm) {
        if (!requireNamespace("rgl", quietly = TRUE)) {
          tryCatch(install.packages("rgl"), error = function(e) {})
        }
      }
    }
    
    if (!requireNamespace("rgl", quietly = TRUE)) {
      msg <- if (is_wasm) {
        shiny::p(
          shiny::strong("3D plot unavailable."),
          shiny::br(),
          "3D plots are not supported in the browser version."
        )
      } else {
        shiny::p(
          shiny::strong("3D plot unavailable."),
          shiny::br(),
          "The ",
          shiny::tags$code("rgl"),
          " package is not installed."
        )
      }
      
      return(shiny::div(
        style = paste(plotSize, "
            display: flex;
            align-items: center;
            justify-content: center;
            border: 1px solid #ddd;
            background: #fafafa;
            color: #666;
          "),
        class = "alert alert-warning",
        msg
      ))
    }
    
    rgl <- asNamespace("rgl")
    
    return(rgl$renderRglwidget({
      if (mode3D() && inherits(distances(), "dist")) {
        cl   <- clusterings()
        proj <- mapping()
        
        shiny::withProgress(message = "Drawing 3D plot", {
          rgl$open3d(useNULL = TRUE)
          shiny::incProgress(0.1)
          
          rgl$bg3d(color = "white")
          rgl$plot3d(
            proj[, 1], proj[, 2], proj[, 3],
            aspect = 1,
            axes   = FALSE,
            col    = pointCols(),
            alpha  = input$pt.opacity / 255,
            cex    = input$pt.cex,
            xlab = "", ylab = "", zlab = ""
          )
          
          shiny::incProgress(0.6)
          
          if ("labelTrees" %in% input$display) {
            rgl$text3d(proj[, 1], proj[, 2], proj[, 3], thinnedTrees())
          }
          
          if (mstSize() > 0) {
            rgl$segments3d(
              proj[t(mstEnds()), ],
              col = if ("mstStrain" %in% input$display) {
                rep(StrainCol(distances(), proj[, 1:3]), each = 2)
              } else {
                "#bbbbbb"
              },
              lty = 1
            )
          }
        })
        
        rgl$rglwidget()
      }
    }))
  })
}

#' Approximate Nearest Neighbour Interchange distance
#' 
#' Use the approach of \insertCite{Li1996;textual}{TreeDist} to approximate the
#' Nearest Neighbour Interchange distance \insertCite{Robinson1971}{TreeDist} 
#' between phylogenetic trees.
#' 
#' In brief, this approximation algorithm works by identifying edges in one
#' tree that do not match edges in the second.  Each of these edges must
#' undergo at least one NNI operation in order to reconcile the trees.
#' Edges that match in both trees need never undergo an NNI operation, and 
#' divide each tree into smaller regions.  By "cutting" matched edges into two,
#' a tree can be divided into a number of regions that solely comprise unmatched
#' edges.
#' 
#' These regions can be viewed as separate trees that need to be reconciled.
#' One way to reconcile these trees is to conduct a series of NNI operations
#' that reduce a tree to a pectinate (caterpillar) tree, then to conduct an 
#' analogue of the mergesort algorithm.  This takes at most _n_ log _n_ + O(_n_)
#' NNI operations, and provides a loose upper bound on the NNI score. 
#' The maximum number of moves for an _n_-leaf tree
#' ([OEIS A182136](https://oeis.org/A182136)) can be calculated exactly for
#' small trees \insertCite{Fack2002}{TreeDist}; this provides a tighter upper
#' bound, but is unavailable for _n_ > 12.
#' `NNIDiameter()` reports the limits on this bound.
#' 
#' 
#' \tabular{rccccccccccccc}{
#'   Leaves:   \tab 1 \tab 2 \tab 3 \tab 4 \tab 5 \tab 6 \tab 7 \tab 8 \tab 9
#'    \tab 10 \tab 11 \tab 12 \tab 13 \cr
#'   Diameter: \tab 0 \tab 0 \tab 0 \tab 1 \tab 3 \tab 5 \tab 7 \tab 10 \tab 12
#'    \tab 15 \tab 18 \tab 21 \tab ? \cr
#'  }
#' 
#' @template tree12Params
#' 
#' @return `NNIDist()` returns, for each pair of trees, a named vector
#'  containing three integers:
#' 
#' - `lower` is a lower bound on the NNI distance, and corresponds
#' to the RF distance between the trees. 
#' 
#' - `tight_upper` is an upper bound on the distance, based on calculated
#' maximum diameters for trees with < 13 leaves.  _NA_ is returned if trees are
#' too different to employ this approach.
#' 
#' - `loose_upper` is a looser upper bound on the distance, using _n_ log _n_ +
#' O(_n_).
#' 
#' 
#' @references 
#' \insertAllCited{}
#' 
#' @examples
#' library("TreeTools", quietly = TRUE)
#' 
#' NNIDist(BalancedTree(7), PectinateTree(7))
#' 
#' NNIDist(BalancedTree(7), as.phylo(0:2, 7))
#' NNIDist(as.phylo(0:2, 7), PectinateTree(7))
#'
#' NNIDist(list(bal = BalancedTree(7), pec = PectinateTree(7)),
#'         as.phylo(0:2, 7))
#'
#' CompareAll(as.phylo(30:33, 8), NNIDist)
#' @template MRS
#'   
#' @family tree distances
#' @export
NNIDist <- function(tree1, tree2 = tree1) {
  .TreeDistance(.NNIDistSingle, tree1, tree2)
}

#' @importFrom TreeTools Postorder RenumberTips
#' @importFrom ape Nnode.phylo
.NNIDistSingle <- function(tree1, tree2, nTip, ...) {
  .CheckMaxTips(nTip)
  if (nrow(tree1[["edge"]]) != nrow(tree2[["edge"]])) {
    stop("Both trees must have the same number of edges. ",
         "Is one rooted and the other unrooted?")
  }
  tree2 <- RenumberTips(tree2, as.character(tree1[["tip.label"]]))
  
  edge1 <- Postorder(tree1[["edge"]])
  edge2 <- Postorder(tree2[["edge"]])
  
  cpp_nni_distance(edge1, edge2, nTip)
}

#' Bounds on the diameter of the NNI distance
#' 
#' @param tree Object of supported class representing a tree or list of trees,
#' or an integer specifying the number of leaves in a tree/trees.
#' 
#' @return `NNIDiameter()` returns a matrix specifying (bounds on) the diameter
#' of the NNI distance metric on the specified tree(s).
#' Columns correspond to:
#' 
#' - `liMin`:  \deqn{n - 3}{_n_ &minus; 3;}, a lower bound on the diameter
#'   (Li _et al._ 1996);
#'   
#' - `fackMin`: Lower bound on diameter following Fack _et al_. (2002), i.e.
#'   \deqn{\log2{N!} / 4}{log&#8322; (_N_!) / 4};
#'   
#' - `min`: The larger of `liMin` and `fackMin`;
#' 
#' - `exact`: The exact value of the diameter, where _n_ &lt; 13;
#' 
#' - `liMax`: Upper bound on diameter following Li _et al._ (1996), i.e. 
#'   \deqn{n \log2{n} + \textrm{O}(n)}{n log&#8322; _n_ + O(_n_)};
#'   
#' - `fackMax`: Upper bound on diameter following Fack _et al_. (2002), i.e.
#'   (\deqn{N - 2}{_N_ &minus; 2}) ceiling(\deqn{\log2{n}}{log&#8322; _N_}) 
#'   + _N_;
#'   
#' - `max`: The smaller of `liMax` and `fackMax`;
#'   
#' where _n_ is the number of leaves, and _N_ the number of internal nodes,
#' i.e. \deqn{n - 2}{_n_ &minus; 2}.
#'   
#' @encoding UTF-8
#' @rdname NNIDist
#' @export
NNIDiameter <- function(tree) UseMethod("NNIDiameter")

.SortingNumber <- function(n_tip) {
  lgN <- ceiling(log2(n_tip))
  n_tip * lgN - 2^lgN + 1
}

.DegenerateDistance <- function(n_tip) { # For working, see nni_distance.cpp
  nodes_in_full <- pmax(ceiling(log2(n_tip) - log2(3)), 0L)
  tips_left <- n_tip - 2^nodes_in_full
  min_backbone_nodes <- pmax(nodes_in_full + ceiling(log2(tips_left)), 0L)
  
  # Return:
  pmax(n_tip - 2, 0) - min_backbone_nodes
}

#' @export
NNIDiameter.numeric <- function(tree) {
  n <- tree - 2L # unrooted tree with n + 2 leaves
  n[tree < 3L] <- NA
  
  liMin <- tree - 3L
  fackMin <- ceiling(lfactorial(n) / 4L / log(2))
  
  liMax <- .SortingNumber(tree) + (2 * .DegenerateDistance(tree))
  fackMax <- ((n - 2L) * ceiling(log2(n))) + n
  cbind(liMin = liMin,
        fackMin = fackMin,
        min = pmax(liMin, fackMin),
        exact = c(0, 1, 3, 5, 7, 10, 12, 15, 18, 21)[n],
        liMax = liMax,
        fackMax = fackMax,
        max = pmin(liMax, fackMax)
  )
}

#' @importFrom TreeTools NTip
#' @export
NNIDiameter.phylo <- function(tree) {
  NNIDiameter(NTip(tree))
}

#' @export
NNIDiameter.multiPhylo <- NNIDiameter.phylo

#' @export
NNIDiameter.list <- function(tree) {
  lapply(tree, NNIDiameter)
}

library("TreeTools", quietly = TRUE)

test_that("NNIDist() handles exceptions", {
  expect_error(NNIDist(list(PectinateTree(7), PectinateTree(8))),
               "trees must contain the same number of leaves")
  expect_error(NNIDist(list(PectinateTree(1:8), PectinateTree(8))),
               "trees must bear identical labels")
  expect_error(NNIDist(list(PectinateTree(1:8),
                            PectinateTree(as.character(1:8)))),
               "trees must bear identical labels")
  expect_error(NNIDist(BalancedTree(5), RootOnNode(BalancedTree(5), 1)))
})

test_that("NNIDist() rejects too many tips (TreeTools >= 2.3.0)", {
  skip_if(TreeDist:::cpp_sl_max_tips() <= 2048L,
          "Compiled against TreeTools < 2.3.0 (SL_MAX_TIPS = 2048)")
  expect_error(NNIDist(PectinateTree(40000), BalancedTree(40000)),
               "not yet supported")
  expect_error(.NNIDistSingle(PectinateTree(32769), BalancedTree(32769), 32769L),
               "not yet supported")
})

test_that("NNIDist() rejects too many tips (TreeTools < 2.3.0)", {
  skip_if(TreeDist:::cpp_sl_max_tips() > 2048L,
          "Compiled against TreeTools >= 2.3.0 (SL_MAX_TIPS > 2048)")
  expect_error(NNIDist(PectinateTree(40000), BalancedTree(40000)),
               "exceed the compiled limit")
  expect_error(.NNIDistSingle(PectinateTree(32769), BalancedTree(32769), 32769L),
               "exceed the compiled limit")
})

test_that("NNIDist() at NNI_MAX_TIPS", {
  maxTips <- 32768
  goingQuickly <- TRUE
  skip_if(goingQuickly)
  
  heapTips <- 16384 + 1
  skip_if_not_installed("testthat", "3.2.2")
  expect_no_error(.NNIDistSingle(PectinateTree(heapTips),
                                 BalancedTree(heapTips), heapTips))
  
  n <- .NNIDistSingle(PectinateTree(maxTips), BalancedTree(maxTips),
                           maxTips)
  expect_gt(n[["best_upper"]], n[["best_lower"]])
  if (!is.na(n[["tight_upper"]])) {
    expect_gte(n[["tight_upper"]], n[["best_upper"]])
  }
  expect_gte(n[["loose_upper"]], n[["best_upper"]])
  expect_gte(n[["fack_upper"]], n[["best_upper"]])
  expect_gte(n[["li_upper"]], n[["best_upper"]])
})

test_that("Simple NNI approximations", {
  nTip <- 6L
  tree1 <- BalancedTree(nTip)
  tree2 <- PectinateTree(nTip)
  edge1 <- Postorder(tree1$edge)
  edge2 <- Postorder(tree2$edge)
  
  Fack <- function(n) ((n - 2) * ceiling(log2(n))) + n
                                         
  Sorting <- function(n) {
    lc <- ceiling(log2(n))
    n * lc - 2 ^ lc + 1
  }
  DegDist <- function(n) {
    nif <- ceiling(log2(n / 3))
    tif <- 2 ^ nif
    tl = n - tif
    mbn <- nif + ceiling(log2(tl / 2)) + 1
    n - 2 - mbn
  }
  Li <- function(n_edges) Sorting(n_edges + 3) + (2 * DegDist(n_edges + 3))
  
  allMatched <- c(lower = 0L, best_lower = 0L, tight_upper = 0L,
                  best_upper = 0L, loose_upper = 0L, fack_upper = 0L,
                  li_upper = 0L)
  oneUnmatched <- c(lower = 1L, best_lower = 1L, tight_upper = 1L,
                    best_upper = 1L, loose_upper = 2L, fack_upper = 2L,
                    li_upper = Li(1))
  fiveUnmatched <- c(lower = 5L, best_lower = 10L, tight_upper = 10L,
                     best_upper = 10L, loose_upper = 18L, fack_upper = 18L,
                     li_upper = Li(5))
  
  Test <- function(tree, expect) {
    expectation <- rep(NA_integer_, 7L)
    names(expectation) <- c("lower", "best_lower", "tight_upper", "best_upper",
                            "loose_upper", "fack_upper", "li_upper")
    expectation[names(expect)] <- expect
    if (is.na(expectation["best_lower"]) && !is.na(expect["tight_upper"])) {
      expectation[c("best_lower", "best_upper")] <- expect["tight_upper"]
    }
    if (is.na(expect["loose_upper"])) {
      expectation["loose_upper"] <- min(expect[c("fack_upper", "li_upper")])
    }
    
    expect_equal(NNIDist(tree1, tree), expectation)
    for (i in c(2L, 3L, 4L, 6L)) {
      tree1i <- RootOnNode(tree1, i)
      j <- 0
      for (t2 in unique(lapply(1:9, RootOnNode, tree = tree))) {
        expect_equal(NNIDist(tree1i, t2), expectation)
      }
    }
  }
  
  expect_equal(NNIDist(BalancedTree(2), PectinateTree(2)), allMatched)
  
  expect_equal(cpp_nni_distance(edge1, edge2, NTip(tree1)), oneUnmatched)
  Test(PectinateTree(nTip), oneUnmatched)

  # Identical trees
  tree1 <- Postorder(read.tree(text = "(((a, b), (c, d)), ((e, f), (g, h)));"))
  tree2 <- Postorder(read.tree(text = "(((a, b), (d, c)), ((h, g), (f, e)));"))
  Test(tree1, allMatched)
  Test(tree2, allMatched)
  
  # Tree names
  output <- NNIDist(list(bal = tree1, pec = tree2), 
                    as.phylo(0:2, tipLabels = letters[1:8]))
  expect_equal(rownames(output), c("bal", "pec"))
  
  # Only root edge is different
  Test(Postorder(ape::read.tree(text="(((a, b), (e, f)), ((c, d), (g, h)));")),
       oneUnmatched)
  
  # Two separate regions of difference one
  Test(read.tree(text="((((a, b), c), d), (e, (f, (g, h))));"),
       oneUnmatched * 2)
  
  # One region of three unmatched edges
  Test(read.tree(text="(((a, e), (c, d)), ((b, f), (g, h)));"),
       c(lower = 3L, tight_upper = 5L, fack_upper = 8L, li_upper = Li(3)))
  
  # One region of four unmatched edges
  Test(ape::read.tree(text="(((a, e), (f, d)), ((b, c), (g, h)));"),
       c(lower = 4L, tight_upper = 7L, fack_upper = 14L, li_upper = Li(4)))
  
  # One region of five unmatched edges
  Test(ape::read.tree(text="(((a, e), (f, d)), ((b, g), (c, h)));"),
       fiveUnmatched)
  
  # Trees with different leaves at root.
  tree1 <- PectinateTree(1:8) # used in Test()
  Test(ape::read.tree(text = "(3, ((5, 6), (7, (1, (2, (4, 8))))));"),
       fiveUnmatched)
  
  # Too different for tight upper bound
  expect_true(is.na(NNIDist(BalancedTree(100), 
                            PectinateTree(100))["tight_upper"]))
  
  # Large, different trees: check that 64 leaf disagreements don't cause crash
  expect_gt(NNIDist(RandomTree(600), RandomTree(600))["li_upper"], 1)
  
})

test_that("NNI with lists of trees", {
  tree1 <- BalancedTree(1:8)
  list1 <- list(tree1, PectinateTree(as.character(1:8)),
                PectinateTree(as.character(c(4:1, 5:8))),
                BalancedTree(c(1:3, 8:4)))
  
  multResult <- NNIDist(tree1, list1)
  expect_equal(NNIDist(tree1, list1[[1]]), multResult[, 1])
  expect_equal(NNIDist(tree1, list1[[2]]), multResult[, 2])
  expect_equal(NNIDist(tree1, list1[[3]]), multResult[, 3])
  expect_equal(NNIDist(tree1, list1[[4]]), multResult[, 4])
  
  expect_equal(NNIDist(tree1, list1), NNIDist(list1, tree1))
  
  # CompareAll
  expect_equal(CompareAll(list1, NNIDist), NNIDist(list1))
  
  expect_equal(
    vapply(NNIDist(list1), function(x) unname(as.matrix(x)[1:4, 4:1]),
           matrix(0,4,4)),
    NNIDist(list1, rev(list1)),
    ignore_attr = TRUE
  )
})

test_that("NNIDiameter() is sane", {
  
  exacts <- NNIDiameter(3:12)
  expect_equal(do.call(rbind, NNIDiameter(lapply(3:12, as.integer))), exacts)
  expect_true(all(exacts[, "min"] <= exacts[, "exact"]))
  expect_true(all(exacts[, "max"] >= exacts[, "exact"]))
  expect_true(is.na(NNIDiameter(13)[, "exact"]))
  expect_true(is.na(NNIDiameter(1)[, "exact"]))
  expect_equal(NNIDiameter(BalancedTree(8))[, "exact"], c(exact = 10L))
  
  FackMin <- function(n) ceiling(0.25 * lfactorial(n) / log(2))
  exacts <- c(0, 0, 0, 1, 3, 5, 7, 10, 12, 15, 18, 21)
  liMaxes <- c(0, 1, 3, 5, 8, 13, 16, 21, 25, 31, 37, 43, 47, 53, 59, 65)
  FackMax <- function(n) n * ceiling(log2(n)) + n - (2 * ceiling(log2(n)))
  n <- 4:8
  expect_equal(NNIDiameter(n), cbind(
    liMin = n - 3L,
    fackMin = FackMin(n - 2L),
    min = pmax(n - 3L, FackMin(4:8 - 2L)),
    exact = exacts[n],
    liMax = liMaxes[n],
    fackMax = FackMax(n - 2L),
    max = pmin(liMaxes[n], FackMax(n - 2L))
  ))

  expect_equal(NNIDiameter(as.phylo(0:1, 6)), NNIDiameter(c(6, 6)))
})

#' Hierarchical Mutual Information
#'
#' Calculate the Hierarchical Mutual Information (\acronym{HMI})
#' between two trees, following the recursive algorithm of
#' \insertCite{Perotti2020;textual}{TreeDist}.
#' 
#' `HierarchicalMutualInfo()` computes the hierarchical mutual content of trees
#' \insertCite{Perotti2015,Perotti2020}{TreeDist}, which accounts for the
#' non-independence of information represented by nested splits.
#' 
#' `tree` is converted to a set of hierarchical partitions, and the mutual
#' information (in bits) is computed recursively; the contribution of a node is
#' given by:
#' 
#' \deqn{I(t,s) = \log_2(n_{ts}) - \dfrac{H_{us} + H_{tv} - H_{uv}}{n_{ts}} +
#' \text{mean}(I_{uv})}
#' 
#' Where:
#' \itemize{
#'   \item \eqn{n_{ts}} is the number of common elements between partitions
#'   \item \eqn{H_{us}, H_{tv}, H_{uv}} are entropy terms from child comparisons 
#'   \item \eqn{I_{uv}} is the recursive \acronym{HMI} for child pairs
#' }
#'
#' @template sprint
#'
#' @param tree,tree1,tree2 An object that can be coerced to an [`HPart`] 
#' object.
# (Not yet implemented: ) object, or a list of such objects.
# (Not yet implemented: ) If \code{tree2} is not provided, distances will be
# calculated between each pair of trees in the list \code{tree1}.
#' @param normalize If `FALSE`, return the raw \acronym{HMI}, in bits.
#' If `TRUE`, normalize to range \[0,1\] by dividing by
#' `max(SelfHMI(tree1), SelfHMI(tree2))`.
#' If a function, divide by `normalize(SelfHMI(tree1), SelfHMI(tree2))`.
#' 
#' @return `HierarchicalMutualInfo()` returns a numeric value representing the
#' hierarchical mutual information between the input trees, in bits,
#' normalized as specified.
#' Higher values indicate more shared hierarchical structure.
#' 
#' @examples
#' library("TreeTools", quietly = TRUE)
#' 
#' tree1 <- BalancedTree(8)
#' tree2 <- PectinateTree(8)
#' 
#' # Calculate HMI between two trees
#' HierarchicalMutualInfo(tree1, tree2)
#' 
#' # HMI normalized against the mean information content of tree1 and tree2
#' HierarchicalMutualInfo(tree1, tree2, normalize = mean)
#' 
#' @references
#' \insertAllCited{}
#' 
#' @family tree distances
#' @export
HierarchicalMutualInfo <- function(tree1, tree2 = NULL, normalize = FALSE) {
  hp1 <- as.HPart(tree1)
  if (is.null(tree2)) {
    if (isFALSE(normalize)) {
      SelfHMI(hp1)
    } else {
      warning("Normalized self-information == 1; did you mean to provide tree2?")
      1
    }
  } else {
    if (inherits(tree1, "phylo") && inherits(tree2, "phylo") &&
        NTip(tree1) != NTip(tree2)) {
      stop("Trees must have the same number of leaves")
    }
    hp2 <- as.HPart(tree2, tree1)
    hmi <- HMI_xptr(hp1, hp2)
    if (isFALSE(normalize)) {
      hmi / log(2)
    } else {
      if (isTRUE(normalize)) {
        normalize <- max
      }
      if (!is.function(normalize)) {
        stop("`normalize` must be logical, or a function")
      }
      denom <- normalize(HH_xptr(hp1), HH_xptr(hp2))
      hmi / denom
    }
  }
}

#' Hierarchical mutual information
#' 
#' An alias of `HierarchicalMutualInfo()`
#' @keywords internal
#' @export
HierarchicalMutualInformation <- HierarchicalMutualInfo

#' @rdname HierarchicalMutualInfo
#' @export
HMI <- HierarchicalMutualInfo

#' @return `SelfHMI()` returns the hierarchical mutual information of a tree
#' compared with itself, i.e. its hierarchical entropy (\acronym{HH}).
#' @examples 
#' # Normalized HMI above is equivalent to:
#' HMI(tree1, tree2) / mean(SelfHMI(tree1), SelfHMI(tree2))
#' @rdname HierarchicalMutualInfo
#' @export
SelfHMI <- function(tree) {
  part <- as.HPart(tree)
  HH_xptr(part) / log(2)
}

#' Self hierarchical mutual information
#' 
#' An alias of `SelfHMI()`
#' @export
#' @keywords internal
HH <- SelfHMI

#' @param precision Numeric; Monte Carlo sampling will terminate once the
#' relative standard error falls below this value.
#' @param minResample Integer specifying minimum number of Monte Carlo samples
#' to conduct.  Avoids early termination when sample size is too small to
#' reliably estimate the standard error of the mean.
#' @return `EHMI()` returns the expected \acronym{HMI} against a uniform
#' shuffling of element labels, estimated by performing Monte Carlo resampling
#' on the same hierarchical structure until the relative standard error of the
#' estimate falls below `precision`.
#' The attributes of the returned object list the variance (`var`),
#' standard deviation (`sd`), standard error of the mean (`sem`) and 
#' relative error (`relativeError`) of the estimate, and the number of Monte
#' Carlo samples used to obtain it (`samples`).
#' @examples
#' # Expected mutual info for this pair of hierarchies
#' EHMI(tree1, tree2, precision = 0.1)
#' @rdname HierarchicalMutualInfo
#' @export
EHMI <- function(tree1, tree2, precision = 0.01, minResample = 36) {
  if (minResample < 2L) stop("Must perform at least one resampling")
  if (precision < 1e-8) stop("Tolerance too low")
  EHMI_xptr(as.HPart(tree1), as.HPart(tree2), as.numeric(precision),
                as.integer(minResample)) / log(2)
}

.AHMISEM <- function(hmi, M, ehmi, ehmi_sem) {
  deriv <- (hmi - M) / (M - ehmi)^2
  abs(deriv) * ehmi_sem
}

#' @details `AHMI()` calculates the adjusted hierarchical mutual information:
#' \deqn{\text{AHMI}(t, s) = \dfrac{I(t, s) - \hat{I}(t, s)}{
#'  \text{mean}(H(t), H(s)) - \hat{I}(t, s)}}
#' Where:
#' - \eqn{I(t, s)} is the hierarchical mutual information between `tree1` and
#'   `tree2`
#' - \eqn{\hat{I}(t, s)} is the expected \acronym{HMI} between `tree1` and
#'   `tree2`, estimated by Monte Carlo sampling
#' - \eqn{H(t), H(s)} is the entropy (self-mutual information) of each tree
#' @param Mean Function by which to combine the self-information of the 
#' two input hierarchies, in order to normalize the \acronym{HMI}.
#' @return `AHMI()` returns the adjusted \acronym{HMI}, normalized such that
#' zero corresponds to the expected \acronym{HMI} given a random shuffling
#' of elements on the same hierarchical structure.  The attribute `sem` gives
#' the standard error of the estimate.
#' @examples
#' # The adjusted HMI normalizes against this expectation
#' AHMI(tree1, tree2, precision = 0.1)
#' @rdname HierarchicalMutualInfo
#' @export
AHMI <- function(tree1, tree2, Mean = max, precision = 0.01, minResample = 36) {
  if (minResample < 2L) stop("Must perform at least one resampling")
  if (precision < 1e-8) stop("Tolerance too low")
  hp1 <- as.HPart(tree1)
  hp2 <- as.HPart(tree2, hp1)
  
  ehmi <- EHMI_xptr(hp1, hp2, as.numeric(precision), as.integer(minResample))
  hmi <- HMI_xptr(hp1, hp2)
  hh1 <- HH_xptr(hp1)
  hh2 <- HH_xptr(hp2)
  M <- Mean(hh1, hh2)
  
  num <- hmi - ehmi[[1]]
  denom <- M - ehmi[[1]]
  # Return:
  structure(if (abs(num) < sqrt(.Machine$double.eps)) 0 else num / denom,
            sem = .AHMISEM(hmi, M, ehmi[[1]], attr(ehmi, "sem")))
}

library("TreeTools", quietly = TRUE)

test_that("Hierarchical Mutual Information", {
  
  # Create test trees
  tree1 <- BalancedTree(8)
  tree2 <- PectinateTree(8)
  tree3 <- BalancedTree(8) # Identical to tree1
  
  # Test basic functionality
  expect_no_error(HierarchicalMutualInfo(tree1, tree2))
  
  # Test that HMI is numeric and non-negative
  hmi <- HierarchicalMutualInfo(tree1, tree2)
  expect_true(is.numeric(hmi))
  expect_true(hmi >= 0)
  
  # Test normalization
  hmi_norm <- HierarchicalMutualInfo(tree1, tree2, normalize = TRUE)
  expect_true(is.numeric(hmi_norm))
  expect_true(hmi_norm >= 0)
  expect_true(hmi_norm <= 1)
  
  # Test symmetry
  hmi12 <- HierarchicalMutualInfo(tree1, tree2)
  hmi21 <- HierarchicalMutualInfo(tree2, tree1)
  expect_equal(hmi12, hmi21, tolerance = 1e-10)
  
  # Test identity property - normalized self-comparison should be 1
  hmi_self1_norm <- HierarchicalMutualInfo(tree1, tree1, normalize = TRUE)
  hmi_self2_norm <- HierarchicalMutualInfo(tree2, tree2, normalize = TRUE)
  
  expect_equal(hmi_self1_norm, 1, tolerance = 1e-10)
  expect_equal(hmi_self2_norm, 1, tolerance = 1e-10)
  
  # Test normalized identity
  hmi_self_norm <- HierarchicalMutualInfo(tree1, tree1, normalize = TRUE)
  expect_equal(hmi_self_norm, 1, tolerance = 1e-10)
  
  # Test error handling
  expect_equal(HierarchicalMutualInfo(tree1), SelfHMI(tree1))
  expect_warning(expect_equal(HierarchicalMutualInfo(tree1, norm = TRUE), 1),
                 "tree2")
  
  # Test with different tip numbers (should error)
  tree_small <- BalancedTree(6)
  expect_error(HierarchicalMutualInfo(tree1, tree_small),
               "number of leaves")
})

test_that("HMI edge cases", {
  bal9 <- BalancedTree(9)
  bal9b <- BalancedTree(paste0("t", c(3:1, 7:9, 6:4)))
  
  expect_lt(HMI(bal9, bal9b), HMI(bal9, bal9))
  
  expect_lt(HMI(bal9, bal9b, normalize = TRUE), 0.05)
  
  expect_equal(AHMI(StarTree(6), BalancedTree(6))[[1]], 0)
  expect_equal(AHMI(StarTree(2), BalancedTree(2)), structure(0, sem = NaN))
})

library("TreeTools")

test_that("HMI fails nicely", {
  bal5 <- as.HPart(BalancedTree(5))
  pec5 <- as.HPart(PectinateTree(5))
  expect_error(HierarchicalMutualInfo(bal5, pec5, normalize = "Error"),
               "`normalize` must be logical, or a function")
  
  # R-level guards in EHMI(); C++ uses ASSERT only
  expect_error(EHMI(BalancedTree(5), PectinateTree(5), precision = 1e-16),
               "Tolerance too low")
  expect_error(EHMI(BalancedTree(5), PectinateTree(5), minResample = 1),
               "Must perform at least one resampl")
})

test_that("HMI examples from Perotti et al. 2015", {
  p1 <- list(list(1, list(2, 3)), list(4, 5, 6))
  p2 <- list(1, list(2, 3), list(4, 5, 6))
  expect_equal(SelfHMI(p1), 1.011 / log(2), tolerance = 0.01)
  expect_equal(SelfHMI(p2), 1.011 / log(2), tolerance = 0.01)
  expect_equal(HMI(p1, p2), log(2) / log(2))
  expect_equal(HMI(p1, p2, normalize = TRUE), 0.685, tolerance = 0.01)
})

test_that("HMI is dependent on root position", {
  bal9 <- BalancedTree(9)
  expect_lt(HMI(RootTree(bal9, 1), RootTree(bal9, 9), normalize = TRUE), 1)
  expect_gt(SelfHMI(RootTree(bal9, 1)), SelfHMI(bal9)) # Pectination -> information
})

test_that("HMI results match hmi.pynb", {
  # Non-hierarchical
  p1 <- list(list(19, 18, 5), list(14, 16, 3), list(7), list(10, 8), list(1, 17, 9, 4, 6, 15), list(2, 13, 11), list(12, 0))
  p2 <- list( list(12, 9), list(4, 2, 0, 7), list(16), list(5), list(8, 3, 1, 14), list(11, 6, 10), list(18, 17, 19), list(13, 15))
  expect_equal(HMI(p1, p2), 0.9410980357245466 / log(2))
  
  
  # Hierarchical
  hp0 <- list(list(23),
    list(list(list(list(list(list(16), list(17)))))), # Tips above order 2 nodes
    list(list(12), list(22, 13)), list(5), list(7), list(24), list(list(list(9),
              list(list(14, 2))),
              list(list(list(list(list(list(27), list(3))))))),
    list(20, 29, 18), list(4), list(26, 15), list(list(10), list(21, 25)),
    list(11), list(list(0, 28), list(1), list(6)), list(19, 8))
  
  hp1 <- list(list(23), list(list(list(list(list(16,  17))))), list(list(12),  list(22, 13)), list(5), list(7), list(24), list(list(list(9),  list(list(14, 2))),  list(list(list(list(list(27, 3)))))), list(20, 29, 18), list(4), list(26, 15), list(list(10),  list(21, 25)), list(11), list(list(0, 28),  list(1),  list(6)), list(19, 8))
  
  hp1Collapsed <- list(
    list(23), list(16, 17), list(list(12), list(22, 13)), list(5), list(7),
    list(24), list(list(list(9), list(14, 2)), list(27, 3)), list(20, 29, 18),
        list(4), list(26, 15), list(list(10), list(21, 25)), list(11),
    list(list(0, 28), list(1), list(6)), list(19, 8))
  
  hp2 <- list(
    list(list(list(0, 25), list(24)), list(6), list(11, 28), list(8)),
    list(list(list(19), list(list(list(list(21), list(4),
                                       list(list(list(list(list(22, 7))))))))),
         list(5)), list(list(3), list(10, 23, 14)),
    list(list(27, 1, 16, 13, 18, 26, 9),
         list(list(list(list(15), list(list(list(list(list(list(12, 17)))))))),
              list(2, 20)), list(29)))
  
  expect_equal(HMI(hp1, hp2), 1.0591260408329395 / log(2))
  
  # expect_equal(HMI(hp0, hp0), 3.0140772805713665 / log(2))
  # Note that hp0 contains [[16], [17]] and [[27], [3]], whereas hp1 has
  # [16, 17] and [27, 3].  I haven't yet worked through why this should give a
  # different result.  But I don't think we are likely to encounter this case
  # in our work.
  expect_equal(HMI(hp1, hp1Collapsed), 2.921657656496707 / log(2))
  expect_equal(HMI(hp2, hp2), 2.606241391162456 / log(2))
  
  expect_equal(SelfHMI(hp1), HMI(hp1, hp1))
  
  ehmi <- structure(0.7806 / log(2), # Calculated from py with tol = 0.001
                    var = 0.01,
                    sd = 0.1, 
                    sem = 0.008,
                    relativeError = 0.01)
  ehmi_cpp <- EHMI(hp1, hp2, precision = 0.01)
  expect_gt(attr(ehmi_cpp, "samples"), 36)
  attr(ehmi_cpp, "samples") <- NULL # Could vary; no point in testing
  expect_equal(ehmi_cpp, ehmi, tolerance = 0.1)
  
  set.seed(13000)
  pyAHMI <- 0.13000 # Calculated with tol = 0.001
  expect_equal(AHMI(hp1, hp2)[[1]], pyAHMI, tolerance = 0.05)
  
  set.seed(1)
  ahmi1 <- AHMI(hp1, hp2)
  set.seed(1)
  expect_equal(AHMI(hp1, hp2), ahmi1)
  nRep <- 100
  ahmis <- replicate(nRep, AHMI(hp1, hp2))
  expect_lt(abs(attr(ahmi1, "sem") - sd(ahmis)), 0.1 * sd(ahmis))
})

test_that("HMI calculated correctly", {
  bal6 <- BalancedTree(6)
  pec6 <- PectinateTree(6)
  
  hp1 <- as.HPart(BalancedTree(6))
  hp2 <- as.HPart(PectinateTree(6))
  expect_equal(capture_output(print(hp2)),
               "Hierarchical partition on 6 leaves: t1, t2, ..., t5, t6")
  expect_equal(capture_output(print(as.HPart(BalancedTree(4)))),
               "Hierarchical partition on 4 leaves: t1, t2, t3, t4")
  expect_equal(HMI_xptr(hp1, hp2), 0.363353185)
  bal8 <- BalancedTree(8)
  pec8 <- PectinateTree(8)
  star8 <- StarTree(8)
  
  expect_equal(HierarchicalMutualInfo(bal6, pec6, normalize = TRUE),
               HMI_xptr(hp1, hp2) / max(HH_xptr(hp1), HH_xptr(hp2)))
  
  hp1 <- build_hpart_from_phylo(BalancedTree(8))
  hp2 <- build_hpart_from_phylo(PectinateTree(8))
  expect_equal(HMI_xptr(hp1, hp1), 1.38629436)
  expect_equal(HMI_xptr(hp1, hp2), 0.3342954)
  
})

test_that("HMI_cpp equals SelfHMI for same partition", {
  set.seed(1)
  tr <- BalancedTree(8)
  hp <- as.HPart(tr)
  expect_equal(SelfHMI(hp), HMI(hp, hp), tolerance = 1e-12)
})

test_that("HMI works with real dataset", {
  ch <- c(1L, 2L, 2L, 2L, 2L, 2L, 1L, 1L, 1L, 1L, 1L)
  tr <- structure(list(
    edge = structure(c(12L, 12L, 13L, 14L, 15L, 16L, 16L, 17L, 17L, 18L, 18L, 
                       15L, 14L, 19L, 20L, 20L, 19L, 13L, 21L, 21L, 1L, 13L, 
                       14L, 15L, 16L, 2L, 17L, 3L, 18L, 4L, 5L, 6L, 19L, 20L,
                       7L, 8L, 9L, 21L, 10L, 11L), dim = c(20L, 2L)),
    Nnode = 10L,
    tip.label = c("Nem", "Sco", "Eun", "Aph", "Chr", "Can", "Hel", "Cha",
                  "Lep", "Ter", "Lin")),
    class = "phylo", order = "preorder")
  chPart <- as.HPart(ch)
  
  # Build HPart from tree, then relabel
  trPart <- as.HPart(tr)
  attr(trPart, "tip.label") <- seq_along(attr(trPart, "tip.label"))
  expect_equal(attr(chPart, "tip.label"), attr(trPart, "tip.label"))
  
  # Because of the difference in levels, this test should NOT pass (!)
  # expect_equal(HMI(chPart, trPart), SelfHMI(chPart))
  
  # Relabel tree first, then build HPart
  tree <- tr
  tree$tip.label <- seq_along(tree[["tip.label"]])
  treePart <- as.HPart(tree)
  treePart
  expect_equal(HMI(trPart, treePart), SelfHMI(treePart))
  expect_equal(HMI(chPart, trPart), HMI(chPart, treePart))
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

binary_entropy_counts <- function(inSplit, nLeaves) {
    .Call(`_TreeDist_binary_entropy_counts`, inSplit, nLeaves)
}

COMCLUST <- function(trees) {
    .Call(`_TreeDist_COMCLUST`, trees)
}

robinson_foulds_all_pairs <- function(tables) {
    .Call(`_TreeDist_robinson_foulds_all_pairs`, tables)
}

robinson_foulds_cross_pairs <- function(tables_a, tables_b) {
    .Call(`_TreeDist_robinson_foulds_cross_pairs`, tables_a, tables_b)
}

consensus_info <- function(trees, phylo, p) {
    .Call(`_TreeDist_consensus_info`, trees, phylo, p)
}

HMI_xptr <- function(ptr1, ptr2) {
    .Call(`_TreeDist_HMI_xptr`, ptr1, ptr2)
}

HH_xptr <- function(ptr) {
    .Call(`_TreeDist_HH_xptr`, ptr)
}

EHMI_xptr <- function(hp1_ptr, hp2_ptr, tolerance = 0.01, minResample = 36L) {
    .Call(`_TreeDist_EHMI_xptr`, hp1_ptr, hp2_ptr, tolerance, minResample)
}

build_hpart_from_phylo <- function(phy) {
    .Call(`_TreeDist_build_hpart_from_phylo`, phy)
}

build_hpart_from_list <- function(tree, n_tip) {
    .Call(`_TreeDist_build_hpart_from_list`, tree, n_tip)
}

hpart_to_edge <- function(hpart_xptr) {
    .Call(`_TreeDist_hpart_to_edge`, hpart_xptr)
}

clone_hpart <- function(hpart_ptr) {
    .Call(`_TreeDist_clone_hpart`, hpart_ptr)
}

relabel_hpart <- function(hpart_ptr, map) {
    invisible(.Call(`_TreeDist_relabel_hpart`, hpart_ptr, map))
}

#' Calculate entropy of integer vector of counts
#' 
#' Wrapper for C++ function; no input checking is performed.
#' [`Ntropy()`] is better suited for use where performance is not critical.
#' @param n a vector of integer counts
#' @return `entropy_int()` returns a numeric corresponding to the entropy of
#' each observation, in bits.
#' @export
#' @keywords internal
entropy_int <- function(n) {
    .Call(`_TreeDist_entropy_int`, n)
}

expected_mi <- function(ni, nj) {
    .Call(`_TreeDist_expected_mi`, ni, nj)
}

cpp_kc_vector <- function(edge, tip_order) {
    .Call(`_TreeDist_cpp_kc_vector`, edge, tip_order)
}

vec_diff_euclidean <- function(vec1, vec2) {
    .Call(`_TreeDist_vec_diff_euclidean`, vec1, vec2)
}

pair_diff_euclidean <- function(vecs) {
    .Call(`_TreeDist_pair_diff_euclidean`, vecs)
}

lapjv <- function(x, maxX) {
    .Call(`_TreeDist_lapjv`, x, maxX)
}

cpp_mast <- function(edge1, edge2, nTip) {
    .Call(`_TreeDist_cpp_mast`, edge1, edge2, nTip)
}

cpp_nni_distance <- function(edge1, edge2, nTip) {
    .Call(`_TreeDist_cpp_nni_distance`, edge1, edge2, nTip)
}

#' Pairwise mutual clustering information — batch computation
#'
#' Internal function. Computes all pairwise MCI scores for a set of trees,
#' using OpenMP threads when available (falling back to single-threaded
#' execution otherwise). No interrupt checking is performed inside the
#' parallel region; the outer R call remains interruptible between batches.
#'
#' @param splits_list A list of split matrices (class `Splits` or `RawMatrix`),
#'   one per tree, all covering the same tip set.  Typically the object
#'   returned by `as.Splits(trees, tipLabels = labs, asSplits = FALSE)`.
#' @param n_tip Integer; number of tips shared by all trees.
#' @return Numeric vector of length `n*(n-1)/2` containing pairwise MCI
#'   scores in `combn(n, 2)` column-major order (i.e. the data payload of
#'   an R `dist` object).
#' @keywords internal
cpp_mutual_clustering_all_pairs <- function(splits_list, n_tip, n_threads = 1L) {
    .Call(`_TreeDist_cpp_mutual_clustering_all_pairs`, splits_list, n_tip, n_threads)
}

#' @keywords internal
cpp_rf_info_all_pairs <- function(splits_list, n_tip, n_threads = 1L) {
    .Call(`_TreeDist_cpp_rf_info_all_pairs`, splits_list, n_tip, n_threads)
}

#' @keywords internal
cpp_msd_all_pairs <- function(splits_list, n_tip, n_threads = 1L) {
    .Call(`_TreeDist_cpp_msd_all_pairs`, splits_list, n_tip, n_threads)
}

#' @keywords internal
cpp_msi_all_pairs <- function(splits_list, n_tip, n_threads = 1L) {
    .Call(`_TreeDist_cpp_msi_all_pairs`, splits_list, n_tip, n_threads)
}

#' @keywords internal
cpp_shared_phylo_all_pairs <- function(splits_list, n_tip, n_threads = 1L) {
    .Call(`_TreeDist_cpp_shared_phylo_all_pairs`, splits_list, n_tip, n_threads)
}

#' @keywords internal
cpp_jaccard_all_pairs <- function(splits_list, n_tip, k = 1.0, allow_conflict = TRUE, n_threads = 1L) {
    .Call(`_TreeDist_cpp_jaccard_all_pairs`, splits_list, n_tip, k, allow_conflict, n_threads)
}

#' @keywords internal
cpp_mutual_clustering_cross_pairs <- function(splits_a, splits_b, n_tip, n_threads = 1L) {
    .Call(`_TreeDist_cpp_mutual_clustering_cross_pairs`, splits_a, splits_b, n_tip, n_threads)
}

#' @keywords internal
cpp_rf_info_cross_pairs <- function(splits_a, splits_b, n_tip, n_threads = 1L) {
    .Call(`_TreeDist_cpp_rf_info_cross_pairs`, splits_a, splits_b, n_tip, n_threads)
}

#' @keywords internal
cpp_msd_cross_pairs <- function(splits_a, splits_b, n_tip, n_threads = 1L) {
    .Call(`_TreeDist_cpp_msd_cross_pairs`, splits_a, splits_b, n_tip, n_threads)
}

#' @keywords internal
cpp_msi_cross_pairs <- function(splits_a, splits_b, n_tip, n_threads = 1L) {
    .Call(`_TreeDist_cpp_msi_cross_pairs`, splits_a, splits_b, n_tip, n_threads)
}

#' @keywords internal
cpp_shared_phylo_cross_pairs <- function(splits_a, splits_b, n_tip, n_threads = 1L) {
    .Call(`_TreeDist_cpp_shared_phylo_cross_pairs`, splits_a, splits_b, n_tip, n_threads)
}

#' @keywords internal
cpp_jaccard_cross_pairs <- function(splits_a, splits_b, n_tip, k = 1.0, allow_conflict = TRUE, n_threads = 1L) {
    .Call(`_TreeDist_cpp_jaccard_cross_pairs`, splits_a, splits_b, n_tip, k, allow_conflict, n_threads)
}

cpp_clustering_entropy_batch <- function(splits_list, n_tip) {
    .Call(`_TreeDist_cpp_clustering_entropy_batch`, splits_list, n_tip)
}

cpp_splitwise_info_batch <- function(splits_list, n_tip) {
    .Call(`_TreeDist_cpp_splitwise_info_batch`, splits_list, n_tip)
}

path_vector <- function(edge) {
    .Call(`_TreeDist_path_vector`, edge)
}

reduce_trees <- function(x, y, original_label) {
    .Call(`_TreeDist_reduce_trees`, x, y, original_label)
}

mismatch_size <- function(x, y) {
    .Call(`_TreeDist_mismatch_size`, x, y)
}

confusion <- function(x, y) {
    .Call(`_TreeDist_confusion`, x, y)
}

keep_and_reroot <- function(tree1, tree2, keep) {
    .Call(`_TreeDist_keep_and_reroot`, tree1, tree2, keep)
}

keep_and_reduce <- function(tree1, tree2, keep) {
    .Call(`_TreeDist_keep_and_reduce`, tree1, tree2, keep)
}

spr_table_7 <- function(sp1, sp2) {
    .Call(`_TreeDist_spr_table_7`, sp1, sp2)
}

#' Transfer consensus (C++ implementation)
#'
#' @param splits_list List of raw matrices (one per tree), each from as.Splits().
#' @param n_tip Number of tips.
#' @param scale Logical: use scaled transfer distance?
#' @param greedy_best_flag Logical: TRUE for "best", FALSE for "first".
#' @param init_majority Logical: TRUE to start from majority-rule splits.
#'
#' @return A `LogicalVector` of length n_splits indicating which pooled splits
#'   are included in the consensus, plus attributes "raw_splits" (a raw matrix
#'   of all unique splits) and "light_side" (integer vector).
#' @keywords internal
cpp_transfer_consensus <- function(splits_list, n_tip, scale, greedy_best_flag, init_majority, n_threads = 1L) {
    .Call(`_TreeDist_cpp_transfer_consensus`, splits_list, n_tip, scale, greedy_best_flag, init_majority, n_threads)
}

#' @keywords internal
cpp_tc_profile <- function(splits_list, n_tip, scale, greedy_best_flag, init_majority, n_iter, n_threads = 1L) {
    .Call(`_TreeDist_cpp_tc_profile`, splits_list, n_tip, scale, greedy_best_flag, init_majority, n_iter, n_threads)
}

#' Per-pair transfer dissimilarity
#'
#' @param x,y Raw matrices representing splits (from as.Splits()).
#' @param nTip Integer: number of tips.
#'
#' @return A list with components:
#'   - score_scaled: scaled transfer dissimilarity (double)
#'   - score_unscaled: unscaled transfer dissimilarity (double)
#'   - `matching_xy`: integer vector, best match in y for each split in x (1-based, NA if sentinel)
#'   - `matching_yx`: integer vector, best match in x for each split in y (1-based, NA if sentinel)
#' @keywords internal
cpp_transfer_dist <- function(x, y, nTip) {
    .Call(`_TreeDist_cpp_transfer_dist`, x, y, nTip)
}

#' @keywords internal
cpp_transfer_dist_scored <- function(x, y, nTip, scale) {
    .Call(`_TreeDist_cpp_transfer_dist_scored`, x, y, nTip, scale)
}

#' All-pairs transfer dissimilarity (OpenMP)
#'
#' @param splits_list List of raw matrices (one per tree).
#' @param n_tip Number of tips.
#' @param scale Logical: use scaled transfer dissimilarity?
#' @param n_threads Number of OpenMP threads.
#'
#' @return Numeric vector of length choose(N,2) in dist order.
#' @keywords internal
cpp_transfer_dist_all_pairs <- function(splits_list, n_tip, scale, n_threads) {
    .Call(`_TreeDist_cpp_transfer_dist_all_pairs`, splits_list, n_tip, scale, n_threads)
}

#' Cross-pairs transfer dissimilarity (OpenMP)
#'
#' @param splits_a,splits_b Lists of raw matrices.
#' @param n_tip Number of tips.
#' @param scale Logical: use scaled transfer dissimilarity?
#' @param n_threads Number of OpenMP threads.
#'
#' @return Numeric matrix of dimension `nA` x `nB`.
#' @keywords internal
cpp_transfer_dist_cross_pairs <- function(splits_a, splits_b, n_tip, scale, n_threads) {
    .Call(`_TreeDist_cpp_transfer_dist_cross_pairs`, splits_a, splits_b, n_tip, scale, n_threads)
}

cpp_mci_impl_score <- function(x, y, n_tips) {
    .Call(`_TreeDist_cpp_mci_impl_score`, x, y, n_tips)
}

cpp_sl_max_tips <- function() {
    .Call(`_TreeDist_cpp_sl_max_tips`)
}

cpp_robinson_foulds_distance <- function(x, y, nTip) {
    .Call(`_TreeDist_cpp_robinson_foulds_distance`, x, y, nTip)
}

cpp_robinson_foulds_info <- function(x, y, nTip) {
    .Call(`_TreeDist_cpp_robinson_foulds_info`, x, y, nTip)
}

cpp_matching_split_distance <- function(x, y, nTip) {
    .Call(`_TreeDist_cpp_matching_split_distance`, x, y, nTip)
}

cpp_jaccard_similarity <- function(x, y, nTip, k, allowConflict) {
    .Call(`_TreeDist_cpp_jaccard_similarity`, x, y, nTip, k, allowConflict)
}

cpp_msi_distance <- function(x, y, nTip) {
    .Call(`_TreeDist_cpp_msi_distance`, x, y, nTip)
}

cpp_mutual_clustering <- function(x, y, nTip) {
    .Call(`_TreeDist_cpp_mutual_clustering`, x, y, nTip)
}

cpp_shared_phylo <- function(x, y, nTip, force_slow = FALSE) {
    .Call(`_TreeDist_cpp_shared_phylo`, x, y, nTip, force_slow)
}

test_that("SPR: keep_and_reroot()", {
  library("TreeTools", quietly = TRUE)
  
  tree1 <- Postorder(BalancedTree(12))
  tree2 <- Postorder(PectinateTree(12))
  keep <- as.logical(tabulate(8:12, 12))
  
  result <- keep_and_reroot(tree1, tree2, keep)
  expect_equal(result[[1]], RootTree(KeepTip(tree1, keep), 1))
  expect_equal(result[[2]], RootTree(KeepTip(tree2, keep), 1))
  
  result1 <- keep_and_reroot(tree1, tree2, 1:12 %in% 4)
  expect_equal(result1[[1]], Preorder(SingleTaxonTree("t4")))
  expect_equal(result1[[2]], Preorder(SingleTaxonTree("t4")))
  
  result0 <- keep_and_reduce(tree1, tree2, logical(12))
  expect_equal(result0[[1]], Preorder(ZeroTaxonTree()))
  
  reduced <- keep_and_reduce(tree1, tree2, keep)
  expect_equal(Preorder(reduced[[1]]), Preorder(DropTip(result[[1]], "t9")))
  expect_equal(Preorder(reduced[[2]]), Preorder(DropTip(result[[2]], "t9")))
  
  result0 <- keep_and_reduce(tree1, tree2, 1:12 %in% 4)
  expect_null(result0[[1]])
})

test_that("SPRDist() handles different inputs", {
  library("TreeTools", quietly = TRUE)
  
  expect_error(
    mismatch_size(as.Splits(c(T, T, F)), as.Splits(c(T, T, T, T))),
    "differ in `nTip"
  )
  expect_error(
    mismatch_size(matrix(as.raw(3), 1, 1), as.Splits(c(T, T, T, T))),
    "nTip attribute"
  )
  expect_error(
    mismatch_size(as.Splits(c(T, T, T, T)), matrix(as.raw(3), 1, 1)),
    "nTip attribute"
  )
  expect_error(
    mismatch_size(as.Splits(matrix(T, 2, 4)), as.Splits(c(T, T, T, T))),
    "number of splits"
  )
  splits <- as.Splits(rbind(c(T, T, T, F, F), c(T, F, F, F, T)))
  Test <- function(s1, s2) {
    expect_equal(length(s1), length(s2))
    nSplits <- length(s1)
    i <- rep(seq_len(nSplits), nSplits)
    j <- rep(seq_len(nSplits), each = nSplits)
    expect_equal(
      mismatch_size(s1, s2),
      TipsInSplits(xor(s1[[i]], s2[[j]]), smallest = TRUE)
    )
  }
  Test(as.Splits(c(T, T, T, F, F)), as.Splits(c(T, F, F, F, T)))

  set.seed(0)
  splits <- as.Splits(t(replicate(10, sample(c(T, F), 99, replace = TRUE))))
  Test(splits[[1]], splits[[2]])
  Test(splits[[1:2]], splits[[2:3]])
  Test(splits, rev(splits))
})

test_that("confusion()", {
  library("TreeTools", quietly = TRUE)
  
  TestConfusion <- function(a, b) {
    i <- rep(seq_along(a), each = length(b))
    j <- rep(seq_along(b), length(a))
    expect_equal(
      confusion(a, b),
      aperm(array(
        c(TipsInSplits(a[[i]] & b[[j]]),
          TipsInSplits(a[[i]] & !b[[j]]),
          TipsInSplits(!a[[i]] & b[[j]]),
          TipsInSplits(!a[[i]] & !b[[j]])
        ),
        c(length(a), length(b), 4)),
        c(3, 2, 1)
      )
    )
  }

  TestConfusion(as.Splits(c(T, T, T, F, F)), as.Splits(c(T, F, F, F, T)))

  set.seed(0)
  splits <- as.Splits(t(replicate(10, sample(c(T, F), 99, replace = TRUE))))
  TestConfusion(splits[[1]], splits[[2]])
  TestConfusion(splits[[1:2]], splits[[2:3]])
  TestConfusion(splits, rev(splits))
})

test_that("SPR shortcuts okay - exhaustive", {
  skip_if_not(getOption("slowMode", FALSE))
  skip_if_not_installed("TBRDist")
  library("TreeTools", quietly = TRUE)
  nTip <- 7
  allTrees <- as.phylo(seq_len(NUnrooted(nTip)), nTip)
  
  sum(apply(combn(length(allTrees), 2), 2, function(ij) {
    reduced <- ReduceTrees(allTrees[[ij[[1]]]], allTrees[[ij[[2]]]])
    r1 <- reduced[[1]]
    if (is.null(r1) || NTip(r1) != nTip) return(NA)
    r2 <- reduced[[2]]
    exact <- TBRDist::USPRDist(r1, r2)
    shortcut <- SPRDist(r1, r2, method = "Rogue")
    equal <- exact == shortcut
    if (!equal) message("Mismatch: ", paste(ij, collapse = ", "))
    expect_true(equal)
    equal
  }), na.rm = TRUE)
})

test_that("SPR shortcuts okay - known answer", {
  skip_if_not(getOption("slowMode", FALSE))
  skip_if_not_installed("TBRDist")
  library("TreeTools", quietly = TRUE)
  set.seed(0)
  trees <- lapply(1:8, function(XX) RandomTree(9, root = TRUE))
    
  exact <- apply(combn(length(trees), 2), 2, function(ij) {
    TBRDist::USPRDist(trees[[ij[[1]]]], trees[[ij[[2]]]])
  })
  opt <- options("sprShortcut" = 0)
  on.exit(options(opt))
  noCuts <- SPRDist(trees, method = "rogue")
  
  options("sprShortcut" = 4)
  cuts4 <- SPRDist(trees, method = "rogue")
  expect_true(all(cuts4 == noCuts))
  
  options("sprShortcut" = 5)
  cuts5 <- SPRDist(trees, method = "rogue")
  expect_true(all(cuts5 <= noCuts))
  expect_true(all(cuts5 >= exact))
  
  options("sprShortcut" = 6)
  cuts6 <- SPRDist(trees, method = "rogue")
  expect_true(all(cuts6 <= noCuts))
  expect_true(all(cuts6 >= exact))
  # Aspirational:
  expect_true(all(cuts6 == exact))
  
  options("sprShortcut" = 7)
  cuts7 <- SPRDist(trees, method = "rogue")
  expect_true(all(cuts7 <= noCuts))
  expect_true(all(cuts7 >= exact))
  # Aspirational:
  expect_true(all(cuts7 == exact))
})

test_that("SPR shortcuts okay - larger trees", {
  skip_if_not(getOption("slowMode", FALSE))
  
  library("TreeTools", quietly = TRUE)
  set.seed(0)
  trees <- lapply(1:8, function(XX) RandomTree(45, root = TRUE))
  opt <- options("sprShortcut" = 0)
  on.exit(options(opt))
  noCuts <- SPRDist(trees, method = "rogue")
  
  options("sprShortcut" = 4)
  cuts4 <- SPRDist(trees, method = "rogue")
  expect_true(all(cuts4 == noCuts))
  
  options("sprShortcut" = 5)
  cuts5 <- SPRDist(trees, method = "rogue")
  expect_true(all(cuts5 <= noCuts))
  
  options("sprShortcut" = 6)
  cuts6 <- SPRDist(trees, method = "rogue")
  expect_true(all(cuts6 <= noCuts))
})

test_that("SPR calculated correctly", {
  library("TreeTools", quietly = TRUE)
  
  expect_exact <- function(x, y, method = "rogue") {
    if (is.character(x)) x <- Tree(x)
    if (is.character(y)) y <- Tree(y)
    expect_equal(
      SPRDist(x, y, method = method)[[1]],
      TBRDist::USPRDist(x, y)
    )
  }
  
  Tree <- function(txt) ape::read.tree(text = txt)

  expect_equal(
    .SPRRogue(
      ape::read.tree(text = "((a, b), (c, d));"),
      ape::read.tree(text = "((a, c), (b, d));")
    )[[1]],
    1L
  )
  expect_equal(
    .SPRRogue(PectinateTree(letters[1:26]), PectinateTree(letters[c(2:26, 1)]))[[1]],
    1L
  )
  
  # Looks simple, but requires ties to be broken suitably
  expect_exact("(a,(d,(b,(c,X))));", "(a,((b,c),(X,d)));", "rogue") # distance = 1
  
  expect_exact("((((b,c),d),e),a);", "(a,(b,((e,c),d)));", "rogue")
  expect_failure(
    expect_exact("((((b,c),d),e),a);", "(a,(b,((e,c),d)));", "deo")
  )
  
  # Passes with ami, joint, vi
  # Fails with viNorm - should tiebreaker be non-normalized?
  expect_exact("(((t21,(((t15,t12),t23),t24)),t18),t19);",
               "((t21,t18),(((t12,(t19,t15)),t23),t24));", "rogue")
  
  # Example AZ33: IJK and DEF are schlepped
  # Passes with joint, ami, viNorm
  # Fails with vi (5, 6 with tiebreaker), conf (7)
  expect_equal(
    .SPRRogue(
      tree1 <- PectinateTree(letters[1:26]),
      tree2 <- Tree(
        "(g, (h, (i, (j, (k, (l, ((m, (c, (b, a))), (n, (o, (p, (q, (r, (s, (t, (u, (v, (w, (x, (y, (z, (f, (e, d))))))))))))))))))))));"
      )
    )[[1]],
    2
  )
  
  # Requires "divide and conquer" step
  expect_equal(
    .SPRRogue(
      tree1 <- PectinateTree(letters[1:26]),
      tree2 <- Tree(
        "(g, (h, (i, (j, (k, (l, (m, (n, (o, (p, (q, (r, (s, (t, (u, (v, (w, (x, (y, (z, (f, ((e, (c, (b, a))), d))))))))))))))))))))));"
      )
    )[[1]],
    2
  )
  
  lockedMid1 <- Tree("((((a1, a2), a3), ((b1, b2), b3)),
                     (((c1, c2), c3), ((d1, d2), d3)));")
  lockedMid2 <- Tree("(((a1, (a2, a3)), (c1, (c2, c3))),
                     ((b1, (b2, b3)), (d1, (d2, d3))));")
  expect_equal(.SPRRogue(lockedMid1, lockedMid2)[[1]], 5)
})

test_that("SPR.dist called safely", {
  skip("#TODO")
  library("TreeTools")
  PhangornSPR <- function(...) unname(phangorn::SPR.dist(...))
  tree1 <- as.phylo(0, 6)
  tree2 <- BalancedTree(6)
  expect_equal(SPRDist(as.phylo(0, 6), BalancedTree(6)),
               PhangornSPR(Postorder(as.phylo(0, 6)),
                           Postorder(BalancedTree(6))))
  expect_equal(SPRDist(as.phylo(0:5, 6), BalancedTree(6)), 
               PhangornSPR(structure(lapply(as.phylo(0:5, 6), Postorder),
                                     class = "multiPhylo"),
                           Postorder(BalancedTree(6))))
  expect_equal(SPRDist(BalancedTree(6), as.phylo(0:5, 6)),
               SPRDist(as.phylo(0:5, 6), BalancedTree(6)))
  expect_equal(SPRDist(BalancedTree(6), PectinateTree(6)),
               SPRDist(list(BalancedTree(6), PectinateTree(6)))[[1]],
               ignore_attr = TRUE)
  
  reduced <- keep_and_reduce(tree1, tree2, keep)
  expect_equal(Preorder(reduced[[1]]), Preorder(DropTip(result[[1]], "t9")))
  expect_equal(Preorder(reduced[[2]]), Preorder(DropTip(result[[2]], "t9")))
})


test_that("SPR deOliveira2008 calculation looks valid", {
  library("TreeTools", quietly = TRUE)
  
  # We do not expect to obtain identical results to phangorn::SPR.dist,
  # because ties are broken in a different arbitrary manner.
  # We're thus left with quite a loose test.
  expect_equal(SPRDist(PectinateTree(letters[1:26]),
                       PectinateTree(letters[c(2:26, 1)]),
                       method = "deOliv"),
               1L)
})


test_that("SPR: Under the hood", {
  expect_error(mismatch_size(as.Splits(c(T, T, F)), as.Splits(c(T, T, T, T))),
               "differ in `nTip")
  expect_error(mismatch_size(matrix(as.raw(3), 1, 1), 
                             as.Splits(c(T, T, T, T))),
               "nTip attribute")
  expect_error(mismatch_size(as.Splits(c(T, T, T, T)),
                             matrix(as.raw(3), 1, 1)),
               "nTip attribute")
  expect_error(mismatch_size(as.Splits(matrix(T, 2, 4)),
                             as.Splits(c(T, T, T, T))),
               "number of splits")
  splits <- as.Splits(rbind(c(T, T, T, F, F),
                            c(T, F, F, F, T)))
  Test <- function (s1, s2) {
    expect_equal(length(s1), length(s2))
    nSplits <- length(s1)
    i <- rep(seq_len(nSplits), nSplits)
    j <- rep(seq_len(nSplits), each = nSplits)
    expect_equal(mismatch_size(s1, s2),
                 TipsInSplits(xor(s1[[i]], s2[[j]]), smallest = TRUE))
  }
  Test(as.Splits(c(T, T, T, F, F)), as.Splits(c(T, F, F, F, T)))
  
  set.seed(0)
  splits <- as.Splits(t(replicate(10, sample(c(T, F), 99, replace = TRUE))))
  Test(splits[[1]], splits[[2]])
  Test(splits[[1:2]], splits[[2:3]])
  Test(splits, rev(splits))
})

test_that("confusion() fails gracefully", {
  x <- as.Splits(c(T, T, T, F, F))
  xNoTip <- x
  attr(xNoTip, "nTip") <- NULL
  xx <- as.Splits(rep(c(T, F, F, F, T), 4))
  expect_error(confusion(x, xx), "differ in `nTip`")
  expect_error(confusion(xNoTip, x), "`x` lacks nTip attribute")
  expect_error(confusion(x, xNoTip), "`y` lacks nTip attribute")
  expect_error(confusion(x, xx), "differ in `nTip`")
  expect_error(confusion(c(x, x), x), "number of splits")
})

test_that("confusion()", {
  TestConfusion <- function (a, b) {
    i <- rep(seq_along(a), each = length(b))
    j <- rep(seq_along(b), length(a))
    expect_equal(
      confusion(a, b),
      aperm(array(c(TipsInSplits(a[[i]] & b[[j]]),
                    TipsInSplits(a[[i]] & !b[[j]]),
                    TipsInSplits(!a[[i]] & b[[j]]),
                    TipsInSplits(!a[[i]] & !b[[j]])),
                  c(length(a), length(b), 4)), c(3, 2, 1))
    )
  }
  
  TestConfusion(as.Splits(c(T, T, T, F, F)), as.Splits(c(T, F, F, F, T)))
  
  set.seed(0)
  splits <- as.Splits(t(replicate(10, sample(c(T, F), 99, replace = TRUE))))
  TestConfusion(splits[[1]], splits[[2]])
  TestConfusion(splits[[1:2]], splits[[2:3]])
  TestConfusion(splits, rev(splits))
})

test_that("SPRDist handles input formats", {
  bal9 <- BalancedTree(9)
  pec9 <- PectinateTree(9)
  
  expect_null(SPRDist(bal9))
  expect_equal(SPRDist(bal9, bal9), 0)
  
  expect_equal(SPRDist(list(bal9, bal9), bal9), c(0, 0))
  expect_equal(SPRDist(c(bal9, bal9), bal9), c(0, 0))
  expect_equal(SPRDist(bal9, list(bal9, bal9)), c(0, 0))
  expect_equal(SPRDist(bal9, c(bal9, bal9)), c(0, 0))
  
  expect_equal(SPRDist(list(bal9, pec9, pec9), list(pec9, bal9)),
               matrix(c(2, 0, 0, 0, 2, 2), 3, 2))
  expect_equal(SPRDist(c(bal9, pec9, pec9), c(pec9, bal9)),
               matrix(c(2, 0, 0, 0, 2, 2), 3, 2))
  
  self <- SPRDist(list(bal9, pec9))
  at <- attributes(self)
  expect_equal(at[["Size"]], 2)
  expect_equal(at[["class"]], "dist")
  expect_equal(at[["Diag"]], FALSE)
  expect_equal(at[["Upper"]], TRUE)
  expect_equal(self[[1]], dist(c(0, 2))[[1]])
})

test_that("SPR deOliveira2008 calculation looks valid", {
  # We do not expect to obtain identical results to phangorn::SPR.dist,
  # because ties are broken in a different arbitrary manner.
  # We're thus left with quite a loose test.
  Tree <- function (txt) ape::read.tree(text = txt)
  
  expect_equal(SPRDist(PectinateTree(letters[1:26]),
                       PectinateTree(letters[c(2:26, 1)]),
                       method = "deOliv"),
               1L)
  
  nTip <- 130
  nSPR <- 35
  
  set.seed(0)
  skip_if_not_installed("TreeSearch")
  tr <- vector("list", nSPR + 1L)
  tr[[1]] <- Postorder(TreeTools::RandomTree(nTip, root = TRUE))
  expect_equal(SPRDist(tr[[1]], tr[[1]]), 0)
  for (i in seq_len(nSPR) + 1L) {
    tr[[i]] <- Postorder(TreeSearch::SPR(tr[[i - 1]]))
  }
  
  testDist <- as.matrix(SPRDist(tr, method = "de Oliv"))
  simDist <- as.matrix(dist(seq_along(tr)))
  biggerThyNeighbour <- sapply(1:nSPR, function(i) sapply(2:nSPR, function(j)
    if (i < j) {
      testDist[i, j] - testDist[i, j - 1]
    } else {
      testDist[i, j - 1] - testDist[i, j]
    }
  ))
  
  errors <- biggerThyNeighbour < 0
  # We may see a few "errors" due to chance, but expect these to be rare
  rare <- nSPR
  expect_lt(sum(errors), rare)
  if (interactive() && any(errors)) {
    testDist[colSums(errors) > 0, ]
  }
  
  overShot <- as.matrix(testDist) > as.matrix(simDist)
  # We may overshoot where there are "knots" in trees and the optimal SPR
  # path is not equivalent to just pruning shared subtrees; these cases
  # ought to be rare.
  rare <- 0.10
  expect_lt(sum(overShot) / length(overShot), rare)
  
  if (interactive()) {
    # View these cases:
    overs <- colSums(overShot) > 0
    overShot[overs, overs]
  }
})

test_that("binary_entropy_counts() fails gracefully", {
  expect_equal(binary_entropy_counts(integer(0), 0), numeric(0))
  expect_equal(binary_entropy_counts(integer(3), 0), rep(NA_real_, 3))
  expect_equal(binary_entropy_counts(integer(3), 2), rep(0, 3))
  expect_equal(binary_entropy_counts(c(1, NA, 2), 2), c(1, NA, 0))
})

library("TreeTools", quietly = TRUE)

# Tests that distance functions handle trees larger than the old SL_MAX_TIPS
# threshold (2048 in TreeTools <= 2.2.0).  These require TreeTools >= 2.3.0
# for heap-backed split storage; they skip gracefully on older builds.

test_that("Distance functions handle trees exceeding TT 2.2.0 limit", {
  skip_on_cran()
  n <- 2049L
  skip_if(TreeDist:::cpp_sl_max_tips() < n,
          "Requires TreeTools >= 2.3.0 (SL_MAX_TIPS > 2048)")

  t1 <- as.phylo(1, n)
  t2 <- as.phylo(2, n)

  expect_gt(RobinsonFoulds(t1, t2)[[1]], 0)
  expect_equal(RobinsonFoulds(t1, t1)[[1]], 0)

  expect_gt(InfoRobinsonFoulds(t1, t2)[[1]], 0)
  expect_equal(InfoRobinsonFoulds(t1, t1)[[1]], 0)

  expect_gt(ClusteringInfoDistance(t1, t2)[[1]], 0)
  expect_equal(ClusteringInfoDistance(t1, t1)[[1]], 0)

  expect_gt(DifferentPhylogeneticInfo(t1, t2)[[1]], 0)
  expect_equal(DifferentPhylogeneticInfo(t1, t1)[[1]], 0)

  expect_gt(MatchingSplitInfoDistance(t1, t2)[[1]], 0)
  expect_equal(MatchingSplitInfoDistance(t1, t1)[[1]], 0)
})

test_that("Distance scores agree across stack/heap storage threshold", {
  skip_on_cran()
  skip_if_not_installed("TreeTools", "2.3.0")
  # SL_STACK_SPLITS in TreeTools 2.3.0+ admits trees of n_tips <= 2048 to
  # stack-backed split storage; n_tips == 2049 forces the heap path.  This
  # test compares each metric on a stack-storage tree pair against the same
  # pair after inserting a cherry next to tip 1, which both pushes them
  # across the threshold and leaves their split structure invariant: each
  # existing split simply gains the new tip on whichever side already
  # contained tip 1, and one perfectly-matched cherry split is added
  # (contributing 0 to any distance).
  stackMax <- 2048L
  skip_if(TreeDist:::cpp_sl_max_tips() <= stackMax,
          "Requires TreeTools >= 2.3.0 (heap-backed storage)")

  t1s <- as.phylo(1234, stackMax)
  t2s <- PectinateTree(stackMax)
  t1h <- AddTip(t1s, where = 1, label = "extra")
  t2h <- AddTip(t2s, where = 1, label = "extra")

  # RF (count of unmatched splits) is invariant under the inserted cherry.
  expect_equal(RobinsonFoulds(t1s, t2s)[[1]],
               RobinsonFoulds(t1h, t2h)[[1]])

  # Per-split info depends on n_tips and on side sizes, so absolute
  # information-theoretic distances drift by a few % when n grows by 1;
  # normalized distances are far less sensitive.
  tol <- 0.002

  expect_equal(InfoRobinsonFoulds(t1s, t2s, normalize = TRUE)[[1]],
               InfoRobinsonFoulds(t1h, t2h, normalize = TRUE)[[1]],
               tolerance = tol)
  expect_equal(ClusteringInfoDistance(t1s, t2s, normalize = TRUE)[[1]],
               ClusteringInfoDistance(t1h, t2h, normalize = TRUE)[[1]],
               tolerance = tol)
  expect_equal(DifferentPhylogeneticInfo(t1s, t2s, normalize = TRUE)[[1]],
               DifferentPhylogeneticInfo(t1h, t2h, normalize = TRUE)[[1]],
               tolerance = tol)
  expect_equal(MatchingSplitInfoDistance(t1s, t2s, normalize = TRUE)[[1]],
               MatchingSplitInfoDistance(t1h, t2h, normalize = TRUE)[[1]],
               tolerance = tol)
})

test_that("RF and IRF work for 8000-tip trees", {
  skip_on_cran()
  skip_if(TreeDist:::cpp_sl_max_tips() < 2049L,
          "Requires TreeTools >= 2.3.0 (SL_MAX_TIPS > 2048)")
  skip_if(!getOption("slowMode", FALSE), "Only runs in slow mode")

  n <- 8000
  t1 <- PectinateTree(n)
  t2 <- BalancedTree(n)

  expect_gt(RobinsonFoulds(t1, t2)[[1]], 0)
  expect_equal(RobinsonFoulds(t1, t1)[[1]], 0)

  expect_gt(InfoRobinsonFoulds(t1, t2)[[1]], 0)
  expect_equal(InfoRobinsonFoulds(t1, t1)[[1]], 0)
})

test_that("Slow lookup path produces identical SPI to fast path", {
  skip_on_cran()
  # Coverage for the n_tips > SL_MAX_TIPS branches in shared_phylo:
  #   - lg2_*_slow fallback functions
  #   - one_overlap<false>, one_overlap_notb<false>, spi_overlap<false>
  # These are only triggered with > 32705 tips in production, but the
  # cpp_shared_phylo entry accepts a force_slow flag so the same code paths
  # can be exercised on small trees.  Both paths must produce identical
  # values: lg2_rooted_lookup() == lg2_rooted[] within the table range.
  t1 <- as.phylo(0, 30)
  t2 <- as.phylo(11, 30)
  splits1 <- as.Splits(t1)
  splits2 <- as.Splits(t2)

  fast <- TreeDist:::cpp_shared_phylo(splits1, splits2, 30L, force_slow = FALSE)
  slow <- TreeDist:::cpp_shared_phylo(splits1, splits2, 30L, force_slow = TRUE)

  expect_equal(fast$score, slow$score)
  expect_equal(fast$matching, slow$matching)

  # Identical-tree case: both paths must report the self-information score.
  fast_self <- TreeDist:::cpp_shared_phylo(splits1, splits1, 30L,
                                           force_slow = FALSE)
  slow_self <- TreeDist:::cpp_shared_phylo(splits1, splits1, 30L,
                                           force_slow = TRUE)
  expect_equal(fast_self$score, slow_self$score)
})

test_that("Tip-count ceiling is enforced correctly", {
  skip_on_cran()
  skip_if(TreeDist:::cpp_sl_max_tips() < 2049L,
          "Requires TreeTools >= 2.3.0 (SL_MAX_TIPS > 2048)")

  expect_no_error(.CheckMaxTips(32767L))
  expect_error(.CheckMaxTips(32768L), "not yet supported")
})

test_that("Day 1985 does not overflow", {
  bigTree <- PectinateTree(2^14 + 1)
  expect_no_error(TreeTools::as.ClusterTable(bigTree))
  expect_equal(RobinsonFoulds(list(bigTree, bigTree))[[1]], 0)
})

test_that("Day 1985 examples", {
  library("TreeTools", quietly = TRUE)
  
  PrepareTree <- function(text) {
    tmp <- ape::read.tree(text = text)
    RenumberTips(tmp, as.character(seq_along(tmp$tip.label)))
  }
  
  TestRF <- function(t1, t2) {
    expect_equal(RobinsonFoulds(t1, t2), 
                 NSplits(t1) + NSplits(t2) - (2 * COMCLUST(list(t1, t2))))
  }
  
  t1 <- PrepareTree("(1, (2, (3, (4, (5, 6)))));")
  t2 <- PrepareTree("(1, (2, ((4, 3), (6, 5))));")
  t3 <- PrepareTree("(1, (((2, 3), 5), (4, 6)));")
  tStar <- PrepareTree("(1, 2, 3, 4, 5, 6);")
  c1 <- as.ClusterTable(t1)
  c2 <- as.ClusterTable(t2)
  c3 <- as.ClusterTable(t3)
  cStar <- as.ClusterTable(tStar)
  t1p <- ape::read.tree(text = write.tree(t1))
  t2p <- ape::read.tree(text = write.tree(t2))
  tStarp <- ape::read.tree(text = write.tree(tStar))
  
  expect_equal(2L, robinson_foulds_all_pairs(list(c1, c2)))
  
  TestRF(t1, t2)
  TestRF(t1, t3)
  TestRF(t2, t3)
  TestRF(t1, tStar)
  TestRF(tStar, tStar)
  expect_equal(c(RobinsonFoulds(t1, c(t2, t3, tStar)), RobinsonFoulds(t2, t3),
                 NSplits(c(t2, t3))),
               as.integer(RobinsonFoulds(c(t1, t2, t3, tStar))))
  
  # Day's figure 2
  t1 <- PrepareTree("((10, 7), (6, (8, 11)), (12, (4, (2, 1))), 14, (5, 9, 13), 3);")
  as.ClusterTable(t1)
  t2 <- PrepareTree("(((2, 4, 5, 7, 9, 10, 12, 13), 1, 14), (6, (8, 11)), 3);")
  as.ClusterTable(t2)
  
  expect_equal(COMCLUST(list(t1, t2)), 2L)
  expect_equal(as.integer(RobinsonFoulds(list(t1, t2))), 7L)

  # Large trees
  twoBiggies <- list(BalancedTree(20000), BalancedTree(20000))
  expect_no_error(COMCLUST(twoBiggies))
  expect_equal(RobinsonFoulds(twoBiggies)[[1]], 0)
})
  
test_that("RobinsonFoulds() with realistic trees", {
  # Test exposes segfault that has arisen whilst modernizing C++
  tr1 <- structure(list(
    edge = structure(c(8L, 8L, 9L, 10L, 10L, 9L, 11L, 11L, 8L, 12L, 12L,
                       1L, 9L, 10L, 2L, 3L, 11L, 4L, 5L, 12L, 6L, 7L),
                     dim = c(11L, 2L)),
    Nnode = 5L, tip.label = c("t1", "t2", "t3", "t4", "t5", "t6", "t7")),
    class = "phylo", order = "preorder")
  tr2 <- structure(list(
    edge = structure(c(8L, 9L, 10L, 10L, 9L, 11L, 11L, 8L, 12L, 12L, 8L,
                       9L, 10L, 1L, 2L, 11L, 3L, 4L, 12L, 5L, 6L, 7L),
                     dim = c(11L, 2L)),
    Nnode = 5L, tip.label = c("t1", "t2", "t3", "t4", "t5", "t6", "t7")),
    class = "phylo", order = "preorder")
  t4 <- list(a = tr1, b = tr2, c = tr1, d = tr2)
  r4 <- TreeTools::RootTree(t4, 1)
  
  expect_equal(as.numeric(RobinsonFoulds(r4)),
               c(8, 0, 8, 8, 0, 8))
})

test_that("ReduceTrees() handles invalid input", {
  library("TreeTools", quietly = TRUE)
  expect_error(ReduceTrees("STRING", as.phylo(1, 4)), "must be a `phylo`")
  expect_error(ReduceTrees(as.phylo(1, 4), logical(2)), "must be a `phylo`")
  expect_error(ReduceTrees(as.phylo(1, 4), as.phylo(1, 5)), "same leaf labels")
  expect_error(ReduceTrees(BalancedTree(5), StarTree(5)), "binary")
  expect_null(reduce_trees(BalancedTree(3)$edge,
                           PectinateTree(3)$edge, letters[1:3])[[1]])
})

test_that("ReduceTrees()", {
  bal9 <- TreeTools::BalancedTree(9)
  pec9 <- TreeTools::PectinateTree(9)
  expect_null(ReduceTrees(bal9, bal9))
  if (interactive()) {
    par(mai = rep(0.1, 4), mfrow = c(2, 2))
    plot(RootTree(bal9, 1)); nodelabels()
    plot(RootTree(pec9, 1)); nodelabels()
  }
  confl <- ReduceTrees(bal9, pec9)
  expect_true(
    all.equal(confl[[1]],
              ape::read.tree(text = "(t1, ((t4, t5), ((t6, t7), t9)));")))
  expect_true(all.equal(confl[[2]], DropTip(pec9, c(2, 3, 8))))
  
  pec9b <- RenumberTips(TreeTools::PectinateTree(paste0('t', c(2:9, 1))),
                        pec9)
  pecred <- ReduceTrees(pec9b, pec9)
  expect_true(all.equal(pecred[[2]], PectinateTree(5)))
  expect_true(all.equal(pecred[[1]], PectinateTree(paste0('t', c(1, 5:2)))))
  
  long1 <- ape::read.tree(
    text = "(a, (b, (c, (d, (e, (f, ((g, (X, Y)), (h, (i, j)))))))));")
  long2 <- ape::read.tree(
    text = "(b, (c, (d, (e, (f, (g, ((h, (X, Y)), (i, (j, a)))))))));")
  longRed <- ReduceTrees(long1, long2)
  expect_true(all.equal(longRed[[1]],
                        DropTip(long1, c("b", "c", "X"))))
  expect_true(all.equal(longRed[[2]],
                        RootTree(DropTip(long2, c("b", "c", "X")), "a")))
  
  long1 <- ape::read.tree(
    text = "(r, (oo, (t, (a, (b, (c, (d, (e, (f, ((g, (X, Y)), (h, (i, j))))))))))));")
  long2 <- ape::read.tree(
    text = "(r, (oo, (t, (b, (c, (d, (e, (f, (g, ((h, (X, Y)), (i, (j, a))))))))))));")
  longRed <- ReduceTrees(long1, long2)
  expect_true(all.equal(longRed[[1]],
                        DropTip(long1, c("t", "oo", "b", "c", "X"))))
  expect_true(all.equal(longRed[[2]],
                        DropTip(long2, c("t", "oo", "b", "c", "X"))))
  
  tree1 <- ape::read.tree(
    text = "(t3, ((t21, t17), (t4, (((((t9, t22), t25), t23), t24), t20))));")
  tree2 <- ape::read.tree(
    text = "(t3, ((t17, t21), (t4, (((((t9, t25), t23), t22), t24), t20))));")
  expect_true(all.equal(ReduceTrees(tree1, tree2)[[1]],
                        DropTip(tree1, c("t4", "t17", "t21", "t20", "t24"))))

  # Lift root AND reduce chain to new root base
  keptRoot1 <- ape::read.tree(text = "((a,(b,((((c,(d,(((e,(((f,((g,(h,((((((i,j),k),l),m),n),o))),p)),q),r)),s),t))),u),v),x))),root);")
  keptRoot2 <- ape::read.tree(text = "((a,(b,(((c,(d,(((e,(((f,((g,(h,(((((((x,i),j),k),l),m),n),o))),p)),q),r)),s),t))),u),v))),root);")
  tree1 <- RenumberTips(keptRoot1, c("root", letters))
  tree2 <- keptRoot2
  expect_equal(c("root", "x", "a", "b") %in% TipLabels(ReduceTrees(tree1, tree2)),
               c(TRUE, TRUE, FALSE, FALSE))
})

#' Hierarchical partition structure
#' 
#' A structure of class `HPart` comprises a pointer to a C++ representation of
#' hierarchical partitions, with the attribute `tip.label` recording the
#' character labels of its leaves.  `HPart` objects with identical tip labels
#' can be compared using [`HierarchicalMutualInfo()`].
#' 
#' 
#' An `HPart` object may be created from various representations of hierarchical
#' structures:
#' 
#' - a tree (possibly phylogenetic) of class `phylo`
#' - A hierarchical list of lists, in which elements are represented by integers
#'   1 to n
#' - A vector, which will be interpreted as a flat structure
#'  in which all elements bearing the same label are assigned to the same cluster
#' 
#' @param tree An object to convert to an HPart structure, in a supported format
#' (see details).
#' @returns `HPart()` returns a structure containing a pointer to a C++
#' representation of a hierarchical partition structure.
#' @name HPart
#' @export
as.HPart <- function(tree, tipLabels) {
  UseMethod("as.HPart")
}

#' @export
#' @rdname HPart
as.HPart.HPart <- function(tree, tipLabels = NULL) {
  if (is.null(tipLabels) || identical(tipLabels, TipLabels(tree))) {
    tree
  } else {
    RenumberTips(tree, tipLabels)
  }
}

#' @rdname HPart
#' @export
as.HPart.default <- function(tree, tipLabels = NULL) {
  if (is.null(dim(tree))) {
    structure(build_hpart_from_list(
      lapply(unique(tree), function(x) as.list(which(tree == x))),
      length(tree)),
      tip.label = seq_along(tree),
      class = "HPart"
    )
  } else {
    stop("no applicable method for 'as.HPart' applied to an object of class \"",
         paste(class(tree), collapse = "\", \""), "\"")
  }
}


#' @rdname HPart
#' @export
as.HPart.list <- function(tree, tipLabels = NULL) {
  # Flatten to verify leaves
  leaves <- unlist(tree, recursive = TRUE)
  if (!all(is.numeric(leaves)) || any(leaves != as.integer(leaves))) {
    stop("All leaves must be integers.")
  }
  tree <- rapply(tree, as.integer, how = "replace")
  if (0 %in% leaves) {
    tree <- rapply(tree, function(x) x + 1L, how = "replace")
    leaves <- leaves + 1
  }
  n_tip <- length(unique(leaves))
  expected <- seq_len(n_tip)
  if (!isTRUE(all.equal(sort(leaves), expected))) {
    stop("Leaves must contain each integer 1..n exactly once")
  }
  
  hpart_ptr <- build_hpart_from_list(tree, n_tip)
  ret <- structure(hpart_ptr, tip.label = as.character(expected), class = "HPart")
  if (!is.null(tipLabels) && !is.list(tipLabels)) {
    RenumberTips(ret, tipLabels)
  }
  ret
}

#' @export
#' @inheritParams TreeTools::as.ClusterTable
#' @rdname HPart
as.HPart.phylo <- function(tree, tipLabels = TipLabels(tree)) {
  if (!identical(TipLabels(tree), tipLabels)) {
    tree <- RenumberTips(tree, tipLabels)
  }
  structure(build_hpart_from_phylo(tree), tip.label = tipLabels,
            class = "HPart")
}

#' @rdname HPart
#' @export
is.HPart <- function(x) {
  inherits(x, "HPart")
}

#' @export
NTip.HPart <- function(phy) {
  length(TipLabels(phy))
}

#' @rdname HPart
#' @export
print.HPart <- function(x, ...) {
  nTip <- NTip(x)
  tips <- TipLabels(x)
  cat("Hierarchical partition on", nTip, "leaves: ")
  if (nTip > 5) {
    cat(paste0(c(tips[1:2], "...", tips[length(tips) - 1:0]), collapse = ", "))
  } else {
    cat(paste0(tips, collapse = ", "))
  }
}

#' @rdname HPart
#' @importFrom ape as.phylo
#' @export
as.phylo.HPart <- function(x, ...) {
  edge <- hpart_to_edge(x)
  labels <- TipLabels(x)
  nNode <- dim(edge)[[1]] - length(labels) + 1
  structure(list(edge = edge, Nnode = nNode, tip.label = labels),
            class = "phylo",
            order = "cladewise")
}

#' @rdname HPart
#' @param x `HPart` object to plot.
#' @param \dots Additional arguments to \code{\link[ape:plot.phylo]{plot.phylo}}.
#' @export
plot.HPart <- function(x, ...) {
  plot(as.phylo(x), ...)
}

#' Clone / duplicate an object
#' `clone()` physically duplicates objects
#' @param x the object to be cloned
#' @param \dots additional parameters for methods
#' @return `clone()` typically returns an object of the same class and "value"
#' as the input `x`.
#' @keywords internal
#' @export
clone <- function(x, ...) UseMethod("clone")

#' @template MRS
#' @rdname clone
#' @inheritParams TreeTools::as.ClusterTable
#' @export
clone.HPart <- function(x, tipLabels = attr(x, "tip.label"), ...) {
  structure(clone_hpart(x), tip.label = tipLabels,
            class = "HPart")
}

#' @importFrom TreeTools MatchStrings
#' @inheritParams TreeTools::RenumberTips
#' @export
RenumberTips.HPart <- function(tree, tipOrder) {
  startOrder <- TipLabels(tree)
  newOrder <- MatchStrings(TipLabels(tipOrder, single = TRUE), startOrder)
  
  if (!identical(newOrder, startOrder)) {
    if (length(newOrder) != length(startOrder)) {
      stop("Tree labels ", paste0(setdiff(startOrder, tipOrder), collapse = ", "),
           " missing from `tipOrder`")
    }
    newIndices <- match(newOrder, startOrder)
    tree <- clone(tree, newOrder)
    relabel_hpart(tree, newIndices - 1L)
    # Return:
    tree
  } else {
    clone(tree)
  }
}

library("TreeTools")

test_that("is.HPart() succeeds", {
  expect_true(is.HPart(as.HPart(TreeTools::BalancedTree(7))))
  expect_true(is.HPart(structure(class = "HPart",
                                 list(list("t1"), list("t2", "t3")))))
  expect_false(is.HPart(structure(class = "NonPart", 
                                  list(list("t1"), list("t2", "t3")))))
})

test_that("as.phylo.HPart", {
  bal7 <- BalancedTree(7)
  hb7 <- as.HPart(bal7)
  expect_equal(Preorder(as.phylo.HPart(hb7)), bal7)
  
  bal17 <- BalancedTree(17)
  hb17 <- as.HPart(bal17)
  expect_equal(Preorder(as.phylo.HPart(hb17)), bal17)
})

test_that("as.HPart.numeric", {
  hpList <- as.HPart(list(list(1, 3, 9),
                          list(2, 4, 8),
                          list(5, 6, 7)))
  expect_equal(class(hpList), "HPart")
  
  hpNum <- as.HPart(c(1, 2, 1, 2, 3, 3, 3, 2, 1))
  expect_equal(class(hpNum), "HPart")
  
  expect_equal(SelfHMI(hpNum), SelfHMI(hpList))
  expect_equal(HMI(hpNum, hpList), SelfHMI(hpNum))
  
  flatP <- as.HPart(list(as.list(1:5), as.list(6:9)))
  hp9 <- as.HPart(BalancedTree(1:9))
  expect_equal(HMI(flatP, hp9), 0.99107606)
  
})

test_that("as.HPart.unimplemented", {
  expect_error(as.HPart(matrix()), "no applicable method")
  expect_error(as.HPart(list(letters, LETTERS)), "leaves must be integers")
  expect_error(as.HPart(list(list(1, 2, 3), list(0, 1, 2))),
               ".eaves must contain each integer")
  expect_error(as.HPart(list(list(1, 2, 3), list(3, 1, 2))),
               ".eaves must contain each integer")
})

test_that("HParts are relabelled correctly", {
  bal7 <- BalancedTree(7)
  hb7 <- as.HPart(bal7)
  
  map <- c(7:4, 1:3)
  mappedLabels <- paste0("t", map)
  
  hbMap <- RenumberTips(hb7, mappedLabels)
  # Here we want only to map the internal node IDs
  attr(hbMap, "tip.label") <- TipLabels(hb7)
  
  bal7tl <- bal7
  bal7tl$tip.label <- bal7$tip[map]
  
  expect_equal(SortTree(Preorder(as.phylo.HPart(hbMap))), SortTree(bal7tl))
})

test_that("plot.HPart", {
  skip_if_not_installed("vdiffr")
  vdiffr::expect_doppelganger("plot-HPart", function() 
    plot(as.HPart(list(list(1, 2, 3), list(4, list(5, 6)))))
  )
})

test_that("Renumber.HPart", {
  expect_error(RenumberTips(as.HPart(list(1, 2, 4, 3)), 4:2),
               "labels 1 missing from `tipOrder`")
})

# Redirect all graphics to a null device for the test session.
# This prevents bare plot() calls and vdiffr snapshot rendering from
# appearing in the interactive graphics device (e.g. the RStudio viewer).
# vdiffr opens its own svglite device on top of this one, so snapshot
# tests are unaffected.
pdf(NULL)
withr::defer(dev.off(), teardown_env())

#' Plot a simple tree
#'
#' Convenience plotting function used in vignettes and documentation.
#'
#' @param tr A tree of class `phylo`, with leaves labelled as integers.
#' @param title `main` title for the plot.
#' @param bold Integer specifying which leaves to print in bold.
#' @param leaveRoom Logical specifying whether to leave space to print
#' tree distances beneath the plotted tree.
#' @param edge.color Additional parameter to `plot.phylo`; will be overridden
#' by `prune` and `graft` as requested.
#' @param prune,graft Integer vectors specifying which edges to highlight as
#' pruned and grafted.
#' @param edge.width,\dots Additional parameters to `plot.phylo`.
#'
#' @template MRS
#' @importFrom ape plot.phylo
#' @keywords internal
#' @export
TreeDistPlot <- function(tr, title = NULL, bold = NULL, leaveRoom = FALSE,
                         prune = integer(0), graft = integer(0),
                         edge.color = "black", edge.width = NULL, ...) {
  
  nEdge <- dim(tr[["edge"]])[[1]]
  if (is.null(tr[["edge.length"]])) {
    tr[["edge.length"]] <- rep(1, nEdge)
  }
  if (is.null(edge.width)) {
    edge.width <- if (is.null(tr[["edge.width"]])) {
      rep(1, nEdge)
    } else {
      tr[["edge.width"]]
    }
  }
  if (length(edge.color) == 1) {
    edge.color <- rep(edge.color, nEdge)
  }
  nTip <- length(tr[["tip.label"]])
  if (all(tr[["tip.label"]] %in% LETTERS)) {
    tr[["tip.label"]] <- match(tr[["tip.label"]], LETTERS)
  } else if (all(tr[["tip.label"]] %in% letters)) {
    tr[["tip.label"]] <- match(tr[["tip.label"]], letters)
  }

  if (length(prune) > 0 || length(graft) > 0) {
    edge.width[c(prune, graft)] <- pmax(3L, edge.width[c(prune, graft)])
    edge.color[prune] <- "#D55E00" # Ternary::cbPalette8[[7]]
    edge.color[graft] <- "#009E73" # Ternary::cbPalette8[[4]]
  }
  tipCols <- c("#000000", "#004949", "#009292", "#FFB6DB", "#490092", "#B66DFF",
               "#6DB6FF", "#B6DBFF", "#920000", "#924900", "#DB6D00", "#24FF24",
               "#FFFF6D") # Ternary::cbPalette15[-c(4, 7)]

  tipNumbers <- tr[["tip.label"]]
  font <- rep(1L, length(tipNumbers))
  if (!is.null(bold)) {
    font[tipNumbers %in% bold] <- 4L
  }
  yLim <- if (leaveRoom) {
    c(-0.4 - 8, # = -0.4 - length(legendSequence)
      nTip) 
  } else {
    c(-0.4, nTip)
  }
  tipInts <- tryCatch(as.integer(tipNumbers),
                      warning = function(e) {
                        warning("Leaves of `tr` must be labelled with integers")
                      })
  
  tr[["tip.label"]] <- LETTERS[tipInts]
  plot.phylo(tr, tip.color = tipCols[tipInts],
             main = title, cex.main = 0.8, font = font, 
             edge.width = edge.width, edge.color = edge.color, 
             y.lim=yLim, ...)
}

library("TreeTools")

test_that("TreeDistPlot() warns", {
  expect_warning(
    expect_warning(
      expect_null(TreeDist::TreeDistPlot(PectinateTree(8))),
      "Leaves.*must be labelled with integers"),
    "fewer than 2 tips" # From plot.phylo: I don't understand why!
  )
})

test_that("TreeDistPlot() works", {
  tr <- PectinateTree(1:11)
  tr$edge.width <- rep(1:2, 10)
  Test1 <- function() {
    TreeDist::TreeDistPlot(tr, title = "Test", 
                         bold = c(2, 4, 6),
                         leaveRoom = TRUE,
                         prune = 1, graft = 10)
  }
  Test2 <- function() {
    TreeDist::TreeDistPlot(tr, title="Crop tightly", 
                           bold = c(2, 4, 6), prune = 11, graft = 10,
                           leaveRoom = FALSE)
  }
  skip_if_not_installed("vdiffr")
  skip_if(packageVersion("graphics") < "4.3")
  skip_if(packageVersion("vdiffr") < "1.0")
  
  vdiffr::expect_doppelganger("Test with space", Test1)
  vdiffr::expect_doppelganger("Test without space", Test2)
  tr$tip.label <- letters[1:11]
  vdiffr::expect_doppelganger("Test-lc-letters", Test2)
  tr$tip.label <- LETTERS[1:11]
  vdiffr::expect_doppelganger("Test-UC-LETTERS", Test2)
  
})

library("TreeTools", quietly = TRUE)
devtools::load_all()
set.seed(0)

message(Sys.time(), ": Starting.")

postTrees <- Postorder(as.phylo(0:200, 182))
message(Sys.time(), ": Distances.")
xx <- PathDist(postTrees)

message(Sys.time(), ": phangorn.")
xx <- phangorn::path.dist(postTrees)
message(Sys.time(), ": End.")

microbenchmark::microbenchmark(times = 1,
xx <- PathDist(postTrees),
xx <- phangorn::path.dist(postTrees)
)

#' k-means++ clustering
#'
#' k-means++ clustering \insertCite{Arthur2007}{TreeDist} improves the speed and
#' accuracy of standard \code{\link[stats]{kmeans}} clustering
#' \insertCite{Hartigan1979}{TreeDist} by preferring initial cluster centres 
#' that are far from others.
#' A scalable version of the algorithm has been proposed for larger data sets
#' \insertCite{Bahmani2012}{TreeDist}, but is not implemented here.
#' 
#' @param x Numeric matrix of data, or an object that can be coerced to such a
#' matrix (such as a numeric vector or a data frame with all numeric columns).
#' @param k Integer specifying the number of clusters, _k_.
#' @param nstart Positive integer specifying how many random sets should be
#' chosen
#' @param \dots additional arguments passed to \code{\link[stats]{kmeans}}
#' @references
#' \insertAllCited{} 
#' 
#' @template MRS
#' @seealso \code{\link[stats]{kmeans}}
#' @examples
#' # Generate random points
#' set.seed(1)
#' x <- cbind(c(rnorm(10, -5), rnorm(5, 1), rnorm(10, 6)),
#'            c(rnorm(5, 0), rnorm(15, 4), rnorm(5, 0)))
#' 
#' # Conventional k-means may perform poorly
#' klusters <- kmeans(x, cent = 5)
#' plot(x, col = klusters$cluster, pch = rep(15:19, each = 5))
#' 
#' # Here, k-means++ recovers a better clustering
#' plusters <- KMeansPP(x, k = 5)
#' plot(x, col = plusters$cluster, pch = rep(15:19, each = 5))
#' @family cluster functions
#' @importFrom stats kmeans
#' @export
KMeansPP <- function(x, k = 2, nstart = 10, ...) UseMethod("KMeansPP")

#' @export
KMeansPP.matrix <- function(x, k = 2, nstart = 10, ...) {
  if (k < 2) {
    return(kmeans(x, centers = k, ...))
  }
  
  n <- dim(x)[[1]]
  nCol <- dim(x)[[2]]
  ret <- list(tot.withinss = Inf)

  # k-means++ seeding needs, at each step, only the Euclidean distance from the
  # chosen centre to all n points: an O(n * nCol) computation per centre.
  # Computing these rows on the fly avoids materializing the full n * n distance
  # matrix (O(n ^ 2) time and memory), so this method scales to large n.
  .DistanceRow <- function(centre) {
    sqrt(.rowSums((x - rep(x[centre, ], each = n)) ^ 2, n, nCol))
  }

  for (start in seq_len(nstart)) {
    centres <- integer(k)
    centres[1L] <- sample.int(n, 1L)
    min_d <- .DistanceRow(centres[1L])

    for (i in 2:k) {
      p <- min_d ^ 2
      if (!any(p != 0)) {
        stop("Not enough distinct data points to compute clustering")
      }
      centres[i] <- sample.int(n, 1L, prob = p)
      min_d <- pmin.int(min_d, .DistanceRow(centres[i]))
    }

    proposal <- kmeans(x, centers = x[centres, ], ...)
    if (proposal[["tot.withinss"]] < ret[["tot.withinss"]]){
      ret <- proposal
    }
  }
  
  # Return:
  ret
}

#' @export
KMeansPP.numeric <- function(x, k = 2, nstart = 10, ...) {
  KMeansPP(matrix(x, ncol = 1), k = k, nstart = nstart, ...)
}

#' @export
KMeansPP.dist <- function(x, k = 2, nstart = 10, ...) {
  if (k < 2) {
    return(kmeans(x, centers = k, ...))
  }
  
  n <- attr(x, "Size")
  ret <- list(tot.withinss = Inf)
  # This method clusters the n rows of the n * n distance matrix as
  # n-dimensional vectors, so it is O(n ^ 2) by construction.  Callers that hold
  # point coordinates should use KMeansPP.matrix(), which seeds in O(n * dim).
  d <- as.matrix(x)

  for (start in seq_len(nstart)) {
    centres <- integer(k)
    centres[1L] <- sample.int(n, 1L)
    min_d <- d[centres[1L], ]

    for (i in 2:k) {
      p <- min_d ^ 2
      if (!any(p != 0)) {
        stop("Not enough distinct data points to compute clustering")
      }
      centres[i] <- sample.int(n, 1L, prob = p)
      min_d <- pmin.int(min_d, d[centres[i], ])
    }

    # Pass the already-materialized matrix d: kmeans() coerces its data argument
    # with as.matrix(), so passing the dist object x would build a second
    # n * n matrix needlessly.
    proposal <- kmeans(d, centers = d[centres, ], ...)
    if (proposal[["tot.withinss"]] < ret[["tot.withinss"]]){
      ret <- proposal
    }
  }
  
  # Return:
  ret
}

test_that("k-means++ fails", {
  indistinct <- dist(rep(1, 100))
  expect_error(KMeansPP(indistinct), "Not enough distinct data points")
  expect_error(
    KMeansPP(as.matrix(indistinct)),
    "Not enough distinct data points"
  )
})

test_that("k-means++ works", {
  set.seed(0)
  two <- c(1:5, 11:15)
  expect_equal(KMeansPP(two, k = 1)$cluster, rep(1, 10))
  expect_equal(KMeansPP(two)$cluster, rep(1:2, each = 5))
  expect_equal(KMeansPP(cbind(numeric(10), two))$cluster,
                        rep(2:1, each = 5))
  five <- cbind(c(rnorm(10, -8), rnorm(5, 0), rnorm(10, 8)),
                c(rnorm(5, -5), rnorm(15, 5), rnorm(5, 0)))
  dists <- dist(five)
  expect_equal(unname(KMeansPP(dists, k = 1)$cluster), rep(1, 25))
  cl <- KMeansPP(dists, k = 5)$cluster
  # plot(five, col = KMeansPP(dists, k = 5)$cluster, pch = rep(15:19, each = 5))
  # plot(five, col = kmeans(dists, cent = 5)$cluster, pch = rep(15:19, each = 5))
  expect_equal(range(cl), c(1, 5))
  expect_equal(length(unique(cl[1:5])), 1)
  expect_equal(length(unique(cl[6:10])), 1)
  expect_equal(length(unique(cl[11:15])), 1)
  expect_equal(length(unique(cl[16:20])), 1)
  expect_equal(length(unique(cl[21:25])), 1)
})

test_that("k-means++ matrix seeding matches materialized distance matrix", {
  # On-the-fly distance rows must select the same centres as the original
  # implementation, which materialized the full n * n distance matrix.  The
  # oracle below is the literal pre-refactor body of KMeansPP.matrix(); both run
  # under the same seed, so the comparison checks that the refactor preserved
  # the RNG draw sequence and chose identical centres (hence cluster and score).
  oracle <- function(x, k, nstart) {
    n <- dim(x)[[1]]
    ret <- list(tot.withinss = Inf)
    d <- as.matrix(dist(x))
    for (start in seq_len(nstart)) {
      centres <- integer(k)
      centres[1L] <- sample.int(n, 1L)
      min_d <- d[centres[1L], ]
      for (i in 2:k) {
        centres[i] <- sample.int(n, 1L, prob = min_d ^ 2)
        min_d <- pmin.int(min_d, d[centres[i], ])
      }
      proposal <- kmeans(x, centers = x[centres, ], iter.max = 100L)
      if (proposal[["tot.withinss"]] < ret[["tot.withinss"]]) {
        ret <- proposal
      }
    }
    ret
  }
  set.seed(1)
  x <- matrix(rnorm(500 * 5), ncol = 5)
  set.seed(42)
  expected <- oracle(x, k = 8, nstart = 5)
  set.seed(42)
  got <- KMeansPP(x, k = 8, nstart = 5, iter.max = 100L)
  expect_identical(got[["cluster"]], expected[["cluster"]])
  expect_identical(got[["tot.withinss"]], expected[["tot.withinss"]])
})

#' Collapse areas of agreement between two trees
#' 
#' `ReduceTrees()` reduces trees according to the tree reduction rules of
#' \insertCite{Allen2001;textual}{TreeDist}:
#' - Collapse identical pendant subtrees;
#' - Compress equivalent internal chains.
#' 
#' @template tree12Params
#' @param check Logical specifying whether to validate input. Specify
#' `FALSE` and you will encounter undefined behaviour if trees are not
#' binary `phylo` objects with identical leaf labels, rooted on leaf 1.
#' 
#' @return `ReduceTrees()` returns a list of two trees, corresponding to 
#' `tree1` and `tree2` after any identical groupings have been collapsed,
#' with tree edges listed in postorder; or `NULL` if the trees are equivalent.
#' @examples
#' tree1 <- TreeTools::BalancedTree(9)
#' tree2 <- TreeTools::PectinateTree(9)
#' 
#' # Set graphical parameters
#' oPar <- par(mai = rep(0.1, 4), mfrow = c(2, 2))
#' 
#' plot(tree1)
#' plot(tree2)
#' 
#' # Reduce trees by collapsing identical clades
#' confl <- ReduceTrees(tree1, tree2)
#' 
#' plot(confl[[1]])
#' plot(confl[[2]])
#' 
#' # Restore graphical parameters
#' par(oPar)
#' @template MRS
#' @importFrom TreeTools NTip PostorderOrder RenumberTips RootTree
#' @export
ReduceTrees <- function(tree1, tree2, check = TRUE) {
  if (check) {
    if (!inherits(tree1, 'phylo')) {
      stop("`tree1` must be a `phylo` object.")
    }
    if (!inherits(tree2, 'phylo')) {
      stop("`tree2` must be a `phylo` object.")
    }
    if (NTip(tree1) != NTip(tree2)) {
      stop("Trees must bear same leaf labels")
    }
    
    tree2 <- RenumberTips(tree2, tree1)
    tree1 <- RootTree(tree1, 1)
    tree2 <- RootTree(tree2, 1)
    if (tree1[["Nnode"]] != tree2[["Nnode"]]
        || tree1[["Nnode"]] != NTip(tree1) - 1L) {
      stop("Trees must be binary")
    }
  }
  ret <- reduce_trees(tree1$edge[PostorderOrder(tree1), ],
                      tree2$edge[PostorderOrder(tree2), ],
                      tree1[["tip.label"]])
  
  # Return:
  if (is.null(ret[[1]])) {
    NULL
  } else {
    ret
  }
}

#' Approximate the Subtree Prune and Regraft distance
#'
#' `SPRDist()` approximates the \acronym{SPR} distance between
#' trees.
#' 
#' The function currently defaults to the heuristic method of
#' \insertCite{deOliveira2008;textual}{TreeDist}, which purports to provide an
#' upper bound on the \acronym{SPR} distance (though exceptions exist).
#' Other approximations 
#' \insertCite{@e.g. @Hickey2008, @Goloboff2008SPR, @Whidden2018}{TreeDist} are
#' not yet implemented.
#'
#' @template tree12ListParams
#' @param method Character specifying which method to use to approximate the
#' SPR distance.  Currently defaults to `"deOliveira"``.
#' `"Rogue"` implements an experimental method whose details are pending
#' publication; this function is under development, and may be modified or
#' removed without notice.  Once formally validated, it is anticipated that this
#' method will become the default.
#' @param symmetric Deprecated (redundant after fix of
#' [phangorn#97](https://github.com/KlausVigo/phangorn/issues/97)).
#'
#' @return `SPRDist()` returns a vector or distance matrix of distances
#' between trees.
#'
#' @references \insertAllCited{}
#'
#' @examples
#' library("TreeTools", quietly = TRUE)
#'
#' # Compare single pair of trees
#' SPRDist(BalancedTree(7), PectinateTree(7))
#'
#' # Compare all pairs of trees
#' SPRDist(as.phylo(30:33, 8))
#'
#' # Compare each tree in one list with each tree in another
#' SPRDist(BalancedTree(7), as.phylo(0:2, 7))
#' SPRDist(as.phylo(0:2, 7), PectinateTree(7))
#'
#' SPRDist(list(bal = BalancedTree(7), pec = PectinateTree(7)),
#'         as.phylo(0:2, 7))
#' @template MRS
#'
#' @seealso Exact calculation with [\pkg{TBRDist}](
#' https://ms609.github.io/TBRDist/reference/TreeRearrangementDistances.html)
#' functions `USPRDist()` and `ReplugDist()`.
#'
#' \pkg{phangorn} function \code{\link[phangorn:treedist]{SPR.dist()}} employs
#' the \insertCite{deOliveira2008;textual}{TreeDist} algorithm but can crash
#' when sent trees of certain formats, and tends to have a longer running time.
#'
#' @family tree distances
#' @importFrom TreeTools PairwiseDistances Postorder
#' @export
SPRDist <- function(tree1, tree2 = NULL, method = "deOliveira", symmetric) {
  UseMethod("SPRDist")
}

#' @rdname SPRDist
#' @export
SPRDist.phylo <- function(tree1, tree2 = NULL, method = "deOliveira", symmetric) {
  if (is.null(tree2)) {
    NULL
  } else if (inherits(tree2, "phylo")) {
    .SPRFunc(method)(tree1, tree2)
  } else {
    vapply(tree2, .SPRFunc(method), double(1), tree1)
  }
}

.SPRFunc <- function(method) {
  switch(pmatch(tolower(gsub("\\s", "", method)),
                c("deoliveira2008", "rogue")),
         .SPRPairDeO, .SPRRogue)
}

#' @rdname SPRDist
#' @export
SPRDist.list <- function(tree1, tree2 = NULL, method = "deOliveira", symmetric) {
  if (is.null(tree2)) {
    PairwiseDistances(
      RootTree(RenumberTips(tree1, tree1), 1),
      .SPRFunc(method),
      check = FALSE
    )
  } else if (inherits(tree2, 'phylo')) {
    vapply(tree1, .SPRFunc(method), double(1), tree2)
  } else {
    vapply(tree2, SPRDist, double(length(tree1)), tree1, method = method)
  }
}

#' @rdname SPRDist
#' @export
SPRDist.multiPhylo <- SPRDist.list

# if x <- which(condition, arr.ind = FALSE), and
# y <-  which(condition, arr.ind = TRUE), then
# .Which1 and .Which2 give y[1] and y[2].
.Which1 <- function (x, nSplits) {
  ret <- x %% nSplits
  if (ret == 0L) {
    nSplits
  } else {
    ret
  }
}
.Which2 <- function (x, nSplits) (x - 1) %/% nSplits + 1L

.SPRExact5 <- function(sp1, sp2) {
  # Trees have shape (r, (p1, p2), (q1, q2)) after reduction.
  # There are two cases:
  if (all(xor(sp1[[1]] , sp1[[2]]) == xor(sp2[[1]] , sp2[[2]]))) {
    # Case 1: `r` has the same label in each tree
    # As the tree cannot be reduced by the reduction rule, we have
    # (X, (A, B), (C, D)) vs (X, (A, C), (B, C)) = 2 moves
    #
    2
  } else {
    # Case 2: `r` has a different label in each tree.
    # Assign `r` the label X in tree 1 and Y in tree 2.
    # (X, (Y, ?), (?, ?)) vs (Y, (X, ?), (?, ?))
    #
    # Then label the sister to X in tree2 X', and Y mutas mutantis
    # Notice that if X' == Y', the unlabelled cherry reduces by reduction rule
    # Hence we have
    # (X, ((Y, Y'), (X', ?))) vs (Y, ((X, X'), (Y', ?))) = 1 moves
    1
  }
}

.SPRExact6 <- function(sp1, sp2) {
  # Surprisingly, there is only one configuration with a distance of 1:
  # (((Lb, Lc), La), (Ra, (Rb, Rc))) vs (((Ra, Rb), La), (Lb, (Lc, Rc)))
  pairs1 <- TipsInSplits(sp1, smallest = TRUE) == 2
  pairs2 <- TipsInSplits(sp2, smallest = TRUE) == 2
  if (all(pairs1) || all(pairs2)) {
    return (2)
  }
  duo1 <- sp1[[pairs1]]
  trio1 <- sp1[[!pairs1]]
  
  .Overlapper <- function(s1, s2) {
    res <- as.logical(xor(s1, s2))
    if (sum(res) == 1) res else !res
  }
  middle1a <- .Overlapper(duo1[[1]], trio1)
  middle1b <- .Overlapper(duo1[[2]], trio1)
  
  duo2 <- sp2[[pairs2]]
  trio2 <- sp2[[!pairs2]]
  middle2a <- .Overlapper(duo2[[1]], trio2)
  middle2b <- .Overlapper(duo2[[2]], trio2)
  
  inMiddleEachTime <- (middle1a | middle1b) & (middle2a | middle2b)
  if (sum(inMiddleEachTime) == 1) {
    La <- if (inMiddleEachTime[middle1a]) middle1a else middle1b
    Lbc1 <- as.logical(duo1[[if (inMiddleEachTime[middle1a]) 1 else 2]])
    if (Lbc1[La]) {
      Lbc1 <- !Lbc1
    }
    if (any(Lbc1[middle2a | middle2b])) {
      # Lb is in the other middle position in tree 2:
      # (((?, ?), La), (Lb, (?, ?)))
      Lbc2 <- as.logical(duo2[[if (La[middle2a]) 1 else 2]])
      if (Lbc2[La]) {
        Lbc2 <- !Lbc2
      }
      if (!any(Lbc2[Lbc1])) {
        #   (((?, ?), La), (Lb, (Lc, ?)))
        #     (((Ra, Rb), La), (Lb, (Lc, Rc))) = 1 !!!
        return(1)
      }
      #   (((Lc, ?), La), (Lb, (?, ?)))
      #     (((Lc, Rc), La), (Lb, (Ra, Rb))) = 2
    }
    # (((?, ?), La), (Rb, (?, ?)))
    #   (((?, Ra), La), (Rb, (?, ?)))
    #     (((Lb, Ra), La), (Rb, (Rc, Lc))) = 2
    #   (((?, ?), La), (Rb, (Ra, ?)))
    #     (((Rc, Lc), La), (Rb, (Ra, Lb))) = 2
    # 
  }
  
  # All other tree pairs have a distance of 2 - see below
  return(2)
  
  # Trees may be one of two shapes: 
  #   ((a1, a2), (b1, b2), (c1, c2))
  #   (((Lb, Lc), La), (Ra, (Rb, Rc)))
  # balanced1 <- all(pairs1)
  # balanced2 <- all(pairs2)
  # if (balanced1 && balanced2) {
    # There's only one possible configuration:
    # ((ab, ac), (ba, bc), (ca, cb)) vs ((ba, ca), (ab, cb), (ac, bc)) = 2
  # }
  # if (!balanced1 && !balanced2) {
    # Both trees have the shape
    # (((Lb, Lc), La), (Ra, (Rb, Rc)))
    # We will use the same labels for Tree 2, matching where possible.
    # if (La1 == La2 && Ra1 == Ra2) {
      # La = La, Ra = Ra:
      # (((Lb, Lc), La), (Ra, (Rb, Rc))), (((Lb, Rb), La), (Ra, (Rc, Lc))) = 2
    # }
    # As we can't match La and Ra, we'll match La if we can.
    # if (La1 != La2 && Ra1 != Ra2) {
      # LO != La, Ra != Ra
      # La and Ra are both in the cherries
      # (((?, La), Lb), (Lc, (?, ?)))
      #   (((Rb, La), Lb), (Lc, (Ra, Rc))) = 2
      # 
      # (((?, ?), Rb), (Rc, (?, ?)))
      #   (((Lc, Ra), Rb), (Rc, (La, Lb))) = 2
      #   
      # (((?, ?), Lb), (Rb, (?, ?)))
      #   (((?, La), Lb), (Rb, (?, ?)))
      #     (((Lc, La), Lb), (Rb, (Ra, Rc))) = 2
      #     (((Ra, La), Lb), (Rb, (Lc, Rc))) = 2
      #     (((Rc, La), Lb), (Rb, (Ra, Lc))) = 2
      #     
      #   (((?, ?), Lb), (Rb, (La, ?)))
      #     (((Ra, Rc), Lb), (Rb, (La, Lc))) = 2
      #     (((Lc, Rc), Lb), (Rb, (La, Ra))) = 2
      #     (((Ra, Lc), Lb), (Rb, (La, Rc))) = 2
      #   
      #   (((?, ?), Lb), (Rb, (?, ?)))
      #     (((Lc, Rc), Lb), (Rb, (La, Ra))) = 2
      #     (((Lc, Ra), Lb), (Rb, (La, Rc))) = 2
      #     (((Ra, Rc), Lb), (Rb, (La, Lc))) = 2
      #     (((Ra, La), Lb), (Rb, (Rc, Lc))) = 2
      #     (((Rc, La), Lb), (Rb, (Ra, Lc))) = 2
      #     (((Lc, La), Lb), (Rb, (Ra, Rc))) = 2
    # }
    # Else exactly one of the bridging leaves is the same; call this La.
    # 
    # (((?, ?), La), (Rb, (?, ?)))
    #   (((?, Ra), La), (Rb, (?, ?)))
    #     (((Lb, Ra), La), (Rb, (Rc, Lc))) = 2
    #   (((?, ?), La), (Rb, (Ra, ?)))
    #     (((Rc, Lc), La), (Rb, (Ra, Lb))) = 2
    # 
    # (((?, ?), La), (Lb, (?, ?)))
    #   (((Lc, ?), La), (Lb, (?, ?)))
    #     (((Lc, Rc), La), (Lb, (Ra, Rb))) = 2
    #   (((?, ?), La), (Lb, (Lc, ?)))
    #     (((Ra, Rb), La), (Lb, (Lc, Rc))) = 1 !!!
    # 
    # 
  # }
}

.SPRExact7 <- function(sp1, sp2) {
  if (length(sp1) != 4L || length(sp2) != 4L) {
    stop("Expected a length-4 raw vector of splits") # nocov
  }
  spr_table_7(sp1, sp2)
}

# Takes a 'Rogue' approach: finds the leaf that introduces the most conflict,
# and nixes it.
#' @importFrom TreeTools FirstMatchingSplit
.SPRRogue <- function(tree1, tree2, check = TRUE) {
  moves <- 0
  
  ProxyDistance <- switch(
    pmatch(toupper(getOption("sprProxy", "C")), c("C", "P", "Q", "R")),
    ClusteringInfoDist,
    PhylogeneticInfoDistance,
    function(x, y) Quartet::QuartetDivergence(Quartet::QuartetStatus(x, y)),
    RobinsonFoulds
  )
  
  reduced <- ReduceTrees(tree1, tree2, check = check)
  
  while (!is.null(reduced)) {
    
    tr1 <- reduced[[1]]
    tr2 <- reduced[[2]]
    nTip <- NTip(tr1)
    if (nTip == 4 && getOption("sprShortcut", Inf) >= 4) {
      return(moves + 1)
    }
    
    sp1 <- as.Splits(tr1)
    sp2 <- as.Splits(tr2, tr1)
    if (nTip == 5 && getOption("sprShortcut", Inf) >= 5) {
      return(moves + .SPRExact5(sp1, sp2))
    }
    if (nTip == 6 && getOption("sprShortcut", Inf) >= 6) {
      return(moves + .SPRExact6(sp1, sp2))
    }
    if (nTip == 7 && getOption("sprShortcut", Inf) >= 7) {
      return(moves + .SPRExact7(sp1, sp2))
    }
    
    firstMatchedSplit <- FirstMatchingSplit(sp1, sp2)
    if (!isFALSE(getOption("sprMatches")) && firstMatchedSplit > 0) {
      # At least one split exists in both trees
      subtips1 <- as.logical(sp1[[firstMatchedSplit]])
      subtips2 <- !subtips1
      
      # Add dummy tip as placeholder for other half of tree
      subtips1[!subtips1][[1]] <- TRUE

      # Repeat for other half-tree
      subtips2[!subtips2][[1]] <- TRUE

      moves1 <- .SPRRogue(
        KeepTipPostorder(tr1, subtips1),
        KeepTipPostorder(tr2, subtips1)
      )
      moves2 <- .SPRRogue(
        KeepTipPostorder(tr1, subtips2),
        KeepTipPostorder(tr2, subtips2)
      )
      return(moves + moves1 + moves2)
    }
    
    labels <- TipLabels(tr1)
    scores <- numeric(length(labels))
    blank <- rep_len(TRUE, length(labels))

    depth <- max(getOption("sprDepth", 1), 1)
    
    
    .ScoreWithout <- function(idx) {
      keep <- blank
      keep[idx] <- FALSE
      
      outcome <- keep_and_reduce(tr1, tr2, keep)
      if (is.null(outcome[[1]])) {
        return(-Inf)
      }
      
      oSpl <- as.Splits(outcome)
      firstMatch <- FirstMatchingSplit(oSpl[[1]], oSpl[[2]])
      
      if (firstMatch > 0) {
        subtips1 <- as.logical(oSpl[[1]][[firstMatch]])
        subtips2 <- !subtips1
        # Anchor to shared edge
        subtips1[!subtips1][[1]] <- TRUE
        subtips2[!subtips2][[1]] <- TRUE
        
        sub1 <- ReduceTrees(KeepTipPostorder(outcome[[1]], subtips1),
                            KeepTipPostorder(outcome[[2]], subtips1))
        
        sub2 <- ReduceTrees(KeepTipPostorder(outcome[[1]], subtips2),
                            KeepTipPostorder(outcome[[2]], subtips2))
        
        # Return:
        ProxyDistance(sub1[[1]], sub1[[2]]) +
          ProxyDistance(sub2[[1]], sub2[[2]])
        
      } else {
        # Return:
        ProxyDistance(outcome[[1]], outcome[[2]])
      }
    }
    for (i in seq_along(labels)) {
      scores[[i]] <- .ScoreWithout(i)
      if (!is.finite(scores[[i]])) break
    }
    if (any(!is.finite(scores))) {
      depth <- 1
    }
    if (depth > 1) {
      pairs <- combn(seq_along(labels), 2)
      nPairs <- dim(pairs)[[2]]
      pairScores <- double(nPairs)
      for (i in seq_len(nPairs)) {
        pairScores[[i]] <- .ScoreWithout(pairs[, i])
        if (!is.finite(pairScores[[i]])) break
      }
    }
    
    drop <- logical(length(labels))
    couldDrop <- scores == min(scores)

    if (depth > 1 && min(pairScores) < min(scores)) {
      pairDrop <- pairScores == min(pairScores)
      dropTions <- pairs[, pairDrop]
      if (any(couldDrop[dropTions]) && !any(!is.finite(pairScores))) {
        # Dropping two at once doesn't give us any benefit over dropping one
        # at a time – but will mean we can't spot a handy pair next time.
        drop[dropTions[which.min(scores[dropTions])]] <- TRUE
      } else {
        # If dropping two gives us a better solution, drop both at once -
        # failing to do so can cause an optimal path to be missed.
        drop[pairs[, which.max(pairDrop)]] <- TRUE
      }
    } else {
      drop[[which.min(scores)]] <- TRUE
    }
    
    reduced <- keep_and_reduce(tr1, tr2, !drop)
    if (length(reduced) == 1L) {
      reduced <- NULL
    }
    
    moves <- sum(moves, drop)
  }
  
  # Return:
  moves
}

# An attempt to reproduce the phangorn results using the algorithm of 
# \insertCite{deOliveira2008;textual}{TreeDist}
# An exact match is unlikely due to the arbitrary breaking of ties when leaves
# are selected for removal.
#' SPR approximation via Oliveira Martins _et al._ (2008)
#' @examples
#' # de Oliveira Martins et al 2008, fig. 7
#' tree1 <- ape::read.tree(text = "((1, 2), ((a, b), (c, d)), (3, (4, (5, (6, 7)))));")
#' tree2 <- ape::read.tree(text = "((1, 2), 3, (4, (5, (((a, b), (c, d)), (6, 7)))));")
#' oPar <- par(mfrow =c(2, 1), mar = rep(0, 4))
#' plot(tree1)
#' plot(tree2)
#' par(oPar)
#' SPRDist(tree1, tree2, method = "deO")
#' @keywords internal
#' @importFrom TreeTools DropTip TipsInSplits KeepTipPostorder
#' @importFrom TreeTools edge_to_splits
.SPRPairDeO <- function(tree1, tree2, check = TRUE) {
  if (check) {
    if (length(tree1[["tip.label"]]) != length(tree2[["tip.label"]])) {
      stop("`tree1` and `tree2` must have the same number of tips.")
    }
  }
  moves <- 0
  
  # Reduce trees (Fig. 7A in deOliveira2008)
  reduced <- ReduceTrees(tree1, tree2, check = check)
    
  while (!is.null(reduced)) {
    tr1 <- reduced[[1]]
    tr2 <- reduced[[2]]
    edge1 <- tr1[["edge"]]
    edge2 <- tr2[["edge"]]
    labels <- tr1[["tip.label"]]
    nTip <- length(labels)
    sp1 <- edge_to_splits(edge1, PostorderOrder(edge1), labels, nTip = nTip)
    sp2 <- edge_to_splits(edge2, PostorderOrder(edge2), labels, nTip = nTip)
    matched <- cpp_robinson_foulds_distance(sp1, sp2, nTip)
    nMatched <- matched[["score"]]
    if (nMatched != length(sp1) * 2) {
      unmatchedSplits <- is.na(matched[["matching"]])
      sp1 <- sp1[[unmatchedSplits]]
      sp2 <- sp2[[-matched$matching[!unmatchedSplits]]]
    }
    
    nSplits <- length(sp1)
    # Compute size of disagreement splits - see Fig. 7C in @deOliv2008
    mmSize <- mismatch_size(sp1, sp2)
    stopifnot(all(mmSize > 0))
    
    # Arbitrary selection of leaves to remove introduces a stochastic element
    minMismatch <- which.min(mmSize)
    
    split1 <- structure(sp1[.Which1(minMismatch, nSplits), , drop = FALSE],
                        nTip = nTip, class = "Splits")
    split2 <- structure(sp2[.Which2(minMismatch, nSplits), , drop = FALSE],
                        nTip = nTip, class = "Splits")
    disagreementSplit <- xor(split1, split2)
    
    keep <- as.logical(disagreementSplit)
    nKeep <- sum(keep)
    if (nKeep < length(keep) / 2) {
      keep <- !keep
      nKeep <- length(keep) - nKeep
    }
    
    reduced <- if (nKeep < 4L) {
      NULL
    } else {
      keep_and_reduce(tr1, tr2, keep)
    }
    
    if (length(reduced) == 1L) {
      reduced <- NULL
    }
    
    moves <- moves + 1 # Usually an underestimate, unless we've lost a chance
                       # to "untangle a knot"
  }
  
  # Return:
  moves
}

# R-level validation wrappers for internal C++ functions.
# C++ uses ASSERT() (compiled out in release); these ensure user-visible errors.
mismatch_size <- function(x, y) {
  if (!inherits(x, "Splits") || is.null(attr(x, "nTip"))) {
    stop("`x` lacks nTip attribute")
  }
  if (!inherits(y, "Splits") || is.null(attr(y, "nTip"))) {
    stop("`y` lacks nTip attribute")
  }
  if (attr(x, "nTip") != attr(y, "nTip")) {
    stop("`x` and `y` differ in `nTip`")
  }
  if (nrow(x) != nrow(y)) {
    stop("`x` and `y` differ in number of splits.")
  }
  .Call(`_TreeDist_mismatch_size`, x, y)
}

confusion <- function(x, y) {
  if (!inherits(x, "Splits") || is.null(attr(x, "nTip"))) {
    stop("`x` lacks nTip attribute")
  }
  if (!inherits(y, "Splits") || is.null(attr(y, "nTip"))) {
    stop("`y` lacks nTip attribute")
  }
  if (attr(x, "nTip") != attr(y, "nTip")) {
    stop("`x` and `y` differ in `nTip`")
  }
  if (nrow(x) != nrow(y)) {
    stop("Input splits must contain same number of splits.")
  }
  .Call(`_TreeDist_confusion`, x, y)
}

#' Solve linear assignment problem using LAPJV
#'
#' Use the algorithm of \insertCite{Jonker1987;textual}{TreeDist} to solve the 
#' Linear Sum Assignment Problem (LSAP).
#' 
#' The Linear Assignment Problem seeks to match each row of a matrix with a 
#' column, such that the cost of the matching is minimized.
#' 
#' The Jonker & Volgenant approach is a faster alternative to the Hungarian
#' algorithm \insertCite{Munkres1957}{TreeDist}, which is implemented in 
#' `clue::solve_LSAP()`.
#' 
#' Note: the JV algorithm expects integers. In order to apply the function
#' to a non-integer _n_, as in the tree distance calculations in this package,
#' each _n_ is multiplied by the largest available integer before applying
#' the JV algorithm.  If two values of _n_ exhibit a trivial difference -- 
#' e.g. due to floating point errors -- then this can lead to interminable
#' run times.  (If numbers of the magnitude of billions differ only in their
#' last significant digit, then the JV algorithm may undergo billions of 
#' iterations.)  To avoid this, integers over 2^22 that differ by a value of
#' 8 or less are treated as equal.
#' 
#' @references \insertAllCited{}
#'
#' @author [C++ code](
#' https://github.com/yongyanghz/LAPJV-algorithm-c/blob/master/src/lap.cpp)
#' by Roy Jonker, MagicLogic Optimization Inc. <roy_jonker@magiclogic.com>, 
#' with contributions from Yong Yang <yongyanglink@gmail.com>, after 
#' [Yi Cao](https://uk.mathworks.com/matlabcentral/profile/authors/69713)
#' 
#' @param x Matrix of costs.
#' @return `LAPJV()` returns a list with two entries: `score`, the score of the
#' optimal matching;
#' and `matching`, the columns matched to each row of the matrix in turn.
#' 
#' @examples 
#' problem <- matrix(c(7, 9, 8, 9, 9,
#'                     2, 8, 5, 7, 9,
#'                     1, 6, 6, 9, 9,
#'                     3, 6, 2, 2, 9), 4, 5, byrow = TRUE)
#'
#' LAPJV(problem)
#' @seealso 
#' Implementations of the Hungarian algorithm exist in \pkg{adagio},
#' \pkg{RcppHungarian}, and \pkg{clue} and \pkg{lpSolve}; for larger matrices,
#' these are substantially slower. (See discussion at [Stack Overflow](
#' https://stackoverflow.com/questions/72806265/).)
#' 
#' The JV algorithm is implemented for square matrices in the Bioconductor
#' package [`GraphAlignment::LinearAssignment()`](
#' https://www.bioconductor.org/packages/release/bioc/html/GraphAlignment.html).
#' 
#' @export
LAPJV <- function(x) {
  dims <- dim(x)
  if (length(dims) == 2L) {
    if (any(dims == 0L)) {
      list(score = 0, matching = integer(0))
    } else {
      lapjv(x, max(x))
    }
  } else {
    stop("x must be a two-dimensional matrix")
  }
}

#' Matching Split Distance
#' 
#' Calculate the 
#' [Matching Split Distance](https://ms609.github.io/TreeDist/articles/Generalized-RF.html#matching-split-distance)
#' \insertCite{Bogdanowicz2012,Lin2012}{TreeDist} for unrooted binary trees.
#' 
#' Trees need not contain identical leaves; scores are based on the leaves that
#' trees hold in common.  Check for unexpected differences in tip labelling
#' with `setdiff(TipLabels(tree1), TipLabels(tree2))`.
#' 
#' @inheritParams TreeDistance
#' 
#' @templateVar returns `MatchingSplitDistance()` returns
#' @template distReturn
#' 
#' @section Normalization:
#' 
#' A normalization value or function must be provided in order to return a
#' normalized value.  If you are aware of a generalised formula, please
#' let me know by
#' \href{https://github.com/ms609/TreeDist/issues/new}{creating a GitHub issue}
#' so that it can be implemented.
#' 
#' @examples 
#' MatchingSplitDistance(lapply(rep(8, 5), ape::rtree), normalize = 16)
#' 
#' MatchingSplitDistance(TreeTools::BalancedTree(6),
#'                       TreeTools::PectinateTree(6),
#'                       reportMatching = TRUE)
#' 
#' VisualizeMatching(MatchingSplitDistance, TreeTools::BalancedTree(6),
#'                   TreeTools::PectinateTree(6))
#' @template MRS
#'  
#' @references \insertAllCited{}
#' 
#' @family tree distances
#' 
#' @export
MatchingSplitDistance <- function(tree1, tree2 = NULL, normalize = FALSE,
                                   reportMatching = FALSE) {
  unnormalized <- CalculateTreeDistance(MatchingSplitDistanceSplits, tree1,
                                        tree2, reportMatching)
  
  # Return:
  NormalizeInfo(unnormalized, tree1, tree2, how = normalize,
                InfoInTree = function(X) stop(
                  "Please specify a function to generate a normalizing constant"
                  ),
                Combine = max)
}

#' @rdname MatchingSplitDistance
#' @inheritParams SharedPhylogeneticInfoSplits
#' @useDynLib TreeDist, .registration = TRUE
#' @export
MatchingSplitDistanceSplits <- function(splits1, splits2, 
                                         nTip = attr(splits1, "nTip"),
                                         normalize = TRUE, 
                                         reportMatching = FALSE) {
  GeneralizedRF(splits1, splits2, nTip, cpp_matching_split_distance,
                maximize = FALSE, reportMatching = reportMatching)
}

#' Information-based generalized Robinson&ndash;Foulds distances
#'
#' Calculate tree similarity and distance measures based on the amount of 
#' phylogenetic or clustering information that two trees hold in common, as
#' proposed in Smith (2020).
#' 
#' 
#' [Generalized Robinson&ndash;Foulds distances](https://ms609.github.io/TreeDist/articles/Robinson-Foulds.html#generalized-robinson-foulds-distances)
#' calculate tree similarity by finding an
#' optimal matching that the similarity between a split on one tree
#' and its pair on a second, considering all possible ways to pair splits 
#' between trees (including leaving a split unpaired).
#' 
#' The methods implemented here use the concepts of 
#' [entropy and information](https://ms609.github.io/TreeDist/articles/information.html)
#' \insertCite{Mackay2003}{TreeDist} to assign a similarity score between each
#' pair of splits.
#' 
#' The returned tree similarity measures state the amount of information, 
#' in bits, that the splits in two trees hold in common 
#' when they are optimally matched, following 
#' \insertCite{SmithDist;textual}{TreeDist}.
#' The complementary tree distance measures state how much information is 
#' different in the splits of two trees, under an optimal matching.
#' Where trees contain different tips, tips present in one tree but not the
#' other are removed before each comparison (as by definition, the trees neither
#' hold information in common nor differ regarding these tips).
#' 
#' # Concepts of information
#'
#' The phylogenetic (Shannon) information content and entropy of a split are
#' defined in 
#' [a separate vignette](https://ms609.github.io/TreeDist/articles/information.html).
#' 
#' Using the mutual (clustering) information
#' \insertCite{Meila2007,Vinh2010}{TreeDist} of two splits to quantify their
#' similarity gives rise to the Mutual Clustering Information measure
#' (`MutualClusteringInfo()`, `MutualClusteringInfoSplits()`);
#' the entropy distance gives the Clustering Information Distance
#' (`ClusteringInfoDistance()`).
#' This approach is optimal in many regards, and is implemented with 
#' normalization in the convenience function `TreeDistance()`.
#' 
#' Using the amount of phylogenetic information common to two splits to measure
#' their similarity gives rise to the Shared Phylogenetic Information similarity
#' measure (`SharedPhylogeneticInfo()`, `SharedPhylogeneticInfoSplits()`).
#' The amount of information distinct to
#' each of a pair of splits provides the complementary Different Phylogenetic
#' Information distance metric (`DifferentPhylogeneticInfo()`).
#' 
#' The Matching Split Information measure (`MatchingSplitInfo()`,
#' `MatchingSplitInfoSplits()`) defines the similarity between a pair of 
#' splits as the phylogenetic information content of the most informative 
#' split that is consistent with both input splits; `MatchingSplitInfoDistance()`
#' is the corresponding measure of tree difference.
#' ([More information here](
#' https://ms609.github.io/TreeDist/articles/Generalized-RF.html).)
#' 
#' # Conversion to distances
#' 
#' To convert similarity measures to distances, it is necessary to 
#' subtract the similarity score from a maximum value.  In order to generate
#' distance _metrics_, these functions subtract the similarity twice from the 
#' total information content (SPI, MSI) or entropy (MCI) of all the splits in 
#' both trees \insertCite{SmithDist}{TreeDist}.
#' 
#' # Normalization
#' 
#' If `normalize = TRUE`, then results will be rescaled such that distance
#' ranges from zero to (in principle) one.
#' The maximum **distance** is the sum of the information content or entropy of
#' each split in each tree; the maximum **similarity** is half this value.
#' (See Vinh _et al._ (2010, table 3) and 
#' \insertCite{SmithDist;textual}{TreeDist} for
#' alternative normalization possibilities.)
#' 
#' Note that a distance value of one (= similarity of zero) will seldom be
#' achieved, as even the most different trees exhibit some similarity.
#' It may thus be helpful to rescale the normalized value such that the
#' _expected_ distance between a random pair of trees equals one.  This can
#' be calculated with `ExpectedVariation()`; or see package
#' '[TreeDistData](
#' https://ms609.github.io/TreeDistData/reference/randomTreeDistances.html)'
#' for a compilation of expected values under different metrics for trees with
#' up to 200 leaves.
#' 
#' Alternatively, use `normalize = `[`pmax`] or [`pmin`] to scale against the
#' information content or entropy of all splits in the most (`pmax`) or
#' least (`pmin`) informative tree in each pair.
#' To calculate the relative similarity against a reference tree that is known
#' to be "correct", use `normalize = SplitwiseInfo(trueTree)` (SPI, MSI) or
#' `ClusteringEntropy(trueTree)` (MCI).
#' For worked examples, see the internal function [`NormalizeInfo()`], which is
#' called from distance functions with the parameter `how = normalize`.
#' .
#' 
#'
#' # Distances between large trees
#' 
#' To balance memory demands and runtime with flexibility, these functions are
#' implemented for trees with up to 2048 leaves.
#' To analyse trees with up to 8192 leaves, you will need to a modified version
#' of the package:
#' `install.packages("BigTreeDist", repos = "https://ms609.github.io/packages/")`
#' Use `library("BigTreeDist")` *instead* of `library("TreeDist")` to load
#' the modified package &ndash; or prefix functions with the package name, e.g.
#' `BigTreeDist::TreeDistance()`.
#' 
#' As an alternative download method,
#' uninstall \pkg{TreeDist} and \pkg{TreeTools} using
#' `remove.packages()`, then use
#'  `devtools::install_github("ms609/TreeTools", ref = "more-leaves")`
#' to install the modified \pkg{TreeTools} package; then, 
#' install \pkg{TreeDist} using
#' `devtools::install_github("ms609/TreeDist", ref = "more-leaves")`.
#' (\pkg{TreeDist} will need building from source _after_ the modified 
#' \pkg{TreeTools} package has been installed, as its code links to values
#' set in the TreeTools source code.)
#' 
#' Trees with over 8192 leaves require further modification of the source code,
#' which the maintainer plans to attempt in the future; please [comment on GitHub](
#' https://github.com/ms609/TreeTools/issues/141) if you would find this useful.
#' 
#' # Parallelism
#' 
#' When `tree2 = NULL` and all trees share the same tip labels, pairwise
#' distance calculation uses a multi-threaded \acronym{OpenMP} batch path
#' automatically.  Control the number of threads with the `"mc.cores"` option:
#' 
#' ```r
#' options(mc.cores = parallel::detectCores())  # use all cores
#' options(mc.cores = 4L)                        # or a fixed number
#' ```
#' 
#' Do **not** call [`StartParallel()`] for these functions: a registered
#' R cluster disables the \acronym{OpenMP} path and replaces it with slower
#' inter-process communication.  See [`StartParallel()`] for full guidance
#' on when an R cluster is appropriate.
#' 
#' @template tree12ListParams
#' 
#' @param normalize If a numeric value is provided, this will be used as a 
#' maximum value against which to rescale results.
#' If `TRUE`, results will be rescaled against a maximum value calculated from
#' the specified tree sizes and topology, as specified in the "Normalization" 
#' section below.
#' If `FALSE`, results will not be rescaled.
#' 
#' @param diag Logical specifying whether to return similarities along the
#' diagonal, i.e. of each tree with itself.  Applies only if `tree2` is
#' a list identical to `tree1`, or `NULL`.
#' 
#' @param reportMatching Logical specifying whether to return the clade
#' matchings as an attribute of the score.
#'
#' @returns
#' If `reportMatching = FALSE`, the functions return a numeric 
#' vector specifying the requested similarities or differences.
#' 
#' If `reportMatching = TRUE`, the functions additionally return an integer
#' vector listing the index of the split in `tree2` that is matched with 
#' each split in `tree1` in the optimal matching.
#' Unmatched splits are denoted `NA`.
#' Use [`VisualizeMatching()`] to plot the optimal matching.
#' 
#' `TreeDistance()` simply returns the clustering information distance (it is
#' an alias of `ClusteringInfoDistance()`).
#'  
#' @examples 
#' tree1 <- ape::read.tree(text="((((a, b), c), d), (e, (f, (g, h))));")
#' tree2 <- ape::read.tree(text="(((a, b), (c, d)), ((e, f), (g, h)));")
#' tree3 <- ape::read.tree(text="((((h, b), c), d), (e, (f, (g, a))));")
#' 
#' # Best possible score is obtained by matching a tree with itself
#' DifferentPhylogeneticInfo(tree1, tree1) # 0, by definition
#' SharedPhylogeneticInfo(tree1, tree1)
#' SplitwiseInfo(tree1) # Maximum shared phylogenetic information
#' 
#' # Best possible score is a function of tree shape; the splits within
#' # balanced trees are more independent and thus contain less information
#' SplitwiseInfo(tree2)
#' 
#' # How similar are two trees?
#' SharedPhylogeneticInfo(tree1, tree2) # Amount of phylogenetic information in common
#' attr(SharedPhylogeneticInfo(tree1, tree2, reportMatching = TRUE), "matching")
#' VisualizeMatching(SharedPhylogeneticInfo, tree1, tree2) # Which clades are matched?
#' 
#' DifferentPhylogeneticInfo(tree1, tree2) # Distance measure
#' DifferentPhylogeneticInfo(tree2, tree1) # The metric is symmetric
#'
#' # Are they more similar than two trees of this shape would be by chance?
#' ExpectedVariation(tree1, tree2, sample=12)["DifferentPhylogeneticInfo", "Estimate"]
#' 
#' # Every split in tree1 conflicts with every split in tree3
#' # Pairs of conflicting splits contain clustering, but not phylogenetic, 
#' # information
#' SharedPhylogeneticInfo(tree1, tree3) # = 0
#' MutualClusteringInfo(tree1, tree3) # > 0
#' 
#' # Distance functions internally convert trees to Splits objects.
#' # Pre-conversion can reduce run time if the same trees will feature in
#' # multiple comparisons
#' splits1 <- TreeTools::as.Splits(tree1)
#' splits2 <- TreeTools::as.Splits(tree2)
#' 
#' SharedPhylogeneticInfoSplits(splits1, splits2)
#' MatchingSplitInfoSplits(splits1, splits2)
#' MutualClusteringInfoSplits(splits1, splits2)
#' @template MRS 
#' 
#' @references
#' \insertAllCited{}
#' 
#' @encoding UTF-8
#' @family tree distances
#' @export
TreeDistance <- function(tree1, tree2 = NULL) {
  ClusteringInfoDistance(tree1, tree2, normalize = TRUE, reportMatching = FALSE)
}

#' @rdname TreeDistance
#' @export
SharedPhylogeneticInfo <- function(tree1, tree2 = NULL, normalize = FALSE,
                                   reportMatching = FALSE, diag = TRUE) {
  
  unnormalized <- CalculateTreeDistance(SharedPhylogeneticInfoSplits, tree1,
                                        tree2, reportMatching = reportMatching)
  
  if (diag && is.null(tree2)) {
    unnormalized <- as.matrix(unnormalized)
    diag(unnormalized) <- SplitwiseInfo(tree1)
    tree2 <- tree1
  }
  
  # Return:
  NormalizeInfo(unnormalized, tree1, tree2, how = normalize,
                InfoInTree = SplitwiseInfo, Combine = .PairMean)
}

#' @rdname TreeDistance
#' @export
DifferentPhylogeneticInfo <- function(tree1, tree2 = NULL, normalize = FALSE,
                                      reportMatching = FALSE) {
  
  # Fast path (all-pairs): same tips, no matching — avoids duplicate as.Splits()
  fast <- .FastDistPath(tree1, tree2, reportMatching,
                        cpp_shared_phylo_all_pairs,
                        cpp_splitwise_info_batch)
  if (!is.null(fast)) {
    spi <- fast[["info"]]
    treesIndependentInfo <- .PairwiseSums(fast[["entropies"]])

    ret <- .FloorNumericalNoise(treesIndependentInfo - spi - spi, treesIndependentInfo)
    ret <- NormalizeInfo(ret, tree1, tree2, how = normalize,
                         infoInBoth = treesIndependentInfo,
                         InfoInTree = SplitwiseInfo, Combine = "+")
    attributes(ret) <- attributes(spi)
    return(ret)
  }

  # Fast path (cross-pairs): same tips, no matching — avoids duplicate as.Splits()
  fast_many <- .FastManyManyPath(tree1, tree2, reportMatching,
                                 cpp_shared_phylo_cross_pairs,
                                 cpp_splitwise_info_batch)
  if (!is.null(fast_many)) {
    spi <- fast_many[["dists"]]
    info1 <- fast_many[["info1"]]
    info2 <- fast_many[["info2"]]
    treesIndependentInfo <- outer(info1, info2, "+")

    ret <- .FloorNumericalNoise(treesIndependentInfo - spi - spi, treesIndependentInfo)
    ret <- NormalizeInfo(ret, tree1, tree2, how = normalize,
                         infoInBoth = treesIndependentInfo,
                         InfoInTree = SplitwiseInfo, Combine = "+")
    return(ret)
  }

  spi <- SharedPhylogeneticInfo(tree1, tree2, normalize = FALSE, diag = FALSE,
                                reportMatching = reportMatching)
  treesIndependentInfo <- .MaxValue(tree1, tree2, SplitwiseInfo)

  ret <- .FloorNumericalNoise(treesIndependentInfo - spi - spi, treesIndependentInfo)
  ret <- NormalizeInfo(ret, tree1, tree2, how = normalize,
                       infoInBoth = treesIndependentInfo,
                       InfoInTree = SplitwiseInfo, Combine = "+")
  attributes(ret) <- attributes(spi)
  
  # Return:
  ret
}

#' @rdname TreeDistance
#' @export
PhylogeneticInfoDistance <- DifferentPhylogeneticInfo

#' @rdname TreeDistance
#' @aliases ClusteringInfoDist
#' @export
ClusteringInfoDistance <- function(tree1, tree2 = NULL, normalize = FALSE,
                                   reportMatching = FALSE) {
  
  # Fast path (all-pairs): same tips, no matching — avoids duplicate as.Splits()
  fast <- .FastDistPath(tree1, tree2, reportMatching,
                        cpp_mutual_clustering_all_pairs,
                        cpp_clustering_entropy_batch)
  if (!is.null(fast)) {
    mci <- fast[["info"]]
    treesIndependentInfo <- .PairwiseSums(fast[["entropies"]])

    ret <- .FloorNumericalNoise(treesIndependentInfo - mci - mci, treesIndependentInfo)
    ret <- NormalizeInfo(ret, tree1, tree2, how = normalize,
                         infoInBoth = treesIndependentInfo,
                         InfoInTree = ClusteringEntropy, Combine = "+")
    attributes(ret) <- attributes(mci)
    return(ret)
  }

  # Fast path (cross-pairs): same tips, no matching — avoids duplicate as.Splits()
  fast_many <- .FastManyManyPath(tree1, tree2, reportMatching,
                                 cpp_mutual_clustering_cross_pairs,
                                 cpp_clustering_entropy_batch)
  if (!is.null(fast_many)) {
    mci <- fast_many[["dists"]]
    info1 <- fast_many[["info1"]]
    info2 <- fast_many[["info2"]]
    treesIndependentInfo <- outer(info1, info2, "+")

    ret <- .FloorNumericalNoise(treesIndependentInfo - mci - mci, treesIndependentInfo)
    ret <- NormalizeInfo(ret, tree1, tree2, how = normalize,
                         infoInBoth = treesIndependentInfo,
                         InfoInTree = ClusteringEntropy, Combine = "+")
    return(ret)
  }

  mci <- MutualClusteringInfo(tree1, tree2, normalize = FALSE, diag = FALSE,
                              reportMatching = reportMatching)
  treesIndependentInfo <- .MaxValue(tree1, tree2, ClusteringEntropy)

  ret <- .FloorNumericalNoise(treesIndependentInfo - mci - mci, treesIndependentInfo)
  ret <- NormalizeInfo(ret, tree1, tree2, how = normalize,
                       infoInBoth = treesIndependentInfo,
                       InfoInTree = ClusteringEntropy, Combine = "+")
  attributes(ret) <- attributes(mci)
  
  # Return:
  ret
}

#' @export
ClusteringInfoDist <- ClusteringInfoDistance

#' @rdname TreeDistance
#' @param samples Integer specifying how many samplings to obtain; 
#' accuracy of estimate increases with `sqrt(samples)`.
#' @importFrom stats sd
#' @importFrom TreeTools as.Splits
#' @export
ExpectedVariation <- function(tree1, tree2, samples = 1e+4) {
  info1 <- SplitwiseInfo(tree1)
  info2 <- SplitwiseInfo(tree2)
  splits1 <- as.Splits(tree1)
  tipLabels <- attr(splits1, "tip.label")
  nTip <- attr(splits1, "nTip")
  splits2 <- as.Splits(tree2, tipLabels)
  
  mutualEstimates <- vapply(seq_len(samples), function(x) {
    resampled2 <- as.Splits(splits2, sample(tipLabels))
    
    c(SharedPhylogeneticInfoSplits(splits1, resampled2),
      MatchingSplitInfoSplits(splits1, resampled2),
      MutualClusteringInfoSplits(splits1, resampled2)
      )
  }, c(SharedPhylogeneticInfo = 0, MatchingSplitInfo = 0,
       MutualClusteringInfo = 0))
  
  mut <- cbind(Estimate = rowMeans(mutualEstimates),
               sd = apply(mutualEstimates, 1, sd), n = samples)
  
  ret <- rbind(
    mut,
    DifferentPhylogeneticInfo = c(info1 + info2 - mut[1, 1] - mut[1, 1],
                                    mut[1, 2] * 2, samples),
    MatchingSplitInfoDistance = c(info1 + info2 - mut[2, 1] - mut[2, 1],
                                     mut[2, 2] * 2, samples),
    ClusteringInfoDistance = c(ClusteringEntropy(tree1) + 
                                 ClusteringEntropy(tree2) - 
                                 mut[3, 1] - mut[3, 1],
                               mut[3, 2] * 2, samples)
  )
  
  # Return:
  cbind(Estimate = ret[, 1],
        "Std. Err." = ret[, "sd"] / sqrt(ret[, "n"]), 
        ret[, 2:3])
}

#' @rdname TreeDistance
#' @aliases MutualClusteringInformation
#' @export
MutualClusteringInfo <- function(tree1, tree2 = NULL, normalize = FALSE,
                                 reportMatching = FALSE, diag = TRUE) {
  unnormalized <- CalculateTreeDistance(Func = MutualClusteringInfoSplits,
                                        tree1, tree2, reportMatching)
  if (diag && is.null(tree2)) {
    unnormalized <- as.matrix(unnormalized)
    diag(unnormalized) <- ClusteringEntropy(tree1)
    tree2 <- tree1
  }
  NormalizeInfo(unnormalized, tree1, tree2, ClusteringEntropy,
                how = normalize, Combine = .PairMean)
}

#' @export
MutualClusteringInformation <- MutualClusteringInfo

#' @rdname TreeDistance
#' @template splits12params
#' @param nTip (Optional) Integer specifying the number of leaves in each split.
#' @export
SharedPhylogeneticInfoSplits <- function(splits1, splits2,
                                         nTip = attr(splits1, "nTip"),
                                         reportMatching = FALSE) {
  GeneralizedRF(splits1, splits2, nTip, cpp_shared_phylo,
                maximize = TRUE, reportMatching = reportMatching)
}

#' @rdname TreeDistance
#' @export
MutualClusteringInfoSplits <- function(splits1, splits2,
                                       nTip = attr(splits1, "nTip"),
                                       reportMatching = FALSE) {
  GeneralizedRF(splits1, splits2, nTip, cpp_mutual_clustering,
                maximize = TRUE, reportMatching = reportMatching)
}

#' Mean of two numbers
#' 
#' Used for normalization and range calculation
#' 
#' @keywords internal
.PairMean <- function(x, y) (x + y) / 2L

## Tests for mutual_clustering_score() branches in src/pairwise_distances.cpp.
##
## These branches are only reachable via cpp_mutual_clustering_all_pairs(),
## invoked automatically by MutualClusteringInfo() when all trees share the
## same tip set and no R-level cluster is active (the fast path in
## .SplitDistanceAllPairs()).  The fast path returns a dist; MutualClusteringInfo
## then converts it to a full matrix and fills the diagonal with ClusteringEntropy
## values.  The off-diagonal entries equal the raw MCI scores from the batch C++
## function.

test_that("cpp_mutual_clustering_all_pairs: orthogonal splits score 0", {
  # Two 8-tip trees, each with exactly one internal split.
  # The splits cross orthogonally: every quadrant contains exactly 2 tips,
  # so a_and_b == A_and_b == a_and_B == A_and_B == 2.
  # This triggers the rounding-error guard (score = max_score), and
  # the LAP assigns that sole pair, yielding MCI = 0.
  tr1 <- ape::read.tree(text = "((t1,t2,t3,t4),(t5,t6,t7,t8));")
  tr2 <- ape::read.tree(text = "((t1,t2,t5,t6),(t3,t4,t7,t8));")
  trees <- structure(list(tr1, tr2), class = "multiPhylo")

  r <- MutualClusteringInfo(trees)

  # r is a 2×2 matrix; the off-diagonal [2,1] is the pairwise MCI
  expect_equal(r[2, 1], 0, tolerance = 1e-10)
  # Agrees with the single-pair path (cpp_mutual_clustering, not the batch fn)
  expect_equal(r[2, 1], MutualClusteringInfo(tr1, tr2), tolerance = 1e-10)
})

test_that("cpp_mutual_clustering_all_pairs: unequal split counts", {
  # Three 6-tip trees with different numbers of non-trivial splits:
  #
  #   tr_1a — 1 split:  {t1,t2,t3} | {t4,t5,t6}
  #   tr_3  — 3 splits: {t1,t4}, {t2,t5}, {t3,t6}
  #   tr_1b — 1 split:  {t1,t2}   | {t3,t4,t5,t6}
  #
  # The batch loop (col < row) processes three pairs using 0-indexed trees:
  #
  #   (col=0 → a=tr_1a[1 split], row=1 → b=tr_3[3 splits]):
  #     a_has_more = FALSE; most_splits = 3.
  #     No split pairs are exact matches, so exact_n = 0 → else branch:
  #       loop fills phantom rows ai=1,2 with max_score
  #       LAP solves the full 3×3 matrix
  #
  #   (col=1 → a=tr_3[3 splits], row=2 → b=tr_1b[1 split]):
  #     a_has_more = TRUE; most_splits = 3, b.n_splits = 1.
  #     For every ai=0..2: padRowAfterCol(ai, 1, max_score)

  tr_1a <- ape::read.tree(text = "((t1,t2,t3),(t4,t5,t6));")
  tr_3  <- ape::read.tree(text = "(((t1,t4),(t2,t5)),(t3,t6));")
  tr_1b <- ape::read.tree(text = "((t1,t2),(t3,t4,t5,t6));")
  trees <- structure(list(tr_1a, tr_3, tr_1b), class = "multiPhylo")

  r <- MutualClusteringInfo(trees)

  # r is a 3×3 matrix; off-diagonal [i, j] with i > j is pair (j, i)
  expect_equal(dim(r), c(3L, 3L))

  # Cross-validate all three off-diagonal pairs against the single-pair path
  # (which uses cpp_mutual_clustering in tree_distances.cpp, not the batch fn)
  expect_equal(r[2, 1], MutualClusteringInfo(tr_1a, tr_3),  tolerance = 1e-9)
  expect_equal(r[3, 1], MutualClusteringInfo(tr_1a, tr_1b), tolerance = 1e-9)
  expect_equal(r[3, 2], MutualClusteringInfo(tr_3,  tr_1b), tolerance = 1e-9)
})

test_that("cpp_rf_info_all_pairs: multi-bin complement (> 64 tips)", {
  # Exercises b_complement[i][bin] = ~b.state[i][bin] at line 236, which
  # only runs when n_bins >= 2 (trees with > 64 tips on 64-bit platforms).
  skip_on_cran()
  trees <- ape::as.phylo(0:2, tipLabels = paste0("t", seq_len(100)))
  r <- InfoRobinsonFoulds(trees)
  expect_true(all(r >= 0))
  expect_equal(as.matrix(r)[2, 1],
               InfoRobinsonFoulds(trees[[1]], trees[[2]]),
               tolerance = 1e-10)
})

test_that("cpp_jaccard_all_pairs: allow_conflict = FALSE", {
  # Random trees have conflicting splits, so with allowConflict = FALSE
  # those pairs score max_score.
  trees <- ape::as.phylo(0:2, tipLabels = paste0("t", seq_len(20)))
  r <- JaccardRobinsonFoulds(trees, allowConflict = FALSE)
  expect_true(all(r >= 0))
  expect_equal(as.matrix(r)[2, 1],
               JaccardRobinsonFoulds(trees[[1]], trees[[2]],
                                     allowConflict = FALSE),
               tolerance = 1e-10)
})

test_that("cpp_jaccard_all_pairs: k = Inf and k != 1", {
  trees <- ape::as.phylo(0:2, tipLabels = paste0("t", seq_len(12)))
  # k = Inf triggers the isinf branch
  r_inf <- JaccardRobinsonFoulds(trees, k = Inf)
  expect_true(all(r_inf >= 0))
  expect_equal(as.matrix(r_inf)[2, 1],
               JaccardRobinsonFoulds(trees[[1]], trees[[2]], k = Inf),
               tolerance = 1e-10)
  # k = 2 triggers the pow() branch
  r_k2 <- JaccardRobinsonFoulds(trees, k = 2)
  expect_true(all(r_k2 >= 0))
  expect_equal(as.matrix(r_k2)[2, 1],
               JaccardRobinsonFoulds(trees[[1]], trees[[2]], k = 2),
               tolerance = 1e-10)
})

test_that("cpp_msd_all_pairs: unequal split counts, no exact matches", {
  # Three 6-tip trees with different numbers of non-trivial splits:
  #
  #   tr_3  — 3 splits: fully resolved
  #   tr_1a — 1 split:  {t1,t2,t3} | {t4,t5,t6}
  #   tr_1b — 1 split:  {t1,t4}    | {t2,t3,t5,t6}
  #
  # tr_1a and tr_1b share no splits (including complements), so their
  # comparison with tr_3 exercises:
  #   - line 361: padRowAfterCol when a.n_splits > b.n_splits
  #     (pair col=0 [tr_3, 3 splits] > row=1 [tr_1a, 1 split])
  #   - lines 393–396: the else branch (exact_n == 0) with unequal counts
  #     (pair col=0 [tr_3] vs row=1 [tr_1a], no shared splits)
  tr_3  <- ape::read.tree(text = "(((t1,t4),(t2,t5)),(t3,t6));")
  tr_1a <- ape::read.tree(text = "((t1,t2,t3),(t4,t5,t6));")
  tr_1b <- ape::read.tree(text = "((t1,t4),(t2,t3,t5,t6));")
  trees <- structure(list(tr_3, tr_1a, tr_1b), class = "multiPhylo")

  r <- MatchingSplitDistance(trees)
  m <- as.matrix(r)

  # Cross-validate batch vs single-pair path
  expect_equal(m[2, 1], MatchingSplitDistance(tr_3,  tr_1a), tolerance = 0)
  expect_equal(m[3, 1], MatchingSplitDistance(tr_3,  tr_1b), tolerance = 0)
  expect_equal(m[3, 2], MatchingSplitDistance(tr_1a, tr_1b), tolerance = 0)
})

test_that("Large trees: batch and single-pair paths agree (lg2 table bounds)", {
  # Exercise the lg2 / lg2_double_factorial / lg2_unrooted tables at indices

  # well beyond small-tree tests.  Catches out-of-bounds table access and
  # ensures the log2 decomposition in add_ic_element remains correct for
  # trees with many splits.
  skip_on_cran()
  set.seed(4728)
  trees <- ape::as.phylo(0:4, tipLabels = paste0("t", seq_len(200)))

  batch <- MutualClusteringInfo(trees)
  n <- length(trees)
  pairs <- which(upper.tri(batch), arr.ind = TRUE)
  for (k in seq_len(nrow(pairs))) {
    i <- pairs[k, 1]
    j <- pairs[k, 2]
    expect_equal(
      batch[i, j],
      MutualClusteringInfo(trees[[i]], trees[[j]]),
      tolerance = 1e-8,
      label = paste0("pair (", i, ",", j, ")")
    )
  }
})

test_that("Batch path handles asymmetric split counts with exact matches", {
  # Trees with different numbers of splits (polytomies) that share some

  # splits exactly.  After exact-match detection removes the shared splits,
  # the reduced LAP matrix is asymmetric (a_unmatched_n != b_unmatched_n),
  # exercising the padAfterRow guard in msd_score and jaccard_score.
  t1 <- ape::read.tree(text = "((t1,t2,t3),(t4,(t5,(t6,(t7,t8)))));")
  t2 <- ape::read.tree(text = "(((t1,t4),(t2,t3)),((t5,t6),(t7,t8)));")
  t3 <- ape::read.tree(text = "((t1,(t2,(t3,t4))),(t5,t6,t7,t8));")
  t4 <- ape::read.tree(text = "(((t1,t2),(t3,t4)),((t5,t7),(t6,t8)));")
  trees <- structure(list(t1, t2, t3, t4), class = "multiPhylo")

  # MSD batch vs per-pair
  msd_b <- as.matrix(MatchingSplitDistance(trees))
  for (i in seq_len(3)) for (j in (i + 1):4) {
    expect_equal(msd_b[i, j],
                 MatchingSplitDistance(trees[[i]], trees[[j]]),
                 tolerance = 1e-10,
                 label = paste0("MSD(", i, ",", j, ")"))
  }

  # Jaccard batch vs per-pair
  jrf_b <- as.matrix(JaccardRobinsonFoulds(trees))
  for (i in seq_len(3)) for (j in (i + 1):4) {
    expect_equal(jrf_b[i, j],
                 JaccardRobinsonFoulds(trees[[i]], trees[[j]]),
                 tolerance = 1e-10,
                 label = paste0("JRF(", i, ",", j, ")"))
  }
})

test_that("InfoRobinsonFoulds(similarity = TRUE) uses batch path", {
  # similarity = TRUE bypasses .FastDistPath() and falls through to
  # CalculateTreeDistance → .SplitDistanceAllPairs / .SplitDistanceManyMany,
  # exercising the IRF dispatch branches in tree_distance_utilities.R.
  trees <- ape::as.phylo(0:9, tipLabels = paste0("t", seq_len(20)))

  # All-pairs
  irf_sim <- InfoRobinsonFoulds(trees, similarity = TRUE)
  expect_true(inherits(irf_sim, "dist"))
  irf_mat <- as.matrix(irf_sim)
  expect_equal(irf_mat[2, 1],
               InfoRobinsonFoulds(trees[[1]], trees[[2]], similarity = TRUE),
               tolerance = 1e-10)

  # Cross-pairs
  tA <- trees[1:4]
  tB <- trees[5:10]
  irf_cross <- InfoRobinsonFoulds(tA, tB, similarity = TRUE)
  expect_equal(dim(irf_cross), c(4L, 6L))
  expect_equal(irf_cross[1, 1],
               InfoRobinsonFoulds(tA[[1]], tB[[1]], similarity = TRUE),
               tolerance = 1e-10)
})

test_that(".FastManyManyPath: cross-pairs distance exercises happy path", {
  # Exercises the happy path through .FastManyManyPath() in tree_distance.R:
  # cluster guard, TipLabels extraction, tip set matching, nTip check.
  tA <- ape::as.phylo(0:4, tipLabels = paste0("t", seq_len(20)))
  tB <- ape::as.phylo(5:9, tipLabels = paste0("t", seq_len(20)))

  cid_cross <- ClusteringInfoDistance(tA, tB)
  expect_equal(dim(cid_cross), c(5L, 5L))
  expect_equal(cid_cross[1, 1],
               ClusteringInfoDistance(tA[[1]], tB[[1]]),
               tolerance = 1e-10)
})

test_that(".FastManyManyPath: guards return NULL for edge cases", {
  tips <- paste0("t", seq_len(8))
  tA <- ape::as.phylo(0:2, tipLabels = tips)
  tB <- ape::as.phylo(3:5, tipLabels = tips)

  # Mismatched tip sets → falls back to slow path (returns non-NULL result)
  tB_diff <- ape::as.phylo(3:5, tipLabels = paste0("s", seq_len(8)))
  expect_true(!is.null(ClusteringInfoDistance(tA, tB_diff)))

  # Small trees (nTip < 4) → falls back to slow path
  tA3 <- ape::as.phylo(0:2, tipLabels = paste0("t", 1:3))
  tB3 <- ape::as.phylo(0:2, tipLabels = paste0("t", 1:3))
  expect_true(!is.null(ClusteringInfoDistance(tA3, tB3)))
})

test_that("SPI batch does not greedily match identical splits", {
  # Regression test for a correctness bug in shared_phylo_score().
  #
  # For the SPI (SharedPhylogeneticInfo) metric, matching identical splits
  # to each other is NOT necessarily globally optimal: spi_overlap(A, B)
  # where B *contains* A can exceed spi_overlap(A, A), so the full LAP
  # can find a better total score by pairing some identical splits with
  # non-identical (containing) partners.
  #
  # A greedy exact-match shortcut (match identical splits first, then
  # solve a reduced LAP) understates SPI, causing DifferentPhylogeneticInfo
  # (= Info1 + Info2 - 2 * SPI) to be overstated (trees appear more
  # different than they are).
  #
  # Counterexample: as.phylo(1, 10) and as.phylo(3, 10) share 5 of 7
  # splits.  The optimal LAP matches s1's 5|5 split to s2's 7|3 split
  # (which contains it) rather than to its identical copy in s2, because
  # spi_overlap({2,6,7,8,9,10}, {2,4,6,7,8,9,10}) > spi_overlap of the
  # identical pair.  The full-LAP SPI (32.012) exceeds the greedy
  # exact-match credit (31.771) even before the reduced LAP adds to it.
  # The same issue affects MatchingSplitInfoSplits (MSI).
  t1 <- ape::as.phylo(1, 10)
  t2 <- ape::as.phylo(3, 10)

  # Verify: optimal matching pairs s1[2] with s2[1] (non-identical),
  # even though s1[2] is identical to s2[2].
  s1 <- TreeTools::as.Splits(t1)
  s2 <- TreeTools::as.Splits(t2)
  result <- SharedPhylogeneticInfoSplits(s1, s2, reportMatching = TRUE)
  matching <- attr(result, "matching")

  # s1[2] is identical to s2[2] (same 5|5 split)
  expect_true(
    identical(as.logical(s1[[2]]), as.logical(s2[[2]])),
    label = "s1[2] and s2[2] are the same bipartition"
  )
  # But the optimal LAP matches s1[2] to s2[1] (a containing 7|3 split)
  expect_equal(matching[2], 1L,
               label = "optimal LAP matches s1[2] to s2[1], not its identical copy")

  # Batch path (all-pairs) must agree with the per-pair path
  trees <- structure(list(t1, t2), class = "multiPhylo")
  batch_spi <- SharedPhylogeneticInfo(trees)
  pair_spi <- SharedPhylogeneticInfo(t1, t2)
  expect_equal(as.matrix(batch_spi)[2, 1], pair_spi, tolerance = 1e-10)

  # The full-LAP SPI must exceed the sum of self-info for the 5 shared
  # splits (what the greedy approach would credit as exact-match score).
  # This proves the greedy is wrong — it understates SPI.
  n <- 10L
  shared_self_info <- sum(vapply(2:6, function(a) {
    k <- sum(as.logical(s1[[a]]))
    log2(TreeTools::NUnrooted(n)) -
      log2(TreeTools::NRooted(k)) - log2(TreeTools::NRooted(n - k))
  }, double(1)))
  expect_gt(pair_spi, shared_self_info,
            label = "full-LAP SPI exceeds greedy exact-match credit")

  # Verify the same holds for MSI (same vulnerability: msi_score formerly
  # used greedy exact-match detection, which is incorrect for MSI).
  batch_msi <- MatchingSplitInfoDistance(trees)
  pair_msi <- MatchingSplitInfoDistance(t1, t2)
  expect_equal(as.matrix(batch_msi)[2, 1], pair_msi, tolerance = 1e-10)
})

## Coverage tests for batch C++ functions, fast paths, and cross-pairs.
##
## These tests target code paths introduced during the OpenMP / sort+merge /
## CostMatrix-pooling optimisation cycle that are not exercised by the
## existing test suite.

library("TreeTools", quietly = TRUE)

# Shared fixtures ----
tips20 <- paste0("t", seq_len(20))
tA <- ape::as.phylo(0:4, tipLabels = tips20)
tB <- ape::as.phylo(5:9, tipLabels = tips20)
trees20 <- ape::as.phylo(0:9, tipLabels = tips20)

tips8 <- paste0("t", seq_len(8))
tA8 <- ape::as.phylo(0:2, tipLabels = tips8)
tB8 <- ape::as.phylo(3:5, tipLabels = tips8)


# DifferentPhylogeneticInfo ----

test_that("DPI all-pairs fast path agrees with per-pair", {
  batch <- DifferentPhylogeneticInfo(trees20)
  m <- as.matrix(batch)
  expect_equal(m[2, 1],
               DifferentPhylogeneticInfo(trees20[[1]], trees20[[2]]),
               tolerance = 1e-10)
  expect_equal(m[5, 3],
               DifferentPhylogeneticInfo(trees20[[3]], trees20[[5]]),
               tolerance = 1e-10)
})

test_that("DPI cross-pairs fast path agrees with per-pair", {
  cross <- DifferentPhylogeneticInfo(tA, tB)
  expect_equal(dim(cross), c(5L, 5L))
  expect_equal(cross[1, 1],
               DifferentPhylogeneticInfo(tA[[1]], tB[[1]]),
               tolerance = 1e-10)
  expect_equal(cross[3, 2],
               DifferentPhylogeneticInfo(tA[[3]], tB[[2]]),
               tolerance = 1e-10)
})


# MatchingSplitInfoDistance ----

test_that("MSID all-pairs fast path agrees with per-pair", {
  batch <- MatchingSplitInfoDistance(trees20)
  m <- as.matrix(batch)
  expect_equal(m[2, 1],
               MatchingSplitInfoDistance(trees20[[1]], trees20[[2]]),
               tolerance = 1e-10)
  expect_equal(m[5, 3],
               MatchingSplitInfoDistance(trees20[[3]], trees20[[5]]),
               tolerance = 1e-10)
})

test_that("MSID cross-pairs fast path agrees with per-pair", {
  cross <- MatchingSplitInfoDistance(tA, tB)
  expect_equal(dim(cross), c(5L, 5L))
  expect_equal(cross[1, 1],
               MatchingSplitInfoDistance(tA[[1]], tB[[1]]),
               tolerance = 1e-10)
  expect_equal(cross[3, 2],
               MatchingSplitInfoDistance(tA[[3]], tB[[2]]),
               tolerance = 1e-10)
})


# InfoRobinsonFoulds (distance mode) ----

test_that("IRF all-pairs fast path agrees with per-pair", {
  batch <- InfoRobinsonFoulds(trees20)
  m <- as.matrix(batch)
  expect_equal(m[2, 1],
               InfoRobinsonFoulds(trees20[[1]], trees20[[2]]),
               tolerance = 1e-10)
  expect_equal(m[5, 3],
               InfoRobinsonFoulds(trees20[[3]], trees20[[5]]),
               tolerance = 1e-10)
})

test_that("IRF cross-pairs fast path agrees with per-pair", {
  cross <- InfoRobinsonFoulds(tA, tB)
  expect_equal(dim(cross), c(5L, 5L))
  expect_equal(cross[1, 1],
               InfoRobinsonFoulds(tA[[1]], tB[[1]]),
               tolerance = 1e-10)
  expect_equal(cross[3, 2],
               InfoRobinsonFoulds(tA[[3]], tB[[2]]),
               tolerance = 1e-10)
})


# ClusteringInfoDistance (all-pairs already tested; add cross-pairs variety) ----

test_that("CID cross-pairs with different-sized collections", {
  tA3 <- ape::as.phylo(0:2, tipLabels = tips20)
  tB7 <- ape::as.phylo(3:9, tipLabels = tips20)
  cross <- ClusteringInfoDistance(tA3, tB7)
  expect_equal(dim(cross), c(3L, 7L))
  expect_equal(cross[2, 5],
               ClusteringInfoDistance(tA3[[2]], tB7[[5]]),
               tolerance = 1e-10)
})


# MSD cross-pairs via CalculateTreeDistance → .SplitDistanceManyMany ----

test_that("MSD cross-pairs batch agrees with per-pair", {
  cross <- MatchingSplitDistance(tA8, tB8)
  expect_equal(dim(cross), c(3L, 3L))
  expect_equal(cross[1, 1],
               MatchingSplitDistance(tA8[[1]], tB8[[1]]),
               tolerance = 1e-10)
  expect_equal(cross[2, 3],
               MatchingSplitDistance(tA8[[2]], tB8[[3]]),
               tolerance = 1e-10)
})


# Nye / Jaccard cross-pairs via CalculateTreeDistance → .SplitDistanceManyMany ----

test_that("NyeSimilarity cross-pairs batch agrees with per-pair", {
  cross <- NyeSimilarity(tA8, tB8)
  expect_equal(dim(cross), c(3L, 3L))
  expect_equal(cross[1, 1],
               NyeSimilarity(tA8[[1]], tB8[[1]]),
               tolerance = 1e-10)
})

test_that("JRF cross-pairs batch with k and allowConflict", {
  # Default k=1, allowConflict=TRUE
  cross1 <- JaccardRobinsonFoulds(tA8, tB8)
  expect_equal(dim(cross1), c(3L, 3L))
  expect_equal(cross1[2, 1],
               JaccardRobinsonFoulds(tA8[[2]], tB8[[1]]),
               tolerance = 1e-10)

  # k = 2
  cross_k2 <- JaccardRobinsonFoulds(tA8, tB8, k = 2)
  expect_equal(cross_k2[1, 1],
               JaccardRobinsonFoulds(tA8[[1]], tB8[[1]], k = 2),
               tolerance = 1e-10)

  # allowConflict = FALSE
  cross_nc <- JaccardRobinsonFoulds(tA8, tB8, allowConflict = FALSE)
  expect_equal(cross_nc[1, 1],
               JaccardRobinsonFoulds(tA8[[1]], tB8[[1]],
                                     allowConflict = FALSE),
               tolerance = 1e-10)

  # k = Inf
  cross_inf <- JaccardRobinsonFoulds(tA8, tB8, k = Inf)
  expect_equal(cross_inf[2, 2],
               JaccardRobinsonFoulds(tA8[[2]], tB8[[2]], k = Inf),
               tolerance = 1e-10)
})


# MSI cross-pairs via CalculateTreeDistance → .SplitDistanceManyMany ----

test_that("MSI cross-pairs batch agrees with per-pair", {
  cross <- MatchingSplitInfo(tA8, tB8)
  expect_equal(dim(cross), c(3L, 3L))
  expect_equal(cross[1, 1],
               MatchingSplitInfo(tA8[[1]], tB8[[1]]),
               tolerance = 1e-10)
})

# SPI cross-pairs via CalculateTreeDistance → .SplitDistanceManyMany ----

test_that("SPI cross-pairs batch agrees with per-pair", {
  cross <- SharedPhylogeneticInfo(tA8, tB8)
  expect_equal(dim(cross), c(3L, 3L))
  expect_equal(cross[1, 1],
               SharedPhylogeneticInfo(tA8[[1]], tB8[[1]]),
               tolerance = 1e-10)
})


# C++ batch entropy/info functions ----

test_that("cpp_clustering_entropy_batch matches R ClusteringEntropy", {
  trees <- ape::as.phylo(0:4, tipLabels = tips20)
  splits_list <- as.Splits(trees)
  nTip <- length(tips20)
  batch <- TreeDist:::cpp_clustering_entropy_batch(splits_list, as.integer(nTip))
  r_ref <- vapply(trees, ClusteringEntropy, double(1))
  expect_equal(batch, unname(r_ref), tolerance = 1e-12)
})

test_that("cpp_splitwise_info_batch matches R SplitwiseInfo", {
  trees <- ape::as.phylo(0:4, tipLabels = tips20)
  splits_list <- as.Splits(trees)
  nTip <- length(tips20)
  batch <- TreeDist:::cpp_splitwise_info_batch(splits_list, as.integer(nTip))
  r_ref <- vapply(trees, SplitwiseInfo, double(1))
  expect_equal(batch, unname(r_ref), tolerance = 1e-8)
})

test_that("cpp_clustering_entropy_batch: polytomy trees", {
  # Trees with a single non-trivial split (polytomy)
  poly1 <- ape::read.tree(text = "((t1,t2,t3),(t4,t5,t6));")
  poly2 <- ape::read.tree(text = "((t1,t2),(t3,t4,t5,t6));")
  trees <- structure(list(poly1, poly2), class = "multiPhylo")
  splits_list <- as.Splits(trees)
  nTip <- 6L
  batch <- TreeDist:::cpp_clustering_entropy_batch(splits_list, nTip)
  r_ref <- vapply(trees, ClusteringEntropy, double(1))
  expect_equal(batch, unname(r_ref), tolerance = 1e-12)
})

test_that("cpp_splitwise_info_batch: polytomy trees", {
  poly1 <- ape::read.tree(text = "((t1,t2,t3),(t4,t5,t6));")
  poly2 <- ape::read.tree(text = "((t1,t2),(t3,t4,t5,t6));")
  trees <- structure(list(poly1, poly2), class = "multiPhylo")
  splits_list <- as.Splits(trees)
  nTip <- 6L
  batch <- TreeDist:::cpp_splitwise_info_batch(splits_list, nTip)
  r_ref <- vapply(trees, SplitwiseInfo, double(1))
  expect_equal(batch, unname(r_ref), tolerance = 1e-8)
})


# KendallColijn batch paths ----

test_that("KC all-pairs uses pair_diff_euclidean", {
  trees <- ape::as.phylo(0:4, tipLabels = tips8)
  kc <- KendallColijn(trees)
  expect_s3_class(kc, "dist")
  expect_equal(attr(kc, "Size"), 5L)
  m <- as.matrix(kc)
  expect_equal(m[2, 1], KendallColijn(trees[[1]], trees[[2]]),
               tolerance = 1e-10)
  expect_equal(m[4, 3], KendallColijn(trees[[3]], trees[[4]]),
               tolerance = 1e-10)
})

test_that("KC cross-pairs uses vec_diff_euclidean", {
  cross <- KendallColijn(tA8, tB8)
  expect_equal(dim(cross), c(3L, 3L))
  expect_equal(cross[1, 1], KendallColijn(tA8[[1]], tB8[[1]]),
               tolerance = 1e-10)
  expect_equal(cross[2, 3], KendallColijn(tA8[[2]], tB8[[3]]),
               tolerance = 1e-10)
})

test_that("KC all-pairs with SplitVector", {
  trees <- ape::as.phylo(0:3, tipLabels = tips8)
  kc <- KendallColijn(trees, Vector = SplitVector)
  expect_s3_class(kc, "dist")
  m <- as.matrix(kc)
  expect_equal(m[2, 1],
               KendallColijn(trees[[1]], trees[[2]], Vector = SplitVector),
               tolerance = 1e-10)
})

test_that("KC cross-pairs with SplitVector", {
  cross <- KendallColijn(tA8, tB8, Vector = SplitVector)
  expect_equal(dim(cross), c(3L, 3L))
  expect_equal(cross[1, 1],
               KendallColijn(tA8[[1]], tB8[[1]], Vector = SplitVector),
               tolerance = 1e-10)
})


# Large-tree cross-pairs (multi-bin >64 tips) ----

test_that("Cross-pairs with >64-tip trees exercise multi-bin path", {
  skip_on_cran()
  tips100 <- paste0("t", seq_len(100))
  tA100 <- ape::as.phylo(0:2, tipLabels = tips100)
  tB100 <- ape::as.phylo(3:5, tipLabels = tips100)

  # CID cross-pairs
  cid <- ClusteringInfoDistance(tA100, tB100)
  expect_equal(dim(cid), c(3L, 3L))
  expect_equal(cid[1, 1],
               ClusteringInfoDistance(tA100[[1]], tB100[[1]]),
               tolerance = 1e-10)

  # MSD cross-pairs
  msd <- MatchingSplitDistance(tA100, tB100)
  expect_equal(dim(msd), c(3L, 3L))
  expect_equal(msd[1, 1],
               MatchingSplitDistance(tA100[[1]], tB100[[1]]),
               tolerance = 1e-10)

  # DPI cross-pairs
  dpi <- DifferentPhylogeneticInfo(tA100, tB100)
  expect_equal(dim(dpi), c(3L, 3L))
  expect_equal(dpi[1, 1],
               DifferentPhylogeneticInfo(tA100[[1]], tB100[[1]]),
               tolerance = 1e-10)
})

test_that("All-pairs entropy batch with >64-tip trees", {
  skip_on_cran()
  tips100 <- paste0("t", seq_len(100))
  trees100 <- ape::as.phylo(0:2, tipLabels = tips100)
  splits_list <- as.Splits(trees100)

  ce <- TreeDist:::cpp_clustering_entropy_batch(splits_list, 100L)
  ce_ref <- vapply(trees100, ClusteringEntropy, double(1))
  expect_equal(ce, unname(ce_ref), tolerance = 1e-12)

  si <- TreeDist:::cpp_splitwise_info_batch(splits_list, 100L)
  si_ref <- vapply(trees100, SplitwiseInfo, double(1))
  expect_equal(si, unname(si_ref), tolerance = 1e-8)
})


# SPI/MSI all-pairs batch dispatch via .SplitDistanceAllPairs ----

test_that("SPI all-pairs batch agrees with per-pair", {
  batch <- SharedPhylogeneticInfo(trees20)
  m <- as.matrix(batch)
  expect_equal(m[2, 1],
               SharedPhylogeneticInfo(trees20[[1]], trees20[[2]]),
               tolerance = 1e-10)
})

test_that("MSI all-pairs batch agrees with per-pair", {
  batch <- MatchingSplitInfo(trees20)
  m <- as.matrix(batch)
  expect_equal(m[2, 1],
               MatchingSplitInfo(trees20[[1]], trees20[[2]]),
               tolerance = 1e-10)
})


# Nye all-pairs batch dispatch ----

test_that("Nye all-pairs batch agrees with per-pair", {
  batch <- NyeSimilarity(trees20)
  m <- as.matrix(batch)
  expect_equal(m[2, 1],
               NyeSimilarity(trees20[[1]], trees20[[2]]),
               tolerance = 1e-10)
})


# Regression test for issue #162 ----
# MutualClusteringInfo matching was not globally optimal for larger trees,
# causing ClusteringInfoDistance to disagree with manual calculation.
# Fixed in v2.11.0 (PR #163).  This test verifies the fix holds in the
# batch path as well as the per-pair path.

test_that("Issue #162: CID by-hand matches function for 33-taxon trees", {
  tree1 <- ape::read.tree(text = "(B,A,((AG,AF),((((C,(E,D)),((F,G),H)),((K,I),J)),((Q,R),((((AE,(AC,AD)),(AB,(N,(P,O)))),(((Y,(Z,AA)),(W,X)),(V,(T,U)))),((M,S),L))))));")
  tree2 <- ape::read.tree(text = "(B,A,((AG,AF),((((C,(E,D)),((F,G),H)),(((Q,R),((((AE,(AC,AD)),AB),(N,(P,O))),((Y,(Z,AA)),((V,(T,U)),(W,X))))),((M,L),S))),((K,I),J))));")

  # Per-pair path: reported matching score must equal MCI value

  mci <- MutualClusteringInfo(tree1, tree2, reportMatching = TRUE)
  expect_equal(sum(attr(mci, "matchedScores")), mci[[1]], tolerance = 1e-10)

  # Per-pair path: CID by hand must equal CID function
  h <- ClusteringEntropy(tree1) + ClusteringEntropy(tree2)
  d_fn <- ClusteringInfoDistance(tree1, tree2, normalize = TRUE)
  d_hand <- (h - 2 * mci[[1]]) / h
  expect_equal(d_fn, d_hand, tolerance = 1e-10)
})

# RobinsonFoulds cross-pairs (Day 1985 ClusterTable batch) ----

test_that("RF cross-pairs fast path agrees with per-pair", {
  cross <- RobinsonFoulds(tA, tB)
  expect_equal(dim(cross), c(5L, 5L))
  expect_equal(cross[1, 1],
               RobinsonFoulds(tA[[1]], tB[[1]]),
               tolerance = 0)
  expect_equal(cross[3, 2],
               RobinsonFoulds(tA[[3]], tB[[2]]),
               tolerance = 0)
  expect_equal(cross[5, 5],
               RobinsonFoulds(tA[[5]], tB[[5]]),
               tolerance = 0)
})

test_that("RF cross-pairs similarity mode agrees with per-pair", {
  cross_sim <- RobinsonFoulds(tA, tB, similarity = TRUE)
  expect_equal(cross_sim[2, 3],
               RobinsonFoulds(tA[[2]], tB[[3]], similarity = TRUE),
               tolerance = 0)
})

test_that("RF cross-pairs normalization agrees with per-pair", {
  cross_norm <- RobinsonFoulds(tA, tB, normalize = TRUE)
  expect_equal(cross_norm[1, 4],
               RobinsonFoulds(tA[[1]], tB[[4]], normalize = TRUE),
               tolerance = 1e-10)
})

test_that("RF cross-pairs handles single-tree inputs via fallback", {
  # Single tree1 → falls back to CalculateTreeDistance
  d <- RobinsonFoulds(tA[[1]], tB)
  expect_equal(length(d), 5L)
  expect_equal(d[1], RobinsonFoulds(tA[[1]], tB[[1]]), tolerance = 0)
})

test_that("RF cross-pairs all-pairs matches cross-pairs diagonal", {
  all_dist <- as.matrix(RobinsonFoulds(trees20))
  # Cross-pairs of trees20 with itself should match all-pairs
  cross <- RobinsonFoulds(trees20, trees20)
  expect_equal(dim(cross), c(10L, 10L))
  for (i in 1:10) {
    for (j in 1:10) {
      expect_equal(cross[i, j], all_dist[i, j], tolerance = 0,
                   info = paste0("i=", i, " j=", j))
    }
  }
})

test_that("RF batch paths use heap allocation for large trees (>8192 tips)", {
  skip_on_cran()
  n <- 8193L
  tr <- list(TreeTools::RandomTree(n), TreeTools::RandomTree(n))
  class(tr) <- "multiPhylo"

  # All-pairs (hits S_heap resize in robinson_foulds_all_pairs)
  ap <- RobinsonFoulds(tr)
  expect_length(ap, 1L)
  expect_gte(ap, 0)

  # Cross-pairs (hits S_heap resize in robinson_foulds_cross_pairs)
  cp <- RobinsonFoulds(tr[1], tr[2])
  expect_equal(cp[1, 1], as.matrix(ap)[1, 2])
})

test_that("Issue #162: batch path agrees with per-pair for 33-taxon trees", {
  tree1 <- ape::read.tree(text = "(B,A,((AG,AF),((((C,(E,D)),((F,G),H)),((K,I),J)),((Q,R),((((AE,(AC,AD)),(AB,(N,(P,O)))),(((Y,(Z,AA)),(W,X)),(V,(T,U)))),((M,S),L))))));")
  tree2 <- ape::read.tree(text = "(B,A,((AG,AF),((((C,(E,D)),((F,G),H)),(((Q,R),((((AE,(AC,AD)),AB),(N,(P,O))),((Y,(Z,AA)),((V,(T,U)),(W,X))))),((M,L),S))),((K,I),J))));")
  tree3 <- PectinateTree(tree1[["tip.label"]])
  trees <- structure(list(tree1, tree2, tree3), class = "multiPhylo")

  # Batch all-pairs via fast path
  batch_cid <- ClusteringInfoDistance(trees)
  m <- as.matrix(batch_cid)

  # All three pairwise CIDs must match per-pair computation
  expect_equal(m[2, 1], ClusteringInfoDistance(tree1, tree2), tolerance = 1e-10)
  expect_equal(m[3, 1], ClusteringInfoDistance(tree1, tree3), tolerance = 1e-10)
  expect_equal(m[3, 2], ClusteringInfoDistance(tree2, tree3), tolerance = 1e-10)
})

test_that("64-bit splits handled ok", {
  SPI <- function(nTip) SharedPhylogeneticInfo(TreeTools::BalancedTree(nTip), 
                                               TreeTools::PectinateTree(nTip),
                                               normalize = pmin)
  expect_gt(SPI(64), SPI(63))
  expect_lt(SPI(64), SPI(65))
  
  expect_gt(SPI(128), SPI(127))
  expect_lt(SPI(128), SPI(129))
})

library("TreeTools", quietly = TRUE)
{ # Reference objects
  bal8 <- BalancedTree(8)
  pec8 <- PectinateTree(8)
  bal8BG <- DropTip(bal8, c(1, 8))
  pec8BG <- DropTip(pec8, c(1, 8))
  bal8CH <- DropTip(bal8, 1:2)
  pec8CH <- DropTip(pec8, 1:2)
}

test_that("Non-identical tips are handled okay", {
  fullDist <- TreeDistance(bal8, pec8)
  expect_equal(TreeDistance(bal8, bal8BG), 0)
  expect_equal(MutualClusteringInfo(bal8, bal8BG), ClusteringEntropy(bal8BG))
  expect_equal(TreeDistance(bal8BG, bal8), 0)
  expect_equal(MutualClusteringInfo(bal8BG, bal8), ClusteringEntropy(bal8BG))
  expect_equal(TreeDistance(bal8BG, bal8CH), 0)
  expect_equal(MutualClusteringInfo(bal8BG, bal8CH),
               ClusteringEntropy(DropTip(bal8, c(1, 2, 8))))
  expect_equal(TreeDistance(bal8, pec8BG), TreeDistance(bal8BG, pec8BG))
  expect_equal(TreeDistance(bal8BG, pec8), TreeDistance(bal8BG, pec8BG))
  expect_equal(MutualClusteringInfo(bal8, pec8BG),
               MutualClusteringInfo(bal8BG, pec8BG))
  expect_equal(TreeDistance(BalancedTree(1:5), BalancedTree(6:9)), NaN)
  expect_equal(MutualClusteringInfo(BalancedTree(1:5), BalancedTree(6:9)), 0)
  
  fullList <- list(
    bal8 = bal8,
    pec8 = pec8,
    
    bal8BG = bal8BG,
    pec8BG = pec8BG,
    bal8CH = bal8CH,
    pec8CH = pec8CH,
    
    balAE = BalancedTree(5),
    balFI = BalancedTree(paste0("t", 6:9))
  )
  
  expect_equal(MutualClusteringInfo(bal8, fullList),
               unlist(vapply(fullList, MutualClusteringInfo, 1, bal8)))
  expect_equal(MutualClusteringInfo(bal8, fullList, normalize = TRUE),
               unlist(vapply(fullList, MutualClusteringInfo, 1, bal8,
                             normalize = TRUE)))
  expect_equal(TreeDistance(bal8, fullList),
               unlist(lapply(fullList, TreeDistance, bal8)))
  expect_equal(TreeDistance(fullList, bal8), TreeDistance(bal8, fullList))
  expect_equal(MutualClusteringInfo(fullList, bal8),
               MutualClusteringInfo(bal8, fullList))
  expect_equal(MutualClusteringInfo(fullList, fullList),
               vapply(fullList, function(t1)
                 vapply(fullList, MutualClusteringInfo, double(1), t1),
                 double(length(fullList)))
               )
  expect_equal(MutualClusteringInfo(fullList),
               vapply(fullList, MutualClusteringInfo,
                      double(length(fullList)), fullList))
  expect_equal(MutualClusteringInfo(fullList, normalize = TRUE),
               vapply(fullList, MutualClusteringInfo,
                      double(length(fullList)), fullList, normalize = TRUE))
  expect_equal(# Lists of trees
    MutualClusteringInfo(fullList[2:3], fullList[1:2], normalize = TRUE),
    vapply(fullList[2:3], MutualClusteringInfo,
           double(2), fullList[1:2], normalize = TRUE)
  )
  expect_equal(
    MutualClusteringInfo(fullList[1:2], fullList[2], normalize = TRUE),
    t(MutualClusteringInfo(fullList[2], fullList[1:2], normalize = TRUE))
  )
  expect_equal(
    MutualClusteringInfo(fullList[2], fullList[1], normalize = TRUE),
    structure(MutualClusteringInfo(fullList[[2]], fullList[[1]],
                                   normalize = TRUE), dim = c(1, 1),
              dimnames = list(names(fullList[2]), names(fullList[1])))
  )
  
  # No overlap
  exp <- as.dist(vapply(fullList, TreeDistance, double(length(fullList)),
                        fullList), diag = FALSE, upper = FALSE)
  attr(exp, "call") <- NULL
  expect_equal(ClusteringInfoDistance(fullList, normalize = TRUE), exp)
  expect_equal(TreeDistance(fullList), exp)
})

library("TreeTools")
{
  # Labels in different order to confound as.Splits
  treeSym8 <- ape::read.tree(text="((e, (f, (g, h))), (((a, b), c), d));")
  treeBal8 <- ape::read.tree(text="(((e, f), (g, h)), ((a, b), (c, d)));")
  treeOpp8 <- ape::read.tree(text="(((a, f), (c, h)), ((g, b), (e, d)));")
  treesSBO8 <- structure(list(treeSym8, treeBal8, treeOpp8), 
                              class = "multiPhylo")
  treesSSBB8 <- structure(list(treeSym8, treeSym8, treeBal8, treeBal8), 
                              class = "multiPhylo")
  
  treeCat8 <- ape::read.tree(text="((((h, g), f), e), (d, (c, (b, a))));")
  treeTac8 <- ape::read.tree(text="((((e, c), g), a), (h, (b, (d, f))));")
  treeStar8 <- ape::read.tree(text="(e, c, g, h, b, a, d, f);")
  
  treeAb.Cdefgh <- ape::read.tree(text="((a, b), (c, d, e, f, g, h));")
  treeAbc.Defgh <- ape::read.tree(text="((a, b, c), (d, e, f, g, h));")
  treeAcd.Befgh <- ape::read.tree(text="((a, c, d), (b, e, f, g, h));")
  treeAbcd.Efgh <- ape::read.tree(text="((a, b, c, d), (e, f, g, h));")
  treeTwoSplits <- ape::read.tree(text="(((a, b), c, d), (e, f, g, h));")
  
  testTrees <- c(treesSBO8, treeCat8, treeTac8, treeStar8, treeAb.Cdefgh,
                 treeAbc.Defgh, treeAbcd.Efgh, treeAcd.Befgh, treeTwoSplits)

  # Matching report matches score
  expect_matching_score <- function(Func) {
    mtch <- Func(BalancedTree(33), PectinateTree(33), reportMatching = TRUE)
    expect_equal(mtch[[1]], Func(BalancedTree(33), PectinateTree(33)))
    matching <- attr(mtch, "matching")
    matchScores <- vapply(seq_along(matching),
           function(i) attr(mtch, "pairScores")[i, matching[[i]]],
           double(1)) # inelegant, but validates cbind logic in function
    expect_equal(matchScores, attr(mtch, "matchedScores"))
    expect_equal(mtch[[1]], sum(matchScores))
  }
}

test_that("Split compatibility is correctly established", {
  expect_true(SplitsCompatible(as.logical(c(0,0,1,1,0)), 
                               as.logical(c(0,0,1,1,0))))
  expect_true(SplitsCompatible( as.logical(c(0,0,1,1,0)), 
                               !as.logical(c(0,0,1,1,0))))
  expect_true(SplitsCompatible(as.logical(c(0,0,1,1,0)), 
                               as.logical(c(1,0,1,1,0))))
  expect_true(SplitsCompatible(!as.logical(c(0,0,1,1,0)), 
                                as.logical(c(1,0,1,1,0))))
  expect_false(SplitsCompatible(as.logical(c(0,0,1,1,0)), 
                                as.logical(c(1,1,0,1,0))))
})

methodsToTest <- list(
  SharedPhylogeneticInfo,
  DifferentPhylogeneticInfo,
  MatchingSplitInfo,
  MatchingSplitInfoDistance,
  MutualClusteringInfo,
  ClusteringInfoDistance,
  NyeSimilarity,
  JaccardRobinsonFoulds,
  MatchingSplitDistance,
  RobinsonFoulds,
  InfoRobinsonFoulds,
  KendallColijn # List last: requires rooted trees.
)

NormalizationTest <- function(FUNC, ...) {
  expect_equal(c(1L, 1L), 
               FUNC(treesSSBB8, normalize = TRUE, ...)[c(1, 6)],
               tolerance = 1e-7)
}

test_that("Size mismatch causes error", {
  treeSym7 <- ape::read.tree(text="((e, (f, g)), (((a, b), c), d));")
  splits7 <- as.Splits(treeSym7)
  splits8 <- as.Splits(treeSym8)
  
  expect_error(MeilaVariationOfInformation(splits7, splits8),
               "Split lengths differ")
  
  # Validation is now in R (.ValidateSplitArgs), not C++
  splits9 <- as.Splits(BalancedTree(9))
  expect_error(.ValidateSplitArgs(splits8, splits9, 8L),
               "same number of tips")
})

test_that("Metrics handle polytomies", {
  polytomy8 <- ape::read.tree(text="(a, b, c, d, e, f, g, h);")
  lapply(list(SharedPhylogeneticInfo, MutualClusteringInfo,
              MatchingSplitDistance, NyeSimilarity),
         function(Func) expect_equal(0, Func(treeSym8, polytomy8)))
})

test_that(".AllTipsSame()", {
  expect_true(.AllTipsSame(1:2, NULL))
  expect_true(.AllTipsSame(list(1:2, 1:2), NULL))
  expect_false(.AllTipsSame(list(1:2, 3:2), NULL))
  expect_true(.AllTipsSame(list(1:2, 1:2, 1:2), NULL))
  expect_false(.AllTipsSame(list(1:2, 3:2, 1:2), NULL))
  expect_false(.AllTipsSame(list(3:2, 1:2, 1:2), NULL))
  expect_true(.AllTipsSame(1:2, 1:2))
  expect_false(.AllTipsSame(1:2, 1))
  
  expect_true(.AllTipsSame(list(1:2), list(1:2)))
  expect_true(.AllTipsSame(1:2, list(1:2)))
  expect_true(.AllTipsSame(1:2, 1:2))
  expect_true(.AllTipsSame(list(1:2), 1:2))
  
  expect_false(.AllTipsSame(list(1:2), list(3:2)))
  expect_false(.AllTipsSame(1:2, list(3:2)))
  expect_false(.AllTipsSame(1:2, 3:2))
  expect_false(.AllTipsSame(list(1:2), 3:2))
  
  expect_true(.AllTipsSame(1:2, list(1:2, 1:2)))
  expect_true(.AllTipsSame(list(1:2), list(1:2, 1:2)))
  expect_true(.AllTipsSame(list(1:2, 1:2), list(1:2, 1:2)))
  expect_true(.AllTipsSame(list(1:2, 1:2), 1:2))
  
  expect_false(.AllTipsSame(1:3, list(1:2, 1:3)))
  expect_false(.AllTipsSame(1:3, list(1:3, 1:2)))
  expect_false(.AllTipsSame(list(1:3), list(1:3, 1:2)))
  
  expect_true(.AllTipsSame(1:4, list(a = 1:4, b = 4:1, c = c(4L, 1L, 2L, 3L))))
})

test_that(".MaxValue() succeeds", {
  list1 <- list(sym = treeSym8, bal = treeBal8)
  list2 <- list(sym = treeSym8, abc = treeAbc.Defgh, abcd = treeAbcd.Efgh)
  dimNames <- list(names(list1), names(list2))
  
  expect_equal(
    MutualClusteringInfo(list1, list2[[2]], normalize = FALSE),
    c(sym = MutualClusteringInfo(treeSym8, treeAbc.Defgh, normalize = FALSE),
      bal = MutualClusteringInfo(treeBal8, treeAbc.Defgh, normalize = FALSE))
  )
  expect_equal(
    as.double(t(MutualClusteringInfo(list1, list2, normalize = FALSE))),
    as.double(c(MutualClusteringInfo(list1[[1]], list2, normalize = FALSE),
                MutualClusteringInfo(list1[[2]], list2, normalize = FALSE)))
  )
  
  expect_equal(
    .MaxValue(list1, list2[[2]], Value = ClusteringEntropy),
    c(sym = .MaxValue(treeSym8, treeAbc.Defgh, Value = ClusteringEntropy),
      bal = .MaxValue(treeBal8, treeAbc.Defgh, Value = ClusteringEntropy))
  )
  expect_equal(
    as.double(t(.MaxValue(list1, list2, ClusteringEntropy))),
    as.double(c(.MaxValue(list1[[1]], list2, ClusteringEntropy),
                .MaxValue(list1[[2]], list2, ClusteringEntropy)))
  )
  
  expect_equal(.MaxValue(list1[[1]], NULL, ClusteringEntropy), double(0))
  
  expect_equal(.MaxValue(list1, NULL, ClusteringEntropy),
               sum(ClusteringEntropy(list1)))
  
  expect_equal(.MaxValue(list2, NULL, ClusteringEntropy),
               c(.MaxValue(list2[-3], NULL, ClusteringEntropy),
                 .MaxValue(list2[-2], NULL, ClusteringEntropy),
                 .MaxValue(list2[-1], NULL, ClusteringEntropy)
                 ))
})

#Func <- ClusteringInfoDistance # FUNC =
test_that("Output dimensions are correct", {
  list1 <- list(sym = treeSym8, bal = treeBal8)
  list2 <- list(sym = treeSym8, abc = treeAbc.Defgh, abcd = treeAbcd.Efgh)
  dimNames <- list(names(list1), names(list2))
  
  Test <- function(Func) {
    allPhylo <- matrix(
      c(Func(treeSym8, treeSym8),      Func(treeBal8, treeSym8),
        Func(treeSym8, treeAbc.Defgh), Func(treeBal8, treeAbc.Defgh),
        Func(treeSym8, treeAbcd.Efgh), Func(treeBal8, treeAbcd.Efgh)),
      2L, 3L, dimnames = dimNames)
    phylo1 <- matrix(c(Func(treeSym8, list2), Func(treeBal8, list2)),
                     byrow = TRUE, 2L, 3L, dimnames = dimNames)
    phylo2 <- matrix(c(Func(list1, treeSym8), Func(list1, treeAbc.Defgh),
                       Func(list1, treeAbcd.Efgh)), 2L, 3L, dimnames = dimNames)
    noPhylo <- Func(list1, list2)
    expect_equal(allPhylo, phylo1)
    expect_equal(allPhylo, phylo2)
    expect_equal(allPhylo, noPhylo)
  }
  
  lapply(methodsToTest, Test)
})

test_that("Small trees work", {
  TestDist <- function(Func) {
    expect_equal(Func(BalancedTree(4), BalancedTree(4)), 0)
    expect_gt(Func(BalancedTree(1:4), BalancedTree(c(1, 3, 2, 4))), 0)
  }
  lapply(list(DifferentPhylogeneticInfo, MatchingSplitInfoDistance,
              ClusteringInfoDistance, MatchingSplitDistance,
              RobinsonFoulds, InfoRobinsonFoulds), TestDist)

  TestSim <- function(Func) {
    expect_gt(Func(BalancedTree(4), BalancedTree(4)), 0)
    expect_equal(Func(BalancedTree(1:4), BalancedTree(c(1, 3, 2, 4))), 0)
  }
  lapply(list(SharedPhylogeneticInfo, MatchingSplitInfo, MutualClusteringInfo),
         TestSim)
})

test_that("RF Distance is correctly calculated", {
  skip_if_not_installed("phangorn")
  PhangRF2 <- function(t1, t2) phangorn::RF.dist(reorder(t1, "cladewise"),
                                                 reorder(t2, "cladewise"))
  RFTest <- function(t1, t2) {
    expect_equal(suppressMessages(PhangRF2(t1, t2)),
                 RobinsonFoulds(t1, t2))
    
    expected <- RobinsonFoulds(t1, t2, reportMatching = TRUE, similarity = TRUE)
    attr(expected, "pairScores") <- attr(expected, "pairScores") == 0L
    expect_equal(expected, RobinsonFouldsMatching(t1, t2))
  }
  RFTest(treeSym8, treeSym8)
  RFTest(treeSym8, treeStar8)
  RFTest(treeStar8, treeStar8)
  RFTest(treeAb.Cdefgh, treeAbc.Defgh)
  RFTest(treeAb.Cdefgh, treeAbcd.Efgh)
  
  # at 2020-10, RF uses Day algorithm if tree2 = null; old algo if tree2 = tree1.
  expect_equal(RobinsonFoulds(testTrees, testTrees),
               as.matrix(RobinsonFoulds(testTrees)),
               ignore_attr = TRUE)
  
  # Invariant to tree description order
  sq_pectinate <- ape::read.tree(text="((((((1, 2), 3), 4), 5), 6), (7, (8, (9, (10, 11)))));")
  shuffle1 <- ape::read.tree(text="(((((1, 5), 2), 6), (3, 4)), ((8, (7, 9)), (10, 11)));")
  shuffle2 <- ape::read.tree(text="(((8, (7, 9)), (10, 11)), ((((1, 5), 2), 6), (3, 4)));")
  RFTest(shuffle1, sq_pectinate)
  RFTest(sq_pectinate, shuffle1)
  RFTest(shuffle1, shuffle2)
  RFTest(shuffle1, sq_pectinate)
  RFTest(shuffle2, sq_pectinate)
})

test_that("Shared Phylogenetic Info is correctly calculated", {
  expect_equal(5.529821, tolerance = 1e-7,
               cpp_shared_phylo(
                 as.Splits(as.logical(c(1, 1, 1, 1, 0, 0, 0, 0))),
                 as.Splits(as.logical(c(1, 1, 1, 1, 0, 0, 0, 0))),
                 8L)$score)
  
  expect_equal(0.2895066, tolerance = 1e-7,
               cpp_shared_phylo(
                 as.Splits(as.logical(c(1, 1, 0, 0, 0, 0, 0, 0))),
                 as.Splits(as.logical(c(0, 0, 1, 1, 0, 0, 0, 0))),
                 8L)$score)
  
  expect_equal(1.137504, tolerance = 1e-6,
               cpp_shared_phylo(
                 as.Splits(as.logical(c(1, 1, 0, 0, 0, 0, 0, 0))),
                 as.Splits(as.logical(c(1, 1, 1, 1, 0, 0, 0, 0))),
                 8L)$score)
  
  expect_equal(3.45943, tolerance = 1e-6,
               cpp_shared_phylo(
                 as.Splits(as.logical(c(1, 1, 0, 0, 0, 0, 0, 0))),
                 as.Splits(as.logical(c(1, 1, 0, 0, 0, 0, 0, 0))),
                 8L)$score)
  
  expect_equal(22.53747, tolerance = 1e-05,
               SharedPhylogeneticInfo(treeSym8, treeSym8, normalize = FALSE))
  
  expect_equal(1, tolerance = 1e-05,
               SharedPhylogeneticInfo(treeSym8, treeSym8, normalize = TRUE))
  
  expect_equal(0,
               SharedPhylogeneticInfo(treeSym8, treeStar8, normalize = TRUE))

  expect_equal(0,
               SharedPhylogeneticInfo(treeStar8, treeStar8, normalize = FALSE))

  expect_equal(NaN, # Division by zero
               SharedPhylogeneticInfo(treeStar8, treeStar8, normalize = TRUE))
  
  expect_equal(13.75284, SharedPhylogeneticInfo(treeSym8, treeBal8), tolerance=1e-05)
  
  expect_equal(DifferentPhylogeneticInfo(treeSym8, treeAcd.Befgh),
               DifferentPhylogeneticInfo(treeAcd.Befgh, treeSym8), tolerance=1e-05)
  
  expect_equal(0, DifferentPhylogeneticInfo(treeSym8, treeSym8, normalize = TRUE))
  
  infoSymBal <- SplitwiseInfo(treeSym8) + SplitwiseInfo(treeBal8)
  expect_equal(infoSymBal - 13.75284 - 13.75284, tolerance = 1e-05,
    DifferentPhylogeneticInfo(treeSym8, treeBal8, normalize = TRUE) * infoSymBal)
  
  expect_equal(22.53747 + SharedPhylogeneticInfo(treeAcd.Befgh, treeAcd.Befgh) - 
                 (2 * SharedPhylogeneticInfo(treeSym8, treeAcd.Befgh)), 
               DifferentPhylogeneticInfo(treeSym8, treeAcd.Befgh), 
               tolerance=1e-06)
  
  expect_equal(-log2(945/10395), 
               SharedPhylogeneticInfo(treeSym8, treeAb.Cdefgh),
               tolerance = 1e-06)
  
  expect_equal(22.53747 + SharedPhylogeneticInfo(treeBal8, treeBal8) - 13.75284 - 13.75284, 
               DifferentPhylogeneticInfo(treeSym8, treeBal8), tolerance=1e-05)
  
  expect_equal(-log2(945/10395),
               SharedPhylogeneticInfo(treeSym8, treeAb.Cdefgh),
               tolerance = 1e-06)
  
  expect_equal(-log2(315/10395),
               SharedPhylogeneticInfo(treeSym8, treeAbc.Defgh),
               tolerance = 1e-06)
  
  expect_equal(0, DifferentPhylogeneticInfo(treeSym8, treeSym8))
  
  expect_equal(SplitwiseInfo(treeSym8) - SplitwiseInfo(treeAcd.Befgh),
               DifferentPhylogeneticInfo(treeSym8, treeAbc.Defgh),
               tolerance = 1e-06)
  
  
  # Test symmetry of small vs large splits
  expect_equal(SharedPhylogeneticInfo(treeSym8, treeAbc.Defgh),
               SharedPhylogeneticInfo(treeAbc.Defgh, treeSym8))
  expect_equal(-log2(225/10395), SharedPhylogeneticInfo(treeSym8, treeAbcd.Efgh))
  expect_equal(-log2(225/10395) - log2(945/10395),
               SharedPhylogeneticInfo(treeSym8, treeTwoSplits),
               tolerance = 1e-7)
  expect_equal(SplitSharedInformation(8, 4, 3),
               SharedPhylogeneticInfo(treeTwoSplits, treeAbc.Defgh),
               tolerance = 1e-7)
  expect_equal(SplitInformation(4, 4) + SplitInformation (3, 5) - 
               (2 * SplitSharedInformation(8, 4, 3)),
               SplitDifferentInformation(8, 4, 3),
               tolerance=1e-07)
  
  expect_equal(SharedPhylogeneticInfo(treeSym8, list(treeSym8, treeBal8)), 
               SharedPhylogeneticInfo(list(treeSym8, treeBal8), treeSym8),
               tolerance = 1e-7)
  
  # Test tree too large to cache
  set.seed(101)
  t1 <- ape::rtree(101)
  t2 <- ape::rtree(101, rooted = FALSE)
  expect_equal(SharedPhylogeneticInfo(t1, t2),
               SharedPhylogeneticInfo(t2, t1))
})

test_that("MatchingSplitInfo() is correctly calculated", {
  BinaryToSplit <- function(binary) matrix(as.logical(binary))
  expect_equal(log2(3),
               MatchingSplitInfoSplits(
                 as.Splits(c(rep(TRUE, 2), rep(FALSE, 6))),
                 as.Splits(c(FALSE, FALSE, rep(TRUE, 2), rep(FALSE, 4)))),
               tolerance = 1e-7)
  expect_equal(log2(3),
               MatchingSplitInfoSplits(
                 as.Splits(c(rep(FALSE, 6), rep(TRUE, 2))),
                 as.Splits(c(FALSE, FALSE, rep(TRUE, 2), rep(FALSE, 4)))),
               tolerance = 1e-7)
  expect_equal(log2(3), cpp_msi_distance(
    as.Splits(c(rep(TRUE, 2), rep(FALSE, 6))),
    as.Splits(c(FALSE, FALSE, rep(TRUE, 2), rep(FALSE, 4))),
    8L)$score, tolerance = 1e-7)
  expect_equal(log2(3), cpp_msi_distance(
    as.Splits(rep(c(FALSE, TRUE), each = 4L)),
    as.Splits(rep(c(FALSE, TRUE), 4L)),
    8L)$score, tolerance = 1e-7)
  
  
  expect_equal(SharedPhylogeneticInfo(treeSym8, treeSym8),
               MatchingSplitInfo(treeSym8, treeSym8), tolerance = 1e-05)
  
  expect_equal(0, MatchingSplitInfo(treeSym8, treeStar8))
  expect_equal(0, MatchingSplitInfo(treeStar8, treeStar8))
  
  expect_equal(MatchingSplitInfo(treeAb.Cdefgh, treeAbc.Defgh),
               MatchingSplitInfo(treeAbc.Defgh, treeAb.Cdefgh))
  expect_equal(MatchingSplitInfo(treeAbcd.Efgh, treeAb.Cdefgh),
               MatchingSplitInfo(treeAb.Cdefgh, treeAbcd.Efgh))
  expect_equal(-(TreeTools::Log2TreesMatchingSplit(2, 5) - Log2Unrooted.int(7)), 
               MatchingSplitInfo(treeAb.Cdefgh, treeAbc.Defgh),
               tolerance = 1e-7)
  expect_true(MatchingSplitInfo(treeSym8, treeBal8) > 
                MatchingSplitInfo(treeSym8, treeOpp8))
  expect_equal(0, MatchingSplitInfoDistance(treeSym8, treeSym8))
  NormalizationTest(MatchingSplitInfo)
})

test_that("Shared Phylogenetic Information is correctly estimated", {
  exp <- ExpectedVariation(treeSym8, treeAbc.Defgh, samples = 1000L)
  tol <- exp[, "Std. Err."] * 2
  # Expected values calculated with 100k samples
  expect_equal(1.175422, exp["SharedPhylogeneticInfo", "Estimate"], 
               tolerance = tol[1])
  expect_equal(3.099776, exp["MatchingSplitInfo", "Estimate"], 
               tolerance = tol[2])
  expect_equal(25.231023, exp["DifferentPhylogeneticInfo", "Estimate"], 
               tolerance = tol[3])
  expect_equal(21.382314, exp["MatchingSplitInfoDistance", "Estimate"], 
               tolerance = tol[4])
  expect_equal(exp[, "sd"], exp[, "Std. Err."] * sqrt(exp[, "n"]))
})

test_that("Clustering information is correctly calculated", {
  expect_equal(Entropy(c(3, 5) / 8) * 2 - Entropy(c(0, 0, 3, 5) / 8),
               cpp_mutual_clustering(
                 as.Splits(as.logical(c(1, 1, 1, 0, 0, 0, 0, 0))),
                 as.Splits(as.logical(c(1, 1, 1, 0, 0, 0, 0, 0))),
                 8L)$score,
               tolerance = 1e-7)
  
  expect_equal(Entropy(c(2, 6) / 8) * 2 - Entropy(c(0, 2, 2, 4) / 8),
               cpp_mutual_clustering(
                 as.Splits(as.logical(c(1, 1, 0, 0, 0, 0, 0, 0))),
                 as.Splits(as.logical(c(0, 0, 1, 1, 0, 0, 0, 0))),
                 8L)$score, tolerance = 1e-7)
  
  expect_equal(Entropy(c(5, 4) / 9) + Entropy(c(3, 6) / 9) -
                      Entropy(c(3, 2, 0, 4) / 9),
               cpp_mutual_clustering(
                 as.Splits(as.logical(c(1, 1, 1, 1, 1, 0, 0, 0, 0))),
                 as.Splits(as.logical(c(0, 0, 1, 1, 1, 0, 0, 0, 0))),
                 9L)$score,
               tolerance = 1e-7)
  
  expect_equal(Entropy(c(4, 4) / 8) * 2 - Entropy(c(2, 2, 2, 2) / 8),
               cpp_mutual_clustering(
                 as.Splits(as.logical(c(1, 1, 1, 1, 0, 0, 0, 0))),
                 as.Splits(as.logical(c(1, 0, 1, 0, 1, 0, 1, 0))),
                 8L)$score,
               tolerance = 1e-7)
  
  expect_equal(Entropy(c(4, 4) / 8) * 2 - Entropy(c(0, 0, 4, 4) / 8),
               cpp_mutual_clustering(
                 as.Splits(as.logical(c(1, 1, 1, 1, 0, 0, 0, 0))),
                 as.Splits(as.logical(c(1, 1, 1, 1, 0, 0, 0, 0))),
                 8L)$score, 
               tolerance = 1e-7)
  
  expect_equal(ClusteringEntropy(treeSym8),
               MutualClusteringInfo(treeSym8, treeSym8),
               tolerance = 1e-05)
  expect_equal(8 * ClusteringEntropy(treeSym8), ClusteringInfo(treeSym8))
  expect_equal(0, MutualClusteringInfo(treeSym8, treeStar8))
  expect_equal(0, MutualClusteringInfo(treeStar8, treeStar8))
  
  expect_equal(TreeDistance(treeSym8, treeBal8),
               ClusteringInfoDistance(treeSym8, treeBal8, normalize = TRUE))
  expect_equal(1, MutualClusteringInfo(treeSym8, treeSym8, normalize = TRUE),
               tolerance = 1e-7)
  expect_true(MutualClusteringInfo(treeSym8, treeBal8, normalize = pmin) >
                MutualClusteringInfo(treeSym8, treeBal8, normalize = pmax))
  expect_equal(ClusteringEntropy(treeSym8) + ClusteringEntropy(treeBal8) -
                 (2 * MutualClusteringInfo(treeBal8, treeSym8)),
               ClusteringInfoDistance(treeSym8, treeBal8), tolerance = 1e-05)
  expect_equal(MutualClusteringInfo(treeAb.Cdefgh, treeAbc.Defgh),
               MutualClusteringInfo(treeAbc.Defgh, treeAb.Cdefgh),
               tolerance = 1e-05)
  
  zeroTree <- TreeTools::ZeroTaxonTree()
  expect_equal(MutualClusteringInfo(zeroTree, zeroTree), 0)
  expect_equal(ClusteringInfoDistance(zeroTree, zeroTree), 0)
  
  
  
  # Different resolution
  randomBif20 <- structure(list(
    edge = structure(c(21L, 21L, 22L, 23L, 24L, 25L, 26L, 27L, 28L, 29L, 30L,
                       31L, 32L, 32L, 31L, 30L, 29L, 33L, 34L, 34L, 33L, 28L, 
                       35L, 36L, 36L, 35L, 27L, 26L, 37L, 37L, 25L, 38L, 38L,
                       39L, 39L, 24L, 23L, 22L, 1L, 22L, 23L, 24L, 25L, 26L, 
                       27L, 28L, 29L, 30L, 31L, 32L, 2L, 14L, 7L, 10L, 33L, 34L,
                       4L, 6L, 8L, 35L, 36L, 13L, 16L, 18L, 17L, 37L, 5L, 15L,
                       38L, 11L, 39L, 12L, 19L, 9L, 3L, 20L),
                     .Dim = c(38L, 2L)), Nnode = 19L,
    tip.label = c("t1", "t2", "t3", "t4", "t5", "t6", "t7", "t8", "t9", "t10",
                  "t11", "t12", "t13", "t14", "t15", "t16", "t17", "t18", "t19",
                  "t20"), br = NULL), class = "phylo")
  # NOT in Preorder.  Preordering the tree will change the matching.
  threeAwayPoly <- structure(
    list(edge = structure(c(21L, 22L, 23L, 24L, 25L, 26L, 27L, 28L, 29L, 29L, 
                            28L, 27L, 26L, 30L, 30L, 30L, 26L, 31L, 31L, 25L, 
                            32L, 33L, 33L, 32L, 25L, 25L, 24L, 34L, 34L, 34L,
                            23L, 22L, 21L, 22L, 23L, 24L, 25L, 26L, 27L, 28L,
                            29L, 2L, 8L, 14L, 10L, 30L, 13L, 16L, 18L, 31L, 4L,
                            6L, 32L, 33L, 15L, 20L, 5L, 7L, 17L, 34L, 11L, 12L,
                            19L, 9L, 3L, 1L), .Dim = c(33L, 2L)),
         tip.label = c("t1", "t2", "t3", "t4", "t5", "t6", "t7", "t8", "t9",
                       "t10", "t11", "t12", "t13", "t14", "t15", "t16", "t17",
                       "t18", "t19", "t20"),
         Nnode = 14L), class = "phylo")
  expect_equal(
    MutualClusteringInfo(threeAwayPoly, randomBif20),
    MutualClusteringInfo(randomBif20, threeAwayPoly))
  match <- MutualClusteringInfo(randomBif20, threeAwayPoly, 
                                reportMatching = TRUE)
  expect_equal(sum(attributes(match)$matchedScores, na.rm = TRUE), match[[1]])
  expect_equal(attr(match, "matching"),
               c(NA, 1L, 2L, 3L, 11L, 4L, 7L, 5L, NA, 6L, NA, 9L, 8L, NA, 10L, 
                 12L, NA)) # Expectation obtained by observation
  
  # Multiple bins, calculated expectation
  library("TreeTools", quietly = TRUE)
  b65m <- lapply(c(1, 2, 70), AddTip, tree = BalancedTree(64))
  self <- ClusteringEntropy(b65m)
  diff <- ClusteringEntropy(b65m[[1]], sum = FALSE)["72"]
  # Copied from C:
  ic_element <- function(nkK, nk, nK, n) {
    if (nkK && nk && nK) {
      if (nkK == nk && nkK == nK && nkK + nkK == n) return (nkK);
      numerator = nkK * n
      denominator = nk * nK
      if (numerator == denominator) return (0);
      nkK * (log2(numerator) - log2(denominator));
    } else 0;
  }
  expect_equal(diff,
               (ic_element(63, 63, 63, 65) +
                ic_element(00, 63, 02, 65) +
                ic_element(00, 02, 63, 65) +
                ic_element(02, 02, 02, 65)) / 65,
               ignore_attr = TRUE)
  new <- (ic_element(65-3, 63, 63, 65) +
          ic_element(1, 63, 02, 65) +
          ic_element(1, 02, 63, 65) +
          ic_element(1, 02, 02, 65)) / 65
  other <- self[1] - diff[1] + new # Calc'd = 20.45412
  expect_equal(other, MutualClusteringInfo(b65m[[1]], b65m[[2]]),
               ignore_attr = TRUE)
  
  expectation <- matrix(other, 3, 3)
  diag(expectation) <- self
  expect_equal(expectation, MutualClusteringInfo(b65m), ignore_attr = TRUE)
  
  
  expect_equal(ClusteringEntropy(BalancedTree(64)),
               MutualClusteringInfo(BalancedTree(64), BalancedTree(64)))
  expect_equal(ClusteringEntropy(BalancedTree(644)),
               MutualClusteringInfo(BalancedTree(644), BalancedTree(644)))
  
  expect_gt(ClusteringEntropy(BalancedTree(64)),
            MutualClusteringInfo(BalancedTree(64), PectinateTree(64)))
  expect_gt(ClusteringEntropy(BalancedTree(644)),
            MutualClusteringInfo(BalancedTree(644), PectinateTree(644)))
  
  NormalizationTest(MutualClusteringInfo)
})

test_that("Matchings are correct", {
  
  # Different resolution: used to cause memory leak
  randomBif20 <- structure(list(
    edge = structure(c(21L, 21L, 22L, 23L, 24L, 25L, 26L, 27L, 28L, 29L, 30L,
                       31L, 32L, 32L, 31L, 30L, 29L, 33L, 34L, 34L, 33L, 28L, 
                       35L, 36L, 36L, 35L, 27L, 26L, 37L, 37L, 25L, 38L, 38L,
                       39L, 39L, 24L, 23L, 22L, 1L, 22L, 23L, 24L, 25L, 26L, 
                       27L, 28L, 29L, 30L, 31L, 32L, 2L, 14L, 7L, 10L, 33L, 34L,
                       4L, 6L, 8L, 35L, 36L, 13L, 16L, 18L, 17L, 37L, 5L, 15L,
                       38L, 11L, 39L, 12L, 19L, 9L, 3L, 20L),
                     .Dim = c(38L, 2L)), Nnode = 19L,
    tip.label = c("t1", "t2", "t3", "t4", "t5", "t6", "t7", "t8", "t9", "t10",
                  "t11", "t12", "t13", "t14", "t15", "t16", "t17", "t18", "t19",
                  "t20"), br = NULL), class = "phylo")
  threeAwayPoly <- structure(
    list(edge = structure(c(21L, 22L, 23L, 24L, 25L, 26L, 27L, 28L, 29L, 29L, 
                            28L, 27L, 26L, 30L, 30L, 30L, 26L, 31L, 31L, 25L, 
                            32L, 33L, 33L, 32L, 25L, 25L, 24L, 34L, 34L, 34L,
                            23L, 22L, 21L, 22L, 23L, 24L, 25L, 26L, 27L, 28L,
                            29L, 2L, 8L, 14L, 10L, 30L, 13L, 16L, 18L, 31L, 4L,
                            6L, 32L, 33L, 15L, 20L, 5L, 7L, 17L, 34L, 11L, 12L,
                            19L, 9L, 3L, 1L), .Dim = c(33L, 2L)),
         tip.label = c("t1", "t2", "t3", "t4", "t5", "t6", "t7", "t8", "t9",
                       "t10", "t11", "t12", "t13", "t14", "t15", "t16", "t17",
                       "t18", "t19", "t20"),
         Nnode = 14L), class = "phylo")
  
  expect_equal(
    MutualClusteringInfo(threeAwayPoly, randomBif20),
    MutualClusteringInfo(randomBif20, threeAwayPoly))
  
  library("TreeTools", quietly = TRUE)
  t1 <- PectinateTree(letters[1:11])
  t2 <- ape::read.tree(text = "(a, (c, (b, (d, e, ((g, h, f), (k, (j, i)))))));")
  t3 <- CollapseNode(PectinateTree(c(letters[11], letters[1:10])), 16:19)
  s0 <- as.Splits(ZeroTaxonTree())
  s1 <- as.Splits(t1)
  s2 <- as.Splits(t2, t1)
  s3 <- as.Splits(t3, t1)
  n1 <- dim(s1)[1]
  n2 <- dim(s2)[1]
  n3 <- dim(s3)[1]
  n <- NTip(s1)
  
  # Plot
  # par(mfrow = 2:1, cex = 0.9, mar = rep(0,4))
  # JRF2T <- function(...) JaccardRobinsonFoulds(..., k = 2)
  # JRF2F <- function(...) JaccardRobinsonFoulds(..., k = 2, allowConflict = FALSE)
  # VisualizeMatching(MatchingSplitDistance, t1, t2, setPar=F)
  # LabelSplits(t2, setNames(1:6, names(s2)), adj = 2)
  # VisualizeMatching(MatchingSplitDistance, t2, t1, setPar=F)
  # LabelSplits(t1, setNames(1:8, names(s1)), adj = 2)
  
  
  Test <- function(CppFn, x12, x21, ...) {
    
    r0 <- CppFn(s0, s0, 0, ...)
    expect_equal(r0$score, 0)
    
    r12 <- CppFn(s1, s2, n, ...)
    r21 <- CppFn(s2, s1, n, ...)
    r13 <- CppFn(s1, s3, n, ...)
    r31 <- CppFn(s3, s1, n, ...)
    expect_equal(r12$score, r21$score)
    expect_equal(r13$score, r31$score)
    
    m12 <- r12$matching
    m21 <- r21$matching
    expect_equal(n1, length(m12))
    expect_equal(length(m12[!is.na(m12)]), length(unique(m12[!is.na(m12)])))
    expect_equal(n2, length(m21))
    expect_equal(length(m21[!is.na(m21)]), length(unique(m21[!is.na(m21)])))
    expect_lte(dim(s1)[1] - dim(s2)[1], sum(is.na(m12)))
    
    m13 <- r13$matching
    m31 <- r31$matching
    expect_equal(n1, length(m13))
    expect_equal(length(m13[!is.na(m13)]), length(unique(m13[!is.na(m13)])))
    expect_equal(n3, length(m31))
    expect_equal(length(m31[!is.na(m31)]), length(unique(m31[!is.na(m31)])))
    expect_lte(dim(s1)[1] - dim(s3)[1], sum(is.na(m13)))
    
    for (i in seq_along(m12)) expect_true(m12[[i]] %in% x12[[i]])
    for (i in seq_along(m21)) expect_true(m21[[i]] %in% x21[[i]])
    
  }
  
  # Expectations by observation
  Test(TreeDist:::cpp_robinson_foulds_distance,
       list(NA, 2, NA, 3, NA, NA, 5, NA),
       list(NA, 2, 4, NA, 7, NA)
       )
  Test(TreeDist:::cpp_robinson_foulds_info,
       list(NA, 2, NA, 3, NA, NA, 5, NA),
       list(NA, 2, 4, NA, 7, NA)
       )
  Test(TreeDist:::cpp_matching_split_distance,
       list(1, 2, 4, 3, NA, NA, 5, 6),
       list(1, 2, 5, 4, 7, 6)
       )
  Test(TreeDist:::cpp_jaccard_similarity,
       list(NA, 2, 1, 3, 4, 6, 5, NA),
       list(3, 2, 4, 5, 7, 6),
       k = 2,
       allowConflict = TRUE)
  Test(TreeDist:::cpp_jaccard_similarity,
       list(NA, 2, 1, 3, NA, 6, 5, 4),
       list(3, 2, 4, 1, 7, 6),
       k = 2,
       allowConflict = FALSE)
  Test(TreeDist:::cpp_msi_distance,
       list(NA, 2, 1, 4, 3, 6, 5, NA),
       list(3, 2, c(4, 5), c(4, 5), c(6, 7), c(7, 6))
       )
  Test(TreeDist:::cpp_shared_phylo,
       list(NA, 2, 4, 3, 1, 6, 5, NA),
       list(5, 2, 4, 3, 7, 6)
       )
  Test(TreeDist:::cpp_mutual_clustering, 
       list(c(NA, 4), 2, 1, 3, NA, 6, 5, c(NA, 4)),
       list(3, 2, 4, c(1, 8), 7, 6)
       )
})

test_that("Matching Split Distance is correctly calculated", {
  expect_equal(MatchingSplitDistance(treeSym8, treeSym8), 0L)
  expect_equal(MatchingSplitDistance(treeStar8, treeSym8), 0L)
  expect_equal(MatchingSplitDistance(treeStar8, treeStar8), 0L)
  match0 <- MatchingSplitDistance(treeStar8, treeStar8, reportMatching = TRUE)
  expect_equal(c(match0, vapply(attributes(match0), length, 0)),
               rep(0L, 5), ignore_attr = TRUE)
  expect_equal(MatchingSplitDistance(treeAb.Cdefgh, treeAbc.Defgh), 1L)
  expect_equal(MatchingSplitDistance(treeAb.Cdefgh, treeAbcd.Efgh), 2L)
  
  splitAB <- as.Splits(c(rep(TRUE, 2), rep(FALSE, 7)))
  splitABC <- as.Splits(c(rep(TRUE, 3), rep(FALSE, 6)))
  splitAEF <- as.Splits(c(TRUE, rep(FALSE, 3), TRUE, TRUE, rep(FALSE, 3)))
  splitABCD <- as.Splits(c(rep(TRUE, 4), rep(FALSE, 5)))
  splitABCDE <- as.Splits(c(rep(TRUE, 5), rep(FALSE, 4)))
  splitAI <- as.Splits(c(TRUE, rep(FALSE, 7), TRUE))
  
  expect_equal(MatchingSplitDistanceSplits(splitAB, splitAI), 2L)
  expect_equal(MatchingSplitDistanceSplits(splitAB, splitABCD), 2L)
  expect_equal(MatchingSplitDistanceSplits(splitAB, splitABCDE), 3L)
  expect_equal(MatchingSplitDistanceSplits(splitABC, splitAEF), 4L)
  expect_equal(MatchingSplitDistanceSplits(splitABC, splitAEF),
               MatchingSplitDistanceSplits(splitAEF, splitABC))
  
  # Invariant to tree description order
  sq_pectinate <- ape::read.tree(
    text = "((((((1, 2), 3), 4), 5), 6), (7, (8, (9, (10, 11)))));"
  )
  shuffle1 <- ape::read.tree(
    text = "(((((1, 5), 2), 6), (3, 4)), ((8, (7, 9)), (10, 11)));"
  )
  shuffle2 <- ape::read.tree(
    text = "(((8, (7, 9)), (10, 11)), ((((1, 5), 2), 6), (3, 4)));"
  )
  expect_equal(MatchingSplitDistance(shuffle1, sq_pectinate),
               MatchingSplitDistance(sq_pectinate, shuffle1))
  expect_equal(MatchingSplitDistance(shuffle1, shuffle2), 0L)
  expect_equal(MatchingSplitDistance(shuffle1, sq_pectinate),
               MatchingSplitDistance(shuffle2, sq_pectinate))
})

test_that("NyeSimilarity is correctly calculated, and matches JRF", {
  listBalSym <- list(treeBal8, treeSym8)
  
  JRF <- function(..., sim = TRUE)
    JaccardRobinsonFoulds(..., k = 1, similarity = sim, allowConflict = TRUE)
  expect_equal(5L, NyeSimilarity(as.Splits(treeSym8), treeSym8))
  
  expect_equal(1, NyeSimilarity(treeSym8, treeSym8, normalize = TRUE))
  expect_equal(1, JRF(treeSym8, treeSym8, normalize = TRUE))
  
  expect_equal(0, NyeSimilarity(treeSym8, treeStar8, normalize = FALSE))
  expect_equal(0, NyeSimilarity(treeSym8, treeStar8, normalize = TRUE))
  expect_equal(0, JRF(treeSym8, treeStar8, normalize = TRUE))
  
  expect_equal(0, NyeSimilarity(treeStar8, treeStar8, normalize = FALSE))
  expect_equal(NaN, NyeSimilarity(treeStar8, treeStar8, normalize = TRUE, 
                                      normalizeMax = FALSE))
  
  expect_equal(c(3.8, 5), NyeSimilarity(treeSym8, listBalSym))
  expect_equal(2 / 3, NyeSimilarity(treeAb.Cdefgh, treeAbc.Defgh), 
               tolerance = 1e-7)
  expect_equal(2 * (1 / 3), tolerance = 1e-7, 
               NyeSimilarity(treeAb.Cdefgh, treeAbc.Defgh, similarity = FALSE))
  
  expect_equal(1L, NyeSimilarity(treeSym8, treeAbcd.Efgh, normalize = FALSE))
  expect_equal(1L / 5L, NyeSimilarity(treeSym8, treeAbcd.Efgh, normalize = 5L))
  expect_equal(0.2, JRF(treeSym8, treeAbcd.Efgh, normalize = 5L * 2L))
  expect_equal(1/3, NyeSimilarity(treeSym8, treeAbcd.Efgh, normalize = TRUE))
  expect_equal(1/3, JRF(treeSym8, treeAbcd.Efgh, normalize = TRUE))
  expect_equal(2/3, NyeSimilarity(treeSym8, treeAbcd.Efgh, similarity = FALSE,
                                  normalize = TRUE))
  expect_equal(2/3, JRF(treeSym8, treeAbcd.Efgh, sim = FALSE, normalize = TRUE))
  
  expect_equal(1L / ((5L + 1L) / 2L),
               NyeSimilarity(treeSym8, treeAbcd.Efgh, normalize = TRUE))
  expect_true(NyeSimilarity(treeSym8, treeBal8) > 
                NyeSimilarity(treeSym8, treeOpp8))
  
  NormalizationTest(NyeSimilarity)
})


test_that("Jaccard RF extremes tend to equivalent functions", {
  expect_equal(JaccardRobinsonFoulds(treeSym8, list(treeBal8, treeSym8),
                                     similarity = TRUE, k = 1L,
                                     allowConflict = TRUE),
               NyeSimilarity(treeSym8, list(treeBal8, treeSym8)) * 2L)
  
  expect_equal(JaccardRobinsonFoulds(treeSym8, list(treeBal8, treeSym8),
                                     similarity = FALSE, k = Inf),
               RobinsonFoulds(treeSym8, list(treeBal8, treeSym8)))
  
  expect_equal(JaccardRobinsonFoulds(treeSym8, list(treeBal8, treeSym8),
                                     similarity = FALSE, k = 999999),
               RobinsonFoulds(treeSym8, list(treeBal8, treeSym8)))
})

test_that("Jaccard RF is correctly calculated", {
  expect_equal(5L * 2L, JaccardRobinsonFoulds(treeSym8, treeSym8,
                                         k = 2, similarity = TRUE))
  expect_equal(c(3.32, 5) * 2L, 
               JaccardRobinsonFoulds(treeSym8, list(treeBal8, treeSym8),
                                     similarity = TRUE, k = 2))
  expect_equal(2 * 2, 3 * JaccardRobinsonFoulds(treeAb.Cdefgh, treeAbc.Defgh,
                                            similarity = TRUE),
               tolerance = 1e-7)
  expect_equal(1, JaccardRobinsonFoulds(treeSym8, treeSym8,
                                        similarity = TRUE, normalize = TRUE))
  expect_equal(0, JaccardRobinsonFoulds(treeSym8, treeSym8,
                                        similarity = FALSE, normalize = TRUE))
  expect_equal(1L * 2L, 
               JaccardRobinsonFoulds(treeSym8, treeAbcd.Efgh, similarity = TRUE,
                                     normalize = FALSE, k = 2))
  expect_equal(1L * 2L / 6L, 
               JaccardRobinsonFoulds(treeSym8, treeAbcd.Efgh, similarity = TRUE,
                                     normalize = TRUE, k = 2))
  expect_lt(JaccardRobinsonFoulds(treeSym8, treeBal8, k = 2),
            JaccardRobinsonFoulds(treeSym8, treeOpp8, k = 2))
  expect_lt(JaccardRobinsonFoulds(treeSym8, treeBal8, k = 3L),
            JaccardRobinsonFoulds(treeSym8, treeBal8, k = 4L))
  expect_lt(JaccardRobinsonFoulds(treeCat8, treeTac8, allowConflict = TRUE),
            JaccardRobinsonFoulds(treeCat8, treeTac8, allowConflict = FALSE))
  
  library("TreeTools", quietly = TRUE)
  expect_equal(0, JaccardRobinsonFoulds(BalancedTree(64), BalancedTree(64)))
  expect_lt(0, JaccardRobinsonFoulds(BalancedTree(64), PectinateTree(64)))
  expect_equal(0, JaccardRobinsonFoulds(BalancedTree(264), BalancedTree(264)))
  expect_lt(0, JaccardRobinsonFoulds(BalancedTree(264), PectinateTree(264)))
})

test_that("RobinsonFoulds() is correctly calculated", {
  skip_if_not_installed("phangorn")
  RF <- function(tree1, tree2) {
    suppressMessages(phangorn::RF.dist(reorder(tree1, "cladewise"),
                                       reorder(tree2, "cladewise")))
  }
  RFTest <- function(tree1, tree2) {
    expect_equal(RF(tree1, tree2), RobinsonFoulds(tree1, tree2))
  }
  
  RFTest(treeSym8, treeSym8)
  RFTest(treeBal8, treeSym8)
  expect_equal(c(4, 0), RobinsonFoulds(treeSym8, list(treeBal8, treeSym8)))
  RFTest(treeAb.Cdefgh, treeAbc.Defgh)
  expect_equal(0, RobinsonFoulds(treeSym8, treeSym8, normalize = TRUE))
  expect_equal(4L / 6L, 
               RobinsonFoulds(treeSym8, treeAbcd.Efgh, normalize = TRUE))
  RFTest(treeSym8, treeOpp8)
  
  RFNtipTest <- function(nTip) {
    backLeaves <- paste0("t", rev(seq_len(nTip)))
    RFTest(TreeTools::PectinateTree(backLeaves), 
           TreeTools::BalancedTree(nTip))
  }
  RFNtipTest(10)
  RFNtipTest(32)
  RFNtipTest(50)
  RFNtipTest(64)
  RFNtipTest(67)
  RFNtipTest(128)
  RFNtipTest(1024)
  RFNtipTest(1027)
  
  NormalizationTest(RobinsonFoulds, similarity = TRUE)
  #TODO we may wish to revise this test once we implement diag = TRUE to 
  #allow similarities to be calculated on the diagonal.
  expect_equal(numeric(0), RobinsonFoulds(treeSym8, normalize = TRUE))
})

test_that("RobinsonFoulds() supports large lists", {
  set.seed(0)
  trees <- lapply(rep(12, 129), TreeTools::RandomTree, root = TRUE)
  
  # The purpose of the test is to verify we don't encounter a stack overflow
  result <- as.matrix(RobinsonFoulds(trees))
  diag(result) <- Inf
  
  # We should also check that the results are correct: Let's pick a cell
  # at random, noting that most RF distances between random trees will
  # be the maximum value.
  minX <- which.min(apply(result, 2, min))
  minY <- apply(result, 2, which.min)[[minX]]
  
  expect_equal(result[minX, minY],
               RobinsonFoulds(trees[[minX]], trees[[minY]]))
})

test_that("Robinson Foulds Info is correctly calculated", {
  expect_equal(22.53747 * 2L, tolerance = 1e-05,
               InfoRobinsonFoulds(treeSym8, treeSym8, similarity = TRUE,
                                  normalize = FALSE))
  expect_equal(0, tolerance = 1e-05,
               InfoRobinsonFoulds(treeSym8, treeSym8, normalize = TRUE))
  expect_equal(1, tolerance = 1e-05,
               InfoRobinsonFoulds(treeSym8, treeSym8, similarity = TRUE, 
                                  normalize = TRUE))
  expect_equal(24.9, tolerance = 0.01, 
               InfoRobinsonFoulds(treeSym8, treeBal8, similarity = TRUE))
  expect_equal(SplitwiseInfo(treeSym8) + SplitwiseInfo(treeBal8) -
                 InfoRobinsonFoulds(treeSym8, treeBal8, similarity = FALSE),
               InfoRobinsonFoulds(treeSym8, treeBal8, similarity = TRUE))
  expect_equal(-log2(945/10395) * 2,
               InfoRobinsonFoulds(treeSym8, treeAb.Cdefgh, similarity = TRUE))
  expect_equal(-log2(945/10395) * 2, 
               InfoRobinsonFoulds(treeSym8, treeAb.Cdefgh, similarity = TRUE))
  expect_equal(-log2(315/10395) * 2, 
               InfoRobinsonFoulds(treeSym8, treeAbc.Defgh, similarity = TRUE))
  
  # Test symmetry of small vs large splits
  expect_equal(InfoRobinsonFoulds(treeSym8, treeAbc.Defgh),
               InfoRobinsonFoulds(treeAbc.Defgh, treeSym8))
  expect_equal(-log2(225/10395) * 2, 
               InfoRobinsonFoulds(treeSym8, treeAbcd.Efgh, similarity = TRUE))
  expect_equal((-log2(225/10395) - log2(945/10395)) * 2,
               InfoRobinsonFoulds(treeSym8, treeTwoSplits, similarity = TRUE))
  expect_equal(InfoRobinsonFoulds(treeSym8, list(treeSym8, treeBal8)), 
               RobinsonFouldsInfo(list(treeSym8, treeBal8), treeSym8))
  
  # Check that large trees work
  library("TreeTools", quietly = TRUE)
  expect_equal(0, InfoRobinsonFoulds(BalancedTree(64), BalancedTree(64)))
  expect_lt(0, InfoRobinsonFoulds(BalancedTree(64), PectinateTree(64)))
  expect_equal(0, InfoRobinsonFoulds(BalancedTree(129), BalancedTree(129)))
  expect_lt(0, InfoRobinsonFoulds(BalancedTree(129), PectinateTree(129)))
})


test_that("Kendall-Colijn distance is correctly calculated", {
  # Expected values calculated using treespace::treeDist(treeSym8, treeBal8)
  expect_equal(2.828427, KendallColijn(treeSym8, treeBal8), tolerance=1e-06)
  expect_equal(2.828427, KendallColijn(treeCat8, treeBal8), tolerance=1e-06)
  expect_equal(7.211103, KendallColijn(treeSym8, treeOpp8), tolerance=1e-06)
  expect_equal(matrix(c(0L, 8L), nrow=2, ncol=2, byrow=TRUE), 
               KendallColijn(list(treeSym8, treeCat8), list(treeCat8, treeTac8)), tolerance=1e-06)
  expect_equal(8L, KendallColijn(treeCat8, treeTac8), tolerance=1e-06)
  expect_equal(0L, KendallColijn(treeSym8, treeCat8), tolerance=1e-06)
  expect_equal(8L, KendallColijn(treeSym8, treeTac8), tolerance=1e-06)
  expect_equal(8L, KendallColijn(treeCat8, treeTac8), tolerance=1e-06)

  expect_equal(5.291503, KendallColijn(treeSym8, treeAb.Cdefgh), tolerance=1e-06)
  expect_equal(4.358899, KendallColijn(treeSym8, treeAbc.Defgh), tolerance=1e-06)
  expect_equal(5L, KendallColijn(treeSym8, treeAcd.Befgh), tolerance=1e-06)
  expect_equal(3.464102, KendallColijn(treeSym8, treeAbcd.Efgh), tolerance=1e-06)
  expect_equal(3L, KendallColijn(treeSym8, treeTwoSplits), tolerance=1e-06)
  expect_equal(2.828427, KendallColijn(treeAbc.Defgh, treeTwoSplits), tolerance=1e-06)
})

test_that("Multiple comparisons are correctly ordered", {
  nTrees <- 6L
  nTip <- 16L
  
  set.seed(0)
  trees <- lapply(rep(nTip, nTrees), ape::rtree, br = NULL)
  trees[[1]] <- TreeTools::BalancedTree(nTip)
  trees[[nTrees - 1L]] <- TreeTools::PectinateTree(nTip)
  class(trees) <- "multiPhylo"
  
  skip_if_not_installed("phangorn")
  PhangRF <- function (trees) {
    phangorn::RF.dist(reorder(trees, "cladewise"))
  }
  
  expect_equal(PhangRF(trees), RobinsonFoulds(trees),
               ignore_attr = TRUE)
  
  # Test CompareAll
  expect_equal(as.matrix(PhangRF(trees)),
               as.matrix(CompareAll(reorder(trees, "cladewise"),
                                    phangorn::RF.dist, 0L)),
               ignore_attr = TRUE)
  
  NNILoose <- function(x, y) NNIDist(x, y)["loose_upper"]
  expect_equal(CompareAll(trees, NNILoose),
               CompareAll(trees, NNIDist)$loose_upper,
               ignore_attr = TRUE)
})

test_that("Normalization occurs as documented", {
  library("TreeTools", quietly = TRUE)
  tree1 <- BalancedTree(8)
  tree2 <- CollapseNode(PectinateTree(8), 12:13)
  
  info1 <- SplitwiseInfo(tree1) # 19.367
  info2 <- SplitwiseInfo(tree2) # 11.963
  
  ent1 <- ClusteringEntropy(tree1) # 4.245
  ent2 <- ClusteringEntropy(tree2) # 2.577
  
  # Phylogenetic information
  spi <- SharedPhylogeneticInfo(tree1, tree2, normalize = FALSE) # 9.64
  dpi <- DifferentPhylogeneticInfo(tree1, tree2, normalize = FALSE) # 12.04
  expect_equal(spi + spi + dpi, info1 + info2)
  
  expect_equal(SharedPhylogeneticInfo(tree1, tree2, normalize = TRUE),
               (spi + spi) / (info1 + info2))
  expect_equal(PhylogeneticInfoDistance(tree1, tree2, normalize = TRUE),
               dpi / (info1 + info2))
  
  # Matching split information
  mmsi <- MatchingSplitInfo(tree1, tree2, normalize = FALSE)
  msid <- MatchingSplitInfoDistance(tree1, tree2, normalize = FALSE)
  expect_equal(mmsi + mmsi + msid, info1 + info2)
  
  expect_equal(MatchingSplitInfo(tree1, tree2, normalize = TRUE),
               (mmsi + mmsi) / (info1 + info2))
  expect_equal(MatchingSplitInfoDistance(tree1, tree2, normalize = TRUE),
               msid / (info1 + info2))
  
  
  # Clustering information
  mci <- MutualClusteringInfo(tree1, tree2, normalize = FALSE)
  cid <- ClusteringInfoDistance(tree1, tree2, normalize = FALSE)
  expect_equal(mci + mci + cid, ent1 + ent2)
  
  expect_equal(MutualClusteringInfo(tree1, tree2, normalize = TRUE),
               (mci + mci) / (ent1 + ent2))
  expect_equal(ClusteringInfoDistance(tree1, tree2, normalize = TRUE),
               cid / (ent1 + ent2))
  
})

test_that("Independent of root position", {
  
  library("TreeTools")
  
  bal8 <- BalancedTree(8)
  pec8 <- PectinateTree(8)
  
  trees <- lapply(list(bal8, RootTree(bal8, "t4"),
                       pec8, RootTree(pec8, "t4")), UnrootTree)
  
  lapply(methodsToTest[-length(methodsToTest)], function(Method) {
    dists <- as.matrix(Method(trees))
    expect_equal(dists[1, 1], dists[1, 2])
    expect_equal(dists[1, 3], dists[1, 4])
    expect_equal(dists[1, 3], dists[2, 4])
    expect_equal(dists[2, 3], dists[2, 4])
    expect_equal(dists[3, 3], dists[3, 4])
  })

  
  Test <- function(Method, score = 0L, ...) {
    expect_equal(Method(trees[[1]], trees[[1]], ...), score)
    expect_equal(Method(trees[[1]], trees[[2]], ...), score)
    expect_equal(Method(trees[[3]], trees[[3]], ...), score)
  }
  
  Test(MASTSize, 8L, rooted = FALSE)
  # Tested further for NNIDist in test-tree_distance_nni.R
  Test(NNIDist, c(lower = 0, best_lower = 0, tight_upper = 0, best_upper = 0,
                  loose_upper = 0, fack_upper = 0, li_upper = 0))
  Test(SPRDist, 0)
  
})

test_that("Matchings are correctly reported", {
  expect_matching_score(MutualClusteringInfo)
  expect_matching_score(SharedPhylogeneticInfo)
  expect_matching_score(MatchingSplitInfo)
})

test_that("Header mutual_clustering_score matches MutualClusteringInfo", {
  skip_if_not_installed("TreeDist")
  library(TreeTools)
  library(TreeDist)

  bal8 <- BalancedTree(8)
  pec8 <- PectinateTree(8)
  star8 <- StarTree(8)

  tips <- TipLabels(bal8)
  n_tip <- length(tips)
  splits_bal <- as.Splits(bal8, tips)
  splits_pec <- as.Splits(pec8, tips)
  splits_star <- as.Splits(star8, tips)

  # Score-only from the installable-header implementation
  impl_score <- TreeDist:::cpp_mci_impl_score

  impl_bal_pec <- impl_score(splits_bal, splits_pec, n_tip)
  impl_bal_bal <- impl_score(splits_bal, splits_bal, n_tip)
  impl_star <- impl_score(splits_bal, splits_star, n_tip)

  # Reference from MutualClusteringInfo (unnormalized score)
  ref_bal_pec <- MutualClusteringInfo(bal8, pec8)
  ref_bal_bal <- MutualClusteringInfo(bal8, bal8)

  expect_equal(impl_bal_pec, ref_bal_pec, tolerance = 1e-10)
  expect_equal(impl_bal_bal, ref_bal_bal, tolerance = 1e-10)
  expect_equal(impl_star, 0)
})

test_that("Header MCI covers exact-match early exit and partial LAP", {
  skip_if_not_installed("TreeDist")
  library(TreeTools)
  library(TreeDist)
  impl_score <- TreeDist:::cpp_mci_impl_score

  # Two identical trees → all splits match exactly (early exit path)
  bal20 <- BalancedTree(20)
  tips <- TipLabels(bal20)
  n_tip <- length(tips)
  splits20 <- as.Splits(bal20, tips)

  result <- impl_score(splits20, splits20, n_tip)
  expect_equal(result, MutualClusteringInfo(bal20, bal20), tolerance = 1e-10)

  # Trees that share some but not all splits → partial match + LAP
  pec20 <- PectinateTree(20)
  splits_pec20 <- as.Splits(pec20, tips)

  result2 <- impl_score(splits20, splits_pec20, n_tip)
  expect_equal(result2, MutualClusteringInfo(bal20, pec20), tolerance = 1e-10)
})

#' Robinson&ndash;Foulds distances, with adjustments for phylogenetic information
#' content
#' 
#' `RobinsonFoulds()` calculates the Robinson&ndash;Foulds distance
#' \insertCite{Robinson1981}{TreeDist}, or the corresponding similarity measure.
#' `InfoRobinsonFoulds()` weights splits according to their phylogenetic
#' information content \insertCite{@§2.1 in @SmithDist}{TreeDist}.
#' Optionally, the matching between identical splits may reported.
#' Generalized Robinson&ndash;Foulds distances (see [`TreeDistance()`])
#' are better suited to most use cases
#' \insertCite{SmithDist,SmithSpace}{TreeDist}.
#' 
#' `RobinsonFoulds()` calculates the standard Robinson&ndash;Foulds distance,
#' i.e. the number of splits that occur in one tree but not the other.
#' `InfoRobinsonFoulds()` calculates the tree similarity or distance by summing 
#' the phylogenetic information content of all splits that are (or are not)
#' identical in both trees.  Consequently, splits that are more likely
#' to be identical by chance alone make a smaller contribution to overall
#' tree distance, because their similarity is less remarkable.
#' 
#' Rapid comparison between multiple pairs of trees employs the
#' \insertCite{Day1985;textual}{TreeDist} linear-time algorithm.
#' 
#' @inheritParams TreeDistance
#' @param similarity Logical specifying whether to report the result as a tree
#' similarity, rather than a difference.
#' 
#' @templateVar returns `RobinsonFoulds()` and `InfoRobinsonFoulds()` return
#' @template distReturn
#' @return If `reportMatching = TRUE`, the `pairScores` attribute 
#' returns a logical matrix specifying whether each pair of splits is identical.
#' 
#' 
#' @section Normalization:
#' 
#' - `RobinsonFoulds()` is normalized against the total number of splits that
#'  are present.
#'  
#' - `InfoRobinsonFoulds()` is normalized against the sum of the phylogenetic 
#' information of all splits in each tree, treated independently.
#'  
#' @references \insertAllCited{}
#' 
#' @examples
#'  # For BalancedTree, PectinateTree, as.phylo:
#' library("TreeTools", quietly = TRUE)
#' balanced7 <- BalancedTree(7)
#' pectinate7 <- PectinateTree(7)
#' RobinsonFoulds(balanced7, pectinate7)
#' RobinsonFoulds(balanced7, pectinate7, normalize = TRUE)
#' VisualizeMatching(RobinsonFouldsMatching, balanced7, pectinate7)
#' 
#' InfoRobinsonFoulds(balanced7, pectinate7)
#' VisualizeMatching(InfoRobinsonFoulds, balanced7, pectinate7)
#' @template MRS
#' 
#' 
#' @family tree distances
#' @seealso Display paired splits: [`VisualizeMatching()`]
#' 
#' @export
#' @encoding UTF-8
#' @name Robinson-Foulds

#' @aliases RobinsonFouldsInfo
#' @rdname Robinson-Foulds
InfoRobinsonFoulds <- function(tree1, tree2 = NULL, similarity = FALSE,
                                normalize = FALSE, reportMatching = FALSE) {
  
  # Fast path for distance (not similarity): avoids duplicate as.Splits()
  if (!similarity) {
    # All-pairs fast path
    fast <- .FastDistPath(tree1, tree2, reportMatching,
                          cpp_rf_info_all_pairs,
                          cpp_splitwise_info_batch)
    if (!is.null(fast)) {
      treesIndependentInfo <- .PairwiseSums(fast[["entropies"]])
      unnormalized <- .FloorNumericalNoise(
        treesIndependentInfo - fast[["info"]] - fast[["info"]],
        treesIndependentInfo)
      ret <- NormalizeInfo(unnormalized, tree1, tree2, how = normalize,
                           InfoInTree = SplitwiseInfo, Combine = "+")
      attributes(ret) <- attributes(fast[["info"]])
      return(ret)
    }

    # Cross-pairs fast path
    fast_many <- .FastManyManyPath(tree1, tree2, reportMatching,
                                   cpp_rf_info_cross_pairs,
                                   cpp_splitwise_info_batch)
    if (!is.null(fast_many)) {
      irf <- fast_many[["dists"]]
      info1 <- fast_many[["info1"]]
      info2 <- fast_many[["info2"]]
      treesIndependentInfo <- outer(info1, info2, "+")

      unnormalized <- .FloorNumericalNoise(treesIndependentInfo - irf - irf,
                                            treesIndependentInfo)
      ret <- NormalizeInfo(unnormalized, tree1, tree2, how = normalize,
                           InfoInTree = SplitwiseInfo, Combine = "+")
      return(ret)
    }
  }

  unnormalized <- CalculateTreeDistance(InfoRobinsonFouldsSplits, tree1, tree2,
                                        reportMatching) * 2

  if (!similarity) {
    treesIndependentInfo <- .MaxValue(tree1, tree2, SplitwiseInfo)
    unnormalized <- .FloorNumericalNoise(treesIndependentInfo - unnormalized,
                                          treesIndependentInfo)
  }
  
  # Return:
  NormalizeInfo(unnormalized, tree1, tree2, how = normalize,
                InfoInTree = SplitwiseInfo, Combine = "+")
}

#' @export
RobinsonFouldsInfo <- InfoRobinsonFoulds

#' @rdname Robinson-Foulds
#' @inheritParams SharedPhylogeneticInfoSplits
#' @export
InfoRobinsonFouldsSplits <- function(splits1, splits2, 
                                      nTip = attr(splits1, "nTip"),
                                      reportMatching = FALSE) {
  
  GeneralizedRF(splits1, splits2, nTip, cpp_robinson_foulds_info,
                maximize = FALSE, reportMatching = reportMatching)
}

#' @rdname Robinson-Foulds
#' @importFrom TreeTools as.ClusterTable NSplits
#' @export
RobinsonFoulds <- function(tree1, tree2 = NULL, similarity = FALSE,
                            normalize = FALSE, reportMatching = FALSE) {
  if (is.null(tree2)) {
    ct <- as.ClusterTable(tree1)
    rf <- robinson_foulds_all_pairs(if (is.list(ct)) ct else list(ct))
    if (similarity) {
      unnormalized <- structure(rf + rf, Size = length(tree1), class = "dist")
    } else {
      splits <- NSplits(tree1)
      nSplits <- outer(splits, splits, "+")
      unnormalized <- structure(nSplits[lower.tri(nSplits)] - rf - rf,
                                Size = length(tree1),
                                class = "dist")
    }
  } else {
    # Fast cross-pairs path: batch C++ via ClusterTable (Day 1985).
    # Only applicable when both inputs are lists of trees with matching
    # tip labels and reportMatching is not requested.
    fast <- NULL
    if (!reportMatching &&
        !inherits(tree1, c("phylo", "Splits")) &&
        !inherits(tree2, c("phylo", "Splits")) &&
        is.null(getOption("TreeDist-cluster"))) {
      lab1 <- TipLabels(tree1)
      lab2 <- TipLabels(tree2)
      if (is.list(lab1)) lab1 <- lab1[[1]]
      if (is.list(lab2)) lab2 <- lab2[[1]]
      if (setequal(lab1, lab2)) {
        ct1 <- as.ClusterTable(tree1, tipLabels = lab1)
        ct2 <- as.ClusterTable(tree2, tipLabels = lab1)
        shared <- robinson_foulds_cross_pairs(
          if (is.list(ct1)) ct1 else list(ct1),
          if (is.list(ct2)) ct2 else list(ct2)
        )
        splits1 <- NSplits(tree1)
        splits2 <- NSplits(tree2)
        if (similarity) {
          fast <- shared + shared
        } else {
          fast <- outer(splits1, splits2, "+") - shared - shared
        }
        dimnames(fast) <- list(names(tree1), names(tree2))
      }
    }
    if (is.null(fast)) {
      unnormalized <- CalculateTreeDistance(RobinsonFouldsSplits, tree1, tree2,
                                            reportMatching)
      if (similarity) {
        unnormalized <- .MaxValue(tree1, tree2, NSplits) - unnormalized
      }
    } else {
      unnormalized <- fast
    }
  }
  
  
  # Return:
  NormalizeInfo(unnormalized, tree1, tree2, how = normalize,
                InfoInTree = NSplits, Combine = `+`)
}

#' @describeIn Robinson-Foulds Matched splits, intended for use with 
#' [`VisualizeMatching()`].
#' @param \dots Not used.
#' @importFrom TreeTools NSplits
#' @export
RobinsonFouldsMatching <- function(tree1, tree2, similarity = FALSE,
                                    normalize = FALSE, ...) {
  ret <- CalculateTreeDistance(RobinsonFouldsSplits, tree1, tree2,
                               reportMatching = TRUE)

  ret <- .MaxValue(tree1, tree2, NSplits) - ret
  
  attr(ret, "pairScores") <- !attr(ret, "pairScores")
  
  # Return:
  ret
}

#' @rdname Robinson-Foulds
#' @inheritParams SharedPhylogeneticInfoSplits
#' @export
RobinsonFouldsSplits <- function(splits1, splits2,
                                  nTip = attr(splits1, "nTip"),
                                  reportMatching = FALSE) {
  GeneralizedRF(splits1, splits2, nTip, cpp_robinson_foulds_distance,
                maximize = FALSE, reportMatching = reportMatching)
}

#' Pseudo-3D plotting
#' 
#' `Plot3()` displays three-dimensional data in two dimensions, reflecting the
#' third dimension with point scaling, overlap and fogging.
#' Points with a lower `z` value are smaller than, fainter than, and overlapped
#' by points with a higher value.
#' 
#' @param x,y,z Coordinates of points to plot.
#' @param fog Numeric from zero (no fading) to one (furthest points are
#' invisible) specifying amount to fade distant points.
#' @param shrink Numeric specifying degree to which size of plotted point
#' should reflect `z` position. `0` denotes no scaling; if `1`, furthest
#' point will have zero size.
#' @param plot.bg Colour with which to fill plot area, used as fog colour.
#' @param bg,cex,col,pch,add,axes,frame.plot,\dots Parameters passed to 
#' [plot.default()].
#' 
#' @examples 
#' Plot3(1:10, 1:10, 1:10, cex = 7, pch = 16, axes = FALSE, asp = 1)
#' # Extreme values of fog and shrink will cause smallest z values to
#' # become invisible.
#' Plot3(1:10, 1:10, 1:10, cex = 7, pch = 16, axes = FALSE, asp = 1,
#'       fog = 1, shrink = 1)
#' @template MRS
#' @importFrom graphics plot points rect
#' @importFrom grDevices colorRamp rgb
#' @export
Plot3 <- function(x, y = NULL, z = NULL,
                   pch = par("pch"), col = par("col"),
                   bg = NA, cex = 1,
                   axes = TRUE,
                   frame.plot = axes,
                   plot.bg = NA, 
                   fog = 1/2, shrink = 1/2,
                   add = FALSE,
                   ...) {
  if (is.null(y)) {
    z <- x[, 3]
    y <- x[, 2]
    x <- x[, 1]
  }
  
  zResolution <- 128L
  n <- length(z)
  zOrder <- order(z)
  bg <- rep_len(bg, n)
  col <- rep_len(col, n)
  cex <- rep_len(cex, n)
  pch <- rep_len(pch, n)
  
  #zStep <- as.integer(cut(z, zResolution))
  zScale <- max(z) - z
  zScale <- zScale / max(zScale)
  
  fogOffset <- zResolution * fog
  bgCol <- if (is.na(plot.bg)) "white" else plot.bg
  .FadeCol <- function(x, fadeAmount) {
    if (is.na(x)) {
      NA_character_
    } else {
      rgb(colorRamp(c(x, bgCol), alpha = TRUE)(fog * fadeAmount),
          maxColorValue = 255)
    }
  }
  fadedCol <- vapply(seq_along(z), function(i) {
    .FadeCol(col[i], zScale[i])
  }, character(1))
  fadedBg <- vapply(seq_along(z), function(i) {
    .FadeCol(bg[i], zScale[i])
  }, character(1))
  if (!add) {
    plot(x, y, type = "n", axes = axes, frame.plot = frame.plot, ...)
    if (!is.na(plot.bg)) {
      rect(par("usr")[[1]], par("usr")[[3]], par("usr")[[2]], par("usr")[[4]],
           col = plot.bg, border = frame.plot)
    }
  }
  points(x[zOrder], y[zOrder],
       cex = cex[zOrder] * (1 - (shrink * zScale[zOrder])),
       pch = pch[zOrder],
       col = fadedCol[zOrder],
       bg = fadedBg[zOrder],
       ...)
}

test_that("Plot3() plots", {
  skip_if_not_installed("vdiffr", minimum_version = "1.0.0")
  library("vdiffr")
  skip_if(packageVersion("graphics") < "4.3")
  
  expect_doppelganger("Simple plot", function() {
    disorder <- c(1, 5, 2, 6, 3, 7, 4, 8, 5, 9, 10)
    Plot3(disorder, disorder, 6 - disorder, pch = 21, cex = 18,
          asp = 1, xlab = "", ylab = "",
          bg = "white",
          plot.bg = "#bbeedd99",
          #shrink = 0,
          frame.plot = FALSE)
  })
  
  
  expect_doppelganger("Plotting order", function() {
    pts <- cbind(1:10, 11:20, c(1, 1, 1, 1, 10, 10, 10, 5, 6, 7))
    Plot3(pts,
          pch = c(3, rep(22, 4), rep(21, 4), 4),
          cex = c(rep(18, 6), 15, rep(18, 3)),
          col = c("red", rep(1, 4), rep(3, 4), 2),
          bg = c(rep("#ff555511", 5), rep("#bbbbff11", 5)),
          fog = 0.8, shrink = 0.8,
          asp = 1, xlab = "", ylab = "", frame.plot = FALSE)
  })
  
})

#' Consensus tree minimizing transfer distance
#'
#' Construct a consensus tree that minimizes the sum of transfer distances
#' to a set of input trees, using a greedy add-and-prune heuristic.
# Note added 2026-06-10; function was only available for 1 release so need
# not stick around too long
#' This function is moving to `ConsTree::Transfer()` and will soon be removed.
#' A copy is retained here temporarily.
#'
#' Unlike the majority-rule consensus, which minimizes Robinson-Foulds
#' distance and can be highly unresolved when phylogenetic signal is low,
#' `TransferConsensus()` uses the finer-grained transfer distance
#' \insertCite{Lemoine2018}{TreeDist} to construct a more resolved consensus
#' tree.
#'
#' The algorithm pools all splits observed across input trees, computes
#' pairwise transfer distances between them, and greedily adds or removes
#' splits to minimize total transfer dissimilarity cost.  The approach
#' follows \insertCite{Takazawa2026;textual}{TreeDist}, reimplemented for
#' 'TreeDist' infrastructure.
#'
#' @param trees An object of class `multiPhylo`: the input trees.
#'   All trees must share the same tip labels.
#' @param scale Logical; if `TRUE` (default), use the scaled transfer
#'   distance (normalized by light-side size minus one).  If `FALSE`, use
#'   the unscaled (raw Hamming) transfer distance.
#' @param greedy Character string specifying the greedy strategy:
#'   `"best"` (default) picks the single highest-benefit action at each step;
#'   `"first"` picks the first improving action encountered (faster but
#'   potentially lower quality).
#' @param init Character string specifying the initial consensus:
#'   `"empty"` (default) starts with no splits (purely additive);
#'   `"majority"` starts with the majority-rule consensus and refines.
#'
#' @return A tree of class `phylo`.
#'
#' @references
#' \insertAllCited{}
#' @family summary trees
#' @importFrom TreeTools as.Splits TipLabels NSplits Consensus StarTree
#' @export
TransferConsensus <- function(trees,
                              scale = TRUE,
                              greedy = c("best", "first"),
                              init = c("empty", "majority")) {
  .Deprecated("ConsTree::Transfer")
  greedy <- match.arg(greedy)
  init <- match.arg(init)

  if (!inherits(trees, "multiPhylo")) {
    stop("`trees` must be an object of class 'multiPhylo'.")
  }
  nTree <- length(trees)
  if (nTree < 2L) stop("Need at least 2 trees.")
  tipLabels <- TipLabels(trees[[1]])
  nTip <- length(tipLabels)
  if (nTip < 4L) {
    return(StarTree(tipLabels))
  }
  .CheckMaxTips(nTip)

  # Convert each tree to a raw split matrix (TreeTools C++ internally).
  # as.Splits() will error if a tree's tips don't match tipLabels.
  splitsList <- lapply(trees, function(tr) {
    sp <- as.Splits(tr, tipLabels)
    unclass(sp)
  })

  # Validate all split matrices have consistent column count
  nCols <- vapply(splitsList, ncol, integer(1L))
  if (length(unique(nCols)) != 1L) {
    stop("Split matrices have inconsistent column counts.")
  }

  # Delegate all work to C++
  nThreads <- max(1L, getOption("TreeDist.threads",
                                 getOption("mc.cores", 1L)))
  res <- cpp_transfer_consensus(
    splitsList, nTip, scale,
    greedy_best_flag = (greedy == "best"),
    init_majority = (init == "majority"),
    n_threads = nThreads
  )

  included <- res$included
  if (!any(included)) {
    return(StarTree(tipLabels))
  }

  rawSplits <- res$raw_splits[included, , drop = FALSE]
  sp <- structure(rawSplits, nTip = nTip, tip.label = tipLabels,
                  class = "Splits")
  as.phylo(sp)
}


#' Profile transfer consensus timing (internal diagnostic)
#' @keywords internal
#' @noRd
tc_profile <- function(trees, scale = TRUE, greedy = "best",
                       init = "empty", n_iter = 1L, n_threads = 1L) {
  if (!inherits(trees, "multiPhylo")) {
    stop("`trees` must be an object of class 'multiPhylo'.")
  }
  if (length(trees) < 2L) stop("Need at least 2 trees.")
  tipLabels <- TipLabels(trees[[1]])
  nTip <- length(tipLabels)
  if (nTip < 4L) stop("Need at least 4 tips for profiling.")
  .CheckMaxTips(nTip)

  splitsList <- lapply(trees, function(tr) unclass(as.Splits(tr, tipLabels)))

  cpp_tc_profile(splitsList, nTip, scale,
                 greedy_best_flag = (greedy == "best"),
                 init_majority = (init == "majority"),
                 n_iter = as.integer(n_iter),
                 n_threads = as.integer(n_threads))
}


# ===========================================================================
# Internal helpers
# ===========================================================================

# popcount lookup for 0:255
.POPCOUNT <- as.integer(sapply(0:255, function(x) sum(as.integer(intToBits(x)))))

#' Pool unique splits, returning an integer (logical) matrix
#' @noRd
.PoolSplits <- function(trees, tipLabels) {
  nTip <- length(tipLabels)
  nTree <- length(trees)

  # Collect all splits as logical matrices and hash to unique set
  env <- new.env(hash = TRUE, parent = emptyenv())
  splitsList <- list()
  rawSplitsList <- list()
  countVec <- integer(0)
  treeMembers <- vector("list", nTree)
  nextIdx <- 1L

  for (i in seq_len(nTree)) {
    sp <- as.Splits(trees[[i]], tipLabels)
    logMat <- as.logical(sp)  # matrix: nSplits x nTip
    # The Splits object is a raw matrix internally; preserve structure
    rawMat <- unclass(sp)
    if (is.null(dim(logMat))) {
      logMat <- matrix(logMat, nrow = 1)
      rawMat <- matrix(rawMat, nrow = 1)
    }
    nSp <- nrow(logMat)
    # Canonicalise: ensure tip 1 is FALSE
    nSp <- nrow(logMat)
    members <- integer(nSp)
    for (j in seq_len(nSp)) {
      row <- logMat[j, ]
      rawRow <- rawMat[j, ]
      if (row[1]) {
        row <- !row
        rawRow <- .FlipRaw(rawRow, nTip)
      }
      key <- paste0(which(row), collapse = ",")
      idx <- env[[key]]
      if (is.null(idx)) {
        env[[key]] <- nextIdx
        splitsList[[nextIdx]] <- row
        rawSplitsList[[nextIdx]] <- rawRow
        countVec[nextIdx] <- 1L
        members[j] <- nextIdx
        nextIdx <- nextIdx + 1L
      } else {
        countVec[idx] <- countVec[idx] + 1L
        members[j] <- idx
      }
    }
    treeMembers[[i]] <- unique(members)
  }

  nSplits <- length(splitsList)
  splitMat <- matrix(FALSE, nrow = nSplits, ncol = nTip)
  nBytes <- length(rawSplitsList[[1]])
  rawMat <- matrix(as.raw(0), nrow = nSplits, ncol = nBytes)
  for (k in seq_len(nSplits)) {
    splitMat[k, ] <- splitsList[[k]]
    rawMat[k, ] <- rawSplitsList[[k]]
  }

  lightSide <- pmin(rowSums(splitMat), nTip - rowSums(splitMat))

  list(
    splits = splitMat,         # logical matrix nSplits x nTip
    rawSplits = rawMat,        # raw matrix nSplits x nBytes
    counts = countVec,
    lightSide = as.integer(lightSide),
    treeMembers = treeMembers
  )
}


.FlipRaw <- function(rawVec, nTip) {
  nBytes <- length(rawVec)
  usedBits <- ((nTip - 1L) %% 8L) + 1L
  lastMask <- as.raw(sum(2^(0:(usedBits - 1L))))
  out <- xor(rawVec, as.raw(0xff))
  out[nBytes] <- out[nBytes] & lastMask
  out
}


#' Pairwise transfer distance matrix using logical split matrix
#' Transfer distance = min(hamming(a, b), n - hamming(a, b))
#' hamming(a,b) when both are logical = sum(xor(a,b))
#' @noRd
.TransferDistMat <- function(splitMat, nTip) {
  # splitMat is logical: nSplits x nTip
  # hamming = number of differing positions
  # Use tcrossprod trick: hamming(a,b) = sum(a) + sum(b) - 2*sum(a&b)
  # = nAgreeing... actually let's just compute XOR directly.
  # Faster: hamming = nTip - 2 * (a %*% t(b) + (1-a) %*% t(1-b))
  #       = nTip - 2 * (a %*% t(b) + (nTip - rowSums(a) - colSums(b) + a %*% t(b)))
  # Simpler: agreement = a %*% t(b) + (1-a) %*% t(1-b)
  #                    = 2 * (a %*% t(b)) - rowSums(a) - rep(rowSums(b)) + nTip
  # hamming = nTip - agreement
  sm <- splitMat + 0L  # convert to integer
  ab <- tcrossprod(sm)  # sm %*% t(sm)
  rs <- rowSums(sm)
  hamming <- nTip - 2L * ab + outer(rs, rs, "+") - nTip
  # Simplifies to: hamming = outer(rs, rs, "+") - 2 * ab
  # Wait, let me re-derive:
  # agreement_ij = sum(a_i == b_j) = sum(a&b) + sum(!a & !b)
  #              = ab[i,j] + (nTip - rs[i] - rs[j] + ab[i,j])
  #              = 2*ab[i,j] + nTip - rs[i] - rs[j]
  # hamming_ij = nTip - agreement_ij = rs[i] + rs[j] - 2*ab[i,j]
  hamming <- outer(rs, rs, "+") - 2L * ab

  # Transfer distance = min(hamming, nTip - hamming)
  pmin(hamming, nTip - hamming)
}


#' Compute TD (transfer dissimilarity cost) for each split
#' @noRd
.ComputeTD <- function(DIST, sentDist, treeMembers, lightSide, nTree, scale) {
  nSplits <- nrow(DIST)
  TD <- numeric(nSplits)
  pMinus1 <- lightSide - 1L

  for (i in seq_len(nTree)) {
    idx <- treeMembers[[i]]
    # For each split b, min distance to any split in tree i
    if (length(idx) == 1L) {
      minD <- DIST[, idx]
    } else {
      minD <- apply(DIST[, idx, drop = FALSE], 1, min)
    }
    # Also consider sentinel distance
    minD <- pmin(minD, sentDist)
    if (scale) {
      TD <- TD + pmin(minD / pMinus1, 1)
    } else {
      TD <- TD + pmin(minD, pMinus1)
    }
  }
  TD
}


#' Compatibility matrix via vectorized bipartition check
#' @noRd
.CompatMat <- function(splitMat, nTip) {
  # Two splits compatible iff one of the 4 intersections (A&B, A&~B, ~A&B, ~A&~B) is empty
  sm <- splitMat + 0L  # integer 0/1 matrix
  notSm <- 1L - sm

  # A&B: a[i,] & b[j,] -> any nonzero -> tcrossprod(sm, sm) > 0
  ab   <- tcrossprod(sm, sm) > 0L
  anb  <- tcrossprod(sm, notSm) > 0L
  nab  <- tcrossprod(notSm, sm) > 0L
  nanb <- tcrossprod(notSm, notSm) > 0L

  # Compatible if at least one intersection is empty
  !ab | !anb | !nab | !nanb
}


#' Initialize MATCH/MATCH2 from currently included splits
#' @noRd
.InitMatches <- function(st, DIST, sentDist, lightSide, scale) {
  nSplits <- length(st$MATCH)
  incIdx <- which(st$incl)
  pMinus1 <- lightSide - 1L

  if (length(incIdx) == 0L) return(invisible())

  # For each split b, find 1st and 2nd closest among included
  subDIST <- DIST[, incIdx, drop = FALSE]
  for (b in seq_len(nSplits)) {
    dists <- subDIST[b, ]
    ord <- order(dists)
    bestDist <- dists[ord[1]]
    thresh <- pMinus1[b]
    if (scale && bestDist / thresh >= 1) next
    if (!scale && bestDist >= thresh) next
    st$MATCH[b] <- incIdx[ord[1]]
    if (length(ord) > 1L) {
      secDist <- dists[ord[2]]
      if (scale && secDist / thresh < 1) {
        st$MATCH2[b] <- incIdx[ord[2]]
      } else if (!scale && secDist < thresh) {
        st$MATCH2[b] <- incIdx[ord[2]]
      }
    }
  }
}


#' Get distance from split b to its match (NA = sentinel)
#' @noRd
.Dist <- function(b, idx, DIST, sentDist) {
  if (is.na(idx)) sentDist[b] else DIST[b, idx]
}


#' Vectorized add-benefit: returns benefit for each candidate
#' @noRd
.AddBenefitVec <- function(candidates, st, DIST, sentDist, TD, counts,
                           lightSide, scale) {
  nSplits <- length(st$MATCH)
  pMinus1 <- lightSide - 1L

  # Current match distances for all splits
  matchDist <- ifelse(is.na(st$MATCH), sentDist, .DiagDist(st$MATCH, DIST, sentDist))

  benefits <- numeric(length(candidates))
  for (ci in seq_along(candidates)) {
    c <- candidates[ci]
    newDist <- DIST[, c]
    if (scale) {
      diff <- (matchDist - newDist) / pMinus1
    } else {
      diff <- matchDist - newDist
    }
    diff[diff < 0] <- 0
    benefits[ci] <- sum(diff * counts) - TD[c]
  }
  benefits
}

#' Helper: get DIST\[b, MATCH\[b\]\] for all b, vectorized
#' @noRd
.DiagDist <- function(matchVec, DIST, sentDist) {
  n <- length(matchVec)
  out <- numeric(n)
  notNA <- !is.na(matchVec)
  if (any(notNA)) {
    out[notNA] <- DIST[cbind(which(notNA), matchVec[notNA])]
  }
  out[!notNA] <- sentDist[!notNA]
  out
}


#' Vectorized remove-benefit
#' @noRd
.RemoveBenefitVec <- function(candidates, st, DIST, sentDist, TD, counts,
                              lightSide, scale) {
  nSplits <- length(st$MATCH)
  pMinus1 <- lightSide - 1L

  # For remove, only splits whose MATCH == candidate are affected
  benefits <- numeric(length(candidates))
  matchDist <- .DiagDist(st$MATCH, DIST, sentDist)
  match2Dist <- .DiagDist(st$MATCH2, DIST, sentDist)

  for (ci in seq_along(candidates)) {
    c <- candidates[ci]
    affected <- st$MATCH == c & !is.na(st$MATCH)
    if (any(affected)) {
      if (scale) {
        fn_cost <- sum((DIST[affected, c] - match2Dist[affected]) /
                         pMinus1[affected] * counts[affected])
      } else {
        fn_cost <- sum((DIST[affected, c] - match2Dist[affected]) *
                         counts[affected])
      }
    } else {
      fn_cost <- 0
    }
    benefits[ci] <- TD[c] + fn_cost
  }
  benefits
}


.DoAdd <- function(branchIdx, st, DIST, sentDist) {
  st$incl[branchIdx] <- TRUE
  nSplits <- length(st$MATCH)

  curMatchDist <- .DiagDist(st$MATCH, DIST, sentDist)
  newDist <- DIST[, branchIdx]

  better <- newDist < curMatchDist
  if (any(better)) {
    st$MATCH2[better] <- st$MATCH[better]
    st$MATCH[better] <- branchIdx
  }

  # Check if it becomes second match for others
  notBetter <- !better
  secDist <- .DiagDist(st$MATCH2, DIST, sentDist)
  betterSec <- notBetter & (newDist < secDist)
  if (any(betterSec)) {
    st$MATCH2[betterSec] <- branchIdx
  }
}


.DoRemove <- function(branchIdx, st, DIST, sentDist, lightSide, scale) {
  st$incl[branchIdx] <- FALSE
  nSplits <- length(st$MATCH)
  curInc <- which(st$incl)
  pMinus1 <- lightSide - 1L

  # Splits whose first match was branchIdx
  affected1 <- which(st$MATCH == branchIdx & !is.na(st$MATCH))
  if (length(affected1)) {
    st$MATCH[affected1] <- st$MATCH2[affected1]
    for (b in affected1) {
      if (is.na(st$MATCH[b])) {
        # Promoted value was sentinel — rescan for actual closest
        st$MATCH[b] <- .FindSecond(b, NA_integer_, curInc, DIST,
                                    pMinus1, scale)
      }
      # Find new second match
      st$MATCH2[b] <- .FindSecond(b, st$MATCH[b], curInc, DIST,
                                  pMinus1, scale)
    }
  }

  # Splits whose second match was branchIdx
  affected2 <- which(st$MATCH2 == branchIdx & !is.na(st$MATCH2))
  if (length(affected2)) {
    for (b in affected2) {
      st$MATCH2[b] <- .FindSecond(b, st$MATCH[b], curInc, DIST,
                                  pMinus1, scale)
    }
  }
}


.FindSecond <- function(b, matchIdx, curInc, DIST, pMinus1, scale) {
  cands <- if (is.na(matchIdx)) curInc else setdiff(curInc, matchIdx)
  if (length(cands) == 0L) return(NA_integer_)
  dists <- DIST[b, cands]
  best <- cands[which.min(dists)]
  bestDist <- DIST[b, best]
  if (scale && bestDist / pMinus1[b] >= 1) return(NA_integer_)
  if (!scale && bestDist >= pMinus1[b]) return(NA_integer_)
  best
}


.IsCompat <- function(idx, incl, compat, nTip) {
  incIdx <- which(incl)
  if (length(incIdx) == 0L) return(TRUE)
  if (length(incIdx) >= nTip - 3L) return(FALSE)
  all(compat[idx, incIdx])
}


.GreedyBest <- function(st, DIST, sentDist, TD, counts, lightSide,
                        compat, sortOrd, scale, nSplits, nTip) {
  repeat {
    # Evaluate all candidates
    addCands <- integer(0)
    remCands <- integer(0)
    for (idx in sortOrd) {
      if (st$incl[idx]) {
        remCands <- c(remCands, idx)
      } else if (.IsCompat(idx, st$incl, compat, nTip)) {
        addCands <- c(addCands, idx)
      }
    }

    bestBen <- 0
    bestIdx <- 0L
    bestAction <- ""

    if (length(addCands)) {
      addBen <- .AddBenefitVec(addCands, st, DIST, sentDist, TD, counts,
                               lightSide, scale)
      mx <- max(addBen)
      if (mx > bestBen) {
        bestBen <- mx
        bestIdx <- addCands[which.max(addBen)]
        bestAction <- "add"
      }
    }
    if (length(remCands)) {
      remBen <- .RemoveBenefitVec(remCands, st, DIST, sentDist, TD, counts,
                                  lightSide, scale)
      mx <- max(remBen)
      if (mx > bestBen) {
        bestBen <- mx
        bestIdx <- remCands[which.max(remBen)]
        bestAction <- "remove"
      }
    }

    if (bestBen <= 0) break
    if (bestAction == "add") {
      .DoAdd(bestIdx, st, DIST, sentDist)
    } else {
      .DoRemove(bestIdx, st, DIST, sentDist, lightSide, scale)
    }
  }
}


.GreedyFirst <- function(st, DIST, sentDist, TD, counts, lightSide,
                         compat, sortOrd, scale, nSplits, nTip) {
  improving <- TRUE
  while (improving) {
    improving <- FALSE
    matchDist <- .DiagDist(st$MATCH, DIST, sentDist)
    match2Dist <- .DiagDist(st$MATCH2, DIST, sentDist)
    pMinus1 <- lightSide - 1L

    for (idx in sortOrd) {
      if (st$incl[idx]) {
        # Quick remove benefit
        affected <- st$MATCH == idx & !is.na(st$MATCH)
        if (any(affected)) {
          if (scale) {
            fn <- sum((DIST[affected, idx] - match2Dist[affected]) /
                        pMinus1[affected] * counts[affected])
          } else {
            fn <- sum((DIST[affected, idx] - match2Dist[affected]) *
                        counts[affected])
          }
        } else {
          fn <- 0
        }
        if (TD[idx] + fn > 0) {
          .DoRemove(idx, st, DIST, sentDist, lightSide, scale)
          improving <- TRUE
          break
        }
      } else if (.IsCompat(idx, st$incl, compat, nTip)) {
        newDist <- DIST[, idx]
        if (scale) {
          diff <- pmax((matchDist - newDist) / pMinus1, 0)
        } else {
          diff <- pmax(matchDist - newDist, 0)
        }
        if (sum(diff * counts) - TD[idx] > 0) {
          .DoAdd(idx, st, DIST, sentDist)
          improving <- TRUE
          break
        }
      }
    }
  }
}


.SplitsToPhylo <- function(rawSplits, included, tipLabels, nTip) {
  selectedIdx <- which(included)
  if (length(selectedIdx) == 0L) {
    return(TreeTools::StarTree(tipLabels))
  }
  selectedRaw <- rawSplits[selectedIdx, , drop = FALSE]
  sp <- structure(selectedRaw, nTip = nTip, tip.label = tipLabels,
                  class = "Splits")
  as.phylo(sp)
}

test_that("Identical trees return fully resolved consensus", {
  tree <- as.phylo(0, nTip = 10)
  trees <- structure(rep(list(tree), 10), class = "multiPhylo")

  tc <- TransferConsensus(trees, greedy = "first")
  expect_equal(NSplits(tc), NSplits(tree))

  tc_best <- TransferConsensus(trees, greedy = "best")
  expect_equal(NSplits(tc_best), NSplits(tree))
})

test_that("Star tree returned when no signal", {
  set.seed(6129)
  # Fully random trees with many tips — negligible split overlap
  trees <- structure(lapply(1:30, function(i) TreeTools::RandomTree(50, root = TRUE)),
                     class = "multiPhylo")
  tc <- TransferConsensus(trees, greedy = "first")

  # Should be very unresolved (0 or near-0 splits)
  expect_lte(NSplits(tc), 5)
})

test_that("Transfer consensus is at least as resolved as majority-rule for structured trees",
{
  # Trees from as.phylo with moderate overlap
  trees <- as.phylo(0:29, nTip = 20)
  tc <- TransferConsensus(trees, greedy = "best")
  mr <- Consensus(trees, p = 0.5)

  # Transfer consensus should be at least as resolved
  expect_gte(NSplits(tc), NSplits(mr))
})

test_that("Both greedy strategies produce valid trees", {
  trees <- as.phylo(1:15, nTip = 12)

  tc_first <- TransferConsensus(trees, greedy = "first")
  tc_best  <- TransferConsensus(trees, greedy = "best")

  expect_s3_class(tc_first, "phylo")
  expect_s3_class(tc_best, "phylo")
  expect_equal(sort(TipLabels(tc_first)), sort(TipLabels(trees[[1]])))
  expect_equal(sort(TipLabels(tc_best)), sort(TipLabels(trees[[1]])))
})

test_that("scale = FALSE (unscaled) works", {
  trees <- as.phylo(1:10, nTip = 10)

  tc_scaled <- TransferConsensus(trees, scale = TRUE)
  tc_unscaled <- TransferConsensus(trees, scale = FALSE)

  expect_s3_class(tc_scaled, "phylo")
  expect_s3_class(tc_unscaled, "phylo")
})

test_that("init = 'majority' works", {
  trees <- as.phylo(0:19, nTip = 15)

  tc_empty <- TransferConsensus(trees, init = "empty")
  tc_maj   <- TransferConsensus(trees, init = "majority")

  expect_s3_class(tc_empty, "phylo")
  expect_s3_class(tc_maj, "phylo")
  # Both should produce reasonable trees
  expect_gte(NSplits(tc_empty), 1)
  expect_gte(NSplits(tc_maj), 1)
})

test_that("Error on bad input", {
  expect_error(TransferConsensus(list(TreeTools::BalancedTree(5))), "multiPhylo")
  expect_error(TransferConsensus(
    structure(list(TreeTools::BalancedTree(5)), class = "multiPhylo")),
    "at least 2")

  # Fewer than 4 tips → star tree
  tiny <- structure(rep(list(ape::read.tree(text = "(a,b,c);")), 3),
                    class = "multiPhylo")
  tc_tiny <- TransferConsensus(tiny)
  expect_equal(NSplits(tc_tiny), 0)

  # tc_profile validation
  expect_error(TreeDist:::tc_profile(list(TreeTools::BalancedTree(5))),
               "multiPhylo")
  expect_error(TreeDist:::tc_profile(
    structure(list(TreeTools::BalancedTree(5)), class = "multiPhylo")),
    "at least 2")
  expect_error(TreeDist:::tc_profile(tiny), "at least 4 tips")
})

test_that("Two-tree consensus returns a valid tree", {
  trees <- as.phylo(1:2, nTip = 8)
  tc <- TransferConsensus(trees)
  expect_s3_class(tc, "phylo")
  expect_equal(length(TipLabels(tc)), 8)
})


# =========================================================================
# C++ edge cases (transfer_consensus.cpp coverage)
# =========================================================================

test_that("TransferConsensus returns star tree for all-star input", {
  star <- StarTree(8)
  trees <- structure(rep(list(star), 3), class = "multiPhylo")
  tc <- TransferConsensus(trees)
  expect_s3_class(tc, "phylo")
  expect_equal(NSplits(tc), 0)
})

test_that("TransferConsensus covers all parameter combinations", {
  trees <- as.phylo(0:9, nTip = 10)

  # greedy="first", init="majority", scale=FALSE
  tc <- TransferConsensus(trees, greedy = "first", init = "majority",
                          scale = FALSE)
  expect_s3_class(tc, "phylo")

  # greedy="best", init="majority", scale=FALSE
  tc2 <- TransferConsensus(trees, greedy = "best", init = "majority",
                           scale = FALSE)
  expect_s3_class(tc2, "phylo")

  # greedy="first", init="empty", scale=FALSE
  tc3 <- TransferConsensus(trees, greedy = "first", init = "empty",
                           scale = FALSE)
  expect_s3_class(tc3, "phylo")
})

test_that("Greedy remove path fires with conflicting majority splits", {
  # Two groups of trees with very different topologies: majority-init seeds
  # splits from the larger group, some of which the greedy then removes.
  t1 <- as.phylo(0, nTip = 12)
  t2 <- as.phylo(100000, nTip = 12)
  trees <- structure(c(rep(list(t1), 6), rep(list(t2), 5)),
                     class = "multiPhylo")

  # greedy="best" exercises do_remove + find_second in GreedyState
  tc_best_s <- TransferConsensus(trees, greedy = "best", init = "majority",
                                 scale = TRUE)
  expect_s3_class(tc_best_s, "phylo")

  tc_best_u <- TransferConsensus(trees, greedy = "best", init = "majority",
                                 scale = FALSE)
  expect_s3_class(tc_best_u, "phylo")

  # greedy="first" exercises the same remove path via greedy_first

  tc_first_s <- TransferConsensus(trees, greedy = "first", init = "majority",
                                  scale = TRUE)
  expect_s3_class(tc_first_s, "phylo")

  tc_first_u <- TransferConsensus(trees, greedy = "first", init = "majority",
                                  scale = FALSE)
  expect_s3_class(tc_first_u, "phylo")
})

test_that("Greedy exercises diverse topologies (extra C++ path coverage)", {
  # Random trees with many tips — sparse split overlap
  set.seed(5123)
  rand_trees <- structure(
    lapply(1:30, function(i) TreeTools::RandomTree(10, root = TRUE)),
    class = "multiPhylo")
  # Diverse indexed trees
  diverse_trees <- as.phylo(seq(0, 500, by = 25), nTip = 20)
  # Three conflicting groups
  mixed_trees <- structure(c(
    rep(list(as.phylo(0, 8)), 5),
    rep(list(as.phylo(300, 8)), 5),
    rep(list(as.phylo(9999, 8)), 3)
  ), class = "multiPhylo")

  for (trees in list(rand_trees, diverse_trees, mixed_trees)) {
    for (gr in c("best", "first")) {
      for (init in c("empty", "majority")) {
        for (sc in c(TRUE, FALSE)) {
          tc <- TransferConsensus(trees, greedy = gr, init = init, scale = sc)
          expect_s3_class(tc, "phylo")
        }
      }
    }
  }
})

test_that("tc_profile() runs without error", {
  trees <- as.phylo(0:4, nTip = 8)

  # Greedy best, empty init
  res <- TreeDist:::tc_profile(trees, scale = TRUE, greedy = "best",
                               init = "empty", n_iter = 1L)
  expect_length(res, 5)
  expect_true(all(res >= 0))
  expect_equal(names(res),
               c("pool_splits", "transfer_dist_mat", "compute_td",
                 "compat_mat", "greedy"))

  # Greedy first, majority init, unscaled
  res2 <- TreeDist:::tc_profile(trees, scale = FALSE, greedy = "first",
                                init = "majority")
  expect_length(res2, 5)
  expect_true(all(res2 >= 0))

  # Multiple iterations
  res3 <- TreeDist:::tc_profile(trees, n_iter = 3L)
  expect_length(res3, 5)

  # Greedy best, majority init
  res4 <- TreeDist:::tc_profile(trees, init = "majority")
  expect_length(res4, 5)
})


# =========================================================================
# R-level internal helper functions (transfer_consensus.R lines 94-551)
# =========================================================================

test_that(".FlipRaw() flips bits and masks correctly", {
  # 8 tips (1 byte, all bits used, mask = 0xFF)
  expect_equal(TreeDist:::.FlipRaw(as.raw(0x07), 8), as.raw(0xf8))

  # 5 tips (1 byte, 5 bits, mask = 0x1F)
  expect_equal(TreeDist:::.FlipRaw(as.raw(0x07), 5), as.raw(0x18))

  # Multi-byte: 10 tips (2 bytes, last byte 2 bits, mask = 0x03)
  result <- TreeDist:::.FlipRaw(as.raw(c(0x00, 0x01)), 10)
  expect_equal(result, as.raw(c(0xff, 0x02)))
})

test_that(".TransferDistMat() computes pairwise transfer distances", {
  splitMat <- matrix(c(
    TRUE, TRUE, FALSE, FALSE, FALSE, FALSE,
    TRUE, TRUE, TRUE, FALSE, FALSE, FALSE,
    FALSE, FALSE, FALSE, TRUE, TRUE, TRUE
  ), nrow = 3, byrow = TRUE)

  DIST <- TreeDist:::.TransferDistMat(splitMat, 6)
  expect_equal(dim(DIST), c(3, 3))
  expect_equal(diag(DIST), c(0, 0, 0))
  expect_true(isSymmetric(DIST))
  # {1,2} vs {1,2,3}: hamming = 1
  expect_equal(DIST[1, 2], 1)
  # {1,2,3} vs {4,5,6}: complements → transfer = 0
  expect_equal(DIST[2, 3], 0)

  # Single split
  DIST1 <- TreeDist:::.TransferDistMat(
    matrix(c(TRUE, TRUE, FALSE, FALSE), nrow = 1), 4)
  expect_equal(dim(DIST1), c(1, 1))
  expect_equal(DIST1[1, 1], 0)
})

test_that(".CompatMat() checks bipartition compatibility", {
  splitMat <- matrix(c(
    TRUE, TRUE, FALSE, FALSE, FALSE, FALSE,
    FALSE, FALSE, TRUE, TRUE, FALSE, FALSE,
    TRUE, FALSE, TRUE, FALSE, FALSE, FALSE
  ), nrow = 3, byrow = TRUE)

  compat <- TreeDist:::.CompatMat(splitMat, 6)
  expect_equal(dim(compat), c(3, 3))
  expect_true(compat[1, 2])    # {1,2} and {3,4}: disjoint → compatible
  expect_false(compat[1, 3])   # {1,2} and {1,3}: all 4 intersections → incompatible
  expect_true(isSymmetric(compat))
})

test_that(".ComputeTD() works for both scale modes", {
  DIST <- matrix(c(0, 1, 3, 1, 0, 2, 3, 2, 0), 3, 3)
  lightSide <- c(3, 2, 3)
  sentDist <- lightSide - 1L
  treeMembers <- list(c(1, 2), c(2, 3))

  td_s <- TreeDist:::.ComputeTD(DIST, sentDist, treeMembers, lightSide, 2, TRUE)
  td_u <- TreeDist:::.ComputeTD(DIST, sentDist, treeMembers, lightSide, 2, FALSE)
  expect_length(td_s, 3)
  expect_length(td_u, 3)
  expect_true(all(td_s >= 0))
  expect_true(all(td_u >= 0))

  # Single member per tree
  td_single <- TreeDist:::.ComputeTD(
    DIST, sentDist, list(c(1), c(3)), lightSide, 2, TRUE)
  expect_length(td_single, 3)
})

test_that(".IsCompat() handles edge cases", {
  compat <- matrix(TRUE, 3, 3)
  compat[1, 3] <- compat[3, 1] <- FALSE

  # Empty included → always compatible
  expect_true(TreeDist:::.IsCompat(1, c(FALSE, FALSE, FALSE), compat, 10))
  # Too many included (>= nTip - 3) → never compatible
  expect_false(TreeDist:::.IsCompat(1, c(TRUE, TRUE, TRUE), compat, 6))
  # Compatible with included set
  expect_true(TreeDist:::.IsCompat(1, c(FALSE, TRUE, FALSE), compat, 10))
  # Incompatible with an included split
  expect_false(TreeDist:::.IsCompat(1, c(FALSE, FALSE, TRUE), compat, 10))
})

test_that(".Dist() returns correct distance", {
  DIST <- matrix(c(0, 2, 3, 2, 0, 4, 3, 4, 0), 3, 3)
  sentDist <- c(5, 6, 7)

  expect_equal(TreeDist:::.Dist(1, 2, DIST, sentDist), 2)
  expect_equal(TreeDist:::.Dist(1, NA, DIST, sentDist), 5)
  expect_equal(TreeDist:::.Dist(3, NA, DIST, sentDist), 7)
})

test_that(".DiagDist() vectorizes correctly", {
  DIST <- matrix(c(0, 2, 3, 2, 0, 4, 3, 4, 0), 3, 3)
  sentDist <- c(5, 6, 7)

  expect_equal(
    TreeDist:::.DiagDist(c(2, NA, 1), DIST, sentDist),
    c(DIST[1, 2], 6, DIST[3, 1])
  )
  expect_equal(TreeDist:::.DiagDist(c(NA, NA, NA), DIST, sentDist), sentDist)
  expect_equal(
    TreeDist:::.DiagDist(c(2, 3, 1), DIST, sentDist),
    c(DIST[1, 2], DIST[2, 3], DIST[3, 1])
  )
})

test_that(".FindSecond() works for both scale modes and edge cases", {
  DIST <- matrix(c(0, 1, 3, 1, 0, 2, 3, 2, 0), 3, 3)
  pMinus1 <- c(4, 4, 4)

  # Unscaled: exclude matchIdx, find best
  expect_equal(
    TreeDist:::.FindSecond(1, 2, c(1, 2, 3), DIST, pMinus1, FALSE), 1)

  # matchIdx = NA → search all
  expect_equal(
    TreeDist:::.FindSecond(1, NA_integer_, c(2, 3), DIST, pMinus1, FALSE), 2)

  # No candidates
  expect_true(is.na(
    TreeDist:::.FindSecond(1, 2, c(2), DIST, pMinus1, FALSE)))

  # All exceed threshold (unscaled)
  DIST_far <- matrix(c(0, 5, 5, 5, 0, 5, 5, 5, 0), 3, 3)
  expect_true(is.na(
    TreeDist:::.FindSecond(1, NA_integer_, c(2, 3), DIST_far, c(2, 2, 2), FALSE)
  ))

  # Scaled: normal case
  expect_equal(
    TreeDist:::.FindSecond(1, NA_integer_, c(2, 3), DIST, pMinus1, TRUE), 2)

  # Scaled: exceeds threshold
  expect_true(is.na(
    TreeDist:::.FindSecond(1, NA_integer_, c(2), DIST_far, c(2, 2, 2), TRUE)
  ))
})

test_that(".DoAdd() updates match state correctly", {
  DIST <- matrix(c(0, 1, 3, 1, 0, 2, 3, 2, 0), 3, 3)
  sentDist <- c(4, 4, 4)

  st <- new.env(parent = emptyenv())
  st$MATCH <- rep(NA_integer_, 3)
  st$MATCH2 <- rep(NA_integer_, 3)
  st$incl <- rep(FALSE, 3)

  # First add: all splits match the newly added split
  TreeDist:::.DoAdd(2, st, DIST, sentDist)
  expect_true(st$incl[2])
  expect_equal(st$MATCH[1], 2)
  expect_equal(st$MATCH[2], 2)
  expect_equal(st$MATCH[3], 2)

  # Second add: closer splits update, old match demotes to second
  TreeDist:::.DoAdd(1, st, DIST, sentDist)
  expect_equal(st$MATCH[1], 1)    # self-match closer
  expect_equal(st$MATCH2[1], 2)   # old match becomes second
  expect_equal(st$MATCH[2], 2)    # split 2 still closest to itself
  expect_equal(st$MATCH2[2], 1)   # split 1 becomes second
})

test_that(".DoRemove() handles sentinel promotion and affected2", {
  DIST <- matrix(c(0, 1, 3, 1, 0, 2, 3, 2, 0), 3, 3)
  sentDist <- c(4, 4, 4)
  lightSide <- c(5, 5, 5)

  # Case 1: sentinel promotion (MATCH2 = NA, remove MATCH, no other included)
  st <- new.env(parent = emptyenv())
  st$MATCH <- c(2L, NA_integer_, 2L)
  st$MATCH2 <- c(NA_integer_, NA_integer_, NA_integer_)
  st$incl <- c(FALSE, TRUE, FALSE)

  TreeDist:::.DoRemove(2, st, DIST, sentDist, lightSide, FALSE)
  expect_true(is.na(st$MATCH[1]))
  expect_true(is.na(st$MATCH[3]))

  # Case 2: sentinel promotion with rescan finding another match
  st2 <- new.env(parent = emptyenv())
  st2$MATCH <- c(2L, 2L, NA_integer_)
  st2$MATCH2 <- c(NA_integer_, NA_integer_, NA_integer_)
  st2$incl <- c(TRUE, TRUE, FALSE)

  TreeDist:::.DoRemove(2, st2, DIST, sentDist, lightSide, FALSE)
  expect_equal(st2$MATCH[1], 1L)  # rescanned, found split 1

  # Case 3: affected2 path (MATCH2 == removed, MATCH != removed)
  st3 <- new.env(parent = emptyenv())
  st3$MATCH <- c(1L, 1L, 1L)
  st3$MATCH2 <- c(2L, 2L, NA_integer_)
  st3$incl <- c(TRUE, TRUE, FALSE)

  TreeDist:::.DoRemove(2, st3, DIST, sentDist, lightSide, FALSE)
  expect_equal(st3$MATCH, c(1L, 1L, 1L))  # unchanged
  expect_true(is.na(st3$MATCH2[1]))
  expect_true(is.na(st3$MATCH2[2]))

  # Case 4: scaled mode
  st4 <- new.env(parent = emptyenv())
  st4$MATCH <- c(2L, NA_integer_, NA_integer_)
  st4$MATCH2 <- c(NA_integer_, NA_integer_, NA_integer_)
  st4$incl <- c(TRUE, TRUE, FALSE)

  TreeDist:::.DoRemove(2, st4, DIST, sentDist, lightSide, TRUE)
  expect_equal(st4$MATCH[1], 1L)
})

test_that(".AddBenefitVec() computes add benefits", {
  DIST <- matrix(c(0, 1, 3, 1, 0, 2, 3, 2, 0), 3, 3)
  lightSide <- c(3, 3, 3)
  sentDist <- lightSide - 1L
  counts <- c(2, 3, 1)
  TD <- c(0.5, 0.3, 0.8)

  st <- new.env(parent = emptyenv())
  st$MATCH <- rep(NA_integer_, 3)
  st$MATCH2 <- rep(NA_integer_, 3)
  st$incl <- rep(FALSE, 3)

  ben_s <- TreeDist:::.AddBenefitVec(
    c(1, 2, 3), st, DIST, sentDist, TD, counts, lightSide, scale = TRUE)
  expect_length(ben_s, 3)

  ben_u <- TreeDist:::.AddBenefitVec(
    c(1, 2, 3), st, DIST, sentDist, TD, counts, lightSide, scale = FALSE)
  expect_length(ben_u, 3)
})

test_that(".RemoveBenefitVec() computes remove benefits", {
  DIST <- matrix(c(0, 1, 3, 1, 0, 2, 3, 2, 0), 3, 3)
  lightSide <- c(3, 3, 3)
  sentDist <- lightSide - 1L
  counts <- c(2, 3, 1)
  TD <- c(0.5, 0.3, 0.8)

  st <- new.env(parent = emptyenv())
  st$MATCH <- c(2L, 1L, 2L)
  st$MATCH2 <- c(NA_integer_, NA_integer_, NA_integer_)
  st$incl <- c(TRUE, TRUE, FALSE)

  ben_s <- TreeDist:::.RemoveBenefitVec(
    c(1, 2), st, DIST, sentDist, TD, counts, lightSide, scale = TRUE)
  expect_length(ben_s, 2)

  ben_u <- TreeDist:::.RemoveBenefitVec(
    c(1, 2), st, DIST, sentDist, TD, counts, lightSide, scale = FALSE)
  expect_length(ben_u, 2)

  # No affected splits (MATCH doesn't reference candidate)
  st2 <- new.env(parent = emptyenv())
  st2$MATCH <- c(1L, 1L, 1L)
  st2$MATCH2 <- rep(NA_integer_, 3)
  st2$incl <- c(TRUE, TRUE, FALSE)

  ben_none <- TreeDist:::.RemoveBenefitVec(
    c(2), st2, DIST, sentDist, TD, counts, lightSide, scale = TRUE)
  # Only TD contribution, no fn_cost (no affected splits)
  expect_equal(ben_none, TD[2])
})

test_that(".PoolSplits() pools and deduplicates splits", {
  trees <- as.phylo(0:4, nTip = 8)
  tipLabels <- TipLabels(trees[[1]])
  pool <- TreeDist:::.PoolSplits(trees, tipLabels)

  expect_true(is.matrix(pool$splits))
  expect_true(is.matrix(pool$rawSplits))
  expect_equal(ncol(pool$splits), length(tipLabels))
  expect_equal(length(pool$counts), nrow(pool$splits))
  expect_equal(length(pool$lightSide), nrow(pool$splits))
  expect_equal(length(pool$treeMembers), length(trees))
  expect_true(all(pool$counts >= 1))
  expect_true(all(pool$lightSide >= 1))
})

test_that(".SplitsToPhylo() converts splits to tree or star", {
  trees <- as.phylo(0:2, nTip = 8)
  tipLabels <- TipLabels(trees[[1]])
  pool <- TreeDist:::.PoolSplits(trees, tipLabels)

  # Some splits included
  included <- rep(FALSE, nrow(pool$rawSplits))
  included[1:min(2, nrow(pool$rawSplits))] <- TRUE
  tr <- TreeDist:::.SplitsToPhylo(pool$rawSplits, included, tipLabels, 8)
  expect_s3_class(tr, "phylo")
  expect_equal(sort(TipLabels(tr)), sort(tipLabels))

  # No splits → star tree
  tr_star <- TreeDist:::.SplitsToPhylo(
    pool$rawSplits, rep(FALSE, nrow(pool$rawSplits)), tipLabels, 8)
  expect_s3_class(tr_star, "phylo")
  expect_equal(NSplits(tr_star), 0)
})

test_that(".InitMatches() initializes match state from included splits", {
  # DIST: included = {1,2}; split 3 close to split 2 (dist 1 < thresh 2)
  DIST <- matrix(c(
    0, 1, 3, 5,
    1, 0, 1, 4,
    3, 1, 0, 1,
    5, 4, 1, 0
  ), 4, 4, byrow = TRUE)
  lightSide <- c(3, 3, 3, 3)
  sentDist <- lightSide - 1L  # all 2

  # Scaled mode
  st <- new.env(parent = emptyenv())
  st$MATCH <- rep(NA_integer_, 4)
  st$MATCH2 <- rep(NA_integer_, 4)
  st$incl <- c(TRUE, TRUE, FALSE, FALSE)

  TreeDist:::.InitMatches(st, DIST, sentDist, lightSide, scale = TRUE)
  # Split 1: self-match (dist 0), second = split 2 (dist 1)
  expect_equal(st$MATCH[1], 1)
  expect_equal(st$MATCH2[1], 2)
  # Split 3: best among included is split 2 (dist 1), 1/2 < 1 → matched
  expect_equal(st$MATCH[3], 2)
  # Split 4: best among included is split 2 (dist 4), 4/2 >= 1 → skipped
  expect_true(is.na(st$MATCH[4]))

  # Unscaled mode
  st2 <- new.env(parent = emptyenv())
  st2$MATCH <- rep(NA_integer_, 4)
  st2$MATCH2 <- rep(NA_integer_, 4)
  st2$incl <- c(TRUE, TRUE, FALSE, FALSE)

  TreeDist:::.InitMatches(st2, DIST, sentDist, lightSide, scale = FALSE)
  expect_equal(st2$MATCH[1], 1)
  expect_equal(st2$MATCH[3], 2)

  # Empty included set → no-op
  st3 <- new.env(parent = emptyenv())
  st3$MATCH <- rep(NA_integer_, 4)
  st3$MATCH2 <- rep(NA_integer_, 4)
  st3$incl <- rep(FALSE, 4)

  TreeDist:::.InitMatches(st3, DIST, sentDist, lightSide, scale = TRUE)
  expect_true(all(is.na(st3$MATCH)))
})


# =========================================================================
# Full R-level greedy pipeline integration tests
# =========================================================================

test_that("R-level GreedyBest pipeline produces valid consensus (scaled)", {
  trees <- as.phylo(0:9, nTip = 8)
  tipLabels <- TipLabels(trees[[1]])
  nTip <- length(tipLabels)
  nTree <- length(trees)

  pool <- TreeDist:::.PoolSplits(trees, tipLabels)
  nSplits <- nrow(pool$splits)
  DIST <- TreeDist:::.TransferDistMat(pool$splits, nTip)
  sentDist <- pool$lightSide - 1L
  TD <- TreeDist:::.ComputeTD(DIST, sentDist, pool$treeMembers, pool$lightSide,
                               nTree, scale = TRUE)
  compat <- TreeDist:::.CompatMat(pool$splits, nTip)
  sortOrd <- order(-pool$counts, seq_len(nSplits))

  st <- new.env(parent = emptyenv())
  st$MATCH <- rep(NA_integer_, nSplits)
  st$MATCH2 <- rep(NA_integer_, nSplits)
  st$incl <- rep(FALSE, nSplits)

  TreeDist:::.GreedyBest(st, DIST, sentDist, TD, pool$counts, pool$lightSide,
                          compat, sortOrd, scale = TRUE, nSplits, nTip)
  expect_true(any(st$incl))

  tr <- TreeDist:::.SplitsToPhylo(pool$rawSplits, st$incl, tipLabels, nTip)
  expect_s3_class(tr, "phylo")
})

test_that("R-level GreedyFirst pipeline produces valid consensus (unscaled)", {
  trees <- as.phylo(0:9, nTip = 8)
  tipLabels <- TipLabels(trees[[1]])
  nTip <- length(tipLabels)
  nTree <- length(trees)

  pool <- TreeDist:::.PoolSplits(trees, tipLabels)
  nSplits <- nrow(pool$splits)
  DIST <- TreeDist:::.TransferDistMat(pool$splits, nTip)
  sentDist <- pool$lightSide - 1L
  TD <- TreeDist:::.ComputeTD(DIST, sentDist, pool$treeMembers, pool$lightSide,
                               nTree, scale = FALSE)
  compat <- TreeDist:::.CompatMat(pool$splits, nTip)
  sortOrd <- order(-pool$counts, seq_len(nSplits))

  st <- new.env(parent = emptyenv())
  st$MATCH <- rep(NA_integer_, nSplits)
  st$MATCH2 <- rep(NA_integer_, nSplits)
  st$incl <- rep(FALSE, nSplits)

  TreeDist:::.GreedyFirst(st, DIST, sentDist, TD, pool$counts, pool$lightSide,
                           compat, sortOrd, scale = FALSE, nSplits, nTip)
  expect_true(any(st$incl))

  tr <- TreeDist:::.SplitsToPhylo(pool$rawSplits, st$incl, tipLabels, nTip)
  expect_s3_class(tr, "phylo")
})

test_that("R-level GreedyBest with majority init exercises InitMatches", {
  trees <- as.phylo(0:14, nTip = 10)
  tipLabels <- TipLabels(trees[[1]])
  nTip <- length(tipLabels)
  nTree <- length(trees)

  pool <- TreeDist:::.PoolSplits(trees, tipLabels)
  nSplits <- nrow(pool$splits)
  DIST <- TreeDist:::.TransferDistMat(pool$splits, nTip)
  sentDist <- pool$lightSide - 1L
  TD <- TreeDist:::.ComputeTD(DIST, sentDist, pool$treeMembers, pool$lightSide,
                               nTree, scale = TRUE)
  compat <- TreeDist:::.CompatMat(pool$splits, nTip)
  sortOrd <- order(-pool$counts, seq_len(nSplits))

  st <- new.env(parent = emptyenv())
  st$MATCH <- rep(NA_integer_, nSplits)
  st$MATCH2 <- rep(NA_integer_, nSplits)

  # Majority init: include splits present in > half of trees
  st$incl <- pool$counts > (nTree / 2)
  if (any(st$incl)) {
    TreeDist:::.InitMatches(st, DIST, sentDist, pool$lightSide, scale = TRUE)
  }

  TreeDist:::.GreedyBest(st, DIST, sentDist, TD, pool$counts, pool$lightSide,
                          compat, sortOrd, scale = TRUE, nSplits, nTip)

  tr <- TreeDist:::.SplitsToPhylo(pool$rawSplits, st$incl, tipLabels, nTip)
  expect_s3_class(tr, "phylo")
})

test_that("R-level GreedyFirst with majority init (unscaled)", {
  trees <- as.phylo(0:14, nTip = 10)
  tipLabels <- TipLabels(trees[[1]])
  nTip <- length(tipLabels)
  nTree <- length(trees)

  pool <- TreeDist:::.PoolSplits(trees, tipLabels)
  nSplits <- nrow(pool$splits)
  DIST <- TreeDist:::.TransferDistMat(pool$splits, nTip)
  sentDist <- pool$lightSide - 1L
  TD <- TreeDist:::.ComputeTD(DIST, sentDist, pool$treeMembers, pool$lightSide,
                               nTree, scale = FALSE)
  compat <- TreeDist:::.CompatMat(pool$splits, nTip)
  sortOrd <- order(-pool$counts, seq_len(nSplits))

  st <- new.env(parent = emptyenv())
  st$MATCH <- rep(NA_integer_, nSplits)
  st$MATCH2 <- rep(NA_integer_, nSplits)
  st$incl <- pool$counts > (nTree / 2)
  if (any(st$incl)) {
    TreeDist:::.InitMatches(st, DIST, sentDist, pool$lightSide, scale = FALSE)
  }

  TreeDist:::.GreedyFirst(st, DIST, sentDist, TD, pool$counts, pool$lightSide,
                           compat, sortOrd, scale = FALSE, nSplits, nTip)

  tr <- TreeDist:::.SplitsToPhylo(pool$rawSplits, st$incl, tipLabels, nTip)
  expect_s3_class(tr, "phylo")
})

test_that("R greedy pipeline matches C++ consensus", {
  trees <- as.phylo(0:9, nTip = 8)
  tipLabels <- TipLabels(trees[[1]])
  nTip <- length(tipLabels)
  nTree <- length(trees)

  # C++ result
  tc_cpp <- TransferConsensus(trees, greedy = "best", scale = TRUE)

  # R-level result
  pool <- TreeDist:::.PoolSplits(trees, tipLabels)
  nSplits <- nrow(pool$splits)
  DIST <- TreeDist:::.TransferDistMat(pool$splits, nTip)
  sentDist <- pool$lightSide - 1L
  TD <- TreeDist:::.ComputeTD(DIST, sentDist, pool$treeMembers, pool$lightSide,
                               nTree, scale = TRUE)
  compat <- TreeDist:::.CompatMat(pool$splits, nTip)
  sortOrd <- order(-pool$counts, seq_len(nSplits))

  st <- new.env(parent = emptyenv())
  st$MATCH <- rep(NA_integer_, nSplits)
  st$MATCH2 <- rep(NA_integer_, nSplits)
  st$incl <- rep(FALSE, nSplits)

  TreeDist:::.GreedyBest(st, DIST, sentDist, TD, pool$counts, pool$lightSide,
                          compat, sortOrd, scale = TRUE, nSplits, nTip)

  tc_r <- TreeDist:::.SplitsToPhylo(pool$rawSplits, st$incl, tipLabels, nTip)

  # Both should produce resolved trees with same tip labels
  expect_s3_class(tc_r, "phylo")
  expect_equal(sort(TipLabels(tc_r)), sort(TipLabels(tc_cpp)))
  # Resolution should be similar (not necessarily identical due to
  # possible differences in canonicalization/sort-order between R and C++)
  expect_gte(NSplits(tc_r), 1)
})

#' Path distance
#' 
#' Calculate the path distance between rooted or unrooted trees.
#' 
#' This function is a faster alternative to the function 
#' \code{\link[phangorn:treedist]{path.dist()}} in the phangorn package,
#' which can crash if the internal representation of trees does not conform to
#' certain (unspecified) expectations, and which treats all trees as unrooted.
#' 
#' The path distance is calculated by tabulating the cladistic difference (=
#' topological distance) between each pair of tips in each tree.
#' A precursor to the path distance \insertCite{Farris1969}{TreeDist}
#' took the mean squared 
#' difference between the elements of each tree's tabulation (Farris, 1973);
#' the method used here is that proposed by
#' \insertCite{Steel1993;textual}{TreeDist}, which takes the square root of this
#' sum.
#' Other precursor measures are described in 
#' \insertCite{Williams1971;textual}{TreeDist} and
#' \insertCite{Phipps1971;textual}{TreeDist}.
#' 
#' If a root node is present, trees are treated as rooted.
#' To avoid counting the root edge twice, use `UnrootTree(tree)` before
#' calculating the path distance.
#' 
#' Use of the path distance is discouraged as it emphasizes 
#' shallow relationships at the expense of deeper (and arguably more
#' fundamental) relationships \insertCite{Farris1973}{TreeDist}.
#' 
#' @template tree12ListParams
#' 
#' @return `PathDist()` returns a vector or distance matrix of distances
#' between trees.
#' 
#' @examples
#' library("TreeTools")
#' 
#' # Treating the two edges to the root node as distinct
#' PathDist(BalancedTree(7), PectinateTree(7))
#' 
#' # Counting those two edges once
#' PathDist(UnrootTree(BalancedTree(7)), UnrootTree(PectinateTree(7)))
#' 
#' PathDist(BalancedTree(7), as.phylo(0:2, 7))
#' PathDist(as.phylo(0:2, 7), PectinateTree(7))
#'
#' PathDist(list(bal = BalancedTree(7), pec = PectinateTree(7)),
#'         as.phylo(0:2, 7))
#'
#' PathDist(as.phylo(30:33, 8))
#'  
#' @references \insertAllCited{}
#' 
#' @template MRS
#' @family tree distances
#' @importFrom TreeTools Postorder
#' @export
PathDist <- function(tree1, tree2 = NULL) {
  if (inherits(tree1, "phylo")) {
    if (inherits(tree2, "phylo")) {
      .PathDist11(tree1, tree2)
    } else {
      .PathDist1Many(tree1, tree2)
    }
  } else if (is.null(tree2)) {
    .PathDistManySelf(tree1)
  } else if (inherits(tree2, "phylo")) {
    .PathDist1Many(tree2, tree1)
  } else {
    .PathDistManyMany(tree1, tree2)
  }
}

.EuclideanDistance <- function(x) sqrt(sum(x * x))

.PathDist11 <- function(tree1, tree2) {
  .EuclideanDistance(PathVector(tree1) - PathVector(RenumberTips(tree2, tree1)))
}

.PathDist1Many <- function(tree1, treeMany) {
  v1 <- PathVector(tree1)
  apply(v1 - vapply(RenumberTips(treeMany, tree1), PathVector, v1), 2,
        .EuclideanDistance)
}

.PathDistManyMany <- function(trees1, trees2) {
  nTip <- NTip(trees1[[1]])
  v1 <- vapply(RenumberTips(trees1, trees1), PathVector,
               integer(nTip * (nTip - 1) / 2))
  v2 <- vapply(RenumberTips(trees2, trees1), PathVector,
               integer(nTip * (nTip - 1) / 2))
  vec_diff_euclidean(v1, v2)
}

.PathDistManySelf <- function(trees) {
  nTip <- NTip(trees[[1]])
  v1 <- vapply(RenumberTips(trees, trees), PathVector,
               integer(nTip * (nTip - 1) / 2))
  
  nTree <- length(trees)
  
  ret <- structure(pair_diff_euclidean(v1),
                   Size = nTree, Diag = FALSE, Upper = FALSE,
                   class = "dist")
  
  # Return:
  ret
}

library("TreeTools")
test_that("Median is calculated", {
  tenTrees <- as.phylo(1:10, 8L)
  expect_equal(4, median(tenTrees, index = TRUE))
  expect_equal(as.phylo(4, 8), median(tenTrees))
  expect_equal(as.phylo(4, 8), median(tenTrees, breakTies = FALSE))
  expect_equal(as.phylo(3, 8),
               median(tenTrees, Distance = RobinsonFoulds, breakTies = TRUE))
  expect_equal(c(3:5, 7), median(tenTrees, Distance = RobinsonFoulds, 
                                 index = TRUE, breakTies = FALSE))
  expect_equal(as.phylo(c(3:5, 7), 8),
               median(tenTrees, Distance = RobinsonFoulds, index = FALSE, 
                      breakTies = FALSE),
               ignore_attr = TRUE)
  expect_equal(as.phylo(c(3:5, 7), 8),
               median(tenTrees, Distance = RobinsonFoulds, breakTies = FALSE),
               ignore_attr = TRUE)
})

---
title: "Comparing sets of trees from different analyses"
author: "[Martin R. Smith](https://smithlabdurham.github.io/)"
output: 
  rmarkdown::html_vignette:
    fig_width: 6
    fig_height: 4
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa-old-doi-prefix.csl
vignette: >
  %\VignetteIndexEntry{Comparing sets of trees from different analyses}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

A common application of tree space analysis is to compare the outputs of
different analyses – for instance, trees obtained from different gene sequences,
or results obtained using different models or methods 
(e.g. Bayesian, maximum likelihood, or parsimony).

## Shiny app

This can be accomplished quickly using the `MapTrees()` graphical user interface:

- <span style="border: solid 2px #e69f00">Load trees</span> from file: Select first tree file

- Select an appropriate sample size

- Select <span style="border: solid 2px #009e73">Replace existing</span>

- Load each additional set of trees from file using <span style="border: solid 2px #d55e00">Add batch to existing</span>

![Load tree batches](Batch1.svg){ width="245px" }


- On the <span style="border: solid 2px #e69f00">Display</span> tab, 
  select <span style="border: solid 2px #009e73">Point symbols: One per batch</span>,
  or <span style="border: solid 2px #d55e00">Colour points by: Batch</span>


![Tree batch styles](Batch2.svg){ width=100% }


## Scripting at the R command line

More control over the mapping can be obtained at the command line:

```{r generate-trees}
# Load trees
library("TreeTools", quietly = TRUE)
batch1 <- as.phylo(1:60, 8) # Generate 60 similar trees 
batch2 <- as.phylo(seq(200, 800, length.out = 30), 8) # A separate batch of 30 trees
styles <- c(1, 2) # Select plotting colours / symbols
treeStyle <- rep(styles, c(length(batch1), length(batch2)))

# Calculate distances
library("TreeDist")
distances <- ClusteringInfoDistance(c(batch1, batch2))

# Construct over-simple 2D PCoA mapping
mapping <- cmdscale(distances, k = 2)
```

```{r plot-mapping, fig.align = "center"}
# Plot mapping
par(mar = rep(0, 4))
plot(mapping,
     asp = 1, # Preserve aspect ratio - do not distort distances
     ann = FALSE, axes = FALSE, # Don't label axes: dimensions are meaningless
     col = treeStyle, # Colour
     pch = treeStyle # Plotting symbol
     )
legend("left", c("Batch 1", "Batch 2"), col = styles, pch = styles)
```

For more robust analyses than the (potentially misleading!) 2D plot above,
consult the companion [vignette](treespace.html).
Note also that mapped areas and their regions of overlap may not correspond
to reality; see the warnings and recommendations in @SmithSpace.

## Comparing trees' dispersal / hypervolume

Interpreting and comparing the areas of tree space from a projection can be
misleading -- the expanded apparent area of Greenland under the Mercator projection
being a familiar example.

![Mercator map projection, with true size of each country overlaid. Credit: [Neil Kaye](https://x.com/neilrkaye)](figures/mercator-projection.jpg)

As such, it is always best to work with original distances when interpreting
whether sets of trees occupy larger or smaller regions of tree space.

### Distances from median

One approach is to plot distances from a median tree:

```{r compare-dist-from-median, fig.align = "center"}
# Calculate median trees
median1 <- median(batch1)
median2 <- median(batch2)

# Compute distance from each tree to the median of its batch
dist1 <- ClusteringInfoDist(batch1, median1)
dist2 <- ClusteringInfoDist(batch2, median2)

# Set resolution of histogram
nBreaks <- 10
breaks <- seq(0, max(dist1, dist2), length.out = nBreaks)

# Plot first distance set
hist(dist1, col = "#00000022", breaks = breaks,
     main = "Distance from median of batch",
     xlab = "Clustering information distance",
     ylim = c(0, 25) # Omit this line to infer Y axis limit from first batch.
     )

# Add second distance set
hist(dist2, col = "#ff000022", breaks = breaks, add = TRUE)

# Add legend
legend("topleft", c("Batch 1", "Batch 2"),
       fill = c("#00000022",  "#ff000022"))
```

In the plotted example, distances to the median tree are greater for batch 2
than batch 1, indicating a more dispersed set of trees that occupies a greater
hypervolume.  Note that the increased frequency at higher distances is expected:
the outer shell of a sphere contains more volume than a layer of equivalent
thickness closer to the centre, and this phenomenon becomes more pronounced as
the dimensionality of tree space increases.

### Consensus resolution

A complementary approach is to identify the resolution of the consensus of each
batch of trees.  This approach shares many of the 
[problems with the Robinson--Foulds distance](
https://ms609.github.io/TreeDist/articles/Robinson-Foulds.html): in particular,
resolution can be decimated by a single "rogue" taxon whose position is poorly
defined [@SmithRogue].
Detecting and removing rogue taxa can provide a more meaningful point of
comparison.

```{r rogue-detection}
# Create tree set with a rogue taxon
batch3 <- AddTipEverywhere(as.phylo(7, 7), "t8")

# Set up plotting area
par(mfrow = c(2, 2), mar = rep(0.4, 4))

# Plot naive strict consensus
plot(consensus(batch1, p = 1))
plot(consensus(batch3, p = 1))

if (requireNamespace("Rogue", quietly = TRUE)) {
  cons1 <- ConsensusWithout(batch1, p = 1,
                            Rogue::QuickRogue(batch1, p = 1)[-1, "taxon"])
  cons3 <- ConsensusWithout(batch3, p = 1,
                            Rogue::QuickRogue(batch3, p = 1)[-1, "taxon"])
  
  # The information content of each tree gives a measure of its resolution,
  # accounting for omitted rogue leaves
  SplitwiseInfo(cons1) # 8.5 bits
  SplitwiseInfo(cons3) # 15.1 bits: higher resolution indicates that these
  # trees are more similar, notwithstanding rogue taxa.
  
  # Plot the trees
  plot(cons1)
  plot(cons3)
} else {
  message("The package 'Rogue' is required to run this example.")
}
```

Whereas a direct interpretation of this analysis is not straightforward, it can
provide a complementary way of understanding the distribution of trees
across tree space.

## References

---
title: "Contextualizing tree distances"
author: "[Martin R. Smith](https://smithlabdurham.github.io/)"
output: rmarkdown::html_vignette
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa-old-doi-prefix.csl
vignette: >
  %\VignetteIndexEntry{Contextualizing tree distances}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

Once you understand [how to use "TreeDist"](Using-TreeDist.html) to calculate
tree distances, the next step is to provide some context for the calculated
distances.

## Normalizing

The maximum value of most tree distance metrics scales with the size of the
trees being compared.
Typically, the resolution of the trees also impacts the range of possible
values.
As such, it can be difficult to interpret the tree distance value without 
suitable context. 

Normalizing a distance metric is one way to render its meaning more obvious.
Selecting an appropriate normalizing constant may require careful consideration
of the purpose to which a tree distance metric is being put.

The default normalization behaviour of each function when `normalize = TRUE` is
listed in the [function reference](
https://ms609.github.io/TreeDist/reference/index.html), or can be viewed by
typing `?FunctionName` in the R terminal.

### Nye _et al._ tree similarity

Let's work through a simple example using the Nye _et al_. [-@Nye2006]
similarity metric to compare two imperfectly-resolved trees.

```{r, fig.width=6, out.width="90%", fig.align="center"}
library("TreeDist")
tree1 <- ape::read.tree(text = '(A, ((B, ((C, D), (E, F))), (G, (H, (I, J, K)))));')
tree2 <- ape::read.tree(text = '(A, (B, (C, D, E, (J, K)), (F, (G, H, I))));')
VisualizeMatching(NyeSimilarity, tree1, tree2,
                  Plot = TreeDistPlot, matchZeros = FALSE)
```

This is a nice metric to start with, because the maximum similarity between 
each pair of splits is defined as one. 
(Astute readers might worry that the minimum similarity is greater than zero --
that's a harder problem to overcome.)

As such, the maximum similarity possible between two 11-leaf trees is 
[`NSplits(11)`](https://ms609.github.io/TreeTools/reference/NSplits.html)
= `r suppressMessages(library("TreeTools")); NSplits(11)`.  

Normalizing against this value tells us how similar the two trees are, compared
to two identical eleven-leaf binary trees.

```{r}
NyeSimilarity(tree1, tree2, normalize = FALSE) / 8
NyeSimilarity(tree1, tree2, normalize = 8)
```

This approach will result in a similarity score less than one if two trees 
are identical, but not fully resolved (i.e. binary).  
As such, we might prefer to compare the tree similarity to the maximum
score possible for two trees of the specified resolution.  This value is given
by the number of splits in the least resolved of the two trees:

```{r}
NyeSimilarity(tree1, tree2,
                  normalize = min(TreeTools::NSplits(list(tree1, tree2))))
```

More concisely, we can provide a normalizing function:

```{r}
NyeSimilarity(tree1, tree2, normalize = min)
```

This approach will produce a similarity of one if one tree is 
a less-resolved version of another (and thus not identical).

If we are comparing lists of trees, this best value will depend on the number
of splits in each pair of trees.  We can use the function `pmin()` to 
select the less resolved of each pair of trees:


```{r}
NyeSimilarity(list(tree1, tree2), list(tree1, tree2), normalize = pmin)
```

To avoid these limitations, we may instead opt to normalize against the 
average number of splits in the two trees. 
This is the default normalization method for
[`NyeSimilarity()`](../reference/NyeSimilarity.html):

```{r}
NyeSimilarity(tree1, tree2, normalize = TRUE)
```

Finally, if `tree1` is a "target" tree -- perhaps one that has been used
to simulate data from, or which is independently known to be true or virtuous --
we may wish to normalize against the best possible match to that tree.

In that case, the best possible score is

```{r}
TreeTools::NSplits(tree1)
```

and our normalized score will be 

```{r}
NyeSimilarity(tree1, tree2, normalize = TreeTools::NSplits(tree1))
```

### Normalizing to random similarity

The diameter (maximum possible distance) of the Nye _et al_. tree similarity
metric is easy to calculate, but this is not the case for all metrics.
For example, the clustering information distance metric [@SmithDist]
ranges in principle from zero to the total clustering entropy present in a 
pair of trees.
But with even a modest number of leaves, no pairs of trees exist in which every
split in one tree is perfectly contradicted by every other split in the other;
as such, any pair of trees will necessarily have some degree of similarity.  
In such a context, it can be relevant to normalize tree similarity against the
_expected_ distance between a pair of random trees, rather than a maximum value
[see @Vinh2010].
On this measure, distances greater than one denote trees that are more different
than expected by chance, whereas a distance of zero denotes identity.

With the quartet divergence, the expected tree distance is readily calculated:
any given quartet has a one in three chance of matching by chance.

```{r}
if (requireNamespace("Quartet", quietly = TRUE)) {
  library("Quartet", exclude = "RobinsonFoulds")
  expectedQD <- 2 / 3
  normalizedQD <- QuartetDivergence(QuartetStatus(tree1, tree2),
                                    similarity = FALSE) / expectedQD
}
```

The expected distance is more difficult to calculate for other metrics, but 
can be approximated by sampling random pairs of trees.  Measured distances
between 10&nbsp;000 pairs of random bifurcating trees with up to 200 leaves are 
available in the data package 
'[TreeDistData](https://github.com/ms609/TreeDistData/)'.
We can view (normalized) distances for a selection of methods:

```{r, fig.width=7, fig.height=4, message=FALSE}
if (requireNamespace("TreeDistData", quietly = TRUE)) {
  library("TreeDistData", exclude = "PairwiseDistances")
  data("randomTreeDistances", package = "TreeDistData")
  
  methods <- c("pid", "cid", "nye", "qd")
  methodCol <- c(pid = "#e15659", cid = "#58a14e", nye = "#edc949",
                 qd = "#af7aa1")
  
  oldPar <- par(cex = 0.7, mar = c(5, 5, 0.01, 0.01))
  nLeaves <- as.integer(dimnames(randomTreeDistances)[[3]])
  plot(nLeaves, type = "n", randomTreeDistances["pid", "mean", ],
       ylim = c(0.54, 1),
       xlab = "Number of leaves",
       ylab = "Normalized distance between random tree pairs")
  
  for (method in methods) {
    dat <- randomTreeDistances[method, , ]
    lines(nLeaves, dat["50%", ], pch = 1, col = methodCol[method])
    polygon(c(nLeaves, rev(nLeaves)), c(dat["25%", ], rev(dat["75%", ])),
            border = NA, col = paste0(methodCol[method], "55"))
    
  }
  
  text(202, randomTreeDistances[methods, "50%", "200"] + 0.02, 
       c("Different phylogenetic information", 
         "Clustering information distance",
         expression(paste(plain("Nye "), italic("et al."))),
         "Quartet divergence"
         ), col = methodCol[methods], pos = 2)
  par(oldPar)
}
```

or use these calculated values to normalize our tree distance:

```{r, eval = FALSE}
expectedCID <- randomTreeDistances["cid", "mean", "9"]
ClusteringInfoDistance(tree1, tree2, normalize = TRUE) / expectedCID
```

## Testing similarity to a known tree

Similarity has two components: precision and accuracy [@Smith2019].
A tree can be 80% similar to a target tree because it contains 80% of the splits
in the target tree, and no incorrect splits -- or because it is a binary tree 
in which 10% of the splits present are resolved incorrectly and are thus 
positively misleading.

In such a comparison, of course, it is more sensible to talk about split
_information_ than just the number of splits: an even split may
contain more information than two very uneven splits, so the absence of 
two information-poor splits may be preferable to the absence of one 
information-rich split.  As such, it is most instructive to think of the
proportion of information that has been correctly resolved: the goal is to find
a tree that is as informative as possible about the true tree.

Ternary diagrams allow us to visualise the quality of a reconstructed tree
with reference to a known "true" tree:

```{r fig.align="center", fig.height=1.8, fig.width=6, out.width="80%"}
testTrees <- list(
  trueTree = ape::read.tree(text = '(a, (b, (c, (d, (e, (f, (g, h)))))));'),
  lackRes = ape::read.tree(text = '(a, (b, c, (d, e, (f, g, h))));'),
  smallErr = ape::read.tree(text = '(a, (c, (b, (d, (f, (e, (g, h)))))));'),
  bigErr = ape::read.tree(text = '(a, (c, (((b, d), (f, h)), (e, g))));')
)
VisualizeMatching(MutualClusteringInfo, testTrees$trueTree, testTrees$lackRes)
points(4, 7.5, pch = 2, cex = 3, col = "#E69F00", xpd = NA)
VisualizeMatching(MutualClusteringInfo, testTrees$trueTree, testTrees$smallErr)
points(4, 7.5, pch = 3, cex = 3, col = "#56B4E9", xpd = NA)
VisualizeMatching(MutualClusteringInfo, testTrees$trueTree, testTrees$bigErr)
points(4, 7.5, pch = 4, cex = 3, col = "#009E73", xpd = NA)
```

Better trees plot vertically towards the "100% shared information" vertex. 
Resolution of trees increases towards the right; trees that are more
resolved may be no better than less-resolved trees if the addition of 
resolution introduces error.

```{r, fig.width=4, fig.align="center", fig.asp=1, out.width="50%"}
if (requireNamespace("Ternary", quietly = TRUE)) {
  library("Ternary")
  oldPar <- par(mar = rep(0.1, 4))
  TernaryPlot(alab = "Absent information", blab = "Shared information",
              clab = "Misinformation",
              lab.cex = 0.8, lab.offset = 0.18,
              point = "left", clockwise = FALSE,
              grid.col = "#dedede", grid.minor.lines = 0,
              axis.labels = 0:10 / 10, axis.col = "#aaaaaa")
  
  HorizontalGrid()
  correct <- MutualClusteringInfo(testTrees$trueTree, testTrees)
  resolved <- ClusteringEntropy(testTrees)
  unresolved <- resolved["trueTree"] - resolved
  incorrect <- resolved - correct
  TernaryPoints(cbind(unresolved, correct, incorrect), 
                pch = 1:4, cex = 2, col = Ternary::cbPalette8[1:4])
  par(oldPar)
}
```


### Example

Here's a noddy real-world example applying this to a simulation-style study.

First, let's generate a starting tree, which will represent our reference
topology:

```{r}
set.seed(0)
trueTree <- TreeTools::RandomTree(20, root = TRUE)
```

Then, let's generate 200 degraded trees.
We'll move away from the true tree by making a TBR move, then reduce 
resolution by taking the consensus of this tree and three trees from its 
immediate neighbourhood (one NNI move away).

```{r}
treeSearchInstalled <- requireNamespace("TreeSearch", quietly = TRUE)
if (treeSearchInstalled) {
  library("TreeSearch", quietly = TRUE) # for TBR, NNI
  oneAway <- structure(lapply(seq_len(200), function(x) {
    tbrTree <- TBR(trueTree)
    ape::consensus(list(tbrTree,
                        NNI(tbrTree),
                        NNI(tbrTree),
                        NNI(tbrTree)))
  }), class = "multiPhylo")
} else {
  message("Install \"TreeSearch\" to run this example")
}
```

And let's generate 200 more trees that are even more degraded.
This time we'll move further (three TBR moves) from the true tree, and reduce 
resolution by taking a consensus with three trees from its wider neighbourhood
(each two NNI moves away).

```{r}
if (treeSearchInstalled) {
  threeAway <- structure(lapply(seq_len(200), function(x) {
    tbrTree <- TBR(TBR(TBR(trueTree)))
    ape::consensus(list(tbrTree, 
                        NNI(NNI(tbrTree)),
                        NNI(NNI(tbrTree)),
                        NNI(NNI(tbrTree))))
  }), class = "multiPhylo")
}
```

Now let's calculate their tree similarity scores.
We need to calculate the amount of information each tree has in common with
the true tree:

```{r}
if (treeSearchInstalled) {
  correct1 <- MutualClusteringInfo(trueTree, oneAway)
  correct3 <- MutualClusteringInfo(trueTree, threeAway)
}
```

The amount of information in each degraded tree:
```{r}
if (treeSearchInstalled) {
  infoInTree1 <- ClusteringEntropy(oneAway)
  infoInTree3 <- ClusteringEntropy(threeAway)
}
```

The amount of information that could have been resolved, but was not:
```{r}
if (treeSearchInstalled) {
  unresolved1 <- ClusteringEntropy(trueTree) - infoInTree1
  unresolved3 <- ClusteringEntropy(trueTree) - infoInTree3
}
```

And the amount of information incorrectly resolved:
```{r}
if (treeSearchInstalled) {
  incorrect1 <- infoInTree1 - correct1
  incorrect3 <- infoInTree3 - correct3
}
```

In preparation for our plot, let's colour our one-away trees 
<span style='background-color:#F5CCB599'>&nbsp;orange&nbsp;</span>, and our
three-away trees 
<span style='background-color:#9CE0E699'>&nbsp;blue&nbsp;</span>:

```{r, collapse=TRUE}
col1 <- hcl(200, alpha = 0.9)
col3 <- hcl(40, alpha = 0.9)
spec1 <- matrix(col2rgb(col1, alpha = TRUE), nrow = 4, ncol = 181)
spec3 <- matrix(col2rgb(col3, alpha = TRUE), nrow = 4, ncol = 181)
spec1[4, ] <- spec3[4, ] <- 0:180
ColToHex <- function(x) rgb(x[1], x[2], x[3], x[4], maxColorValue = 255)
spec1 <- apply(spec1, 2, ColToHex)
spec3 <- apply(spec3, 2, ColToHex)
```

Now we can plot this information on a ternary diagram.

```{r, fig.width=7, fig.align="center", fig.asp=5/7, out.width="70%"}
if (treeSearchInstalled && requireNamespace("Ternary", quietly = TRUE)) {
  layout(matrix(c(1, 2), ncol = 2), widths = c(5, 2))
  oldPar <- par(mar = rep(0, 4))
  TernaryPlot(alab = "Information absent in degraded tree", 
              blab = "\n\nCorrect information in degraded tree", 
              clab = "Misinformation in degraded tree",
              point = "left", clockwise = FALSE, grid.minor.lines = 0,
              axis.labels = 0:10 / 10)
  
  HorizontalGrid()
  
  coords1 <- cbind(unresolved1, correct1, incorrect1)
  coords3 <- cbind(unresolved3, correct3, incorrect3)
  
  ColourTernary(TernaryDensity(coords1, resolution = 20), spectrum = spec1)
  ColourTernary(TernaryDensity(coords3, resolution = 20), spectrum = spec3)
  
  TernaryDensityContour(coords3, col = col3, nlevels = 4)
  TernaryDensityContour(coords1, col = col1, nlevels = 4)
  
  if (requireNamespace("kdensity", quietly = TRUE)) {
    library("kdensity")
    HorizontalKDE <- function(dat, col, add = FALSE) {
      lty <- 1
      lwd <- 2
      kde <- kdensity(dat)
      kdeRange <- kdensity:::get_range(kde)
      if (add) {
        lines(kde(kdeRange), kdeRange, col = col, lty = lty, lwd = lwd)
      } else {
        plot(kde(kdeRange), kdeRange, col = col, lty = lty, lwd = lwd, 
             ylim = c(0, 1), main = "", axes = FALSE, type = "l")
      }
      # abline(h = 0:10 / 10) # Useful for confirming alignment
    }
  
    par(mar = c(1.8, 0, 1.8, 0)) # align plot limits with ternary plot
    HorizontalKDE(correct1 / infoInTree1, col1, add = FALSE)
    HorizontalKDE(correct3 / infoInTree3, col3, add = TRUE)
    mtext("\u2192 Normalized tree quality \u2192", 2)
  }
  par(oldPar)
} else {
  message("Install \"TreeSearch\" and \"Ternary\" to generate this plot")
}
```

In the ternary plot, the vertical direction corresponds to the normalized tree
quality, as depicted in the accompanying histogram.

## What next?

You may wish to:

- Explore the [Ternary package](https://ms609.github.io/Ternary/)

- [Interpret tree distance metrics](https://ms609.github.io/TreeDistData/articles/09-expected-similarity.html)

- Compare trees with [different tips](different-leaves.html)

- Review [available distance measures](https://ms609.github.io/TreeDist/index.html)
  and the corresponding [TreeDist functions](https://ms609.github.io/TreeDist/reference/index.html#section-tree-distance-measures)

- Construct [tree spaces](treespace.html) to visualize landscapes of
  phylogenetic trees


## References

---
title: "Trees with different leaves"
author: "[Martin R. Smith](https://smithlabdurham.github.io/)"
output: rmarkdown::html_vignette
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa-old-doi-prefix.csl
vignette: >
  %\VignetteIndexEntry{Trees with different leaves}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

Sometimes one may wish to compare trees whose tip sets only partially overlap:
that is, certain leaves are missing in one tree or the other.

Whilst this process is computationally trivial, interpreting the resultant
distances can require a little thought.

Let's generate some simple trees to follow this through in practice:

```{r initialize}
library("TreeTools", quietly = TRUE)
balAL <- BalancedTree(letters[1:12])
balAJ <- DropTip(balAL, letters[11:12])
balCL <- DropTip(balAL, letters[1:2])
balGL <- DropTip(balAL, letters[1:6])

pecAL <- PectinateTree(letters[1:12])
pecAF <- DropTip(pecAL, letters[7:12])
pecAJ <- DropTip(pecAL, letters[11:12])
pecCL <- DropTip(pecAL, letters[1:2])

treeList <- list(balAL = balAL, balAJ = balAJ, balCL = balCL, balGL = balGL,
                 pecAL = pecAL, pecAJ = pecAJ, pecCL = pecCL, pecAF = pecAF)

# Define a function to plot two trees
Plot2 <- function(t1, t2, ..., main2 = "") {
  oPar <- par(mfrow = c(1, 2), mar = rep(1, 4), cex = 0.9)
  on.exit(par(oPar)) # Restore original parameters
  plot(t1, ...)
  plot(t2, main = main2)
}
```


First let's consider the scenario where we are identifying two identical trees
-- except that some leaves present in one tree (here, `a` and `b`) are
missing from the other.

```{r drop-some-plot, fig.width = 8}
Plot2(balAL, balCL,
      main = "balAL", main2 = "balCL",
      font = c(rep(4, 2), rep(3, 10)) # Emphasize missing tips
      )
```

From an information theoretic perspective, all the information present in
our reduced tree is also present in the complete tree.
The information held in common between the trees is thus equal to the
information held in common if only the common leaves are retained:

```{r drop-some}
library("TreeDist")
commonTips <- intersect(TipLabels(balAL), TipLabels(balCL))

# How much information is in tree balCL?
ClusteringEntropy(balCL)
# All the information in balCL is also present in balAL
MutualClusteringInfo(balCL, balAL)
# balAL also contains information about leaves A and B,
# so contains more information in total
ClusteringEntropy(balAL)
```

_Some_ information is held in common between any pair of trees with
more than a few leaves.  Two random trees with many leaves may thus have
more information in common than two identical trees with only a few leaves,
simply because they have more information overall.
As such, it is clearly necessary to perform some form of normalization before
comparing tree distances.
The obvious choice -- and the default, if `normalize = TRUE` -- 
is to normalize against the maximum similarity that could be obtained,
given the set of leaves that a pair of trees have in common.
A `NaN` result is returned when a pair of trees has no leaves in common.

```{r full-list}
# Information in common (bits)
raw <- MutualClusteringInfo(treeList)
heatmap(raw, symm = TRUE)
# Normalized
normalized <- MutualClusteringInfo(treeList, normalize = TRUE)
heatmap(normalized, symm = TRUE)
```

A more nuanced perspective may be obtained by computing the range of distances
that could in principle be obtained if missing tips were added to each tree.
This is not yet implemented in 'TreeDist'.

---
title: "Generalized Robinson-Foulds distances"
author: "[Martin R. Smith](https://smithlabdurham.github.io/)"
output: rmarkdown::html_vignette
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa-old-doi-prefix.csl
vignette: >
  %\VignetteIndexEntry{Generalized Robinson-Foulds distances}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r init, message=FALSE, warning=FALSE, echo = FALSE}
library('ape')
library('TreeDist')
tree1 <- read.tree(text='((A, B), ((C, (D, E)), (F, (G, (H, I)))));')
tree2 <- read.tree(text='((A, B), ((C, D, (E, I)), (F, (G, H))));')
AtoJ <-  read.tree(text='(((((A, B), C), D), E), (F, (G, (H, (I, J)))));')
swapAJ <-  read.tree(text='(((((J, B), C), D), E), (F, (G, (H, (I, A)))));')
```

This document outlines the similarity measures employed by the 
generalized Robinson–Foulds distances implemented in this package.

Generalized RF distances are introduced [elsewhere](Robinson-Foulds.html); 
before you read further, you may also wish to revisit how to
[use the 'TreeDist' package](Using-TreeDist.html), and relevant 
[principles of information theory](information.html).

## Shared phylogenetic information

Under the shared phylogenetic information tree distance measure [@SmithDist], 
pairs of splits are assigned a similarity score that corresponds to the amount
of phylogenetic information [_sensu_ @Steel2006] that they share in common
(see [separate vignette](information.html)), a concept introduced (though not 
developed) by Nelson [-@Nelson1979].

```{r, fig.width=6, fig.align='center'}
VisualizeMatching(SharedPhylogeneticInfo, tree1, tree2, 
                  Plot = TreeDistPlot, matchZeros = FALSE)
SharedPhylogeneticInfo(tree1, tree2)
```

This distance is measured in bits; on this measure, the total information 
content of a tree is given by 
```{r, total-phylo-info}
SplitwiseInfo(tree1)
```

### Conflicting splits can nevertheless be instructive

Shared phylogenetic information assigns zero similarity to incompatible splits,
i.e. those that cannot both occur on a single tree.
This leads to problematic behaviour in certain cases: for example,
swapping the position of two distant leaves (as with leaves 'A' and 'J' below)
can disproportionately reduce similarity -- in this example, to zero.

```{r, fig.width=6, fig.align='center'}
VisualizeMatching(SharedPhylogeneticInfo, AtoJ, swapAJ,
                  Plot = TreeDistPlot, matchZeros = FALSE, prune = c(5, 18))
```

## Mutual clustering information

Scoring each pair of splits according to their mutual clustering information
[@SmithDist]  (see [separate vignette](information.html)) results in a
information-based tree distance metric that recognizes similarity in tree
structure even when every possible pairing of splits conflicts:

```{r, fig.width=6, fig.align='center'}
VisualizeMatching(MutualClusteringInfo, AtoJ, swapAJ,
                  Plot = TreeDistPlot, matchZeros = FALSE, prune = c(5, 18))
MutualClusteringInfo(AtoJ, swapAJ)
```

Because no pair of non-trivial splits has zero mutual clustering information,
even a dissimilar pairing (such as `HI|ABCDEFG` &Rightarrow; `EI|ABCDFGH` below) 
is (slightly) preferable to leaving a split unpaired.

```{r, fig.width=6, fig.align='center'}
VisualizeMatching(MutualClusteringInfo, tree1, tree2, 
                  Plot = TreeDistPlot, matchZeros = FALSE)
```

The total mutual clustering information in a single tree is given by
```{r total-mci}
ClusteringEntropy(tree1)
```

## Nye _et al._ tree similarity metric

The Nye _et al_. [-@Nye2006] tree similarity metric scores pairs by
considering the elements held in common between subsets of each split.

Consider a pair of splits `ABCDEF|GHIJ` and `ABCDEIJ|FGH`.
These can be aligned thus:

```
ABCDEF  | GHIJ
ABCDE IJ|FGH
```

The first pair of subsets, `ABCDEF` and `ABCDEIJ`, have five elements in common 
(`ABCDE`), and together encompass eight elements (`ABCDEFIJ`).  Their 
_subset score_ is thus $\frac{5}{8}$.

The second pair of subsets, `GHIJ` and `FGH`, have two elements (`GH`) in common,
of the five total (`FGHIJ`), and hence receive a subset score of $\frac{2}{5}$.

This split alignment then receives an _alignment score_ corresponding to the 
lower of the two subset scores, $\frac{2}{5}$.

We must now consider the other alignment of this pair of splits,


```
ABCDEF  |     GHIJ
     FGH|ABCDE  IJ
```

This yields subset scores of $\frac{1}{8}$ and $\frac{2}{9}$, and thus has an
alignment score of $\frac{1}{8}$.
This alignment gives a lower score than the other, so is disregarded.
The pair of splits is allocated a similarity score corresponding to the 
better alignment: $\frac{2}{5}$.

As such, splits that match exactly will receive a similarity score of 1,
in a manner analogous to the Robinson–Foulds distance.  (This is despite the
fact that some splits are more likely to match than others.)
It is not possible for a pair of splits to receive a zero similarity 
score.

```{r, fig.width=6, fig.align='center'}
VisualizeMatching(NyeSimilarity, tree1, tree2, 
                  Plot = TreeDistPlot, matchZeros = FALSE)
NyeSimilarity(tree1, tree2, normalize = FALSE)
```

### Jaccard–Robinson–Foulds metric

B&ouml;cker _et al_. [-@Bocker2013] employ the same split similarity calculation
as Nye _et al._ (above), which they suggest ought to be raised to an
arbitrary exponent in order to down-weight the contribution of paired splits 
that are not identical.
In order for the metric to converge to the Robinson–Foulds metric as the
exponent grows towards infinity, the resulting score is then doubled.

```{r, fig.width=6, fig.align='center'}
JaccardRobinsonFoulds(tree1, tree2, k = 1)
VisualizeMatching(JaccardRobinsonFoulds, tree1, tree2,
                  Plot = TreeDistPlot, matchZeros = FALSE)

JRF2 <- function(...) JaccardRobinsonFoulds(k = 2, ...)
JRF2(tree1, tree2)
VisualizeMatching(JRF2, tree1, tree2,
                  Plot = TreeDistPlot, matchZeros = FALSE)
```

The figure below shows how the JRF distance between the two trees plotted above
varies with the value of the exponent _k_, relative to the Nye _et al._ and 
Robinson–Foulds distances between the trees:

```{r, fig.width=6, fig.height = 4,fig.align = 'center', echo = FALSE}
oldPar <- par(mar = c(4, 4, 0, 0))
x <- seq(1, 20, by = 0.1)
cbPalette8 <- c("#000000", "#E69F00", "#56B4E9", "#009E73", "#F0E442",
                "#0072B2", "#D55E00", "#CC79A7") # dput(Ternary::cbPalette8)
plot(0, xlim = range(x), ylim = c(3.5, 9.5),
     xlab = expression(italic('k')),
     ylab = 'Jaccard\u2013Robinson\u2013Foulds distance',
     pch = 3, type = 'n')

abline(h = RobinsonFoulds(tree1, tree2),
       lty = 'dashed', col = cbPalette8[6])
abline(h = NyeSimilarity(tree1, tree2, similarity = FALSE),
       lty = 'dotted', col = cbPalette8[8])

lines(x = x, y = vapply(x, JaccardRobinsonFoulds, tree1 = tree1, tree2 = tree2,
                        allowConflict = TRUE, 0), col = cbPalette8[5])
lines(x = x, y = vapply(x, JaccardRobinsonFoulds, tree1 = tree1, tree2 = tree2,
                        0), col = cbPalette8[4])

legend('right', c('Robinson\u2013Foulds', 'JRF, no conflict', 'JRF, conflict ok', 'Nye et al.'),
       lty = c('dashed', 'solid', 'solid', 'dotted'), 
       col = cbPalette8[c(6, 4, 5, 8)], bty = 'n')
par(oldPar)
```

The theoretical and practical performance of the JRF metric, and its speed
of calculation, are best at lower values of _k_ [@SmithDist], raising the
question of whether an exponent is useful.

B&ouml;cker _et al_. [-@Bocker2013] suggest that 'reasonable' matchings 
exhibit a property they term _arboreality_.
Their definition of an arboreal matching supposes that trees are rooted, in 
which case each split corresponds to a clade.
In an arboreal matching on a rooted tree, no pairing of splits conflicts with
any other pairing of splits.
Consider the case where splits 'A' and 'B' in tree 1 are paired 
respectively with splits 'C' and 'D' in tree 2.  If 'A' is paired with 'C' 
and 'B' with 'D', then to avoid conflict:

- if A is nested within B, then C must be nested within D

- if B is nested within A, then D must be nested within C

- if A and B do not overlap, then C and D must not overlap.

Equivalent statements for unrooted trees are a little harder to express.

Unfortunately, constructing arboreal matchings is NP-complete, making an optimal
arboreal matching slow to find.
A faster alternative is to prohibit pairings of contradictory splits, though 
distances generated under such an approach are theoretically less coherent and 
practically no more effective than those calculated when contradictory splits
may be paired, so the only advantage of this approach is a slight increase in 
calculation speed [@SmithDist].

## Matching Split Distance

Bogdanowicz & Giaro [-@Bogdanowicz2012] propose an alternative distance, which
they term the Matching Split Distance.  (This approach was independently 
proposed by Lin _et al._ [-@Lin2012].)

```{r, fig.width=6, fig.align='center'}
MatchingSplitDistance(tree1, tree2)
VisualizeMatching(MatchingSplitDistance, tree1, tree2,
                  Plot = TreeDistPlot, matchZeros = FALSE)
```

Note that the visualization shows the difference, rather than the similarity,
between splits.

Similar to the Nye _et al_. similarity metric, this method compares the subsets
implied by a pair of splits.
Here, the relevant quantity is the number of elements that must be moved from 
one subset to another in order to make the two splits identical.
With the pair of splits

```
ABCDEF  | GHIJ
ABCDE IJ|FGH
```

three leaves ('F', 'I' and 'J') must be moved before the splits are identical;
as such, the pair of splits are assigned a difference score of three.  
Formally, where $S_i$ splits $n$ leaves into bipartitions $A_i$ and $B_i$, the
difference score is calculated by

$n - m$

where $m$ counts the number of leaves that already match, and is defined as

$m = \max\{|A_1 \cap A_2| + |B_1 \cap B_2|, |A_1 \cap B_2| + |B_1 \cap A_2|\}$

```{r, fig.width=6, fig.align='center'}
MatchingSplitDistance(read.tree(text='((a, b, c, d, e, f), (g, h, i, j));'),
                      read.tree(text='((a, b, c, d, e, i, j), (g, h, f));'))
```

This distance is difficult to normalize, as it is not easy to calculate its
maximum value.

### Information theoretic alternative

In the matching split distance, $m$ represents a simple count of the number of
shared taxa.  An alternative is to measure the phylogenetic information content
of the largest split consistent with $S_1$ and $S_2$:

$m = \max\{h(A_1 \cap A_2 | B_1 \cap B_2), h(A_1 \cap B_2 | B_1 \cap A_2)\}$

The most information-rich split consistent with

```
ABCDEF  | GHIJ
ABCDE IJ|FGH  
```

is `ABCDE | GH`, which contains 
```{r}
TreeTools::SplitInformation(5, 2)
```
bits of phylogenetic information. This value can be used as a similarity score
for this pairing of splits.

```{r, fig.width=6, fig.align='center'}
MatchingSplitInfoDistance(tree1, tree2)
VisualizeMatching(MatchingSplitInfoDistance, tree1, tree2,
                  Plot = TreeDistPlot, matchZeros = FALSE)
```


## References

---
title: "Extending the Robinson-Foulds metric"
author: "[Martin R. Smith](https://smithlabdurham.github.io/)"
output: rmarkdown::html_vignette
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa-old-doi-prefix.csl
vignette: >
  %\VignetteIndexEntry{Extending the Robinson-Foulds metric}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r init, echo=FALSE, warning=FALSE, message=FALSE}
library('ape')
library('TreeDist')
standardMargin <- c(0.4, 0.4, 0.8, 0.4)
cbPalette8 <- c("#000000", "#E69F00", "#56B4E9", "#009E73", "#F0E442",
                "#0072B2", "#D55E00", "#CC79A7")
TwoTreePlot <- function() par(mfrow = c(1, 2), mar = standardMargin, cex = 0.9)

Plot <- function(tree, tree2 = NULL, highlight = character(0),
                  prune = if (is.null(tree2)) integer(0) 
                          else list(integer(0), integer(0))
                  ) {
  ABCDEFGHIJK <- LETTERS[1:11]
  if (is.null(tree2)) {
    if(!all(tree$tip.label %in% 1:11)) {
      tree$tip.label <- match(tree$tip.label, ABCDEFGHIJK)
    }
    
    highlightTip <- match(highlight, ABCDEFGHIJK)
    TreeDistPlot(tree, bold = highlightTip, prune = prune,
                 graft = which(tree$edge[, 2] == 
                                 match(highlightTip, tree$tip.label)))
  } else {
    par(mfrow = c(1, 2), mar = rep(0.4, 4))
    Plot(tree, highlight = highlight, prune = prune[[1]])
    Plot(tree2, highlight = highlight, prune = prune[[2]])
  }
}
```

# The Robinson–Foulds distance

An intuitive way to calculate the similarity between two trees is to calculate
the resolution of their strict consensus, which corresponds to the number
of splits that occur in both trees [@Schuh1980].
The corresponding distance measure, the Robinson–Foulds distance
[@Robinson1981], counts the number of splits that are unique to one of the 
two trees.

It is important to remember that counting splits is not the same as counting
clades, edges or nodes.  If a tree is drawn as rooted (even without a root 
edge),
then the number of splits is one less than the number of edges or clades,
and two less than the number of nodes.
For example, the trees below have one split, two edges and three nodes in 
common.

```{R, echo=FALSE, fig.width = 6, fig.align = 'center'}
startPar <- TwoTreePlot()
palette <- colorspace::qualitative_hcl(5, c=42, l=88)
TreeDistPlot(balancedTree <- 
               read.tree(text="(((1, 2), (3, 4)), ((5, 6), (7, 8)));"))
edgelabels(1:5, c(1, 2, 5, 9, 12), adj = c(1.5, 0.5), bg = palette[1]) # Splits
edgelabels(1:6, edge = c(1, 2, 5, 8, 9, 12), bg = palette[2],
           adj = c(-0.5, 0.5)) # Internal edges
nodelabels(1:7, bg = palette[3], adj=c(1, 0.5)) # Nodes
nodelabels(1:6, node = 10:15, bg = palette[4], adj=c(-1, 0.5)) # Clades
legend('bottomleft', col=palette[1:2], pch=15, bty='n',
       c('Splits (5)', 'Internal edges (6)'))
legend('bottomright', col=palette[3:4], pch=15, bty='n',
       c('Nodes (7)', 'Clades (6)'))

TreeDistPlot(read.tree(text="((1, 2, 3, 4), (5, 6, 7, 8));"))
edgelabels(1, 1, adj = c(1.5, 0.5), bg = palette[1])
edgelabels(1:2, edge = c(1, 6), adj = c(-0.5, 0.5), bg = palette[2])
nodelabels(1:3, bg = palette[3], adj = c(1, 0.5))
nodelabels(1:2, node=10:11, bg = palette[4], adj = c(-1, 0.5))

```

The simplicity of counting splits is appealing, but limited by the underlying
assumption that all splits are equivalent.

As an example, a split that separates eight leaves into two sets of four
(as in the right-hand tree above) has a $\frac{1}{`r choose(8, 4) / 2`}$
chance of being compatible with the reference tree.
In contrast, a split that separates two leaves from the other six has a
$\frac{1}{7}$ chance of matching the reference tree: the similarity observed 
is five times more likely to have arisen by chance.
In other words, failure to match an even split is less noteworthy than failure
to match an uneven one.

As a consequence, trees whose splits are less even will, on average, exhibit
higher Robinson–Foulds distances with comparison trees.
Compare a balanced and an unbalanced eight-taxon tree:

```{r two-trees, echo=FALSE, fig.height = 3, fig.width=6, fig.align = 'center'}
TwoTreePlot()
TreeDistPlot(balancedTree, edge.color = cbPalette8[3])
ape::edgelabels(1:5, c(1, 2, 5, 9, 12), bg = palette[1], adj=c(0.5, -0.25))
ape::edgelabels(c('4:4', rep('2:6', 4)), c(1, 2, 5, 9, 12),
                adj=c(0.5, 1.25), frame = 'n')
ape::edgelabels(c(5.53, rep(3.46, 4)), c(1, 2, 5, 9, 12),
                adj=c(0.5, 2.75), frame = 'n')
legend('topleft', 'Balanced tree', bty='n', inset=c(-0.05, -0.03))

TreeDistPlot(caterpillarTree <- 
               ape::read.tree(text="(1, (2, (3, (4, (5, (6, (7, 8)))))));"),
             edge.color = cbPalette8[2])
ape::edgelabels(1:5, bg = palette[1], c(4, 6, 8, 10, 12), adj = c(0.5, -0.25))
ape::edgelabels(c('2|6', '3|5', '4|4', '3|5', '2|6'), edge = c(4, 6, 8, 10, 12),
                frame = 'n', adj = c(0.5, 1.25))
ape::edgelabels(c(3.46, 5.04, 5.53, 5.04, 3.46), edge = c(4, 6, 8, 10, 12),
                frame = 'n', adj = c(0.5, 2.75))
legend('topleft', 'Asymmetric tree', bty = 'n', inset = c(-0.05, -0.03))
```

Each tree divides the eight taxa into five splits.
The [phylogenetic information content](information.html) of a split is a 
function of the probability that the split will match a uniformly chosen random
tree, i.e. the proportion of eight-leaf binary trees that contain the split in
question.
(Information content, in bits, is defined as $-\log_2(\textrm{probability})$.)
This, in turn, is a function of the evenness of the split:

```{r ic-of-splits, display='asis', echo=FALSE, warnings=FALSE}
splitSmall <- 2:4
splitLarge <- 8L - splitSmall

rootedTrees <- c(
  '1' = 1,
  '2' = 1 * 1,
  '3' = 3 * 1,
  '4' = 5 * 3 * 1,
  '5' = 7 * 5 * 3 * 1,
  '6' = 9 * 7 * 5 * 3 * 1
)

matchingTrees <- rootedTrees[splitSmall] * rootedTrees[splitLarge]
names(matchingTrees) <- paste0("Split size: ", splitSmall, ':', splitLarge)

matchingP <- matchingTrees / (11 * 9 * 7 * 5 * 3 * 1 * 1)

ic <- -log2(matchingP)

knitr::kable(cbind(
      'Matching trees' = paste(matchingTrees, "/ 10 395"),
      '_P_(Match in random tree)' = signif(matchingP, 3),
      'Phylogenetic information content' = paste(signif(ic, 3), 'bits')))
```

In the first tree, split 1 is even, dividing four taxa from four others (`4|4`);
splits 2--5 are maximally uneven (`2|6`).  If each split is treated as 
independent, then the total information content of the five splits is
`r signif(sum(ic[c(4, 2, 2, 2, 2) - 1]), 4)`&nbsp;bits,
whereas that of the five splits in the second tree, of sizes 
`2|6`, `3|5`, `4|4`, `3|5` and `2|6`, is
`r signif(sum(ic[c(2, 3, 4, 3, 2) - 1]), 4)`&nbsp;bits.
Put another way, a random tree will
on average share more splits with the balanced tree (whose splits are 
predominantly uneven and thus likely to be matched) than the asymmetric tree 
(which contains more even splits that are less likely to occur in a random 
tree).

Indeed, of the 10&nbsp;395 eight-leaf trees, many more bear at least one split 
in common with a balanced tree than with an asymmetric tree:

```{r all-8-tip-trees, echo=FALSE, cache=TRUE, fig.width=4, fig.height=4, fig.align='center'}
calculate <- FALSE
if (calculate) {
  # Generate all eight-leaf trees
  all8 <- as.phylo(seq_len(NUnrooted(8)), tipLabels = 1:8)
  inBalanced <- RobinsonFoulds(all8, balancedTree, similarity = TRUE) / 2
  inCaterpillar <- RobinsonFoulds(all8, caterpillarTree, similarity = TRUE) / 2
} else {
  inBalanced <- rep(0:5, c(7088, 2708, 512, 76, 10, 1))
  inCaterpillar <- rep(0:5, c(8162, 1808, 350, 64, 10, 1))
}

par(cex = 0.8)
sch <- hist(inCaterpillar + 0.7, breaks = 0:18 / 3 - (1/6),
            main = "8-leaf trees with N common splits", cex.main = 1,
            xlim = c(0, 6), axes = FALSE,
            xlab = "Splits in common", ylab = "Number of trees")
sbh <- hist(inBalanced + 0.4, breaks = 0:18 / 3 - (1/6), plot = FALSE)
plot(sch, col = paste0(cbPalette8[2], "44"), add = TRUE)
plot(sbh, col = paste0(cbPalette8[3], "44"), add = TRUE)
text(1/6, 100, paste0("Balanced: ", sum(inBalanced == 0)), 
     pos = 4, srt = 90, cex = 0.7)
text(0.5, 100, paste0("Asymmetric: ", sum(inCaterpillar == 0)), 
     pos = 4, srt = 90, cex = 0.7)


legend("topright", pch = 22,
       pt.cex = 2, col = "black",
       pt.bg = paste0(cbPalette8[2:3], "44"), bty = "n",
       c("Asymmetric", "Balanced"))

axis(1, at = 0:5 + 0.5, labels = 0:5)
axis(2)
```


# Information-corrected Robinson–Foulds distance

This differing information content can be accommodated by weighting each
split according to the amount of phylogenetic information it contains
[@SmithDist].
The two tree pairs below both have a Robinson–Foulds distance of two,
but the first pair differ with regard to 
an uneven split (`ABCDEF|GH`), so obtain a total difference
of 22.54 &minus; (3.46 + 5.04 + 5.53 + 5.04) = 3.46&nbsp;bits:

```{r, fig.height = 3, fig.width=6, fig.align = 'center'}
tree1 <- ape::read.tree(text='(1, (2, (3, (4, (5, (6, (7, 8)))))));')
tree2 <- ape::read.tree(text='(1, (2, (3, (4, (5, (7, (6, 8)))))));')
tree3 <- ape::read.tree(text='(1, (2, (3, (5, (4, (6, (7, 8)))))));')

VisualizeMatching(InfoRobinsonFoulds, tree1, tree2, 
                  Plot = TreeDistPlot, prune = 12)
```

whereas the second pair differ in the resolution of a more even, and thus more
information-rich, split (`ABCD|EFGH`), and so receive a distance score of 
5.53&nbsp;bits:

```{r, fig.height = 3, fig.width=6, fig.align = 'center'}
VisualizeMatching(InfoRobinsonFoulds, tree1, tree3, 
                  Plot = TreeDistPlot, prune = 8)
```

# Generalized Robinson–Foulds distances

Even when accounting for the information content of splits in this way,
the Robinson–Foulds distance is readily saturated: the maximum value can be 
obtained by moving a single leaf.

```{r, fig.height = 3, fig.width=6, fig.align = 'center'}
tree1 <- ape::read.tree(text='(1, (2, (3, (4, (5, (6, (7, 8)))))));')
tree2 <- ape::read.tree(text='(8, (1, (2, (3, (4, (5, (6, 7)))))));')

VisualizeMatching(RobinsonFouldsMatching, tree1, tree2, Plot = TreeDistPlot)
```

Generalized Robinson–Foulds distances [@Nye2006;@Bocker2013] seek to address
this issue. 
This category of metrics aim to acknowledge semblances between similar-but-not-quite-identical pairs of splits, which would contribute 
zero to tree similarity under the standard Robinson–Foulds measure.

Generalized RF distances work by finding a _matching_ that pairs splits from
one tree with splits in the other.
Each pairing is scored according to the similarity of the paired splits;
the sum of these scores is the score of the matching.
The tree distance is given by the score of the optimal matching.

## Constructing a matching

Let's consider two trees that differ in the position of one wildcard leaf,
and in the resolution of a clade:

```{r, fig.height = 3, fig.width=6, fig.align = 'center'}
tree1 <- ape::read.tree(text='((A, B), ((C, (D, E)), (F, (G, (H, I)))));')
tree2 <- ape::read.tree(text='((A, B), ((C, D, (E, I)), (F, (G, H))));')

Plot(tree1, tree2, highlight = 'I', prune = list(8, integer(0)))
```

These trees obtain a Robinson–Foulds distance of nine: a large distances, as
the maximum possible for trees of this resolution is eleven.
`AB|CDEFGHI` is the only split in common between the two trees:

```{r, fig.height = 3, fig.width=6, fig.align = 'center'}
TwoTreePlot()
VisualizeMatching(RobinsonFouldsMatching, tree1, tree2)
```

This distance score is higher than might be expected, given how much the trees 
have in common;
removing the single leaf 'I' results in two trees that differ only 
in the resolution of a single node:

```{r, fig.height = 3, fig.width=6, fig.align = 'center'}
TwoTreePlot()
VisualizeMatching(RobinsonFouldsMatching,
                  drop.tip(tree1, 'I'),
                  drop.tip(tree2, 'I'))
```

This hidden similarity can be better reflected if similar, but non-identical,
splits are assigned non-zero similarity scores.

There are various ways to score the similarity between two splits. 
One is to build on the idea introduced above, where identical splits are
scored according to their phylogenetic information content.
Non-matching splits can be scored according to the amount of phylogenetic 
information that they hold in common, which is a function of the proportion of 
trees that are consistent with both splits.
(A full explanation is provided in the discussion of
[Generalized Robinson–Foulds distances](Generalized-RF.html).)

```{r, fig.height = 3, fig.width=6, fig.align = 'center'}
TwoTreePlot()
VisualizeMatching(SharedPhylogeneticInfo, tree1, tree2)
```

Here, the split `AB|CDEFGHI` occurs in both trees,
and, as it happens, makes the largest contribution to the tree similarity 
score (3.70) for this particular pair of trees.
This is the same contribution it would
have made to the information-corrected Robinson–Foulds similarity.

The split `ABCDEF|GHI` in the left-hand tree is paired with the split
`ABCDEFI|GH` in the right-hand tree.  Had `ABCDEF|GHI` been available in the 
right-hand tree, then this perfect match would have been assigned a similarity 
of `SplitInformation(3, 6)` = 5.57&nbsp;bits.  The partial match is instead 
allocated a lower score of 2.12&nbsp;bits.  Pairings of incompatible splits,
i.e. those that cannot co-exist on a tree, such as `ABCDEFG|HI` - `ABCDFGH|EI`,
have no phylogenetic information in common.
([clustering information](information.html) is an alternative way to think
about split similarity that recognizes similarity even between incompatible
splits.)

The matching depicted above is one of many.
It happens to be optimal: an optimal matching can be found by considering the
similarity score of each possible pairing, and solving a linear assignment
problem to find the optimal set of pairings.

We can view the splits in each tree, named according to the number of their
associated node:

```{r}
summary(TreeTools::as.Splits(tree1, LETTERS[1:9]))
summary(TreeTools::as.Splits(tree2, LETTERS[1:9]))
```

We can then see the similarity scores for each pair of splits, along with the 
optimal matching:

```{r}
attributes(SharedPhylogeneticInfo(tree1, tree2, reportMatching = TRUE))
```

`..` denotes that the fifth matching contributes zero to similarity score; an 
alternative optimal matching would leave these splits unpaired.

## What next?

- Alternatives measures of split similarity, such as 
[mutual clustering information](information.html), give rise to other
[Generalized Robinson–Foulds distances](Generalized-RF.html), and can be used
to generate meaningful [tree spaces](treespace.html).

```{r echo=FALSE}
par(startPar)
```

## References

---
title: "Comparing splits using information theory"
author: "[Martin R. Smith](https://smithlabdurham.github.io/)"
output: rmarkdown::html_vignette
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa-old-doi-prefix.csl
vignette: >
  %\VignetteIndexEntry{Comparing splits using information theory}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

To understand the [information-based metrics](Generalized-RF.html) implemented
in '[TreeDist](Using-TreeDist.html)', it is useful to recall some basic concepts
of information theory.

For an introduction, see @Mackay2003 or an introductory video to the clustering
information distance:

[![Introduction to the Clustering Info Distance](CID_talk.png)](https://durham.cloud.panopto.eu/Panopto/Pages/Viewer.aspx?id=ca5ede19-d21a-40ce-8b9e-ac6e00d7e2c0)


## Splits

Each internal edge in a tree represents a split that divides its leaves
into two partitions.  Intuitively, some splits are more instructive than
others.  For example, the fact that mammals and reptiles represent two separate
groups is profound enough that it is worth teaching to schoolchildren; much 
less information is represented by a split that identifies two species of bat as
more closely related to one another than to any other mammal or reptile.


## Quantifying information

How can we formalize the intuition that some splits contain more information
than others?  More generally, how can we quantify an amount of information?

Information is usually measured in _bits_.  One bit is the amount of information
generated by tossing a fair coin: to record the outcome of a coin toss, I must
record either a `H` or a `T`, and with each of the two symbols equally likely,
there is no way to compress the results of multiple tosses.

The Shannon [-@Shannon1948] information content of an outcome $x$ is defined 
to be $h(x) = -\log_2{P(x)}$, which simplifies to $\log_2{n}$ when all $n$ 
outcomes are equally likely.
Thus, the outcome of a fair coin toss delivers
$\log_2{2} = 1\textrm{ bit}$ of information;
the outcome of rolling a fair six-sided die contains
$\log_2{6} \approx 2.58\textrm{ bits}$ of information;
and the outcome of selecting at random one of the 105 unrooted binary six-leaf
trees is $\log_2{105} \approx 6.71\textrm{ bits}$.

Unlikely outcomes are more surprising, and thus contain more information than
likely outcomes.  The information content of rolling a twelve on two fair
six-sided dice is $-\log_2{\frac{1}{36}} \approx 5.16\textrm{ bits}$, whereas a seven,
which could be produced by six of the 36 possible rolls (`1 & 6`, `2 & 5`, ...),
is less surprising, and thus contains less information: 
$-\log_2{\frac{6}{36}} \approx 2.58\textrm{ bits}$.
An additional 2.58 bits of information would be required to establish which of
the six possible rolls produced the seven.


### Application to splits

The split $S_1 =$ `AB|CDEF` is found in 15 of the 105 six-leaf trees; 
as such, the probability that a randomly drawn tree contains $S_1$ is 
$P(S_1) = \frac{15}{105}$, and the information content $h(S_1) = 
-\log_2{\frac{15}{105}} \approx 2.81\textrm{ bits}$. @Steel2006 dub this 
quantity the **phylogenetic information content**.

Likewise, the split $S_2 =$ `ABC|DEF` occurs in nine of the 105 six-leaf trees,
so $h(S_2) = -\log_2{\frac{9}{105}} \approx 3.54\textrm{ bits}$.
Three six-leaf trees contain both splits, so in combination the splits deliver
$h(S_1,S_2) = -\log_2{\frac{3}{105}} \approx 5.13\textrm{ bits}$ of information.

Because $h(S_1,S_2) < h(S_1) + h(S_2)$, some of the information in $S_1$ is
also present in $S_2$.
The information in common between $S_1$ and $S_2$ is $h_{shared}(S_1, S_2) =
h(S_1) + h(S_2) - h(S_1,S_2) \approx 1.22\textrm{ bits}$.
The information unique to $S_1$ and $S_2$ is $h_{different}(S_1,S_2) = 
2h(S_1,S_2) - h(S_1) - h(S_2) \approx 3.91\textrm{ bits}$.

These quantities can be calculated using functions in the 
'[TreeTools](https://ms609.github.io/TreeTools/)' package.

```{r}
library("TreeTools", quietly = TRUE)
library("TreeDist")
treesMatchingSplit <- c(
  AB.CDEF = TreesMatchingSplit(2, 4),
  ABC.DEF = TreesMatchingSplit(3, 3)
)
treesMatchingSplit

proportionMatchingSplit <- treesMatchingSplit / NUnrooted(6)
proportionMatchingSplit

splitInformation <- -log2(proportionMatchingSplit)
splitInformation

treesMatchingBoth <- TreesConsistentWithTwoSplits(6, 2, 3)
combinedInformation <- -log2(treesMatchingBoth / NUnrooted(6))

sharedInformation <- sum(splitInformation) - combinedInformation
sharedInformation

# Or more concisely:
SplitSharedInformation(n = 6, 2, 3)
```


<!--The more similar the splits, the more information they will have in common;
the shared information is maximised when $S_1 = S_2$, and $h_{common} = h(S_1)$.
Likewise, more even splits contain more information than less even splits
(i.e. _h_(`AB|CDEF`) < _h_(`ABC|DEF`)).-->

## Entropy

Entropy is the average information content of each outcome, weighted by its
probability: $\sum{-p \log_2(p)}$.  Where all $n$ outcomes are equiprobable, 
this simplifies to $\log_2{n}$.

Consider a case in which Jane rolls a dice, and makes two true statements
about the outcome $x$:

$S_1$: "Is the roll even?".

- Two equally-possible outcomes: yes or no
- Entropy: $H(S_1) = \log_2{2} = 1\textrm{ bit}$.

$S_2$: "Is the roll greater than 3?"

- Two equally-possible outcomes: yes or no
- Entropy: $H(S_2) = \log_2{2} = 1\textrm{ bit}$.


The joint entropy of $S_1$ and $S_2$ is the entropy of the association 
matrix that considers each possible outcome:

 &nbsp;        | $S_1: x$ odd            | $S_1: x$ even
-------------- | ----------------------- | -------------------
$S_2: x \le 3$ | $x \in {1, 3}; p = \frac{2}{6}$ | $x = 2; p = \frac{1}{6}$ |
$S_2: x > 3$   | $x = 5; p = \frac{1}{6}$    | $x \in {4, 6}; p = \frac{2}{6}$ |

$\begin{aligned}
H(S_1, S_2) = \frac{2}{3}\log_2{\frac{2}{3}} + \frac{1}{3}\log_2{\frac{1}{3}} + \frac{1}{3}\log_2{\frac{1}{3}} + \frac{2}{3}\log_2{\frac{2}{3}} \approx
1.84 \textrm{ bits}
\end{aligned}$
 

Note that this less than the $\log_2{6} \approx 2.58\textrm{ bits}$ we require 
to determine the exact value of the roll: knowledge of $S_1$ and $S_2$ is not 
guaranteed to be sufficient to unambiguously identify $x$.


The mutual information between $S_1$ and $S_2$ describes how much knowledge of
$S_1$ reduces our uncertainty in $S_2$ (or _vice versa_).
So if we learn that $S_1$ is 'even', we become a little more confident
that $S_2$ is 'greater than three'. 

The mutual information $I(S_1;S_2)$, denoted in blue below, corresponds to the
sum of the individual entropies, minus the joint entropy:

\begin{aligned}
I(S_1;S_2) = H(S_1) + H(S_2) - H(S_1, S_2)
\end{aligned}

If two statements have high mutual information, then once you have heard one
statement, you already have a good idea what the outcome of the other statement
will be, and thus learn little new on hearing it.


The entropy distance, also termed the variation of information [@Meila2007],
corresponds to the information that $S_1$ and $S_2$ do _not_ have in common
(denoted below in yellow):

\begin{aligned}
H_D(S_1, S_2) = H(S_1, S_2) - I(S_1;S_2) = 2H(S_1, S_2) - H(S_1) - H(S_2)
\end{aligned}

The higher the entropy distance, the harder it is to predict the outcome of one
statement from the other; the maximum entropy distance occurs when the two
statements are entirely independent.

```{r mackay-8-1, echo=FALSE, fig.width=4, out.width='50%', fig.height=3, fig.align='center'}
library('TreeDist')
H <- function(inBracket) {
  expression(paste(italic('H'), plain('('), italic(inBracket), plain(')')))
}
oldPar <- par(mar = c(3.1, 0.1, 0, 0.1))
joint <- Entropy(c(1, 2, 1, 2) / 6)
plot(0, type = 'n', xlim = c(0, joint), ylim = c(5, 0), axes = FALSE)
axis(1, at = c(0, 0.5, 1, 1.5, round(joint, 2)))
mtext('Entropy / bits', 1, 2)
rect(joint - 1, 3.1, 1, 3.9, col = "#56B4E9")
rect(0, 0.1, joint - 1, 0.9, col = "#F0E442", border = NA)
rect(1, 1.1, joint, 1.9, col = "#F0E442", border = NA)
rect(joint - 1, 1.1, 1, 1.9, col = "#56B4E9", border = NA)
rect(joint - 1, 0.1, 1, 0.9, col = "#56B4E9", border = NA)
text(1, 3.5, pos=4,
     expression(paste(italic(I), plain('('), italic('A;B'), plain(')'))))



rect(0, 2.1, joint, 2.9)
text(joint / 2, 2.5, 
     expression(paste(italic('H'), plain('('), italic('A, B'), plain(')'))))

rect(0, 0.1, 1, 0.9)
text(0.5, 0.5, 
     expression(paste(italic('H'), plain('('), italic(A), plain(')'))))


rect(joint - 1, 1.1, joint - 0, 1.9)
text(joint - 0.5, 1.5, 
     expression(paste(italic('H'), plain('('), italic(B), plain(')'))))


rect(0, 4.1, joint - 1, 4.9, col = "#F0E442")
rect(1, 4.1, joint, 4.9,     col = "#F0E442")
text(joint / 2, 4.5, 
     expression(paste(italic('H'['D']), plain('('), italic('A, B'), plain(')'))))
par(oldPar)
```

### Application to splits

A split divides leaves into two partitions.  If we arbitrarily label these
partitions 'A' and 'B', and select a leaf at random, we can view the partition
label associated with the leaf. 
If 60/100 leaves belong to partition 'A', and 40/100 to 'B', then the a leaf
drawn at random has a 40% chance of bearing the label 'A'; the split has an
entropy of $-\frac{60}{100}\log_2{\frac{60}{100}}-\frac{40}{100}\log_2{\frac{40}{100}} \approx 0.97\textrm{ bits}$.

Now consider a different split, perhaps in a different tree, that assigns
50 leaves from 'A' to a partition 'C', leaving the remaining 10 leaves from 'A',
along with the 40 from 'B', in partition 'D'.
This split has $-\frac{50}{100}\log_2{\frac{50}{100}}-\frac{50}{100}\log_2{\frac{50}{100}} = 1\textrm{ bit}$ of entropy.  
Put these together, and a randomly selected leaf may now bear one of three
possible labellings:

- 'A' and 'C': 50 leaves
- 'A' and 'D': 10 leaves
- 'B' and 'D': 40 leaves.

The two splits thus have a joint entropy of 
$-\frac{50}{100}\log_2{\frac{50}{100}}
-\frac{10}{100}\log_2{\frac{10}{100}}
-\frac{40}{100}\log_2{\frac{40}{100}} \approx 1.36\textrm{ bits} < 0.97 + 1$.

The joint entropy is less than the sum of the individual entropies because the
two splits contain some mutual information: for instance, if a leaf bears the
label 'B', we can be certain that it will also bear the label 'D'.
The more similar the splits are, and the more they agree in their division of
leaves, the more mutual information they will exhibit.
I term this the **clustering information**, in contradistinction to the 
concept of phylogenetic information discussed above.

More formally, let split $S$ divides $n$ leaves into two partitions $A$ and $B$.
The probability that a randomly chosen leaf $x$ is in partition $k$ is 
$P(x \in k) = \frac{|k|}{n}$.  $S$ thus corresponds to a random variable 
with entropy $H(S) = -\frac{|A|}{n} \log_2{\frac{|A|}{n}} - 
\frac{|B|}{n}\log_2{\frac{|B|}{n}}$ [@Meila2007].
The joint entropy of two splits, $S_1$ and $S_2$, corresponds to the entropy of
the association matrix of probabilities that a randomly selected leaf
belongs to each pair of partitions:

 &nbsp;          | $S_1: x \in A_1$               | $S_1: x \in B_1$
---------------- | ------------------------------ | ----------------------
$S_2: x \in A_2$ | $P(A_1,A_2) =
\frac{|A_1 \cap A_2|}{n}$ | $P(B_1,A_2) = \frac{|B_1 \cap A_2|}{n}$ |
$S_2: x \in B_2$ | $P(A_1,B_2) =
\frac{|A_1 \cap B_2|}{n}$ | $P(B_1,B_2) = \frac{|B_1 \cap B_2|}{n}$ |


$H(S_1, S_2) =
P(A_1,A_2) \log_2 {P(A_1,A_2)} + 
P(B_1,A_2) \log_2 {P(B_1,A_2)}$

$+ P(A_1,B_2)\log_2{P(A_1,B_2)} + 
P(B_1,B_2)\log_2{P(B_1,B_2)}$

These values can then be substituted into the definitions of mutual information
and entropy distance given above.

As $S_1$ and $S_2$ become more different, the disposition of $S_1$ gives less
information about the configuration of $S_2$, and the mutual information 
decreases accordingly.

## References

---
title: "Analysing landscapes of phylogenetic trees"
author: "[Martin R. Smith](https://smithlabdurham.github.io/)"
output: 
  rmarkdown::html_vignette:
    fig_width: 6
    fig_height: 4
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa-old-doi-prefix.csl
vignette: >
  %\VignetteIndexEntry{Analysing landscapes of phylogenetic trees}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

Landscapes of trees are [mappings of tree space](treespace.html) that are
contoured according to some optimality criterion -- often, but not necessarily,
a tree's score under a phylogenetic reconstruction technique [@Bastert2002].
Detecting "islands" or "terraces" of trees can illuminate the nature of the
space of optimal trees and thus inform tree search strategy [@Maddison1991;
@Sanderson2011].

For simplicity (and to avoid scoring trees against a dataset), this example
uses a tree's balance (measured using the total cophenetic index) as its score
[@Mir2013].
We assume that mappings have already been shown to be [adequate](
treespace.html) [@SmithSpace].

A landscape is most simply visualized by colouring each tree according to its
score:

```{r col-trees-by-score}
# Load required libraries
library("TreeTools", quietly = TRUE)
library("TreeDist")

# Generate a set of trees
trees <- as.phylo(as.TreeNumber(BalancedTree(16)) + 0:100 - 15, 16)

# Create a 2D mapping
distances <- ClusteringInfoDist(trees)
mapping <- cmdscale(distances, 2)

# Score trees according to their balance
scores <- TotalCopheneticIndex(trees)

# Normalize scores
scoreMax <- TCIContext(trees[[1]])[["maximum"]]
scoreMin <- TCIContext(trees[[1]])[["minimum"]]
scores <- scores - scoreMin
scores <- scores / (scoreMax - scoreMin)

# Generate colour palette
col <- colorRamp(c("orange", "blue"))(scores)
rgbCol <- rgb(col, maxColorValue = 255)

# Plot trees, coloured by their score
plot(
  mapping,
  asp = 1, # Preserve aspect ratio - do not distort distances
  ann = FALSE, axes = FALSE, # Don't label axes: dimensions are meaningless
  col = rgbCol, # Colour trees by score
  pch = 16 # Plotting character: Filled circle
)

# Add a legend
PlotTools::SpectrumLegend(
  "left",
  title = "Tree balance",
  palette = rgb(colorRamp(c("orange", "blue"))(0:100 / 100) / 255),
  legend = floor(seq(scoreMax, scoreMin, length.out = 6)),
  bty = "n"
)
```

A more sophisticated output can be produced using contours, interpolating
between adjacent trees. This example uses a simple inverse distance weighting
function for interpolation; more sophisticated techniques such as
[kriging](https://desktop.arcgis.com/en/arcmap/10.3/tools/3d-analyst-toolbox/how-kriging-works.htm) or (in continuous tree spaces) the use of geodesics
[@Khodaei2022] can produce even better results.

```{r contoured-plot}
# Use an inverse distance weighting to interpolate between measured points
Predict <- function (x, y) {
  Distance <- function (a, b) {
    apply(a, 2, function (pt) sqrt(colSums((pt - b) ^ 2)))
  }
  predXY <- rbind(x, y)
  dists <- Distance(t(mapping), predXY)
  invDist <- 1 / dists
  weightings <- invDist / rowSums(invDist)

  # Return:
  colSums(scores * t(weightings))
}

# Generate grid for contour plot
resolution <- 32
xLim <- range(mapping[, 1]) * 1.1
yLim <- range(mapping[, 2]) * 1.11
x <- seq(xLim[1], xLim[2], length.out = resolution)
y <- seq(yLim[1], yLim[2], length.out = resolution)
z <- outer(x, y, Predict) # Predicted values for each grid square

# Plot
filled.contour(
  x, y, z,
  asp = 1, # Preserve aspect ratio - do not distort distances
  ann = FALSE, axes = FALSE, # Don't label axes: dimensions are meaningless
  plot.axes = {points(mapping, xpd = NA)} # Use filled.contour coordinates
)
```

A variety of R add-on packages facilitate three-dimensional plots.

```{r plot-3d, message = FALSE}
if (requireNamespace("plotly", quietly = TRUE)) {
  library("plotly", quietly = TRUE)
  fig <- plot_ly(x = x, y = y, z = z)
  fig <- fig %>% add_surface()
  fig
} else {
  print("Run `install.packages('plotly')` to view this output")
}
```

(Use the mouse to reorient)

# References

---
title: "Tree space analysis"
author: "[Martin R. Smith](https://smithlabdurham.github.io/)"
output: 
  rmarkdown::html_vignette:
    fig_width: 6
    fig_height: 4
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa-old-doi-prefix.csl
vignette: >
  %\VignetteIndexEntry{Tree space analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r secret-header, echo=FALSE}
set.seed(0)
protoclust <- if (requireNamespace("protoclust", quietly = TRUE)) {
  protoclust::protoclust
} else {
  hclust
}
```

It can be instructive to visualize the distribution of trees by mapping their
spatial relationships.
This can be a helpful means to address whether discrete islands exist in a
sample of trees, or whether analytical runs have converged.
Such analysis is relatively simple to conduct, but a few common oversights
can mislead interpretation.

## Tree space analysis via user interface

Tree space mapping and analysis is made simple with the Shiny app included in
the "TreeDist" R package.  Simply install [R](https://www.r-project.org/) or
[RStudio](https://posit.co/), then copy the code below into the R command
line:

```r
install.packages("TreeDist")
TreeDist::MapTrees()
```

This will allow you to conduct and evaluate basic tree space mappings
from tree lists saved in most common file formats; see an 
[outline of the basic functionality](https://ms609.github.io/TreeDist/reference/MapTrees.html).
To avoid misinterpreting tree space, it's worth having a broad idea of what
an analysis involves, and some potential pitfalls.

## Avoiding common pitfalls in tree space analysis

Here's an example analysis of a series of 200 trees from an ordered list.
The list corresponds to a mixed-base representation of trees (see 
`TreeTools::as.TreeNumber()`), so is expected to contain some structure as we 
jump from one "class" of tree to another. 
Let's see whether we can visualize and corroborate this structure.

First we'll generate the trees, and load some colours with which we might
identify them.

```{r generate-trees}
library("TreeTools", quietly = TRUE)
treeNumbers <- c(1:220)
trees <- as.phylo(treeNumbers, 8)
spectrum <- hcl.colors(220, "plasma")
treeCols <- spectrum[treeNumbers]
```

### Using a suitable distance metric

Now we need to calculate the distance between each pair of trees in our list.
The choice of distance metric is important [@SmithSpace].
The widely used Robinson–Foulds distance is, unfortunately, unsuitable for
tree space analysis.
The clustering information distance [@SmithDist] is a reliable alternative
that is fast to calculate:

```{r calculate-distances, message=FALSE}
library("TreeDist")
distances <- ClusteringInfoDistance(trees)
```

The reader is encouraged to repeat the exercise with other distances:

```r
distances <- RobinsonFoulds(trees)
distances <- PhylogeneticInfoDistance(trees)
distances <- as.dist(Quartet::QuartetDivergence(
  Quartet::ManyToManyQuartetAgreement(trees), similarity = FALSE))
```
### Mapping distances

Then we need to reduce the dimensionality of these distances.
We'll start out with a 12-dimensional mapping; if needed, we can always 
drop higher dimensions.  

Principal coordinates analysis is quick and performs very well:

```{r map}
mapping <- cmdscale(distances, k = 12)
```

Alternative mapping methods do exist, and sometimes give slightly better 
mappings.  `isoMDS()` performs non-metric multidimensional scaling (MDS)
with the Kruskal-1 stress function [@Kruskal1964]:
```r
kruskal <- MASS::isoMDS(distances, k = 12)
mapping <- kruskal$points
```

whereas `sammon()`, one of many metric MDS methods, uses Sammon's stress 
function [@Sammon1969]:
```
sammon <- MASS::sammon(distances, k = 12)
mapping <- sammon$points
```

That's a good start.  It is tempting to plot the first two dimensions
arising from this mapping and be done:

```{r plot-mapping-2d, fig.asp = 1, fig.width = 3, fig.align="center"}
par(mar = rep(0, 4))
plot(mapping,
     asp = 1, # Preserve aspect ratio - do not distort distances
     ann = FALSE, axes = FALSE, # Don't label axes: dimensions are meaningless
     col = treeCols, pch = 16
     )
```

### Identifying clusters

A quick visual inspection suggests at least two clusters, with the possibility
of further subdivision of the brighter trees.
But visual inspection can be 
highly misleading [@SmithSpace].  We must take a statistical approach.
A combination of partitioning around medoids and hierarchical clustering with
minimax linkage will typically find a clustering solution that is close to 
optimal, if one exists [@SmithSpace]; if suitably initialized, k-means++
clustering [@Arthur2007] can also be worthwhile.


```{r clustering, fig.align="center"}
possibleClusters <- 2:10

# Partitioning around medoids
pamClusters <- lapply(possibleClusters, function(k) cluster::pam(distances, k = k))
pamSils <- vapply(pamClusters, function(pamCluster) {
  mean(cluster::silhouette(pamCluster)[, 3])
}, double(1))

bestPam <- which.max(pamSils)
pamSil <- pamSils[bestPam]
pamCluster <- pamClusters[[bestPam]]$cluster

# Hierarchical clustering
hTree <- protoclust(distances)
hClusters <- lapply(possibleClusters, function(k) cutree(hTree, k = k))
hSils <- vapply(hClusters, function(hCluster) {
  mean(cluster::silhouette(hCluster, distances)[, 3])
}, double(1))

bestH <- which.max(hSils)
hSil <- hSils[bestH]
hCluster <- hClusters[[bestH]]

# k-means++ clustering
kClusters <- lapply(possibleClusters, function(k) KMeansPP(distances, k = k))
kSils <- vapply(kClusters, function(kCluster) {
  mean(cluster::silhouette(kCluster$cluster, distances)[, 3])
}, double(1))

bestK <- which.max(kSils)
kSil <- kSils[bestK]
kCluster <- kClusters[[bestK]]$cluster

plot(pamSils ~ possibleClusters,
     xlab = "Number of clusters", ylab = "Silhouette coefficient",
     ylim = range(c(pamSils, hSils)))
points(hSils ~ possibleClusters, pch = 2, col = 2)
points(kSils ~ possibleClusters, pch = 3, col = 3)
legend("topright", c("PAM", "Hierarchical", "k-means++"),
       pch = 1:3, col = 1:3)
```

Silhouette coefficients of < 0.25 suggest that structure is not meaningful; 
\> 0.5 denotes good evidence of clustering, and \> 0.7 strong evidence
[@Kaufman1990].
The evidence for the visually apparent clustering is not as strong as it first
appears.  Let's explore our two-cluster hierarchical clustering solution anyway.

```{r chosen-cluster}
nClusters <- 2
whichResult <- match(nClusters, possibleClusters)
cluster <- hClusters[[whichResult]]
```


We can visualize the clustering solution as a tree:

```{r h-tree, fig.align="center"}
class(hTree) <- "hclust"
par(mar = c(0, 0, 0, 0))
plot(hTree, labels = FALSE, main = "")
points(seq_along(trees), rep(1, length(trees)), pch = 16,
       col = spectrum[hTree$order])
```

Another thing we may wish to do is to take the consensus of each cluster:

```{r cluster-consensus, fig.align="center"}
par(mfrow = c(1, 2), mar = rep(0.2, 4))
col1 <- spectrum[mean(treeNumbers[cluster == 1])]
col2 <- spectrum[mean(treeNumbers[cluster == 2])]
plot(consensus(trees[cluster == 1], p = 0.5),
     edge.color = col1, edge.width = 2, tip.color = col1)
plot(consensus(trees[cluster == 2], p = 0.5),
     edge.color = col2, edge.width = 2, tip.color = col2)
```

Here, we learn that the two clusters are distinguished by the position of `t7`.


### Identifying islands

Besides clustering, we can also define 'islands' in tree space that are
separated by a 'moat', such that all trees on one island are separated from
all trees on another by at least a certain distance [@Silva2021].

```{r island-id, fig.asp = 1, fig.width = 3, fig.align="center"}
par(mar = rep(0, 4))
# set a threshold corresponding to the width of the "moat" between islands
threshold <- 1.8
island <- Islands(distances, threshold)

# See how many trees are on each island
table(island)

# Let's ignore the small islands for now
largeIsle <- Islands(distances, threshold, smallest = 5)

# Colour trees according to their island
plot(mapping,
     asp = 1, # Preserve aspect ratio - do not distort distances
     ann = FALSE, axes = FALSE, # Don't label axes: dimensions are meaningless
     col = ifelse(is.na(largeIsle), "grey", largeIsle + 1),
     pch = 16
     )
```

Let's view the consensus of each large island cluster:

```{r island-consensus, fig.align="center"}
par(mfrow = c(1, 2), mar = rep(0.2, 4))
plot(consensus(trees[!is.na(largeIsle) & largeIsle == 1], p = 0.5),
     edge.color = 2, edge.width = 2, tip.color = 2)
plot(consensus(trees[!is.na(largeIsle) & largeIsle == 2], p = 0.5),
     edge.color = 3, edge.width = 2, tip.color = 3)
```


### Validating a mapping

Now let's evaluate whether our map of tree space is representative.
First we want to know how many dimensions are necessary to adequately
represent the true distances between trees.  We hope for a trustworthiness ×
continuity score of \> 0.9 for a usable mapping, or \> 0.95 for a good one.

```{r how-many-dims, fig.align="center"}
txc <- vapply(seq_len(ncol(mapping)), function(k) {
  newDist <- dist(mapping[, seq_len(k)])
  MappingQuality(distances, newDist, 10)["TxC"]
}, 0)
plot(txc, xlab = "Dimension")
abline(h = 0.9, lty = 2)
```

We are going to need at least five dimensions to adequately represent the
distances between trees.

To help establish visually what structures are more likely to be genuine,
we might also choose to calculate a minimum spanning tree:

```{r calculate-MST}
mstEnds <- MSTEdges(distances)
```

Let's plot the first five dimensions of our tree space, highlighting the convex
hulls of our clusters:

```{r plot-mapping-5d, fig.asp = 1, fig.align="center"}
nDim <- which.max(txc > 0.9)
plotSeq <- matrix(0, nDim, nDim)
plotSeq[upper.tri(plotSeq)] <- seq_len(nDim * (nDim - 1) / 2)
plotSeq <- t(plotSeq[-nDim, -1])
plotSeq[nDim * 1:3] <- (nDim * (nDim - 1) / 2) + 1:3
layout(plotSeq)
par(mar = rep(0.1, 4))

for (i in 2:nDim) for (j in seq_len(i - 1)) {
  # Set up blank plot
  plot(mapping[, j], mapping[, i], ann = FALSE, axes = FALSE, frame.plot = TRUE,
       type = "n", asp = 1, xlim = range(mapping), ylim = range(mapping))
  
  # Plot MST
  MSTSegments(mapping[, c(j, i)], mstEnds,
              col = StrainCol(distances, mapping[, c(j, i)]))
  
  # Add points
  points(mapping[, j], mapping[, i], pch = 16, col = treeCols)

  # Mark clusters
  for (clI in unique(cluster)) {
    inCluster <- cluster == clI
    clusterX <- mapping[inCluster, j]
    clusterY <- mapping[inCluster, i]
    hull <- chull(clusterX, clusterY)
    polygon(clusterX[hull], clusterY[hull], lty = 1, lwd = 2,
            border = "#54de25bb")
    text(mean(clusterX), mean(clusterY), clI, col = "#54de25bb", font = 2)
  }
}
# Annotate dimensions
plot(0, 0, type = "n", ann = FALSE, axes = FALSE)
text(0, 0, "Dimension 2")
plot(0, 0, type = "n", ann = FALSE, axes = FALSE)
text(0, 0, "Dimension 3")
plot(0, 0, type = "n", ann = FALSE, axes = FALSE)
text(0, 0, "Dimension 4")
```

Our clusters, so distinct in dimension 1, overlap strongly in every other
dimension. The fact that the minimum spanning tree moves between clusters also
underlines the fact that they are not as well defined as they appear by eye.

Note that cluster membership, as well as the precise shape of tree space, 
is a function of the tree distance metric.  The phylogenetic information
distance recovers a different pair of clusters, which may not correspond to
those that are most apparent from a simple visual inspection of the
two-dimensional tree space plot:


```{r pid, fig.asp = 1, fig.width = 4, fig.align = "center", echo = FALSE, message = FALSE}
library("TreeDist")
pid_distances <- PhylogeneticInfoDistance(trees)
pid_mapping <- cmdscale(pid_distances, k = 6)
pid_cluster <- cutree(protoclust(pid_distances), k = 2)

par(mar = rep(0, 4))
plot(pid_mapping, ann = FALSE, axes = FALSE, asp = 1,
     col = treeCols, pch = 16)
MSTEdges(pid_distances, TRUE, pid_mapping[, 1], pid_mapping[, 2],
         col = "#bbbbbb", lty = 1)


pid_clusters <- seq_along(unique(pid_cluster))
for (clI in pid_clusters) {
  inCluster <- pid_cluster == clI
  clusterX <- pid_mapping[inCluster, 1]
  clusterY <- pid_mapping[inCluster, 2]
  hull <- chull(clusterX, clusterY)
  polygon(clusterX[hull], clusterY[hull], lty = 1, lwd = 2,
          border = "#54de25bb")
}
```

### Comparing cluster size

It is tempting to compare the size of clusters by calculating the area of 
convex hulls on a two-dimensional mapping.  However, mapped areas do [not
necessarily correspond](https://xkcd.com/2489/) to true hypervolumes.

Accuracy may be improved by comparing higher dimensions of projections using
the "[hypervolume](https://www.benjaminblonder.org/hypervolume_faq.html)"
package, though the same considerations apply [@Blonder2018].
Interpretation of overlap statistics is detailed in @Mammola2019.

```{r hypervolume, message = FALSE}
hypervolumeInstalled <- requireNamespace("hypervolume", quietly = TRUE)
if (hypervolumeInstalled) {
  library("hypervolume")
  hv1 <- hypervolume_gaussian(pid_mapping[pid_cluster == 1, 1:3],
                              verbose = FALSE)
  hv2 <- hypervolume_gaussian(pid_mapping[pid_cluster == 2, 1:3],
                              verbose = FALSE)
  hv_dist <- hypervolume_distance(hv1, hv2)
  capture.output(
    hyperset <- hypervolume_set(hv1, hv2, verbose = FALSE,
                              check.memory = FALSE)
  ) -> XX_VerboseNotRespected
  hv_overlap <- hypervolume_overlap_statistics(hyperset)
  hv_dist
  hv_overlap
} else {
  print("Install the 'hypervolume' package to run this example")
}

```

If the objective is to quantify the spread of different clusters, other metrics
may be easier to interpret than the clusters' hypervolume (e.g. @Smith2022).

The divergence of outlying points can be measures using the sum of ranges:

```{r sum-of-ranges}
SumOfRanges(pid_mapping, pid_cluster)
```

The overall size of a cluster can be measured using the sum of variances,
or the mean distance from the centroid or median:

```{r cluster-size}
SumOfVariances(pid_mapping, pid_cluster)
MeanCentroidDistance(pid_mapping, pid_cluster)
DistanceFromMedian(pid_mapping, pid_cluster)
```

The density of points within a cluster can be measured using the mean nearest-neighbour distance or the mean minimum spanning tree edge length:

```{r cluster-density}
MeanNN(pid_mapping, pid_cluster)
MeanMSTEdge(pid_mapping, pid_cluster)
```



## Self-organizing maps

An alternative approach to visualizing tree space is to create emergent
self-organizing maps [@Kohonen1982;@Ultsch2003;@Thrun2016],
which map high-dimensional data into two dimensions,
then add a third dimension to indicate distance between data points: 
nearby points occur in valleys, and are separated by ridges from more distant
data points.

```{r umatrix, fig.asp = 1, fig.align = "center"}
umatrixInstalled <- requireNamespace("Umatrix", quietly = TRUE)
if (umatrixInstalled) {
  map <- Umatrix::esomTrain(as.matrix(distances), Key = seq_along(trees),
                            Epochs = 5, # Increase for better results
                            Lines = 42,
                            Columns = 42,
                            Toroid = FALSE)
  
  Umatrix::plotMatrix(Matrix = map$Umatrix,
                      Toroid = FALSE, FixedRatio = TRUE,
                      TransparentContours = FALSE, Clean = TRUE) +
  ggplot2::geom_point(data = data.frame(x = map$BestMatches[, 3],
                                        y = map$BestMatches[, 2]),
                      shape = 19, color = treeCols, size = 2)

} else {
  message("Install the 'Umatrix' package to run this example")
}
```

## What next?

You may wish to:

- [Analyse landscapes](landscapes.html) of phylogenetic trees

- [Provide context](using-distances.html) for tree distances

- Review [available distance measures](https://ms609.github.io/TreeDist/index.html)
  and the corresponding [functions](https://ms609.github.io/TreeDist/reference/index.html#section-tree-distance-measures)

- [Interpret tree distance metrics](https://ms609.github.io/TreeDistData/articles/09-expected-similarity.html)

- Compare the distribution of different [sets of trees](compare-treesets.html)

# References

---
title: "Calculate tree similarity with 'TreeDist'"
author: "[Martin R. Smith](https://smithlabdurham.github.io/)"
output: rmarkdown::html_vignette
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa-old-doi-prefix.csl
vignette: >
  %\VignetteIndexEntry{Calculate tree similarity with 'TreeDist'}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

This document should contain all you need to get started measuring tree 
distances with 'TreeDist'.  If you get stuck, please 
[let me know](https://github.com/ms609/TreeDist/issues/new?title=Suggestion:+)
so I can improve this documentation.

## Loading trees

Instructions for loading phylogenetic trees into R can be found in a [separate vignette](https://ms609.github.io/TreeTools/articles/load-trees.html).
For these examples, we'll enter two simple trees by hand:

```{r set-up}
tree1 <- ape::read.tree(text = '(A, ((B, (C, (D, E))), ((F, G), (H, I))));')
tree2 <- ape::read.tree(text = '(A, ((B, (C, (D, (H, I)))), ((F, G), E)));')
```

## Calculating distances

We can calculate distances between pairs of trees using the 'TreeDist' package.

First we'll install the package.
We can either install the stable version from the CRAN repository:

```r
install.packages('TreeDist')
```

or the development version, from GitHub -- which will contain the latest
features but may not be as extensively tested:

```r 
devtools::install_github('ms609/TreeDist')
```

Then we'll load the package in to R's working environment:

```{r load-package, message=FALSE}
library('TreeDist')
```

Now the package's functions are available within R.
Let's proceed to calculate some tree distances.

### Pairs of trees

Calculating the distance between two trees is as simple as:

```{r measure-distance}
distance <- TreeDistance(tree1, tree2)
```

The convenience function `TreeDistance()` returns the variation of clustering
information between two trees, [normalized](using-distances.html#normalizing)
against the total information content of all splits.  

### Multiple comparisons

If you have more than two trees to compare, you can send a list of trees 
(class: `list` or `multiPhylo`) to the distance comparison function.
The function will then calculate the distance between each tree in the first
list and each tree in the second.

```{r multi-trees}
oneTree <- ape::rtree(11)
twoTrees <- structure(list(one = ape::rtree(11), two = ape::rtree(11)),
                      class = 'multiPhylo')
threeTrees <- list(a = ape::rtree(11), b = ape::rtree(11), c = ape::rtree(11))

TreeDistance(oneTree, twoTrees)
TreeDistance(twoTrees, threeTrees)
```

## Visualizing a matching

[Generalized Robinson–Foulds metrics](Generalized-RF.html),
such as the variation of clustering information,
rely on matching each split within a tree with another split in the other tree.  
We can view an optimal matching:

```{r visualise-matching, fig.align='center', fig.width=8, out.width='90%'}
#| fig.alt: >
#|   Pair of similar phylogenetic trees with matched splits highlighted
#|   according to the amount of clustering information in common.
VisualizeMatching(ClusteringInfoDistance, tree1, tree2)
```

This shows the six splits in tree 1, and the paired splits in tree two.  
Each split is labelled with a measure of its similarity, which is its 
contribution to the total tree similarity score.

We can view this information in a format accessible for further examination in R with:

```{r write-vis-matching, fig.align='center', fig.width=8, out.width='90%'}
ClusteringInfoDistance(tree1, tree2, reportMatching = TRUE)
```

Here, the `pairScores` attribute lists the score of each possible matching of 
splits.

We can identify the splits with:

```{r}
splits <- as.character(TreeTools::as.Splits(tree2))
splits
```

The names of the splits correspond to the number of an associated node in the
original tree:

```{r named-splits, fig.align='center', fig.width=7, fig.height = 5, out.width='80%'}
#| fig.alt: >
#|   Phylogenetic tree with nodes numbered, and labelled with the splits
#|   to which they correspond.
oldPar <- par(mar = rep(0, 4))
plot(tree2)
ape::nodelabels()
ape::nodelabels(splits, as.integer(names(splits)), 
                adj = c(1.1, -0.2), cex = 0.8, frame = 'none')
```
```{r, echo=FALSE}
par(oldPar)
```

Note that strictly, (informative) splits are associated with (internal) edges.
To avoid listing the same split twice, nodes close to the root (here, 10 and 11)
will not be associated with a split.

## What next?

You may wish to:

- [Provide context](using-distances.html) for tree distances

- Compare trees with [different tips](different-leaves.html)

- Review [available distance measures](https://ms609.github.io/TreeDist/index.html)
  and the corresponding [functions](https://ms609.github.io/TreeDist/reference/index.html#section-tree-distance-measures)

- [Interpret tree distance metrics](https://ms609.github.io/TreeDistData/articles/09-expected-similarity.html)

- Visualize [tree landscapes](treespace.html) using distance-based tree spaces

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_information.R
\name{MCITree}
\alias{MCITree}
\title{Maximum Clade Information Tree}
\usage{
MCITree(trees, info = "phylogenetic", check.tips = TRUE)
}
\arguments{
\item{trees}{List of \code{phylo} objects, optionally with class \code{multiPhylo}.}

\item{info}{Abbreviation of "phylogenetic" or "clustering", specifying
the concept of information to employ.}

\item{check.tips}{Logical specifying whether to renumber leaves such that
leaf numbering is consistent in all trees.}
}
\value{
\code{MCITree()} returns the tree with the highest information
content, selected from \code{trees}.
}
\description{
Analogous to the Maximum Clade Credibility tree:
select the tree from a posterior distribution whose clades have the
highest information content.
Generate the MCC tree by specifying \code{info = "credibility"}.
}
\examples{
library("TreeTools", quietly = TRUE)
trees <- as.phylo(24:40, 16)

# Maximum Clade Information tree
mci <- MCITree(trees)
SplitwiseInfo(mci)
plot(mci)
p <- SplitFrequency(mci, trees) / length(trees)
LabelSplits(mci, round(p * 100), "\%", bg = SupportColor(p))

\donttest{
# Compare with Maximum Clade Credibility tree
mcc <- MCITree(trees, "credibility")
plot(mcc)
p <- SplitFrequency(mcc, trees) / length(trees)
LabelSplits(mcc, round(p * 100), "\%", bg = SupportColor(p))
SplitwiseInfo(mcc)
}

}
\seealso{
Other summary trees:
\code{\link[=TransferConsensus]{TransferConsensus()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{summary trees}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Plot3.R
\name{Plot3}
\alias{Plot3}
\title{Pseudo-3D plotting}
\usage{
Plot3(
  x,
  y = NULL,
  z = NULL,
  pch = par("pch"),
  col = par("col"),
  bg = NA,
  cex = 1,
  axes = TRUE,
  frame.plot = axes,
  plot.bg = NA,
  fog = 1/2,
  shrink = 1/2,
  add = FALSE,
  ...
)
}
\arguments{
\item{x, y, z}{Coordinates of points to plot.}

\item{bg, cex, col, pch, add, axes, frame.plot, \dots}{Parameters passed to
\code{\link[=plot.default]{plot.default()}}.}

\item{plot.bg}{Colour with which to fill plot area, used as fog colour.}

\item{fog}{Numeric from zero (no fading) to one (furthest points are
invisible) specifying amount to fade distant points.}

\item{shrink}{Numeric specifying degree to which size of plotted point
should reflect \code{z} position. \code{0} denotes no scaling; if \code{1}, furthest
point will have zero size.}
}
\description{
\code{Plot3()} displays three-dimensional data in two dimensions, reflecting the
third dimension with point scaling, overlap and fogging.
Points with a lower \code{z} value are smaller than, fainter than, and overlapped
by points with a higher value.
}
\examples{
Plot3(1:10, 1:10, 1:10, cex = 7, pch = 16, axes = FALSE, asp = 1)
# Extreme values of fog and shrink will cause smallest z values to
# become invisible.
Plot3(1:10, 1:10, 1:10, cex = 7, pch = 16, axes = FALSE, asp = 1,
      fog = 1, shrink = 1)
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hierarchical_mutual_information.R
\name{HierarchicalMutualInfo}
\alias{HierarchicalMutualInfo}
\alias{HMI}
\alias{SelfHMI}
\alias{EHMI}
\alias{AHMI}
\title{Hierarchical Mutual Information}
\usage{
HierarchicalMutualInfo(tree1, tree2 = NULL, normalize = FALSE)

HMI(tree1, tree2 = NULL, normalize = FALSE)

SelfHMI(tree)

EHMI(tree1, tree2, precision = 0.01, minResample = 36)

AHMI(tree1, tree2, Mean = max, precision = 0.01, minResample = 36)
}
\arguments{
\item{normalize}{If \code{FALSE}, return the raw \acronym{HMI}, in bits.
If \code{TRUE}, normalize to range [0,1] by dividing by
\code{max(SelfHMI(tree1), SelfHMI(tree2))}.
If a function, divide by \code{normalize(SelfHMI(tree1), SelfHMI(tree2))}.}

\item{tree, tree1, tree2}{An object that can be coerced to an \code{\link{HPart}}
object.}

\item{precision}{Numeric; Monte Carlo sampling will terminate once the
relative standard error falls below this value.}

\item{minResample}{Integer specifying minimum number of Monte Carlo samples
to conduct.  Avoids early termination when sample size is too small to
reliably estimate the standard error of the mean.}

\item{Mean}{Function by which to combine the self-information of the
two input hierarchies, in order to normalize the \acronym{HMI}.}
}
\value{
\code{HierarchicalMutualInfo()} returns a numeric value representing the
hierarchical mutual information between the input trees, in bits,
normalized as specified.
Higher values indicate more shared hierarchical structure.

\code{SelfHMI()} returns the hierarchical mutual information of a tree
compared with itself, i.e. its hierarchical entropy (\acronym{HH}).

\code{EHMI()} returns the expected \acronym{HMI} against a uniform
shuffling of element labels, estimated by performing Monte Carlo resampling
on the same hierarchical structure until the relative standard error of the
estimate falls below \code{precision}.
The attributes of the returned object list the variance (\code{var}),
standard deviation (\code{sd}), standard error of the mean (\code{sem}) and
relative error (\code{relativeError}) of the estimate, and the number of Monte
Carlo samples used to obtain it (\code{samples}).

\code{AHMI()} returns the adjusted \acronym{HMI}, normalized such that
zero corresponds to the expected \acronym{HMI} given a random shuffling
of elements on the same hierarchical structure.  The attribute \code{sem} gives
the standard error of the estimate.
}
\description{
Calculate the Hierarchical Mutual Information (\acronym{HMI})
between two trees, following the recursive algorithm of
\insertCite{Perotti2020;textual}{TreeDist}.

This function was written during a code sprint: its documentation and test
cases have not yet been carefully scrutinized, and its implementation may
change without notice.
Please alert the maintainer to any issues you encounter.
}
\details{
\code{HierarchicalMutualInfo()} computes the hierarchical mutual content of trees
\insertCite{Perotti2015,Perotti2020}{TreeDist}, which accounts for the
non-independence of information represented by nested splits.

\code{tree} is converted to a set of hierarchical partitions, and the mutual
information (in bits) is computed recursively; the contribution of a node is
given by:

\deqn{I(t,s) = \log_2(n_{ts}) - \dfrac{H_{us} + H_{tv} - H_{uv}}{n_{ts}} +
\text{mean}(I_{uv})}

Where:
\itemize{
\item \eqn{n_{ts}} is the number of common elements between partitions
\item \eqn{H_{us}, H_{tv}, H_{uv}} are entropy terms from child comparisons
\item \eqn{I_{uv}} is the recursive \acronym{HMI} for child pairs
}

\code{AHMI()} calculates the adjusted hierarchical mutual information:
\deqn{\text{AHMI}(t, s) = \dfrac{I(t, s) - \hat{I}(t, s)}{
 \text{mean}(H(t), H(s)) - \hat{I}(t, s)}}
Where:
\itemize{
\item \eqn{I(t, s)} is the hierarchical mutual information between \code{tree1} and
\code{tree2}
\item \eqn{\hat{I}(t, s)} is the expected \acronym{HMI} between \code{tree1} and
\code{tree2}, estimated by Monte Carlo sampling
\item \eqn{H(t), H(s)} is the entropy (self-mutual information) of each tree
}
}
\examples{
library("TreeTools", quietly = TRUE)

tree1 <- BalancedTree(8)
tree2 <- PectinateTree(8)

# Calculate HMI between two trees
HierarchicalMutualInfo(tree1, tree2)

# HMI normalized against the mean information content of tree1 and tree2
HierarchicalMutualInfo(tree1, tree2, normalize = mean)

# Normalized HMI above is equivalent to:
HMI(tree1, tree2) / mean(SelfHMI(tree1), SelfHMI(tree2))
# Expected mutual info for this pair of hierarchies
EHMI(tree1, tree2, precision = 0.1)
# The adjusted HMI normalizes against this expectation
AHMI(tree1, tree2, precision = 0.1)
}
\references{
\insertAllCited{}
}
\seealso{
Other tree distances:
\code{\link[=JaccardRobinsonFoulds]{JaccardRobinsonFoulds()}},
\code{\link[=KendallColijn]{KendallColijn()}},
\code{\link[=MASTSize]{MASTSize()}},
\code{\link[=MatchingSplitDistance]{MatchingSplitDistance()}},
\code{\link[=NNIDist]{NNIDist()}},
\code{\link[=NyeSimilarity]{NyeSimilarity()}},
\code{\link[=PathDist]{PathDist()}},
\code{\link{Robinson-Foulds}},
\code{\link[=SPRDist]{SPRDist()}},
\code{\link[=TransferDist]{TransferDist()}},
\code{\link[=TreeDistance]{TreeDistance()}}
}
\concept{tree distances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_info.R
\name{.PairMean}
\alias{.PairMean}
\title{Mean of two numbers}
\usage{
.PairMean(x, y)
}
\description{
Used for normalization and range calculation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_info.R, R/tree_distance_msi.R
\encoding{UTF-8}
\name{TreeDistance}
\alias{TreeDistance}
\alias{SharedPhylogeneticInfo}
\alias{DifferentPhylogeneticInfo}
\alias{PhylogeneticInfoDistance}
\alias{ClusteringInfoDistance}
\alias{ClusteringInfoDist}
\alias{ExpectedVariation}
\alias{MutualClusteringInfo}
\alias{MutualClusteringInformation}
\alias{SharedPhylogeneticInfoSplits}
\alias{MutualClusteringInfoSplits}
\alias{MatchingSplitInfo}
\alias{MatchingSplitInfoDistance}
\alias{MatchingSplitInfoSplits}
\title{Information-based generalized Robinson–Foulds distances}
\usage{
TreeDistance(tree1, tree2 = NULL)

SharedPhylogeneticInfo(
  tree1,
  tree2 = NULL,
  normalize = FALSE,
  reportMatching = FALSE,
  diag = TRUE
)

DifferentPhylogeneticInfo(
  tree1,
  tree2 = NULL,
  normalize = FALSE,
  reportMatching = FALSE
)

PhylogeneticInfoDistance(
  tree1,
  tree2 = NULL,
  normalize = FALSE,
  reportMatching = FALSE
)

ClusteringInfoDistance(
  tree1,
  tree2 = NULL,
  normalize = FALSE,
  reportMatching = FALSE
)

ExpectedVariation(tree1, tree2, samples = 10000)

MutualClusteringInfo(
  tree1,
  tree2 = NULL,
  normalize = FALSE,
  reportMatching = FALSE,
  diag = TRUE
)

SharedPhylogeneticInfoSplits(
  splits1,
  splits2,
  nTip = attr(splits1, "nTip"),
  reportMatching = FALSE
)

MutualClusteringInfoSplits(
  splits1,
  splits2,
  nTip = attr(splits1, "nTip"),
  reportMatching = FALSE
)

MatchingSplitInfo(
  tree1,
  tree2 = NULL,
  normalize = FALSE,
  reportMatching = FALSE,
  diag = TRUE
)

MatchingSplitInfoDistance(
  tree1,
  tree2 = NULL,
  normalize = FALSE,
  reportMatching = FALSE
)

MatchingSplitInfoSplits(
  splits1,
  splits2,
  nTip = attr(splits1, "nTip"),
  reportMatching = FALSE
)
}
\arguments{
\item{tree1, tree2}{Trees of class \code{phylo}, with leaves labelled identically,
or lists of such trees to undergo pairwise comparison.  Where implemented,
\code{tree2 = NULL} will compute distances between each pair of trees in the list
\code{tree1} using a fast algorithm based on
\insertCite{Day1985;textual}{TreeDist}.}

\item{normalize}{If a numeric value is provided, this will be used as a
maximum value against which to rescale results.
If \code{TRUE}, results will be rescaled against a maximum value calculated from
the specified tree sizes and topology, as specified in the "Normalization"
section below.
If \code{FALSE}, results will not be rescaled.}

\item{reportMatching}{Logical specifying whether to return the clade
matchings as an attribute of the score.}

\item{diag}{Logical specifying whether to return similarities along the
diagonal, i.e. of each tree with itself.  Applies only if \code{tree2} is
a list identical to \code{tree1}, or \code{NULL}.}

\item{samples}{Integer specifying how many samplings to obtain;
accuracy of estimate increases with \code{sqrt(samples)}.}

\item{splits1, splits2}{Logical matrices where each row corresponds to a leaf,
either listed in the same order or bearing identical names (in any sequence),
and each column corresponds to a split, such that each leaf is identified as
a member of the ingroup (\code{TRUE}) or outgroup (\code{FALSE}) of the respective
split.}

\item{nTip}{(Optional) Integer specifying the number of leaves in each split.}
}
\value{
If \code{reportMatching = FALSE}, the functions return a numeric
vector specifying the requested similarities or differences.

If \code{reportMatching = TRUE}, the functions additionally return an integer
vector listing the index of the split in \code{tree2} that is matched with
each split in \code{tree1} in the optimal matching.
Unmatched splits are denoted \code{NA}.
Use \code{\link[=VisualizeMatching]{VisualizeMatching()}} to plot the optimal matching.

\code{TreeDistance()} simply returns the clustering information distance (it is
an alias of \code{ClusteringInfoDistance()}).
}
\description{
Calculate tree similarity and distance measures based on the amount of
phylogenetic or clustering information that two trees hold in common, as
proposed in Smith (2020).
}
\details{
\href{https://ms609.github.io/TreeDist/articles/Robinson-Foulds.html#generalized-robinson-foulds-distances}{Generalized Robinson–Foulds distances}
calculate tree similarity by finding an
optimal matching that the similarity between a split on one tree
and its pair on a second, considering all possible ways to pair splits
between trees (including leaving a split unpaired).

The methods implemented here use the concepts of
\href{https://ms609.github.io/TreeDist/articles/information.html}{entropy and information}
\insertCite{Mackay2003}{TreeDist} to assign a similarity score between each
pair of splits.

The returned tree similarity measures state the amount of information,
in bits, that the splits in two trees hold in common
when they are optimally matched, following
\insertCite{SmithDist;textual}{TreeDist}.
The complementary tree distance measures state how much information is
different in the splits of two trees, under an optimal matching.
Where trees contain different tips, tips present in one tree but not the
other are removed before each comparison (as by definition, the trees neither
hold information in common nor differ regarding these tips).
}
\section{Concepts of information}{
The phylogenetic (Shannon) information content and entropy of a split are
defined in
\href{https://ms609.github.io/TreeDist/articles/information.html}{a separate vignette}.

Using the mutual (clustering) information
\insertCite{Meila2007,Vinh2010}{TreeDist} of two splits to quantify their
similarity gives rise to the Mutual Clustering Information measure
(\code{MutualClusteringInfo()}, \code{MutualClusteringInfoSplits()});
the entropy distance gives the Clustering Information Distance
(\code{ClusteringInfoDistance()}).
This approach is optimal in many regards, and is implemented with
normalization in the convenience function \code{TreeDistance()}.

Using the amount of phylogenetic information common to two splits to measure
their similarity gives rise to the Shared Phylogenetic Information similarity
measure (\code{SharedPhylogeneticInfo()}, \code{SharedPhylogeneticInfoSplits()}).
The amount of information distinct to
each of a pair of splits provides the complementary Different Phylogenetic
Information distance metric (\code{DifferentPhylogeneticInfo()}).

The Matching Split Information measure (\code{MatchingSplitInfo()},
\code{MatchingSplitInfoSplits()}) defines the similarity between a pair of
splits as the phylogenetic information content of the most informative
split that is consistent with both input splits; \code{MatchingSplitInfoDistance()}
is the corresponding measure of tree difference.
(\href{https://ms609.github.io/TreeDist/articles/Generalized-RF.html}{More information here}.)
}

\section{Conversion to distances}{
To convert similarity measures to distances, it is necessary to
subtract the similarity score from a maximum value.  In order to generate
distance \emph{metrics}, these functions subtract the similarity twice from the
total information content (SPI, MSI) or entropy (MCI) of all the splits in
both trees \insertCite{SmithDist}{TreeDist}.
}

\section{Normalization}{
If \code{normalize = TRUE}, then results will be rescaled such that distance
ranges from zero to (in principle) one.
The maximum \strong{distance} is the sum of the information content or entropy of
each split in each tree; the maximum \strong{similarity} is half this value.
(See Vinh \emph{et al.} (2010, table 3) and
\insertCite{SmithDist;textual}{TreeDist} for
alternative normalization possibilities.)

Note that a distance value of one (= similarity of zero) will seldom be
achieved, as even the most different trees exhibit some similarity.
It may thus be helpful to rescale the normalized value such that the
\emph{expected} distance between a random pair of trees equals one.  This can
be calculated with \code{ExpectedVariation()}; or see package
'\href{https://ms609.github.io/TreeDistData/reference/randomTreeDistances.html}{TreeDistData}'
for a compilation of expected values under different metrics for trees with
up to 200 leaves.

Alternatively, use \verb{normalize = }\code{\link{pmax}} or \code{\link{pmin}} to scale against the
information content or entropy of all splits in the most (\code{pmax}) or
least (\code{pmin}) informative tree in each pair.
To calculate the relative similarity against a reference tree that is known
to be "correct", use \code{normalize = SplitwiseInfo(trueTree)} (SPI, MSI) or
\code{ClusteringEntropy(trueTree)} (MCI).
For worked examples, see the internal function \code{\link[=NormalizeInfo]{NormalizeInfo()}}, which is
called from distance functions with the parameter \code{how = normalize}.
.
}

\section{Distances between large trees}{
To balance memory demands and runtime with flexibility, these functions are
implemented for trees with up to 2048 leaves.
To analyse trees with up to 8192 leaves, you will need to a modified version
of the package:
\code{install.packages("BigTreeDist", repos = "https://ms609.github.io/packages/")}
Use \code{library("BigTreeDist")} \emph{instead} of \code{library("TreeDist")} to load
the modified package – or prefix functions with the package name, e.g.
\code{BigTreeDist::TreeDistance()}.

As an alternative download method,
uninstall \pkg{TreeDist} and \pkg{TreeTools} using
\code{remove.packages()}, then use
\code{devtools::install_github("ms609/TreeTools", ref = "more-leaves")}
to install the modified \pkg{TreeTools} package; then,
install \pkg{TreeDist} using
\code{devtools::install_github("ms609/TreeDist", ref = "more-leaves")}.
(\pkg{TreeDist} will need building from source \emph{after} the modified
\pkg{TreeTools} package has been installed, as its code links to values
set in the TreeTools source code.)

Trees with over 8192 leaves require further modification of the source code,
which the maintainer plans to attempt in the future; please \href{https://github.com/ms609/TreeTools/issues/141}{comment on GitHub} if you would find this useful.
}

\section{Parallelism}{
When \code{tree2 = NULL} and all trees share the same tip labels, pairwise
distance calculation uses a multi-threaded \acronym{OpenMP} batch path
automatically.  Control the number of threads with the \code{"mc.cores"} option:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(mc.cores = parallel::detectCores())  # use all cores
options(mc.cores = 4L)                        # or a fixed number
}\if{html}{\out{</div>}}

Do \strong{not} call \code{\link[=StartParallel]{StartParallel()}} for these functions: a registered
R cluster disables the \acronym{OpenMP} path and replaces it with slower
inter-process communication.  See \code{\link[=StartParallel]{StartParallel()}} for full guidance
on when an R cluster is appropriate.
}

\examples{
tree1 <- ape::read.tree(text="((((a, b), c), d), (e, (f, (g, h))));")
tree2 <- ape::read.tree(text="(((a, b), (c, d)), ((e, f), (g, h)));")
tree3 <- ape::read.tree(text="((((h, b), c), d), (e, (f, (g, a))));")

# Best possible score is obtained by matching a tree with itself
DifferentPhylogeneticInfo(tree1, tree1) # 0, by definition
SharedPhylogeneticInfo(tree1, tree1)
SplitwiseInfo(tree1) # Maximum shared phylogenetic information

# Best possible score is a function of tree shape; the splits within
# balanced trees are more independent and thus contain less information
SplitwiseInfo(tree2)

# How similar are two trees?
SharedPhylogeneticInfo(tree1, tree2) # Amount of phylogenetic information in common
attr(SharedPhylogeneticInfo(tree1, tree2, reportMatching = TRUE), "matching")
VisualizeMatching(SharedPhylogeneticInfo, tree1, tree2) # Which clades are matched?

DifferentPhylogeneticInfo(tree1, tree2) # Distance measure
DifferentPhylogeneticInfo(tree2, tree1) # The metric is symmetric

# Are they more similar than two trees of this shape would be by chance?
ExpectedVariation(tree1, tree2, sample=12)["DifferentPhylogeneticInfo", "Estimate"]

# Every split in tree1 conflicts with every split in tree3
# Pairs of conflicting splits contain clustering, but not phylogenetic, 
# information
SharedPhylogeneticInfo(tree1, tree3) # = 0
MutualClusteringInfo(tree1, tree3) # > 0

# Distance functions internally convert trees to Splits objects.
# Pre-conversion can reduce run time if the same trees will feature in
# multiple comparisons
splits1 <- TreeTools::as.Splits(tree1)
splits2 <- TreeTools::as.Splits(tree2)

SharedPhylogeneticInfoSplits(splits1, splits2)
MatchingSplitInfoSplits(splits1, splits2)
MutualClusteringInfoSplits(splits1, splits2)
}
\references{
\insertAllCited{}
}
\seealso{
Other tree distances:
\code{\link[=HierarchicalMutualInfo]{HierarchicalMutualInfo()}},
\code{\link[=JaccardRobinsonFoulds]{JaccardRobinsonFoulds()}},
\code{\link[=KendallColijn]{KendallColijn()}},
\code{\link[=MASTSize]{MASTSize()}},
\code{\link[=MatchingSplitDistance]{MatchingSplitDistance()}},
\code{\link[=NNIDist]{NNIDist()}},
\code{\link[=NyeSimilarity]{NyeSimilarity()}},
\code{\link[=PathDist]{PathDist()}},
\code{\link{Robinson-Foulds}},
\code{\link[=SPRDist]{SPRDist()}},
\code{\link[=TransferDist]{TransferDist()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree distances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_utilities.R
\name{.TreeDistance}
\alias{.TreeDistance}
\title{Calculate distance between trees, or lists of trees}
\usage{
.TreeDistance(Func, tree1, tree2, checks = TRUE, ...)
}
\arguments{
\item{checks}{Logical specifying whether to perform basic sanity checks to
avoid crashes in C++.}
}
\description{
Calculate distance between trees, or lists of trees
}
\seealso{
\code{\link[=CalculateTreeDistance]{CalculateTreeDistance()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hierarchical_mutual_information.R
\name{HierarchicalMutualInformation}
\alias{HierarchicalMutualInformation}
\title{Hierarchical mutual information}
\usage{
HierarchicalMutualInformation(tree1, tree2 = NULL, normalize = FALSE)
}
\description{
An alias of \code{HierarchicalMutualInfo()}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/3Dplot.R
\name{.ThreeDPlotServer}
\alias{.ThreeDPlotServer}
\title{Helper function for shiny 3D plotting}
\usage{
.ThreeDPlotServer(
  input,
  output,
  distances,
  clusterings,
  mapping,
  mstEnds,
  mstSize,
  pointCols,
  thinnedTrees,
  StrainCol,
  PlotSize,
  mode3D
)
}
\description{
Helper function for shiny 3D plotting
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MSTSegments.R
\name{MSTSegments}
\alias{MSTSegments}
\alias{StrainCol}
\title{Add minimum spanning tree to plot, colouring by stress}
\usage{
MSTSegments(mapping, mstEnds, ...)

StrainCol(
  distances,
  mapping,
  mstEnds = MSTEdges(distances),
  palette = rev(hcl.colors(256L, "RdYlBu"))
)
}
\arguments{
\item{mapping}{Two-column matrix giving \emph{x} and \emph{y} coordinates of plotted
points.}

\item{mstEnds}{Two-column matrix identifying rows of \code{mapping} at end of
each edge of the MST, as output by \code{\link[TreeTools:MSTEdges]{TreeTools::MSTEdges()}}.}

\item{\dots}{Additional arguments to \code{\link[=segments]{segments()}}.}

\item{distances}{Matrix or \code{dist} object giving original distances between
each pair of points.}

\item{palette}{Vector of colours with which to colour edges.}
}
\value{
\code{StrainCol()} returns a vector in which each entry is selected from
\code{palette}, with an attribute \code{logStrain} denoting the logarithm of the
mapped over original distance, shifted such that the median value is zero.
Palette colours are assigned centred on the median value, with entries
early in \code{palette} assigned to edges in which the ratio of mapped
distance to original distance is small.
}
\description{
To identify strain in a multidimensional scaling of distances, it can be
useful to plot a minimum spanning tree
\insertCite{Gower1966,SmithSpace}{TreeDist}.  Colouring each edge of the
tree according to its strain can identify areas where the mapping is
stretched or compressed.
}
\examples{
set.seed(0)
library("TreeTools", quietly = TRUE)
distances <- ClusteringInfoDist(as.phylo(5:16, 8))
mapping <- cmdscale(distances, k = 2)
mstEnds <- MSTEdges(distances)

# Set up blank plot
plot(mapping, asp = 1, frame.plot = FALSE, ann = FALSE, axes = FALSE,
     type = "n")
# Add MST
MSTSegments(mapping, mstEnds, 
            col = StrainCol(distances, mapping, mstEnds))
# Add points at end so they overprint the MST
points(mapping)
PlotTools::SpectrumLegend(
 "bottomleft",
 legend = c("Extended", "Median", "Contracted"),
 bty = "n",     # No box
 y.intersp = 2, # Expand in Y direction
 palette = hcl.colors(256L, "RdYlBu", rev = TRUE)
)
}
\references{
\insertAllCited{}
}
\seealso{
Other tree space functions:
\code{\link[=Islands]{Islands()}},
\code{\link[=MapTrees]{MapTrees()}},
\code{\link[=MappingQuality]{MappingQuality()}},
\code{\link[=SpectralEigens]{SpectralEigens()}},
\code{\link{cluster-statistics}},
\code{\link[=median.multiPhylo]{median.multiPhylo()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree space functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{cpp_transfer_dist_cross_pairs}
\alias{cpp_transfer_dist_cross_pairs}
\title{Cross-pairs transfer dissimilarity (OpenMP)}
\usage{
cpp_transfer_dist_cross_pairs(splits_a, splits_b, n_tip, scale, n_threads)
}
\arguments{
\item{splits_a, splits_b}{Lists of raw matrices.}

\item{n_tip}{Number of tips.}

\item{scale}{Logical: use scaled transfer dissimilarity?}

\item{n_threads}{Number of OpenMP threads.}
}
\value{
Numeric matrix of dimension \code{nA} x \code{nB}.
}
\description{
Cross-pairs transfer dissimilarity (OpenMP)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_path.R
\name{PathDist}
\alias{PathDist}
\title{Path distance}
\usage{
PathDist(tree1, tree2 = NULL)
}
\arguments{
\item{tree1, tree2}{Trees of class \code{phylo}, with leaves labelled identically,
or lists of such trees to undergo pairwise comparison.  Where implemented,
\code{tree2 = NULL} will compute distances between each pair of trees in the list
\code{tree1} using a fast algorithm based on
\insertCite{Day1985;textual}{TreeDist}.}
}
\value{
\code{PathDist()} returns a vector or distance matrix of distances
between trees.
}
\description{
Calculate the path distance between rooted or unrooted trees.
}
\details{
This function is a faster alternative to the function
\code{\link[phangorn:treedist]{path.dist()}} in the phangorn package,
which can crash if the internal representation of trees does not conform to
certain (unspecified) expectations, and which treats all trees as unrooted.

The path distance is calculated by tabulating the cladistic difference (=
topological distance) between each pair of tips in each tree.
A precursor to the path distance \insertCite{Farris1969}{TreeDist}
took the mean squared
difference between the elements of each tree's tabulation (Farris, 1973);
the method used here is that proposed by
\insertCite{Steel1993;textual}{TreeDist}, which takes the square root of this
sum.
Other precursor measures are described in
\insertCite{Williams1971;textual}{TreeDist} and
\insertCite{Phipps1971;textual}{TreeDist}.

If a root node is present, trees are treated as rooted.
To avoid counting the root edge twice, use \code{UnrootTree(tree)} before
calculating the path distance.

Use of the path distance is discouraged as it emphasizes
shallow relationships at the expense of deeper (and arguably more
fundamental) relationships \insertCite{Farris1973}{TreeDist}.
}
\examples{
library("TreeTools")

# Treating the two edges to the root node as distinct
PathDist(BalancedTree(7), PectinateTree(7))

# Counting those two edges once
PathDist(UnrootTree(BalancedTree(7)), UnrootTree(PectinateTree(7)))

PathDist(BalancedTree(7), as.phylo(0:2, 7))
PathDist(as.phylo(0:2, 7), PectinateTree(7))

PathDist(list(bal = BalancedTree(7), pec = PectinateTree(7)),
        as.phylo(0:2, 7))

PathDist(as.phylo(30:33, 8))
 
}
\references{
\insertAllCited{}
}
\seealso{
Other tree distances:
\code{\link[=HierarchicalMutualInfo]{HierarchicalMutualInfo()}},
\code{\link[=JaccardRobinsonFoulds]{JaccardRobinsonFoulds()}},
\code{\link[=KendallColijn]{KendallColijn()}},
\code{\link[=MASTSize]{MASTSize()}},
\code{\link[=MatchingSplitDistance]{MatchingSplitDistance()}},
\code{\link[=NNIDist]{NNIDist()}},
\code{\link[=NyeSimilarity]{NyeSimilarity()}},
\code{\link{Robinson-Foulds}},
\code{\link[=SPRDist]{SPRDist()}},
\code{\link[=TransferDist]{TransferDist()}},
\code{\link[=TreeDistance]{TreeDistance()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree distances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hierarchical_mutual_information.R
\name{HH}
\alias{HH}
\title{Self hierarchical mutual information}
\usage{
HH(tree)
}
\description{
An alias of \code{SelfHMI()}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{cpp_transfer_consensus}
\alias{cpp_transfer_consensus}
\title{Transfer consensus (C++ implementation)}
\usage{
cpp_transfer_consensus(
  splits_list,
  n_tip,
  scale,
  greedy_best_flag,
  init_majority,
  n_threads = 1L
)
}
\arguments{
\item{splits_list}{List of raw matrices (one per tree), each from as.Splits().}

\item{n_tip}{Number of tips.}

\item{scale}{Logical: use scaled transfer distance?}

\item{greedy_best_flag}{Logical: TRUE for "best", FALSE for "first".}

\item{init_majority}{Logical: TRUE to start from majority-rule splits.}
}
\value{
A \code{LogicalVector} of length n_splits indicating which pooled splits
are included in the consensus, plus attributes "raw_splits" (a raw matrix
of all unique splits) and "light_side" (integer vector).
}
\description{
Transfer consensus (C++ implementation)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/VisualizeMatching.R
\encoding{UTF-8}
\name{VisualizeMatching}
\alias{VisualizeMatching}
\alias{VisualiseMatching}
\alias{PlotMatching}
\alias{DisplayMatching}
\title{Visualize a matching}
\usage{
VisualizeMatching(
  Func,
  tree1,
  tree2,
  setPar = TRUE,
  precision = 3L,
  Plot = plot.phylo,
  matchZeros = TRUE,
  plainEdges = FALSE,
  edge.cex = par("cex"),
  value.cex = edge.cex * 0.8,
  edge.frame = "rect",
  edge.width = 1,
  edge.color = "black",
  ...
)
}
\arguments{
\item{Func}{Function used to construct tree similarity.}

\item{tree1, tree2}{Trees of class \code{phylo}, with identical leaf labels.}

\item{setPar}{Logical specifying whether graphical parameters should be
set to display trees side by side.}

\item{precision}{Integer specifying number of significant figures to display
when reporting matching scores.}

\item{Plot}{Function to use to plot trees.}

\item{matchZeros}{Logical specifying whether to pair splits with zero
similarity (\code{TRUE}), or leave them unpaired (\code{FALSE}).}

\item{plainEdges}{Logical specifying whether to plot edges with a uniform
width and colour (\code{TRUE}), or whether to draw edge widths according to the
similarity of the associated splits (\code{FALSE}).}

\item{edge.cex}{Character expansion for edge labels.
If \code{FALSE}, suppress edge labels.}

\item{value.cex}{Character expansion for values on edge labels.
If \code{FALSE}, values are not displayed.}

\item{edge.frame}{Character specifying the kind of frame to be printed around
the text of the edge labels.  Choose an abbreviation of \code{"rect"}, \code{"circle"},
or \code{"none"}.}

\item{edge.width, edge.color, \dots}{Additional parameters to send to \code{Plot()}.}
}
\value{
\code{VisualizeMatching()} invisibly returns the matching of splits
between \code{tree1} and \code{tree2} (i.e.
\code{Func(tree1, tree2, reportMatching = TRUE)})
}
\description{
Depict the splits that are matched between two trees using a specified
\href{https://ms609.github.io/TreeDist/articles/Generalized-RF.html}{Generalized Robinson–Foulds}
similarity measure.
}
\details{
Note that when visualizing a Robinson–Foulds distance (using
\code{Func = RobinsonFouldsMatching}), matched splits are assigned a \emph{similarity}
score of 1, which is deducted from the total number of splits to calculate
the Robinson–Foulds \emph{distance}.  Unmatched splits thus contribute one to
total tree distance.
}
\examples{
tree1 <- TreeTools::BalancedTree(6)
tree2 <- TreeTools::PectinateTree(6)

VisualizeMatching(RobinsonFouldsMatching, tree1, tree2)
matching <- VisualizeMatching(SharedPhylogeneticInfo, tree1, tree2,
                              matchZeros = FALSE)
attributes(matching)
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_utilities.R
\name{NormalizeInfo}
\alias{NormalizeInfo}
\title{Normalize tree distances}
\usage{
NormalizeInfo(
  unnormalized,
  tree1,
  tree2,
  InfoInTree,
  infoInBoth = NULL,
  how = TRUE,
  Combine = "+",
  ...
)
}
\arguments{
\item{unnormalized}{Numeric value, vector or matrix to be normalized.}

\item{tree1, tree2}{Trees from which \code{unnormalized} was calculated.}

\item{InfoInTree}{Function to calculate the information content of each tree.}

\item{infoInBoth}{Optional numeric specifying information content of both
trees independently. If unspecified (\code{NULL}), this will be calculated using
the method specified by \code{how}.}

\item{how}{Method for normalization, perhaps specified using the \code{normalize}
argument to a tree distance function.  See details for options.}

\item{\dots}{Additional parameters to \code{InfoInTree()} or \code{how()}.}
}
\value{
\code{NormalizeInfo()} returns an object corresponding to the normalized
values of \code{unnormalized}.
}
\description{
\code{NormalizeInfo()} is an internal function used to normalize information
against a reference, such as the total information present in a pair of
trees.
}
\details{
The unnormalized value(s) are normalized by dividing by a denominator
calculated based on the \code{how} parameter.  Valid options include:

\describe{
\item{\code{FALSE}}{No normalization is performed; the unnormalized values
are returned.}
\item{\code{TRUE}}{Unless \code{infoInBoth} is specified, the information in
each tree is computed using \code{InfoInTree()}, and the two values combined
using \code{Combine()}.}
\item{A numeric value, vector or matrix}{\code{how} is used as the denominator;
the returned value is \code{unnormalized / how}.}
\item{A function}{Unless \code{infoInBoth} is specified, the information in
each tree is computed using \code{InfoInTree()}, and the two values combined
using \code{how}.  \code{NormalizeInfo(how = Func)} is thus equivalent to
\code{NormalizeInfo(how = TRUE, Combine = Func)}.}
}
}
\examples{
library("TreeTools", quietly = TRUE)
pair1 <- c(BalancedTree(9), StarTree(9))
pair2 <- c(BalancedTree(9), PectinateTree(9))

# We'll let the number of nodes define the total information in a tree
Nnode(pair1)
Nnode(pair2)

# Let's normalize a unit distance
rawDist <- cbind(c(1, 1), c(1, 1))

# With `Combine = "+"`, the maximum distance is the sum of
# the information in each tree
denominator <- outer(Nnode(pair1), Nnode(pair2), "+")

NormalizeInfo(rawDist, pair1, pair2, InfoInTree = ape::Nnode, Combine = "+")
rawDist / denominator


# A denominator can be specified manually using `how`:
NormalizeInfo(rawDist, pair1, pair2, InfoInTree = ape::Nnode, how = 16)
rawDist / 16


# `how` also allows the denominator to be computed from trees:
outer(Nnode(pair1), Nnode(pair2), pmin)
NormalizeInfo(rawDist, pair1, pair2, InfoInTree = ape::Nnode, how = pmin)
rawDist / outer(Nnode(pair1), Nnode(pair2), pmin)

}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_utilities.R
\name{CalculateTreeDistance}
\alias{CalculateTreeDistance}
\title{Wrapper for tree distance calculations}
\usage{
CalculateTreeDistance(Func, tree1, tree2 = NULL, reportMatching = FALSE, ...)
}
\arguments{
\item{Func}{Tree distance function.}

\item{tree1, tree2}{Trees of class \code{phylo}, with leaves labelled identically,
or lists of such trees to undergo pairwise comparison.  Where implemented,
\code{tree2 = NULL} will compute distances between each pair of trees in the list
\code{tree1} using a fast algorithm based on
\insertCite{Day1985;textual}{TreeDist}.}

\item{reportMatching}{Logical specifying whether to return the clade
matchings as an attribute of the score.}

\item{\dots}{Additional arguments to \code{Func}.}
}
\description{
Calls tree distance functions from trees or lists of trees
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_spr.R
\name{SPRDist}
\alias{SPRDist}
\alias{SPRDist.phylo}
\alias{SPRDist.list}
\alias{SPRDist.multiPhylo}
\title{Approximate the Subtree Prune and Regraft distance}
\usage{
SPRDist(tree1, tree2 = NULL, method = "deOliveira", symmetric)

\method{SPRDist}{phylo}(tree1, tree2 = NULL, method = "deOliveira", symmetric)

\method{SPRDist}{list}(tree1, tree2 = NULL, method = "deOliveira", symmetric)

\method{SPRDist}{multiPhylo}(tree1, tree2 = NULL, method = "deOliveira", symmetric)
}
\arguments{
\item{tree1, tree2}{Trees of class \code{phylo}, with leaves labelled identically,
or lists of such trees to undergo pairwise comparison.  Where implemented,
\code{tree2 = NULL} will compute distances between each pair of trees in the list
\code{tree1} using a fast algorithm based on
\insertCite{Day1985;textual}{TreeDist}.}

\item{method}{Character specifying which method to use to approximate the
SPR distance.  Currently defaults to \verb{"deOliveira"``. }"Rogue"` implements an experimental method whose details are pending
publication; this function is under development, and may be modified or
removed without notice.  Once formally validated, it is anticipated that this
method will become the default.}

\item{symmetric}{Deprecated (redundant after fix of
\href{https://github.com/KlausVigo/phangorn/issues/97}{phangorn#97}).}
}
\value{
\code{SPRDist()} returns a vector or distance matrix of distances
between trees.
}
\description{
\code{SPRDist()} approximates the \acronym{SPR} distance between
trees.
}
\details{
The function currently defaults to the heuristic method of
\insertCite{deOliveira2008;textual}{TreeDist}, which purports to provide an
upper bound on the \acronym{SPR} distance (though exceptions exist).
Other approximations
\insertCite{@e.g. @Hickey2008, @Goloboff2008SPR, @Whidden2018}{TreeDist} are
not yet implemented.
}
\examples{
library("TreeTools", quietly = TRUE)

# Compare single pair of trees
SPRDist(BalancedTree(7), PectinateTree(7))

# Compare all pairs of trees
SPRDist(as.phylo(30:33, 8))

# Compare each tree in one list with each tree in another
SPRDist(BalancedTree(7), as.phylo(0:2, 7))
SPRDist(as.phylo(0:2, 7), PectinateTree(7))

SPRDist(list(bal = BalancedTree(7), pec = PectinateTree(7)),
        as.phylo(0:2, 7))
}
\references{
\insertAllCited{}
}
\seealso{
Exact calculation with \href{https://ms609.github.io/TBRDist/reference/TreeRearrangementDistances.html}{\pkg{TBRDist}}
functions \code{USPRDist()} and \code{ReplugDist()}.

\pkg{phangorn} function \code{\link[phangorn:treedist]{SPR.dist()}} employs
the \insertCite{deOliveira2008;textual}{TreeDist} algorithm but can crash
when sent trees of certain formats, and tends to have a longer running time.

Other tree distances:
\code{\link[=HierarchicalMutualInfo]{HierarchicalMutualInfo()}},
\code{\link[=JaccardRobinsonFoulds]{JaccardRobinsonFoulds()}},
\code{\link[=KendallColijn]{KendallColijn()}},
\code{\link[=MASTSize]{MASTSize()}},
\code{\link[=MatchingSplitDistance]{MatchingSplitDistance()}},
\code{\link[=NNIDist]{NNIDist()}},
\code{\link[=NyeSimilarity]{NyeSimilarity()}},
\code{\link[=PathDist]{PathDist()}},
\code{\link{Robinson-Foulds}},
\code{\link[=TransferDist]{TransferDist()}},
\code{\link[=TreeDistance]{TreeDistance()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree distances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parallel.R
\name{StartParallel}
\alias{StartParallel}
\alias{SetParallel}
\alias{GetParallel}
\alias{StopParallel}
\title{Calculate distances in parallel}
\usage{
StartParallel(...)

SetParallel(cl)

GetParallel(cl)

StopParallel(quietly = FALSE)
}
\arguments{
\item{\dots}{Parameters to pass to \code{\link[=makeCluster]{makeCluster()}}.}

\item{cl}{An existing cluster.}

\item{quietly}{Logical; if \code{TRUE}, do not warn when no cluster was running.}
}
\value{
\code{StartParallel()} and \code{SetParallel()} return the previous value of
\code{options("TreeDist-cluster")}.

\code{GetParallel()} returns the currently specified cluster.

\code{StopParallel()} returns \code{TRUE} if a cluster was destroyed,
\code{FALSE} otherwise.
}
\description{
Accelerate distance calculation by employing multiple \acronym{CPU} workers.
}
\details{
\subsection{OpenMP (recommended for all split-based metrics)}{

When the package is built with \acronym{OpenMP} support (the default on
Linux and Windows; optional on macOS), all pairwise split-based distance
calculations use an efficient multi-threaded batch path automatically —
no cluster setup is required.  The affected functions are:
\itemize{
\item \code{\link[=ClusteringInfoDistance]{ClusteringInfoDistance()}} / \code{\link[=MutualClusteringInfo]{MutualClusteringInfo()}}
\item \code{\link[=SharedPhylogeneticInfo]{SharedPhylogeneticInfo()}} / \code{\link[=DifferentPhylogeneticInfo]{DifferentPhylogeneticInfo()}}
\item \code{\link[=MatchingSplitInfo]{MatchingSplitInfo()}} / \code{\link[=MatchingSplitInfoDistance]{MatchingSplitInfoDistance()}}
\item \code{\link[=MatchingSplitDistance]{MatchingSplitDistance()}}
\item \code{\link[=InfoRobinsonFoulds]{InfoRobinsonFoulds()}}
\item \code{\link[=NyeSimilarity]{NyeSimilarity()}}
\item \code{\link[=JaccardRobinsonFoulds]{JaccardRobinsonFoulds()}}
}

The number of \acronym{OpenMP} threads is controlled by the standard
\code{"mc.cores"} option:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(mc.cores = parallel::detectCores())  # use all available cores
options(mc.cores = 4L)                        # or a fixed number
}\if{html}{\out{</div>}}

The default is \code{1} (single-threaded).
}

\subsection{R parallel cluster}{

\code{StartParallel()} creates an R socket cluster (via \code{\link[=makeCluster]{makeCluster()}}) and
registers it for use by TreeDist.  \code{SetParallel()} registers a pre-existing
cluster.  \code{StopParallel()} stops the cluster and releases resources.

\strong{When to use \code{StartParallel()}:} for metrics that do not have an
\acronym{OpenMP} batch path, namely tree-object-based distances such as
\code{\link[=NNIDist]{NNIDist()}} and \code{\link[=MASTSize]{MASTSize()}} / \code{\link[=MASTInfo]{MASTInfo()}}, or any function called
via \code{\link[=CompareAll]{CompareAll()}}.  R-cluster parallelism carries a serialisation overhead
of ~2–3 s, so it is only beneficial for large problems.

\strong{When \emph{not} to use \code{StartParallel()}:} for the split-based metrics listed
above.  Registering a cluster disables the \acronym{OpenMP} batch path for
those functions, replacing a thread-local C++ loop with inter-process
communication — which is slower at every problem size measured.  Call
\code{StopParallel()} before computing split-based distances if a cluster is
active.
}
}
\examples{
# OpenMP parallelism: set mc.cores before calling any split-based metric.
options(mc.cores = 2L)
# MutualClusteringInfo(trees)  # uses 2 OpenMP threads automatically
options(mc.cores = NULL)  # restore default (single-threaded)

if (interactive()) {
  # R cluster: beneficial for NNIDist, MASTSize/MASTInfo, CompareAll(), etc.
  # Do NOT activate while computing split-based distances (MCI, SPI, MSI, …)
  # as it bypasses the faster OpenMP path.
  library("TreeTools", quietly = TRUE)
  nCores <- ceiling(parallel::detectCores() / 2)
  StartParallel(nCores) # Takes a few seconds to set up processes
  GetParallel()
  CompareAll(as.phylo(0:6, 100), NNIDist)
  StopParallel() # Returns system resources
}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Information.R
\encoding{UTF-8}
\name{MeilaVariationOfInformation}
\alias{MeilaVariationOfInformation}
\alias{MeilaMutualInformation}
\title{Use variation of clustering information to compare pairs of splits}
\usage{
MeilaVariationOfInformation(split1, split2)

MeilaMutualInformation(split1, split2)
}
\arguments{
\item{split1, split2}{Logical vectors listing leaves in a consistent order,
identifying each leaf as a member of the ingroup (\code{TRUE}) or outgroup
(\code{FALSE}) of the split in question.}
}
\value{
\code{MeilaVariationOfInformation()} returns the variation of (clustering)
information, measured in bits.

\code{MeilaMutualInformation()} returns the mutual information,
measured in bits.
}
\description{
Compare a pair of splits viewed as clusterings of taxa, using the variation
of clustering information proposed by \insertCite{Meila2007}{TreeDist}.
}
\details{
This is equivalent to the mutual clustering information
\insertCite{Vinh2010}{TreeDist}.
For the total information content, multiply the VoI by the number of leaves.
}
\examples{
# Maximum variation = information content of each split separately
A <- TRUE
B <- FALSE
MeilaVariationOfInformation(c(A, A, A, B, B, B), c(A, A, A, A, A, A))
Entropy(c(3, 3) / 6) + Entropy(c(0, 6) / 6)

# Minimum variation = 0
MeilaVariationOfInformation(c(A, A, A, B, B, B), c(A, A, A, B, B, B))

# Not always possible for two evenly-sized splits to reach maximum
# variation of information
Entropy(c(3, 3) / 6) * 2  # = 2
MeilaVariationOfInformation(c(A, A, A,B ,B, B), c(A, B, A, B, A, B)) # < 2

# Phylogenetically uninformative groupings contain spliting information
Entropy(c(1, 5) / 6)
MeilaVariationOfInformation(c(B, A, A, A, A, A), c(A, A, A, A, A, B))
}
\references{
\insertAllCited{}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HPart.R
\name{clone}
\alias{clone}
\alias{clone.HPart}
\title{Clone / duplicate an object
\code{clone()} physically duplicates objects}
\usage{
clone(x, ...)

\method{clone}{HPart}(x, tipLabels = attr(x, "tip.label"), ...)
}
\arguments{
\item{x}{the object to be cloned}

\item{\dots}{additional parameters for methods}

\item{tipLabels}{Character vector specifying sequence in which to order
tip labels.}
}
\value{
\code{clone()} typically returns an object of the same class and "value"
as the input \code{x}.
}
\description{
Clone / duplicate an object
\code{clone()} physically duplicates objects
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{TreeDistPlot}
\alias{TreeDistPlot}
\title{Plot a simple tree}
\usage{
TreeDistPlot(
  tr,
  title = NULL,
  bold = NULL,
  leaveRoom = FALSE,
  prune = integer(0),
  graft = integer(0),
  edge.color = "black",
  edge.width = NULL,
  ...
)
}
\arguments{
\item{tr}{A tree of class \code{phylo}, with leaves labelled as integers.}

\item{title}{\code{main} title for the plot.}

\item{bold}{Integer specifying which leaves to print in bold.}

\item{leaveRoom}{Logical specifying whether to leave space to print
tree distances beneath the plotted tree.}

\item{prune, graft}{Integer vectors specifying which edges to highlight as
pruned and grafted.}

\item{edge.color}{Additional parameter to \code{plot.phylo}; will be overridden
by \code{prune} and \code{graft} as requested.}

\item{edge.width, \dots}{Additional parameters to \code{plot.phylo}.}
}
\description{
Convenience plotting function used in vignettes and documentation.
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kmeanspp.R
\name{KMeansPP}
\alias{KMeansPP}
\title{k-means++ clustering}
\usage{
KMeansPP(x, k = 2, nstart = 10, ...)
}
\arguments{
\item{x}{Numeric matrix of data, or an object that can be coerced to such a
matrix (such as a numeric vector or a data frame with all numeric columns).}

\item{k}{Integer specifying the number of clusters, \emph{k}.}

\item{nstart}{Positive integer specifying how many random sets should be
chosen}

\item{\dots}{additional arguments passed to \code{\link[stats]{kmeans}}}
}
\description{
k-means++ clustering \insertCite{Arthur2007}{TreeDist} improves the speed and
accuracy of standard \code{\link[stats]{kmeans}} clustering
\insertCite{Hartigan1979}{TreeDist} by preferring initial cluster centres
that are far from others.
A scalable version of the algorithm has been proposed for larger data sets
\insertCite{Bahmani2012}{TreeDist}, but is not implemented here.
}
\examples{
# Generate random points
set.seed(1)
x <- cbind(c(rnorm(10, -5), rnorm(5, 1), rnorm(10, 6)),
           c(rnorm(5, 0), rnorm(15, 4), rnorm(5, 0)))

# Conventional k-means may perform poorly
klusters <- kmeans(x, cent = 5)
plot(x, col = klusters$cluster, pch = rep(15:19, each = 5))

# Here, k-means++ recovers a better clustering
plusters <- KMeansPP(x, k = 5)
plot(x, col = plusters$cluster, pch = rep(15:19, each = 5))
}
\references{
\insertAllCited{}
}
\seealso{
\code{\link[stats]{kmeans}}

Other cluster functions:
\code{\link{cluster-statistics}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{cluster functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance.R
\name{SplitsCompatible}
\alias{SplitsCompatible}
\title{Are splits compatible?}
\usage{
SplitsCompatible(split1, split2)
}
\arguments{
\item{split1, split2}{Logical vectors listing leaves in a consistent order,
identifying each leaf as a member of the ingroup (\code{TRUE}) or outgroup
(\code{FALSE}) of the split in question.}
}
\value{
\code{SplitsCompatible()} returns a logical specifying whether the splits
provided are compatible with one another.
}
\description{
Determine whether splits are compatible (concave); i.e. they can both occur
on a single tree.
}
\examples{
A <- TRUE
B <- FALSE
SplitsCompatible(c(A, A, A, B, B, B),
                 c(A, A, B, B, B, B))
SplitsCompatible(c(A, A, A, B, B, B),
                 c(A, A, B, B, B, A))
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_transfer.R
\name{TransferDist}
\alias{TransferDist}
\alias{TransferDistance}
\alias{TransferDistSplits}
\title{Transfer dissimilarity between phylogenetic trees}
\usage{
TransferDist(
  tree1,
  tree2 = NULL,
  scale = TRUE,
  normalize = FALSE,
  reportMatching = FALSE
)

TransferDistance(
  tree1,
  tree2 = NULL,
  scale = TRUE,
  normalize = FALSE,
  reportMatching = FALSE
)

TransferDistSplits(
  splits1,
  splits2,
  nTip = attr(splits1, "nTip"),
  scale = TRUE,
  reportMatching = FALSE
)
}
\arguments{
\item{tree1, tree2}{Trees of class \code{phylo}, with leaves labelled identically,
or lists of such trees to undergo pairwise comparison.  Where implemented,
\code{tree2 = NULL} will compute distances between each pair of trees in the list
\code{tree1} using a fast algorithm based on
\insertCite{Day1985;textual}{TreeDist}.}

\item{scale}{Logical; if \code{TRUE} (default), use the scaled transfer
dissimilarity.  If \code{FALSE}, use the unscaled transfer dissimilarity.}

\item{normalize}{If a numeric value is provided, this will be used as a
maximum value against which to rescale results.
If \code{TRUE}, results will be rescaled against a maximum value calculated from
the specified tree sizes and topology, as specified in the "Normalization"
section below.
If \code{FALSE}, results will not be rescaled.}

\item{reportMatching}{Logical specifying whether to return the clade
matchings as an attribute of the score.}

\item{splits1, splits2}{Logical matrices where each row corresponds to a leaf,
either listed in the same order or bearing identical names (in any sequence),
and each column corresponds to a split, such that each leaf is identified as
a member of the ingroup (\code{TRUE}) or outgroup (\code{FALSE}) of the respective
split.}

\item{nTip}{(Optional) Integer specifying the number of leaves in each split.}
}
\value{
\code{TransferDist()} returns an object of class \code{dist} (if \code{tree2} is
\code{NULL}), a numeric matrix (if both \code{tree1} and \code{tree2} are lists), or a
numeric value (for a single pair).  If \code{reportMatching = TRUE}, the
return value carries \code{matching} and \code{pairScores} attributes.
}
\description{
Compute the transfer dissimilarity between phylogenetic trees, as defined
by \insertCite{Takazawa2026;textual}{TreeDist}.
The transfer dissimilarity uses the transfer distance
\insertCite{Lemoine2018}{TreeDist} to compare bipartitions, providing a
finer-grained measure than the Robinson--Foulds distance.  Each split in
each tree is scored by how many taxa must be moved to match its closest
counterpart in the other tree, and these scores are summed.
}
\details{
The \code{scaled} variant divides each split's contribution by its depth minus
one, giving equal weight to all splits regardless of their depth (analogous
to the Robinson--Foulds distance).  The \code{unscaled} variant uses raw transfer
distances, giving more weight to deep splits.  Neither variant satisfies
the triangle inequality for trees with six or more tips.
}
\section{Normalization}{


When \code{normalize = TRUE}, the scaled transfer dissimilarity is divided by
\code{2 * (n - 3)}, placing it in the range [0, 1].  The unscaled version is
divided by the maximum possible unscaled dissimilarity
(following \insertCite{Takazawa2026;textual}{TreeDist}).
}

\examples{
library(TreeTools)
TransferDist(BalancedTree(8), PectinateTree(8))
TransferDist(BalancedTree(8), PectinateTree(8), scale = FALSE)

# All-pairs
TransferDist(as.phylo(0:5, 8))

}
\references{
\insertAllCited{}
}
\seealso{
Other tree distances:
\code{\link[=HierarchicalMutualInfo]{HierarchicalMutualInfo()}},
\code{\link[=JaccardRobinsonFoulds]{JaccardRobinsonFoulds()}},
\code{\link[=KendallColijn]{KendallColijn()}},
\code{\link[=MASTSize]{MASTSize()}},
\code{\link[=MatchingSplitDistance]{MatchingSplitDistance()}},
\code{\link[=NNIDist]{NNIDist()}},
\code{\link[=NyeSimilarity]{NyeSimilarity()}},
\code{\link[=PathDist]{PathDist()}},
\code{\link{Robinson-Foulds}},
\code{\link[=SPRDist]{SPRDist()}},
\code{\link[=TreeDistance]{TreeDistance()}}
}
\concept{tree distances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_spr.R
\name{.SPRPairDeO}
\alias{.SPRPairDeO}
\title{SPR approximation via Oliveira Martins \emph{et al.} (2008)}
\usage{
.SPRPairDeO(tree1, tree2, check = TRUE)
}
\description{
SPR approximation via Oliveira Martins \emph{et al.} (2008)
}
\examples{
# de Oliveira Martins et al 2008, fig. 7
tree1 <- ape::read.tree(text = "((1, 2), ((a, b), (c, d)), (3, (4, (5, (6, 7)))));")
tree2 <- ape::read.tree(text = "((1, 2), 3, (4, (5, (((a, b), (c, d)), (6, 7)))));")
oPar <- par(mfrow =c(2, 1), mar = rep(0, 4))
plot(tree1)
plot(tree2)
par(oPar)
SPRDist(tree1, tree2, method = "deO")
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lap.R
\name{LAPJV}
\alias{LAPJV}
\title{Solve linear assignment problem using LAPJV}
\usage{
LAPJV(x)
}
\arguments{
\item{x}{Matrix of costs.}
}
\value{
\code{LAPJV()} returns a list with two entries: \code{score}, the score of the
optimal matching;
and \code{matching}, the columns matched to each row of the matrix in turn.
}
\description{
Use the algorithm of \insertCite{Jonker1987;textual}{TreeDist} to solve the
Linear Sum Assignment Problem (LSAP).
}
\details{
The Linear Assignment Problem seeks to match each row of a matrix with a
column, such that the cost of the matching is minimized.

The Jonker & Volgenant approach is a faster alternative to the Hungarian
algorithm \insertCite{Munkres1957}{TreeDist}, which is implemented in
\code{clue::solve_LSAP()}.

Note: the JV algorithm expects integers. In order to apply the function
to a non-integer \emph{n}, as in the tree distance calculations in this package,
each \emph{n} is multiplied by the largest available integer before applying
the JV algorithm.  If two values of \emph{n} exhibit a trivial difference --
e.g. due to floating point errors -- then this can lead to interminable
run times.  (If numbers of the magnitude of billions differ only in their
last significant digit, then the JV algorithm may undergo billions of
iterations.)  To avoid this, integers over 2^22 that differ by a value of
8 or less are treated as equal.
}
\examples{
problem <- matrix(c(7, 9, 8, 9, 9,
                    2, 8, 5, 7, 9,
                    1, 6, 6, 9, 9,
                    3, 6, 2, 2, 9), 4, 5, byrow = TRUE)

LAPJV(problem)
}
\references{
\insertAllCited{}
}
\seealso{
Implementations of the Hungarian algorithm exist in \pkg{adagio},
\pkg{RcppHungarian}, and \pkg{clue} and \pkg{lpSolve}; for larger matrices,
these are substantially slower. (See discussion at \href{https://stackoverflow.com/questions/72806265/}{Stack Overflow}.)

The JV algorithm is implemented for square matrices in the Bioconductor
package \href{https://www.bioconductor.org/packages/release/bioc/html/GraphAlignment.html}{\code{GraphAlignment::LinearAssignment()}}.
}
\author{
\href{https://github.com/yongyanghz/LAPJV-algorithm-c/blob/master/src/lap.cpp}{C++ code}
by Roy Jonker, MagicLogic Optimization Inc. \href{mailto:roy_jonker@magiclogic.com}{roy_jonker@magiclogic.com},
with contributions from Yong Yang \href{mailto:yongyanglink@gmail.com}{yongyanglink@gmail.com}, after
\href{https://uk.mathworks.com/matlabcentral/profile/authors/69713}{Yi Cao}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_msd.R
\name{MatchingSplitDistance}
\alias{MatchingSplitDistance}
\alias{MatchingSplitDistanceSplits}
\title{Matching Split Distance}
\usage{
MatchingSplitDistance(
  tree1,
  tree2 = NULL,
  normalize = FALSE,
  reportMatching = FALSE
)

MatchingSplitDistanceSplits(
  splits1,
  splits2,
  nTip = attr(splits1, "nTip"),
  normalize = TRUE,
  reportMatching = FALSE
)
}
\arguments{
\item{tree1, tree2}{Trees of class \code{phylo}, with leaves labelled identically,
or lists of such trees to undergo pairwise comparison.  Where implemented,
\code{tree2 = NULL} will compute distances between each pair of trees in the list
\code{tree1} using a fast algorithm based on
\insertCite{Day1985;textual}{TreeDist}.}

\item{normalize}{If a numeric value is provided, this will be used as a
maximum value against which to rescale results.
If \code{TRUE}, results will be rescaled against a maximum value calculated from
the specified tree sizes and topology, as specified in the "Normalization"
section below.
If \code{FALSE}, results will not be rescaled.}

\item{reportMatching}{Logical specifying whether to return the clade
matchings as an attribute of the score.}

\item{splits1, splits2}{Logical matrices where each row corresponds to a leaf,
either listed in the same order or bearing identical names (in any sequence),
and each column corresponds to a split, such that each leaf is identified as
a member of the ingroup (\code{TRUE}) or outgroup (\code{FALSE}) of the respective
split.}

\item{nTip}{(Optional) Integer specifying the number of leaves in each split.}
}
\value{
\code{MatchingSplitDistance()} returns an array of numerics providing the
distances between each  pair of trees in \code{tree1} and \code{tree2},
or \code{splits1} and \code{splits2}.
}
\description{
Calculate the
\href{https://ms609.github.io/TreeDist/articles/Generalized-RF.html#matching-split-distance}{Matching Split Distance}
\insertCite{Bogdanowicz2012,Lin2012}{TreeDist} for unrooted binary trees.
}
\details{
Trees need not contain identical leaves; scores are based on the leaves that
trees hold in common.  Check for unexpected differences in tip labelling
with \code{setdiff(TipLabels(tree1), TipLabels(tree2))}.
}
\section{Normalization}{


A normalization value or function must be provided in order to return a
normalized value.  If you are aware of a generalised formula, please
let me know by
\href{https://github.com/ms609/TreeDist/issues/new}{creating a GitHub issue}
so that it can be implemented.
}

\examples{
MatchingSplitDistance(lapply(rep(8, 5), ape::rtree), normalize = 16)

MatchingSplitDistance(TreeTools::BalancedTree(6),
                      TreeTools::PectinateTree(6),
                      reportMatching = TRUE)

VisualizeMatching(MatchingSplitDistance, TreeTools::BalancedTree(6),
                  TreeTools::PectinateTree(6))
}
\references{
\insertAllCited{}
}
\seealso{
Other tree distances:
\code{\link[=HierarchicalMutualInfo]{HierarchicalMutualInfo()}},
\code{\link[=JaccardRobinsonFoulds]{JaccardRobinsonFoulds()}},
\code{\link[=KendallColijn]{KendallColijn()}},
\code{\link[=MASTSize]{MASTSize()}},
\code{\link[=NNIDist]{NNIDist()}},
\code{\link[=NyeSimilarity]{NyeSimilarity()}},
\code{\link[=PathDist]{PathDist()}},
\code{\link{Robinson-Foulds}},
\code{\link[=SPRDist]{SPRDist()}},
\code{\link[=TransferDist]{TransferDist()}},
\code{\link[=TreeDistance]{TreeDistance()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree distances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_nye.R
\encoding{UTF-8}
\name{JaccardRobinsonFoulds}
\alias{JaccardRobinsonFoulds}
\alias{JaccardSplitSimilarity}
\title{Jaccard–Robinson–Foulds metric}
\usage{
JaccardRobinsonFoulds(
  tree1,
  tree2 = NULL,
  k = 1L,
  allowConflict = TRUE,
  similarity = FALSE,
  normalize = FALSE,
  reportMatching = FALSE
)

JaccardSplitSimilarity(
  splits1,
  splits2,
  nTip = attr(splits1, "nTip"),
  k = 1L,
  allowConflict = TRUE,
  reportMatching = FALSE
)
}
\arguments{
\item{tree1, tree2}{Trees of class \code{phylo}, with leaves labelled identically,
or lists of such trees to undergo pairwise comparison.  Where implemented,
\code{tree2 = NULL} will compute distances between each pair of trees in the list
\code{tree1} using a fast algorithm based on
\insertCite{Day1985;textual}{TreeDist}.}

\item{k}{An arbitrary exponent to which to raise the Jaccard index.
Integer values greater than one are anticipated by Böcker \emph{et al}.
The Nye \emph{et al}. metric uses \code{k = 1}.
As k increases towards infinity, the metric converges to the Robinson–Foulds
metric.}

\item{allowConflict}{Logical specifying whether to allow conflicting splits
to be paired. If \code{FALSE}, such pairings will be allocated a similarity
score of zero.}

\item{similarity}{Logical specifying whether to report the result as a tree
similarity, rather than a difference.}

\item{normalize}{If a numeric value is provided, this will be used as a
maximum value against which to rescale results.
If \code{TRUE}, results will be rescaled against a maximum value calculated from
the specified tree sizes and topology, as specified in the "Normalization"
section below.
If \code{FALSE}, results will not be rescaled.}

\item{reportMatching}{Logical specifying whether to return the clade
matchings as an attribute of the score.}

\item{splits1, splits2}{Logical matrices where each row corresponds to a leaf,
either listed in the same order or bearing identical names (in any sequence),
and each column corresponds to a split, such that each leaf is identified as
a member of the ingroup (\code{TRUE}) or outgroup (\code{FALSE}) of the respective
split.}

\item{nTip}{(Optional) Integer specifying the number of leaves in each split.}
}
\value{
\code{JaccardRobinsonFoulds()} returns an array of numerics providing the
distances between each  pair of trees in \code{tree1} and \code{tree2},
or \code{splits1} and \code{splits2}.
}
\description{
Calculate the
\href{https://ms609.github.io/TreeDist/articles/Generalized-RF.html#jaccard-robinson-foulds-metric}{Jaccard–Robinson–Foulds metric}
\insertCite{Bocker2013}{TreeDist}, a
\href{https://ms609.github.io/TreeDist/articles/Robinson-Foulds.html#generalized-robinson-foulds-distances}{Generalized Robinson–Foulds metric}.
}
\details{
In short, the Jaccard–Robinson–Foulds
metric is a generalized Robinson-Foulds
metric: it finds the optimal matching that pairs each split in one tree with
a similar split in the second.
Matchings are scored according to the size of the largest split that is
consistent with both of them, normalized against the Jaccard index, and
raised to an arbitrary exponent.
A more detailed explanation is provided in the
\href{https://ms609.github.io/TreeDist/articles/Generalized-RF.html#jaccard-robinson-foulds-metric}{vignettes}.

By default, conflicting splits may be paired.

Note that the settings \verb{k = 1, allowConflict = TRUE, similarity = TRUE}
give the similarity metric of \insertCite{Nye2006;textual}{TreeDist};
a slightly faster implementation of this metric is available as
\code{\link[=NyeSimilarity]{NyeSimilarity()}}.

The examples section below details how to visualize matchings with
non-default parameter values.

Trees need not contain identical leaves; scores are based on the leaves that
trees hold in common.  Check for unexpected differences in tip labelling
with \code{setdiff(TipLabels(tree1), TipLabels(tree2))}.
}
\section{Normalization}{


If \code{normalize = TRUE}, then results will be rescaled from zero to one by
dividing by the maximum possible value for trees of the given topologies,
which is equal to the sum of the number of splits in each tree.
You may wish to normalize instead against the maximum number of splits
present in a pair of trees with \emph{n} leaves, by specifying
\code{normalize = n - 3}.
}

\examples{
set.seed(2)
tree1 <- ape::rtree(10)
tree2 <- ape::rtree(10)
JaccardRobinsonFoulds(tree1, tree2, k = 2, allowConflict = FALSE)
JaccardRobinsonFoulds(tree1, tree2, k = 2, allowConflict = TRUE)

JRF2 <- function(tree1, tree2, ...) 
  JaccardRobinsonFoulds(tree1, tree2, k = 2, allowConflict = FALSE, ...)
  
VisualizeMatching(JRF2, tree1, tree2, matchZeros = FALSE)
}
\references{
\insertAllCited{}
}
\seealso{
Other tree distances:
\code{\link[=HierarchicalMutualInfo]{HierarchicalMutualInfo()}},
\code{\link[=KendallColijn]{KendallColijn()}},
\code{\link[=MASTSize]{MASTSize()}},
\code{\link[=MatchingSplitDistance]{MatchingSplitDistance()}},
\code{\link[=NNIDist]{NNIDist()}},
\code{\link[=NyeSimilarity]{NyeSimilarity()}},
\code{\link[=PathDist]{PathDist()}},
\code{\link{Robinson-Foulds}},
\code{\link[=SPRDist]{SPRDist()}},
\code{\link[=TransferDist]{TransferDist()}},
\code{\link[=TreeDistance]{TreeDistance()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree distances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_nni.R
\encoding{UTF-8}
\name{NNIDist}
\alias{NNIDist}
\alias{NNIDiameter}
\title{Approximate Nearest Neighbour Interchange distance}
\usage{
NNIDist(tree1, tree2 = tree1)

NNIDiameter(tree)
}
\arguments{
\item{tree1, tree2}{Single trees of class \code{phylo} to undergo comparison.}

\item{tree}{Object of supported class representing a tree or list of trees,
or an integer specifying the number of leaves in a tree/trees.}
}
\value{
\code{NNIDist()} returns, for each pair of trees, a named vector
containing three integers:
\itemize{
\item \code{lower} is a lower bound on the NNI distance, and corresponds
to the RF distance between the trees.
\item \code{tight_upper} is an upper bound on the distance, based on calculated
maximum diameters for trees with < 13 leaves.  \emph{NA} is returned if trees are
too different to employ this approach.
\item \code{loose_upper} is a looser upper bound on the distance, using \emph{n} log \emph{n} +
O(\emph{n}).
}

\code{NNIDiameter()} returns a matrix specifying (bounds on) the diameter
of the NNI distance metric on the specified tree(s).
Columns correspond to:
\itemize{
\item \code{liMin}:  \deqn{n - 3}{_n_ &minus; 3;}, a lower bound on the diameter
(Li \emph{et al.} 1996);
\item \code{fackMin}: Lower bound on diameter following Fack \emph{et al}. (2002), i.e.
\deqn{\log2{N!} / 4}{log&#8322; (_N_!) / 4};
\item \code{min}: The larger of \code{liMin} and \code{fackMin};
\item \code{exact}: The exact value of the diameter, where \emph{n} < 13;
\item \code{liMax}: Upper bound on diameter following Li \emph{et al.} (1996), i.e.
\deqn{n \log2{n} + \textrm{O}(n)}{n log&#8322; _n_ + O(_n_)};
\item \code{fackMax}: Upper bound on diameter following Fack \emph{et al}. (2002), i.e.
(\deqn{N - 2}{_N_ &minus; 2}) ceiling(\deqn{\log2{n}}{log&#8322; _N_})
\itemize{
\item \emph{N};
}
\item \code{max}: The smaller of \code{liMax} and \code{fackMax};
}

where \emph{n} is the number of leaves, and \emph{N} the number of internal nodes,
i.e. \deqn{n - 2}{_n_ &minus; 2}.
}
\description{
Use the approach of \insertCite{Li1996;textual}{TreeDist} to approximate the
Nearest Neighbour Interchange distance \insertCite{Robinson1971}{TreeDist}
between phylogenetic trees.
}
\details{
In brief, this approximation algorithm works by identifying edges in one
tree that do not match edges in the second.  Each of these edges must
undergo at least one NNI operation in order to reconcile the trees.
Edges that match in both trees need never undergo an NNI operation, and
divide each tree into smaller regions.  By "cutting" matched edges into two,
a tree can be divided into a number of regions that solely comprise unmatched
edges.

These regions can be viewed as separate trees that need to be reconciled.
One way to reconcile these trees is to conduct a series of NNI operations
that reduce a tree to a pectinate (caterpillar) tree, then to conduct an
analogue of the mergesort algorithm.  This takes at most \emph{n} log \emph{n} + O(\emph{n})
NNI operations, and provides a loose upper bound on the NNI score.
The maximum number of moves for an \emph{n}-leaf tree
(\href{https://oeis.org/A182136}{OEIS A182136}) can be calculated exactly for
small trees \insertCite{Fack2002}{TreeDist}; this provides a tighter upper
bound, but is unavailable for \emph{n} > 12.
\code{NNIDiameter()} reports the limits on this bound.

\tabular{rccccccccccccc}{
Leaves:   \tab 1 \tab 2 \tab 3 \tab 4 \tab 5 \tab 6 \tab 7 \tab 8 \tab 9
\tab 10 \tab 11 \tab 12 \tab 13 \cr
Diameter: \tab 0 \tab 0 \tab 0 \tab 1 \tab 3 \tab 5 \tab 7 \tab 10 \tab 12
\tab 15 \tab 18 \tab 21 \tab ? \cr
}
}
\examples{
library("TreeTools", quietly = TRUE)

NNIDist(BalancedTree(7), PectinateTree(7))

NNIDist(BalancedTree(7), as.phylo(0:2, 7))
NNIDist(as.phylo(0:2, 7), PectinateTree(7))

NNIDist(list(bal = BalancedTree(7), pec = PectinateTree(7)),
        as.phylo(0:2, 7))

CompareAll(as.phylo(30:33, 8), NNIDist)
}
\references{
\insertAllCited{}
}
\seealso{
Other tree distances:
\code{\link[=HierarchicalMutualInfo]{HierarchicalMutualInfo()}},
\code{\link[=JaccardRobinsonFoulds]{JaccardRobinsonFoulds()}},
\code{\link[=KendallColijn]{KendallColijn()}},
\code{\link[=MASTSize]{MASTSize()}},
\code{\link[=MatchingSplitDistance]{MatchingSplitDistance()}},
\code{\link[=NyeSimilarity]{NyeSimilarity()}},
\code{\link[=PathDist]{PathDist()}},
\code{\link{Robinson-Foulds}},
\code{\link[=SPRDist]{SPRDist()}},
\code{\link[=TransferDist]{TransferDist()}},
\code{\link[=TreeDistance]{TreeDistance()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree distances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spectral_clustering.R
\name{SpectralEigens}
\alias{SpectralEigens}
\alias{SpectralClustering}
\title{Eigenvalues for spectral clustering}
\usage{
SpectralEigens(D, nn = 10L, nEig = 2L)

SpectralClustering(D, nn = 10L, nEig = 2L)
}
\arguments{
\item{D}{Square matrix or \code{dist} object containing Euclidean distances
between data points.}

\item{nn}{Integer specifying number of nearest neighbours to consider}

\item{nEig}{Integer specifying number of eigenvectors to retain.}
}
\value{
\code{SpectralEigens()} returns spectral eigenvalues that can then be
clustered using a method of choice.
}
\description{
Spectral clustering emphasizes nearest neighbours when forming clusters;
it avoids some of the issues that arise from clustering around means /
medoids.
}
\examples{
library("TreeTools", quietly = TRUE)
trees <- as.phylo(0:18, nTip = 8)
distances <- ClusteringInfoDistance(trees)
eigens <- SpectralEigens(distances)
# Perform clustering:
clusts <- KMeansPP(dist(eigens), k = 3)
plot(eigens, pch = 15, col = clusts$cluster)
plot(cmdscale(distances), pch = 15, col = clusts$cluster)
}
\seealso{
Other tree space functions:
\code{\link[=Islands]{Islands()}},
\code{\link[=MSTSegments]{MSTSegments()}},
\code{\link[=MapTrees]{MapTrees()}},
\code{\link[=MappingQuality]{MappingQuality()}},
\code{\link{cluster-statistics}},
\code{\link[=median.multiPhylo]{median.multiPhylo()}}
}
\author{
Adapted by MRS from script by \href{https://rpubs.com/nurakawa/spectral-clustering}{Nura Kawa}
}
\concept{tree space functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Information.R
\name{SplitSharedInformation}
\alias{SplitSharedInformation}
\alias{SplitDifferentInformation}
\alias{TreesConsistentWithTwoSplits}
\alias{LnTreesConsistentWithTwoSplits}
\alias{Log2TreesConsistentWithTwoSplits}
\title{Shared information content of two splits}
\usage{
SplitSharedInformation(n, A1, A2 = A1)

SplitDifferentInformation(n, A1, A2 = A1)

TreesConsistentWithTwoSplits(n, A1, A2 = A1)

LnTreesConsistentWithTwoSplits(n, A1, A2 = A1)

Log2TreesConsistentWithTwoSplits(n, A1, A2 = A1)

Log2TreesConsistentWithTwoSplits(n, A1, A2 = A1)
}
\arguments{
\item{n}{Integer specifying the number of leaves}

\item{A1, A2}{Integers specifying the number of taxa in \emph{A1} and \emph{A2},
once the splits have been arranged such that \emph{A1} fully overlaps with \emph{A2}.}
}
\value{
\code{TreesConsistentWithTwoSplits()} returns the number of unrooted bifurcating
trees consistent with two splits.

\code{SplitSharedInformation()} returns the phylogenetic information that two splits
have in common \insertCite{Meila2007}{TreeDist}, in bits.

\code{SplitDifferentInformation()} returns the amount of phylogenetic information
distinct to one of the two splits, in bits.
}
\description{
Calculate the phylogenetic information shared, or not shared, between two
splits.
See the \href{https://ms609.github.io/TreeDist/articles/information.html}{accompanying vignette}
for definitions.
}
\details{
Split \emph{S1} divides \emph{n} leaves into two splits, \emph{A1} and \emph{B1}.
Split \emph{S2} divides the same leaves into the splits \emph{A2} and \emph{B2}.

Splits must be named such that \emph{A1} fully overlaps with \emph{A2}:
that is to say, all taxa in \emph{A1} are also in \emph{A2}, or \emph{vice versa}.
Thus, all taxa in the smaller of \emph{A1} and \emph{A2} also occur in the larger.
}
\section{Functions}{
\itemize{
\item \code{SplitDifferentInformation()}: Different information between two splits.

\item \code{TreesConsistentWithTwoSplits()}: Number of trees consistent with two
splits.

\item \code{LnTreesConsistentWithTwoSplits()}: Natural logarithm of
\code{TreesConsistentWithTwoSplits()}.

\item \code{Log2TreesConsistentWithTwoSplits()}: Base two logarithm of
\code{TreesConsistentWithTwoSplits()}.

\item \code{Log2TreesConsistentWithTwoSplits()}: Base 2 logarithm of
\code{TreesConsistentWithTwoSplits()}.

}}
\examples{
  # Eight leaves, labelled A to H.
  # Split 1: ABCD|EFGH
  # Split 2: ABC|DEFGH
  # Let A1 = ABCD (four taxa), and A2 = ABC (three taxa).
  # A1 and A2 overlap (both contain ABC).
  
  TreesConsistentWithTwoSplits(n = 8, A1 = 4, A2 = 3)
  SplitSharedInformation(n = 8, A1 = 4, A2 = 3)
  SplitDifferentInformation(n = 8, A1 = 4, A2 = 3)

  # If splits are identical, then their shared information is the same
  # as the information of either split:
  SplitSharedInformation(n = 8, A1 = 3, A2 = 3)
  TreeTools::SplitInformation(3, 5)
}
\references{
\insertAllCited{}
}
\seealso{
Other information functions:
\code{\link[=SplitEntropy]{SplitEntropy()}},
\code{\link{TreeInfo}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{information functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance.R
\encoding{UTF-8}
\name{GeneralizedRF}
\alias{GeneralizedRF}
\title{Generalized Robinson–Foulds distance}
\usage{
GeneralizedRF(
  splits1,
  splits2,
  nTip,
  PairScorer,
  maximize,
  reportMatching,
  ...
)
}
\arguments{
\item{splits1, splits2}{Logical matrices where each row corresponds to a leaf,
either listed in the same order or bearing identical names (in any sequence),
and each column corresponds to a split, such that each leaf is identified as
a member of the ingroup (\code{TRUE}) or outgroup (\code{FALSE}) of the respective
split.}

\item{nTip}{Integer specifying the number of leaves in each split.}

\item{PairScorer}{function taking four arguments, \code{splits1}, \code{splits2},
\code{nSplits1}, \code{nSplits2}, which should return the score of each pair of splits
in a two-dimensional matrix.  Additional parameters may be specified via
\code{...}.}

\item{maximize}{Logical specifying whether the optimal matching maximizes
or minimizes the scores obtained by \code{PairScorer()}.}

\item{reportMatching}{Logical specifying whether to return the clade
matchings as an attribute of the score.}

\item{\dots}{Additional parameters to \code{PairScorer()}.}
}
\value{
A numeric value specifying the score of the tree pairs under the
specified pair scorer. If \code{reportMatching = TRUE}, attribute also list:
\itemize{
\item \code{matching}: which split in \code{splits2} is optimally matched to each split in
\code{split1} (\code{NA} if not matched);
\item \code{matchedSplits}: Textual representation of each match
\item \code{matchedScores}: Scores for matched split.
\item \code{pairScores}: Calculated scores for each possible matching of each split.
}
}
\description{
An internal function to calculate Generalized Robinson–Foulds
distances from splits.
}
\details{
Note that no checks will be made to confirm that \code{splits1} and \code{splits2}
contain the same leaves in the same order.
This is the responsibility of the calling function.
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_mast.R
\name{.MASTSizeEdges}
\alias{.MASTSizeEdges}
\title{Calculate MAST size from edge matrices.}
\usage{
.MASTSizeEdges(edge1, edge2, nTip)
}
\arguments{
\item{edge1}{Edge matrix of tree 1. \strong{Must be in postorder!}}

\item{edge2}{Edge matrix of tree 2.}

\item{nTip}{Integer specifying the number of leaves in each split.}
}
\description{
Internal function.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cluster_stats.R
\name{cluster-statistics}
\alias{cluster-statistics}
\alias{SumOfRanges}
\alias{SumOfVariances}
\alias{SumOfVars}
\alias{MeanCentroidDistance}
\alias{MeanCentDist}
\alias{MeanCentroidDist}
\alias{DistanceFromMedian}
\alias{DistFromMed}
\alias{MeanNN}
\alias{MeanMSTEdge}
\title{Cluster size statistics}
\usage{
SumOfRanges(x, cluster = 1)

SumOfVariances(x, cluster = 1)

SumOfVars(x, cluster = 1)

MeanCentroidDistance(x, cluster = 1, Average = mean)

MeanCentDist(x, cluster = 1, Average = mean)

MeanCentroidDist(x, cluster = 1, Average = mean)

DistanceFromMedian(x, cluster = 1, Average = mean)

DistFromMed(x, cluster = 1, Average = mean)

MeanNN(x, cluster = 1, Average = mean)

MeanMSTEdge(x, cluster = 1)
}
\arguments{
\item{x}{Matrix in which each row lists the coordinates of a point
in a Euclidian space; or, where supported, \code{dist} object specifying
distances between each pair of points.}

\item{cluster}{Optional integer vector specifying the cluster or group to
which each row in \code{x} belongs.}

\item{Average}{Function to use to summarize distances. Defaults to \code{mean};
specifying \code{median} returns a value akin to the median absolute divergence
(see \code{\link{mad}}).}
}
\value{
\code{SumOfRanges()} returns a numeric specifying the sum of ranges
within each cluster across all dimensions.

\code{SumOfVariances()} returns a numeric specifying the sum of variances
within each cluster across all dimensions.

\code{MeanCentroidDistance()} returns a numeric specifying the mean
distance from the centroid to points in each cluster.

\code{DistanceFromMedian()} returns a numeric specifying the mean distance
of each point (except the median) from the median point of its cluster.

\code{MeanNN()} returns a numeric specifying the mean distance from each
point within a cluster to its nearest neighbour.

\code{MeanMSTEdge()} returns a numeric specifying the mean length of an
edge in the minimum spanning tree of points within each cluster.
}
\description{
Cluster size statistics
}
\examples{
points <- rbind(matrix(1:16, 4), rep(1, 4), matrix(1:32, 8, 4) / 10)
cluster <- rep(1:3, c(4, 1, 8))

plot(
  points[, 1:2], # Plot first two dimensions of four-dimensional space
  col = cluster, pch = cluster, # Style by cluster membership
  asp = 1, # Fix aspect ratio to avoid distortion
  ann = FALSE, frame = FALSE # Simple axes
)

SumOfRanges(points, cluster)
SumOfVariances(points, cluster)
MeanCentroidDistance(points, cluster)
DistanceFromMedian(points, cluster)
MeanNN(points, cluster)
MeanMSTEdge(points, cluster)
}
\seealso{
Other tree space functions:
\code{\link[=Islands]{Islands()}},
\code{\link[=MSTSegments]{MSTSegments()}},
\code{\link[=MapTrees]{MapTrees()}},
\code{\link[=MappingQuality]{MappingQuality()}},
\code{\link[=SpectralEigens]{SpectralEigens()}},
\code{\link[=median.multiPhylo]{median.multiPhylo()}}

Other cluster functions:
\code{\link[=KMeansPP]{KMeansPP()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{cluster functions}
\concept{tree space functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shiny.R
\name{MapTrees}
\alias{MapTrees}
\alias{Project}
\title{Graphical user interface for mapping distances and analysing
tree space}
\usage{
MapTrees()

Project()
}
\description{
\code{MapTrees()} launches a "Shiny" application for the visualization and
evaluation of tree spaces.
}
\section{Input tab}{
The input tab allows for the upload of sets of phylogenetic trees from file.
Trees at the start or end of a file can be excluded, and the number of trees
can be brought down to a manageable number by uniformly subsampling every
_n_th tree.  Samples of c. 100 trees can be analysed in seconds;
analysis of larger samples will take longer, particularly with slower
methods (e.g. quartet distances; Kruskal-1 MDS; large minimum spanning trees).

Different batches can be plotted with different colours / symbols.

If each tree is associated with a property -- for example, the data or method
used to generate it, or its stratigraphic congruence -- a list of properties
for each tree, with one entry per line/row, can be uploaded along with
the trees.  Points in tree space can then be styled according to the
corresponding property.

If trees are subsampled (using the "Sample every" slider), then the values
in the tree properties file can also be subsampled accordingly.
Unfortunately there is not yet support for multiple point property files;
one file will be applied to all trees, in the sequence that they were added
to memory.
}

\section{Analysis tab}{
Select from a suite of distance methods: clustering information and
phylogenetic information are quick and satisfactory; quartet is slow but
gives slightly better mappings; path is very fast but may not reflect
evolutionary signal very well; and Robinson--Foulds should probably never
be used for analysis; it is included for comparison purposes.

Principle components mappings should suffice for most purposes;
Sammon and Kruskal mappings are slower and seldom differ by much,
in character or quality, but may emphasize outliers more.

Partitioning around medoids or minimax-linkage hierarchical clustering
will typically find a close-to-optimal clustering where one exists;
select additional methods for a more exhaustive search.
To avoid redundant calculation, clusterings are only updated when
"recalculate clustering" is clicked, or the "maximum cluster number" slider
is modified; clustering solutions using more than this many clusters are
not considered
Clusterings with silhouette coefficients < 0.25 are unlikely to represent
genuine structure and are not reported or depicted.
}

\section{Display tab}{
Up to 15 dimensions can be depicted; the quality of a mapping -- that is,
the faithfulness of mapped distances to true tree-to-tree distances --
is quantified by the product of the Trustworthiness and Continuity metrics,
which should exceed 0.9 (at least).

An interactive 3D plot can be explored by dragging the mouse and scrolling,
but do be careful to check that three dimensions are enough to depict your
data accurately.

The minimum spanning tree is the shortest possible line selecting the chosen
subsample of trees; if it takes a convoluted zig-zagging route, then the
mapping is doing a poor job of reflecting true tree to tree distances.

Convex hulls are the smallest polygons enclosing all points in each cluster;
they are handy for spotting clusters, but their area does not correspond
to a genuine quantity, so should not be interpreted.

Tree numbers correspond to the sequence of trees in their original input
file, before subsampling.

Each tree is denoted by a point, whose symbol can be styled according to
cluster membership or according to the file that contains the tree,
with each click of "Add to existing" on the input tab constituting a
new batch with a new symbol.

Points can be coloured according to a category -- the cluster or batch to
which they belong, or custom data provided in the Point Property File
on the input tab -- or continuously, either by the sequence in which they
were added to memory, or according to custom data.
}

\section{Exporting tree spaces}{
A mapping can be saved to PDF or as a PNG bitmap at the size selected.
}

\section{References}{
A list of references employed when constructing the tree space is populated
according to the methods used; it would be appropriate to cite and briefly
discuss these studies in any publication using figures generated using
this application.  The application itself can be cited using
\insertCite{SmithDist,SmithSpace;textual}{TreeDist}
}

\references{
\insertAllCited{}
}
\seealso{
Full detail of tree space analysis in R is provided in the accompanying
\href{https://ms609.github.io/TreeDist/articles/treespace.html}{vignette}.

Other tree space functions:
\code{\link[=Islands]{Islands()}},
\code{\link[=MSTSegments]{MSTSegments()}},
\code{\link[=MappingQuality]{MappingQuality()}},
\code{\link[=SpectralEigens]{SpectralEigens()}},
\code{\link{cluster-statistics}},
\code{\link[=median.multiPhylo]{median.multiPhylo()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree space functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_kendall-colijn.R
\encoding{UTF-8}
\name{KendallColijn}
\alias{KendallColijn}
\alias{KCVector}
\alias{PathVector}
\alias{SplitVector}
\alias{KCDiameter}
\title{Kendall–Colijn distance}
\usage{
KendallColijn(tree1, tree2 = NULL, Vector = KCVector)

KCVector(tree)

PathVector(tree)

SplitVector(tree)

KCDiameter(tree)
}
\arguments{
\item{tree1, tree2}{Trees of class \code{phylo}, with leaves labelled identically,
or lists of such trees to undergo pairwise comparison.  Where implemented,
\code{tree2 = NULL} will compute distances between each pair of trees in the list
\code{tree1} using a fast algorithm based on
\insertCite{Day1985;textual}{TreeDist}.}

\item{Vector}{Function converting a tree to a numeric vector.

\code{KCVector}, the default, returns the number of edges between the common
ancestor of each pair of leaves and the root of the tree
\insertCite{@per @Kendall2016}{TreeDist}.

\code{PathVector} returns the number of edges between each pair of leaves
\insertCite{@per @Steel1993}{TreeDist}.

\code{SplitVector} returns the size of the smallest split that contains each
pair of leaves (per \insertCite{SmithSpace;nobrackets}{TreeDist}).}

\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}
}
\value{
\code{KendallColijn()} returns an array of numerics providing the
distances between each  pair of trees in \code{tree1} and \code{tree2},
or \code{splits1} and \code{splits2}.

\code{KCDiameter()} returns the value of the Kendall & Colijn's (2016)
metric distance between two pectinate trees with \emph{n} leaves ordered in
the opposite direction, which I suggest (without any attempt at a proof) may
be a useful proxy for the diameter (i.e. maximum value) of the K–C
metric.
}
\description{
Calculate the Kendall–Colijn tree distance, a measure related to the
path difference.
}
\details{
The Kendall–Colijn distance works by measuring, for each pair of
leaves, the distance from the most recent common ancestor of those leaves
and the root node.
For a given tree, this produces a vector of values recording the
distance-from-the-root of each most recent common ancestor of each pair of
leaves.

Two trees are compared by taking the Euclidean distance between the
respective vectors.  This is calculated by taking the square root of the sum
of the squares of the differences between the vectors.

An analogous distance can be created from any vector representation of a
tree.
The split size vector metric \insertCite{SmithSpace}{TreeDist} is an attempt
to mimic the Kendall Colijn metric in situations where the position of
the root should not be afforded special significance; and the path distance
\insertCite{Steel1993}{TreeDist} is a familiar alternative whose underlying
vector measures the distance of the last common ancestor of each pair
of leaves from the leaves themselves, i.e. the length of the path from one
leaf to another.

None of these vector-based methods performs as well as other tree distances
in measuring similarities in the relationships implied by a pair of trees
\insertCite{SmithDist}{TreeDist}; in particular, the Kendall Colijn
metric is strongly influenced by tree balance, and may not be appropriate
for a suite of common applications \insertCite{SmithSpace}{TreeDist}.
}
\section{Functions}{
\itemize{
\item \code{KCVector()}: Creates a vector that characterises a rooted tree,
as described in \insertCite{Kendall2016;textual}{TreeDist}.

\item \code{PathVector()}: Creates a vector reporting the number of edges
between each pair of leaves, per the path metric of
\insertCite{Steel1993;textual}{TreeDist}.

\item \code{SplitVector()}: Creates a vector reporting the smallest split
containing each pair of leaves, per the metric proposed in
\insertCite{SmithSpace;textual}{TreeDist}.

}}
\examples{
KendallColijn(TreeTools::BalancedTree(8), TreeTools::PectinateTree(8))

set.seed(0)
KendallColijn(TreeTools::BalancedTree(8), lapply(rep(8, 3), ape::rtree))
KendallColijn(lapply(rep(8, 4), ape::rtree))

KendallColijn(lapply(rep(8, 4), ape::rtree), Vector = SplitVector)

# Notice that changing tree shape close to the root results in much
# larger differences
tree1 <- ape::read.tree(text = "(a, (b, (c, (d, (e, (f, (g, h)))))));")
tree2 <- ape::read.tree(text = "(a, ((b, c), (d, (e, (f, (g, h))))));")
tree3 <- ape::read.tree(text = "(a, (b, (c, (d, (e, ((f, g), h))))));")
trees <- c(tree1, tree2, tree3)
KendallColijn(trees)
KendallColijn(trees, Vector = SplitVector)
KCDiameter(4)
KCDiameter(trees)
}
\references{
\insertAllCited{}
}
\seealso{
\href{https://CRAN.R-project.org/package=treespace/vignettes/introduction.html}{\code{treespace::treeDist}}
is a more sophisticated, if more cumbersome, implementation that supports
lambda > 0, i.e. use of edge lengths in tree comparison.

Other tree distances:
\code{\link[=HierarchicalMutualInfo]{HierarchicalMutualInfo()}},
\code{\link[=JaccardRobinsonFoulds]{JaccardRobinsonFoulds()}},
\code{\link[=MASTSize]{MASTSize()}},
\code{\link[=MatchingSplitDistance]{MatchingSplitDistance()}},
\code{\link[=NNIDist]{NNIDist()}},
\code{\link[=NyeSimilarity]{NyeSimilarity()}},
\code{\link[=PathDist]{PathDist()}},
\code{\link{Robinson-Foulds}},
\code{\link[=SPRDist]{SPRDist()}},
\code{\link[=TransferDist]{TransferDist()}},
\code{\link[=TreeDistance]{TreeDistance()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree distances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Information.R
\encoding{UTF-8}
\name{SplitEntropy}
\alias{SplitEntropy}
\title{Entropy of two splits}
\usage{
SplitEntropy(split1, split2 = split1)
}
\arguments{
\item{split1, split2}{Logical vectors listing leaves in a consistent order,
identifying each leaf as a member of the ingroup (\code{TRUE}) or outgroup
(\code{FALSE}) of the split in question.}
}
\value{
A numeric vector listing, in bits:
\itemize{
\item \code{H1} The entropy of split 1;
\item \code{H2} The entropy of split 2;
\item \code{H12} The joint entropy of both splits;
\item \code{I} The mutual information of the splits;
\item \code{Hd} The entropy distance (variation of information) of the splits.
}
}
\description{
Calculate the entropy, joint entropy, entropy distance and information
content of two splits, treating each split as a division of \emph{n} leaves into
two groups.
Further details are available in a
\href{https://ms609.github.io/TreeDist/articles/information.html}{vignette},
\insertCite{Mackay2003;textual}{TreeDist} and
\insertCite{Meila2007;textual}{TreeDist}.
}
\examples{
A <- TRUE
B <- FALSE
SplitEntropy(c(A, A, A, B, B, B), c(A, A, B, B, B, B))
}
\references{
\insertAllCited{}
}
\seealso{
Other information functions:
\code{\link[=SplitSharedInformation]{SplitSharedInformation()}},
\code{\link{TreeInfo}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{information functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{cpp_mutual_clustering_all_pairs}
\alias{cpp_mutual_clustering_all_pairs}
\title{Pairwise mutual clustering information — batch computation}
\usage{
cpp_mutual_clustering_all_pairs(splits_list, n_tip, n_threads = 1L)
}
\arguments{
\item{splits_list}{A list of split matrices (class \code{Splits} or \code{RawMatrix}),
one per tree, all covering the same tip set.  Typically the object
returned by \code{as.Splits(trees, tipLabels = labs, asSplits = FALSE)}.}

\item{n_tip}{Integer; number of tips shared by all trees.}
}
\value{
Numeric vector of length \code{n*(n-1)/2} containing pairwise MCI
scores in \code{combn(n, 2)} column-major order (i.e. the data payload of
an R \code{dist} object).
}
\description{
Internal function. Computes all pairwise MCI scores for a set of trees,
using OpenMP threads when available (falling back to single-threaded
execution otherwise). No interrupt checking is performed inside the
parallel region; the outer R call remains interruptible between batches.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TreeDist-package.R
\docType{package}
\encoding{UTF-8}
\name{TreeDist-package}
\alias{TreeDist}
\alias{TreeDist-package}
\title{TreeDist: Distances between Phylogenetic Trees}
\description{
'TreeDist' is an R package that implements a suite of metrics that quantify the
topological distance between pairs of unweighted phylogenetic trees.
It also includes a simple "Shiny" application to allow the visualization of
distance-based tree spaces, and functions to calculate the information content
of trees and splits.
}
\details{
"TreeDist" primarily employs metrics in the category of
"generalized Robinson–Foulds distances": they are based on comparing splits
(bipartitions) between trees, and thus reflect the relationship data within
trees, with no reference to branch lengths.
Detailed documentation and usage instructions are
\href{https://ms609.github.io/TreeDist/}{available online} or in the vignettes.
\subsection{Generalized RF distances}{

The \href{https://ms609.github.io/TreeDist/articles/Robinson-Foulds.html}{Robinson–Foulds distance}
simply tallies the number of non-trivial splits (sometimes inaccurately
termed clades, nodes or edges) that occur in both trees -- any splits that are
not perfectly identical contributes one point to the distance score of zero,
however similar or different they are.
By overlooking potential similarities between almost-identical splits,
this conservative approach has undesirable properties.

\href{https://ms609.github.io/TreeDist/articles/Generalized-RF.html}{"Generalized" RF metrics}
generate \emph{matchings} that pair each split in one tree with a similar split
in the other.
Each pair of splits is assigned a similarity score; the sum of these scores in
the optimal matching then quantifies the similarity between two trees.

Different ways of calculating the the similarity between a pair of splits
lead to different tree distance metrics, implemented in the functions below:
\itemize{
\item \href{https://ms609.github.io/TreeDist/reference/TreeDistance.html}{\code{MutualClusteringInfo()}}, \href{https://ms609.github.io/TreeDist/reference/TreeDistance.html}{\code{SharedPhylogeneticInfo()}}
\itemize{
\item Smith (2020) scores matchings based on the amount of information
that one partition contains about the other.  The Mutual Phylogenetic
Information assigns zero similarity to split pairs that cannot
both exist on a single tree; The Mutual
Clustering Information metric is more forgiving, and exhibits more
desirable behaviour; it is the recommended metric for tree comparison.
(Its complement, \href{https://ms609.github.io/TreeDist/reference/TreeDistance.html}{\code{ClusteringInfoDistance()}}, returns a tree
distance.)
}
\item \href{https://ms609.github.io/TreeDist/reference/NyeSimilarity.html}{\code{NyeSimilarity()}}
\itemize{
\item Nye \emph{et al.} (2006) score matchings according to the size of the largest
split that is consistent with both of them, normalized against
the Jaccard index.  This approach is extended by Böcker \emph{et al}. (2013)
with the Jaccard–Robinson–Foulds metric (function
\href{https://ms609.github.io/TreeDist/reference/JaccardRobinsonFoulds.html}{\code{JaccardRobinsonFoulds()}}).
}
\item \href{https://ms609.github.io/TreeDist/reference/MatchingSplitDistance.html}{\code{MatchingSplitDistance()}}
\itemize{
\item Bogdanowicz and Giaro (2012) and  Lin \emph{et al.} (2012) independently proposed
counting the number of "mismatched" leaves in a pair of splits.
\href{https://ms609.github.io/TreeDist/reference/TreeDistance.html}{\code{MatchingSplitInfoDistance()}}
provides an information-based equivalent (Smith 2020).
}
}

The package also implements the variation of the path distance
proposed by Kendal and Colijn (2016) (function
\href{https://ms609.github.io/TreeDist/reference/KendallColijn.html}{\code{KendallColijn()}}),
approximations of the Nearest-Neighbour Interchange (NNI) distance (function
\href{https://ms609.github.io/TreeDist/reference/NNIDist.html}{\code{NNIDist()}};
following Li \emph{et al.} (1996)), and calculates the size (function
\href{https://ms609.github.io/TreeDist/reference/MASTSize.html}{\code{MASTSize()}}) and
information content (function
\href{https://ms609.github.io/TreeDist/reference/MASTSize.html}{\code{MASTInfo()}}) of the
Maximum Agreement Subtree.

For an implementation of the Tree Bisection and Reconnection (TBR) distance, see
the package '\href{https://ms609.github.io/TBRDist/index.html}{TBRDist}'.
}
}
\section{Tree space analysis}{
Map tree spaces and readily visualize mapped landscapes, avoiding
common analytical pitfalls (Smith, forthcoming),
using the inbuilt graphical user interface:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{TreeDist::MapTrees()
}\if{html}{\out{</div>}}

Serious analysts should consult the
\href{https://ms609.github.io/TreeDist/articles/treespace.html}{vignette}
for a command-line interface.
}

\references{
\itemize{
\item \insertRef{Bocker2013}{TreeDist}
\item \insertRef{Bogdanowicz2012}{TreeDist}
\item \insertRef{Kendall2016}{TreeDist}
\item \insertRef{Li1996}{TreeDist}
\item \insertRef{Lin2012}{TreeDist}
\item \insertRef{Nye2006}{TreeDist}
\item \insertRef{SmithDist}{TreeDist}
\item \insertRef{SmithSpace}{TreeDist}
}
}
\seealso{
Further documentation is available in the
\href{https://ms609.github.io/TreeDist/articles/}{package vignettes}, visible from
R using \code{vignette(package = "TreeDist")}.

Other R packages implementing tree distance functions include:
\itemize{
\item \href{https://emmanuelparadis.github.io/}{ape}:
\itemize{
\item \code{cophenetic.phylo()}: Cophenetic distance
\item \code{dist.topo()}: Path (topological) distance, Robinson–Foulds distance.
}
\item \href{https://cran.r-project.org/package=phangorn}{phangorn}
\itemize{
\item \code{treedist()}: Path, Robinson–Foulds and approximate SPR distances.
}
\item \href{https://ms609.github.io/Quartet/}{Quartet}: Triplet and Quartet distances,
using the tqDist algorithm.
\item \href{https://ms609.github.io/TBRDist/}{TBRDist}: TBR and SPR distances on
unrooted trees, using the 'uspr' C library.
\item \href{https://cran.r-project.org/package=distory}{distory} (unmaintained):
Geodesic distance
}
}
\author{
\strong{Maintainer}: Martin R. Smith \email{martin.smith@durham.ac.uk} (\href{https://orcid.org/0000-0001-5660-1727}{ORCID}) [copyright holder, programmer]

Authors:
\itemize{
  \item Martin R. Smith \email{martin.smith@durham.ac.uk} (\href{https://orcid.org/0000-0001-5660-1727}{ORCID}) [copyright holder, programmer]
}

Other contributors:
\itemize{
  \item Roy Jonker \email{roy_jonker@magiclogic.com} (LAP algorithm) [programmer, copyright holder]
  \item Yong Yang \email{yongyanglink@gmail.com} (LAP algorithm) [contributor, copyright holder]
  \item Yi Cao (LAP algorithm) [contributor, copyright holder]
  \item Neil Kaye (Mercator image) [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{cpp_transfer_dist_all_pairs}
\alias{cpp_transfer_dist_all_pairs}
\title{All-pairs transfer dissimilarity (OpenMP)}
\usage{
cpp_transfer_dist_all_pairs(splits_list, n_tip, scale, n_threads)
}
\arguments{
\item{splits_list}{List of raw matrices (one per tree).}

\item{n_tip}{Number of tips.}

\item{scale}{Logical: use scaled transfer dissimilarity?}

\item{n_threads}{Number of OpenMP threads.}
}
\value{
Numeric vector of length choose(N,2) in dist order.
}
\description{
All-pairs transfer dissimilarity (OpenMP)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Information.R
\encoding{UTF-8}
\name{AllSplitPairings}
\alias{AllSplitPairings}
\title{Variation of information for all split pairings}
\usage{
AllSplitPairings(n)
}
\arguments{
\item{n}{Integer specifying the number of leaves in a tree.}
}
\value{
\code{AllSplitPairings()} returns a named vector. The name of each
element corresponds to a certain variation of information, in bits; the
value of each element specifies the number of pairings of non-trivial
splits that give rise to that variation of information.
Split \code{AB|CD}  is treated as distinct from \code{CD|AB}.  If pairing
\code{AB|CD}=\code{CD|AB} is considered equivalent to \code{CD|AB}=\code{CD|AB} (etc), then
values should be divided by four.
}
\description{
Calculate the variation of clustering information
\insertCite{Meila2007}{TreeDist} for each possible pairing of
non-trivial splits on \emph{n} leaves \insertCite{SmithDist}{TreeDist},
tabulating the number of pairings with each similarity.
}
\examples{
AllSplitPairings(6)
# Treat equivalent splits as identical by dividing by four:
AllSplitPairings(6) / 4L
}
\references{
\insertAllCited{}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transfer_consensus.R
\name{TransferConsensus}
\alias{TransferConsensus}
\title{Consensus tree minimizing transfer distance}
\usage{
TransferConsensus(
  trees,
  scale = TRUE,
  greedy = c("best", "first"),
  init = c("empty", "majority")
)
}
\arguments{
\item{trees}{An object of class \code{multiPhylo}: the input trees.
All trees must share the same tip labels.}

\item{scale}{Logical; if \code{TRUE} (default), use the scaled transfer
distance (normalized by light-side size minus one).  If \code{FALSE}, use
the unscaled (raw Hamming) transfer distance.}

\item{greedy}{Character string specifying the greedy strategy:
\code{"best"} (default) picks the single highest-benefit action at each step;
\code{"first"} picks the first improving action encountered (faster but
potentially lower quality).}

\item{init}{Character string specifying the initial consensus:
\code{"empty"} (default) starts with no splits (purely additive);
\code{"majority"} starts with the majority-rule consensus and refines.}
}
\value{
A tree of class \code{phylo}.
}
\description{
Construct a consensus tree that minimizes the sum of transfer distances
to a set of input trees, using a greedy add-and-prune heuristic.
This function is moving to \code{ConsTree::Transfer()} and will soon be removed.
A copy is retained here temporarily.
}
\details{
Unlike the majority-rule consensus, which minimizes Robinson-Foulds
distance and can be highly unresolved when phylogenetic signal is low,
\code{TransferConsensus()} uses the finer-grained transfer distance
\insertCite{Lemoine2018}{TreeDist} to construct a more resolved consensus
tree.

The algorithm pools all splits observed across input trees, computes
pairwise transfer distances between them, and greedily adds or removes
splits to minimize total transfer dissimilarity cost.  The approach
follows \insertCite{Takazawa2026;textual}{TreeDist}, reimplemented for
'TreeDist' infrastructure.
}
\references{
\insertAllCited{}
}
\seealso{
Other summary trees:
\code{\link[=MCITree]{MCITree()}}
}
\concept{summary trees}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_rf.R
\encoding{UTF-8}
\name{Robinson-Foulds}
\alias{Robinson-Foulds}
\alias{InfoRobinsonFoulds}
\alias{RobinsonFouldsInfo}
\alias{InfoRobinsonFouldsSplits}
\alias{RobinsonFoulds}
\alias{RobinsonFouldsMatching}
\alias{RobinsonFouldsSplits}
\title{Robinson–Foulds distances, with adjustments for phylogenetic information
content}
\usage{
InfoRobinsonFoulds(
  tree1,
  tree2 = NULL,
  similarity = FALSE,
  normalize = FALSE,
  reportMatching = FALSE
)

InfoRobinsonFouldsSplits(
  splits1,
  splits2,
  nTip = attr(splits1, "nTip"),
  reportMatching = FALSE
)

RobinsonFoulds(
  tree1,
  tree2 = NULL,
  similarity = FALSE,
  normalize = FALSE,
  reportMatching = FALSE
)

RobinsonFouldsMatching(
  tree1,
  tree2,
  similarity = FALSE,
  normalize = FALSE,
  ...
)

RobinsonFouldsSplits(
  splits1,
  splits2,
  nTip = attr(splits1, "nTip"),
  reportMatching = FALSE
)
}
\arguments{
\item{tree1, tree2}{Trees of class \code{phylo}, with leaves labelled identically,
or lists of such trees to undergo pairwise comparison.  Where implemented,
\code{tree2 = NULL} will compute distances between each pair of trees in the list
\code{tree1} using a fast algorithm based on
\insertCite{Day1985;textual}{TreeDist}.}

\item{similarity}{Logical specifying whether to report the result as a tree
similarity, rather than a difference.}

\item{normalize}{If a numeric value is provided, this will be used as a
maximum value against which to rescale results.
If \code{TRUE}, results will be rescaled against a maximum value calculated from
the specified tree sizes and topology, as specified in the "Normalization"
section below.
If \code{FALSE}, results will not be rescaled.}

\item{reportMatching}{Logical specifying whether to return the clade
matchings as an attribute of the score.}

\item{splits1, splits2}{Logical matrices where each row corresponds to a leaf,
either listed in the same order or bearing identical names (in any sequence),
and each column corresponds to a split, such that each leaf is identified as
a member of the ingroup (\code{TRUE}) or outgroup (\code{FALSE}) of the respective
split.}

\item{nTip}{(Optional) Integer specifying the number of leaves in each split.}

\item{\dots}{Not used.}
}
\value{
\code{RobinsonFoulds()} and \code{InfoRobinsonFoulds()} return an array of numerics providing the
distances between each  pair of trees in \code{tree1} and \code{tree2},
or \code{splits1} and \code{splits2}.

If \code{reportMatching = TRUE}, the \code{pairScores} attribute
returns a logical matrix specifying whether each pair of splits is identical.
}
\description{
\code{RobinsonFoulds()} calculates the Robinson–Foulds distance
\insertCite{Robinson1981}{TreeDist}, or the corresponding similarity measure.
\code{InfoRobinsonFoulds()} weights splits according to their phylogenetic
information content \insertCite{@§2.1 in @SmithDist}{TreeDist}.
Optionally, the matching between identical splits may reported.
Generalized Robinson–Foulds distances (see \code{\link[=TreeDistance]{TreeDistance()}})
are better suited to most use cases
\insertCite{SmithDist,SmithSpace}{TreeDist}.
}
\details{
\code{RobinsonFoulds()} calculates the standard Robinson–Foulds distance,
i.e. the number of splits that occur in one tree but not the other.
\code{InfoRobinsonFoulds()} calculates the tree similarity or distance by summing
the phylogenetic information content of all splits that are (or are not)
identical in both trees.  Consequently, splits that are more likely
to be identical by chance alone make a smaller contribution to overall
tree distance, because their similarity is less remarkable.

Rapid comparison between multiple pairs of trees employs the
\insertCite{Day1985;textual}{TreeDist} linear-time algorithm.
}
\section{Functions}{
\itemize{
\item \code{RobinsonFouldsMatching()}: Matched splits, intended for use with
\code{\link[=VisualizeMatching]{VisualizeMatching()}}.

}}
\section{Normalization}{

\itemize{
\item \code{RobinsonFoulds()} is normalized against the total number of splits that
are present.
\item \code{InfoRobinsonFoulds()} is normalized against the sum of the phylogenetic
information of all splits in each tree, treated independently.
}
}

\examples{
 # For BalancedTree, PectinateTree, as.phylo:
library("TreeTools", quietly = TRUE)
balanced7 <- BalancedTree(7)
pectinate7 <- PectinateTree(7)
RobinsonFoulds(balanced7, pectinate7)
RobinsonFoulds(balanced7, pectinate7, normalize = TRUE)
VisualizeMatching(RobinsonFouldsMatching, balanced7, pectinate7)

InfoRobinsonFoulds(balanced7, pectinate7)
VisualizeMatching(InfoRobinsonFoulds, balanced7, pectinate7)
}
\references{
\insertAllCited{}
}
\seealso{
Display paired splits: \code{\link[=VisualizeMatching]{VisualizeMatching()}}

Other tree distances:
\code{\link[=HierarchicalMutualInfo]{HierarchicalMutualInfo()}},
\code{\link[=JaccardRobinsonFoulds]{JaccardRobinsonFoulds()}},
\code{\link[=KendallColijn]{KendallColijn()}},
\code{\link[=MASTSize]{MASTSize()}},
\code{\link[=MatchingSplitDistance]{MatchingSplitDistance()}},
\code{\link[=NNIDist]{NNIDist()}},
\code{\link[=NyeSimilarity]{NyeSimilarity()}},
\code{\link[=PathDist]{PathDist()}},
\code{\link[=SPRDist]{SPRDist()}},
\code{\link[=TransferDist]{TransferDist()}},
\code{\link[=TreeDistance]{TreeDistance()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree distances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_utilities.R
\name{CompareAll}
\alias{CompareAll}
\title{Distances between each pair of trees}
\usage{
CompareAll(x, Func, FUN.VALUE = Func(x[[1]], x[[1]], ...), ...)
}
\arguments{
\item{x}{List of trees, in the format expected by \code{Func()}.}

\item{Func}{distance function returning distance between two trees,
e.g. \code{\link[phangorn:treedist]{path.dist()}}.}

\item{FUN.VALUE}{Format of output of \code{Func()}, to be passed to \code{\link[=vapply]{vapply()}}.
If unspecified, calculated by running \code{Func(x[[1]], x[[1]])}.}

\item{\dots}{Additional parameters to pass to \code{Func()}.}
}
\value{
\code{CompareAll()} returns a distance matrix of class \code{dist} detailing
the distance between each pair of trees.
Identical trees are assumed to have zero distance.
}
\description{
Calculate the distance between each tree in a list, and each other tree
in the same list.
}
\details{
\code{CompareAll()} is not limited to tree comparisons:
\code{Func} can be any symmetric function.
}
\examples{
# Generate a list of trees to compare
library("TreeTools", quietly = TRUE)
trees <- list(bal1 = BalancedTree(1:8), 
              pec1 = PectinateTree(1:8),
              pec2 = PectinateTree(c(4:1, 5:8)))

# Compare each tree with each other tree
CompareAll(trees, NNIDist)
  
# Providing FUN.VALUE yields a modest speed gain:
dist <- CompareAll(trees, NNIDist, FUN.VALUE = integer(7))
  
# View distances as a matrix
as.matrix(dist$lower)
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{pairwise tree distances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{cpp_transfer_dist}
\alias{cpp_transfer_dist}
\title{Per-pair transfer dissimilarity}
\usage{
cpp_transfer_dist(x, y, nTip)
}
\arguments{
\item{x, y}{Raw matrices representing splits (from as.Splits()).}

\item{nTip}{Integer: number of tips.}
}
\value{
A list with components:
\itemize{
\item score_scaled: scaled transfer dissimilarity (double)
\item score_unscaled: unscaled transfer dissimilarity (double)
\item \code{matching_xy}: integer vector, best match in y for each split in x (1-based, NA if sentinel)
\item \code{matching_yx}: integer vector, best match in x for each split in y (1-based, NA if sentinel)
}
}
\description{
Per-pair transfer dissimilarity
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Reduce.R
\name{ReduceTrees}
\alias{ReduceTrees}
\title{Collapse areas of agreement between two trees}
\usage{
ReduceTrees(tree1, tree2, check = TRUE)
}
\arguments{
\item{tree1, tree2}{Single trees of class \code{phylo} to undergo comparison.}

\item{check}{Logical specifying whether to validate input. Specify
\code{FALSE} and you will encounter undefined behaviour if trees are not
binary \code{phylo} objects with identical leaf labels, rooted on leaf 1.}
}
\value{
\code{ReduceTrees()} returns a list of two trees, corresponding to
\code{tree1} and \code{tree2} after any identical groupings have been collapsed,
with tree edges listed in postorder; or \code{NULL} if the trees are equivalent.
}
\description{
\code{ReduceTrees()} reduces trees according to the tree reduction rules of
\insertCite{Allen2001;textual}{TreeDist}:
\itemize{
\item Collapse identical pendant subtrees;
\item Compress equivalent internal chains.
}
}
\examples{
tree1 <- TreeTools::BalancedTree(9)
tree2 <- TreeTools::PectinateTree(9)

# Set graphical parameters
oPar <- par(mai = rep(0.1, 4), mfrow = c(2, 2))

plot(tree1)
plot(tree2)

# Reduce trees by collapsing identical clades
confl <- ReduceTrees(tree1, tree2)

plot(confl[[1]])
plot(confl[[2]])

# Restore graphical parameters
par(oPar)
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_utilities.R
\name{ReportMatching}
\alias{ReportMatching}
\title{List clades as text}
\usage{
ReportMatching(splits1, splits2, realMatch = TRUE)
}
\arguments{
\item{splits1, splits2}{Logical matrices with columns specifying membership
of each corresponding matched clade.}
}
\value{
\code{ReportMatching} returns a character vector describing each pairing
in a matching.
}
\description{
List clades as text
}
\seealso{
\code{\link{VisualizeMatching}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_nye.R
\encoding{UTF-8}
\name{NyeSimilarity}
\alias{NyeSimilarity}
\alias{NyeSplitSimilarity}
\title{Nye \emph{et al}. (2006) tree comparison}
\usage{
NyeSimilarity(
  tree1,
  tree2 = NULL,
  similarity = TRUE,
  normalize = FALSE,
  normalizeMax = !is.logical(normalize),
  reportMatching = FALSE,
  diag = TRUE
)

NyeSplitSimilarity(
  splits1,
  splits2,
  nTip = attr(splits1, "nTip"),
  reportMatching = FALSE
)
}
\arguments{
\item{tree1, tree2}{Trees of class \code{phylo}, with leaves labelled identically,
or lists of such trees to undergo pairwise comparison.  Where implemented,
\code{tree2 = NULL} will compute distances between each pair of trees in the list
\code{tree1} using a fast algorithm based on
\insertCite{Day1985;textual}{TreeDist}.}

\item{similarity}{Logical specifying whether to report the result as a tree
similarity, rather than a difference.}

\item{normalize}{If a numeric value is provided, this will be used as a
maximum value against which to rescale results.
If \code{TRUE}, results will be rescaled against a maximum value calculated from
the specified tree sizes and topology, as specified in the "Normalization"
section below.
If \code{FALSE}, results will not be rescaled.}

\item{normalizeMax}{When calculating similarity, normalize against the
maximum number of splits that could have been present (\code{TRUE}),
or the number of splits that were actually observed (\code{FALSE})?
Defaults to the number of splits in the better-resolved tree; set
\code{normalize = pmin.int} to use the number of splits in the less resolved
tree.}

\item{reportMatching}{Logical specifying whether to return the clade
matchings as an attribute of the score.}

\item{diag}{Logical specifying whether to return similarities along the
diagonal, i.e. of each tree with itself.  Applies only if \code{tree2} is
a list identical to \code{tree1}, or \code{NULL}.}

\item{splits1, splits2}{Logical matrices where each row corresponds to a leaf,
either listed in the same order or bearing identical names (in any sequence),
and each column corresponds to a split, such that each leaf is identified as
a member of the ingroup (\code{TRUE}) or outgroup (\code{FALSE}) of the respective
split.}

\item{nTip}{(Optional) Integer specifying the number of leaves in each split.}
}
\value{
\code{NyeSimilarity()} returns an array of numerics providing the
distances between each  pair of trees in \code{tree1} and \code{tree2},
or \code{splits1} and \code{splits2}.
}
\description{
\code{NyeSimilarity()} and \code{NyeSplitSimilarity()} implement the
\href{https://ms609.github.io/TreeDist/articles/Robinson-Foulds.html#generalized-robinson-foulds-distances}{Generalized Robinson–Foulds}
tree comparison metric of \insertCite{Nye2006;textual}{TreeDist}.
In short, this finds the optimal matching that pairs each branch from
one tree with a branch in the second, where matchings are scored according to
the size of the largest split that is consistent with both of them,
normalized against the Jaccard index.
A more detailed account is available in the
\href{https://ms609.github.io/TreeDist/articles/Generalized-RF.html#nye-et-al--tree-similarity-metric}{vignettes}.
}
\details{
The measure is defined as a similarity score.  If \code{similarity = FALSE}, the
similarity score will be converted into a distance by doubling it and
subtracting it from the number of splits present in both trees.
This ensures consistency with \code{JaccardRobinsonFoulds}.

Note that \code{NyeSimilarity(tree1, tree2)} is equivalent to, but
slightly faster than, \code{\link{JaccardRobinsonFoulds}
(tree1, tree2, k = 1, allowConflict = TRUE)}.
}
\section{Normalization}{


If \code{normalize = TRUE} and \code{similarity = TRUE}, then results will be rescaled
from zero to one by dividing by the mean number of splits in the two trees
being compared.

You may wish to normalize instead against the number of splits present
in the smaller tree, which represents the maximum value possible for a pair
of trees with the specified topologies (\code{normalize = pmin.int}); the
number of splits in the most resolved tree (\code{normalize = pmax.int});
or the maximum value possible for any pair of trees with  \emph{n} leaves,
\emph{n} - 3 (\code{normalize = TreeTools::NTip(tree1) - 3L}).

If \code{normalize = TRUE} and \code{similarity = FALSE}, then results will be rescaled
from zero to one by dividing by the total number of splits in the pair
of trees being considered.

Trees need not contain identical leaves; scores are based on the leaves that
trees hold in common.  Check for unexpected differences in tip labelling
with \code{setdiff(TipLabels(tree1), TipLabels(tree2))}.
}

\examples{
library("TreeTools")
NyeSimilarity(BalancedTree(8), PectinateTree(8))
VisualizeMatching(NyeSimilarity, BalancedTree(8), PectinateTree(8))
NyeSimilarity(as.phylo(0:5, nTip = 8), PectinateTree(8))
NyeSimilarity(as.phylo(0:5, nTip = 8), similarity = FALSE)
}
\references{
\insertAllCited{}
}
\seealso{
Other tree distances:
\code{\link[=HierarchicalMutualInfo]{HierarchicalMutualInfo()}},
\code{\link[=JaccardRobinsonFoulds]{JaccardRobinsonFoulds()}},
\code{\link[=KendallColijn]{KendallColijn()}},
\code{\link[=MASTSize]{MASTSize()}},
\code{\link[=MatchingSplitDistance]{MatchingSplitDistance()}},
\code{\link[=NNIDist]{NNIDist()}},
\code{\link[=PathDist]{PathDist()}},
\code{\link{Robinson-Foulds}},
\code{\link[=SPRDist]{SPRDist()}},
\code{\link[=TransferDist]{TransferDist()}},
\code{\link[=TreeDistance]{TreeDistance()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree distances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{entropy_int}
\alias{entropy_int}
\title{Calculate entropy of integer vector of counts}
\usage{
entropy_int(n)
}
\arguments{
\item{n}{a vector of integer counts}
}
\value{
\code{entropy_int()} returns a numeric corresponding to the entropy of
each observation, in bits.
}
\description{
Wrapper for C++ function; no input checking is performed.
\code{\link[=Ntropy]{Ntropy()}} is better suited for use where performance is not critical.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/median.R
\encoding{UTF-8}
\name{median.multiPhylo}
\alias{median.multiPhylo}
\title{Median of a set of trees}
\usage{
\method{median}{multiPhylo}(
  x,
  na.rm = FALSE,
  Distance = ClusteringInfoDistance,
  index = FALSE,
  breakTies = TRUE,
  ...
)
}
\arguments{
\item{x}{Object of class \code{multiPhylo} containing phylogenetic trees.}

\item{na.rm, \dots}{Unused; included for consistency with default function..}

\item{Distance}{Function to calculate distances between each pair
of trees in \code{x}.}

\item{index}{Logical: if \code{TRUE}, return the index of the median tree(s);
if \code{FALSE}, return the tree itself.}

\item{breakTies}{Logical: if \code{TRUE}, return a single tree with the minimum
score; if \code{FALSE}, return all tied trees.}
}
\value{
\code{median()} returns an object of class \code{phylo}
corresponding to the geometric median of a set of trees:
that is, the tree whose average distance from all other trees in the set
is lowest.
If multiple trees tie in their average distance, the first will be returned,
unless \code{breakTies = FALSE}, in which case an object of class \code{multiPhylo}
containing all such trees will be returned.
}
\description{
Calculate the single binary tree that represents the geometric median --
an "average" -- of a forest of tree topologies.
}
\details{
The geometric median is the tree that exhibits the shortest average distance
from each other tree topology in the set.
It represents an "average" of a set of trees, though note that an unsampled
tree may be closer to the geometric "centre of gravity" of the input set --
such a tree would not be considered.

The result will depend on the metric chosen to calculate distances between
tree topologies. In the absence of a natural metric of tree topologies,
the default choice is \code{\link[=ClusteringInfoDistance]{ClusteringInfoDistance()}} -- which discards
branch length information.
If specifying a different function, be sure that it returns a difference,
rather than a similarity.
}
\examples{
library("TreeTools", quietly = TRUE)
tenTrees <- as.phylo(1:10, nTip = 8)

# Default settings:
median(tenTrees)

# Robinson-Foulds distances include ties:
median(tenTrees, Distance = RobinsonFoulds, breakTies = FALSE)

# Be sure to use a distance function, rather than a similarity:
NyeDistance <- function(...) NyeSimilarity(..., similarity = FALSE)
median(tenTrees, Distance = NyeDistance)

# To analyse a list of trees that is not of class multiPhylo:
treeList <- lapply(1:10, as.phylo, nTip = 8)
class(treeList)
median(structure(treeList, class = "multiPhylo"))
}
\seealso{
Consensus methods:
\code{\link[ape:consensus]{ape::consensus()}},
\code{\link[TreeTools:ConsensusWithout]{TreeTools::ConsensusWithout()}}

Other tree space functions:
\code{\link[=Islands]{Islands()}},
\code{\link[=MSTSegments]{MSTSegments()}},
\code{\link[=MapTrees]{MapTrees()}},
\code{\link[=MappingQuality]{MappingQuality()}},
\code{\link[=SpectralEigens]{SpectralEigens()}},
\code{\link{cluster-statistics}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree space functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HPart.R
\name{HPart}
\alias{HPart}
\alias{as.HPart}
\alias{as.HPart.HPart}
\alias{as.HPart.default}
\alias{as.HPart.list}
\alias{as.HPart.phylo}
\alias{is.HPart}
\alias{print.HPart}
\alias{as.phylo.HPart}
\alias{plot.HPart}
\title{Hierarchical partition structure}
\usage{
as.HPart(tree, tipLabels)

\method{as.HPart}{HPart}(tree, tipLabels = NULL)

\method{as.HPart}{default}(tree, tipLabels = NULL)

\method{as.HPart}{list}(tree, tipLabels = NULL)

\method{as.HPart}{phylo}(tree, tipLabels = TipLabels(tree))

is.HPart(x)

\method{print}{HPart}(x, ...)

\method{as.phylo}{HPart}(x, ...)

\method{plot}{HPart}(x, ...)
}
\arguments{
\item{tree}{An object to convert to an HPart structure, in a supported format
(see details).}

\item{tipLabels}{Character vector specifying sequence in which to order
tip labels.}

\item{x}{\code{HPart} object to plot.}

\item{\dots}{Additional arguments to \code{\link[ape:plot.phylo]{plot.phylo}}.}
}
\value{
\code{HPart()} returns a structure containing a pointer to a C++
representation of a hierarchical partition structure.
}
\description{
A structure of class \code{HPart} comprises a pointer to a C++ representation of
hierarchical partitions, with the attribute \code{tip.label} recording the
character labels of its leaves.  \code{HPart} objects with identical tip labels
can be compared using \code{\link[=HierarchicalMutualInfo]{HierarchicalMutualInfo()}}.
}
\details{
An \code{HPart} object may be created from various representations of hierarchical
structures:
\itemize{
\item a tree (possibly phylogenetic) of class \code{phylo}
\item A hierarchical list of lists, in which elements are represented by integers
1 to n
\item A vector, which will be interpreted as a flat structure
in which all elements bearing the same label are assigned to the same cluster
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_utilities.R
\name{Entropy}
\alias{Entropy}
\alias{Ntropy}
\title{Entropy in bits}
\usage{
Entropy(...)

Ntropy(...)
}
\arguments{
\item{\dots}{Series of numerics, or single numeric vector, specifying
probabilities of outcomes (for \code{Entropy()}) or counts (for \code{Ntropy()}).}
}
\value{
\code{Entropy()} and \code{Ntropy()} return the entropy of the specified
probabilities or counts, in bits.
}
\description{
Calculate the entropy of a vector of probabilities, in bits.
Probabilities should sum to one.
Probabilities equalling zero will be ignored.
}
\examples{
Entropy(1/2, 0, 1/2) # = 1
Entropy(rep(1/4, 4)) # = 2
Ntropy(c(2, 2, 0, 2, 2)) # = 2
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_distance_mast.R
\name{MASTSize}
\alias{MASTSize}
\alias{MASTInfo}
\title{Maximum Agreement Subtree size}
\usage{
MASTSize(tree1, tree2 = tree1, rooted = TRUE)

MASTInfo(tree1, tree2 = tree1, rooted = TRUE)
}
\arguments{
\item{tree1, tree2}{Trees of class \code{phylo}, or lists of such trees to undergo
pairwise comparison.}

\item{rooted}{Logical specifying whether to treat the trees as rooted.}
}
\value{
\code{MASTSize()} returns an integer specifying the number of leaves in
the maximum agreement subtree.

\code{MASTInfo()} returns a vector or matrix listing the phylogenetic
information content, in bits, of the maximum agreement subtree.
}
\description{
Calculate the size or phylogenetic information content
\insertCite{Steel2006}{TreeDist}
of the maximum agreement subtree between two phylogenetic trees, i.e.
the largest tree that can be obtained from both \code{tree1} and \code{tree2} by
deleting, but not rearranging, leaves, using the algorithm of
\insertCite{Valiente2009;textual}{TreeDist}.
}
\details{
Implemented for trees with up to 4096 tips.  Contact the maintainer if you
need to process larger trees.
}
\examples{
 # for as.phylo, BalancedTree, PectinateTree:
library("TreeTools", quietly = TRUE)

MASTSize(PectinateTree(8), BalancedTree(8))
MASTInfo(PectinateTree(8), BalancedTree(8))

MASTSize(BalancedTree(7), as.phylo(0:3, 7))
MASTSize(as.phylo(0:3, 7), PectinateTree(7))

MASTInfo(BalancedTree(7), as.phylo(0:3, 7))
MASTInfo(as.phylo(0:3, 7), PectinateTree(7))

MASTSize(list(Bal = BalancedTree(7), Pec = PectinateTree(7)),
         as.phylo(0:3, 7))
MASTInfo(list(Bal = BalancedTree(7), Pec = PectinateTree(7)),
         as.phylo(0:3, 7))

CompareAll(as.phylo(0:4, 8), MASTSize)
CompareAll(as.phylo(0:4, 8), MASTInfo)
}
\references{
\insertAllCited{}
}
\seealso{
\code{\link[phangorn:mast]{phangorn::mast()}}, a slower implementation that also lists the
leaves contained within the subtree.

Other tree distances:
\code{\link[=HierarchicalMutualInfo]{HierarchicalMutualInfo()}},
\code{\link[=JaccardRobinsonFoulds]{JaccardRobinsonFoulds()}},
\code{\link[=KendallColijn]{KendallColijn()}},
\code{\link[=MatchingSplitDistance]{MatchingSplitDistance()}},
\code{\link[=NNIDist]{NNIDist()}},
\code{\link[=NyeSimilarity]{NyeSimilarity()}},
\code{\link[=PathDist]{PathDist()}},
\code{\link{Robinson-Foulds}},
\code{\link[=SPRDist]{SPRDist()}},
\code{\link[=TransferDist]{TransferDist()}},
\code{\link[=TreeDistance]{TreeDistance()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree distances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trustworthiness.R
\name{MappingQuality}
\alias{MappingQuality}
\alias{ProjectionQuality}
\title{Faithfulness of mapped distances}
\usage{
MappingQuality(original, mapped, neighbours = 10L)

ProjectionQuality(original, mapped, neighbours = 10L)
}
\arguments{
\item{original, mapped}{Square matrix or \code{dist} object containing
original / mapped pairwise distances.}

\item{neighbours}{Integer specifying number of nearest neighbours to use in
calculation.  This should typically be small relative to the number of
points.}
}
\value{
\code{MappingQuality()} returns a named vector of length four,
containing the entries: \code{Trustworthiness}, \code{Continuity}, \code{TxC}
(the product of these values), and \code{sqrtTxC} (its square root).
}
\description{
\code{MappingQuality()} calculates the trustworthiness and continuity
of mapped distances \insertCite{Venna2001,Kaski2003}{TreeDist}.
Trustworthiness measures, on a scale from 0--1,
the degree to which points that are nearby in a mapping are truly close
neighbours; continuity, the extent to which points that are truly nearby
retain their close spatial proximity in a mapping.
}
\examples{
library("TreeTools", quietly = TRUE)
trees <- as.phylo(0:10, nTip = 8)
distances <- ClusteringInfoDistance(trees)
mapping <- cmdscale(distances)
MappingQuality(distances, dist(mapping), 4)
}
\references{
\insertAllCited{}
}
\seealso{
Other tree space functions:
\code{\link[=Islands]{Islands()}},
\code{\link[=MSTSegments]{MSTSegments()}},
\code{\link[=MapTrees]{MapTrees()}},
\code{\link[=SpectralEigens]{SpectralEigens()}},
\code{\link{cluster-statistics}},
\code{\link[=median.multiPhylo]{median.multiPhylo()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree space functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_information.R
\encoding{UTF-8}
\name{TreeInfo}
\alias{TreeInfo}
\alias{SplitwiseInfo}
\alias{ClusteringEntropy}
\alias{ClusteringInfo}
\alias{ClusteringEntropy.phylo}
\alias{ClusteringEntropy.list}
\alias{ClusteringEntropy.multiPhylo}
\alias{ClusteringEntropy.Splits}
\alias{ClusteringInfo.phylo}
\alias{ClusteringInfo.list}
\alias{ClusteringInfo.multiPhylo}
\alias{ClusteringInfo.Splits}
\alias{ConsensusInfo}
\title{Information content of splits within a tree}
\usage{
SplitwiseInfo(x, p = NULL, sum = TRUE)

ClusteringEntropy(x, p = NULL, sum = TRUE)

ClusteringInfo(x, p = NULL, sum = TRUE)

\method{ClusteringEntropy}{phylo}(x, p = NULL, sum = TRUE)

\method{ClusteringEntropy}{list}(x, p = NULL, sum = TRUE)

\method{ClusteringEntropy}{multiPhylo}(x, p = NULL, sum = TRUE)

\method{ClusteringEntropy}{Splits}(x, p = NULL, sum = TRUE)

\method{ClusteringInfo}{phylo}(x, p = NULL, sum = TRUE)

\method{ClusteringInfo}{list}(x, p = NULL, sum = TRUE)

\method{ClusteringInfo}{multiPhylo}(x, p = NULL, sum = TRUE)

\method{ClusteringInfo}{Splits}(x, p = NULL, sum = TRUE)

ConsensusInfo(trees, info = "phylogenetic", p = 0.5, check.tips = TRUE)
}
\arguments{
\item{x}{A tree of class \code{phylo}, a list of trees, or a \code{multiPhylo} object.}

\item{p}{Scalar from 0.5 to 1 specifying minimum proportion of trees that
must contain a split for it to appear within the consensus.}

\item{sum}{Logical: if \code{TRUE}, sum the information content of each split to
provide the total splitwise information content of the tree.}

\item{trees}{List of \code{phylo} objects, optionally with class \code{multiPhylo}.}

\item{info}{Abbreviation of "phylogenetic" or "clustering", specifying
the concept of information to employ.}

\item{check.tips}{Logical specifying whether to renumber leaves such that
leaf numbering is consistent in all trees.}
}
\value{
\code{SplitwiseInfo()}, \code{ClusteringInfo()} and \code{ClusteringEntropy()}
return the splitwise information content of the tree -- or of each split
in turn, if \code{sum = FALSE} -- in bits.

\code{ConsensusInfo()} returns the splitwise information content of the
majority rule consensus of \code{trees}.
}
\description{
Sum the entropy (\code{ClusteringEntropy()}), clustering information content
(\code{ClusteringInfo()}), or phylogenetic information content (\code{SplitwiseInfo()})
across each split within a phylogenetic tree,
or the consensus of a set of phylogenetic trees (\code{ConsensusInfo()}).
This value will be greater than the total information
content of the tree where a tree contains multiple splits, as
these splits are not independent and thus contain mutual information that is
counted more than once
}
\section{Clustering information}{


Clustering entropy addresses the question "how much information is contained
in the splits within a tree". Its approach is complementary to the
phylogenetic information content, used in \code{\link[=SplitwiseInfo]{SplitwiseInfo()}}.
In essence, it asks, given a split that subdivides the leaves of a tree into
two partitions, how easy it is to predict which partition a randomly drawn
leaf belongs to \insertCite{@Meila2007; @Vinh2010}{TreeDist}.

Formally, the entropy of a split \emph{S} that divides \emph{n} leaves into two
partitions of sizes \emph{a} and \emph{b} is given by
\emph{H(S)} = - \emph{a/n} log \emph{a/n} - \emph{b/n} log \emph{b/n}.

Base 2 logarithms are conventionally used, such that entropy is measured in
bits.
Entropy denotes the number of bits that are necessary to encode the outcome
of a random variable: here, the random variable is "what partition does a
randomly selected leaf belong to".

An even split has an entropy of 1 bit: there is no better way of encoding
an outcome than using one bit to specify which of the two partitions the
randomly selected leaf belongs to.

An uneven split has a lower entropy: membership of the larger partition is
common, and thus less surprising; it can be signified using fewer bits in an
optimal compression system.

If this sounds confusing, let's consider creating a code to transmit the
cluster label of two randomly selected leaves.  One straightforward
option would be to use
\itemize{
\item \code{00} = "Both leaves belong to partition A"
\item \code{11} = "Both leaves belong to partition B"
\item \code{01} = 'First leaf in A, second in B`
\item \code{10} = 'First leaf in B, second in A`
}

This code uses two bits to transmit the partition labels of two leaves.
If partitions A and B are equiprobable, this is the optimal code; our
entropy -- the average information content required per leaf -- is 1 bit.

Alternatively, we could use the (suboptimal) code
\itemize{
\item \code{0} = "Both leaves belong to partition A"
\item \code{111} = "Both leaves belong to partition B"
\item \code{101} = 'First leaf in A, second in B`
\item \code{110} = 'First leaf in B, second in A`
}

If A is much larger than B, then most pairs of leaves will require just
a single bit (code \code{0}). The additional bits when 1+ leaf belongs to B
may be required sufficiently rarely that the average message
requires fewer than two bits for two leaves, so the entropy is less than
1 bit.  (The optimal coding strategy will depend on the exact sizes
of A and B.)

As entropy measures the bits required to transmit the cluster label of each
leaf \insertCite{@Vinh2010: p. 2840}{TreeDist}, the information content of
a split is its entropy multiplied by the number of leaves.
}

\section{Phylogenetic information}{


Phylogenetic information expresses the information content of a split
in terms of the probability that a uniformly selected tree will contain it
\insertCite{Thorley1998}{TreeDist}.
}

\section{Consensus information}{


The information content of splits in a consensus tree is calculated by
interpreting support values (i.e. the proportion of trees containing
each split in the consensus) as probabilities that the true tree contains
that split, following \insertCite{SmithCons;textual}{TreeDist}.
}

\examples{
library("TreeTools", quietly = TRUE)

SplitwiseInfo(PectinateTree(8))
tree <- read.tree(text = "(a, b, (c, (d, e, (f, g)0.8))0.9);")
SplitwiseInfo(tree)
SplitwiseInfo(tree, TRUE)

# Clustering entropy of an even split = 1 bit
ClusteringEntropy(TreeTools::as.Splits(c(rep(TRUE, 4), rep(FALSE, 4))))

# Clustering entropy of an uneven split < 1 bit
ClusteringEntropy(TreeTools::as.Splits(c(rep(TRUE, 2), rep(FALSE, 6))))

tree1 <- TreeTools::BalancedTree(8)
tree2 <- TreeTools::PectinateTree(8)

ClusteringInfo(tree1)
ClusteringEntropy(tree1)
ClusteringInfo(list(one = tree1, two = tree2))

ClusteringInfo(tree1) + ClusteringInfo(tree2)
ClusteringEntropy(tree1) + ClusteringEntropy(tree2)
ClusteringInfoDistance(tree1, tree2)
MutualClusteringInfo(tree1, tree2)

# Clustering entropy with uncertain splits
tree <- ape::read.tree(text = "(a, b, (c, (d, e, (f, g)0.8))0.9);")
ClusteringInfo(tree)
ClusteringInfo(tree, TRUE)

# Support-weighted information content of a consensus tree
set.seed(0)
trees <- list(RandomTree(8), RootTree(BalancedTree(8), 1), PectinateTree(8))
cons <- consensus(trees, p = 0.5)
p <- SplitFrequency(cons, trees) / length(trees)
plot(cons)
LabelSplits(cons, signif(SplitwiseInfo(cons, p, sum = FALSE), 4))
ConsensusInfo(trees)
LabelSplits(cons, signif(ClusteringInfo(cons, p, sum = FALSE), 4))
ConsensusInfo(trees, "clustering")
}
\references{
\insertAllCited{}
}
\seealso{
An introduction to the phylogenetic information content of a split is given
in \href{https://ms609.github.io/TreeTools/reference/SplitInformation.html}{\code{SplitInformation()}}
and in a \href{https://ms609.github.io/TreeDist/articles/information.html}{package vignette}.

Other information functions:
\code{\link[=SplitEntropy]{SplitEntropy()}},
\code{\link[=SplitSharedInformation]{SplitSharedInformation()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{information functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Islands.R
\name{Islands}
\alias{Islands}
\title{Find islands from distance matrix}
\usage{
Islands(D, threshold, dense = TRUE, smallest = 0)
}
\arguments{
\item{D}{Square matrix or \code{dist} object containing Euclidean distances
between data points.}

\item{threshold}{Elements greater than \code{threshold} distance units will not be
assigned to the same island.}

\item{dense}{Logical; if \code{FALSE}, each island will be named according to the
index of its lowest-indexed member; if \code{TRUE}, each island will be numbered
sequentially from \code{1}, ordered by the index of the lowest-indexed member.}

\item{smallest}{Integer; Islands comprising no more than \code{smallest} elements
will be assigned to island \code{NA}.}
}
\value{
\code{Islands()} returns a vector listing the island to which
each element is assigned.
}
\description{
\code{Islands()} assigns a set of objects to islands, such that all elements
within an island can form a connected graph in which each edge is no longer
than \code{threshold} distance units \insertRef{Silva2021}{TreeDist}.
}
\examples{
library("TreeTools", quietly = TRUE)
# Generate a set of trees
trees <- as.phylo(as.TreeNumber(BalancedTree(16)) + c(-(40:20), 70:105), 16)

# Calculate distances between trees
distances <- ClusteringInfoDist(trees)
summary(distances)

# Assign trees to islands
isle <- Islands(distances, quantile(distances, 0.1))
table(isle)

# Indicate island membership on 2D mapping of tree distances
mapping <- cmdscale(distances, 2)
plot(mapping, col = isle + 1,
     asp = 1, # Preserve aspect ratio - do not distort distances
     ann = FALSE, axes = FALSE, # Don't label axes: dimensions are meaningless)
     pch = 16 # Plotting character: Filled circle
)

# Compare strict consensus with island consensus trees
oPar <- par(mfrow = c(2, 2), mai = rep(0.1, 4))
plot(Consensus(trees), main = "Strict")
plot(Consensus(trees[isle == 1]), edge.col = 2, main = "Island 1")
plot(Consensus(trees[isle == 2]), edge.col = 3, main = "Island 2")
plot(Consensus(trees[isle == 3]), edge.col = 4, main = "Island 3")

# Restore graphical parameters
par(oPar)
}
\references{
\insertAllCited{}
}
\seealso{
Other tree space functions:
\code{\link[=MSTSegments]{MSTSegments()}},
\code{\link[=MapTrees]{MapTrees()}},
\code{\link[=MappingQuality]{MappingQuality()}},
\code{\link[=SpectralEigens]{SpectralEigens()}},
\code{\link{cluster-statistics}},
\code{\link[=median.multiPhylo]{median.multiPhylo()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree space functions}
